#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
"""
Tools to help interfacing with the Game Boy hardware
"""

__pdoc__ = {
    "constants": False,
    "manager": False,
}
__all__ = ["BotSupportManager"]

from .manager import BotSupportManager
