#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

# from .base_mbc import BaseMBC, ROMOnly
from .cartridge import load_cartridge

# from .mbc1 import MBC1
# from .mbc2 import MBC2
# from .mbc3 import MBC3
# from .mbc5 import MBC5
