#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
"""
Plugins that extend PyBoy's functionality. The only publicly exposed, are the game wrappers.
"""

__pdoc__ = {
    # docs exclude
    "window_headless": False,
    "window_open_gl": False,
    "screen_recorder": False,
    "rewind": False,
    "window_dummy": False,
    "disable_input": False,
    "manager_gen": False,
    "auto_pause": False,
    "manager": False,
    "record_replay": False,
    "screenshot_recorder": False,
    "debug": False,
    "window_sdl2": False,
    # docs exclude end
}
