/* Generated by Cython 3.0.8 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "extra_compile_args": [
            "-O3",
            "-DCYTHON_INLINE=inline __attribute__ ((__unused__)) __attribute__((always_inline))"
        ],
        "include_dirs": [
            "/Users/runner/hostedtoolcache/Python/3.8.18/x64/lib/python3.8/site-packages/numpy/core/include"
        ],
        "name": "pyboy.core.opcodes",
        "sources": [
            "pyboy/core/opcodes.py"
        ]
    },
    "module_name": "pyboy.core.opcodes"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_8" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x030008F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyboy__core__opcodes
#define __PYX_HAVE_API__pyboy__core__opcodes
/* Early includes */
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stddef.h>
#include "pythread.h"
#include <time.h>

    #if CYTHON_COMPILING_IN_PYPY
    #ifdef _MSC_VER
    #pragma message ("This module uses CPython specific internals of 'array.array', which are not available in PyPy.")
    #else
    #warning This module uses CPython specific internals of 'array.array', which are not available in PyPy.
    #endif
    #endif
    
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const wchar_t *u)
{
    const wchar_t *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#endif
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "pyboy/core/opcodes.py",
  "<stringsource>",
  "contextvars.pxd",
  "time.pxd",
  "array.pxd",
  "pyboy/logging/logging.pxd",
  "pyboy/core/bootrom.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "pyboy/utils.pxd",
  "pyboy/core/cartridge/rtc.pxd",
  "pyboy/core/cartridge/base_mbc.pxd",
  "pyboy/core/interaction.pxd",
  "pyboy/core/lcd.pxd",
  "pyboy/core/ram.pxd",
  "pyboy/core/sound.pxd",
  "pyboy/core/timer.pxd",
  "pyboy/core/mb.pxd",
  "pyboy/core/cpu.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_incr_aligned(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_decr_aligned(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_incr_aligned(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_decr_aligned(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_incr_aligned(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd)
    #define __pyx_atomic_incr_aligned(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_aligned(value) _InterlockedExchangeAdd(value, -1)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
#ifndef _ARRAYARRAY_H
struct arrayobject;
typedef struct arrayobject arrayobject;
#endif
struct __pyx_obj_5pyboy_7logging_7logging_Logger;
struct __pyx_obj_5pyboy_4core_7bootrom_BootROM;
struct __pyx_obj_5pyboy_5utils_IntIOInterface;
struct __pyx_obj_5pyboy_5utils_IntIOWrapper;
struct __pyx_obj_5pyboy_5utils_WindowEvent;
struct __pyx_obj_5pyboy_5utils_WindowEventMouse;
struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC;
struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC;
struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_ROMOnly;
struct __pyx_obj_5pyboy_4core_11interaction_Interaction;
struct __pyx_obj_5pyboy_4core_3lcd_LCD;
struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister;
struct __pyx_obj_5pyboy_4core_3lcd_STATRegister;
struct __pyx_obj_5pyboy_4core_3lcd_LCDCRegister;
struct __pyx_obj_5pyboy_4core_3lcd_Renderer;
struct __pyx_obj_5pyboy_4core_3lcd_CGBLCD;
struct __pyx_obj_5pyboy_4core_3lcd_CGBRenderer;
struct __pyx_obj_5pyboy_4core_3lcd_VBKregister;
struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister;
struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister;
struct __pyx_obj_5pyboy_4core_3ram_RAM;
struct __pyx_obj_5pyboy_4core_5sound_Sound;
struct __pyx_obj_5pyboy_4core_5sound_ToneChannel;
struct __pyx_obj_5pyboy_4core_5sound_SweepChannel;
struct __pyx_obj_5pyboy_4core_5sound_WaveChannel;
struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel;
struct __pyx_obj_5pyboy_4core_5timer_Timer;
struct __pyx_obj_5pyboy_4core_2mb_Motherboard;
struct __pyx_obj_5pyboy_4core_2mb_HDMA;
struct __pyx_obj_5pyboy_4core_3cpu_CPU;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};
struct __pyx_ctuple_int__and_int;
typedef struct __pyx_ctuple_int__and_int __pyx_ctuple_int__and_int;
struct __pyx_ctuple_int__and_int__and_int__and_int__and_int;
typedef struct __pyx_ctuple_int__and_int__and_int__and_int__and_int __pyx_ctuple_int__and_int__and_int__and_int__and_int;

/* "pyboy/core/lcd.pxd":63
 *     cdef void load_state(self, IntIOInterface, int) noexcept
 * 
 *     cdef (int, int) getwindowpos(self) noexcept nogil             # <<<<<<<<<<<<<<
 *     cdef (int, int) getviewport(self) noexcept nogil
 * 
 */
struct __pyx_ctuple_int__and_int {
  int f0;
  int f1;
};

/* "pyboy/core/lcd.pxd":236
 *         bg_priority = uint8_t,
 *     )
 *     cdef (int, int, int, int, int) _cgb_get_background_map_attributes(self, LCD, int) noexcept nogil             # <<<<<<<<<<<<<<
 * 
 * cdef class CGBLCD(LCD):
 */
struct __pyx_ctuple_int__and_int__and_int__and_int__and_int {
  int f0;
  int f1;
  int f2;
  int f3;
  int f4;
};

/* "pyboy/logging/logging.pxd":6
 * cdef int CRITICAL, FATAL, ERROR, WARNING, INFO, DEBUG
 * 
 * cdef class Logger:             # <<<<<<<<<<<<<<
 *   cdef str name
 * 
 */
struct __pyx_obj_5pyboy_7logging_7logging_Logger {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_7logging_7logging_Logger *__pyx_vtab;
  PyObject *name;
};


/* "pyboy/core/bootrom.pxd":13
 * cdef Logger logger
 * 
 * cdef class BootROM:             # <<<<<<<<<<<<<<
 *     cdef uint8_t[:] bootrom
 *     cdef uint8_t getitem(self, uint16_t) noexcept nogil
 */
struct __pyx_obj_5pyboy_4core_7bootrom_BootROM {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_7bootrom_BootROM *__pyx_vtab;
  __Pyx_memviewslice bootrom;
};


/* "utils.pxd":12
 * # Buffer classes
 * 
 * cdef class IntIOInterface:             # <<<<<<<<<<<<<<
 *     cdef int64_t write(self, uint8_t) noexcept
 *     cdef int64_t write_16bit(self, uint16_t) noexcept
 */
struct __pyx_obj_5pyboy_5utils_IntIOInterface {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_5utils_IntIOInterface *__pyx_vtab;
};


/* "utils.pxd":28
 * 
 * 
 * cdef class IntIOWrapper(IntIOInterface):             # <<<<<<<<<<<<<<
 *     cdef object buffer
 * 
 */
struct __pyx_obj_5pyboy_5utils_IntIOWrapper {
  struct __pyx_obj_5pyboy_5utils_IntIOInterface __pyx_base;
  PyObject *buffer;
};


/* "utils.pxd":40
 * # Temporarily placed here to not be exposed on public API
 * 
 * cdef class WindowEvent:             # <<<<<<<<<<<<<<
 *     cdef public int event
 * 
 */
struct __pyx_obj_5pyboy_5utils_WindowEvent {
  PyObject_HEAD
  int event;
};


/* "utils.pxd":43
 *     cdef public int event
 * 
 * cdef class WindowEventMouse(WindowEvent):             # <<<<<<<<<<<<<<
 *     cdef public int window_id
 *     cdef public int mouse_x
 */
struct __pyx_obj_5pyboy_5utils_WindowEventMouse {
  struct __pyx_obj_5pyboy_5utils_WindowEvent __pyx_base;
  int window_id;
  int mouse_x;
  int mouse_y;
  int mouse_scroll_x;
  int mouse_scroll_y;
  int mouse_button;
};


/* "pyboy/core/cartridge/rtc.pxd":18
 * cdef Logger logger
 * 
 * cdef class RTC:             # <<<<<<<<<<<<<<
 *     cdef str filename
 *     cdef bint latch_enabled
 */
struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_9cartridge_3rtc_RTC *__pyx_vtab;
  PyObject *filename;
  int latch_enabled;
  double timezero;
  uint64_t sec_latch;
  uint64_t min_latch;
  uint64_t hour_latch;
  uint64_t day_latch_low;
  uint64_t day_latch_high;
  uint64_t day_carry;
  uint64_t halt;
};


/* "pyboy/core/cartridge/base_mbc.pxd":15
 * cdef Logger logger
 * 
 * cdef class BaseMBC:             # <<<<<<<<<<<<<<
 *     cdef str filename
 *     cdef str gamename
 */
struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *__pyx_vtab;
  PyObject *filename;
  PyObject *gamename;
  __Pyx_memviewslice rombanks;
  __Pyx_memviewslice rambanks;
  uint8_t carttype;
  int battery;
  int rtc_enabled;
  struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC *rtc;
  uint8_t memorymodel;
  int rambank_enabled;
  int external_ram_count;
  int external_rom_count;
  int rambank_initialized;
  uint16_t rambank_selected;
  uint16_t rombank_selected;
  int cgb;
};


/* "pyboy/core/cartridge/base_mbc.pxd":44
 *     cdef void overrideitem(self, int, uint16_t, uint8_t) noexcept nogil
 * 
 * cdef class ROMOnly(BaseMBC):             # <<<<<<<<<<<<<<
 *     cdef void setitem(self, uint16_t, uint8_t) noexcept nogil
 */
struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_ROMOnly {
  struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC __pyx_base;
};


/* "pyboy/core/interaction.pxd":19
 * 
 * 
 * cdef class Interaction:             # <<<<<<<<<<<<<<
 *     cdef uint8_t directional, standard
 *     cdef bint key_event(self, WindowEvent) noexcept
 */
struct __pyx_obj_5pyboy_4core_11interaction_Interaction {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_11interaction_Interaction *__pyx_vtab;
  uint8_t directional;
  uint8_t standard;
};


/* "pyboy/core/lcd.pxd":26
 * cdef Logger logger
 * 
 * cdef class LCD:             # <<<<<<<<<<<<<<
 *     cdef bint disable_renderer
 *     cdef uint8_t[8 * 1024] VRAM0
 */
struct __pyx_obj_5pyboy_4core_3lcd_LCD {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_LCD *__pyx_vtab;
  int disable_renderer;
  uint8_t VRAM0[(8 * 0x400)];
  uint8_t OAM[0xA0];
  uint8_t SCY;
  uint8_t SCX;
  uint8_t WY;
  uint8_t WX;
  uint8_t next_stat_mode;
  int frame_done;
  uint8_t max_ly;
  uint8_t LY;
  uint8_t LYC;
  uint64_t clock;
  uint64_t clock_target;
  struct __pyx_obj_5pyboy_4core_3lcd_LCDCRegister *_LCDC;
  struct __pyx_obj_5pyboy_4core_3lcd_STATRegister *_STAT;
  struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister *BGP;
  struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister *OBP0;
  struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister *OBP1;
  struct __pyx_obj_5pyboy_4core_3lcd_Renderer *renderer;
  int cgb;
  int double_speed;
  uint8_t VRAM1[(8 * 0x400)];
  struct __pyx_obj_5pyboy_4core_3lcd_VBKregister *vbk;
  struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister *bcps;
  struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister *bcpd;
  struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister *ocps;
  struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister *ocpd;
};


/* "pyboy/core/lcd.pxd":77
 * 
 * 
 * cdef class PaletteRegister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t value
 *     cdef uint32_t[4] lookup
 */
struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteRegister *__pyx_vtab;
  uint8_t value;
  uint32_t lookup[4];
  uint32_t palette_mem_rgb[4];
};


/* "pyboy/core/lcd.pxd":87
 *     cdef inline uint32_t getcolor(self, uint8_t) noexcept nogil
 * 
 * cdef class STATRegister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t value
 *     cdef uint8_t _mode
 */
struct __pyx_obj_5pyboy_4core_3lcd_STATRegister {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_STATRegister *__pyx_vtab;
  uint8_t value;
  uint8_t _mode;
};


/* "pyboy/core/lcd.pxd":94
 *     cdef void set(self, uint64_t) noexcept nogil
 * 
 * cdef class LCDCRegister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t value
 * 
 */
struct __pyx_obj_5pyboy_4core_3lcd_LCDCRegister {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_LCDCRegister *__pyx_vtab;
  uint8_t value;
  int lcd_enable;
  int windowmap_select;
  int window_enable;
  int tiledata_select;
  int backgroundmap_select;
  int sprite_height;
  int sprite_enable;
  int background_enable;
  int cgb_master_priority;
};


/* "pyboy/core/lcd.pxd":110
 * 
 * 
 * cdef class Renderer:             # <<<<<<<<<<<<<<
 *     cdef uint8_t[:] _tilecache0_state, _tilecache1_state, _spritecache0_state, _spritecache1_state
 *     cdef str color_format
 */
struct __pyx_obj_5pyboy_4core_3lcd_Renderer {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_Renderer *__pyx_vtab;
  __Pyx_memviewslice _tilecache0_state;
  __Pyx_memviewslice _tilecache1_state;
  __Pyx_memviewslice _spritecache0_state;
  __Pyx_memviewslice _spritecache1_state;
  PyObject *color_format;
  PyObject *buffer_dims;
  int clearcache;
  int double_speed;
  int cgb;
  arrayobject *_screenbuffer_raw;
  PyObject *_screenbuffer_ptr;
  arrayobject *_tilecache0_raw;
  arrayobject *_spritecache0_raw;
  arrayobject *_spritecache1_raw;
  __Pyx_memviewslice _screenbuffer;
  __Pyx_memviewslice _tilecache0;
  __Pyx_memviewslice _spritecache0;
  __Pyx_memviewslice _spritecache1;
  int sprites_to_render[10];
  int ly_window;
  int _scanlineparameters[0x90][5];
  arrayobject *_tilecache1_raw;
  __Pyx_memviewslice _tilecache1;
};


/* "pyboy/core/lcd.pxd":238
 *     cdef (int, int, int, int, int) _cgb_get_background_map_attributes(self, LCD, int) noexcept nogil
 * 
 * cdef class CGBLCD(LCD):             # <<<<<<<<<<<<<<
 *     pass
 *     # cdef uint8_t[8 * 1024] VRAM1
 */
struct __pyx_obj_5pyboy_4core_3lcd_CGBLCD {
  struct __pyx_obj_5pyboy_4core_3lcd_LCD __pyx_base;
};


/* "pyboy/core/lcd.pxd":247
 *     # cdef PaletteColorRegister ocpd
 * 
 * cdef class CGBRenderer(Renderer):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5pyboy_4core_3lcd_CGBRenderer {
  struct __pyx_obj_5pyboy_4core_3lcd_Renderer __pyx_base;
};


/* "pyboy/core/lcd.pxd":250
 *     pass
 * 
 * cdef class VBKregister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t active_bank
 * 
 */
struct __pyx_obj_5pyboy_4core_3lcd_VBKregister {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_VBKregister *__pyx_vtab;
  uint8_t active_bank;
};


/* "pyboy/core/lcd.pxd":256
 *     cdef uint8_t get(self) noexcept nogil
 * 
 * cdef class PaletteIndexRegister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t value
 *     cdef int auto_inc
 */
struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteIndexRegister *__pyx_vtab;
  uint8_t value;
  int auto_inc;
  int index;
  int hl;
};


/* "pyboy/core/lcd.pxd":270
 *     cdef void load_state(self, IntIOInterface, int) noexcept
 * 
 * cdef class PaletteColorRegister:             # <<<<<<<<<<<<<<
 *     cdef uint16_t[8 * 4] palette_mem
 *     cdef uint32_t[8 * 4] palette_mem_rgb
 */
struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteColorRegister *__pyx_vtab;
  uint16_t palette_mem[(8 * 4)];
  uint32_t palette_mem_rgb[(8 * 4)];
  struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister *index_reg;
};


/* "pyboy/core/ram.pxd":15
 * cdef Logger logger
 * 
 * cdef class RAM:             # <<<<<<<<<<<<<<
 *     @cython.locals(n=int)
 *     cdef void save_state(self, IntIOInterface) noexcept
 */
struct __pyx_obj_5pyboy_4core_3ram_RAM {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3ram_RAM *__pyx_vtab;
  __Pyx_memviewslice internal_ram0;
  uint8_t non_io_internal_ram0[0x60];
  uint8_t io_ports[0x4C];
  uint8_t internal_ram1[0x7F];
  uint8_t non_io_internal_ram1[0x34];
  int cgb;
};


/* "pyboy/core/sound.pxd":18
 * cdef int CPU_FREQ
 * 
 * cdef class Sound:             # <<<<<<<<<<<<<<
 *     cdef void save_state(self, IntIOInterface) noexcept
 *     cdef void load_state(self, IntIOInterface, int) noexcept
 */
struct __pyx_obj_5pyboy_4core_5sound_Sound {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_5sound_Sound *__pyx_vtab;
  int enabled;
  int emulate;
  int device;
  int sample_rate;
  int sampleclocks;
  PyObject *audiobuffer;
  PyObject *audiobuffer_p;
  int clock;
  int poweron;
  struct __pyx_obj_5pyboy_4core_5sound_SweepChannel *sweepchannel;
  struct __pyx_obj_5pyboy_4core_5sound_ToneChannel *tonechannel;
  struct __pyx_obj_5pyboy_4core_5sound_WaveChannel *wavechannel;
  struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel *noisechannel;
  int leftnoise;
  int leftwave;
  int lefttone;
  int leftsweep;
  int rightnoise;
  int rightwave;
  int righttone;
  int rightsweep;
};


/* "pyboy/core/sound.pxd":52
 * 
 * 
 * cdef class ToneChannel:             # <<<<<<<<<<<<<<
 *     cdef uint8_t[4][8] wavetables
 * 
 */
struct __pyx_obj_5pyboy_4core_5sound_ToneChannel {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_5sound_ToneChannel *__pyx_vtab;
  uint8_t wavetables[4][8];
  uint8_t wavsel;
  uint8_t sndlen;
  uint8_t envini;
  uint8_t envdir;
  uint8_t envper;
  uint16_t sndper;
  uint8_t uselen;
  int enable;
  int lengthtimer;
  int periodtimer;
  int envelopetimer;
  int period;
  int waveframe;
  int frametimer;
  int frame;
  int volume;
};


/* "pyboy/core/sound.pxd":83
 * 
 * 
 * cdef class SweepChannel(ToneChannel):             # <<<<<<<<<<<<<<
 *     # Register Values
 *     cdef uint8_t swpper # Register 0 bits 6-4: Sweep period
 */
struct __pyx_obj_5pyboy_4core_5sound_SweepChannel {
  struct __pyx_obj_5pyboy_4core_5sound_ToneChannel __pyx_base;
  uint8_t swpper;
  uint8_t swpdir;
  uint8_t swpmag;
  int sweeptimer;
  int sweepenable;
  int shadow;
};


/* "pyboy/core/sound.pxd":101
 * 
 * 
 * cdef class WaveChannel:             # <<<<<<<<<<<<<<
 *     # Memory for wave sample
 *     cdef uint8_t[16] wavetable
 */
struct __pyx_obj_5pyboy_4core_5sound_WaveChannel {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_5sound_WaveChannel *__pyx_vtab;
  uint8_t wavetable[16];
  uint8_t dacpow;
  uint8_t sndlen;
  uint8_t volreg;
  uint16_t sndper;
  uint8_t uselen;
  int enable;
  int lengthtimer;
  int periodtimer;
  int period;
  int waveframe;
  int frametimer;
  int frame;
  int volumeshift;
};


/* "pyboy/core/sound.pxd":134
 * 
 * 
 * cdef class NoiseChannel:             # <<<<<<<<<<<<<<
 *     cdef uint8_t[8] DIVTABLE
 * 
 */
struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_5sound_NoiseChannel *__pyx_vtab;
  uint8_t DIVTABLE[8];
  uint8_t sndlen;
  uint8_t envini;
  uint8_t envdir;
  uint8_t envper;
  uint8_t clkpow;
  uint8_t regwid;
  uint8_t clkdiv;
  uint8_t uselen;
  int enable;
  int lengthtimer;
  int periodtimer;
  int envelopetimer;
  int period;
  int shiftregister;
  int lfsrfeed;
  int frametimer;
  int frame;
  int volume;
};


/* "pyboy/core/timer.pxd":17
 * cdef Logger logger
 * 
 * cdef class Timer:             # <<<<<<<<<<<<<<
 *     cdef uint64_t DIV, TIMA, TMA, TAC
 *     cdef uint16_t DIV_counter, TIMA_counter
 */
struct __pyx_obj_5pyboy_4core_5timer_Timer {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_5timer_Timer *__pyx_vtab;
  uint64_t DIV;
  uint64_t TIMA;
  uint64_t TMA;
  uint64_t TAC;
  uint16_t DIV_counter;
  uint16_t TIMA_counter;
  uint64_t dividers[4];
};


/* "pyboy/core/mb.pxd":30
 * 
 * 
 * cdef class Motherboard:             # <<<<<<<<<<<<<<
 *     cdef pyboy.core.interaction.Interaction interaction
 *     cdef pyboy.core.bootrom.BootROM bootrom
 */
struct __pyx_obj_5pyboy_4core_2mb_Motherboard {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *__pyx_vtab;
  struct __pyx_obj_5pyboy_4core_11interaction_Interaction *interaction;
  struct __pyx_obj_5pyboy_4core_7bootrom_BootROM *bootrom;
  struct __pyx_obj_5pyboy_4core_3ram_RAM *ram;
  struct __pyx_obj_5pyboy_4core_3lcd_LCD *lcd;
  struct __pyx_obj_5pyboy_4core_3cpu_CPU *cpu;
  struct __pyx_obj_5pyboy_4core_5timer_Timer *timer;
  struct __pyx_obj_5pyboy_4core_5sound_Sound *sound;
  struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *cartridge;
  int bootrom_enabled;
  char serialbuffer[0x400];
  uint16_t serialbuffer_count;
  struct __pyx_obj_5pyboy_4core_2mb_HDMA *hdma;
  uint8_t key1;
  int double_speed;
  int cgb;
  int breakpoints_enabled;
  PyObject *breakpoints_list;
  int breakpoint_latch;
};


/* "pyboy/core/mb.pxd":73
 *     cdef void load_state(self, IntIOInterface) noexcept
 * 
 * cdef class HDMA:             # <<<<<<<<<<<<<<
 *     cdef uint8_t hdma1
 *     cdef uint8_t hdma2
 */
struct __pyx_obj_5pyboy_4core_2mb_HDMA {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_2mb_HDMA *__pyx_vtab;
  uint8_t hdma1;
  uint8_t hdma2;
  uint8_t hdma3;
  uint8_t hdma4;
  uint8_t hdma5;
  uint8_t _hdma5;
  int transfer_active;
  uint16_t curr_src;
  uint16_t curr_dst;
};


/* "cpu.pxd":27
 * 
 * 
 * cdef class CPU:             # <<<<<<<<<<<<<<
 *     cdef bint is_stuck
 *     cdef bint interrupt_master_enable, interrupt_queued, halted, stopped
 */
struct __pyx_obj_5pyboy_4core_3cpu_CPU {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU *__pyx_vtab;
  int is_stuck;
  int interrupt_master_enable;
  int interrupt_queued;
  int halted;
  int stopped;
  uint8_t interrupts_flag;
  uint8_t interrupts_enabled;
  uint8_t interrupts_flag_register;
  uint8_t interrupts_enabled_register;
  int16_t A;
  int16_t F;
  int16_t B;
  int16_t C;
  int16_t D;
  int16_t E;
  int HL;
  int SP;
  int PC;
  struct __pyx_obj_5pyboy_4core_2mb_Motherboard *mb;
};


/* "View.MemoryView":114
 * @cython.collection_type("sequence")
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":302
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":337
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview:             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int_type acquisition_count;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":952
 * @cython.collection_type("sequence")
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "pyboy/logging/logging.pxd":6
 * cdef int CRITICAL, FATAL, ERROR, WARNING, INFO, DEBUG
 * 
 * cdef class Logger:             # <<<<<<<<<<<<<<
 *   cdef str name
 * 
 */

struct __pyx_vtabstruct_5pyboy_7logging_7logging_Logger {
  void (*critical)(struct __pyx_obj_5pyboy_7logging_7logging_Logger *, PyObject *, int __pyx_skip_dispatch, ...);
  void (*error)(struct __pyx_obj_5pyboy_7logging_7logging_Logger *, PyObject *, int __pyx_skip_dispatch, ...);
  void (*warning)(struct __pyx_obj_5pyboy_7logging_7logging_Logger *, PyObject *, int __pyx_skip_dispatch, ...);
  void (*info)(struct __pyx_obj_5pyboy_7logging_7logging_Logger *, PyObject *, int __pyx_skip_dispatch, ...);
  void (*debug)(struct __pyx_obj_5pyboy_7logging_7logging_Logger *, PyObject *, int __pyx_skip_dispatch, ...);
};
static struct __pyx_vtabstruct_5pyboy_7logging_7logging_Logger *__pyx_vtabptr_5pyboy_7logging_7logging_Logger;


/* "pyboy/core/bootrom.pxd":13
 * cdef Logger logger
 * 
 * cdef class BootROM:             # <<<<<<<<<<<<<<
 *     cdef uint8_t[:] bootrom
 *     cdef uint8_t getitem(self, uint16_t) noexcept nogil
 */

struct __pyx_vtabstruct_5pyboy_4core_7bootrom_BootROM {
  uint8_t (*getitem)(struct __pyx_obj_5pyboy_4core_7bootrom_BootROM *, uint16_t);
};
static struct __pyx_vtabstruct_5pyboy_4core_7bootrom_BootROM *__pyx_vtabptr_5pyboy_4core_7bootrom_BootROM;


/* "utils.pxd":12
 * # Buffer classes
 * 
 * cdef class IntIOInterface:             # <<<<<<<<<<<<<<
 *     cdef int64_t write(self, uint8_t) noexcept
 *     cdef int64_t write_16bit(self, uint16_t) noexcept
 */

struct __pyx_vtabstruct_5pyboy_5utils_IntIOInterface {
  int64_t (*write)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *, uint8_t);
  int64_t (*write_16bit)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *, uint16_t);
  int64_t (*write_32bit)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *, uint32_t);
  int64_t (*write_64bit)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *, uint64_t);
  uint8_t (*read)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  uint16_t (*read_16bit)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  uint32_t (*read_32bit)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  uint64_t (*read_64bit)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*seek)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int64_t);
  void (*flush)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  uint64_t (*tell)(struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
};
static struct __pyx_vtabstruct_5pyboy_5utils_IntIOInterface *__pyx_vtabptr_5pyboy_5utils_IntIOInterface;


/* "utils.pxd":28
 * 
 * 
 * cdef class IntIOWrapper(IntIOInterface):             # <<<<<<<<<<<<<<
 *     cdef object buffer
 * 
 */

struct __pyx_vtabstruct_5pyboy_5utils_IntIOWrapper {
  struct __pyx_vtabstruct_5pyboy_5utils_IntIOInterface __pyx_base;
};
static struct __pyx_vtabstruct_5pyboy_5utils_IntIOWrapper *__pyx_vtabptr_5pyboy_5utils_IntIOWrapper;


/* "pyboy/core/cartridge/rtc.pxd":18
 * cdef Logger logger
 * 
 * cdef class RTC:             # <<<<<<<<<<<<<<
 *     cdef str filename
 *     cdef bint latch_enabled
 */

struct __pyx_vtabstruct_5pyboy_4core_9cartridge_3rtc_RTC {
  void (*stop)(struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC *);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
  void (*latch_rtc)(struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC *);
  void (*writecommand)(struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC *, uint8_t);
  uint8_t (*getregister)(struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC *, uint8_t);
  void (*setregister)(struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC *, uint8_t, uint8_t);
};
static struct __pyx_vtabstruct_5pyboy_4core_9cartridge_3rtc_RTC *__pyx_vtabptr_5pyboy_4core_9cartridge_3rtc_RTC;


/* "pyboy/core/cartridge/base_mbc.pxd":15
 * cdef Logger logger
 * 
 * cdef class BaseMBC:             # <<<<<<<<<<<<<<
 *     cdef str filename
 *     cdef str gamename
 */

struct __pyx_vtabstruct_5pyboy_4core_9cartridge_8base_mbc_BaseMBC {
  void (*save_state)(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
  void (*save_ram)(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_ram)(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*init_rambanks)(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *, uint8_t);
  PyObject *(*getgamename)(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *, __Pyx_memviewslice);
  uint8_t (*getitem)(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *, uint16_t);
  void (*setitem)(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *, uint16_t, uint8_t);
  void (*overrideitem)(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *, int, uint16_t, uint8_t);
};
static struct __pyx_vtabstruct_5pyboy_4core_9cartridge_8base_mbc_BaseMBC *__pyx_vtabptr_5pyboy_4core_9cartridge_8base_mbc_BaseMBC;


/* "pyboy/core/cartridge/base_mbc.pxd":44
 *     cdef void overrideitem(self, int, uint16_t, uint8_t) noexcept nogil
 * 
 * cdef class ROMOnly(BaseMBC):             # <<<<<<<<<<<<<<
 *     cdef void setitem(self, uint16_t, uint8_t) noexcept nogil
 */

struct __pyx_vtabstruct_5pyboy_4core_9cartridge_8base_mbc_ROMOnly {
  struct __pyx_vtabstruct_5pyboy_4core_9cartridge_8base_mbc_BaseMBC __pyx_base;
};
static struct __pyx_vtabstruct_5pyboy_4core_9cartridge_8base_mbc_ROMOnly *__pyx_vtabptr_5pyboy_4core_9cartridge_8base_mbc_ROMOnly;


/* "pyboy/core/interaction.pxd":19
 * 
 * 
 * cdef class Interaction:             # <<<<<<<<<<<<<<
 *     cdef uint8_t directional, standard
 *     cdef bint key_event(self, WindowEvent) noexcept
 */

struct __pyx_vtabstruct_5pyboy_4core_11interaction_Interaction {
  int (*key_event)(struct __pyx_obj_5pyboy_4core_11interaction_Interaction *, struct __pyx_obj_5pyboy_5utils_WindowEvent *);
  uint8_t (*pull)(struct __pyx_obj_5pyboy_4core_11interaction_Interaction *, uint8_t);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_11interaction_Interaction *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_11interaction_Interaction *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
};
static struct __pyx_vtabstruct_5pyboy_4core_11interaction_Interaction *__pyx_vtabptr_5pyboy_4core_11interaction_Interaction;


/* "pyboy/core/lcd.pxd":26
 * cdef Logger logger
 * 
 * cdef class LCD:             # <<<<<<<<<<<<<<
 *     cdef bint disable_renderer
 *     cdef uint8_t[8 * 1024] VRAM0
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_LCD {
  uint8_t (*tick)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *, int);
  uint64_t (*cycles_to_interrupt)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *);
  void (*set_lcdc)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *, uint8_t);
  uint8_t (*get_lcdc)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *);
  void (*set_stat)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *, uint8_t);
  uint8_t (*get_stat)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *);
  int (*cycles_to_mode0)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
  __pyx_ctuple_int__and_int (*getwindowpos)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *);
  __pyx_ctuple_int__and_int (*getviewport)(struct __pyx_obj_5pyboy_4core_3lcd_LCD *);
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_LCD *__pyx_vtabptr_5pyboy_4core_3lcd_LCD;


/* "pyboy/core/lcd.pxd":77
 * 
 * 
 * cdef class PaletteRegister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t value
 *     cdef uint32_t[4] lookup
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteRegister {
  int (*set)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister *, uint64_t);
  uint8_t (*get)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister *);
  uint32_t (*getcolor)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister *, uint8_t);
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteRegister *__pyx_vtabptr_5pyboy_4core_3lcd_PaletteRegister;


/* "pyboy/core/lcd.pxd":87
 *     cdef inline uint32_t getcolor(self, uint8_t) noexcept nogil
 * 
 * cdef class STATRegister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t value
 *     cdef uint8_t _mode
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_STATRegister {
  uint8_t (*set_mode)(struct __pyx_obj_5pyboy_4core_3lcd_STATRegister *, uint8_t);
  uint8_t (*update_LYC)(struct __pyx_obj_5pyboy_4core_3lcd_STATRegister *, uint8_t, uint8_t);
  void (*set)(struct __pyx_obj_5pyboy_4core_3lcd_STATRegister *, uint64_t);
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_STATRegister *__pyx_vtabptr_5pyboy_4core_3lcd_STATRegister;


/* "pyboy/core/lcd.pxd":94
 *     cdef void set(self, uint64_t) noexcept nogil
 * 
 * cdef class LCDCRegister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t value
 * 
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_LCDCRegister {
  void (*set)(struct __pyx_obj_5pyboy_4core_3lcd_LCDCRegister *, uint64_t);
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_LCDCRegister *__pyx_vtabptr_5pyboy_4core_3lcd_LCDCRegister;


/* "pyboy/core/lcd.pxd":110
 * 
 * 
 * cdef class Renderer:             # <<<<<<<<<<<<<<
 *     cdef uint8_t[:] _tilecache0_state, _tilecache1_state, _spritecache0_state, _spritecache1_state
 *     cdef str color_format
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_Renderer {
  void (*invalidate_tile)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, int, int);
  void (*blank_screen)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_4core_3lcd_LCD *);
  void (*scanline)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_4core_3lcd_LCD *, int);
  void (*scanline_sprites)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_4core_3lcd_LCD *, int, __Pyx_memviewslice, int);
  void (*sort_sprites)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, int);
  void (*clear_cache)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *);
  void (*clear_tilecache0)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *);
  void (*clear_tilecache1)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *);
  void (*clear_spritecache0)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *);
  void (*clear_spritecache1)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *);
  void (*update_tilecache0)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_4core_3lcd_LCD *, int, int);
  void (*update_tilecache1)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_4core_3lcd_LCD *, int, int);
  void (*update_spritecache0)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_4core_3lcd_LCD *, int, int);
  void (*update_spritecache1)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_4core_3lcd_LCD *, int, int);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
  __pyx_ctuple_int__and_int__and_int__and_int__and_int (*_cgb_get_background_map_attributes)(struct __pyx_obj_5pyboy_4core_3lcd_Renderer *, struct __pyx_obj_5pyboy_4core_3lcd_LCD *, int);
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_Renderer *__pyx_vtabptr_5pyboy_4core_3lcd_Renderer;


/* "pyboy/core/lcd.pxd":238
 *     cdef (int, int, int, int, int) _cgb_get_background_map_attributes(self, LCD, int) noexcept nogil
 * 
 * cdef class CGBLCD(LCD):             # <<<<<<<<<<<<<<
 *     pass
 *     # cdef uint8_t[8 * 1024] VRAM1
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_CGBLCD {
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_LCD __pyx_base;
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_CGBLCD *__pyx_vtabptr_5pyboy_4core_3lcd_CGBLCD;


/* "pyboy/core/lcd.pxd":247
 *     # cdef PaletteColorRegister ocpd
 * 
 * cdef class CGBRenderer(Renderer):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_CGBRenderer {
  struct __pyx_vtabstruct_5pyboy_4core_3lcd_Renderer __pyx_base;
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_CGBRenderer *__pyx_vtabptr_5pyboy_4core_3lcd_CGBRenderer;


/* "pyboy/core/lcd.pxd":250
 *     pass
 * 
 * cdef class VBKregister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t active_bank
 * 
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_VBKregister {
  void (*set)(struct __pyx_obj_5pyboy_4core_3lcd_VBKregister *, uint8_t);
  uint8_t (*get)(struct __pyx_obj_5pyboy_4core_3lcd_VBKregister *);
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_VBKregister *__pyx_vtabptr_5pyboy_4core_3lcd_VBKregister;


/* "pyboy/core/lcd.pxd":256
 *     cdef uint8_t get(self) noexcept nogil
 * 
 * cdef class PaletteIndexRegister:             # <<<<<<<<<<<<<<
 *     cdef uint8_t value
 *     cdef int auto_inc
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteIndexRegister {
  void (*set)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister *, uint8_t);
  uint8_t (*get)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister *);
  int (*getindex)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister *);
  void (*shouldincrement)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister *);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteIndexRegister *__pyx_vtabptr_5pyboy_4core_3lcd_PaletteIndexRegister;


/* "pyboy/core/lcd.pxd":270
 *     cdef void load_state(self, IntIOInterface, int) noexcept
 * 
 * cdef class PaletteColorRegister:             # <<<<<<<<<<<<<<
 *     cdef uint16_t[8 * 4] palette_mem
 *     cdef uint32_t[8 * 4] palette_mem_rgb
 */

struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteColorRegister {
  uint32_t (*cgb_to_rgb)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister *, uint16_t, uint8_t);
  void (*set)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister *, uint16_t);
  uint16_t (*get)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister *);
  uint32_t (*getcolor)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister *, uint8_t, uint8_t);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
};
static struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteColorRegister *__pyx_vtabptr_5pyboy_4core_3lcd_PaletteColorRegister;


/* "pyboy/core/ram.pxd":15
 * cdef Logger logger
 * 
 * cdef class RAM:             # <<<<<<<<<<<<<<
 *     @cython.locals(n=int)
 *     cdef void save_state(self, IntIOInterface) noexcept
 */

struct __pyx_vtabstruct_5pyboy_4core_3ram_RAM {
  void (*save_state)(struct __pyx_obj_5pyboy_4core_3ram_RAM *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_3ram_RAM *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
};
static struct __pyx_vtabstruct_5pyboy_4core_3ram_RAM *__pyx_vtabptr_5pyboy_4core_3ram_RAM;


/* "pyboy/core/sound.pxd":18
 * cdef int CPU_FREQ
 * 
 * cdef class Sound:             # <<<<<<<<<<<<<<
 *     cdef void save_state(self, IntIOInterface) noexcept
 *     cdef void load_state(self, IntIOInterface, int) noexcept
 */

struct __pyx_vtabstruct_5pyboy_4core_5sound_Sound {
  void (*save_state)(struct __pyx_obj_5pyboy_4core_5sound_Sound *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_5sound_Sound *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
  uint8_t (*get)(struct __pyx_obj_5pyboy_4core_5sound_Sound *, uint8_t);
  void (*set)(struct __pyx_obj_5pyboy_4core_5sound_Sound *, uint8_t, uint8_t);
  void (*sync)(struct __pyx_obj_5pyboy_4core_5sound_Sound *);
  void (*enqueue_sound)(struct __pyx_obj_5pyboy_4core_5sound_Sound *, int);
  void (*stop)(struct __pyx_obj_5pyboy_4core_5sound_Sound *);
};
static struct __pyx_vtabstruct_5pyboy_4core_5sound_Sound *__pyx_vtabptr_5pyboy_4core_5sound_Sound;


/* "pyboy/core/sound.pxd":52
 * 
 * 
 * cdef class ToneChannel:             # <<<<<<<<<<<<<<
 *     cdef uint8_t[4][8] wavetables
 * 
 */

struct __pyx_vtabstruct_5pyboy_4core_5sound_ToneChannel {
  uint8_t (*getreg)(struct __pyx_obj_5pyboy_4core_5sound_ToneChannel *, uint8_t);
  void (*setreg)(struct __pyx_obj_5pyboy_4core_5sound_ToneChannel *, uint8_t, uint8_t);
  void (*run)(struct __pyx_obj_5pyboy_4core_5sound_ToneChannel *, uint64_t);
  uint8_t (*sample)(struct __pyx_obj_5pyboy_4core_5sound_ToneChannel *);
  void (*trigger)(struct __pyx_obj_5pyboy_4core_5sound_ToneChannel *);
  void (*tickframe)(struct __pyx_obj_5pyboy_4core_5sound_ToneChannel *);
};
static struct __pyx_vtabstruct_5pyboy_4core_5sound_ToneChannel *__pyx_vtabptr_5pyboy_4core_5sound_ToneChannel;


/* "pyboy/core/sound.pxd":83
 * 
 * 
 * cdef class SweepChannel(ToneChannel):             # <<<<<<<<<<<<<<
 *     # Register Values
 *     cdef uint8_t swpper # Register 0 bits 6-4: Sweep period
 */

struct __pyx_vtabstruct_5pyboy_4core_5sound_SweepChannel {
  struct __pyx_vtabstruct_5pyboy_4core_5sound_ToneChannel __pyx_base;
  int (*sweep)(struct __pyx_obj_5pyboy_4core_5sound_SweepChannel *, int);
};
static struct __pyx_vtabstruct_5pyboy_4core_5sound_SweepChannel *__pyx_vtabptr_5pyboy_4core_5sound_SweepChannel;


/* "pyboy/core/sound.pxd":101
 * 
 * 
 * cdef class WaveChannel:             # <<<<<<<<<<<<<<
 *     # Memory for wave sample
 *     cdef uint8_t[16] wavetable
 */

struct __pyx_vtabstruct_5pyboy_4core_5sound_WaveChannel {
  uint8_t (*getreg)(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel *, uint8_t);
  void (*setreg)(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel *, uint8_t, uint8_t);
  void (*run)(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel *, uint64_t);
  uint8_t (*sample)(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel *);
  void (*trigger)(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel *);
  void (*tickframe)(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel *);
  uint8_t (*getwavebyte)(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel *, uint8_t);
  void (*setwavebyte)(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel *, uint8_t, uint8_t);
};
static struct __pyx_vtabstruct_5pyboy_4core_5sound_WaveChannel *__pyx_vtabptr_5pyboy_4core_5sound_WaveChannel;


/* "pyboy/core/sound.pxd":134
 * 
 * 
 * cdef class NoiseChannel:             # <<<<<<<<<<<<<<
 *     cdef uint8_t[8] DIVTABLE
 * 
 */

struct __pyx_vtabstruct_5pyboy_4core_5sound_NoiseChannel {
  uint8_t (*getreg)(struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel *, uint8_t);
  void (*setreg)(struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel *, uint8_t, uint8_t);
  void (*run)(struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel *, uint64_t);
  uint8_t (*sample)(struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel *);
  void (*trigger)(struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel *);
  void (*tickframe)(struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel *);
};
static struct __pyx_vtabstruct_5pyboy_4core_5sound_NoiseChannel *__pyx_vtabptr_5pyboy_4core_5sound_NoiseChannel;


/* "pyboy/core/timer.pxd":17
 * cdef Logger logger
 * 
 * cdef class Timer:             # <<<<<<<<<<<<<<
 *     cdef uint64_t DIV, TIMA, TMA, TAC
 *     cdef uint16_t DIV_counter, TIMA_counter
 */

struct __pyx_vtabstruct_5pyboy_4core_5timer_Timer {
  void (*reset)(struct __pyx_obj_5pyboy_4core_5timer_Timer *);
  int (*tick)(struct __pyx_obj_5pyboy_4core_5timer_Timer *, uint64_t);
  uint64_t (*cycles_to_interrupt)(struct __pyx_obj_5pyboy_4core_5timer_Timer *);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_5timer_Timer *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_5timer_Timer *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
};
static struct __pyx_vtabstruct_5pyboy_4core_5timer_Timer *__pyx_vtabptr_5pyboy_4core_5timer_Timer;


/* "pyboy/core/mb.pxd":30
 * 
 * 
 * cdef class Motherboard:             # <<<<<<<<<<<<<<
 *     cdef pyboy.core.interaction.Interaction interaction
 *     cdef pyboy.core.bootrom.BootROM bootrom
 */

struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard {
  int (*processing_frame)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *);
  void (*buttonevent)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *, struct __pyx_obj_5pyboy_5utils_WindowEvent *);
  void (*stop)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *, int);
  int (*tick)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *);
  void (*switch_speed)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *);
  int (*breakpoint_reached)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *);
  uint8_t (*getitem)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *, uint16_t);
  void (*setitem)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *, uint16_t, uint8_t);
  void (*transfer_DMA)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *, uint8_t);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_2mb_Motherboard *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
};
static struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *__pyx_vtabptr_5pyboy_4core_2mb_Motherboard;


/* "pyboy/core/mb.pxd":73
 *     cdef void load_state(self, IntIOInterface) noexcept
 * 
 * cdef class HDMA:             # <<<<<<<<<<<<<<
 *     cdef uint8_t hdma1
 *     cdef uint8_t hdma2
 */

struct __pyx_vtabstruct_5pyboy_4core_2mb_HDMA {
  void (*set_hdma5)(struct __pyx_obj_5pyboy_4core_2mb_HDMA *, uint8_t, struct __pyx_obj_5pyboy_4core_2mb_Motherboard *);
  int (*tick)(struct __pyx_obj_5pyboy_4core_2mb_HDMA *, struct __pyx_obj_5pyboy_4core_2mb_Motherboard *);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_2mb_HDMA *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_2mb_HDMA *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
};
static struct __pyx_vtabstruct_5pyboy_4core_2mb_HDMA *__pyx_vtabptr_5pyboy_4core_2mb_HDMA;


/* "cpu.pxd":27
 * 
 * 
 * cdef class CPU:             # <<<<<<<<<<<<<<
 *     cdef bint is_stuck
 *     cdef bint interrupt_master_enable, interrupt_queued, halted, stopped
 */

struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU {
  int (*check_interrupts)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *);
  void (*set_interruptflag)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int);
  int (*handle_interrupt)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, uint8_t, uint16_t);
  uint8_t (*fetch_and_execute)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *);
  int (*tick)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *);
  void (*save_state)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *);
  void (*load_state)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, struct __pyx_obj_5pyboy_5utils_IntIOInterface *, int);
  void (*set_bc)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, uint16_t);
  void (*set_de)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, uint16_t);
  PyObject *(*dump_state)(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, PyObject *);
};
static struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU *__pyx_vtabptr_5pyboy_4core_3cpu_CPU;


/* "View.MemoryView":114
 * @cython.collection_type("sequence")
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":337
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview:             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
  PyObject *(*_get_base)(struct __pyx_memoryview_obj *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":952
 * @cython.collection_type("sequence")
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

CYTHON_UNUSED static int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* DivInt[Py_ssize_t].proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define __Pyx_UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

CYTHON_UNUSED static int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PySequenceMultiply.proto */
#define __Pyx_PySequence_Multiply_Left(mul, seq)  __Pyx_PySequence_Multiply(seq, mul)
static CYTHON_INLINE PyObject* __Pyx_PySequence_Multiply(PyObject *seq, Py_ssize_t mul);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* RaiseUnboundLocalError.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_8
#define __PYX_HAVE_RT_ImportType_proto_3_0_8
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_8(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_8(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_8 {
   __Pyx_ImportType_CheckSize_Error_3_0_8 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_8 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_8 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_8(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_8 check_size);
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* ArrayAPI.proto */
#ifndef _ARRAYARRAY_H
#define _ARRAYARRAY_H
typedef struct arraydescr {
    int typecode;
    int itemsize;
    PyObject * (*getitem)(struct arrayobject *, Py_ssize_t);
    int (*setitem)(struct arrayobject *, Py_ssize_t, PyObject *);
#if PY_MAJOR_VERSION >= 3
    char *formats;
#endif
} arraydescr;
struct arrayobject {
    PyObject_HEAD
    Py_ssize_t ob_size;
    union {
        char *ob_item;
        float *as_floats;
        double *as_doubles;
        int *as_ints;
        unsigned int *as_uints;
        unsigned char *as_uchars;
        signed char *as_schars;
        char *as_chars;
        unsigned long *as_ulongs;
        long *as_longs;
#if PY_MAJOR_VERSION >= 3
        unsigned long long *as_ulonglongs;
        long long *as_longlongs;
#endif
        short *as_shorts;
        unsigned short *as_ushorts;
        Py_UNICODE *as_pyunicodes;
        void *as_voidptr;
    } data;
    Py_ssize_t allocated;
    struct arraydescr *ob_descr;
    PyObject *weakreflist;
#if PY_MAJOR_VERSION >= 3
        int ob_exports;
#endif
};
#ifndef NO_NEWARRAY_INLINE
static CYTHON_INLINE PyObject * newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr) {
    arrayobject *op;
    size_t nbytes;
    if (size < 0) {
        PyErr_BadInternalCall();
        return NULL;
    }
    nbytes = size * descr->itemsize;
    if (nbytes / descr->itemsize != (size_t)size) {
        return PyErr_NoMemory();
    }
    op = (arrayobject *) type->tp_alloc(type, 0);
    if (op == NULL) {
        return NULL;
    }
    op->ob_descr = descr;
    op->allocated = size;
    op->weakreflist = NULL;
    __Pyx_SET_SIZE(op, size);
    if (size <= 0) {
        op->data.ob_item = NULL;
    }
    else {
        op->data.ob_item = PyMem_NEW(char, nbytes);
        if (op->data.ob_item == NULL) {
            Py_DECREF(op);
            return PyErr_NoMemory();
        }
    }
    return (PyObject *) op;
}
#else
PyObject* newarrayobject(PyTypeObject *type, Py_ssize_t size,
    struct arraydescr *descr);
#endif
static CYTHON_INLINE int resize(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    PyMem_Resize(items, char, (size_t)(n * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = n;
    return 0;
}
static CYTHON_INLINE int resize_smart(arrayobject *self, Py_ssize_t n) {
    void *items = (void*) self->data.ob_item;
    Py_ssize_t newsize;
    if (n < self->allocated && n*4 > self->allocated) {
        __Pyx_SET_SIZE(self, n);
        return 0;
    }
    newsize = n + (n / 2) + 1;
    if (newsize <= n) {
        PyErr_NoMemory();
        return -1;
    }
    PyMem_Resize(items, char, (size_t)(newsize * self->ob_descr->itemsize));
    if (items == NULL) {
        PyErr_NoMemory();
        return -1;
    }
    self->data.ob_item = (char*) items;
    __Pyx_SET_SIZE(self, n);
    self->allocated = newsize;
    return 0;
}
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int_type *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int_type *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (&memview->acquisition_count)
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XCLEAR_MEMVIEW(slice, have_gil) __Pyx_XCLEAR_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XCLEAR_MEMVIEW(__Pyx_memviewslice *, int, int);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_3_0_8(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview__get_base(struct __pyx_memoryview_obj *__pyx_v_self); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice__get_base(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/

/* Module declarations from "cython.view" */

/* Module declarations from "cython.dataclasses" */

/* Module declarations from "cython" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "pyboy.logging.logging" */
static int *__pyx_vp_5pyboy_7logging_7logging_CRITICAL = 0;
#define __pyx_v_5pyboy_7logging_7logging_CRITICAL (*__pyx_vp_5pyboy_7logging_7logging_CRITICAL)
static int *__pyx_vp_5pyboy_7logging_7logging_FATAL = 0;
#define __pyx_v_5pyboy_7logging_7logging_FATAL (*__pyx_vp_5pyboy_7logging_7logging_FATAL)
static int *__pyx_vp_5pyboy_7logging_7logging_ERROR = 0;
#define __pyx_v_5pyboy_7logging_7logging_ERROR (*__pyx_vp_5pyboy_7logging_7logging_ERROR)
static int *__pyx_vp_5pyboy_7logging_7logging_WARNING = 0;
#define __pyx_v_5pyboy_7logging_7logging_WARNING (*__pyx_vp_5pyboy_7logging_7logging_WARNING)
static int *__pyx_vp_5pyboy_7logging_7logging_INFO = 0;
#define __pyx_v_5pyboy_7logging_7logging_INFO (*__pyx_vp_5pyboy_7logging_7logging_INFO)
static int *__pyx_vp_5pyboy_7logging_7logging_DEBUG = 0;
#define __pyx_v_5pyboy_7logging_7logging_DEBUG (*__pyx_vp_5pyboy_7logging_7logging_DEBUG)

/* Module declarations from "pyboy.core.bootrom" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_7bootrom_logger = 0;
#define __pyx_v_5pyboy_4core_7bootrom_logger (*__pyx_vp_5pyboy_4core_7bootrom_logger)

/* Module declarations from "cpython.version" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "libc.time" */

/* Module declarations from "cpython.time" */
static CYTHON_INLINE double __pyx_f_7cpython_4time_time(void); /*proto*/
static CYTHON_INLINE int __pyx_f_7cpython_4time__raise_from_errno(void); /*proto*/

/* Module declarations from "pyboy.utils" */

/* Module declarations from "pyboy" */

/* Module declarations from "pyboy.core.cartridge.rtc" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_9cartridge_3rtc_logger = 0;
#define __pyx_v_5pyboy_4core_9cartridge_3rtc_logger (*__pyx_vp_5pyboy_4core_9cartridge_3rtc_logger)

/* Module declarations from "pyboy.core.cartridge.base_mbc" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_9cartridge_8base_mbc_logger = 0;
#define __pyx_v_5pyboy_4core_9cartridge_8base_mbc_logger (*__pyx_vp_5pyboy_4core_9cartridge_8base_mbc_logger)

/* Module declarations from "pyboy.core.cartridge" */

/* Module declarations from "pyboy.core.interaction" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_11interaction_logger = 0;
#define __pyx_v_5pyboy_4core_11interaction_logger (*__pyx_vp_5pyboy_4core_11interaction_logger)
static uint8_t *__pyx_vp_5pyboy_4core_11interaction_P10 = 0;
#define __pyx_v_5pyboy_4core_11interaction_P10 (*__pyx_vp_5pyboy_4core_11interaction_P10)
static uint8_t *__pyx_vp_5pyboy_4core_11interaction_P11 = 0;
#define __pyx_v_5pyboy_4core_11interaction_P11 (*__pyx_vp_5pyboy_4core_11interaction_P11)
static uint8_t *__pyx_vp_5pyboy_4core_11interaction_P12 = 0;
#define __pyx_v_5pyboy_4core_11interaction_P12 (*__pyx_vp_5pyboy_4core_11interaction_P12)
static uint8_t *__pyx_vp_5pyboy_4core_11interaction_P13 = 0;
#define __pyx_v_5pyboy_4core_11interaction_P13 (*__pyx_vp_5pyboy_4core_11interaction_P13)
static uint8_t *__pyx_vp_5pyboy_4core_11interaction_P14 = 0;
#define __pyx_v_5pyboy_4core_11interaction_P14 (*__pyx_vp_5pyboy_4core_11interaction_P14)
static uint8_t *__pyx_vp_5pyboy_4core_11interaction_P15 = 0;
#define __pyx_v_5pyboy_4core_11interaction_P15 (*__pyx_vp_5pyboy_4core_11interaction_P15)

/* Module declarations from "array" */

/* Module declarations from "cpython.array" */
static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *, char *, Py_ssize_t); /*proto*/

/* Module declarations from "pyboy.core.lcd" */
static uint8_t *__pyx_vp_5pyboy_4core_3lcd_INTR_VBLANK = 0;
#define __pyx_v_5pyboy_4core_3lcd_INTR_VBLANK (*__pyx_vp_5pyboy_4core_3lcd_INTR_VBLANK)
static uint8_t *__pyx_vp_5pyboy_4core_3lcd_INTR_LCDC = 0;
#define __pyx_v_5pyboy_4core_3lcd_INTR_LCDC (*__pyx_vp_5pyboy_4core_3lcd_INTR_LCDC)
static uint8_t *__pyx_vp_5pyboy_4core_3lcd_INTR_TIMER = 0;
#define __pyx_v_5pyboy_4core_3lcd_INTR_TIMER (*__pyx_vp_5pyboy_4core_3lcd_INTR_TIMER)
static uint8_t *__pyx_vp_5pyboy_4core_3lcd_INTR_SERIAL = 0;
#define __pyx_v_5pyboy_4core_3lcd_INTR_SERIAL (*__pyx_vp_5pyboy_4core_3lcd_INTR_SERIAL)
static uint8_t *__pyx_vp_5pyboy_4core_3lcd_INTR_HIGHTOLOW = 0;
#define __pyx_v_5pyboy_4core_3lcd_INTR_HIGHTOLOW (*__pyx_vp_5pyboy_4core_3lcd_INTR_HIGHTOLOW)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_LCDC = 0;
#define __pyx_v_5pyboy_4core_3lcd_LCDC (*__pyx_vp_5pyboy_4core_3lcd_LCDC)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_STAT = 0;
#define __pyx_v_5pyboy_4core_3lcd_STAT (*__pyx_vp_5pyboy_4core_3lcd_STAT)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_SCY = 0;
#define __pyx_v_5pyboy_4core_3lcd_SCY (*__pyx_vp_5pyboy_4core_3lcd_SCY)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_SCX = 0;
#define __pyx_v_5pyboy_4core_3lcd_SCX (*__pyx_vp_5pyboy_4core_3lcd_SCX)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_LY = 0;
#define __pyx_v_5pyboy_4core_3lcd_LY (*__pyx_vp_5pyboy_4core_3lcd_LY)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_LYC = 0;
#define __pyx_v_5pyboy_4core_3lcd_LYC (*__pyx_vp_5pyboy_4core_3lcd_LYC)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_DMA = 0;
#define __pyx_v_5pyboy_4core_3lcd_DMA (*__pyx_vp_5pyboy_4core_3lcd_DMA)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_BGP = 0;
#define __pyx_v_5pyboy_4core_3lcd_BGP (*__pyx_vp_5pyboy_4core_3lcd_BGP)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_OBP0 = 0;
#define __pyx_v_5pyboy_4core_3lcd_OBP0 (*__pyx_vp_5pyboy_4core_3lcd_OBP0)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_OBP1 = 0;
#define __pyx_v_5pyboy_4core_3lcd_OBP1 (*__pyx_vp_5pyboy_4core_3lcd_OBP1)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_WY = 0;
#define __pyx_v_5pyboy_4core_3lcd_WY (*__pyx_vp_5pyboy_4core_3lcd_WY)
static uint16_t *__pyx_vp_5pyboy_4core_3lcd_WX = 0;
#define __pyx_v_5pyboy_4core_3lcd_WX (*__pyx_vp_5pyboy_4core_3lcd_WX)
static int *__pyx_vp_5pyboy_4core_3lcd_ROWS = 0;
#define __pyx_v_5pyboy_4core_3lcd_ROWS (*__pyx_vp_5pyboy_4core_3lcd_ROWS)
static int *__pyx_vp_5pyboy_4core_3lcd_COLS = 0;
#define __pyx_v_5pyboy_4core_3lcd_COLS (*__pyx_vp_5pyboy_4core_3lcd_COLS)
static int *__pyx_vp_5pyboy_4core_3lcd_TILES = 0;
#define __pyx_v_5pyboy_4core_3lcd_TILES (*__pyx_vp_5pyboy_4core_3lcd_TILES)
static int *__pyx_vp_5pyboy_4core_3lcd_FRAME_CYCLES = 0;
#define __pyx_v_5pyboy_4core_3lcd_FRAME_CYCLES (*__pyx_vp_5pyboy_4core_3lcd_FRAME_CYCLES)
static int *__pyx_vp_5pyboy_4core_3lcd_VIDEO_RAM = 0;
#define __pyx_v_5pyboy_4core_3lcd_VIDEO_RAM (*__pyx_vp_5pyboy_4core_3lcd_VIDEO_RAM)
static int *__pyx_vp_5pyboy_4core_3lcd_OBJECT_ATTRIBUTE_MEMORY = 0;
#define __pyx_v_5pyboy_4core_3lcd_OBJECT_ATTRIBUTE_MEMORY (*__pyx_vp_5pyboy_4core_3lcd_OBJECT_ATTRIBUTE_MEMORY)
static uint32_t *__pyx_vp_5pyboy_4core_3lcd_COL0_FLAG = 0;
#define __pyx_v_5pyboy_4core_3lcd_COL0_FLAG (*__pyx_vp_5pyboy_4core_3lcd_COL0_FLAG)
static uint8_t *__pyx_vp_5pyboy_4core_3lcd_BG_PRIORITY_FLAG = 0;
#define __pyx_v_5pyboy_4core_3lcd_BG_PRIORITY_FLAG (*__pyx_vp_5pyboy_4core_3lcd_BG_PRIORITY_FLAG)
static uint8_t *__pyx_vp_5pyboy_4core_3lcd_CGB_NUM_PALETTES = 0;
#define __pyx_v_5pyboy_4core_3lcd_CGB_NUM_PALETTES (*__pyx_vp_5pyboy_4core_3lcd_CGB_NUM_PALETTES)
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_3lcd_logger = 0;
#define __pyx_v_5pyboy_4core_3lcd_logger (*__pyx_vp_5pyboy_4core_3lcd_logger)

/* Module declarations from "pyboy.core.ram" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_3ram_logger = 0;
#define __pyx_v_5pyboy_4core_3ram_logger (*__pyx_vp_5pyboy_4core_3ram_logger)

/* Module declarations from "pyboy.core.sound" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_5sound_logger = 0;
#define __pyx_v_5pyboy_4core_5sound_logger (*__pyx_vp_5pyboy_4core_5sound_logger)
static int *__pyx_vp_5pyboy_4core_5sound_SOUND_DESYNC_THRESHOLD = 0;
#define __pyx_v_5pyboy_4core_5sound_SOUND_DESYNC_THRESHOLD (*__pyx_vp_5pyboy_4core_5sound_SOUND_DESYNC_THRESHOLD)
static int *__pyx_vp_5pyboy_4core_5sound_CPU_FREQ = 0;
#define __pyx_v_5pyboy_4core_5sound_CPU_FREQ (*__pyx_vp_5pyboy_4core_5sound_CPU_FREQ)

/* Module declarations from "pyboy.core.timer" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_5timer_logger = 0;
#define __pyx_v_5pyboy_4core_5timer_logger (*__pyx_vp_5pyboy_4core_5timer_logger)

/* Module declarations from "pyboy.core.mb" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_2mb_logger = 0;
#define __pyx_v_5pyboy_4core_2mb_logger (*__pyx_vp_5pyboy_4core_2mb_logger)
static uint16_t *__pyx_vp_5pyboy_4core_2mb_STAT = 0;
#define __pyx_v_5pyboy_4core_2mb_STAT (*__pyx_vp_5pyboy_4core_2mb_STAT)
static uint16_t *__pyx_vp_5pyboy_4core_2mb_LY = 0;
#define __pyx_v_5pyboy_4core_2mb_LY (*__pyx_vp_5pyboy_4core_2mb_LY)
static uint16_t *__pyx_vp_5pyboy_4core_2mb_LYC = 0;
#define __pyx_v_5pyboy_4core_2mb_LYC (*__pyx_vp_5pyboy_4core_2mb_LYC)
static short *__pyx_vp_5pyboy_4core_2mb_VBLANK = 0;
#define __pyx_v_5pyboy_4core_2mb_VBLANK (*__pyx_vp_5pyboy_4core_2mb_VBLANK)
static short *__pyx_vp_5pyboy_4core_2mb_LCDC = 0;
#define __pyx_v_5pyboy_4core_2mb_LCDC (*__pyx_vp_5pyboy_4core_2mb_LCDC)
static short *__pyx_vp_5pyboy_4core_2mb_TIMER = 0;
#define __pyx_v_5pyboy_4core_2mb_TIMER (*__pyx_vp_5pyboy_4core_2mb_TIMER)
static short *__pyx_vp_5pyboy_4core_2mb_SERIAL = 0;
#define __pyx_v_5pyboy_4core_2mb_SERIAL (*__pyx_vp_5pyboy_4core_2mb_SERIAL)
static short *__pyx_vp_5pyboy_4core_2mb_HIGHTOLOW = 0;
#define __pyx_v_5pyboy_4core_2mb_HIGHTOLOW (*__pyx_vp_5pyboy_4core_2mb_HIGHTOLOW)
static int *__pyx_vp_5pyboy_4core_2mb_INTR_VBLANK = 0;
#define __pyx_v_5pyboy_4core_2mb_INTR_VBLANK (*__pyx_vp_5pyboy_4core_2mb_INTR_VBLANK)
static int *__pyx_vp_5pyboy_4core_2mb_INTR_LCDC = 0;
#define __pyx_v_5pyboy_4core_2mb_INTR_LCDC (*__pyx_vp_5pyboy_4core_2mb_INTR_LCDC)
static int *__pyx_vp_5pyboy_4core_2mb_INTR_TIMER = 0;
#define __pyx_v_5pyboy_4core_2mb_INTR_TIMER (*__pyx_vp_5pyboy_4core_2mb_INTR_TIMER)
static int *__pyx_vp_5pyboy_4core_2mb_INTR_SERIAL = 0;
#define __pyx_v_5pyboy_4core_2mb_INTR_SERIAL (*__pyx_vp_5pyboy_4core_2mb_INTR_SERIAL)
static int *__pyx_vp_5pyboy_4core_2mb_INTR_HIGHTOLOW = 0;
#define __pyx_v_5pyboy_4core_2mb_INTR_HIGHTOLOW (*__pyx_vp_5pyboy_4core_2mb_INTR_HIGHTOLOW)
static int *__pyx_vp_5pyboy_4core_2mb_STATE_VERSION = 0;
#define __pyx_v_5pyboy_4core_2mb_STATE_VERSION (*__pyx_vp_5pyboy_4core_2mb_STATE_VERSION)

/* Module declarations from "pyboy.core.cpu" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger **__pyx_vp_5pyboy_4core_3cpu_logger = 0;
#define __pyx_v_5pyboy_4core_3cpu_logger (*__pyx_vp_5pyboy_4core_3cpu_logger)
static uint16_t *__pyx_vp_5pyboy_4core_3cpu_IF_ADDRESS = 0;
#define __pyx_v_5pyboy_4core_3cpu_IF_ADDRESS (*__pyx_vp_5pyboy_4core_3cpu_IF_ADDRESS)
static uint16_t *__pyx_vp_5pyboy_4core_3cpu_IE_ADDRESS = 0;
#define __pyx_v_5pyboy_4core_3cpu_IE_ADDRESS (*__pyx_vp_5pyboy_4core_3cpu_IE_ADDRESS)
static int16_t *__pyx_vp_5pyboy_4core_3cpu_FLAGC = 0;
#define __pyx_v_5pyboy_4core_3cpu_FLAGC (*__pyx_vp_5pyboy_4core_3cpu_FLAGC)
static int16_t *__pyx_vp_5pyboy_4core_3cpu_FLAGH = 0;
#define __pyx_v_5pyboy_4core_3cpu_FLAGH (*__pyx_vp_5pyboy_4core_3cpu_FLAGH)
static int16_t *__pyx_vp_5pyboy_4core_3cpu_FLAGN = 0;
#define __pyx_v_5pyboy_4core_3cpu_FLAGN (*__pyx_vp_5pyboy_4core_3cpu_FLAGN)
static int16_t *__pyx_vp_5pyboy_4core_3cpu_FLAGZ = 0;
#define __pyx_v_5pyboy_4core_3cpu_FLAGZ (*__pyx_vp_5pyboy_4core_3cpu_FLAGZ)
static uint8_t *__pyx_vp_5pyboy_4core_3cpu_INTR_VBLANK = 0;
#define __pyx_v_5pyboy_4core_3cpu_INTR_VBLANK (*__pyx_vp_5pyboy_4core_3cpu_INTR_VBLANK)
static uint8_t *__pyx_vp_5pyboy_4core_3cpu_INTR_LCDC = 0;
#define __pyx_v_5pyboy_4core_3cpu_INTR_LCDC (*__pyx_vp_5pyboy_4core_3cpu_INTR_LCDC)
static uint8_t *__pyx_vp_5pyboy_4core_3cpu_INTR_TIMER = 0;
#define __pyx_v_5pyboy_4core_3cpu_INTR_TIMER (*__pyx_vp_5pyboy_4core_3cpu_INTR_TIMER)
static uint8_t *__pyx_vp_5pyboy_4core_3cpu_INTR_SERIAL = 0;
#define __pyx_v_5pyboy_4core_3cpu_INTR_SERIAL (*__pyx_vp_5pyboy_4core_3cpu_INTR_SERIAL)
static uint8_t *__pyx_vp_5pyboy_4core_3cpu_INTR_HIGHTOLOW = 0;
#define __pyx_v_5pyboy_4core_3cpu_INTR_HIGHTOLOW (*__pyx_vp_5pyboy_4core_3cpu_INTR_HIGHTOLOW)

/* Module declarations from "pyboy.core" */

/* Module declarations from "pyboy.core.opcodes" */
static struct __pyx_obj_5pyboy_7logging_7logging_Logger *__pyx_v_5pyboy_4core_7opcodes_logger = 0;
static uint16_t __pyx_v_5pyboy_4core_7opcodes_FLAGC;
static uint16_t __pyx_v_5pyboy_4core_7opcodes_FLAGH;
static uint16_t __pyx_v_5pyboy_4core_7opcodes_FLAGN;
static uint16_t __pyx_v_5pyboy_4core_7opcodes_FLAGZ;
static uint8_t __pyx_v_5pyboy_4core_7opcodes_OPCODE_LENGTHS[0x200];
static PyObject *__pyx_collections_abc_Sequence = 0;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static uint8_t __pyx_f_5pyboy_4core_7opcodes_no_opcode(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_NOP_00(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_01(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_02(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_03(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_04(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_05(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_06(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLCA_07(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_08(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_09(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_0A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_0B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_0C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_0D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_0E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRCA_0F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_STOP_10(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_11(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_12(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_13(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_14(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_15(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_16(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLA_17(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_18(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_19(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_1A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_1B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_1C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_1D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_1E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRA_1F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_20(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_21(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_22(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_23(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_24(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_25(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_26(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DAA_27(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_28(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_29(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_2A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_2B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_2C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_2D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_2E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CPL_2F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_30(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_31(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_32(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_33(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_34(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_35(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_36(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SCF_37(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_38(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_39(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_3A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_3B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_3C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_3D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_3E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CCF_3F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_40(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_41(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_42(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_43(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_44(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_45(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_46(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_47(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_48(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_49(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_50(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_51(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_52(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_53(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_54(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_55(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_56(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_57(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_58(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_59(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_60(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_61(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_62(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_63(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_64(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_65(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_66(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_67(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_68(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_69(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_70(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_71(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_72(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_73(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_74(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_75(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_HALT_76(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_77(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_78(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_79(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_80(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_81(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_82(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_83(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_84(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_85(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_86(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_87(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_88(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_89(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_90(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_91(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_92(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_93(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_94(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_95(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_96(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_97(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_98(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_99(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_A8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_A9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_B8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_B9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_C0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_POP_C1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_C2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_C3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_C4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_PUSH_C5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_C6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_C7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_C8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_C9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_CA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_PREFIX_CB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_CC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_CD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_CE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_CF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_D0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_POP_D1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_D2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_D4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_PUSH_D5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_D6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_D7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_D8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RETI_D9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_DA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_DC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_DE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_DF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LDH_E0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_POP_E1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_E2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_PUSH_E5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_E6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_E7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_E8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_E9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_EA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_EE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_EF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LDH_F0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_POP_F1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_F2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_DI_F3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_PUSH_F5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_F6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_F7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_F8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_F9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_FA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_EI_FB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_FE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_FF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_100(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_101(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_102(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_103(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_104(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_105(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_106(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_107(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_108(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_109(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_110(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_111(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_112(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_113(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_114(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_115(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_116(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_117(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_118(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_119(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_120(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_121(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_122(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_123(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_124(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_125(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_126(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_127(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_128(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_129(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_130(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_131(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_132(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_133(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_134(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_135(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_136(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_137(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_138(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_139(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_140(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_141(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_142(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_143(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_144(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_145(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_146(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_147(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_148(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_149(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_150(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_151(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_152(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_153(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_154(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_155(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_156(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_157(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_158(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_159(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_160(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_161(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_162(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_163(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_164(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_165(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_166(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_167(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_168(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_169(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_170(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_171(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_172(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_173(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_174(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_175(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_176(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_177(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_178(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_179(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_180(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_181(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_182(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_183(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_184(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_185(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_186(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_187(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_188(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_189(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_190(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_191(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_192(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_193(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_194(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_195(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_196(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_197(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_198(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_199(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1ED(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *); /*proto*/
static int __Pyx_carray_from_py_uint8_t(PyObject *, uint8_t *, Py_ssize_t); /*proto*/
static int __pyx_array_allocate_buffer(struct __pyx_array_obj *); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static int assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, PyObject *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, PyObject *); /*proto*/
static int __pyx_memoryview_err_no_memory(void); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyboy.core.opcodes"
extern int __pyx_module_is_main_pyboy__core__opcodes;
int __pyx_module_is_main_pyboy__core__opcodes = 0;

/* Implementation of "pyboy.core.opcodes" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_OverflowError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin___import__;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_RuntimeError;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = ": ";
static const char __pyx_k_B[] = "B";
static const char __pyx_k_O[] = "O";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_DI[] = "DI";
static const char __pyx_k_EI[] = "EI";
static const char __pyx_k__2[] = ".";
static const char __pyx_k__3[] = "*";
static const char __pyx_k__6[] = "'";
static const char __pyx_k__7[] = ")";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_CCF[] = "CCF";
static const char __pyx_k_CPL[] = "CPL";
static const char __pyx_k_DAA[] = "DAA";
static const char __pyx_k_NOP[] = "NOP";
static const char __pyx_k_RET[] = "RET";
static const char __pyx_k_RLA[] = "RLA";
static const char __pyx_k_RRA[] = "RRA";
static const char __pyx_k_SCF[] = "SCF";
static const char __pyx_k__21[] = "";
static const char __pyx_k__22[] = "?";
static const char __pyx_k_abc[] = "abc";
static const char __pyx_k_and[] = " and ";
static const char __pyx_k_got[] = " (got ";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_CP_A[] = "CP A";
static const char __pyx_k_CP_B[] = "CP B";
static const char __pyx_k_CP_C[] = "CP C";
static const char __pyx_k_CP_D[] = "CP D";
static const char __pyx_k_CP_E[] = "CP E";
static const char __pyx_k_CP_H[] = "CP H";
static const char __pyx_k_CP_L[] = "CP L";
static const char __pyx_k_HALT[] = "HALT";
static const char __pyx_k_OR_A[] = "OR A";
static const char __pyx_k_OR_B[] = "OR B";
static const char __pyx_k_OR_C[] = "OR C";
static const char __pyx_k_OR_D[] = "OR D";
static const char __pyx_k_OR_E[] = "OR E";
static const char __pyx_k_OR_H[] = "OR H";
static const char __pyx_k_OR_L[] = "OR L";
static const char __pyx_k_RETI[] = "RETI";
static const char __pyx_k_RLCA[] = "RLCA";
static const char __pyx_k_RL_A[] = "RL A";
static const char __pyx_k_RL_B[] = "RL B";
static const char __pyx_k_RL_C[] = "RL C";
static const char __pyx_k_RL_D[] = "RL D";
static const char __pyx_k_RL_E[] = "RL E";
static const char __pyx_k_RL_H[] = "RL H";
static const char __pyx_k_RL_L[] = "RL L";
static const char __pyx_k_RRCA[] = "RRCA";
static const char __pyx_k_RR_A[] = "RR A";
static const char __pyx_k_RR_B[] = "RR B";
static const char __pyx_k_RR_C[] = "RR C";
static const char __pyx_k_RR_D[] = "RR D";
static const char __pyx_k_RR_E[] = "RR E";
static const char __pyx_k_RR_H[] = "RR H";
static const char __pyx_k_RR_L[] = "RR L";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_AND_A[] = "AND A";
static const char __pyx_k_AND_B[] = "AND B";
static const char __pyx_k_AND_C[] = "AND C";
static const char __pyx_k_AND_D[] = "AND D";
static const char __pyx_k_AND_E[] = "AND E";
static const char __pyx_k_AND_H[] = "AND H";
static const char __pyx_k_AND_L[] = "AND L";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_CP_HL[] = "CP (HL)";
static const char __pyx_k_CP_d8[] = "CP d8";
static const char __pyx_k_DEC_A[] = "DEC A";
static const char __pyx_k_DEC_B[] = "DEC B";
static const char __pyx_k_DEC_C[] = "DEC C";
static const char __pyx_k_DEC_D[] = "DEC D";
static const char __pyx_k_DEC_E[] = "DEC E";
static const char __pyx_k_DEC_H[] = "DEC H";
static const char __pyx_k_DEC_L[] = "DEC L";
static const char __pyx_k_FLAGC[] = "FLAGC";
static const char __pyx_k_FLAGH[] = "FLAGH";
static const char __pyx_k_FLAGN[] = "FLAGN";
static const char __pyx_k_FLAGZ[] = "FLAGZ";
static const char __pyx_k_INC_A[] = "INC A";
static const char __pyx_k_INC_B[] = "INC B";
static const char __pyx_k_INC_C[] = "INC C";
static const char __pyx_k_INC_D[] = "INC D";
static const char __pyx_k_INC_E[] = "INC E";
static const char __pyx_k_INC_H[] = "INC H";
static const char __pyx_k_INC_L[] = "INC L";
static const char __pyx_k_JP_HL[] = "JP (HL)";
static const char __pyx_k_JR_r8[] = "JR r8";
static const char __pyx_k_OR_HL[] = "OR (HL)";
static const char __pyx_k_OR_d8[] = "OR d8";
static const char __pyx_k_RET_C[] = "RET C";
static const char __pyx_k_RET_Z[] = "RET Z";
static const char __pyx_k_RLC_A[] = "RLC A";
static const char __pyx_k_RLC_B[] = "RLC B";
static const char __pyx_k_RLC_C[] = "RLC C";
static const char __pyx_k_RLC_D[] = "RLC D";
static const char __pyx_k_RLC_E[] = "RLC E";
static const char __pyx_k_RLC_H[] = "RLC H";
static const char __pyx_k_RLC_L[] = "RLC L";
static const char __pyx_k_RL_HL[] = "RL (HL)";
static const char __pyx_k_RRC_A[] = "RRC A";
static const char __pyx_k_RRC_B[] = "RRC B";
static const char __pyx_k_RRC_C[] = "RRC C";
static const char __pyx_k_RRC_D[] = "RRC D";
static const char __pyx_k_RRC_E[] = "RRC E";
static const char __pyx_k_RRC_H[] = "RRC H";
static const char __pyx_k_RRC_L[] = "RRC L";
static const char __pyx_k_RR_HL[] = "RR (HL)";
static const char __pyx_k_SLA_A[] = "SLA A";
static const char __pyx_k_SLA_B[] = "SLA B";
static const char __pyx_k_SLA_C[] = "SLA C";
static const char __pyx_k_SLA_D[] = "SLA D";
static const char __pyx_k_SLA_E[] = "SLA E";
static const char __pyx_k_SLA_H[] = "SLA H";
static const char __pyx_k_SLA_L[] = "SLA L";
static const char __pyx_k_SRA_A[] = "SRA A";
static const char __pyx_k_SRA_B[] = "SRA B";
static const char __pyx_k_SRA_C[] = "SRA C";
static const char __pyx_k_SRA_D[] = "SRA D";
static const char __pyx_k_SRA_E[] = "SRA E";
static const char __pyx_k_SRA_H[] = "SRA H";
static const char __pyx_k_SRA_L[] = "SRA L";
static const char __pyx_k_SRL_A[] = "SRL A";
static const char __pyx_k_SRL_B[] = "SRL B";
static const char __pyx_k_SRL_C[] = "SRL C";
static const char __pyx_k_SRL_D[] = "SRL D";
static const char __pyx_k_SRL_E[] = "SRL E";
static const char __pyx_k_SRL_H[] = "SRL H";
static const char __pyx_k_SRL_L[] = "SRL L";
static const char __pyx_k_SUB_A[] = "SUB A";
static const char __pyx_k_SUB_B[] = "SUB B";
static const char __pyx_k_SUB_C[] = "SUB C";
static const char __pyx_k_SUB_D[] = "SUB D";
static const char __pyx_k_SUB_E[] = "SUB E";
static const char __pyx_k_SUB_H[] = "SUB H";
static const char __pyx_k_SUB_L[] = "SUB L";
static const char __pyx_k_XOR_A[] = "XOR A";
static const char __pyx_k_XOR_B[] = "XOR B";
static const char __pyx_k_XOR_C[] = "XOR C";
static const char __pyx_k_XOR_D[] = "XOR D";
static const char __pyx_k_XOR_E[] = "XOR E";
static const char __pyx_k_XOR_H[] = "XOR H";
static const char __pyx_k_XOR_L[] = "XOR L";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_pyboy[] = "pyboy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_utils[] = "utils";
static const char __pyx_k_AND_HL[] = "AND (HL)";
static const char __pyx_k_AND_d8[] = "AND d8";
static const char __pyx_k_DEC_BC[] = "DEC BC";
static const char __pyx_k_DEC_DE[] = "DEC DE";
static const char __pyx_k_DEC_HL[] = "DEC HL";
static const char __pyx_k_DEC_SP[] = "DEC SP";
static const char __pyx_k_INC_BC[] = "INC BC";
static const char __pyx_k_INC_DE[] = "INC DE";
static const char __pyx_k_INC_HL[] = "INC HL";
static const char __pyx_k_INC_SP[] = "INC SP";
static const char __pyx_k_JP_a16[] = "JP a16";
static const char __pyx_k_LD_A_A[] = "LD A,A";
static const char __pyx_k_LD_A_B[] = "LD A,B";
static const char __pyx_k_LD_A_C[] = "LD A,C";
static const char __pyx_k_LD_A_D[] = "LD A,D";
static const char __pyx_k_LD_A_E[] = "LD A,E";
static const char __pyx_k_LD_A_H[] = "LD A,H";
static const char __pyx_k_LD_A_L[] = "LD A,L";
static const char __pyx_k_LD_B_A[] = "LD B,A";
static const char __pyx_k_LD_B_B[] = "LD B,B";
static const char __pyx_k_LD_B_C[] = "LD B,C";
static const char __pyx_k_LD_B_D[] = "LD B,D";
static const char __pyx_k_LD_B_E[] = "LD B,E";
static const char __pyx_k_LD_B_H[] = "LD B,H";
static const char __pyx_k_LD_B_L[] = "LD B,L";
static const char __pyx_k_LD_C_A[] = "LD C,A";
static const char __pyx_k_LD_C_B[] = "LD C,B";
static const char __pyx_k_LD_C_C[] = "LD C,C";
static const char __pyx_k_LD_C_D[] = "LD C,D";
static const char __pyx_k_LD_C_E[] = "LD C,E";
static const char __pyx_k_LD_C_H[] = "LD C,H";
static const char __pyx_k_LD_C_L[] = "LD C,L";
static const char __pyx_k_LD_D_A[] = "LD D,A";
static const char __pyx_k_LD_D_B[] = "LD D,B";
static const char __pyx_k_LD_D_C[] = "LD D,C";
static const char __pyx_k_LD_D_D[] = "LD D,D";
static const char __pyx_k_LD_D_E[] = "LD D,E";
static const char __pyx_k_LD_D_H[] = "LD D,H";
static const char __pyx_k_LD_D_L[] = "LD D,L";
static const char __pyx_k_LD_E_A[] = "LD E,A";
static const char __pyx_k_LD_E_B[] = "LD E,B";
static const char __pyx_k_LD_E_C[] = "LD E,C";
static const char __pyx_k_LD_E_D[] = "LD E,D";
static const char __pyx_k_LD_E_E[] = "LD E,E";
static const char __pyx_k_LD_E_H[] = "LD E,H";
static const char __pyx_k_LD_E_L[] = "LD E,L";
static const char __pyx_k_LD_H_A[] = "LD H,A";
static const char __pyx_k_LD_H_B[] = "LD H,B";
static const char __pyx_k_LD_H_C[] = "LD H,C";
static const char __pyx_k_LD_H_D[] = "LD H,D";
static const char __pyx_k_LD_H_E[] = "LD H,E";
static const char __pyx_k_LD_H_H[] = "LD H,H";
static const char __pyx_k_LD_H_L[] = "LD H,L";
static const char __pyx_k_LD_L_A[] = "LD L,A";
static const char __pyx_k_LD_L_B[] = "LD L,B";
static const char __pyx_k_LD_L_C[] = "LD L,C";
static const char __pyx_k_LD_L_D[] = "LD L,D";
static const char __pyx_k_LD_L_E[] = "LD L,E";
static const char __pyx_k_LD_L_H[] = "LD L,H";
static const char __pyx_k_LD_L_L[] = "LD L,L";
static const char __pyx_k_POP_AF[] = "POP AF";
static const char __pyx_k_POP_BC[] = "POP BC";
static const char __pyx_k_POP_DE[] = "POP DE";
static const char __pyx_k_POP_HL[] = "POP HL";
static const char __pyx_k_RET_NC[] = "RET NC";
static const char __pyx_k_RET_NZ[] = "RET NZ";
static const char __pyx_k_RLC_HL[] = "RLC (HL)";
static const char __pyx_k_RRC_HL[] = "RRC (HL)";
static const char __pyx_k_SLA_HL[] = "SLA (HL)";
static const char __pyx_k_SRA_HL[] = "SRA (HL)";
static const char __pyx_k_SRL_HL[] = "SRL (HL)";
static const char __pyx_k_STOP_0[] = "STOP 0";
static const char __pyx_k_SUB_HL[] = "SUB (HL)";
static const char __pyx_k_SUB_d8[] = "SUB d8";
static const char __pyx_k_SWAP_A[] = "SWAP A";
static const char __pyx_k_SWAP_B[] = "SWAP B";
static const char __pyx_k_SWAP_C[] = "SWAP C";
static const char __pyx_k_SWAP_D[] = "SWAP D";
static const char __pyx_k_SWAP_E[] = "SWAP E";
static const char __pyx_k_SWAP_H[] = "SWAP H";
static const char __pyx_k_SWAP_L[] = "SWAP L";
static const char __pyx_k_XOR_HL[] = "XOR (HL)";
static const char __pyx_k_XOR_d8[] = "XOR d8";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_ADC_A_A[] = "ADC A,A";
static const char __pyx_k_ADC_A_B[] = "ADC A,B";
static const char __pyx_k_ADC_A_C[] = "ADC A,C";
static const char __pyx_k_ADC_A_D[] = "ADC A,D";
static const char __pyx_k_ADC_A_E[] = "ADC A,E";
static const char __pyx_k_ADC_A_H[] = "ADC A,H";
static const char __pyx_k_ADC_A_L[] = "ADC A,L";
static const char __pyx_k_ADD_A_A[] = "ADD A,A";
static const char __pyx_k_ADD_A_B[] = "ADD A,B";
static const char __pyx_k_ADD_A_C[] = "ADD A,C";
static const char __pyx_k_ADD_A_D[] = "ADD A,D";
static const char __pyx_k_ADD_A_E[] = "ADD A,E";
static const char __pyx_k_ADD_A_H[] = "ADD A,H";
static const char __pyx_k_ADD_A_L[] = "ADD A,L";
static const char __pyx_k_BIT_0_A[] = "BIT 0,A";
static const char __pyx_k_BIT_0_B[] = "BIT 0,B";
static const char __pyx_k_BIT_0_C[] = "BIT 0,C";
static const char __pyx_k_BIT_0_D[] = "BIT 0,D";
static const char __pyx_k_BIT_0_E[] = "BIT 0,E";
static const char __pyx_k_BIT_0_H[] = "BIT 0,H";
static const char __pyx_k_BIT_0_L[] = "BIT 0,L";
static const char __pyx_k_BIT_1_A[] = "BIT 1,A";
static const char __pyx_k_BIT_1_B[] = "BIT 1,B";
static const char __pyx_k_BIT_1_C[] = "BIT 1,C";
static const char __pyx_k_BIT_1_D[] = "BIT 1,D";
static const char __pyx_k_BIT_1_E[] = "BIT 1,E";
static const char __pyx_k_BIT_1_H[] = "BIT 1,H";
static const char __pyx_k_BIT_1_L[] = "BIT 1,L";
static const char __pyx_k_BIT_2_A[] = "BIT 2,A";
static const char __pyx_k_BIT_2_B[] = "BIT 2,B";
static const char __pyx_k_BIT_2_C[] = "BIT 2,C";
static const char __pyx_k_BIT_2_D[] = "BIT 2,D";
static const char __pyx_k_BIT_2_E[] = "BIT 2,E";
static const char __pyx_k_BIT_2_H[] = "BIT 2,H";
static const char __pyx_k_BIT_2_L[] = "BIT 2,L";
static const char __pyx_k_BIT_3_A[] = "BIT 3,A";
static const char __pyx_k_BIT_3_B[] = "BIT 3,B";
static const char __pyx_k_BIT_3_C[] = "BIT 3,C";
static const char __pyx_k_BIT_3_D[] = "BIT 3,D";
static const char __pyx_k_BIT_3_E[] = "BIT 3,E";
static const char __pyx_k_BIT_3_H[] = "BIT 3,H";
static const char __pyx_k_BIT_3_L[] = "BIT 3,L";
static const char __pyx_k_BIT_4_A[] = "BIT 4,A";
static const char __pyx_k_BIT_4_B[] = "BIT 4,B";
static const char __pyx_k_BIT_4_C[] = "BIT 4,C";
static const char __pyx_k_BIT_4_D[] = "BIT 4,D";
static const char __pyx_k_BIT_4_E[] = "BIT 4,E";
static const char __pyx_k_BIT_4_H[] = "BIT 4,H";
static const char __pyx_k_BIT_4_L[] = "BIT 4,L";
static const char __pyx_k_BIT_5_A[] = "BIT 5,A";
static const char __pyx_k_BIT_5_B[] = "BIT 5,B";
static const char __pyx_k_BIT_5_C[] = "BIT 5,C";
static const char __pyx_k_BIT_5_D[] = "BIT 5,D";
static const char __pyx_k_BIT_5_E[] = "BIT 5,E";
static const char __pyx_k_BIT_5_H[] = "BIT 5,H";
static const char __pyx_k_BIT_5_L[] = "BIT 5,L";
static const char __pyx_k_BIT_6_A[] = "BIT 6,A";
static const char __pyx_k_BIT_6_B[] = "BIT 6,B";
static const char __pyx_k_BIT_6_C[] = "BIT 6,C";
static const char __pyx_k_BIT_6_D[] = "BIT 6,D";
static const char __pyx_k_BIT_6_E[] = "BIT 6,E";
static const char __pyx_k_BIT_6_H[] = "BIT 6,H";
static const char __pyx_k_BIT_6_L[] = "BIT 6,L";
static const char __pyx_k_BIT_7_A[] = "BIT 7,A";
static const char __pyx_k_BIT_7_B[] = "BIT 7,B";
static const char __pyx_k_BIT_7_C[] = "BIT 7,C";
static const char __pyx_k_BIT_7_D[] = "BIT 7,D";
static const char __pyx_k_BIT_7_E[] = "BIT 7,E";
static const char __pyx_k_BIT_7_H[] = "BIT 7,H";
static const char __pyx_k_BIT_7_L[] = "BIT 7,L";
static const char __pyx_k_JR_C_r8[] = "JR C,r8";
static const char __pyx_k_JR_Z_r8[] = "JR Z,r8";
static const char __pyx_k_LD_A_BC[] = "LD A,(BC)";
static const char __pyx_k_LD_A_DE[] = "LD A,(DE)";
static const char __pyx_k_LD_A_HL[] = "LD A,(HL+)";
static const char __pyx_k_LD_A_d8[] = "LD A,d8";
static const char __pyx_k_LD_BC_A[] = "LD (BC),A";
static const char __pyx_k_LD_B_HL[] = "LD B,(HL)";
static const char __pyx_k_LD_B_d8[] = "LD B,d8";
static const char __pyx_k_LD_C_HL[] = "LD C,(HL)";
static const char __pyx_k_LD_C_d8[] = "LD C,d8";
static const char __pyx_k_LD_DE_A[] = "LD (DE),A";
static const char __pyx_k_LD_D_HL[] = "LD D,(HL)";
static const char __pyx_k_LD_D_d8[] = "LD D,d8";
static const char __pyx_k_LD_E_HL[] = "LD E,(HL)";
static const char __pyx_k_LD_E_d8[] = "LD E,d8";
static const char __pyx_k_LD_HL_A[] = "LD (HL+),A";
static const char __pyx_k_LD_HL_B[] = "LD (HL),B";
static const char __pyx_k_LD_HL_C[] = "LD (HL),C";
static const char __pyx_k_LD_HL_D[] = "LD (HL),D";
static const char __pyx_k_LD_HL_E[] = "LD (HL),E";
static const char __pyx_k_LD_HL_H[] = "LD (HL),H";
static const char __pyx_k_LD_HL_L[] = "LD (HL),L";
static const char __pyx_k_LD_H_HL[] = "LD H,(HL)";
static const char __pyx_k_LD_H_d8[] = "LD H,d8";
static const char __pyx_k_LD_L_HL[] = "LD L,(HL)";
static const char __pyx_k_LD_L_d8[] = "LD L,d8";
static const char __pyx_k_PUSH_AF[] = "PUSH AF";
static const char __pyx_k_PUSH_BC[] = "PUSH BC";
static const char __pyx_k_PUSH_DE[] = "PUSH DE";
static const char __pyx_k_PUSH_HL[] = "PUSH HL";
static const char __pyx_k_RES_0_A[] = "RES 0,A";
static const char __pyx_k_RES_0_B[] = "RES 0,B";
static const char __pyx_k_RES_0_C[] = "RES 0,C";
static const char __pyx_k_RES_0_D[] = "RES 0,D";
static const char __pyx_k_RES_0_E[] = "RES 0,E";
static const char __pyx_k_RES_0_H[] = "RES 0,H";
static const char __pyx_k_RES_0_L[] = "RES 0,L";
static const char __pyx_k_RES_1_A[] = "RES 1,A";
static const char __pyx_k_RES_1_B[] = "RES 1,B";
static const char __pyx_k_RES_1_C[] = "RES 1,C";
static const char __pyx_k_RES_1_D[] = "RES 1,D";
static const char __pyx_k_RES_1_E[] = "RES 1,E";
static const char __pyx_k_RES_1_H[] = "RES 1,H";
static const char __pyx_k_RES_1_L[] = "RES 1,L";
static const char __pyx_k_RES_2_A[] = "RES 2,A";
static const char __pyx_k_RES_2_B[] = "RES 2,B";
static const char __pyx_k_RES_2_C[] = "RES 2,C";
static const char __pyx_k_RES_2_D[] = "RES 2,D";
static const char __pyx_k_RES_2_E[] = "RES 2,E";
static const char __pyx_k_RES_2_H[] = "RES 2,H";
static const char __pyx_k_RES_2_L[] = "RES 2,L";
static const char __pyx_k_RES_3_A[] = "RES 3,A";
static const char __pyx_k_RES_3_B[] = "RES 3,B";
static const char __pyx_k_RES_3_C[] = "RES 3,C";
static const char __pyx_k_RES_3_D[] = "RES 3,D";
static const char __pyx_k_RES_3_E[] = "RES 3,E";
static const char __pyx_k_RES_3_H[] = "RES 3,H";
static const char __pyx_k_RES_3_L[] = "RES 3,L";
static const char __pyx_k_RES_4_A[] = "RES 4,A";
static const char __pyx_k_RES_4_B[] = "RES 4,B";
static const char __pyx_k_RES_4_C[] = "RES 4,C";
static const char __pyx_k_RES_4_D[] = "RES 4,D";
static const char __pyx_k_RES_4_E[] = "RES 4,E";
static const char __pyx_k_RES_4_H[] = "RES 4,H";
static const char __pyx_k_RES_4_L[] = "RES 4,L";
static const char __pyx_k_RES_5_A[] = "RES 5,A";
static const char __pyx_k_RES_5_B[] = "RES 5,B";
static const char __pyx_k_RES_5_C[] = "RES 5,C";
static const char __pyx_k_RES_5_D[] = "RES 5,D";
static const char __pyx_k_RES_5_E[] = "RES 5,E";
static const char __pyx_k_RES_5_H[] = "RES 5,H";
static const char __pyx_k_RES_5_L[] = "RES 5,L";
static const char __pyx_k_RES_6_A[] = "RES 6,A";
static const char __pyx_k_RES_6_B[] = "RES 6,B";
static const char __pyx_k_RES_6_C[] = "RES 6,C";
static const char __pyx_k_RES_6_D[] = "RES 6,D";
static const char __pyx_k_RES_6_E[] = "RES 6,E";
static const char __pyx_k_RES_6_H[] = "RES 6,H";
static const char __pyx_k_RES_6_L[] = "RES 6,L";
static const char __pyx_k_RES_7_A[] = "RES 7,A";
static const char __pyx_k_RES_7_B[] = "RES 7,B";
static const char __pyx_k_RES_7_C[] = "RES 7,C";
static const char __pyx_k_RES_7_D[] = "RES 7,D";
static const char __pyx_k_RES_7_E[] = "RES 7,E";
static const char __pyx_k_RES_7_H[] = "RES 7,H";
static const char __pyx_k_RES_7_L[] = "RES 7,L";
static const char __pyx_k_RST_00H[] = "RST 00H";
static const char __pyx_k_RST_08H[] = "RST 08H";
static const char __pyx_k_RST_10H[] = "RST 10H";
static const char __pyx_k_RST_18H[] = "RST 18H";
static const char __pyx_k_RST_20H[] = "RST 20H";
static const char __pyx_k_RST_28H[] = "RST 28H";
static const char __pyx_k_RST_30H[] = "RST 30H";
static const char __pyx_k_RST_38H[] = "RST 38H";
static const char __pyx_k_SBC_A_A[] = "SBC A,A";
static const char __pyx_k_SBC_A_B[] = "SBC A,B";
static const char __pyx_k_SBC_A_C[] = "SBC A,C";
static const char __pyx_k_SBC_A_D[] = "SBC A,D";
static const char __pyx_k_SBC_A_E[] = "SBC A,E";
static const char __pyx_k_SBC_A_H[] = "SBC A,H";
static const char __pyx_k_SBC_A_L[] = "SBC A,L";
static const char __pyx_k_SET_0_A[] = "SET 0,A";
static const char __pyx_k_SET_0_B[] = "SET 0,B";
static const char __pyx_k_SET_0_C[] = "SET 0,C";
static const char __pyx_k_SET_0_D[] = "SET 0,D";
static const char __pyx_k_SET_0_E[] = "SET 0,E";
static const char __pyx_k_SET_0_H[] = "SET 0,H";
static const char __pyx_k_SET_0_L[] = "SET 0,L";
static const char __pyx_k_SET_1_A[] = "SET 1,A";
static const char __pyx_k_SET_1_B[] = "SET 1,B";
static const char __pyx_k_SET_1_C[] = "SET 1,C";
static const char __pyx_k_SET_1_D[] = "SET 1,D";
static const char __pyx_k_SET_1_E[] = "SET 1,E";
static const char __pyx_k_SET_1_H[] = "SET 1,H";
static const char __pyx_k_SET_1_L[] = "SET 1,L";
static const char __pyx_k_SET_2_A[] = "SET 2,A";
static const char __pyx_k_SET_2_B[] = "SET 2,B";
static const char __pyx_k_SET_2_C[] = "SET 2,C";
static const char __pyx_k_SET_2_D[] = "SET 2,D";
static const char __pyx_k_SET_2_E[] = "SET 2,E";
static const char __pyx_k_SET_2_H[] = "SET 2,H";
static const char __pyx_k_SET_2_L[] = "SET 2,L";
static const char __pyx_k_SET_3_A[] = "SET 3,A";
static const char __pyx_k_SET_3_B[] = "SET 3,B";
static const char __pyx_k_SET_3_C[] = "SET 3,C";
static const char __pyx_k_SET_3_D[] = "SET 3,D";
static const char __pyx_k_SET_3_E[] = "SET 3,E";
static const char __pyx_k_SET_3_H[] = "SET 3,H";
static const char __pyx_k_SET_3_L[] = "SET 3,L";
static const char __pyx_k_SET_4_A[] = "SET 4,A";
static const char __pyx_k_SET_4_B[] = "SET 4,B";
static const char __pyx_k_SET_4_C[] = "SET 4,C";
static const char __pyx_k_SET_4_D[] = "SET 4,D";
static const char __pyx_k_SET_4_E[] = "SET 4,E";
static const char __pyx_k_SET_4_H[] = "SET 4,H";
static const char __pyx_k_SET_4_L[] = "SET 4,L";
static const char __pyx_k_SET_5_A[] = "SET 5,A";
static const char __pyx_k_SET_5_B[] = "SET 5,B";
static const char __pyx_k_SET_5_C[] = "SET 5,C";
static const char __pyx_k_SET_5_D[] = "SET 5,D";
static const char __pyx_k_SET_5_E[] = "SET 5,E";
static const char __pyx_k_SET_5_H[] = "SET 5,H";
static const char __pyx_k_SET_5_L[] = "SET 5,L";
static const char __pyx_k_SET_6_A[] = "SET 6,A";
static const char __pyx_k_SET_6_B[] = "SET 6,B";
static const char __pyx_k_SET_6_C[] = "SET 6,C";
static const char __pyx_k_SET_6_D[] = "SET 6,D";
static const char __pyx_k_SET_6_E[] = "SET 6,E";
static const char __pyx_k_SET_6_H[] = "SET 6,H";
static const char __pyx_k_SET_6_L[] = "SET 6,L";
static const char __pyx_k_SET_7_A[] = "SET 7,A";
static const char __pyx_k_SET_7_B[] = "SET 7,B";
static const char __pyx_k_SET_7_C[] = "SET 7,C";
static const char __pyx_k_SET_7_D[] = "SET 7,D";
static const char __pyx_k_SET_7_E[] = "SET 7,E";
static const char __pyx_k_SET_7_H[] = "SET 7,H";
static const char __pyx_k_SET_7_L[] = "SET 7,L";
static const char __pyx_k_SWAP_HL[] = "SWAP (HL)";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_logging[] = "logging";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_ADC_A_HL[] = "ADC A,(HL)";
static const char __pyx_k_ADC_A_d8[] = "ADC A,d8";
static const char __pyx_k_ADD_A_HL[] = "ADD A,(HL)";
static const char __pyx_k_ADD_A_d8[] = "ADD A,d8";
static const char __pyx_k_BIT_0_HL[] = "BIT 0,(HL)";
static const char __pyx_k_BIT_1_HL[] = "BIT 1,(HL)";
static const char __pyx_k_BIT_2_HL[] = "BIT 2,(HL)";
static const char __pyx_k_BIT_3_HL[] = "BIT 3,(HL)";
static const char __pyx_k_BIT_4_HL[] = "BIT 4,(HL)";
static const char __pyx_k_BIT_5_HL[] = "BIT 5,(HL)";
static const char __pyx_k_BIT_6_HL[] = "BIT 6,(HL)";
static const char __pyx_k_BIT_7_HL[] = "BIT 7,(HL)";
static const char __pyx_k_CALL_a16[] = "CALL a16";
static const char __pyx_k_DEC_HL_2[] = "DEC (HL)";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_INC_HL_2[] = "INC (HL)";
static const char __pyx_k_JP_C_a16[] = "JP C,a16";
static const char __pyx_k_JP_Z_a16[] = "JP Z,a16";
static const char __pyx_k_JR_NC_r8[] = "JR NC,r8";
static const char __pyx_k_JR_NZ_r8[] = "JR NZ,r8";
static const char __pyx_k_LDH_A_a8[] = "LDH A,(a8)";
static const char __pyx_k_LDH_a8_A[] = "LDH (a8),A";
static const char __pyx_k_LD_A_C_2[] = "LD A,(C)";
static const char __pyx_k_LD_A_a16[] = "LD A,(a16)";
static const char __pyx_k_LD_C_A_2[] = "LD (C),A";
static const char __pyx_k_LD_HL_d8[] = "LD (HL),d8";
static const char __pyx_k_LD_SP_HL[] = "LD SP,HL";
static const char __pyx_k_LD_a16_A[] = "LD (a16),A";
static const char __pyx_k_RES_0_HL[] = "RES 0,(HL)";
static const char __pyx_k_RES_1_HL[] = "RES 1,(HL)";
static const char __pyx_k_RES_2_HL[] = "RES 2,(HL)";
static const char __pyx_k_RES_3_HL[] = "RES 3,(HL)";
static const char __pyx_k_RES_4_HL[] = "RES 4,(HL)";
static const char __pyx_k_RES_5_HL[] = "RES 5,(HL)";
static const char __pyx_k_RES_6_HL[] = "RES 6,(HL)";
static const char __pyx_k_RES_7_HL[] = "RES 7,(HL)";
static const char __pyx_k_SBC_A_HL[] = "SBC A,(HL)";
static const char __pyx_k_SBC_A_d8[] = "SBC A,d8";
static const char __pyx_k_SET_0_HL[] = "SET 0,(HL)";
static const char __pyx_k_SET_1_HL[] = "SET 1,(HL)";
static const char __pyx_k_SET_2_HL[] = "SET 2,(HL)";
static const char __pyx_k_SET_3_HL[] = "SET 3,(HL)";
static const char __pyx_k_SET_4_HL[] = "SET 4,(HL)";
static const char __pyx_k_SET_5_HL[] = "SET 5,(HL)";
static const char __pyx_k_SET_6_HL[] = "SET 6,(HL)";
static const char __pyx_k_SET_7_HL[] = "SET 7,(HL)";
static const char __pyx_k_Sequence[] = "Sequence";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_register[] = "register";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_ADD_HL_BC[] = "ADD HL,BC";
static const char __pyx_k_ADD_HL_DE[] = "ADD HL,DE";
static const char __pyx_k_ADD_HL_HL[] = "ADD HL,HL";
static const char __pyx_k_ADD_HL_SP[] = "ADD HL,SP";
static const char __pyx_k_ADD_SP_r8[] = "ADD SP,r8";
static const char __pyx_k_JP_NC_a16[] = "JP NC,a16";
static const char __pyx_k_JP_NZ_a16[] = "JP NZ,a16";
static const char __pyx_k_LD_A_HL_2[] = "LD A,(HL-)";
static const char __pyx_k_LD_A_HL_3[] = "LD A,(HL)";
static const char __pyx_k_LD_BC_d16[] = "LD BC,d16";
static const char __pyx_k_LD_DE_d16[] = "LD DE,d16";
static const char __pyx_k_LD_HL_A_2[] = "LD (HL-),A";
static const char __pyx_k_LD_HL_A_3[] = "LD (HL),A";
static const char __pyx_k_LD_HL_d16[] = "LD HL,d16";
static const char __pyx_k_LD_SP_d16[] = "LD SP,d16";
static const char __pyx_k_LD_a16_SP[] = "LD (a16),SP";
static const char __pyx_k_PREFIX_CB[] = "PREFIX CB";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_CALL_C_a16[] = "CALL C,a16";
static const char __pyx_k_CALL_Z_a16[] = "CALL Z,a16";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_get_logger[] = "get_logger";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_CALL_NC_a16[] = "CALL NC,a16";
static const char __pyx_k_CALL_NZ_a16[] = "CALL NZ,a16";
static const char __pyx_k_LD_HL_SP_r8[] = "LD HL,SP+r8";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_CPU_COMMANDS[] = "CPU_COMMANDS";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_version_info[] = "version_info";
static const char __pyx_k_OverflowError[] = "OverflowError";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_OPCODE_LENGTHS[] = "OPCODE_LENGTHS";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_collections_abc[] = "collections.abc";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_CB_cannot_be_called[] = "CB cannot be called!";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_Invalid_shape_in_axis[] = "Invalid shape in axis ";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_Cannot_index_with_type[] = "Cannot index with type '";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_Dimension_d_is_not_direct[] = "Dimension %d is not direct";
static const char __pyx_k_Index_out_of_bounds_axis_d[] = "Index out of bounds (axis %d)";
static const char __pyx_k_Step_may_not_be_zero_axis_d[] = "Step may not be zero (axis %d)";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_All_dimensions_preceding_dimensi[] = "All dimensions preceding dimension %d must be indexed and not sliced";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Cannot_transpose_memoryview_with[] = "Cannot transpose memoryview with indirect dimensions";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got ";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis ";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension ";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
/* #### Code section: decls ### */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_7logging_7logging_Logger;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_7bootrom_BootROM;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_5utils_IntIOInterface;
  PyTypeObject *__pyx_ptype_5pyboy_5utils_IntIOWrapper;
  PyTypeObject *__pyx_ptype_5pyboy_5utils_WindowEvent;
  PyTypeObject *__pyx_ptype_5pyboy_5utils_WindowEventMouse;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_9cartridge_3rtc_RTC;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_BaseMBC;
  PyTypeObject *__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_ROMOnly;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_11interaction_Interaction;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_5array_array;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_LCD;
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_PaletteRegister;
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_STATRegister;
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_LCDCRegister;
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_Renderer;
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_CGBLCD;
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_CGBRenderer;
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_VBKregister;
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_PaletteIndexRegister;
  PyTypeObject *__pyx_ptype_5pyboy_4core_3lcd_PaletteColorRegister;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_3ram_RAM;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_5sound_Sound;
  PyTypeObject *__pyx_ptype_5pyboy_4core_5sound_ToneChannel;
  PyTypeObject *__pyx_ptype_5pyboy_4core_5sound_SweepChannel;
  PyTypeObject *__pyx_ptype_5pyboy_4core_5sound_WaveChannel;
  PyTypeObject *__pyx_ptype_5pyboy_4core_5sound_NoiseChannel;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_5timer_Timer;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_2mb_Motherboard;
  PyTypeObject *__pyx_ptype_5pyboy_4core_2mb_HDMA;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5pyboy_4core_3cpu_CPU;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type___pyx_array;
  PyObject *__pyx_type___pyx_MemviewEnum;
  PyObject *__pyx_type___pyx_memoryview;
  PyObject *__pyx_type___pyx_memoryviewslice;
  #endif
  PyTypeObject *__pyx_array_type;
  PyTypeObject *__pyx_MemviewEnum_type;
  PyTypeObject *__pyx_memoryview_type;
  PyTypeObject *__pyx_memoryviewslice_type;
  PyObject *__pyx_kp_u_;
  PyObject *__pyx_kp_u_ADC_A_A;
  PyObject *__pyx_kp_u_ADC_A_B;
  PyObject *__pyx_kp_u_ADC_A_C;
  PyObject *__pyx_kp_u_ADC_A_D;
  PyObject *__pyx_kp_u_ADC_A_E;
  PyObject *__pyx_kp_u_ADC_A_H;
  PyObject *__pyx_kp_u_ADC_A_HL;
  PyObject *__pyx_kp_u_ADC_A_L;
  PyObject *__pyx_kp_u_ADC_A_d8;
  PyObject *__pyx_kp_u_ADD_A_A;
  PyObject *__pyx_kp_u_ADD_A_B;
  PyObject *__pyx_kp_u_ADD_A_C;
  PyObject *__pyx_kp_u_ADD_A_D;
  PyObject *__pyx_kp_u_ADD_A_E;
  PyObject *__pyx_kp_u_ADD_A_H;
  PyObject *__pyx_kp_u_ADD_A_HL;
  PyObject *__pyx_kp_u_ADD_A_L;
  PyObject *__pyx_kp_u_ADD_A_d8;
  PyObject *__pyx_kp_u_ADD_HL_BC;
  PyObject *__pyx_kp_u_ADD_HL_DE;
  PyObject *__pyx_kp_u_ADD_HL_HL;
  PyObject *__pyx_kp_u_ADD_HL_SP;
  PyObject *__pyx_kp_u_ADD_SP_r8;
  PyObject *__pyx_kp_u_AND_A;
  PyObject *__pyx_kp_u_AND_B;
  PyObject *__pyx_kp_u_AND_C;
  PyObject *__pyx_kp_u_AND_D;
  PyObject *__pyx_kp_u_AND_E;
  PyObject *__pyx_kp_u_AND_H;
  PyObject *__pyx_kp_u_AND_HL;
  PyObject *__pyx_kp_u_AND_L;
  PyObject *__pyx_kp_u_AND_d8;
  PyObject *__pyx_n_s_ASCII;
  PyObject *__pyx_kp_s_All_dimensions_preceding_dimensi;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_u_B;
  PyObject *__pyx_kp_u_BIT_0_A;
  PyObject *__pyx_kp_u_BIT_0_B;
  PyObject *__pyx_kp_u_BIT_0_C;
  PyObject *__pyx_kp_u_BIT_0_D;
  PyObject *__pyx_kp_u_BIT_0_E;
  PyObject *__pyx_kp_u_BIT_0_H;
  PyObject *__pyx_kp_u_BIT_0_HL;
  PyObject *__pyx_kp_u_BIT_0_L;
  PyObject *__pyx_kp_u_BIT_1_A;
  PyObject *__pyx_kp_u_BIT_1_B;
  PyObject *__pyx_kp_u_BIT_1_C;
  PyObject *__pyx_kp_u_BIT_1_D;
  PyObject *__pyx_kp_u_BIT_1_E;
  PyObject *__pyx_kp_u_BIT_1_H;
  PyObject *__pyx_kp_u_BIT_1_HL;
  PyObject *__pyx_kp_u_BIT_1_L;
  PyObject *__pyx_kp_u_BIT_2_A;
  PyObject *__pyx_kp_u_BIT_2_B;
  PyObject *__pyx_kp_u_BIT_2_C;
  PyObject *__pyx_kp_u_BIT_2_D;
  PyObject *__pyx_kp_u_BIT_2_E;
  PyObject *__pyx_kp_u_BIT_2_H;
  PyObject *__pyx_kp_u_BIT_2_HL;
  PyObject *__pyx_kp_u_BIT_2_L;
  PyObject *__pyx_kp_u_BIT_3_A;
  PyObject *__pyx_kp_u_BIT_3_B;
  PyObject *__pyx_kp_u_BIT_3_C;
  PyObject *__pyx_kp_u_BIT_3_D;
  PyObject *__pyx_kp_u_BIT_3_E;
  PyObject *__pyx_kp_u_BIT_3_H;
  PyObject *__pyx_kp_u_BIT_3_HL;
  PyObject *__pyx_kp_u_BIT_3_L;
  PyObject *__pyx_kp_u_BIT_4_A;
  PyObject *__pyx_kp_u_BIT_4_B;
  PyObject *__pyx_kp_u_BIT_4_C;
  PyObject *__pyx_kp_u_BIT_4_D;
  PyObject *__pyx_kp_u_BIT_4_E;
  PyObject *__pyx_kp_u_BIT_4_H;
  PyObject *__pyx_kp_u_BIT_4_HL;
  PyObject *__pyx_kp_u_BIT_4_L;
  PyObject *__pyx_kp_u_BIT_5_A;
  PyObject *__pyx_kp_u_BIT_5_B;
  PyObject *__pyx_kp_u_BIT_5_C;
  PyObject *__pyx_kp_u_BIT_5_D;
  PyObject *__pyx_kp_u_BIT_5_E;
  PyObject *__pyx_kp_u_BIT_5_H;
  PyObject *__pyx_kp_u_BIT_5_HL;
  PyObject *__pyx_kp_u_BIT_5_L;
  PyObject *__pyx_kp_u_BIT_6_A;
  PyObject *__pyx_kp_u_BIT_6_B;
  PyObject *__pyx_kp_u_BIT_6_C;
  PyObject *__pyx_kp_u_BIT_6_D;
  PyObject *__pyx_kp_u_BIT_6_E;
  PyObject *__pyx_kp_u_BIT_6_H;
  PyObject *__pyx_kp_u_BIT_6_HL;
  PyObject *__pyx_kp_u_BIT_6_L;
  PyObject *__pyx_kp_u_BIT_7_A;
  PyObject *__pyx_kp_u_BIT_7_B;
  PyObject *__pyx_kp_u_BIT_7_C;
  PyObject *__pyx_kp_u_BIT_7_D;
  PyObject *__pyx_kp_u_BIT_7_E;
  PyObject *__pyx_kp_u_BIT_7_H;
  PyObject *__pyx_kp_u_BIT_7_HL;
  PyObject *__pyx_kp_u_BIT_7_L;
  PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
  PyObject *__pyx_kp_u_CALL_C_a16;
  PyObject *__pyx_kp_u_CALL_NC_a16;
  PyObject *__pyx_kp_u_CALL_NZ_a16;
  PyObject *__pyx_kp_u_CALL_Z_a16;
  PyObject *__pyx_kp_u_CALL_a16;
  PyObject *__pyx_kp_u_CB_cannot_be_called;
  PyObject *__pyx_n_u_CCF;
  PyObject *__pyx_n_u_CPL;
  PyObject *__pyx_n_s_CPU_COMMANDS;
  PyObject *__pyx_kp_u_CP_A;
  PyObject *__pyx_kp_u_CP_B;
  PyObject *__pyx_kp_u_CP_C;
  PyObject *__pyx_kp_u_CP_D;
  PyObject *__pyx_kp_u_CP_E;
  PyObject *__pyx_kp_u_CP_H;
  PyObject *__pyx_kp_u_CP_HL;
  PyObject *__pyx_kp_u_CP_L;
  PyObject *__pyx_kp_u_CP_d8;
  PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
  PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
  PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
  PyObject *__pyx_kp_u_Cannot_index_with_type;
  PyObject *__pyx_kp_s_Cannot_transpose_memoryview_with;
  PyObject *__pyx_n_u_DAA;
  PyObject *__pyx_kp_u_DEC_A;
  PyObject *__pyx_kp_u_DEC_B;
  PyObject *__pyx_kp_u_DEC_BC;
  PyObject *__pyx_kp_u_DEC_C;
  PyObject *__pyx_kp_u_DEC_D;
  PyObject *__pyx_kp_u_DEC_DE;
  PyObject *__pyx_kp_u_DEC_E;
  PyObject *__pyx_kp_u_DEC_H;
  PyObject *__pyx_kp_u_DEC_HL;
  PyObject *__pyx_kp_u_DEC_HL_2;
  PyObject *__pyx_kp_u_DEC_L;
  PyObject *__pyx_kp_u_DEC_SP;
  PyObject *__pyx_n_u_DI;
  PyObject *__pyx_kp_s_Dimension_d_is_not_direct;
  PyObject *__pyx_n_u_EI;
  PyObject *__pyx_n_s_Ellipsis;
  PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
  PyObject *__pyx_n_s_FLAGC;
  PyObject *__pyx_n_s_FLAGH;
  PyObject *__pyx_n_s_FLAGN;
  PyObject *__pyx_n_s_FLAGZ;
  PyObject *__pyx_n_u_HALT;
  PyObject *__pyx_kp_u_INC_A;
  PyObject *__pyx_kp_u_INC_B;
  PyObject *__pyx_kp_u_INC_BC;
  PyObject *__pyx_kp_u_INC_C;
  PyObject *__pyx_kp_u_INC_D;
  PyObject *__pyx_kp_u_INC_DE;
  PyObject *__pyx_kp_u_INC_E;
  PyObject *__pyx_kp_u_INC_H;
  PyObject *__pyx_kp_u_INC_HL;
  PyObject *__pyx_kp_u_INC_HL_2;
  PyObject *__pyx_kp_u_INC_L;
  PyObject *__pyx_kp_u_INC_SP;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_n_s_IndexError;
  PyObject *__pyx_kp_s_Index_out_of_bounds_axis_d;
  PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
  PyObject *__pyx_kp_u_Invalid_mode_expected_c_or_fortr;
  PyObject *__pyx_kp_u_Invalid_shape_in_axis;
  PyObject *__pyx_kp_u_JP_C_a16;
  PyObject *__pyx_kp_u_JP_HL;
  PyObject *__pyx_kp_u_JP_NC_a16;
  PyObject *__pyx_kp_u_JP_NZ_a16;
  PyObject *__pyx_kp_u_JP_Z_a16;
  PyObject *__pyx_kp_u_JP_a16;
  PyObject *__pyx_kp_u_JR_C_r8;
  PyObject *__pyx_kp_u_JR_NC_r8;
  PyObject *__pyx_kp_u_JR_NZ_r8;
  PyObject *__pyx_kp_u_JR_Z_r8;
  PyObject *__pyx_kp_u_JR_r8;
  PyObject *__pyx_kp_u_LDH_A_a8;
  PyObject *__pyx_kp_u_LDH_a8_A;
  PyObject *__pyx_kp_u_LD_A_A;
  PyObject *__pyx_kp_u_LD_A_B;
  PyObject *__pyx_kp_u_LD_A_BC;
  PyObject *__pyx_kp_u_LD_A_C;
  PyObject *__pyx_kp_u_LD_A_C_2;
  PyObject *__pyx_kp_u_LD_A_D;
  PyObject *__pyx_kp_u_LD_A_DE;
  PyObject *__pyx_kp_u_LD_A_E;
  PyObject *__pyx_kp_u_LD_A_H;
  PyObject *__pyx_kp_u_LD_A_HL;
  PyObject *__pyx_kp_u_LD_A_HL_2;
  PyObject *__pyx_kp_u_LD_A_HL_3;
  PyObject *__pyx_kp_u_LD_A_L;
  PyObject *__pyx_kp_u_LD_A_a16;
  PyObject *__pyx_kp_u_LD_A_d8;
  PyObject *__pyx_kp_u_LD_BC_A;
  PyObject *__pyx_kp_u_LD_BC_d16;
  PyObject *__pyx_kp_u_LD_B_A;
  PyObject *__pyx_kp_u_LD_B_B;
  PyObject *__pyx_kp_u_LD_B_C;
  PyObject *__pyx_kp_u_LD_B_D;
  PyObject *__pyx_kp_u_LD_B_E;
  PyObject *__pyx_kp_u_LD_B_H;
  PyObject *__pyx_kp_u_LD_B_HL;
  PyObject *__pyx_kp_u_LD_B_L;
  PyObject *__pyx_kp_u_LD_B_d8;
  PyObject *__pyx_kp_u_LD_C_A;
  PyObject *__pyx_kp_u_LD_C_A_2;
  PyObject *__pyx_kp_u_LD_C_B;
  PyObject *__pyx_kp_u_LD_C_C;
  PyObject *__pyx_kp_u_LD_C_D;
  PyObject *__pyx_kp_u_LD_C_E;
  PyObject *__pyx_kp_u_LD_C_H;
  PyObject *__pyx_kp_u_LD_C_HL;
  PyObject *__pyx_kp_u_LD_C_L;
  PyObject *__pyx_kp_u_LD_C_d8;
  PyObject *__pyx_kp_u_LD_DE_A;
  PyObject *__pyx_kp_u_LD_DE_d16;
  PyObject *__pyx_kp_u_LD_D_A;
  PyObject *__pyx_kp_u_LD_D_B;
  PyObject *__pyx_kp_u_LD_D_C;
  PyObject *__pyx_kp_u_LD_D_D;
  PyObject *__pyx_kp_u_LD_D_E;
  PyObject *__pyx_kp_u_LD_D_H;
  PyObject *__pyx_kp_u_LD_D_HL;
  PyObject *__pyx_kp_u_LD_D_L;
  PyObject *__pyx_kp_u_LD_D_d8;
  PyObject *__pyx_kp_u_LD_E_A;
  PyObject *__pyx_kp_u_LD_E_B;
  PyObject *__pyx_kp_u_LD_E_C;
  PyObject *__pyx_kp_u_LD_E_D;
  PyObject *__pyx_kp_u_LD_E_E;
  PyObject *__pyx_kp_u_LD_E_H;
  PyObject *__pyx_kp_u_LD_E_HL;
  PyObject *__pyx_kp_u_LD_E_L;
  PyObject *__pyx_kp_u_LD_E_d8;
  PyObject *__pyx_kp_u_LD_HL_A;
  PyObject *__pyx_kp_u_LD_HL_A_2;
  PyObject *__pyx_kp_u_LD_HL_A_3;
  PyObject *__pyx_kp_u_LD_HL_B;
  PyObject *__pyx_kp_u_LD_HL_C;
  PyObject *__pyx_kp_u_LD_HL_D;
  PyObject *__pyx_kp_u_LD_HL_E;
  PyObject *__pyx_kp_u_LD_HL_H;
  PyObject *__pyx_kp_u_LD_HL_L;
  PyObject *__pyx_kp_u_LD_HL_SP_r8;
  PyObject *__pyx_kp_u_LD_HL_d16;
  PyObject *__pyx_kp_u_LD_HL_d8;
  PyObject *__pyx_kp_u_LD_H_A;
  PyObject *__pyx_kp_u_LD_H_B;
  PyObject *__pyx_kp_u_LD_H_C;
  PyObject *__pyx_kp_u_LD_H_D;
  PyObject *__pyx_kp_u_LD_H_E;
  PyObject *__pyx_kp_u_LD_H_H;
  PyObject *__pyx_kp_u_LD_H_HL;
  PyObject *__pyx_kp_u_LD_H_L;
  PyObject *__pyx_kp_u_LD_H_d8;
  PyObject *__pyx_kp_u_LD_L_A;
  PyObject *__pyx_kp_u_LD_L_B;
  PyObject *__pyx_kp_u_LD_L_C;
  PyObject *__pyx_kp_u_LD_L_D;
  PyObject *__pyx_kp_u_LD_L_E;
  PyObject *__pyx_kp_u_LD_L_H;
  PyObject *__pyx_kp_u_LD_L_HL;
  PyObject *__pyx_kp_u_LD_L_L;
  PyObject *__pyx_kp_u_LD_L_d8;
  PyObject *__pyx_kp_u_LD_SP_HL;
  PyObject *__pyx_kp_u_LD_SP_d16;
  PyObject *__pyx_kp_u_LD_a16_A;
  PyObject *__pyx_kp_u_LD_a16_SP;
  PyObject *__pyx_n_s_MemoryError;
  PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
  PyObject *__pyx_kp_s_MemoryView_of_r_object;
  PyObject *__pyx_n_u_NOP;
  PyObject *__pyx_n_b_O;
  PyObject *__pyx_n_s_OPCODE_LENGTHS;
  PyObject *__pyx_kp_u_OR_A;
  PyObject *__pyx_kp_u_OR_B;
  PyObject *__pyx_kp_u_OR_C;
  PyObject *__pyx_kp_u_OR_D;
  PyObject *__pyx_kp_u_OR_E;
  PyObject *__pyx_kp_u_OR_H;
  PyObject *__pyx_kp_u_OR_HL;
  PyObject *__pyx_kp_u_OR_L;
  PyObject *__pyx_kp_u_OR_d8;
  PyObject *__pyx_kp_u_Out_of_bounds_on_buffer_access_a;
  PyObject *__pyx_n_s_OverflowError;
  PyObject *__pyx_kp_u_POP_AF;
  PyObject *__pyx_kp_u_POP_BC;
  PyObject *__pyx_kp_u_POP_DE;
  PyObject *__pyx_kp_u_POP_HL;
  PyObject *__pyx_kp_u_PREFIX_CB;
  PyObject *__pyx_kp_u_PUSH_AF;
  PyObject *__pyx_kp_u_PUSH_BC;
  PyObject *__pyx_kp_u_PUSH_DE;
  PyObject *__pyx_kp_u_PUSH_HL;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_kp_u_RES_0_A;
  PyObject *__pyx_kp_u_RES_0_B;
  PyObject *__pyx_kp_u_RES_0_C;
  PyObject *__pyx_kp_u_RES_0_D;
  PyObject *__pyx_kp_u_RES_0_E;
  PyObject *__pyx_kp_u_RES_0_H;
  PyObject *__pyx_kp_u_RES_0_HL;
  PyObject *__pyx_kp_u_RES_0_L;
  PyObject *__pyx_kp_u_RES_1_A;
  PyObject *__pyx_kp_u_RES_1_B;
  PyObject *__pyx_kp_u_RES_1_C;
  PyObject *__pyx_kp_u_RES_1_D;
  PyObject *__pyx_kp_u_RES_1_E;
  PyObject *__pyx_kp_u_RES_1_H;
  PyObject *__pyx_kp_u_RES_1_HL;
  PyObject *__pyx_kp_u_RES_1_L;
  PyObject *__pyx_kp_u_RES_2_A;
  PyObject *__pyx_kp_u_RES_2_B;
  PyObject *__pyx_kp_u_RES_2_C;
  PyObject *__pyx_kp_u_RES_2_D;
  PyObject *__pyx_kp_u_RES_2_E;
  PyObject *__pyx_kp_u_RES_2_H;
  PyObject *__pyx_kp_u_RES_2_HL;
  PyObject *__pyx_kp_u_RES_2_L;
  PyObject *__pyx_kp_u_RES_3_A;
  PyObject *__pyx_kp_u_RES_3_B;
  PyObject *__pyx_kp_u_RES_3_C;
  PyObject *__pyx_kp_u_RES_3_D;
  PyObject *__pyx_kp_u_RES_3_E;
  PyObject *__pyx_kp_u_RES_3_H;
  PyObject *__pyx_kp_u_RES_3_HL;
  PyObject *__pyx_kp_u_RES_3_L;
  PyObject *__pyx_kp_u_RES_4_A;
  PyObject *__pyx_kp_u_RES_4_B;
  PyObject *__pyx_kp_u_RES_4_C;
  PyObject *__pyx_kp_u_RES_4_D;
  PyObject *__pyx_kp_u_RES_4_E;
  PyObject *__pyx_kp_u_RES_4_H;
  PyObject *__pyx_kp_u_RES_4_HL;
  PyObject *__pyx_kp_u_RES_4_L;
  PyObject *__pyx_kp_u_RES_5_A;
  PyObject *__pyx_kp_u_RES_5_B;
  PyObject *__pyx_kp_u_RES_5_C;
  PyObject *__pyx_kp_u_RES_5_D;
  PyObject *__pyx_kp_u_RES_5_E;
  PyObject *__pyx_kp_u_RES_5_H;
  PyObject *__pyx_kp_u_RES_5_HL;
  PyObject *__pyx_kp_u_RES_5_L;
  PyObject *__pyx_kp_u_RES_6_A;
  PyObject *__pyx_kp_u_RES_6_B;
  PyObject *__pyx_kp_u_RES_6_C;
  PyObject *__pyx_kp_u_RES_6_D;
  PyObject *__pyx_kp_u_RES_6_E;
  PyObject *__pyx_kp_u_RES_6_H;
  PyObject *__pyx_kp_u_RES_6_HL;
  PyObject *__pyx_kp_u_RES_6_L;
  PyObject *__pyx_kp_u_RES_7_A;
  PyObject *__pyx_kp_u_RES_7_B;
  PyObject *__pyx_kp_u_RES_7_C;
  PyObject *__pyx_kp_u_RES_7_D;
  PyObject *__pyx_kp_u_RES_7_E;
  PyObject *__pyx_kp_u_RES_7_H;
  PyObject *__pyx_kp_u_RES_7_HL;
  PyObject *__pyx_kp_u_RES_7_L;
  PyObject *__pyx_n_u_RET;
  PyObject *__pyx_n_u_RETI;
  PyObject *__pyx_kp_u_RET_C;
  PyObject *__pyx_kp_u_RET_NC;
  PyObject *__pyx_kp_u_RET_NZ;
  PyObject *__pyx_kp_u_RET_Z;
  PyObject *__pyx_n_u_RLA;
  PyObject *__pyx_n_u_RLCA;
  PyObject *__pyx_kp_u_RLC_A;
  PyObject *__pyx_kp_u_RLC_B;
  PyObject *__pyx_kp_u_RLC_C;
  PyObject *__pyx_kp_u_RLC_D;
  PyObject *__pyx_kp_u_RLC_E;
  PyObject *__pyx_kp_u_RLC_H;
  PyObject *__pyx_kp_u_RLC_HL;
  PyObject *__pyx_kp_u_RLC_L;
  PyObject *__pyx_kp_u_RL_A;
  PyObject *__pyx_kp_u_RL_B;
  PyObject *__pyx_kp_u_RL_C;
  PyObject *__pyx_kp_u_RL_D;
  PyObject *__pyx_kp_u_RL_E;
  PyObject *__pyx_kp_u_RL_H;
  PyObject *__pyx_kp_u_RL_HL;
  PyObject *__pyx_kp_u_RL_L;
  PyObject *__pyx_n_u_RRA;
  PyObject *__pyx_n_u_RRCA;
  PyObject *__pyx_kp_u_RRC_A;
  PyObject *__pyx_kp_u_RRC_B;
  PyObject *__pyx_kp_u_RRC_C;
  PyObject *__pyx_kp_u_RRC_D;
  PyObject *__pyx_kp_u_RRC_E;
  PyObject *__pyx_kp_u_RRC_H;
  PyObject *__pyx_kp_u_RRC_HL;
  PyObject *__pyx_kp_u_RRC_L;
  PyObject *__pyx_kp_u_RR_A;
  PyObject *__pyx_kp_u_RR_B;
  PyObject *__pyx_kp_u_RR_C;
  PyObject *__pyx_kp_u_RR_D;
  PyObject *__pyx_kp_u_RR_E;
  PyObject *__pyx_kp_u_RR_H;
  PyObject *__pyx_kp_u_RR_HL;
  PyObject *__pyx_kp_u_RR_L;
  PyObject *__pyx_kp_u_RST_00H;
  PyObject *__pyx_kp_u_RST_08H;
  PyObject *__pyx_kp_u_RST_10H;
  PyObject *__pyx_kp_u_RST_18H;
  PyObject *__pyx_kp_u_RST_20H;
  PyObject *__pyx_kp_u_RST_28H;
  PyObject *__pyx_kp_u_RST_30H;
  PyObject *__pyx_kp_u_RST_38H;
  PyObject *__pyx_n_s_RuntimeError;
  PyObject *__pyx_kp_u_SBC_A_A;
  PyObject *__pyx_kp_u_SBC_A_B;
  PyObject *__pyx_kp_u_SBC_A_C;
  PyObject *__pyx_kp_u_SBC_A_D;
  PyObject *__pyx_kp_u_SBC_A_E;
  PyObject *__pyx_kp_u_SBC_A_H;
  PyObject *__pyx_kp_u_SBC_A_HL;
  PyObject *__pyx_kp_u_SBC_A_L;
  PyObject *__pyx_kp_u_SBC_A_d8;
  PyObject *__pyx_n_u_SCF;
  PyObject *__pyx_kp_u_SET_0_A;
  PyObject *__pyx_kp_u_SET_0_B;
  PyObject *__pyx_kp_u_SET_0_C;
  PyObject *__pyx_kp_u_SET_0_D;
  PyObject *__pyx_kp_u_SET_0_E;
  PyObject *__pyx_kp_u_SET_0_H;
  PyObject *__pyx_kp_u_SET_0_HL;
  PyObject *__pyx_kp_u_SET_0_L;
  PyObject *__pyx_kp_u_SET_1_A;
  PyObject *__pyx_kp_u_SET_1_B;
  PyObject *__pyx_kp_u_SET_1_C;
  PyObject *__pyx_kp_u_SET_1_D;
  PyObject *__pyx_kp_u_SET_1_E;
  PyObject *__pyx_kp_u_SET_1_H;
  PyObject *__pyx_kp_u_SET_1_HL;
  PyObject *__pyx_kp_u_SET_1_L;
  PyObject *__pyx_kp_u_SET_2_A;
  PyObject *__pyx_kp_u_SET_2_B;
  PyObject *__pyx_kp_u_SET_2_C;
  PyObject *__pyx_kp_u_SET_2_D;
  PyObject *__pyx_kp_u_SET_2_E;
  PyObject *__pyx_kp_u_SET_2_H;
  PyObject *__pyx_kp_u_SET_2_HL;
  PyObject *__pyx_kp_u_SET_2_L;
  PyObject *__pyx_kp_u_SET_3_A;
  PyObject *__pyx_kp_u_SET_3_B;
  PyObject *__pyx_kp_u_SET_3_C;
  PyObject *__pyx_kp_u_SET_3_D;
  PyObject *__pyx_kp_u_SET_3_E;
  PyObject *__pyx_kp_u_SET_3_H;
  PyObject *__pyx_kp_u_SET_3_HL;
  PyObject *__pyx_kp_u_SET_3_L;
  PyObject *__pyx_kp_u_SET_4_A;
  PyObject *__pyx_kp_u_SET_4_B;
  PyObject *__pyx_kp_u_SET_4_C;
  PyObject *__pyx_kp_u_SET_4_D;
  PyObject *__pyx_kp_u_SET_4_E;
  PyObject *__pyx_kp_u_SET_4_H;
  PyObject *__pyx_kp_u_SET_4_HL;
  PyObject *__pyx_kp_u_SET_4_L;
  PyObject *__pyx_kp_u_SET_5_A;
  PyObject *__pyx_kp_u_SET_5_B;
  PyObject *__pyx_kp_u_SET_5_C;
  PyObject *__pyx_kp_u_SET_5_D;
  PyObject *__pyx_kp_u_SET_5_E;
  PyObject *__pyx_kp_u_SET_5_H;
  PyObject *__pyx_kp_u_SET_5_HL;
  PyObject *__pyx_kp_u_SET_5_L;
  PyObject *__pyx_kp_u_SET_6_A;
  PyObject *__pyx_kp_u_SET_6_B;
  PyObject *__pyx_kp_u_SET_6_C;
  PyObject *__pyx_kp_u_SET_6_D;
  PyObject *__pyx_kp_u_SET_6_E;
  PyObject *__pyx_kp_u_SET_6_H;
  PyObject *__pyx_kp_u_SET_6_HL;
  PyObject *__pyx_kp_u_SET_6_L;
  PyObject *__pyx_kp_u_SET_7_A;
  PyObject *__pyx_kp_u_SET_7_B;
  PyObject *__pyx_kp_u_SET_7_C;
  PyObject *__pyx_kp_u_SET_7_D;
  PyObject *__pyx_kp_u_SET_7_E;
  PyObject *__pyx_kp_u_SET_7_H;
  PyObject *__pyx_kp_u_SET_7_HL;
  PyObject *__pyx_kp_u_SET_7_L;
  PyObject *__pyx_kp_u_SLA_A;
  PyObject *__pyx_kp_u_SLA_B;
  PyObject *__pyx_kp_u_SLA_C;
  PyObject *__pyx_kp_u_SLA_D;
  PyObject *__pyx_kp_u_SLA_E;
  PyObject *__pyx_kp_u_SLA_H;
  PyObject *__pyx_kp_u_SLA_HL;
  PyObject *__pyx_kp_u_SLA_L;
  PyObject *__pyx_kp_u_SRA_A;
  PyObject *__pyx_kp_u_SRA_B;
  PyObject *__pyx_kp_u_SRA_C;
  PyObject *__pyx_kp_u_SRA_D;
  PyObject *__pyx_kp_u_SRA_E;
  PyObject *__pyx_kp_u_SRA_H;
  PyObject *__pyx_kp_u_SRA_HL;
  PyObject *__pyx_kp_u_SRA_L;
  PyObject *__pyx_kp_u_SRL_A;
  PyObject *__pyx_kp_u_SRL_B;
  PyObject *__pyx_kp_u_SRL_C;
  PyObject *__pyx_kp_u_SRL_D;
  PyObject *__pyx_kp_u_SRL_E;
  PyObject *__pyx_kp_u_SRL_H;
  PyObject *__pyx_kp_u_SRL_HL;
  PyObject *__pyx_kp_u_SRL_L;
  PyObject *__pyx_kp_u_STOP_0;
  PyObject *__pyx_kp_u_SUB_A;
  PyObject *__pyx_kp_u_SUB_B;
  PyObject *__pyx_kp_u_SUB_C;
  PyObject *__pyx_kp_u_SUB_D;
  PyObject *__pyx_kp_u_SUB_E;
  PyObject *__pyx_kp_u_SUB_H;
  PyObject *__pyx_kp_u_SUB_HL;
  PyObject *__pyx_kp_u_SUB_L;
  PyObject *__pyx_kp_u_SUB_d8;
  PyObject *__pyx_kp_u_SWAP_A;
  PyObject *__pyx_kp_u_SWAP_B;
  PyObject *__pyx_kp_u_SWAP_C;
  PyObject *__pyx_kp_u_SWAP_D;
  PyObject *__pyx_kp_u_SWAP_E;
  PyObject *__pyx_kp_u_SWAP_H;
  PyObject *__pyx_kp_u_SWAP_HL;
  PyObject *__pyx_kp_u_SWAP_L;
  PyObject *__pyx_n_s_Sequence;
  PyObject *__pyx_kp_s_Step_may_not_be_zero_axis_d;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s_View_MemoryView;
  PyObject *__pyx_kp_u_XOR_A;
  PyObject *__pyx_kp_u_XOR_B;
  PyObject *__pyx_kp_u_XOR_C;
  PyObject *__pyx_kp_u_XOR_D;
  PyObject *__pyx_kp_u_XOR_E;
  PyObject *__pyx_kp_u_XOR_H;
  PyObject *__pyx_kp_u_XOR_HL;
  PyObject *__pyx_kp_u_XOR_L;
  PyObject *__pyx_kp_u_XOR_d8;
  PyObject *__pyx_kp_u__2;
  PyObject *__pyx_kp_u__21;
  PyObject *__pyx_n_s__22;
  PyObject *__pyx_n_s__3;
  PyObject *__pyx_kp_u__6;
  PyObject *__pyx_kp_u__7;
  PyObject *__pyx_n_s_abc;
  PyObject *__pyx_n_s_allocate_buffer;
  PyObject *__pyx_kp_u_and;
  PyObject *__pyx_n_s_array;
  PyObject *__pyx_n_s_base;
  PyObject *__pyx_n_s_c;
  PyObject *__pyx_n_u_c;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_kp_s_collections_abc;
  PyObject *__pyx_kp_s_contiguous_and_direct;
  PyObject *__pyx_kp_s_contiguous_and_indirect;
  PyObject *__pyx_n_s_count;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_dtype_is_object;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_error;
  PyObject *__pyx_n_s_flags;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_s_fortran;
  PyObject *__pyx_n_u_fortran;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_get_logger;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_kp_u_got;
  PyObject *__pyx_kp_u_got_differing_extents_in_dimensi;
  PyObject *__pyx_n_s_id;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_itemsize;
  PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
  PyObject *__pyx_n_s_logger;
  PyObject *__pyx_n_s_logging;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_memview;
  PyObject *__pyx_n_s_mode;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_ndim;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_n_s_obj;
  PyObject *__pyx_n_s_pack;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_pyboy;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_capi;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_Enum;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_register;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_shape;
  PyObject *__pyx_n_s_size;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_s_step;
  PyObject *__pyx_n_s_stop;
  PyObject *__pyx_kp_s_strided_and_direct;
  PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
  PyObject *__pyx_kp_s_strided_and_indirect;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_struct;
  PyObject *__pyx_n_s_sys;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_kp_s_unable_to_allocate_array_data;
  PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
  PyObject *__pyx_n_s_unpack;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_utils;
  PyObject *__pyx_n_s_version_info;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_3;
  PyObject *__pyx_int_4;
  PyObject *__pyx_int_8;
  PyObject *__pyx_int_112105877;
  PyObject *__pyx_int_136983863;
  PyObject *__pyx_int_184977713;
  PyObject *__pyx_int_neg_1;
  PyObject *__pyx_slice__5;
  PyObject *__pyx_tuple__4;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__15;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__17;
  PyObject *__pyx_tuple__18;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_codeobj__19;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_7logging_7logging_Logger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_7bootrom_BootROM);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_5utils_IntIOInterface);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_5utils_IntIOWrapper);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_5utils_WindowEvent);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_5utils_WindowEventMouse);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_9cartridge_3rtc_RTC);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_BaseMBC);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_ROMOnly);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_11interaction_Interaction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_5array_array);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_LCD);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_PaletteRegister);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_STATRegister);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_LCDCRegister);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_Renderer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_CGBLCD);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_CGBRenderer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_VBKregister);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_PaletteIndexRegister);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3lcd_PaletteColorRegister);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3ram_RAM);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_5sound_Sound);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_5sound_ToneChannel);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_5sound_SweepChannel);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_5sound_WaveChannel);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_5sound_NoiseChannel);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_5timer_Timer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_2mb_Motherboard);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_2mb_HDMA);
  Py_CLEAR(clear_module_state->__pyx_ptype_5pyboy_4core_3cpu_CPU);
  Py_CLEAR(clear_module_state->__pyx_array_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_array);
  Py_CLEAR(clear_module_state->__pyx_MemviewEnum_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_MemviewEnum);
  Py_CLEAR(clear_module_state->__pyx_memoryview_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_memoryview);
  Py_CLEAR(clear_module_state->__pyx_memoryviewslice_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_memoryviewslice);
  Py_CLEAR(clear_module_state->__pyx_kp_u_);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADC_A_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADC_A_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADC_A_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADC_A_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADC_A_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADC_A_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADC_A_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADC_A_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADC_A_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_A_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_A_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_A_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_A_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_A_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_A_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_A_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_A_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_A_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_HL_BC);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_HL_DE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_HL_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_HL_SP);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ADD_SP_r8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AND_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AND_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AND_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AND_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AND_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AND_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AND_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AND_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AND_d8);
  Py_CLEAR(clear_module_state->__pyx_n_s_ASCII);
  Py_CLEAR(clear_module_state->__pyx_kp_s_All_dimensions_preceding_dimensi);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_u_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_0_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_0_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_0_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_0_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_0_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_0_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_0_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_0_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_1_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_1_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_1_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_1_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_1_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_1_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_1_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_1_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_2_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_2_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_2_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_2_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_2_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_2_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_2_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_2_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_3_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_3_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_3_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_3_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_3_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_3_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_3_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_3_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_4_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_4_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_4_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_4_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_4_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_4_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_4_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_4_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_5_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_5_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_5_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_5_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_5_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_5_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_5_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_5_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_6_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_6_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_6_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_6_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_6_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_6_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_6_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_6_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_7_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_7_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_7_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_7_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_7_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_7_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_7_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_BIT_7_L);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Buffer_view_does_not_expose_stri);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CALL_C_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CALL_NC_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CALL_NZ_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CALL_Z_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CALL_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CB_cannot_be_called);
  Py_CLEAR(clear_module_state->__pyx_n_u_CCF);
  Py_CLEAR(clear_module_state->__pyx_n_u_CPL);
  Py_CLEAR(clear_module_state->__pyx_n_s_CPU_COMMANDS);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CP_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CP_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CP_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CP_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CP_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CP_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CP_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CP_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_CP_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Can_only_create_a_buffer_that_is);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_assign_to_read_only_memor);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_create_writable_memory_vi);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_index_with_type);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_transpose_memoryview_with);
  Py_CLEAR(clear_module_state->__pyx_n_u_DAA);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_BC);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_DE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_HL_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_DEC_SP);
  Py_CLEAR(clear_module_state->__pyx_n_u_DI);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Dimension_d_is_not_direct);
  Py_CLEAR(clear_module_state->__pyx_n_u_EI);
  Py_CLEAR(clear_module_state->__pyx_n_s_Ellipsis);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Empty_shape_tuple_for_cython_arr);
  Py_CLEAR(clear_module_state->__pyx_n_s_FLAGC);
  Py_CLEAR(clear_module_state->__pyx_n_s_FLAGH);
  Py_CLEAR(clear_module_state->__pyx_n_s_FLAGN);
  Py_CLEAR(clear_module_state->__pyx_n_s_FLAGZ);
  Py_CLEAR(clear_module_state->__pyx_n_u_HALT);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_BC);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_DE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_HL_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_INC_SP);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Index_out_of_bounds_axis_d);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Indirect_dimensions_not_supporte);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_mode_expected_c_or_fortr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_shape_in_axis);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JP_C_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JP_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JP_NC_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JP_NZ_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JP_Z_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JP_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JR_C_r8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JR_NC_r8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JR_NZ_r8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JR_Z_r8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_JR_r8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LDH_A_a8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LDH_a8_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_BC);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_C_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_DE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_HL_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_HL_3);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_a16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_A_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_BC_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_BC_d16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_B_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_B_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_B_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_B_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_B_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_B_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_B_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_B_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_B_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_A_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_C_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_DE_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_DE_d16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_D_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_D_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_D_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_D_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_D_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_D_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_D_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_D_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_D_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_E_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_E_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_E_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_E_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_E_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_E_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_E_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_E_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_E_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_A_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_A_3);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_SP_r8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_d16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_HL_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_H_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_H_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_H_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_H_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_H_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_H_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_H_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_H_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_H_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_L_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_L_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_L_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_L_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_L_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_L_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_L_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_L_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_L_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_SP_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_SP_d16);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_a16_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_LD_a16_SP);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_MemoryView_of_r_at_0x_x);
  Py_CLEAR(clear_module_state->__pyx_kp_s_MemoryView_of_r_object);
  Py_CLEAR(clear_module_state->__pyx_n_u_NOP);
  Py_CLEAR(clear_module_state->__pyx_n_b_O);
  Py_CLEAR(clear_module_state->__pyx_n_s_OPCODE_LENGTHS);
  Py_CLEAR(clear_module_state->__pyx_kp_u_OR_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_OR_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_OR_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_OR_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_OR_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_OR_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_OR_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_OR_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_OR_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_OverflowError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_POP_AF);
  Py_CLEAR(clear_module_state->__pyx_kp_u_POP_BC);
  Py_CLEAR(clear_module_state->__pyx_kp_u_POP_DE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_POP_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_PREFIX_CB);
  Py_CLEAR(clear_module_state->__pyx_kp_u_PUSH_AF);
  Py_CLEAR(clear_module_state->__pyx_kp_u_PUSH_BC);
  Py_CLEAR(clear_module_state->__pyx_kp_u_PUSH_DE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_PUSH_HL);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_0_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_0_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_0_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_0_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_0_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_0_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_0_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_0_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_1_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_1_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_1_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_1_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_1_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_1_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_1_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_1_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_2_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_2_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_2_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_2_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_2_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_2_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_2_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_2_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_3_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_3_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_3_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_3_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_3_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_3_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_3_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_3_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_4_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_4_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_4_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_4_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_4_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_4_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_4_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_4_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_5_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_5_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_5_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_5_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_5_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_5_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_5_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_5_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_6_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_6_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_6_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_6_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_6_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_6_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_6_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_6_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_7_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_7_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_7_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_7_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_7_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_7_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_7_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RES_7_L);
  Py_CLEAR(clear_module_state->__pyx_n_u_RET);
  Py_CLEAR(clear_module_state->__pyx_n_u_RETI);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RET_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RET_NC);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RET_NZ);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RET_Z);
  Py_CLEAR(clear_module_state->__pyx_n_u_RLA);
  Py_CLEAR(clear_module_state->__pyx_n_u_RLCA);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RLC_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RLC_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RLC_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RLC_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RLC_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RLC_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RLC_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RLC_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RL_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RL_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RL_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RL_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RL_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RL_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RL_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RL_L);
  Py_CLEAR(clear_module_state->__pyx_n_u_RRA);
  Py_CLEAR(clear_module_state->__pyx_n_u_RRCA);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RRC_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RRC_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RRC_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RRC_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RRC_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RRC_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RRC_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RRC_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RR_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RR_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RR_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RR_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RR_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RR_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RR_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RR_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RST_00H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RST_08H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RST_10H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RST_18H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RST_20H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RST_28H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RST_30H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_RST_38H);
  Py_CLEAR(clear_module_state->__pyx_n_s_RuntimeError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SBC_A_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SBC_A_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SBC_A_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SBC_A_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SBC_A_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SBC_A_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SBC_A_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SBC_A_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SBC_A_d8);
  Py_CLEAR(clear_module_state->__pyx_n_u_SCF);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_0_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_0_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_0_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_0_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_0_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_0_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_0_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_0_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_1_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_1_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_1_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_1_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_1_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_1_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_1_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_1_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_2_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_2_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_2_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_2_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_2_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_2_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_2_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_2_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_3_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_3_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_3_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_3_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_3_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_3_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_3_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_3_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_4_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_4_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_4_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_4_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_4_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_4_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_4_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_4_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_5_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_5_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_5_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_5_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_5_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_5_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_5_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_5_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_6_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_6_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_6_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_6_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_6_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_6_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_6_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_6_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_7_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_7_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_7_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_7_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_7_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_7_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_7_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SET_7_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SLA_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SLA_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SLA_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SLA_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SLA_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SLA_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SLA_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SLA_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRA_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRA_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRA_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRA_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRA_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRA_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRA_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRA_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRL_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRL_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRL_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRL_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRL_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRL_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRL_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SRL_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_STOP_0);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SUB_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SUB_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SUB_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SUB_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SUB_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SUB_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SUB_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SUB_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SUB_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SWAP_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SWAP_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SWAP_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SWAP_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SWAP_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SWAP_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SWAP_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SWAP_L);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sequence);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Step_may_not_be_zero_axis_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Unable_to_convert_item_to_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s_View_MemoryView);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XOR_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XOR_B);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XOR_C);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XOR_D);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XOR_E);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XOR_H);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XOR_HL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XOR_L);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XOR_d8);
  Py_CLEAR(clear_module_state->__pyx_kp_u__2);
  Py_CLEAR(clear_module_state->__pyx_kp_u__21);
  Py_CLEAR(clear_module_state->__pyx_n_s__22);
  Py_CLEAR(clear_module_state->__pyx_n_s__3);
  Py_CLEAR(clear_module_state->__pyx_kp_u__6);
  Py_CLEAR(clear_module_state->__pyx_kp_u__7);
  Py_CLEAR(clear_module_state->__pyx_n_s_abc);
  Py_CLEAR(clear_module_state->__pyx_n_s_allocate_buffer);
  Py_CLEAR(clear_module_state->__pyx_kp_u_and);
  Py_CLEAR(clear_module_state->__pyx_n_s_array);
  Py_CLEAR(clear_module_state->__pyx_n_s_base);
  Py_CLEAR(clear_module_state->__pyx_n_s_c);
  Py_CLEAR(clear_module_state->__pyx_n_u_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_kp_s_collections_abc);
  Py_CLEAR(clear_module_state->__pyx_kp_s_contiguous_and_direct);
  Py_CLEAR(clear_module_state->__pyx_kp_s_contiguous_and_indirect);
  Py_CLEAR(clear_module_state->__pyx_n_s_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_dtype_is_object);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_error);
  Py_CLEAR(clear_module_state->__pyx_n_s_flags);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_fortran);
  Py_CLEAR(clear_module_state->__pyx_n_u_fortran);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got_differing_extents_in_dimensi);
  Py_CLEAR(clear_module_state->__pyx_n_s_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_itemsize);
  Py_CLEAR(clear_module_state->__pyx_kp_s_itemsize_0_for_cython_array);
  Py_CLEAR(clear_module_state->__pyx_n_s_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_logging);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_memview);
  Py_CLEAR(clear_module_state->__pyx_n_s_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_ndim);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_n_s_obj);
  Py_CLEAR(clear_module_state->__pyx_n_s_pack);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyboy);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_capi);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Enum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_register);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_shape);
  Py_CLEAR(clear_module_state->__pyx_n_s_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_s_step);
  Py_CLEAR(clear_module_state->__pyx_n_s_stop);
  Py_CLEAR(clear_module_state->__pyx_kp_s_strided_and_direct);
  Py_CLEAR(clear_module_state->__pyx_kp_s_strided_and_direct_or_indirect);
  Py_CLEAR(clear_module_state->__pyx_kp_s_strided_and_indirect);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_struct);
  Py_CLEAR(clear_module_state->__pyx_n_s_sys);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_kp_s_unable_to_allocate_array_data);
  Py_CLEAR(clear_module_state->__pyx_kp_s_unable_to_allocate_shape_and_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_utils);
  Py_CLEAR(clear_module_state->__pyx_n_s_version_info);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_3);
  Py_CLEAR(clear_module_state->__pyx_int_4);
  Py_CLEAR(clear_module_state->__pyx_int_8);
  Py_CLEAR(clear_module_state->__pyx_int_112105877);
  Py_CLEAR(clear_module_state->__pyx_int_136983863);
  Py_CLEAR(clear_module_state->__pyx_int_184977713);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_slice__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__4);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__15);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__17);
  Py_CLEAR(clear_module_state->__pyx_tuple__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_codeobj__19);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_7logging_7logging_Logger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_7bootrom_BootROM);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_5utils_IntIOInterface);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_5utils_IntIOWrapper);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_5utils_WindowEvent);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_5utils_WindowEventMouse);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_9cartridge_3rtc_RTC);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_BaseMBC);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_ROMOnly);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_11interaction_Interaction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_5array_array);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_LCD);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_PaletteRegister);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_STATRegister);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_LCDCRegister);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_Renderer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_CGBLCD);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_CGBRenderer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_VBKregister);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_PaletteIndexRegister);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3lcd_PaletteColorRegister);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3ram_RAM);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_5sound_Sound);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_5sound_ToneChannel);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_5sound_SweepChannel);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_5sound_WaveChannel);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_5sound_NoiseChannel);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_5timer_Timer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_2mb_Motherboard);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_2mb_HDMA);
  Py_VISIT(traverse_module_state->__pyx_ptype_5pyboy_4core_3cpu_CPU);
  Py_VISIT(traverse_module_state->__pyx_array_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_array);
  Py_VISIT(traverse_module_state->__pyx_MemviewEnum_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_MemviewEnum);
  Py_VISIT(traverse_module_state->__pyx_memoryview_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_memoryview);
  Py_VISIT(traverse_module_state->__pyx_memoryviewslice_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_memoryviewslice);
  Py_VISIT(traverse_module_state->__pyx_kp_u_);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADC_A_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADC_A_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADC_A_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADC_A_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADC_A_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADC_A_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADC_A_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADC_A_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADC_A_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_A_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_A_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_A_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_A_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_A_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_A_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_A_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_A_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_A_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_HL_BC);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_HL_DE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_HL_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_HL_SP);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ADD_SP_r8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AND_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AND_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AND_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AND_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AND_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AND_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AND_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AND_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AND_d8);
  Py_VISIT(traverse_module_state->__pyx_n_s_ASCII);
  Py_VISIT(traverse_module_state->__pyx_kp_s_All_dimensions_preceding_dimensi);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_u_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_0_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_0_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_0_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_0_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_0_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_0_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_0_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_0_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_1_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_1_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_1_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_1_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_1_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_1_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_1_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_1_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_2_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_2_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_2_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_2_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_2_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_2_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_2_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_2_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_3_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_3_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_3_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_3_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_3_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_3_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_3_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_3_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_4_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_4_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_4_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_4_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_4_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_4_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_4_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_4_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_5_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_5_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_5_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_5_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_5_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_5_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_5_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_5_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_6_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_6_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_6_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_6_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_6_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_6_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_6_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_6_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_7_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_7_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_7_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_7_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_7_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_7_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_7_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_BIT_7_L);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Buffer_view_does_not_expose_stri);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CALL_C_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CALL_NC_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CALL_NZ_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CALL_Z_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CALL_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CB_cannot_be_called);
  Py_VISIT(traverse_module_state->__pyx_n_u_CCF);
  Py_VISIT(traverse_module_state->__pyx_n_u_CPL);
  Py_VISIT(traverse_module_state->__pyx_n_s_CPU_COMMANDS);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CP_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CP_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CP_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CP_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CP_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CP_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CP_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CP_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_CP_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Can_only_create_a_buffer_that_is);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_assign_to_read_only_memor);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_create_writable_memory_vi);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_index_with_type);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_transpose_memoryview_with);
  Py_VISIT(traverse_module_state->__pyx_n_u_DAA);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_BC);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_DE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_HL_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_DEC_SP);
  Py_VISIT(traverse_module_state->__pyx_n_u_DI);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Dimension_d_is_not_direct);
  Py_VISIT(traverse_module_state->__pyx_n_u_EI);
  Py_VISIT(traverse_module_state->__pyx_n_s_Ellipsis);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Empty_shape_tuple_for_cython_arr);
  Py_VISIT(traverse_module_state->__pyx_n_s_FLAGC);
  Py_VISIT(traverse_module_state->__pyx_n_s_FLAGH);
  Py_VISIT(traverse_module_state->__pyx_n_s_FLAGN);
  Py_VISIT(traverse_module_state->__pyx_n_s_FLAGZ);
  Py_VISIT(traverse_module_state->__pyx_n_u_HALT);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_BC);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_DE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_HL_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_INC_SP);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Index_out_of_bounds_axis_d);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Indirect_dimensions_not_supporte);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_mode_expected_c_or_fortr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_shape_in_axis);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JP_C_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JP_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JP_NC_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JP_NZ_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JP_Z_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JP_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JR_C_r8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JR_NC_r8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JR_NZ_r8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JR_Z_r8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_JR_r8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LDH_A_a8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LDH_a8_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_BC);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_C_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_DE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_HL_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_HL_3);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_a16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_A_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_BC_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_BC_d16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_B_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_B_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_B_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_B_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_B_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_B_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_B_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_B_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_B_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_A_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_C_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_DE_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_DE_d16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_D_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_D_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_D_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_D_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_D_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_D_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_D_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_D_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_D_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_E_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_E_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_E_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_E_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_E_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_E_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_E_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_E_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_E_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_A_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_A_3);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_SP_r8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_d16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_HL_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_H_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_H_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_H_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_H_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_H_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_H_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_H_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_H_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_H_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_L_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_L_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_L_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_L_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_L_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_L_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_L_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_L_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_L_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_SP_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_SP_d16);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_a16_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_LD_a16_SP);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_MemoryView_of_r_at_0x_x);
  Py_VISIT(traverse_module_state->__pyx_kp_s_MemoryView_of_r_object);
  Py_VISIT(traverse_module_state->__pyx_n_u_NOP);
  Py_VISIT(traverse_module_state->__pyx_n_b_O);
  Py_VISIT(traverse_module_state->__pyx_n_s_OPCODE_LENGTHS);
  Py_VISIT(traverse_module_state->__pyx_kp_u_OR_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_OR_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_OR_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_OR_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_OR_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_OR_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_OR_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_OR_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_OR_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_OverflowError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_POP_AF);
  Py_VISIT(traverse_module_state->__pyx_kp_u_POP_BC);
  Py_VISIT(traverse_module_state->__pyx_kp_u_POP_DE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_POP_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_PREFIX_CB);
  Py_VISIT(traverse_module_state->__pyx_kp_u_PUSH_AF);
  Py_VISIT(traverse_module_state->__pyx_kp_u_PUSH_BC);
  Py_VISIT(traverse_module_state->__pyx_kp_u_PUSH_DE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_PUSH_HL);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_0_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_0_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_0_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_0_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_0_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_0_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_0_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_0_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_1_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_1_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_1_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_1_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_1_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_1_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_1_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_1_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_2_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_2_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_2_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_2_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_2_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_2_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_2_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_2_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_3_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_3_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_3_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_3_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_3_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_3_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_3_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_3_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_4_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_4_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_4_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_4_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_4_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_4_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_4_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_4_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_5_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_5_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_5_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_5_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_5_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_5_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_5_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_5_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_6_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_6_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_6_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_6_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_6_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_6_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_6_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_6_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_7_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_7_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_7_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_7_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_7_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_7_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_7_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RES_7_L);
  Py_VISIT(traverse_module_state->__pyx_n_u_RET);
  Py_VISIT(traverse_module_state->__pyx_n_u_RETI);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RET_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RET_NC);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RET_NZ);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RET_Z);
  Py_VISIT(traverse_module_state->__pyx_n_u_RLA);
  Py_VISIT(traverse_module_state->__pyx_n_u_RLCA);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RLC_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RLC_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RLC_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RLC_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RLC_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RLC_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RLC_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RLC_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RL_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RL_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RL_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RL_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RL_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RL_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RL_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RL_L);
  Py_VISIT(traverse_module_state->__pyx_n_u_RRA);
  Py_VISIT(traverse_module_state->__pyx_n_u_RRCA);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RRC_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RRC_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RRC_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RRC_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RRC_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RRC_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RRC_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RRC_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RR_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RR_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RR_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RR_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RR_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RR_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RR_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RR_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RST_00H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RST_08H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RST_10H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RST_18H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RST_20H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RST_28H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RST_30H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_RST_38H);
  Py_VISIT(traverse_module_state->__pyx_n_s_RuntimeError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SBC_A_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SBC_A_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SBC_A_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SBC_A_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SBC_A_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SBC_A_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SBC_A_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SBC_A_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SBC_A_d8);
  Py_VISIT(traverse_module_state->__pyx_n_u_SCF);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_0_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_0_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_0_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_0_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_0_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_0_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_0_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_0_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_1_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_1_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_1_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_1_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_1_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_1_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_1_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_1_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_2_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_2_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_2_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_2_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_2_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_2_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_2_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_2_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_3_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_3_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_3_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_3_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_3_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_3_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_3_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_3_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_4_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_4_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_4_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_4_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_4_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_4_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_4_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_4_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_5_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_5_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_5_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_5_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_5_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_5_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_5_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_5_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_6_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_6_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_6_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_6_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_6_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_6_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_6_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_6_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_7_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_7_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_7_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_7_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_7_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_7_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_7_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SET_7_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SLA_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SLA_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SLA_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SLA_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SLA_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SLA_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SLA_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SLA_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRA_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRA_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRA_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRA_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRA_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRA_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRA_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRA_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRL_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRL_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRL_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRL_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRL_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRL_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRL_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SRL_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_STOP_0);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SUB_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SUB_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SUB_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SUB_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SUB_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SUB_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SUB_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SUB_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SUB_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SWAP_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SWAP_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SWAP_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SWAP_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SWAP_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SWAP_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SWAP_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SWAP_L);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sequence);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Step_may_not_be_zero_axis_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Unable_to_convert_item_to_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s_View_MemoryView);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XOR_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XOR_B);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XOR_C);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XOR_D);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XOR_E);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XOR_H);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XOR_HL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XOR_L);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XOR_d8);
  Py_VISIT(traverse_module_state->__pyx_kp_u__2);
  Py_VISIT(traverse_module_state->__pyx_kp_u__21);
  Py_VISIT(traverse_module_state->__pyx_n_s__22);
  Py_VISIT(traverse_module_state->__pyx_n_s__3);
  Py_VISIT(traverse_module_state->__pyx_kp_u__6);
  Py_VISIT(traverse_module_state->__pyx_kp_u__7);
  Py_VISIT(traverse_module_state->__pyx_n_s_abc);
  Py_VISIT(traverse_module_state->__pyx_n_s_allocate_buffer);
  Py_VISIT(traverse_module_state->__pyx_kp_u_and);
  Py_VISIT(traverse_module_state->__pyx_n_s_array);
  Py_VISIT(traverse_module_state->__pyx_n_s_base);
  Py_VISIT(traverse_module_state->__pyx_n_s_c);
  Py_VISIT(traverse_module_state->__pyx_n_u_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_kp_s_collections_abc);
  Py_VISIT(traverse_module_state->__pyx_kp_s_contiguous_and_direct);
  Py_VISIT(traverse_module_state->__pyx_kp_s_contiguous_and_indirect);
  Py_VISIT(traverse_module_state->__pyx_n_s_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_dtype_is_object);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_error);
  Py_VISIT(traverse_module_state->__pyx_n_s_flags);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_fortran);
  Py_VISIT(traverse_module_state->__pyx_n_u_fortran);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got_differing_extents_in_dimensi);
  Py_VISIT(traverse_module_state->__pyx_n_s_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_itemsize);
  Py_VISIT(traverse_module_state->__pyx_kp_s_itemsize_0_for_cython_array);
  Py_VISIT(traverse_module_state->__pyx_n_s_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_logging);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_memview);
  Py_VISIT(traverse_module_state->__pyx_n_s_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_ndim);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_n_s_obj);
  Py_VISIT(traverse_module_state->__pyx_n_s_pack);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyboy);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_capi);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Enum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_register);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_shape);
  Py_VISIT(traverse_module_state->__pyx_n_s_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_s_step);
  Py_VISIT(traverse_module_state->__pyx_n_s_stop);
  Py_VISIT(traverse_module_state->__pyx_kp_s_strided_and_direct);
  Py_VISIT(traverse_module_state->__pyx_kp_s_strided_and_direct_or_indirect);
  Py_VISIT(traverse_module_state->__pyx_kp_s_strided_and_indirect);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_struct);
  Py_VISIT(traverse_module_state->__pyx_n_s_sys);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_kp_s_unable_to_allocate_array_data);
  Py_VISIT(traverse_module_state->__pyx_kp_s_unable_to_allocate_shape_and_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_utils);
  Py_VISIT(traverse_module_state->__pyx_n_s_version_info);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_3);
  Py_VISIT(traverse_module_state->__pyx_int_4);
  Py_VISIT(traverse_module_state->__pyx_int_8);
  Py_VISIT(traverse_module_state->__pyx_int_112105877);
  Py_VISIT(traverse_module_state->__pyx_int_136983863);
  Py_VISIT(traverse_module_state->__pyx_int_184977713);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_slice__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__4);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__15);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__17);
  Py_VISIT(traverse_module_state->__pyx_tuple__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_codeobj__19);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_7logging_7logging_Logger __pyx_mstate_global->__pyx_ptype_5pyboy_7logging_7logging_Logger
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_7bootrom_BootROM __pyx_mstate_global->__pyx_ptype_5pyboy_4core_7bootrom_BootROM
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_5utils_IntIOInterface __pyx_mstate_global->__pyx_ptype_5pyboy_5utils_IntIOInterface
#define __pyx_ptype_5pyboy_5utils_IntIOWrapper __pyx_mstate_global->__pyx_ptype_5pyboy_5utils_IntIOWrapper
#define __pyx_ptype_5pyboy_5utils_WindowEvent __pyx_mstate_global->__pyx_ptype_5pyboy_5utils_WindowEvent
#define __pyx_ptype_5pyboy_5utils_WindowEventMouse __pyx_mstate_global->__pyx_ptype_5pyboy_5utils_WindowEventMouse
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_9cartridge_3rtc_RTC __pyx_mstate_global->__pyx_ptype_5pyboy_4core_9cartridge_3rtc_RTC
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_BaseMBC __pyx_mstate_global->__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_BaseMBC
#define __pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_ROMOnly __pyx_mstate_global->__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_ROMOnly
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_11interaction_Interaction __pyx_mstate_global->__pyx_ptype_5pyboy_4core_11interaction_Interaction
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_5array_array __pyx_mstate_global->__pyx_ptype_7cpython_5array_array
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_3lcd_LCD __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_LCD
#define __pyx_ptype_5pyboy_4core_3lcd_PaletteRegister __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_PaletteRegister
#define __pyx_ptype_5pyboy_4core_3lcd_STATRegister __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_STATRegister
#define __pyx_ptype_5pyboy_4core_3lcd_LCDCRegister __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_LCDCRegister
#define __pyx_ptype_5pyboy_4core_3lcd_Renderer __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_Renderer
#define __pyx_ptype_5pyboy_4core_3lcd_CGBLCD __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_CGBLCD
#define __pyx_ptype_5pyboy_4core_3lcd_CGBRenderer __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_CGBRenderer
#define __pyx_ptype_5pyboy_4core_3lcd_VBKregister __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_VBKregister
#define __pyx_ptype_5pyboy_4core_3lcd_PaletteIndexRegister __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_PaletteIndexRegister
#define __pyx_ptype_5pyboy_4core_3lcd_PaletteColorRegister __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3lcd_PaletteColorRegister
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_3ram_RAM __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3ram_RAM
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_5sound_Sound __pyx_mstate_global->__pyx_ptype_5pyboy_4core_5sound_Sound
#define __pyx_ptype_5pyboy_4core_5sound_ToneChannel __pyx_mstate_global->__pyx_ptype_5pyboy_4core_5sound_ToneChannel
#define __pyx_ptype_5pyboy_4core_5sound_SweepChannel __pyx_mstate_global->__pyx_ptype_5pyboy_4core_5sound_SweepChannel
#define __pyx_ptype_5pyboy_4core_5sound_WaveChannel __pyx_mstate_global->__pyx_ptype_5pyboy_4core_5sound_WaveChannel
#define __pyx_ptype_5pyboy_4core_5sound_NoiseChannel __pyx_mstate_global->__pyx_ptype_5pyboy_4core_5sound_NoiseChannel
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_5timer_Timer __pyx_mstate_global->__pyx_ptype_5pyboy_4core_5timer_Timer
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_2mb_Motherboard __pyx_mstate_global->__pyx_ptype_5pyboy_4core_2mb_Motherboard
#define __pyx_ptype_5pyboy_4core_2mb_HDMA __pyx_mstate_global->__pyx_ptype_5pyboy_4core_2mb_HDMA
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5pyboy_4core_3cpu_CPU __pyx_mstate_global->__pyx_ptype_5pyboy_4core_3cpu_CPU
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type___pyx_array __pyx_mstate_global->__pyx_type___pyx_array
#define __pyx_type___pyx_MemviewEnum __pyx_mstate_global->__pyx_type___pyx_MemviewEnum
#define __pyx_type___pyx_memoryview __pyx_mstate_global->__pyx_type___pyx_memoryview
#define __pyx_type___pyx_memoryviewslice __pyx_mstate_global->__pyx_type___pyx_memoryviewslice
#endif
#define __pyx_array_type __pyx_mstate_global->__pyx_array_type
#define __pyx_MemviewEnum_type __pyx_mstate_global->__pyx_MemviewEnum_type
#define __pyx_memoryview_type __pyx_mstate_global->__pyx_memoryview_type
#define __pyx_memoryviewslice_type __pyx_mstate_global->__pyx_memoryviewslice_type
#define __pyx_kp_u_ __pyx_mstate_global->__pyx_kp_u_
#define __pyx_kp_u_ADC_A_A __pyx_mstate_global->__pyx_kp_u_ADC_A_A
#define __pyx_kp_u_ADC_A_B __pyx_mstate_global->__pyx_kp_u_ADC_A_B
#define __pyx_kp_u_ADC_A_C __pyx_mstate_global->__pyx_kp_u_ADC_A_C
#define __pyx_kp_u_ADC_A_D __pyx_mstate_global->__pyx_kp_u_ADC_A_D
#define __pyx_kp_u_ADC_A_E __pyx_mstate_global->__pyx_kp_u_ADC_A_E
#define __pyx_kp_u_ADC_A_H __pyx_mstate_global->__pyx_kp_u_ADC_A_H
#define __pyx_kp_u_ADC_A_HL __pyx_mstate_global->__pyx_kp_u_ADC_A_HL
#define __pyx_kp_u_ADC_A_L __pyx_mstate_global->__pyx_kp_u_ADC_A_L
#define __pyx_kp_u_ADC_A_d8 __pyx_mstate_global->__pyx_kp_u_ADC_A_d8
#define __pyx_kp_u_ADD_A_A __pyx_mstate_global->__pyx_kp_u_ADD_A_A
#define __pyx_kp_u_ADD_A_B __pyx_mstate_global->__pyx_kp_u_ADD_A_B
#define __pyx_kp_u_ADD_A_C __pyx_mstate_global->__pyx_kp_u_ADD_A_C
#define __pyx_kp_u_ADD_A_D __pyx_mstate_global->__pyx_kp_u_ADD_A_D
#define __pyx_kp_u_ADD_A_E __pyx_mstate_global->__pyx_kp_u_ADD_A_E
#define __pyx_kp_u_ADD_A_H __pyx_mstate_global->__pyx_kp_u_ADD_A_H
#define __pyx_kp_u_ADD_A_HL __pyx_mstate_global->__pyx_kp_u_ADD_A_HL
#define __pyx_kp_u_ADD_A_L __pyx_mstate_global->__pyx_kp_u_ADD_A_L
#define __pyx_kp_u_ADD_A_d8 __pyx_mstate_global->__pyx_kp_u_ADD_A_d8
#define __pyx_kp_u_ADD_HL_BC __pyx_mstate_global->__pyx_kp_u_ADD_HL_BC
#define __pyx_kp_u_ADD_HL_DE __pyx_mstate_global->__pyx_kp_u_ADD_HL_DE
#define __pyx_kp_u_ADD_HL_HL __pyx_mstate_global->__pyx_kp_u_ADD_HL_HL
#define __pyx_kp_u_ADD_HL_SP __pyx_mstate_global->__pyx_kp_u_ADD_HL_SP
#define __pyx_kp_u_ADD_SP_r8 __pyx_mstate_global->__pyx_kp_u_ADD_SP_r8
#define __pyx_kp_u_AND_A __pyx_mstate_global->__pyx_kp_u_AND_A
#define __pyx_kp_u_AND_B __pyx_mstate_global->__pyx_kp_u_AND_B
#define __pyx_kp_u_AND_C __pyx_mstate_global->__pyx_kp_u_AND_C
#define __pyx_kp_u_AND_D __pyx_mstate_global->__pyx_kp_u_AND_D
#define __pyx_kp_u_AND_E __pyx_mstate_global->__pyx_kp_u_AND_E
#define __pyx_kp_u_AND_H __pyx_mstate_global->__pyx_kp_u_AND_H
#define __pyx_kp_u_AND_HL __pyx_mstate_global->__pyx_kp_u_AND_HL
#define __pyx_kp_u_AND_L __pyx_mstate_global->__pyx_kp_u_AND_L
#define __pyx_kp_u_AND_d8 __pyx_mstate_global->__pyx_kp_u_AND_d8
#define __pyx_n_s_ASCII __pyx_mstate_global->__pyx_n_s_ASCII
#define __pyx_kp_s_All_dimensions_preceding_dimensi __pyx_mstate_global->__pyx_kp_s_All_dimensions_preceding_dimensi
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_u_B __pyx_mstate_global->__pyx_n_u_B
#define __pyx_kp_u_BIT_0_A __pyx_mstate_global->__pyx_kp_u_BIT_0_A
#define __pyx_kp_u_BIT_0_B __pyx_mstate_global->__pyx_kp_u_BIT_0_B
#define __pyx_kp_u_BIT_0_C __pyx_mstate_global->__pyx_kp_u_BIT_0_C
#define __pyx_kp_u_BIT_0_D __pyx_mstate_global->__pyx_kp_u_BIT_0_D
#define __pyx_kp_u_BIT_0_E __pyx_mstate_global->__pyx_kp_u_BIT_0_E
#define __pyx_kp_u_BIT_0_H __pyx_mstate_global->__pyx_kp_u_BIT_0_H
#define __pyx_kp_u_BIT_0_HL __pyx_mstate_global->__pyx_kp_u_BIT_0_HL
#define __pyx_kp_u_BIT_0_L __pyx_mstate_global->__pyx_kp_u_BIT_0_L
#define __pyx_kp_u_BIT_1_A __pyx_mstate_global->__pyx_kp_u_BIT_1_A
#define __pyx_kp_u_BIT_1_B __pyx_mstate_global->__pyx_kp_u_BIT_1_B
#define __pyx_kp_u_BIT_1_C __pyx_mstate_global->__pyx_kp_u_BIT_1_C
#define __pyx_kp_u_BIT_1_D __pyx_mstate_global->__pyx_kp_u_BIT_1_D
#define __pyx_kp_u_BIT_1_E __pyx_mstate_global->__pyx_kp_u_BIT_1_E
#define __pyx_kp_u_BIT_1_H __pyx_mstate_global->__pyx_kp_u_BIT_1_H
#define __pyx_kp_u_BIT_1_HL __pyx_mstate_global->__pyx_kp_u_BIT_1_HL
#define __pyx_kp_u_BIT_1_L __pyx_mstate_global->__pyx_kp_u_BIT_1_L
#define __pyx_kp_u_BIT_2_A __pyx_mstate_global->__pyx_kp_u_BIT_2_A
#define __pyx_kp_u_BIT_2_B __pyx_mstate_global->__pyx_kp_u_BIT_2_B
#define __pyx_kp_u_BIT_2_C __pyx_mstate_global->__pyx_kp_u_BIT_2_C
#define __pyx_kp_u_BIT_2_D __pyx_mstate_global->__pyx_kp_u_BIT_2_D
#define __pyx_kp_u_BIT_2_E __pyx_mstate_global->__pyx_kp_u_BIT_2_E
#define __pyx_kp_u_BIT_2_H __pyx_mstate_global->__pyx_kp_u_BIT_2_H
#define __pyx_kp_u_BIT_2_HL __pyx_mstate_global->__pyx_kp_u_BIT_2_HL
#define __pyx_kp_u_BIT_2_L __pyx_mstate_global->__pyx_kp_u_BIT_2_L
#define __pyx_kp_u_BIT_3_A __pyx_mstate_global->__pyx_kp_u_BIT_3_A
#define __pyx_kp_u_BIT_3_B __pyx_mstate_global->__pyx_kp_u_BIT_3_B
#define __pyx_kp_u_BIT_3_C __pyx_mstate_global->__pyx_kp_u_BIT_3_C
#define __pyx_kp_u_BIT_3_D __pyx_mstate_global->__pyx_kp_u_BIT_3_D
#define __pyx_kp_u_BIT_3_E __pyx_mstate_global->__pyx_kp_u_BIT_3_E
#define __pyx_kp_u_BIT_3_H __pyx_mstate_global->__pyx_kp_u_BIT_3_H
#define __pyx_kp_u_BIT_3_HL __pyx_mstate_global->__pyx_kp_u_BIT_3_HL
#define __pyx_kp_u_BIT_3_L __pyx_mstate_global->__pyx_kp_u_BIT_3_L
#define __pyx_kp_u_BIT_4_A __pyx_mstate_global->__pyx_kp_u_BIT_4_A
#define __pyx_kp_u_BIT_4_B __pyx_mstate_global->__pyx_kp_u_BIT_4_B
#define __pyx_kp_u_BIT_4_C __pyx_mstate_global->__pyx_kp_u_BIT_4_C
#define __pyx_kp_u_BIT_4_D __pyx_mstate_global->__pyx_kp_u_BIT_4_D
#define __pyx_kp_u_BIT_4_E __pyx_mstate_global->__pyx_kp_u_BIT_4_E
#define __pyx_kp_u_BIT_4_H __pyx_mstate_global->__pyx_kp_u_BIT_4_H
#define __pyx_kp_u_BIT_4_HL __pyx_mstate_global->__pyx_kp_u_BIT_4_HL
#define __pyx_kp_u_BIT_4_L __pyx_mstate_global->__pyx_kp_u_BIT_4_L
#define __pyx_kp_u_BIT_5_A __pyx_mstate_global->__pyx_kp_u_BIT_5_A
#define __pyx_kp_u_BIT_5_B __pyx_mstate_global->__pyx_kp_u_BIT_5_B
#define __pyx_kp_u_BIT_5_C __pyx_mstate_global->__pyx_kp_u_BIT_5_C
#define __pyx_kp_u_BIT_5_D __pyx_mstate_global->__pyx_kp_u_BIT_5_D
#define __pyx_kp_u_BIT_5_E __pyx_mstate_global->__pyx_kp_u_BIT_5_E
#define __pyx_kp_u_BIT_5_H __pyx_mstate_global->__pyx_kp_u_BIT_5_H
#define __pyx_kp_u_BIT_5_HL __pyx_mstate_global->__pyx_kp_u_BIT_5_HL
#define __pyx_kp_u_BIT_5_L __pyx_mstate_global->__pyx_kp_u_BIT_5_L
#define __pyx_kp_u_BIT_6_A __pyx_mstate_global->__pyx_kp_u_BIT_6_A
#define __pyx_kp_u_BIT_6_B __pyx_mstate_global->__pyx_kp_u_BIT_6_B
#define __pyx_kp_u_BIT_6_C __pyx_mstate_global->__pyx_kp_u_BIT_6_C
#define __pyx_kp_u_BIT_6_D __pyx_mstate_global->__pyx_kp_u_BIT_6_D
#define __pyx_kp_u_BIT_6_E __pyx_mstate_global->__pyx_kp_u_BIT_6_E
#define __pyx_kp_u_BIT_6_H __pyx_mstate_global->__pyx_kp_u_BIT_6_H
#define __pyx_kp_u_BIT_6_HL __pyx_mstate_global->__pyx_kp_u_BIT_6_HL
#define __pyx_kp_u_BIT_6_L __pyx_mstate_global->__pyx_kp_u_BIT_6_L
#define __pyx_kp_u_BIT_7_A __pyx_mstate_global->__pyx_kp_u_BIT_7_A
#define __pyx_kp_u_BIT_7_B __pyx_mstate_global->__pyx_kp_u_BIT_7_B
#define __pyx_kp_u_BIT_7_C __pyx_mstate_global->__pyx_kp_u_BIT_7_C
#define __pyx_kp_u_BIT_7_D __pyx_mstate_global->__pyx_kp_u_BIT_7_D
#define __pyx_kp_u_BIT_7_E __pyx_mstate_global->__pyx_kp_u_BIT_7_E
#define __pyx_kp_u_BIT_7_H __pyx_mstate_global->__pyx_kp_u_BIT_7_H
#define __pyx_kp_u_BIT_7_HL __pyx_mstate_global->__pyx_kp_u_BIT_7_HL
#define __pyx_kp_u_BIT_7_L __pyx_mstate_global->__pyx_kp_u_BIT_7_L
#define __pyx_kp_s_Buffer_view_does_not_expose_stri __pyx_mstate_global->__pyx_kp_s_Buffer_view_does_not_expose_stri
#define __pyx_kp_u_CALL_C_a16 __pyx_mstate_global->__pyx_kp_u_CALL_C_a16
#define __pyx_kp_u_CALL_NC_a16 __pyx_mstate_global->__pyx_kp_u_CALL_NC_a16
#define __pyx_kp_u_CALL_NZ_a16 __pyx_mstate_global->__pyx_kp_u_CALL_NZ_a16
#define __pyx_kp_u_CALL_Z_a16 __pyx_mstate_global->__pyx_kp_u_CALL_Z_a16
#define __pyx_kp_u_CALL_a16 __pyx_mstate_global->__pyx_kp_u_CALL_a16
#define __pyx_kp_u_CB_cannot_be_called __pyx_mstate_global->__pyx_kp_u_CB_cannot_be_called
#define __pyx_n_u_CCF __pyx_mstate_global->__pyx_n_u_CCF
#define __pyx_n_u_CPL __pyx_mstate_global->__pyx_n_u_CPL
#define __pyx_n_s_CPU_COMMANDS __pyx_mstate_global->__pyx_n_s_CPU_COMMANDS
#define __pyx_kp_u_CP_A __pyx_mstate_global->__pyx_kp_u_CP_A
#define __pyx_kp_u_CP_B __pyx_mstate_global->__pyx_kp_u_CP_B
#define __pyx_kp_u_CP_C __pyx_mstate_global->__pyx_kp_u_CP_C
#define __pyx_kp_u_CP_D __pyx_mstate_global->__pyx_kp_u_CP_D
#define __pyx_kp_u_CP_E __pyx_mstate_global->__pyx_kp_u_CP_E
#define __pyx_kp_u_CP_H __pyx_mstate_global->__pyx_kp_u_CP_H
#define __pyx_kp_u_CP_HL __pyx_mstate_global->__pyx_kp_u_CP_HL
#define __pyx_kp_u_CP_L __pyx_mstate_global->__pyx_kp_u_CP_L
#define __pyx_kp_u_CP_d8 __pyx_mstate_global->__pyx_kp_u_CP_d8
#define __pyx_kp_s_Can_only_create_a_buffer_that_is __pyx_mstate_global->__pyx_kp_s_Can_only_create_a_buffer_that_is
#define __pyx_kp_s_Cannot_assign_to_read_only_memor __pyx_mstate_global->__pyx_kp_s_Cannot_assign_to_read_only_memor
#define __pyx_kp_s_Cannot_create_writable_memory_vi __pyx_mstate_global->__pyx_kp_s_Cannot_create_writable_memory_vi
#define __pyx_kp_u_Cannot_index_with_type __pyx_mstate_global->__pyx_kp_u_Cannot_index_with_type
#define __pyx_kp_s_Cannot_transpose_memoryview_with __pyx_mstate_global->__pyx_kp_s_Cannot_transpose_memoryview_with
#define __pyx_n_u_DAA __pyx_mstate_global->__pyx_n_u_DAA
#define __pyx_kp_u_DEC_A __pyx_mstate_global->__pyx_kp_u_DEC_A
#define __pyx_kp_u_DEC_B __pyx_mstate_global->__pyx_kp_u_DEC_B
#define __pyx_kp_u_DEC_BC __pyx_mstate_global->__pyx_kp_u_DEC_BC
#define __pyx_kp_u_DEC_C __pyx_mstate_global->__pyx_kp_u_DEC_C
#define __pyx_kp_u_DEC_D __pyx_mstate_global->__pyx_kp_u_DEC_D
#define __pyx_kp_u_DEC_DE __pyx_mstate_global->__pyx_kp_u_DEC_DE
#define __pyx_kp_u_DEC_E __pyx_mstate_global->__pyx_kp_u_DEC_E
#define __pyx_kp_u_DEC_H __pyx_mstate_global->__pyx_kp_u_DEC_H
#define __pyx_kp_u_DEC_HL __pyx_mstate_global->__pyx_kp_u_DEC_HL
#define __pyx_kp_u_DEC_HL_2 __pyx_mstate_global->__pyx_kp_u_DEC_HL_2
#define __pyx_kp_u_DEC_L __pyx_mstate_global->__pyx_kp_u_DEC_L
#define __pyx_kp_u_DEC_SP __pyx_mstate_global->__pyx_kp_u_DEC_SP
#define __pyx_n_u_DI __pyx_mstate_global->__pyx_n_u_DI
#define __pyx_kp_s_Dimension_d_is_not_direct __pyx_mstate_global->__pyx_kp_s_Dimension_d_is_not_direct
#define __pyx_n_u_EI __pyx_mstate_global->__pyx_n_u_EI
#define __pyx_n_s_Ellipsis __pyx_mstate_global->__pyx_n_s_Ellipsis
#define __pyx_kp_s_Empty_shape_tuple_for_cython_arr __pyx_mstate_global->__pyx_kp_s_Empty_shape_tuple_for_cython_arr
#define __pyx_n_s_FLAGC __pyx_mstate_global->__pyx_n_s_FLAGC
#define __pyx_n_s_FLAGH __pyx_mstate_global->__pyx_n_s_FLAGH
#define __pyx_n_s_FLAGN __pyx_mstate_global->__pyx_n_s_FLAGN
#define __pyx_n_s_FLAGZ __pyx_mstate_global->__pyx_n_s_FLAGZ
#define __pyx_n_u_HALT __pyx_mstate_global->__pyx_n_u_HALT
#define __pyx_kp_u_INC_A __pyx_mstate_global->__pyx_kp_u_INC_A
#define __pyx_kp_u_INC_B __pyx_mstate_global->__pyx_kp_u_INC_B
#define __pyx_kp_u_INC_BC __pyx_mstate_global->__pyx_kp_u_INC_BC
#define __pyx_kp_u_INC_C __pyx_mstate_global->__pyx_kp_u_INC_C
#define __pyx_kp_u_INC_D __pyx_mstate_global->__pyx_kp_u_INC_D
#define __pyx_kp_u_INC_DE __pyx_mstate_global->__pyx_kp_u_INC_DE
#define __pyx_kp_u_INC_E __pyx_mstate_global->__pyx_kp_u_INC_E
#define __pyx_kp_u_INC_H __pyx_mstate_global->__pyx_kp_u_INC_H
#define __pyx_kp_u_INC_HL __pyx_mstate_global->__pyx_kp_u_INC_HL
#define __pyx_kp_u_INC_HL_2 __pyx_mstate_global->__pyx_kp_u_INC_HL_2
#define __pyx_kp_u_INC_L __pyx_mstate_global->__pyx_kp_u_INC_L
#define __pyx_kp_u_INC_SP __pyx_mstate_global->__pyx_kp_u_INC_SP
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_n_s_IndexError __pyx_mstate_global->__pyx_n_s_IndexError
#define __pyx_kp_s_Index_out_of_bounds_axis_d __pyx_mstate_global->__pyx_kp_s_Index_out_of_bounds_axis_d
#define __pyx_kp_s_Indirect_dimensions_not_supporte __pyx_mstate_global->__pyx_kp_s_Indirect_dimensions_not_supporte
#define __pyx_kp_u_Invalid_mode_expected_c_or_fortr __pyx_mstate_global->__pyx_kp_u_Invalid_mode_expected_c_or_fortr
#define __pyx_kp_u_Invalid_shape_in_axis __pyx_mstate_global->__pyx_kp_u_Invalid_shape_in_axis
#define __pyx_kp_u_JP_C_a16 __pyx_mstate_global->__pyx_kp_u_JP_C_a16
#define __pyx_kp_u_JP_HL __pyx_mstate_global->__pyx_kp_u_JP_HL
#define __pyx_kp_u_JP_NC_a16 __pyx_mstate_global->__pyx_kp_u_JP_NC_a16
#define __pyx_kp_u_JP_NZ_a16 __pyx_mstate_global->__pyx_kp_u_JP_NZ_a16
#define __pyx_kp_u_JP_Z_a16 __pyx_mstate_global->__pyx_kp_u_JP_Z_a16
#define __pyx_kp_u_JP_a16 __pyx_mstate_global->__pyx_kp_u_JP_a16
#define __pyx_kp_u_JR_C_r8 __pyx_mstate_global->__pyx_kp_u_JR_C_r8
#define __pyx_kp_u_JR_NC_r8 __pyx_mstate_global->__pyx_kp_u_JR_NC_r8
#define __pyx_kp_u_JR_NZ_r8 __pyx_mstate_global->__pyx_kp_u_JR_NZ_r8
#define __pyx_kp_u_JR_Z_r8 __pyx_mstate_global->__pyx_kp_u_JR_Z_r8
#define __pyx_kp_u_JR_r8 __pyx_mstate_global->__pyx_kp_u_JR_r8
#define __pyx_kp_u_LDH_A_a8 __pyx_mstate_global->__pyx_kp_u_LDH_A_a8
#define __pyx_kp_u_LDH_a8_A __pyx_mstate_global->__pyx_kp_u_LDH_a8_A
#define __pyx_kp_u_LD_A_A __pyx_mstate_global->__pyx_kp_u_LD_A_A
#define __pyx_kp_u_LD_A_B __pyx_mstate_global->__pyx_kp_u_LD_A_B
#define __pyx_kp_u_LD_A_BC __pyx_mstate_global->__pyx_kp_u_LD_A_BC
#define __pyx_kp_u_LD_A_C __pyx_mstate_global->__pyx_kp_u_LD_A_C
#define __pyx_kp_u_LD_A_C_2 __pyx_mstate_global->__pyx_kp_u_LD_A_C_2
#define __pyx_kp_u_LD_A_D __pyx_mstate_global->__pyx_kp_u_LD_A_D
#define __pyx_kp_u_LD_A_DE __pyx_mstate_global->__pyx_kp_u_LD_A_DE
#define __pyx_kp_u_LD_A_E __pyx_mstate_global->__pyx_kp_u_LD_A_E
#define __pyx_kp_u_LD_A_H __pyx_mstate_global->__pyx_kp_u_LD_A_H
#define __pyx_kp_u_LD_A_HL __pyx_mstate_global->__pyx_kp_u_LD_A_HL
#define __pyx_kp_u_LD_A_HL_2 __pyx_mstate_global->__pyx_kp_u_LD_A_HL_2
#define __pyx_kp_u_LD_A_HL_3 __pyx_mstate_global->__pyx_kp_u_LD_A_HL_3
#define __pyx_kp_u_LD_A_L __pyx_mstate_global->__pyx_kp_u_LD_A_L
#define __pyx_kp_u_LD_A_a16 __pyx_mstate_global->__pyx_kp_u_LD_A_a16
#define __pyx_kp_u_LD_A_d8 __pyx_mstate_global->__pyx_kp_u_LD_A_d8
#define __pyx_kp_u_LD_BC_A __pyx_mstate_global->__pyx_kp_u_LD_BC_A
#define __pyx_kp_u_LD_BC_d16 __pyx_mstate_global->__pyx_kp_u_LD_BC_d16
#define __pyx_kp_u_LD_B_A __pyx_mstate_global->__pyx_kp_u_LD_B_A
#define __pyx_kp_u_LD_B_B __pyx_mstate_global->__pyx_kp_u_LD_B_B
#define __pyx_kp_u_LD_B_C __pyx_mstate_global->__pyx_kp_u_LD_B_C
#define __pyx_kp_u_LD_B_D __pyx_mstate_global->__pyx_kp_u_LD_B_D
#define __pyx_kp_u_LD_B_E __pyx_mstate_global->__pyx_kp_u_LD_B_E
#define __pyx_kp_u_LD_B_H __pyx_mstate_global->__pyx_kp_u_LD_B_H
#define __pyx_kp_u_LD_B_HL __pyx_mstate_global->__pyx_kp_u_LD_B_HL
#define __pyx_kp_u_LD_B_L __pyx_mstate_global->__pyx_kp_u_LD_B_L
#define __pyx_kp_u_LD_B_d8 __pyx_mstate_global->__pyx_kp_u_LD_B_d8
#define __pyx_kp_u_LD_C_A __pyx_mstate_global->__pyx_kp_u_LD_C_A
#define __pyx_kp_u_LD_C_A_2 __pyx_mstate_global->__pyx_kp_u_LD_C_A_2
#define __pyx_kp_u_LD_C_B __pyx_mstate_global->__pyx_kp_u_LD_C_B
#define __pyx_kp_u_LD_C_C __pyx_mstate_global->__pyx_kp_u_LD_C_C
#define __pyx_kp_u_LD_C_D __pyx_mstate_global->__pyx_kp_u_LD_C_D
#define __pyx_kp_u_LD_C_E __pyx_mstate_global->__pyx_kp_u_LD_C_E
#define __pyx_kp_u_LD_C_H __pyx_mstate_global->__pyx_kp_u_LD_C_H
#define __pyx_kp_u_LD_C_HL __pyx_mstate_global->__pyx_kp_u_LD_C_HL
#define __pyx_kp_u_LD_C_L __pyx_mstate_global->__pyx_kp_u_LD_C_L
#define __pyx_kp_u_LD_C_d8 __pyx_mstate_global->__pyx_kp_u_LD_C_d8
#define __pyx_kp_u_LD_DE_A __pyx_mstate_global->__pyx_kp_u_LD_DE_A
#define __pyx_kp_u_LD_DE_d16 __pyx_mstate_global->__pyx_kp_u_LD_DE_d16
#define __pyx_kp_u_LD_D_A __pyx_mstate_global->__pyx_kp_u_LD_D_A
#define __pyx_kp_u_LD_D_B __pyx_mstate_global->__pyx_kp_u_LD_D_B
#define __pyx_kp_u_LD_D_C __pyx_mstate_global->__pyx_kp_u_LD_D_C
#define __pyx_kp_u_LD_D_D __pyx_mstate_global->__pyx_kp_u_LD_D_D
#define __pyx_kp_u_LD_D_E __pyx_mstate_global->__pyx_kp_u_LD_D_E
#define __pyx_kp_u_LD_D_H __pyx_mstate_global->__pyx_kp_u_LD_D_H
#define __pyx_kp_u_LD_D_HL __pyx_mstate_global->__pyx_kp_u_LD_D_HL
#define __pyx_kp_u_LD_D_L __pyx_mstate_global->__pyx_kp_u_LD_D_L
#define __pyx_kp_u_LD_D_d8 __pyx_mstate_global->__pyx_kp_u_LD_D_d8
#define __pyx_kp_u_LD_E_A __pyx_mstate_global->__pyx_kp_u_LD_E_A
#define __pyx_kp_u_LD_E_B __pyx_mstate_global->__pyx_kp_u_LD_E_B
#define __pyx_kp_u_LD_E_C __pyx_mstate_global->__pyx_kp_u_LD_E_C
#define __pyx_kp_u_LD_E_D __pyx_mstate_global->__pyx_kp_u_LD_E_D
#define __pyx_kp_u_LD_E_E __pyx_mstate_global->__pyx_kp_u_LD_E_E
#define __pyx_kp_u_LD_E_H __pyx_mstate_global->__pyx_kp_u_LD_E_H
#define __pyx_kp_u_LD_E_HL __pyx_mstate_global->__pyx_kp_u_LD_E_HL
#define __pyx_kp_u_LD_E_L __pyx_mstate_global->__pyx_kp_u_LD_E_L
#define __pyx_kp_u_LD_E_d8 __pyx_mstate_global->__pyx_kp_u_LD_E_d8
#define __pyx_kp_u_LD_HL_A __pyx_mstate_global->__pyx_kp_u_LD_HL_A
#define __pyx_kp_u_LD_HL_A_2 __pyx_mstate_global->__pyx_kp_u_LD_HL_A_2
#define __pyx_kp_u_LD_HL_A_3 __pyx_mstate_global->__pyx_kp_u_LD_HL_A_3
#define __pyx_kp_u_LD_HL_B __pyx_mstate_global->__pyx_kp_u_LD_HL_B
#define __pyx_kp_u_LD_HL_C __pyx_mstate_global->__pyx_kp_u_LD_HL_C
#define __pyx_kp_u_LD_HL_D __pyx_mstate_global->__pyx_kp_u_LD_HL_D
#define __pyx_kp_u_LD_HL_E __pyx_mstate_global->__pyx_kp_u_LD_HL_E
#define __pyx_kp_u_LD_HL_H __pyx_mstate_global->__pyx_kp_u_LD_HL_H
#define __pyx_kp_u_LD_HL_L __pyx_mstate_global->__pyx_kp_u_LD_HL_L
#define __pyx_kp_u_LD_HL_SP_r8 __pyx_mstate_global->__pyx_kp_u_LD_HL_SP_r8
#define __pyx_kp_u_LD_HL_d16 __pyx_mstate_global->__pyx_kp_u_LD_HL_d16
#define __pyx_kp_u_LD_HL_d8 __pyx_mstate_global->__pyx_kp_u_LD_HL_d8
#define __pyx_kp_u_LD_H_A __pyx_mstate_global->__pyx_kp_u_LD_H_A
#define __pyx_kp_u_LD_H_B __pyx_mstate_global->__pyx_kp_u_LD_H_B
#define __pyx_kp_u_LD_H_C __pyx_mstate_global->__pyx_kp_u_LD_H_C
#define __pyx_kp_u_LD_H_D __pyx_mstate_global->__pyx_kp_u_LD_H_D
#define __pyx_kp_u_LD_H_E __pyx_mstate_global->__pyx_kp_u_LD_H_E
#define __pyx_kp_u_LD_H_H __pyx_mstate_global->__pyx_kp_u_LD_H_H
#define __pyx_kp_u_LD_H_HL __pyx_mstate_global->__pyx_kp_u_LD_H_HL
#define __pyx_kp_u_LD_H_L __pyx_mstate_global->__pyx_kp_u_LD_H_L
#define __pyx_kp_u_LD_H_d8 __pyx_mstate_global->__pyx_kp_u_LD_H_d8
#define __pyx_kp_u_LD_L_A __pyx_mstate_global->__pyx_kp_u_LD_L_A
#define __pyx_kp_u_LD_L_B __pyx_mstate_global->__pyx_kp_u_LD_L_B
#define __pyx_kp_u_LD_L_C __pyx_mstate_global->__pyx_kp_u_LD_L_C
#define __pyx_kp_u_LD_L_D __pyx_mstate_global->__pyx_kp_u_LD_L_D
#define __pyx_kp_u_LD_L_E __pyx_mstate_global->__pyx_kp_u_LD_L_E
#define __pyx_kp_u_LD_L_H __pyx_mstate_global->__pyx_kp_u_LD_L_H
#define __pyx_kp_u_LD_L_HL __pyx_mstate_global->__pyx_kp_u_LD_L_HL
#define __pyx_kp_u_LD_L_L __pyx_mstate_global->__pyx_kp_u_LD_L_L
#define __pyx_kp_u_LD_L_d8 __pyx_mstate_global->__pyx_kp_u_LD_L_d8
#define __pyx_kp_u_LD_SP_HL __pyx_mstate_global->__pyx_kp_u_LD_SP_HL
#define __pyx_kp_u_LD_SP_d16 __pyx_mstate_global->__pyx_kp_u_LD_SP_d16
#define __pyx_kp_u_LD_a16_A __pyx_mstate_global->__pyx_kp_u_LD_a16_A
#define __pyx_kp_u_LD_a16_SP __pyx_mstate_global->__pyx_kp_u_LD_a16_SP
#define __pyx_n_s_MemoryError __pyx_mstate_global->__pyx_n_s_MemoryError
#define __pyx_kp_s_MemoryView_of_r_at_0x_x __pyx_mstate_global->__pyx_kp_s_MemoryView_of_r_at_0x_x
#define __pyx_kp_s_MemoryView_of_r_object __pyx_mstate_global->__pyx_kp_s_MemoryView_of_r_object
#define __pyx_n_u_NOP __pyx_mstate_global->__pyx_n_u_NOP
#define __pyx_n_b_O __pyx_mstate_global->__pyx_n_b_O
#define __pyx_n_s_OPCODE_LENGTHS __pyx_mstate_global->__pyx_n_s_OPCODE_LENGTHS
#define __pyx_kp_u_OR_A __pyx_mstate_global->__pyx_kp_u_OR_A
#define __pyx_kp_u_OR_B __pyx_mstate_global->__pyx_kp_u_OR_B
#define __pyx_kp_u_OR_C __pyx_mstate_global->__pyx_kp_u_OR_C
#define __pyx_kp_u_OR_D __pyx_mstate_global->__pyx_kp_u_OR_D
#define __pyx_kp_u_OR_E __pyx_mstate_global->__pyx_kp_u_OR_E
#define __pyx_kp_u_OR_H __pyx_mstate_global->__pyx_kp_u_OR_H
#define __pyx_kp_u_OR_HL __pyx_mstate_global->__pyx_kp_u_OR_HL
#define __pyx_kp_u_OR_L __pyx_mstate_global->__pyx_kp_u_OR_L
#define __pyx_kp_u_OR_d8 __pyx_mstate_global->__pyx_kp_u_OR_d8
#define __pyx_kp_u_Out_of_bounds_on_buffer_access_a __pyx_mstate_global->__pyx_kp_u_Out_of_bounds_on_buffer_access_a
#define __pyx_n_s_OverflowError __pyx_mstate_global->__pyx_n_s_OverflowError
#define __pyx_kp_u_POP_AF __pyx_mstate_global->__pyx_kp_u_POP_AF
#define __pyx_kp_u_POP_BC __pyx_mstate_global->__pyx_kp_u_POP_BC
#define __pyx_kp_u_POP_DE __pyx_mstate_global->__pyx_kp_u_POP_DE
#define __pyx_kp_u_POP_HL __pyx_mstate_global->__pyx_kp_u_POP_HL
#define __pyx_kp_u_PREFIX_CB __pyx_mstate_global->__pyx_kp_u_PREFIX_CB
#define __pyx_kp_u_PUSH_AF __pyx_mstate_global->__pyx_kp_u_PUSH_AF
#define __pyx_kp_u_PUSH_BC __pyx_mstate_global->__pyx_kp_u_PUSH_BC
#define __pyx_kp_u_PUSH_DE __pyx_mstate_global->__pyx_kp_u_PUSH_DE
#define __pyx_kp_u_PUSH_HL __pyx_mstate_global->__pyx_kp_u_PUSH_HL
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_kp_u_RES_0_A __pyx_mstate_global->__pyx_kp_u_RES_0_A
#define __pyx_kp_u_RES_0_B __pyx_mstate_global->__pyx_kp_u_RES_0_B
#define __pyx_kp_u_RES_0_C __pyx_mstate_global->__pyx_kp_u_RES_0_C
#define __pyx_kp_u_RES_0_D __pyx_mstate_global->__pyx_kp_u_RES_0_D
#define __pyx_kp_u_RES_0_E __pyx_mstate_global->__pyx_kp_u_RES_0_E
#define __pyx_kp_u_RES_0_H __pyx_mstate_global->__pyx_kp_u_RES_0_H
#define __pyx_kp_u_RES_0_HL __pyx_mstate_global->__pyx_kp_u_RES_0_HL
#define __pyx_kp_u_RES_0_L __pyx_mstate_global->__pyx_kp_u_RES_0_L
#define __pyx_kp_u_RES_1_A __pyx_mstate_global->__pyx_kp_u_RES_1_A
#define __pyx_kp_u_RES_1_B __pyx_mstate_global->__pyx_kp_u_RES_1_B
#define __pyx_kp_u_RES_1_C __pyx_mstate_global->__pyx_kp_u_RES_1_C
#define __pyx_kp_u_RES_1_D __pyx_mstate_global->__pyx_kp_u_RES_1_D
#define __pyx_kp_u_RES_1_E __pyx_mstate_global->__pyx_kp_u_RES_1_E
#define __pyx_kp_u_RES_1_H __pyx_mstate_global->__pyx_kp_u_RES_1_H
#define __pyx_kp_u_RES_1_HL __pyx_mstate_global->__pyx_kp_u_RES_1_HL
#define __pyx_kp_u_RES_1_L __pyx_mstate_global->__pyx_kp_u_RES_1_L
#define __pyx_kp_u_RES_2_A __pyx_mstate_global->__pyx_kp_u_RES_2_A
#define __pyx_kp_u_RES_2_B __pyx_mstate_global->__pyx_kp_u_RES_2_B
#define __pyx_kp_u_RES_2_C __pyx_mstate_global->__pyx_kp_u_RES_2_C
#define __pyx_kp_u_RES_2_D __pyx_mstate_global->__pyx_kp_u_RES_2_D
#define __pyx_kp_u_RES_2_E __pyx_mstate_global->__pyx_kp_u_RES_2_E
#define __pyx_kp_u_RES_2_H __pyx_mstate_global->__pyx_kp_u_RES_2_H
#define __pyx_kp_u_RES_2_HL __pyx_mstate_global->__pyx_kp_u_RES_2_HL
#define __pyx_kp_u_RES_2_L __pyx_mstate_global->__pyx_kp_u_RES_2_L
#define __pyx_kp_u_RES_3_A __pyx_mstate_global->__pyx_kp_u_RES_3_A
#define __pyx_kp_u_RES_3_B __pyx_mstate_global->__pyx_kp_u_RES_3_B
#define __pyx_kp_u_RES_3_C __pyx_mstate_global->__pyx_kp_u_RES_3_C
#define __pyx_kp_u_RES_3_D __pyx_mstate_global->__pyx_kp_u_RES_3_D
#define __pyx_kp_u_RES_3_E __pyx_mstate_global->__pyx_kp_u_RES_3_E
#define __pyx_kp_u_RES_3_H __pyx_mstate_global->__pyx_kp_u_RES_3_H
#define __pyx_kp_u_RES_3_HL __pyx_mstate_global->__pyx_kp_u_RES_3_HL
#define __pyx_kp_u_RES_3_L __pyx_mstate_global->__pyx_kp_u_RES_3_L
#define __pyx_kp_u_RES_4_A __pyx_mstate_global->__pyx_kp_u_RES_4_A
#define __pyx_kp_u_RES_4_B __pyx_mstate_global->__pyx_kp_u_RES_4_B
#define __pyx_kp_u_RES_4_C __pyx_mstate_global->__pyx_kp_u_RES_4_C
#define __pyx_kp_u_RES_4_D __pyx_mstate_global->__pyx_kp_u_RES_4_D
#define __pyx_kp_u_RES_4_E __pyx_mstate_global->__pyx_kp_u_RES_4_E
#define __pyx_kp_u_RES_4_H __pyx_mstate_global->__pyx_kp_u_RES_4_H
#define __pyx_kp_u_RES_4_HL __pyx_mstate_global->__pyx_kp_u_RES_4_HL
#define __pyx_kp_u_RES_4_L __pyx_mstate_global->__pyx_kp_u_RES_4_L
#define __pyx_kp_u_RES_5_A __pyx_mstate_global->__pyx_kp_u_RES_5_A
#define __pyx_kp_u_RES_5_B __pyx_mstate_global->__pyx_kp_u_RES_5_B
#define __pyx_kp_u_RES_5_C __pyx_mstate_global->__pyx_kp_u_RES_5_C
#define __pyx_kp_u_RES_5_D __pyx_mstate_global->__pyx_kp_u_RES_5_D
#define __pyx_kp_u_RES_5_E __pyx_mstate_global->__pyx_kp_u_RES_5_E
#define __pyx_kp_u_RES_5_H __pyx_mstate_global->__pyx_kp_u_RES_5_H
#define __pyx_kp_u_RES_5_HL __pyx_mstate_global->__pyx_kp_u_RES_5_HL
#define __pyx_kp_u_RES_5_L __pyx_mstate_global->__pyx_kp_u_RES_5_L
#define __pyx_kp_u_RES_6_A __pyx_mstate_global->__pyx_kp_u_RES_6_A
#define __pyx_kp_u_RES_6_B __pyx_mstate_global->__pyx_kp_u_RES_6_B
#define __pyx_kp_u_RES_6_C __pyx_mstate_global->__pyx_kp_u_RES_6_C
#define __pyx_kp_u_RES_6_D __pyx_mstate_global->__pyx_kp_u_RES_6_D
#define __pyx_kp_u_RES_6_E __pyx_mstate_global->__pyx_kp_u_RES_6_E
#define __pyx_kp_u_RES_6_H __pyx_mstate_global->__pyx_kp_u_RES_6_H
#define __pyx_kp_u_RES_6_HL __pyx_mstate_global->__pyx_kp_u_RES_6_HL
#define __pyx_kp_u_RES_6_L __pyx_mstate_global->__pyx_kp_u_RES_6_L
#define __pyx_kp_u_RES_7_A __pyx_mstate_global->__pyx_kp_u_RES_7_A
#define __pyx_kp_u_RES_7_B __pyx_mstate_global->__pyx_kp_u_RES_7_B
#define __pyx_kp_u_RES_7_C __pyx_mstate_global->__pyx_kp_u_RES_7_C
#define __pyx_kp_u_RES_7_D __pyx_mstate_global->__pyx_kp_u_RES_7_D
#define __pyx_kp_u_RES_7_E __pyx_mstate_global->__pyx_kp_u_RES_7_E
#define __pyx_kp_u_RES_7_H __pyx_mstate_global->__pyx_kp_u_RES_7_H
#define __pyx_kp_u_RES_7_HL __pyx_mstate_global->__pyx_kp_u_RES_7_HL
#define __pyx_kp_u_RES_7_L __pyx_mstate_global->__pyx_kp_u_RES_7_L
#define __pyx_n_u_RET __pyx_mstate_global->__pyx_n_u_RET
#define __pyx_n_u_RETI __pyx_mstate_global->__pyx_n_u_RETI
#define __pyx_kp_u_RET_C __pyx_mstate_global->__pyx_kp_u_RET_C
#define __pyx_kp_u_RET_NC __pyx_mstate_global->__pyx_kp_u_RET_NC
#define __pyx_kp_u_RET_NZ __pyx_mstate_global->__pyx_kp_u_RET_NZ
#define __pyx_kp_u_RET_Z __pyx_mstate_global->__pyx_kp_u_RET_Z
#define __pyx_n_u_RLA __pyx_mstate_global->__pyx_n_u_RLA
#define __pyx_n_u_RLCA __pyx_mstate_global->__pyx_n_u_RLCA
#define __pyx_kp_u_RLC_A __pyx_mstate_global->__pyx_kp_u_RLC_A
#define __pyx_kp_u_RLC_B __pyx_mstate_global->__pyx_kp_u_RLC_B
#define __pyx_kp_u_RLC_C __pyx_mstate_global->__pyx_kp_u_RLC_C
#define __pyx_kp_u_RLC_D __pyx_mstate_global->__pyx_kp_u_RLC_D
#define __pyx_kp_u_RLC_E __pyx_mstate_global->__pyx_kp_u_RLC_E
#define __pyx_kp_u_RLC_H __pyx_mstate_global->__pyx_kp_u_RLC_H
#define __pyx_kp_u_RLC_HL __pyx_mstate_global->__pyx_kp_u_RLC_HL
#define __pyx_kp_u_RLC_L __pyx_mstate_global->__pyx_kp_u_RLC_L
#define __pyx_kp_u_RL_A __pyx_mstate_global->__pyx_kp_u_RL_A
#define __pyx_kp_u_RL_B __pyx_mstate_global->__pyx_kp_u_RL_B
#define __pyx_kp_u_RL_C __pyx_mstate_global->__pyx_kp_u_RL_C
#define __pyx_kp_u_RL_D __pyx_mstate_global->__pyx_kp_u_RL_D
#define __pyx_kp_u_RL_E __pyx_mstate_global->__pyx_kp_u_RL_E
#define __pyx_kp_u_RL_H __pyx_mstate_global->__pyx_kp_u_RL_H
#define __pyx_kp_u_RL_HL __pyx_mstate_global->__pyx_kp_u_RL_HL
#define __pyx_kp_u_RL_L __pyx_mstate_global->__pyx_kp_u_RL_L
#define __pyx_n_u_RRA __pyx_mstate_global->__pyx_n_u_RRA
#define __pyx_n_u_RRCA __pyx_mstate_global->__pyx_n_u_RRCA
#define __pyx_kp_u_RRC_A __pyx_mstate_global->__pyx_kp_u_RRC_A
#define __pyx_kp_u_RRC_B __pyx_mstate_global->__pyx_kp_u_RRC_B
#define __pyx_kp_u_RRC_C __pyx_mstate_global->__pyx_kp_u_RRC_C
#define __pyx_kp_u_RRC_D __pyx_mstate_global->__pyx_kp_u_RRC_D
#define __pyx_kp_u_RRC_E __pyx_mstate_global->__pyx_kp_u_RRC_E
#define __pyx_kp_u_RRC_H __pyx_mstate_global->__pyx_kp_u_RRC_H
#define __pyx_kp_u_RRC_HL __pyx_mstate_global->__pyx_kp_u_RRC_HL
#define __pyx_kp_u_RRC_L __pyx_mstate_global->__pyx_kp_u_RRC_L
#define __pyx_kp_u_RR_A __pyx_mstate_global->__pyx_kp_u_RR_A
#define __pyx_kp_u_RR_B __pyx_mstate_global->__pyx_kp_u_RR_B
#define __pyx_kp_u_RR_C __pyx_mstate_global->__pyx_kp_u_RR_C
#define __pyx_kp_u_RR_D __pyx_mstate_global->__pyx_kp_u_RR_D
#define __pyx_kp_u_RR_E __pyx_mstate_global->__pyx_kp_u_RR_E
#define __pyx_kp_u_RR_H __pyx_mstate_global->__pyx_kp_u_RR_H
#define __pyx_kp_u_RR_HL __pyx_mstate_global->__pyx_kp_u_RR_HL
#define __pyx_kp_u_RR_L __pyx_mstate_global->__pyx_kp_u_RR_L
#define __pyx_kp_u_RST_00H __pyx_mstate_global->__pyx_kp_u_RST_00H
#define __pyx_kp_u_RST_08H __pyx_mstate_global->__pyx_kp_u_RST_08H
#define __pyx_kp_u_RST_10H __pyx_mstate_global->__pyx_kp_u_RST_10H
#define __pyx_kp_u_RST_18H __pyx_mstate_global->__pyx_kp_u_RST_18H
#define __pyx_kp_u_RST_20H __pyx_mstate_global->__pyx_kp_u_RST_20H
#define __pyx_kp_u_RST_28H __pyx_mstate_global->__pyx_kp_u_RST_28H
#define __pyx_kp_u_RST_30H __pyx_mstate_global->__pyx_kp_u_RST_30H
#define __pyx_kp_u_RST_38H __pyx_mstate_global->__pyx_kp_u_RST_38H
#define __pyx_n_s_RuntimeError __pyx_mstate_global->__pyx_n_s_RuntimeError
#define __pyx_kp_u_SBC_A_A __pyx_mstate_global->__pyx_kp_u_SBC_A_A
#define __pyx_kp_u_SBC_A_B __pyx_mstate_global->__pyx_kp_u_SBC_A_B
#define __pyx_kp_u_SBC_A_C __pyx_mstate_global->__pyx_kp_u_SBC_A_C
#define __pyx_kp_u_SBC_A_D __pyx_mstate_global->__pyx_kp_u_SBC_A_D
#define __pyx_kp_u_SBC_A_E __pyx_mstate_global->__pyx_kp_u_SBC_A_E
#define __pyx_kp_u_SBC_A_H __pyx_mstate_global->__pyx_kp_u_SBC_A_H
#define __pyx_kp_u_SBC_A_HL __pyx_mstate_global->__pyx_kp_u_SBC_A_HL
#define __pyx_kp_u_SBC_A_L __pyx_mstate_global->__pyx_kp_u_SBC_A_L
#define __pyx_kp_u_SBC_A_d8 __pyx_mstate_global->__pyx_kp_u_SBC_A_d8
#define __pyx_n_u_SCF __pyx_mstate_global->__pyx_n_u_SCF
#define __pyx_kp_u_SET_0_A __pyx_mstate_global->__pyx_kp_u_SET_0_A
#define __pyx_kp_u_SET_0_B __pyx_mstate_global->__pyx_kp_u_SET_0_B
#define __pyx_kp_u_SET_0_C __pyx_mstate_global->__pyx_kp_u_SET_0_C
#define __pyx_kp_u_SET_0_D __pyx_mstate_global->__pyx_kp_u_SET_0_D
#define __pyx_kp_u_SET_0_E __pyx_mstate_global->__pyx_kp_u_SET_0_E
#define __pyx_kp_u_SET_0_H __pyx_mstate_global->__pyx_kp_u_SET_0_H
#define __pyx_kp_u_SET_0_HL __pyx_mstate_global->__pyx_kp_u_SET_0_HL
#define __pyx_kp_u_SET_0_L __pyx_mstate_global->__pyx_kp_u_SET_0_L
#define __pyx_kp_u_SET_1_A __pyx_mstate_global->__pyx_kp_u_SET_1_A
#define __pyx_kp_u_SET_1_B __pyx_mstate_global->__pyx_kp_u_SET_1_B
#define __pyx_kp_u_SET_1_C __pyx_mstate_global->__pyx_kp_u_SET_1_C
#define __pyx_kp_u_SET_1_D __pyx_mstate_global->__pyx_kp_u_SET_1_D
#define __pyx_kp_u_SET_1_E __pyx_mstate_global->__pyx_kp_u_SET_1_E
#define __pyx_kp_u_SET_1_H __pyx_mstate_global->__pyx_kp_u_SET_1_H
#define __pyx_kp_u_SET_1_HL __pyx_mstate_global->__pyx_kp_u_SET_1_HL
#define __pyx_kp_u_SET_1_L __pyx_mstate_global->__pyx_kp_u_SET_1_L
#define __pyx_kp_u_SET_2_A __pyx_mstate_global->__pyx_kp_u_SET_2_A
#define __pyx_kp_u_SET_2_B __pyx_mstate_global->__pyx_kp_u_SET_2_B
#define __pyx_kp_u_SET_2_C __pyx_mstate_global->__pyx_kp_u_SET_2_C
#define __pyx_kp_u_SET_2_D __pyx_mstate_global->__pyx_kp_u_SET_2_D
#define __pyx_kp_u_SET_2_E __pyx_mstate_global->__pyx_kp_u_SET_2_E
#define __pyx_kp_u_SET_2_H __pyx_mstate_global->__pyx_kp_u_SET_2_H
#define __pyx_kp_u_SET_2_HL __pyx_mstate_global->__pyx_kp_u_SET_2_HL
#define __pyx_kp_u_SET_2_L __pyx_mstate_global->__pyx_kp_u_SET_2_L
#define __pyx_kp_u_SET_3_A __pyx_mstate_global->__pyx_kp_u_SET_3_A
#define __pyx_kp_u_SET_3_B __pyx_mstate_global->__pyx_kp_u_SET_3_B
#define __pyx_kp_u_SET_3_C __pyx_mstate_global->__pyx_kp_u_SET_3_C
#define __pyx_kp_u_SET_3_D __pyx_mstate_global->__pyx_kp_u_SET_3_D
#define __pyx_kp_u_SET_3_E __pyx_mstate_global->__pyx_kp_u_SET_3_E
#define __pyx_kp_u_SET_3_H __pyx_mstate_global->__pyx_kp_u_SET_3_H
#define __pyx_kp_u_SET_3_HL __pyx_mstate_global->__pyx_kp_u_SET_3_HL
#define __pyx_kp_u_SET_3_L __pyx_mstate_global->__pyx_kp_u_SET_3_L
#define __pyx_kp_u_SET_4_A __pyx_mstate_global->__pyx_kp_u_SET_4_A
#define __pyx_kp_u_SET_4_B __pyx_mstate_global->__pyx_kp_u_SET_4_B
#define __pyx_kp_u_SET_4_C __pyx_mstate_global->__pyx_kp_u_SET_4_C
#define __pyx_kp_u_SET_4_D __pyx_mstate_global->__pyx_kp_u_SET_4_D
#define __pyx_kp_u_SET_4_E __pyx_mstate_global->__pyx_kp_u_SET_4_E
#define __pyx_kp_u_SET_4_H __pyx_mstate_global->__pyx_kp_u_SET_4_H
#define __pyx_kp_u_SET_4_HL __pyx_mstate_global->__pyx_kp_u_SET_4_HL
#define __pyx_kp_u_SET_4_L __pyx_mstate_global->__pyx_kp_u_SET_4_L
#define __pyx_kp_u_SET_5_A __pyx_mstate_global->__pyx_kp_u_SET_5_A
#define __pyx_kp_u_SET_5_B __pyx_mstate_global->__pyx_kp_u_SET_5_B
#define __pyx_kp_u_SET_5_C __pyx_mstate_global->__pyx_kp_u_SET_5_C
#define __pyx_kp_u_SET_5_D __pyx_mstate_global->__pyx_kp_u_SET_5_D
#define __pyx_kp_u_SET_5_E __pyx_mstate_global->__pyx_kp_u_SET_5_E
#define __pyx_kp_u_SET_5_H __pyx_mstate_global->__pyx_kp_u_SET_5_H
#define __pyx_kp_u_SET_5_HL __pyx_mstate_global->__pyx_kp_u_SET_5_HL
#define __pyx_kp_u_SET_5_L __pyx_mstate_global->__pyx_kp_u_SET_5_L
#define __pyx_kp_u_SET_6_A __pyx_mstate_global->__pyx_kp_u_SET_6_A
#define __pyx_kp_u_SET_6_B __pyx_mstate_global->__pyx_kp_u_SET_6_B
#define __pyx_kp_u_SET_6_C __pyx_mstate_global->__pyx_kp_u_SET_6_C
#define __pyx_kp_u_SET_6_D __pyx_mstate_global->__pyx_kp_u_SET_6_D
#define __pyx_kp_u_SET_6_E __pyx_mstate_global->__pyx_kp_u_SET_6_E
#define __pyx_kp_u_SET_6_H __pyx_mstate_global->__pyx_kp_u_SET_6_H
#define __pyx_kp_u_SET_6_HL __pyx_mstate_global->__pyx_kp_u_SET_6_HL
#define __pyx_kp_u_SET_6_L __pyx_mstate_global->__pyx_kp_u_SET_6_L
#define __pyx_kp_u_SET_7_A __pyx_mstate_global->__pyx_kp_u_SET_7_A
#define __pyx_kp_u_SET_7_B __pyx_mstate_global->__pyx_kp_u_SET_7_B
#define __pyx_kp_u_SET_7_C __pyx_mstate_global->__pyx_kp_u_SET_7_C
#define __pyx_kp_u_SET_7_D __pyx_mstate_global->__pyx_kp_u_SET_7_D
#define __pyx_kp_u_SET_7_E __pyx_mstate_global->__pyx_kp_u_SET_7_E
#define __pyx_kp_u_SET_7_H __pyx_mstate_global->__pyx_kp_u_SET_7_H
#define __pyx_kp_u_SET_7_HL __pyx_mstate_global->__pyx_kp_u_SET_7_HL
#define __pyx_kp_u_SET_7_L __pyx_mstate_global->__pyx_kp_u_SET_7_L
#define __pyx_kp_u_SLA_A __pyx_mstate_global->__pyx_kp_u_SLA_A
#define __pyx_kp_u_SLA_B __pyx_mstate_global->__pyx_kp_u_SLA_B
#define __pyx_kp_u_SLA_C __pyx_mstate_global->__pyx_kp_u_SLA_C
#define __pyx_kp_u_SLA_D __pyx_mstate_global->__pyx_kp_u_SLA_D
#define __pyx_kp_u_SLA_E __pyx_mstate_global->__pyx_kp_u_SLA_E
#define __pyx_kp_u_SLA_H __pyx_mstate_global->__pyx_kp_u_SLA_H
#define __pyx_kp_u_SLA_HL __pyx_mstate_global->__pyx_kp_u_SLA_HL
#define __pyx_kp_u_SLA_L __pyx_mstate_global->__pyx_kp_u_SLA_L
#define __pyx_kp_u_SRA_A __pyx_mstate_global->__pyx_kp_u_SRA_A
#define __pyx_kp_u_SRA_B __pyx_mstate_global->__pyx_kp_u_SRA_B
#define __pyx_kp_u_SRA_C __pyx_mstate_global->__pyx_kp_u_SRA_C
#define __pyx_kp_u_SRA_D __pyx_mstate_global->__pyx_kp_u_SRA_D
#define __pyx_kp_u_SRA_E __pyx_mstate_global->__pyx_kp_u_SRA_E
#define __pyx_kp_u_SRA_H __pyx_mstate_global->__pyx_kp_u_SRA_H
#define __pyx_kp_u_SRA_HL __pyx_mstate_global->__pyx_kp_u_SRA_HL
#define __pyx_kp_u_SRA_L __pyx_mstate_global->__pyx_kp_u_SRA_L
#define __pyx_kp_u_SRL_A __pyx_mstate_global->__pyx_kp_u_SRL_A
#define __pyx_kp_u_SRL_B __pyx_mstate_global->__pyx_kp_u_SRL_B
#define __pyx_kp_u_SRL_C __pyx_mstate_global->__pyx_kp_u_SRL_C
#define __pyx_kp_u_SRL_D __pyx_mstate_global->__pyx_kp_u_SRL_D
#define __pyx_kp_u_SRL_E __pyx_mstate_global->__pyx_kp_u_SRL_E
#define __pyx_kp_u_SRL_H __pyx_mstate_global->__pyx_kp_u_SRL_H
#define __pyx_kp_u_SRL_HL __pyx_mstate_global->__pyx_kp_u_SRL_HL
#define __pyx_kp_u_SRL_L __pyx_mstate_global->__pyx_kp_u_SRL_L
#define __pyx_kp_u_STOP_0 __pyx_mstate_global->__pyx_kp_u_STOP_0
#define __pyx_kp_u_SUB_A __pyx_mstate_global->__pyx_kp_u_SUB_A
#define __pyx_kp_u_SUB_B __pyx_mstate_global->__pyx_kp_u_SUB_B
#define __pyx_kp_u_SUB_C __pyx_mstate_global->__pyx_kp_u_SUB_C
#define __pyx_kp_u_SUB_D __pyx_mstate_global->__pyx_kp_u_SUB_D
#define __pyx_kp_u_SUB_E __pyx_mstate_global->__pyx_kp_u_SUB_E
#define __pyx_kp_u_SUB_H __pyx_mstate_global->__pyx_kp_u_SUB_H
#define __pyx_kp_u_SUB_HL __pyx_mstate_global->__pyx_kp_u_SUB_HL
#define __pyx_kp_u_SUB_L __pyx_mstate_global->__pyx_kp_u_SUB_L
#define __pyx_kp_u_SUB_d8 __pyx_mstate_global->__pyx_kp_u_SUB_d8
#define __pyx_kp_u_SWAP_A __pyx_mstate_global->__pyx_kp_u_SWAP_A
#define __pyx_kp_u_SWAP_B __pyx_mstate_global->__pyx_kp_u_SWAP_B
#define __pyx_kp_u_SWAP_C __pyx_mstate_global->__pyx_kp_u_SWAP_C
#define __pyx_kp_u_SWAP_D __pyx_mstate_global->__pyx_kp_u_SWAP_D
#define __pyx_kp_u_SWAP_E __pyx_mstate_global->__pyx_kp_u_SWAP_E
#define __pyx_kp_u_SWAP_H __pyx_mstate_global->__pyx_kp_u_SWAP_H
#define __pyx_kp_u_SWAP_HL __pyx_mstate_global->__pyx_kp_u_SWAP_HL
#define __pyx_kp_u_SWAP_L __pyx_mstate_global->__pyx_kp_u_SWAP_L
#define __pyx_n_s_Sequence __pyx_mstate_global->__pyx_n_s_Sequence
#define __pyx_kp_s_Step_may_not_be_zero_axis_d __pyx_mstate_global->__pyx_kp_s_Step_may_not_be_zero_axis_d
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_kp_s_Unable_to_convert_item_to_object __pyx_mstate_global->__pyx_kp_s_Unable_to_convert_item_to_object
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s_View_MemoryView __pyx_mstate_global->__pyx_n_s_View_MemoryView
#define __pyx_kp_u_XOR_A __pyx_mstate_global->__pyx_kp_u_XOR_A
#define __pyx_kp_u_XOR_B __pyx_mstate_global->__pyx_kp_u_XOR_B
#define __pyx_kp_u_XOR_C __pyx_mstate_global->__pyx_kp_u_XOR_C
#define __pyx_kp_u_XOR_D __pyx_mstate_global->__pyx_kp_u_XOR_D
#define __pyx_kp_u_XOR_E __pyx_mstate_global->__pyx_kp_u_XOR_E
#define __pyx_kp_u_XOR_H __pyx_mstate_global->__pyx_kp_u_XOR_H
#define __pyx_kp_u_XOR_HL __pyx_mstate_global->__pyx_kp_u_XOR_HL
#define __pyx_kp_u_XOR_L __pyx_mstate_global->__pyx_kp_u_XOR_L
#define __pyx_kp_u_XOR_d8 __pyx_mstate_global->__pyx_kp_u_XOR_d8
#define __pyx_kp_u__2 __pyx_mstate_global->__pyx_kp_u__2
#define __pyx_kp_u__21 __pyx_mstate_global->__pyx_kp_u__21
#define __pyx_n_s__22 __pyx_mstate_global->__pyx_n_s__22
#define __pyx_n_s__3 __pyx_mstate_global->__pyx_n_s__3
#define __pyx_kp_u__6 __pyx_mstate_global->__pyx_kp_u__6
#define __pyx_kp_u__7 __pyx_mstate_global->__pyx_kp_u__7
#define __pyx_n_s_abc __pyx_mstate_global->__pyx_n_s_abc
#define __pyx_n_s_allocate_buffer __pyx_mstate_global->__pyx_n_s_allocate_buffer
#define __pyx_kp_u_and __pyx_mstate_global->__pyx_kp_u_and
#define __pyx_n_s_array __pyx_mstate_global->__pyx_n_s_array
#define __pyx_n_s_base __pyx_mstate_global->__pyx_n_s_base
#define __pyx_n_s_c __pyx_mstate_global->__pyx_n_s_c
#define __pyx_n_u_c __pyx_mstate_global->__pyx_n_u_c
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_kp_s_collections_abc __pyx_mstate_global->__pyx_kp_s_collections_abc
#define __pyx_kp_s_contiguous_and_direct __pyx_mstate_global->__pyx_kp_s_contiguous_and_direct
#define __pyx_kp_s_contiguous_and_indirect __pyx_mstate_global->__pyx_kp_s_contiguous_and_indirect
#define __pyx_n_s_count __pyx_mstate_global->__pyx_n_s_count
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_dtype_is_object __pyx_mstate_global->__pyx_n_s_dtype_is_object
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_error __pyx_mstate_global->__pyx_n_s_error
#define __pyx_n_s_flags __pyx_mstate_global->__pyx_n_s_flags
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_s_fortran __pyx_mstate_global->__pyx_n_s_fortran
#define __pyx_n_u_fortran __pyx_mstate_global->__pyx_n_u_fortran
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_get_logger __pyx_mstate_global->__pyx_n_s_get_logger
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_kp_u_got __pyx_mstate_global->__pyx_kp_u_got
#define __pyx_kp_u_got_differing_extents_in_dimensi __pyx_mstate_global->__pyx_kp_u_got_differing_extents_in_dimensi
#define __pyx_n_s_id __pyx_mstate_global->__pyx_n_s_id
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_itemsize __pyx_mstate_global->__pyx_n_s_itemsize
#define __pyx_kp_s_itemsize_0_for_cython_array __pyx_mstate_global->__pyx_kp_s_itemsize_0_for_cython_array
#define __pyx_n_s_logger __pyx_mstate_global->__pyx_n_s_logger
#define __pyx_n_s_logging __pyx_mstate_global->__pyx_n_s_logging
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_memview __pyx_mstate_global->__pyx_n_s_memview
#define __pyx_n_s_mode __pyx_mstate_global->__pyx_n_s_mode
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_ndim __pyx_mstate_global->__pyx_n_s_ndim
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_n_s_obj __pyx_mstate_global->__pyx_n_s_obj
#define __pyx_n_s_pack __pyx_mstate_global->__pyx_n_s_pack
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_pyboy __pyx_mstate_global->__pyx_n_s_pyboy
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_capi __pyx_mstate_global->__pyx_n_s_pyx_capi
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_Enum __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Enum
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_register __pyx_mstate_global->__pyx_n_s_register
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_shape __pyx_mstate_global->__pyx_n_s_shape
#define __pyx_n_s_size __pyx_mstate_global->__pyx_n_s_size
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_s_step __pyx_mstate_global->__pyx_n_s_step
#define __pyx_n_s_stop __pyx_mstate_global->__pyx_n_s_stop
#define __pyx_kp_s_strided_and_direct __pyx_mstate_global->__pyx_kp_s_strided_and_direct
#define __pyx_kp_s_strided_and_direct_or_indirect __pyx_mstate_global->__pyx_kp_s_strided_and_direct_or_indirect
#define __pyx_kp_s_strided_and_indirect __pyx_mstate_global->__pyx_kp_s_strided_and_indirect
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_struct __pyx_mstate_global->__pyx_n_s_struct
#define __pyx_n_s_sys __pyx_mstate_global->__pyx_n_s_sys
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_kp_s_unable_to_allocate_array_data __pyx_mstate_global->__pyx_kp_s_unable_to_allocate_array_data
#define __pyx_kp_s_unable_to_allocate_shape_and_str __pyx_mstate_global->__pyx_kp_s_unable_to_allocate_shape_and_str
#define __pyx_n_s_unpack __pyx_mstate_global->__pyx_n_s_unpack
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_utils __pyx_mstate_global->__pyx_n_s_utils
#define __pyx_n_s_version_info __pyx_mstate_global->__pyx_n_s_version_info
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_3 __pyx_mstate_global->__pyx_int_3
#define __pyx_int_4 __pyx_mstate_global->__pyx_int_4
#define __pyx_int_8 __pyx_mstate_global->__pyx_int_8
#define __pyx_int_112105877 __pyx_mstate_global->__pyx_int_112105877
#define __pyx_int_136983863 __pyx_mstate_global->__pyx_int_136983863
#define __pyx_int_184977713 __pyx_mstate_global->__pyx_int_184977713
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_slice__5 __pyx_mstate_global->__pyx_slice__5
#define __pyx_tuple__4 __pyx_mstate_global->__pyx_tuple__4
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__15 __pyx_mstate_global->__pyx_tuple__15
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__17 __pyx_mstate_global->__pyx_tuple__17
#define __pyx_tuple__18 __pyx_mstate_global->__pyx_tuple__18
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_codeobj__19 __pyx_mstate_global->__pyx_codeobj__19
/* #### Code section: module_code ### */

/* "carray.from_py":79
 * 
 * @cname("__Pyx_carray_from_py_uint8_t")
 * cdef int __Pyx_carray_from_py_uint8_t(object o, base_type *v, Py_ssize_t length) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i = length
 *     try:
 */

static int __Pyx_carray_from_py_uint8_t(PyObject *__pyx_v_o, uint8_t *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_item = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  uint8_t __pyx_t_11;
  char const *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_carray_from_py_uint8_t", 1);

  /* "carray.from_py":80
 * @cname("__Pyx_carray_from_py_uint8_t")
 * cdef int __Pyx_carray_from_py_uint8_t(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length             # <<<<<<<<<<<<<<
 *     try:
 *         i = len(o)
 */
  __pyx_v_i = __pyx_v_length;

  /* "carray.from_py":81
 * cdef int __Pyx_carray_from_py_uint8_t(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "carray.from_py":82
 *     cdef Py_ssize_t i = length
 *     try:
 *         i = len(o)             # <<<<<<<<<<<<<<
 *     except (TypeError, OverflowError):
 *         pass
 */
      __pyx_t_4 = PyObject_Length(__pyx_v_o); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 82, __pyx_L3_error)
      __pyx_v_i = __pyx_t_4;

      /* "carray.from_py":81
 * cdef int __Pyx_carray_from_py_uint8_t(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "carray.from_py":83
 *     try:
 *         i = len(o)
 *     except (TypeError, OverflowError):             # <<<<<<<<<<<<<<
 *         pass
 *     if i == length:
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_TypeError, __pyx_builtin_OverflowError);
    if (__pyx_t_5) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "carray.from_py":81
 * cdef int __Pyx_carray_from_py_uint8_t(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "carray.from_py":85
 *     except (TypeError, OverflowError):
 *         pass
 *     if i == length:             # <<<<<<<<<<<<<<
 *         for i, item in enumerate(o):
 *             if i >= length:
 */
  __pyx_t_6 = (__pyx_v_i == __pyx_v_length);
  if (__pyx_t_6) {

    /* "carray.from_py":86
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */
    __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
      __pyx_t_7 = __pyx_v_o; __Pyx_INCREF(__pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 86, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_7);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 86, __pyx_L1_error)
            #endif
            if (__pyx_t_8 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(1, 86, __pyx_L1_error)
          #else
          __pyx_t_10 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 86, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 86, __pyx_L1_error)
            #endif
            if (__pyx_t_8 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(1, 86, __pyx_L1_error)
          #else
          __pyx_t_10 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 86, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        }
      } else {
        __pyx_t_10 = __pyx_t_9(__pyx_t_7);
        if (unlikely(!__pyx_t_10)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(1, 86, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_10);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_10);
      __pyx_t_10 = 0;
      __pyx_v_i = __pyx_t_4;
      __pyx_t_4 = (__pyx_t_4 + 1);

      /* "carray.from_py":87
 *     if i == length:
 *         for i, item in enumerate(o):
 *             if i >= length:             # <<<<<<<<<<<<<<
 *                 break
 *             v[i] = item
 */
      __pyx_t_6 = (__pyx_v_i >= __pyx_v_length);
      if (__pyx_t_6) {

        /* "carray.from_py":88
 *         for i, item in enumerate(o):
 *             if i >= length:
 *                 break             # <<<<<<<<<<<<<<
 *             v[i] = item
 *         else:
 */
        goto __pyx_L11_break;

        /* "carray.from_py":87
 *     if i == length:
 *         for i, item in enumerate(o):
 *             if i >= length:             # <<<<<<<<<<<<<<
 *                 break
 *             v[i] = item
 */
      }

      /* "carray.from_py":89
 *             if i >= length:
 *                 break
 *             v[i] = item             # <<<<<<<<<<<<<<
 *         else:
 *             i += 1  # convert index to length
 */
      __pyx_t_11 = __Pyx_PyInt_As_uint8_t(__pyx_v_item); if (unlikely((__pyx_t_11 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 89, __pyx_L1_error)
      (__pyx_v_v[__pyx_v_i]) = __pyx_t_11;

      /* "carray.from_py":86
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L13_for_else;
    __pyx_L11_break:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L14_for_end;
    /*else*/ {
      __pyx_L13_for_else:;

      /* "carray.from_py":91
 *             v[i] = item
 *         else:
 *             i += 1  # convert index to length             # <<<<<<<<<<<<<<
 *             if i == length:
 *                 return 0
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "carray.from_py":92
 *         else:
 *             i += 1  # convert index to length
 *             if i == length:             # <<<<<<<<<<<<<<
 *                 return 0
 * 
 */
      __pyx_t_6 = (__pyx_v_i == __pyx_v_length);
      if (__pyx_t_6) {

        /* "carray.from_py":93
 *             i += 1  # convert index to length
 *             if i == length:
 *                 return 0             # <<<<<<<<<<<<<<
 * 
 *     PyErr_Format(
 */
        __pyx_r = 0;
        goto __pyx_L0;

        /* "carray.from_py":92
 *         else:
 *             i += 1  # convert index to length
 *             if i == length:             # <<<<<<<<<<<<<<
 *                 return 0
 * 
 */
      }
    }
    __pyx_L14_for_end:;

    /* "carray.from_py":85
 *     except (TypeError, OverflowError):
 *         pass
 *     if i == length:             # <<<<<<<<<<<<<<
 *         for i, item in enumerate(o):
 *             if i >= length:
 */
  }

  /* "carray.from_py":98
 *         IndexError,
 *         ("too many values found during array assignment, expected %zd"
 *          if i >= length else             # <<<<<<<<<<<<<<
 *          "not enough values found during array assignment, expected %zd, got %zd"),
 *         length, i)
 */
  __pyx_t_6 = (__pyx_v_i >= __pyx_v_length);
  if (__pyx_t_6) {
    __pyx_t_12 = ((char const *)"too many values found during array assignment, expected %zd");
  } else {
    __pyx_t_12 = ((char const *)"not enough values found during array assignment, expected %zd, got %zd");
  }

  /* "carray.from_py":95
 *                 return 0
 * 
 *     PyErr_Format(             # <<<<<<<<<<<<<<
 *         IndexError,
 *         ("too many values found during array assignment, expected %zd"
 */
  __pyx_t_7 = PyErr_Format(__pyx_builtin_IndexError, __pyx_t_12, __pyx_v_length, __pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "carray.from_py":79
 * 
 * @cname("__Pyx_carray_from_py_uint8_t")
 * cdef int __Pyx_carray_from_py_uint8_t(object o, base_type *v, Py_ssize_t length) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i = length
 *     try:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("carray.from_py.__Pyx_carray_from_py_uint8_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":131
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    values[3] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)__pyx_n_s_c));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_shape)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_itemsize)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(1, 131, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_format)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(1, 131, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mode);
          if (value) { values[3] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 131, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 132, __pyx_L3_error)
    } else {

      /* "View.MemoryView":132
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, __pyx_nargs); __PYX_ERR(1, 131, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(1, 131, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(1, 131, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":131
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_dim;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  char *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":137
 *         cdef Py_ssize_t dim
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 137, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 137, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":138
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":140
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 */
  __pyx_t_2 = (!(__pyx_v_self->ndim != 0));
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":141
 * 
 *         if not self.ndim:
 *             raise ValueError, "Empty shape tuple for cython.array"             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Empty_shape_tuple_for_cython_arr, 0, 0);
    __PYX_ERR(1, 141, __pyx_L1_error)

    /* "View.MemoryView":140
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 */
  }

  /* "View.MemoryView":143
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 */
  __pyx_t_2 = (__pyx_v_itemsize <= 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":144
 * 
 *         if itemsize <= 0:
 *             raise ValueError, "itemsize <= 0 for cython.array"             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_itemsize_0_for_cython_array, 0, 0);
    __PYX_ERR(1, 144, __pyx_L1_error)

    /* "View.MemoryView":143
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 */
  }

  /* "View.MemoryView":146
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "View.MemoryView":147
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_n_s_ASCII};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 147, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":146
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":148
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_v_format))) __PYX_ERR(1, 148, __pyx_L1_error)
  __pyx_t_4 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":149
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 149, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(1, 149, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_8;

  /* "View.MemoryView":152
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":153
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":155
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError, "unable to allocate shape and strides."
 * 
 */
  __pyx_t_3 = (!(__pyx_v_self->_shape != 0));
  if (unlikely(__pyx_t_3)) {

    /* "View.MemoryView":156
 * 
 *         if not self._shape:
 *             raise MemoryError, "unable to allocate shape and strides."             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_Raise(__pyx_builtin_MemoryError, __pyx_kp_s_unable_to_allocate_shape_and_str, 0, 0);
    __PYX_ERR(1, 156, __pyx_L1_error)

    /* "View.MemoryView":155
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError, "unable to allocate shape and strides."
 * 
 */
  }

  /* "View.MemoryView":159
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 */
  __pyx_t_7 = 0;
  __pyx_t_4 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_4);
  __pyx_t_1 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 159, __pyx_L1_error)
      #endif
      if (__pyx_t_1 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely((0 < 0))) __PYX_ERR(1, 159, __pyx_L1_error)
    #else
    __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 159, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_7;
    __pyx_t_7 = (__pyx_t_7 + 1);

    /* "View.MemoryView":160
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 *             self._shape[idx] = dim
 */
    __pyx_t_3 = (__pyx_v_dim <= 0);
    if (unlikely(__pyx_t_3)) {

      /* "View.MemoryView":161
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_Invalid_shape_in_axis);
      __pyx_t_9 += 22;
      __Pyx_GIVEREF(__pyx_kp_u_Invalid_shape_in_axis);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Invalid_shape_in_axis);
      __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_idx, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u_);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u_);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u_);
      __pyx_t_6 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_dim, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u__2);
      __pyx_t_9 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__2);
      PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_kp_u__2);
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_5, 5, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_t_6, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 161, __pyx_L1_error)

      /* "View.MemoryView":160
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":162
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":159
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "View.MemoryView":165
 * 
 *         cdef char order
 *         if mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 165, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "View.MemoryView":166
 *         cdef char order
 *         if mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         elif mode == 'fortran':
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":167
 *         if mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         elif mode == 'fortran':
 *             order = b'F'
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":165
 * 
 *         cdef char order
 *         if mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L11;
  }

  /* "View.MemoryView":168
 *             order = b'C'
 *             self.mode = u'c'
 *         elif mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(1, 168, __pyx_L1_error)
  if (likely(__pyx_t_3)) {

    /* "View.MemoryView":169
 *             self.mode = u'c'
 *         elif mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         else:
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":170
 *         elif mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError, f"Invalid mode, expected 'c' or 'fortran', got {mode}"
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":168
 *             order = b'C'
 *             self.mode = u'c'
 *         elif mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L11;
  }

  /* "View.MemoryView":172
 *             self.mode = u'fortran'
 *         else:
 *             raise ValueError, f"Invalid mode, expected 'c' or 'fortran', got {mode}"             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides, itemsize, self.ndim, order)
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_mode, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Invalid_mode_expected_c_or_fortr, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_t_6, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 172, __pyx_L1_error)
  }
  __pyx_L11:;

  /* "View.MemoryView":174
 *             raise ValueError, f"Invalid mode, expected 'c' or 'fortran', got {mode}"
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides, itemsize, self.ndim, order)             # <<<<<<<<<<<<<<
 * 
 *         self.free_data = allocate_buffer
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":176
 *         self.len = fill_contig_strides_array(self._shape, self._strides, itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 * 
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":177
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 * 
 *         if allocate_buffer:
 */
  __pyx_t_6 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_3;

  /* "View.MemoryView":179
 *         self.dtype_is_object = format == b'O'
 * 
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 *             _allocate_buffer(self)
 * 
 */
  if (__pyx_v_allocate_buffer) {

    /* "View.MemoryView":180
 * 
 *         if allocate_buffer:
 *             _allocate_buffer(self)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_t_7 = __pyx_array_allocate_buffer(__pyx_v_self); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 180, __pyx_L1_error)

    /* "View.MemoryView":179
 *         self.dtype_is_object = format == b'O'
 * 
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 *             _allocate_buffer(self)
 * 
 */
  }

  /* "View.MemoryView":131
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":182
 *             _allocate_buffer(self)
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 */

/* Python wrapper */
CYTHON_UNUSED static int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
CYTHON_UNUSED static int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char *__pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (unlikely(__pyx_v_info == NULL)) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":184
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):
 *             if self.mode == u"c":
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":185
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):             # <<<<<<<<<<<<<<
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_t_1 = ((__pyx_v_flags & ((PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS) | PyBUF_ANY_CONTIGUOUS)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":186
 *         cdef int bufmode = -1
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):
 *             if self.mode == u"c":             # <<<<<<<<<<<<<<
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":
 */
    __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 186, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":187
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *             elif self.mode == u"fortran":
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
      __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

      /* "View.MemoryView":186
 *         cdef int bufmode = -1
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):
 *             if self.mode == u"c":             # <<<<<<<<<<<<<<
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":188
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):
 */
    __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 188, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":189
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *             if not (flags & bufmode):
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 */
      __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

      /* "View.MemoryView":188
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):
 */
    }
    __pyx_L4:;

    /* "View.MemoryView":190
 *             elif self.mode == u"fortran":
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data
 */
    __pyx_t_1 = (!((__pyx_v_flags & __pyx_v_bufmode) != 0));
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":191
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Can_only_create_a_buffer_that_is, 0, 0);
      __PYX_ERR(1, 191, __pyx_L1_error)

      /* "View.MemoryView":190
 *             elif self.mode == u"fortran":
 *                 bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *             if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data
 */
    }

    /* "View.MemoryView":185
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if flags & (PyBUF_C_CONTIGUOUS | PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS):             # <<<<<<<<<<<<<<
 *             if self.mode == u"c":
 *                 bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  }

  /* "View.MemoryView":192
 *             if not (flags & bufmode):
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 * 
 */
  __pyx_t_2 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_2;

  /* "View.MemoryView":193
 *                 raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  __pyx_t_3 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_3;

  /* "View.MemoryView":195
 *         info.len = self.len
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.ndim = self.ndim
 *             info.shape = self._shape
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":196
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *             info.shape = self._shape
 *             info.strides = self._strides
 */
    __pyx_t_4 = __pyx_v_self->ndim;
    __pyx_v_info->ndim = __pyx_t_4;

    /* "View.MemoryView":197
 *         if flags & PyBUF_STRIDES:
 *             info.ndim = self.ndim
 *             info.shape = self._shape             # <<<<<<<<<<<<<<
 *             info.strides = self._strides
 *         else:
 */
    __pyx_t_5 = __pyx_v_self->_shape;
    __pyx_v_info->shape = __pyx_t_5;

    /* "View.MemoryView":198
 *             info.ndim = self.ndim
 *             info.shape = self._shape
 *             info.strides = self._strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.ndim = 1
 */
    __pyx_t_5 = __pyx_v_self->_strides;
    __pyx_v_info->strides = __pyx_t_5;

    /* "View.MemoryView":195
 *         info.len = self.len
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.ndim = self.ndim
 *             info.shape = self._shape
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":200
 *             info.strides = self._strides
 *         else:
 *             info.ndim = 1             # <<<<<<<<<<<<<<
 *             info.shape = &self.len if flags & PyBUF_ND else NULL
 *             info.strides = NULL
 */
  /*else*/ {
    __pyx_v_info->ndim = 1;

    /* "View.MemoryView":201
 *         else:
 *             info.ndim = 1
 *             info.shape = &self.len if flags & PyBUF_ND else NULL             # <<<<<<<<<<<<<<
 *             info.strides = NULL
 * 
 */
    __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
    if (__pyx_t_1) {
      __pyx_t_5 = (&__pyx_v_self->len);
    } else {
      __pyx_t_5 = NULL;
    }
    __pyx_v_info->shape = __pyx_t_5;

    /* "View.MemoryView":202
 *             info.ndim = 1
 *             info.shape = &self.len if flags & PyBUF_ND else NULL
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.suboffsets = NULL
 */
    __pyx_v_info->strides = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":204
 *             info.strides = NULL
 * 
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":205
 * 
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 *         info.format = self.format if flags & PyBUF_FORMAT else NULL
 */
  __pyx_t_3 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_3;

  /* "View.MemoryView":206
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 *         info.format = self.format if flags & PyBUF_FORMAT else NULL
 *         info.obj = self
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":207
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 *         info.format = self.format if flags & PyBUF_FORMAT else NULL             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {
    __pyx_t_2 = __pyx_v_self->format;
  } else {
    __pyx_t_2 = NULL;
  }
  __pyx_v_info->format = __pyx_t_2;

  /* "View.MemoryView":208
 *         info.readonly = 0
 *         info.format = self.format if flags & PyBUF_FORMAT else NULL
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":182
 *             _allocate_buffer(self)
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":210
 *         info.obj = self
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_t_1;
  int __pyx_t_2;

  /* "View.MemoryView":211
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 */
  __pyx_t_1 = (__pyx_v_self->callback_free_data != NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":212
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":211
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":213
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 */
  if (__pyx_v_self->free_data) {
  } else {
    __pyx_t_1 = __pyx_v_self->free_data;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->data != NULL);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":214
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)
 */
    if (__pyx_v_self->dtype_is_object) {

      /* "View.MemoryView":215
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)             # <<<<<<<<<<<<<<
 *             free(self.data)
 *         PyObject_Free(self._shape)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":214
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)
 */
    }

    /* "View.MemoryView":216
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":213
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":217
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":210
 *         info.obj = self
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
}

/* "View.MemoryView":219
 *         PyObject_Free(self._shape)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def memview(self):
 *         return self.get_memview()
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":221
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":219
 *         PyObject_Free(self._shape)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def memview(self):
 *         return self.get_memview()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":224
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_memview", 1);

  /* "View.MemoryView":225
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":226
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self))) __PYX_ERR(1, 226, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 226, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(1, 226, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":224
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":228
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;

  /* "View.MemoryView":229
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":228
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":231
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 1);

  /* "View.MemoryView":232
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":231
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":234
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "View.MemoryView":235
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":234
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":237
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 1);

  /* "View.MemoryView":238
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely((PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0))) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":237
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":248
 * 
 * @cname("__pyx_array_allocate_buffer")
 * cdef int _allocate_buffer(array self) except -1:             # <<<<<<<<<<<<<<
 * 
 * 
 */

static int __pyx_array_allocate_buffer(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_i;
  PyObject **__pyx_v_p;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":254
 *     cdef PyObject **p
 * 
 *     self.free_data = True             # <<<<<<<<<<<<<<
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:
 */
  __pyx_v_self->free_data = 1;

  /* "View.MemoryView":255
 * 
 *     self.free_data = True
 *     self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *     if not self.data:
 *         raise MemoryError, "unable to allocate array data."
 */
  __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

  /* "View.MemoryView":256
 *     self.free_data = True
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:             # <<<<<<<<<<<<<<
 *         raise MemoryError, "unable to allocate array data."
 * 
 */
  __pyx_t_1 = (!(__pyx_v_self->data != 0));
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":257
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:
 *         raise MemoryError, "unable to allocate array data."             # <<<<<<<<<<<<<<
 * 
 *     if self.dtype_is_object:
 */
    __Pyx_Raise(__pyx_builtin_MemoryError, __pyx_kp_s_unable_to_allocate_array_data, 0, 0);
    __PYX_ERR(1, 257, __pyx_L1_error)

    /* "View.MemoryView":256
 *     self.free_data = True
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:             # <<<<<<<<<<<<<<
 *         raise MemoryError, "unable to allocate array data."
 * 
 */
  }

  /* "View.MemoryView":259
 *         raise MemoryError, "unable to allocate array data."
 * 
 *     if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):
 */
  if (__pyx_v_self->dtype_is_object) {

    /* "View.MemoryView":260
 * 
 *     if self.dtype_is_object:
 *         p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *         for i in range(self.len // self.itemsize):
 *             p[i] = Py_None
 */
    __pyx_v_p = ((PyObject **)__pyx_v_self->data);

    /* "View.MemoryView":261
 *     if self.dtype_is_object:
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):             # <<<<<<<<<<<<<<
 *             p[i] = Py_None
 *             Py_INCREF(Py_None)
 */
    if (unlikely(__pyx_v_self->itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 261, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_self->itemsize == (Py_ssize_t)-1)  && unlikely(__Pyx_UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(1, 261, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_self->itemsize);
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":262
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):
 *             p[i] = Py_None             # <<<<<<<<<<<<<<
 *             Py_INCREF(Py_None)
 *     return 0
 */
      (__pyx_v_p[__pyx_v_i]) = Py_None;

      /* "View.MemoryView":263
 *         for i in range(self.len // self.itemsize):
 *             p[i] = Py_None
 *             Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
      Py_INCREF(Py_None);
    }

    /* "View.MemoryView":259
 *         raise MemoryError, "unable to allocate array data."
 * 
 *     if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):
 */
  }

  /* "View.MemoryView":264
 *             p[i] = Py_None
 *             Py_INCREF(Py_None)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":248
 * 
 * @cname("__pyx_array_allocate_buffer")
 * cdef int _allocate_buffer(array self) except -1:             # <<<<<<<<<<<<<<
 * 
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._allocate_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":268
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format, char *c_mode, char *buf):             # <<<<<<<<<<<<<<
 *     cdef array result
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_c_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  PyObject *__pyx_v_mode = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array_cwrapper", 1);

  /* "View.MemoryView":270
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format, char *c_mode, char *buf):
 *     cdef array result
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.             # <<<<<<<<<<<<<<
 * 
 *     if buf is NULL:
 */
  __pyx_t_2 = ((__pyx_v_c_mode[0]) == 'f');
  if (__pyx_t_2) {
    __Pyx_INCREF(__pyx_n_s_fortran);
    __pyx_t_1 = __pyx_n_s_fortran;
  } else {
    __Pyx_INCREF(__pyx_n_s_c);
    __pyx_t_1 = __pyx_n_s_c;
  }
  __pyx_v_mode = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":272
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 * 
 *     if buf is NULL:             # <<<<<<<<<<<<<<
 *         result = array.__new__(array, shape, itemsize, format, mode)
 *     else:
 */
  __pyx_t_2 = (__pyx_v_buf == NULL);
  if (__pyx_t_2) {

    /* "View.MemoryView":273
 * 
 *     if buf is NULL:
 *         result = array.__new__(array, shape, itemsize, format, mode)             # <<<<<<<<<<<<<<
 *     else:
 *         result = array.__new__(array, shape, itemsize, format, mode, allocate_buffer=False)
 */
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_shape)) __PYX_ERR(1, 273, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 273, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3)) __PYX_ERR(1, 273, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_mode);
    __Pyx_GIVEREF(__pyx_v_mode);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_mode)) __PYX_ERR(1, 273, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = ((PyObject *)__pyx_tp_new_array(((PyTypeObject *)__pyx_array_type), __pyx_t_4, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 273, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":272
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 * 
 *     if buf is NULL:             # <<<<<<<<<<<<<<
 *         result = array.__new__(array, shape, itemsize, format, mode)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":275
 *         result = array.__new__(array, shape, itemsize, format, mode)
 *     else:
 *         result = array.__new__(array, shape, itemsize, format, mode, allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
  /*else*/ {
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape)) __PYX_ERR(1, 275, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(1, 275, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4)) __PYX_ERR(1, 275, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_mode);
    __Pyx_GIVEREF(__pyx_v_mode);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_mode)) __PYX_ERR(1, 275, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(1, 275, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_tp_new_array(((PyTypeObject *)__pyx_array_type), __pyx_t_1, __pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 275, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":276
 *     else:
 *         result = array.__new__(array, shape, itemsize, format, mode, allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":278
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":268
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format, char *c_mode, char *buf):             # <<<<<<<<<<<<<<
 *     cdef array result
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_mode);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":304
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 304, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(1, 304, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 304, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "View.MemoryView":305
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":304
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":306
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "View.MemoryView":307
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":306
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_2;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136983863);
    __Pyx_GIVEREF(__pyx_int_136983863);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136983863)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_136983863);
    __Pyx_GIVEREF(__pyx_int_136983863);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_136983863)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x82a3537, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":349
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_obj)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_flags)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(1, 349, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 349, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 349, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, __pyx_nargs); __PYX_ERR(1, 349, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_intptr_t __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "View.MemoryView":350
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":351
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":352
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_obj != Py_None);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":353
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_3 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 353, __pyx_L1_error)

    /* "View.MemoryView":354
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->view.obj) == NULL);
    if (__pyx_t_1) {

      /* "View.MemoryView":355
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":356
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":354
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":352
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":358
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:
 */
  __pyx_t_1 = (!__PYX_CYTHON_ATOMICS_ENABLED());
  if (__pyx_t_1) {

    /* "View.MemoryView":360
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */
    __pyx_t_1 = (__pyx_memoryview_thread_locks_used < 8);
    if (__pyx_t_1) {

      /* "View.MemoryView":361
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 */
      __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

      /* "View.MemoryView":362
 *             if __pyx_memoryview_thread_locks_used < 8:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 */
      __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

      /* "View.MemoryView":360
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */
    }

    /* "View.MemoryView":363
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */
    __pyx_t_1 = (__pyx_v_self->lock == NULL);
    if (__pyx_t_1) {

      /* "View.MemoryView":364
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *                 if self.lock is NULL:
 *                     raise MemoryError
 */
      __pyx_v_self->lock = PyThread_allocate_lock();

      /* "View.MemoryView":365
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      __pyx_t_1 = (__pyx_v_self->lock == NULL);
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":366
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
        PyErr_NoMemory(); __PYX_ERR(1, 366, __pyx_L1_error)

        /* "View.MemoryView":365
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      }

      /* "View.MemoryView":363
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */
    }

    /* "View.MemoryView":358
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < 8:
 */
  }

  /* "View.MemoryView":368
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":369
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = ((__pyx_v_self->view.format[0]) == 'O');
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_self->view.format[1]) == '\x00');
    __pyx_t_1 = __pyx_t_2;
    __pyx_L12_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":368
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L11;
  }

  /* "View.MemoryView":371
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         assert <Py_intptr_t><void*>(&self.acquisition_count) % sizeof(__pyx_atomic_int_type) == 0
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L11:;

  /* "View.MemoryView":373
 *             self.dtype_is_object = dtype_is_object
 * 
 *         assert <Py_intptr_t><void*>(&self.acquisition_count) % sizeof(__pyx_atomic_int_type) == 0             # <<<<<<<<<<<<<<
 *         self.typeinfo = NULL
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_4 = ((Py_intptr_t)((void *)(&__pyx_v_self->acquisition_count)));
    __pyx_t_5 = (sizeof(__pyx_atomic_int_type));
    if (unlikely(__pyx_t_5 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 373, __pyx_L1_error)
    }
    __pyx_t_1 = ((__pyx_t_4 % __pyx_t_5) == 0);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(1, 373, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(1, 373, __pyx_L1_error)
  #endif

  /* "View.MemoryView":374
 * 
 *         assert <Py_intptr_t><void*>(&self.acquisition_count) % sizeof(__pyx_atomic_int_type) == 0
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":349
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":376
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyThread_type_lock __pyx_t_5;
  PyThread_type_lock __pyx_t_6;

  /* "View.MemoryView":377
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  if (__pyx_t_1) {

    /* "View.MemoryView":378
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":377
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":379
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_1 = (((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None);
  if (__pyx_t_1) {

    /* "View.MemoryView":381
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":382
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":379
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":386
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_1 = (__pyx_v_self->lock != NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":387
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_2 = __pyx_memoryview_thread_locks_used;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":388
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_1 = ((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock);
      if (__pyx_t_1) {

        /* "View.MemoryView":389
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":390
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_1 = (__pyx_v_i != __pyx_memoryview_thread_locks_used);
        if (__pyx_t_1) {

          /* "View.MemoryView":392
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_5 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":391
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_5;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_6;

          /* "View.MemoryView":390
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":393
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":388
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":395
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":386
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":376
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
}

/* "View.MemoryView":397
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_pointer", 1);

  /* "View.MemoryView":399
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":401
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 401, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 401, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(1, 401, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 401, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(1, 401, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 401, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":402
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 402, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(1, 402, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":401
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":404
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":397
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":407
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "View.MemoryView":408
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  if (__pyx_t_1) {

    /* "View.MemoryView":409
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":408
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":411
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 411, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 411, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_indices = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "View.MemoryView":414
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 414, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":415
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":414
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":417
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_5 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_5 == ((char *)NULL))) __PYX_ERR(1, 417, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_5;

    /* "View.MemoryView":418
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":407
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":420
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError, "Cannot assign to read-only memoryview"
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":421
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError, "Cannot assign to read-only memoryview"
 * 
 */
  if (unlikely(__pyx_v_self->view.readonly)) {

    /* "View.MemoryView":422
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError, "Cannot assign to read-only memoryview"             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_Cannot_assign_to_read_only_memor, 0, 0);
    __PYX_ERR(1, 422, __pyx_L1_error)

    /* "View.MemoryView":421
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError, "Cannot assign to read-only memoryview"
 * 
 */
  }

  /* "View.MemoryView":424
 *             raise TypeError, "Cannot assign to read-only memoryview"
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_1 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(__pyx_t_1 != Py_None)) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 424, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 424, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_2;
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":426
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(1, 426, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":427
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_obj = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "View.MemoryView":428
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(1, 428, __pyx_L1_error)
    if (__pyx_t_4) {

      /* "View.MemoryView":429
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_1, __pyx_v_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":428
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":431
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 431, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 431, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_3), __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 431, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":426
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":433
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":420
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError, "Cannot assign to read-only memoryview"
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":435
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":436
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "View.MemoryView":437
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":438
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 438, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":439
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 439, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":438
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 438, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj)) __PYX_ERR(1, 438, __pyx_L4_error);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6)) __PYX_ERR(1, 438, __pyx_L4_error);
        __Pyx_GIVEREF(__pyx_t_7);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7)) __PYX_ERR(1, 438, __pyx_L4_error);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 438, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":437
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":440
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(1, 440, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);

        /* "View.MemoryView":441
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;

      /* "View.MemoryView":437
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":436
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":443
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":435
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":445
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  __Pyx_memviewslice __pyx_v_msrc;
  __Pyx_memviewslice __pyx_v_mdst;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 1);

  /* "View.MemoryView":448
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 *         cdef __Pyx_memviewslice msrc = get_slice_from_memview(src, &src_slice)[0]             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mdst = get_slice_from_memview(dst, &dst_slice)[0]
 * 
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(1, 448, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 448, __pyx_L1_error)
  __pyx_v_msrc = (__pyx_t_1[0]);

  /* "View.MemoryView":449
 *         cdef __Pyx_memviewslice src_slice
 *         cdef __Pyx_memviewslice msrc = get_slice_from_memview(src, &src_slice)[0]
 *         cdef __Pyx_memviewslice mdst = get_slice_from_memview(dst, &dst_slice)[0]             # <<<<<<<<<<<<<<
 * 
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(1, 449, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 449, __pyx_L1_error)
  __pyx_v_mdst = (__pyx_t_1[0]);

  /* "View.MemoryView":451
 *         cdef __Pyx_memviewslice mdst = get_slice_from_memview(dst, &dst_slice)[0]
 * 
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_memoryview_copy_contents(__pyx_v_msrc, __pyx_v_mdst, __pyx_t_3, __pyx_t_4, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 451, __pyx_L1_error)

  /* "View.MemoryView":445
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":453
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 1);

  /* "View.MemoryView":455
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":460
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 460, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":462
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_2 = (((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array)));
  if (__pyx_t_2) {

    /* "View.MemoryView":463
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":464
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_2 = (__pyx_v_tmp == NULL);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":465
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(1, 465, __pyx_L1_error)

      /* "View.MemoryView":464
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":466
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":462
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":468
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":470
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":471
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    if (__pyx_v_self->dtype_is_object) {

      /* "View.MemoryView":472
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":471
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":474
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 474, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":478
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_2 = (__pyx_v_self->view.suboffsets != NULL);
    if (__pyx_t_2) {

      /* "View.MemoryView":479
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_4 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 479, __pyx_L6_error)

      /* "View.MemoryView":478
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":480
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":483
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":453
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":485
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_indexed", 1);

  /* "View.MemoryView":486
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(1, 486, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":487
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":485
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":489
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 1);

  /* "View.MemoryView":492
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":495
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":496
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":497
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError, "Unable to convert item to object"
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 497, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 497, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 497, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":496
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":501
 *             raise ValueError, "Unable to convert item to object"
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_9 = __Pyx_ssize_strlen(__pyx_v_self->view.format); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(1, 501, __pyx_L5_except_error)
      __pyx_t_10 = (__pyx_t_9 == 1);
      if (__pyx_t_10) {

        /* "View.MemoryView":502
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 502, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":501
 *             raise ValueError, "Unable to convert item to object"
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":503
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "View.MemoryView":498
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Unable to convert item to object"
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 498, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_6);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(1, 498, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_1);

      /* "View.MemoryView":499
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError, "Unable to convert item to object"             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Unable_to_convert_item_to_object, 0, 0);
      __PYX_ERR(1, 499, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "View.MemoryView":496
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":489
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":505
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  char *__pyx_t_10;
  char *__pyx_t_11;
  char *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 1);

  /* "View.MemoryView":508
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":513
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  if (__pyx_t_2) {

    /* "View.MemoryView":514
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 514, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(1, 514, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":513
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_1, __pyx_v_value};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(1, 516, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":518
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_7 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(1, 518, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_8 = __pyx_v_bytesvalue;
  __pyx_t_10 = PyBytes_AS_STRING(__pyx_t_8);
  __pyx_t_11 = (__pyx_t_10 + PyBytes_GET_SIZE(__pyx_t_8));
  for (__pyx_t_12 = __pyx_t_10; __pyx_t_12 < __pyx_t_11; __pyx_t_12++) {
    __pyx_t_9 = __pyx_t_12;
    __pyx_v_c = (__pyx_t_9[0]);

    /* "View.MemoryView":519
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_7;

    /* "View.MemoryView":518
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_7 = (__pyx_t_7 + 1);

    /* "View.MemoryView":519
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "View.MemoryView":505
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":521
 *             itemp[i] = c
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 */

/* Python wrapper */
CYTHON_UNUSED static int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
CYTHON_UNUSED static int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  char *__pyx_t_4;
  void *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (unlikely(__pyx_v_info == NULL)) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":523
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":524
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Cannot_create_writable_memory_vi, 0, 0);
    __PYX_ERR(1, 524, __pyx_L1_error)

    /* "View.MemoryView":523
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 */
  }

  /* "View.MemoryView":526
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":527
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_3 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_3;

    /* "View.MemoryView":526
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":529
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":531
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":532
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_3 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_3;

    /* "View.MemoryView":531
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":534
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":536
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":537
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_3 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_3;

    /* "View.MemoryView":536
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":539
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":541
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":542
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":541
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":544
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":546
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_5 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_5;

  /* "View.MemoryView":547
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_6 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":548
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_7 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_7;

  /* "View.MemoryView":549
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_7 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_7;

  /* "View.MemoryView":550
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":551
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":521
 *             itemp[i] = c
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":554
 * 
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":556
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(1, 556, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":557
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 557, __pyx_L1_error)

  /* "View.MemoryView":558
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":554
 * 
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":560
 *         return result
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base(self):
 *         return self._get_base()
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":562
 *     @property
 *     def base(self):
 *         return self._get_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef _get_base(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->_get_base(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":560
 *         return result
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base(self):
 *         return self._get_base()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.base.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":564
 *         return self._get_base()
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

static PyObject *__pyx_memoryview__get_base(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_base", 1);

  /* "View.MemoryView":565
 * 
 *     cdef _get_base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":564
 *         return self._get_base()
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":567
 *         return self.obj
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_7genexpr__pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":569
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
      __pyx_t_2 = __pyx_t_4;
      __pyx_7genexpr__pyx_v_length = (__pyx_t_2[0]);
      __pyx_t_5 = PyInt_FromSsize_t(__pyx_7genexpr__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 569, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 569, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":567
 *         return self.obj
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":571
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def strides(self):
 *         if self.view.strides == NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_8genexpr1__pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":573
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError, "Buffer view does not expose strides"
 */
  __pyx_t_1 = (__pyx_v_self->view.strides == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":575
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError, "Buffer view does not expose strides"             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Buffer_view_does_not_expose_stri, 0, 0);
    __PYX_ERR(1, 575, __pyx_L1_error)

    /* "View.MemoryView":573
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError, "Buffer view does not expose strides"
 */
  }

  /* "View.MemoryView":577
 *             raise ValueError, "Buffer view does not expose strides"
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_8genexpr1__pyx_v_stride = (__pyx_t_3[0]);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_8genexpr1__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(1, 577, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  } /* exit inner scope */
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":571
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def strides(self):
 *         if self.view.strides == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":579
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_8genexpr2__pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":581
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.suboffsets == NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":582
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PySequence_Multiply(__pyx_tuple__4, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":581
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":584
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.suboffsets; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_8genexpr2__pyx_v_suboffset = (__pyx_t_3[0]);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_8genexpr2__pyx_v_suboffset); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(1, 584, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  } /* exit inner scope */
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":579
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":586
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ndim(self):
 *         return self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":588
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":586
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ndim(self):
 *         return self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":590
 *         return self.view.ndim
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def itemsize(self):
 *         return self.view.itemsize
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":592
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":590
 *         return self.view.ndim
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def itemsize(self):
 *         return self.view.itemsize
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":594
 *         return self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nbytes(self):
 *         return self.size * self.view.itemsize
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":596
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":594
 *         return self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nbytes(self):
 *         return self.size * self.view.itemsize
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":598
 *         return self.size * self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def size(self):
 *         if self._size is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "View.MemoryView":600
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  if (__pyx_t_1) {

    /* "View.MemoryView":601
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":603
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
      __pyx_t_2 = __pyx_t_4;
      __pyx_t_5 = PyInt_FromSsize_t((__pyx_t_2[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "View.MemoryView":604
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_5 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_5);
      __pyx_t_5 = 0;
    }

    /* "View.MemoryView":606
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":600
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":608
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":598
 *         return self.size * self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def size(self):
 *         if self._size is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":610
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":611
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.ndim >= 1);
  if (__pyx_t_1) {

    /* "View.MemoryView":612
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":611
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":614
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":610
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":616
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "View.MemoryView":617
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":618
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":617
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(1, 617, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(1, 617, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":616
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":620
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 1);

  /* "View.MemoryView":621
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 621, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":620
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":624
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_c_contig", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_c_contig", 0))) return NULL;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_c_contig", 1);

  /* "View.MemoryView":627
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 627, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":628
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":624
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":630
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_f_contig", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_f_contig", 0))) return NULL;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_f_contig", 1);

  /* "View.MemoryView":633
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 633, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":634
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":630
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":636
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "View.MemoryView":638
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":640
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":641
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 641, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":646
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":636
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":648
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy_fortran", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy_fortran", 0))) return NULL;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_fortran", 1);

  /* "View.MemoryView":650
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":652
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":653
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 653, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":658
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":648
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":662
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 1);

  /* "View.MemoryView":663
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o)) __PYX_ERR(1, 663, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 663, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(1, 663, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":664
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":665
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":662
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":668
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":669
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o) noexcept:
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":668
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o) noexcept:             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":671
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_idx;
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unellipsify", 1);

  /* "View.MemoryView":677
 *     """
 *     cdef Py_ssize_t idx
 *     tup = <tuple>index if isinstance(index, tuple) else (index,)             # <<<<<<<<<<<<<<
 * 
 *     result = [slice(None)] * ndim
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_index); 
  if (__pyx_t_2) {
    __Pyx_INCREF(((PyObject*)__pyx_v_index));
    __pyx_t_1 = __pyx_v_index;
  } else {
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index)) __PYX_ERR(1, 677, __pyx_L1_error);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_v_tup = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":679
 *     tup = <tuple>index if isinstance(index, tuple) else (index,)
 * 
 *     result = [slice(None)] * ndim             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_1 = PyList_New(1 * ((__pyx_v_ndim<0) ? 0:__pyx_v_ndim)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_v_ndim; __pyx_temp++) {
      __Pyx_INCREF(__pyx_slice__5);
      __Pyx_GIVEREF(__pyx_slice__5);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_1, __pyx_temp, __pyx_slice__5)) __PYX_ERR(1, 679, __pyx_L1_error);
    }
  }
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":680
 * 
 *     result = [slice(None)] * ndim
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     idx = 0
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":681
 *     result = [slice(None)] * ndim
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     idx = 0
 *     for item in tup:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":682
 *     have_slices = False
 *     seen_ellipsis = False
 *     idx = 0             # <<<<<<<<<<<<<<
 *     for item in tup:
 *         if item is Ellipsis:
 */
  __pyx_v_idx = 0;

  /* "View.MemoryView":683
 *     seen_ellipsis = False
 *     idx = 0
 *     for item in tup:             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  if (unlikely(__pyx_v_tup == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 683, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_4 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 683, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(1, 683, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 683, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":684
 *     idx = 0
 *     for item in tup:
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    if (__pyx_t_2) {

      /* "View.MemoryView":685
 *     for item in tup:
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True
 */
      __pyx_t_2 = (!__pyx_v_seen_ellipsis);
      if (__pyx_t_2) {

        /* "View.MemoryView":686
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             have_slices = True
 */
        if (unlikely(__pyx_v_tup == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(1, 686, __pyx_L1_error)
        }
        __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v_tup); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 686, __pyx_L1_error)
        __pyx_v_idx = (__pyx_v_idx + (__pyx_v_ndim - __pyx_t_5));

        /* "View.MemoryView":687
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":685
 *     for item in tup:
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True
 */
      }

      /* "View.MemoryView":688
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if isinstance(item, slice):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":684
 *     idx = 0
 *     for item in tup:
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *                 have_slices = True
 *             elif not PyIndex_Check(item):
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      if (__pyx_t_2) {

        /* "View.MemoryView":691
 *         else:
 *             if isinstance(item, slice):
 *                 have_slices = True             # <<<<<<<<<<<<<<
 *             elif not PyIndex_Check(item):
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 */
        __pyx_v_have_slices = 1;

        /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *                 have_slices = True
 *             elif not PyIndex_Check(item):
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":692
 *             if isinstance(item, slice):
 *                 have_slices = True
 *             elif not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item
 */
      __pyx_t_2 = (!(PyIndex_Check(__pyx_v_item) != 0));
      if (unlikely(__pyx_t_2)) {

        /* "View.MemoryView":693
 *                 have_slices = True
 *             elif not PyIndex_Check(item):
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"             # <<<<<<<<<<<<<<
 *             result[idx] = item
 *         idx += 1
 */
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 693, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = 0;
        __pyx_t_6 = 127;
        __Pyx_INCREF(__pyx_kp_u_Cannot_index_with_type);
        __pyx_t_5 += 24;
        __Pyx_GIVEREF(__pyx_kp_u_Cannot_index_with_type);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Cannot_index_with_type);
        __pyx_t_7 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_item)), __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 693, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
        __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_INCREF(__pyx_kp_u__6);
        __pyx_t_5 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__6);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__6);
        __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 693, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_builtin_TypeError, __pyx_t_7, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(1, 693, __pyx_L1_error)

        /* "View.MemoryView":692
 *             if isinstance(item, slice):
 *                 have_slices = True
 *             elif not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item
 */
      }
      __pyx_L7:;

      /* "View.MemoryView":694
 *             elif not PyIndex_Check(item):
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item             # <<<<<<<<<<<<<<
 *         idx += 1
 * 
 */
      if (unlikely((__Pyx_SetItemInt(__pyx_v_result, __pyx_v_idx, __pyx_v_item, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0))) __PYX_ERR(1, 694, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "View.MemoryView":695
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item
 *         idx += 1             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - idx
 */
    __pyx_v_idx = (__pyx_v_idx + 1);

    /* "View.MemoryView":683
 *     seen_ellipsis = False
 *     idx = 0
 *     for item in tup:             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":697
 *         idx += 1
 * 
 *     nslices = ndim - idx             # <<<<<<<<<<<<<<
 *     return have_slices or nslices, tuple(result)
 * 
 */
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_v_idx);

  /* "View.MemoryView":698
 * 
 *     nslices = ndim - idx
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_7 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __pyx_t_7;
  __pyx_t_7 = 0;
  __pyx_L9_bool_binop_done:;
  __pyx_t_7 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(1, 698, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7)) __PYX_ERR(1, 698, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":671
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static int assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  int __pyx_r;
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":701
 * 
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError, "Indirect dimensions not supported"
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":702
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Indirect dimensions not supported"
 *     return 0  # return type just used as an error flag
 */
    __pyx_t_4 = (__pyx_v_suboffset >= 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError, "Indirect dimensions not supported"             # <<<<<<<<<<<<<<
 *     return 0  # return type just used as an error flag
 * 
 */
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Indirect_dimensions_not_supporte, 0, 0);
      __PYX_ERR(1, 703, __pyx_L1_error)

      /* "View.MemoryView":702
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Indirect dimensions not supported"
 *     return 0  # return type just used as an error flag
 */
    }
  }

  /* "View.MemoryView":704
 *         if suboffset >= 0:
 *             raise ValueError, "Indirect dimensions not supported"
 *     return 0  # return type just used as an error flag             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef int assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim) except -1:             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":711
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  Py_ssize_t __pyx_v_cindex;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_3;
  char *__pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memview_slice", 1);

  /* "View.MemoryView":712
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":719
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":723
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_memview->view.ndim > 0);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(1, 723, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(1, 723, __pyx_L1_error)
  #endif

  /* "View.MemoryView":725
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  if (__pyx_t_1) {

    /* "View.MemoryView":726
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 726, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":727
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":725
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":729
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":730
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":736
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_3 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_3;

  /* "View.MemoryView":737
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_4;

  /* "View.MemoryView":742
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step, cindex
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":743
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step, cindex
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":747
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             cindex = index
 */
  __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_2 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 747, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 747, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(1, 747, __pyx_L1_error)
        #else
        __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 747, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 747, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(1, 747, __pyx_L1_error)
        #else
        __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 747, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 747, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_v_dim = __pyx_t_5;
    __pyx_t_5 = (__pyx_t_5 + 1);

    /* "View.MemoryView":748
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             cindex = index
 *             slice_memviewslice(
 */
    __pyx_t_1 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":749
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             cindex = index             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 749, __pyx_L1_error)
      __pyx_v_cindex = __pyx_t_9;

      /* "View.MemoryView":750
 *         if PyIndex_Check(index):
 *             cindex = index
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_10 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_cindex, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(1, 750, __pyx_L1_error)

      /* "View.MemoryView":748
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             cindex = index
 *             slice_memviewslice(
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_1 = (__pyx_v_index == Py_None);
    if (__pyx_t_1) {

      /* "View.MemoryView":757
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":758
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":759
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":760
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":762
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else {
        __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_t_8); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 762, __pyx_L1_error)
        __pyx_t_9 = __pyx_t_11;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_9 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_9;

      /* "View.MemoryView":763
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 763, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 763, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else {
        __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_t_8); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 763, __pyx_L1_error)
        __pyx_t_9 = __pyx_t_11;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_9 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_9;

      /* "View.MemoryView":764
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(1, 764, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else {
        __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_t_8); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 764, __pyx_L1_error)
        __pyx_t_9 = __pyx_t_11;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_9 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_9;

      /* "View.MemoryView":766
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = (__pyx_t_8 != Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":767
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = (__pyx_t_8 != Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 768, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = (__pyx_t_8 != Py_None);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":770
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_10 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(1, 770, __pyx_L1_error)

      /* "View.MemoryView":776
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":747
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             cindex = index
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  if (__pyx_t_1) {

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);

    /* "View.MemoryView":780
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 780, __pyx_L1_error) }

    /* "View.MemoryView":781
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 781, __pyx_L1_error) }

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_memoryview_type))))) __PYX_ERR(1, 779, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF((PyObject *)__pyx_r);

    /* "View.MemoryView":785
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 784, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_memoryview_type))))) __PYX_ERR(1, 784, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":711
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":793
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":813
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = (!__pyx_v_is_slice);
  if (__pyx_t_1) {

    /* "View.MemoryView":815
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = (__pyx_v_start < 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":816
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":815
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":817
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = (!__pyx_t_1);
    if (__pyx_t_2) {

      /* "View.MemoryView":818
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(PyExc_IndexError, __pyx_kp_s_Index_out_of_bounds_axis_d, __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 818, __pyx_L1_error)

      /* "View.MemoryView":817
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":813
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":821
 *     else:
 * 
 *         if have_step:             # <<<<<<<<<<<<<<
 *             negative_step = step < 0
 *             if step == 0:
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_have_step != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":822
 * 
 *         if have_step:
 *             negative_step = step < 0             # <<<<<<<<<<<<<<
 *             if step == 0:
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 */
      __pyx_v_negative_step = (__pyx_v_step < 0);

      /* "View.MemoryView":823
 *         if have_step:
 *             negative_step = step < 0
 *             if step == 0:             # <<<<<<<<<<<<<<
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 *         else:
 */
      __pyx_t_2 = (__pyx_v_step == 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":824
 *             negative_step = step < 0
 *             if step == 0:
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             negative_step = False
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(PyExc_ValueError, __pyx_kp_s_Step_may_not_be_zero_axis_d, __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 824, __pyx_L1_error)

        /* "View.MemoryView":823
 *         if have_step:
 *             negative_step = step < 0
 *             if step == 0:             # <<<<<<<<<<<<<<
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 *         else:
 */
      }

      /* "View.MemoryView":821
 *     else:
 * 
 *         if have_step:             # <<<<<<<<<<<<<<
 *             negative_step = step < 0
 *             if step == 0:
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":826
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 *         else:
 *             negative_step = False             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    /*else*/ {
      __pyx_v_negative_step = 0;

      /* "View.MemoryView":827
 *         else:
 *             negative_step = False
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;
    }
    __pyx_L6:;

    /* "View.MemoryView":830
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":831
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = (__pyx_v_start < 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":832
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":833
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = (__pyx_v_start < 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":834
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":833
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":831
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L9;
      }

      /* "View.MemoryView":835
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = (__pyx_v_start >= __pyx_v_shape);
      if (__pyx_t_2) {

        /* "View.MemoryView":836
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        if (__pyx_v_negative_step) {

          /* "View.MemoryView":837
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":836
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L11;
        }

        /* "View.MemoryView":839
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L11:;

        /* "View.MemoryView":835
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L9:;

      /* "View.MemoryView":830
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":841
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      if (__pyx_v_negative_step) {

        /* "View.MemoryView":842
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":841
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":844
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L12:;
    }
    __pyx_L8:;

    /* "View.MemoryView":846
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":847
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = (__pyx_v_stop < 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":848
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":849
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = (__pyx_v_stop < 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":850
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":849
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":847
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L14;
      }

      /* "View.MemoryView":851
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = (__pyx_v_stop > __pyx_v_shape);
      if (__pyx_t_2) {

        /* "View.MemoryView":852
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":851
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L14:;

      /* "View.MemoryView":846
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L13;
    }

    /* "View.MemoryView":854
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      if (__pyx_v_negative_step) {

        /* "View.MemoryView":855
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":854
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L16;
      }

      /* "View.MemoryView":857
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 * 
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L16:;
    }
    __pyx_L13:;

    /* "View.MemoryView":861
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":863
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":864
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":863
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":866
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = (__pyx_v_new_shape < 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":867
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":866
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":870
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":871
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":872
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":875
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_suboffset_dim[0]) < 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":876
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":875
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L19;
  }

  /* "View.MemoryView":878
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L19:;

  /* "View.MemoryView":880
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = (__pyx_v_suboffset >= 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":881
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = (!__pyx_v_is_slice);
    if (__pyx_t_2) {

      /* "View.MemoryView":882
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = (__pyx_v_new_ndim == 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":883
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(PyExc_IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":882
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L22;
      }

      /* "View.MemoryView":885
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(PyExc_IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":886
 *             else:
 *                 _err_dim(PyExc_IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(PyExc_IndexError, __pyx_kp_s_All_dimensions_preceding_dimensi, __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 885, __pyx_L1_error)
      }
      __pyx_L22:;

      /* "View.MemoryView":881
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L21;
    }

    /* "View.MemoryView":888
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L21:;

    /* "View.MemoryView":880
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":890
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":793
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":896
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pybuffer_index", 1);

  /* "View.MemoryView":898
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":899
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":902
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len // itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = (__pyx_v_view->ndim == 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":903
 * 
 *     if view.ndim == 0:
 *         shape = view.len // itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 903, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(__Pyx_UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(1, 903, __pyx_L1_error)
    }
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":904
 *     if view.ndim == 0:
 *         shape = view.len // itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":902
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len // itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":906
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":907
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":908
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = (__pyx_v_view->suboffsets != NULL);
    if (__pyx_t_2) {

      /* "View.MemoryView":909
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":908
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":911
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = (__pyx_v_index < 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":912
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":913
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
    __pyx_t_2 = (__pyx_v_index < 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":914
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = 0;
      __pyx_t_4 = 127;
      __Pyx_INCREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
      __pyx_t_1 += 37;
      __Pyx_GIVEREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Out_of_bounds_on_buffer_access_a);
      __pyx_t_5 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_dim, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u__7);
      __pyx_t_1 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__7);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__7);
      __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_builtin_IndexError, __pyx_t_5, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(1, 914, __pyx_L1_error)

      /* "View.MemoryView":913
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
    }

    /* "View.MemoryView":911
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":916
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
  __pyx_t_2 = (__pyx_v_index >= __pyx_v_shape);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":917
 * 
 *     if index >= shape:
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
    __pyx_t_1 += 37;
    __Pyx_GIVEREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Out_of_bounds_on_buffer_access_a);
    __pyx_t_3 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_dim, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_1 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__7);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u__7);
    __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_builtin_IndexError, __pyx_t_3, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 917, __pyx_L1_error)

    /* "View.MemoryView":916
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
  }

  /* "View.MemoryView":919
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":920
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = (__pyx_v_suboffset >= 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":921
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":920
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":923
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":896
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":929
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) except -1 nogil:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":930
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) except -1 nogil:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":932
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":933
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":937
 * 
 *     cdef int i, j
 *     for i in range(ndim // 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":938
 *     cdef int i, j
 *     for i in range(ndim // 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":939
 *     for i in range(ndim // 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":940
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":942
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = ((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = ((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":943
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
      __pyx_t_9 = __pyx_memoryview_err(PyExc_ValueError, __pyx_kp_s_Cannot_transpose_memoryview_with); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 943, __pyx_L1_error)

      /* "View.MemoryView":942
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":945
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":929
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) except -1 nogil:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":963
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {

  /* "View.MemoryView":964
 * 
 *     def __dealloc__(self):
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XCLEAR_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":963
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":966
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 1);

  /* "View.MemoryView":967
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->to_object_func != NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":968
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":967
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":970
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 970, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":966
 *         __PYX_XCLEAR_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":972
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 1);

  /* "View.MemoryView":973
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->to_dtype_func != NULL);
  if (__pyx_t_1) {

    /* "View.MemoryView":974
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 974, __pyx_L1_error)

    /* "View.MemoryView":973
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":976
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _get_base(self):
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":972
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":978
 *             memoryview.assign_item_from_object(self, itemp, value)
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

static PyObject *__pyx_memoryviewslice__get_base(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_base", 1);

  /* "View.MemoryView":979
 * 
 *     cdef _get_base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":978
 *             memoryview.assign_item_from_object(self, itemp, value)
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 1);

  /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_memviewslice.memview) == Py_None);
  if (__pyx_t_1) {

    /* "View.MemoryView":1008
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1013
 * 
 * 
 *     result = _memoryviewslice.__new__(_memoryviewslice, None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None)) __PYX_ERR(1, 1013, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0)) __PYX_ERR(1, 1013, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = ((PyObject *)__pyx_tp_new__memoryviewslice(((PyTypeObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1015
 *     result = _memoryviewslice.__new__(_memoryviewslice, None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1016
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview)._get_base()
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1018
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview)._get_base()             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->__pyx_vtab)->_get_base(((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1019
 * 
 *     result.from_object = (<memoryview> memviewslice.memview)._get_base()
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1021
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1022
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1023
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1024
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1025
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1028
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1030
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1033
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1036
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1037
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = (__pyx_v_suboffset >= 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1039
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1040
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1042
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1043
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1044
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1046
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1047
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1049
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 1);

  /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  if (__pyx_t_1) {

    /* "View.MemoryView":1056
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 1056, __pyx_L1_error)
    __pyx_t_2 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1057
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1059
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1060
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst) noexcept:             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;

  /* "View.MemoryView":1067
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1068
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1069
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1071
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1072
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1074
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1075
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1076
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1077
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    __pyx_t_6 = (__pyx_v_suboffsets != 0);
    if (__pyx_t_6) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst) noexcept:             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
}

/* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy", 1);

  /* "View.MemoryView":1083
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1084
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *(*__pyx_t_2)(char *);
  int (*__pyx_t_3)(char *, PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 1);

  /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  if (__pyx_t_1) {

    /* "View.MemoryView":1095
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_2 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_2;

    /* "View.MemoryView":1096
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_3;

    /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1098
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1099
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1101
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1103
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) noexcept nogil:             # <<<<<<<<<<<<<<
 *     return -arg if arg < 0 else arg
 * 
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;

  /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) noexcept nogil:
 *     return -arg if arg < 0 else arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  __pyx_t_2 = (__pyx_v_arg < 0);
  if (__pyx_t_2) {
    __pyx_t_1 = (-__pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) noexcept nogil:             # <<<<<<<<<<<<<<
 *     return -arg if arg < 0 else arg
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1113
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1118
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1119
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1121
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1122
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = ((__pyx_v_mslice->shape[__pyx_v_i]) > 1);
    if (__pyx_t_2) {

      /* "View.MemoryView":1123
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1124
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1122
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1126
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1127
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = ((__pyx_v_mslice->shape[__pyx_v_i]) > 1);
    if (__pyx_t_2) {

      /* "View.MemoryView":1128
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1129
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1127
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1131
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = (abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride));
  if (__pyx_t_2) {

    /* "View.MemoryView":1132
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1131
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1134
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1113
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1137
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;

  /* "View.MemoryView":1144
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1145
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1146
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1147
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1149
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         if (src_stride > 0 and dst_stride > 0 and
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = (__pyx_v_ndim == 1);
  if (__pyx_t_1) {

    /* "View.MemoryView":1150
 * 
 *     if ndim == 1:
 *         if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = (__pyx_v_src_stride > 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_dst_stride > 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1151
 *     if ndim == 1:
 *         if (src_stride > 0 and dst_stride > 0 and
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 *         else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1150
 * 
 *     if ndim == 1:
 *         if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1152
 *         if (src_stride > 0 and dst_stride > 0 and
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 *             memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *         else:
 *             for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1150
 * 
 *     if ndim == 1:
 *         if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1154
 *             memcpy(dst_data, src_data, itemsize * dst_extent)
 *         else:
 *             for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                 memcpy(dst_data, src_data, itemsize)
 *                 src_data += src_stride
 */
    /*else*/ {
      __pyx_t_3 = __pyx_v_dst_extent;
      __pyx_t_4 = __pyx_t_3;
      for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
        __pyx_v_i = __pyx_t_5;

        /* "View.MemoryView":1155
 *         else:
 *             for i in range(dst_extent):
 *                 memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                 src_data += src_stride
 *                 dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1156
 *             for i in range(dst_extent):
 *                 memcpy(dst_data, src_data, itemsize)
 *                 src_data += src_stride             # <<<<<<<<<<<<<<
 *                 dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1157
 *                 memcpy(dst_data, src_data, itemsize)
 *                 src_data += src_stride
 *                 dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1149
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         if (src_stride > 0 and dst_stride > 0 and
 *             <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1159
 *                 dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_3 = __pyx_v_dst_extent;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":1160
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1164
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1165
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1137
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1167
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) noexcept nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1170
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) noexcept nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1167
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) noexcept nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1174
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) noexcept nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1176
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) noexcept nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1178
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1179
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1181
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1174
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) noexcept nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1184
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) noexcept nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1193
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = (__pyx_v_order == 'F');
  if (__pyx_t_1) {

    /* "View.MemoryView":1194
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1195
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1196
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1193
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1198
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1199
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1200
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1202
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1184
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) noexcept nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1205
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":1216
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1217
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1219
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err_no_memory()
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1220
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err_no_memory()
 * 
 */
  __pyx_t_2 = (!(__pyx_v_result != 0));
  if (__pyx_t_2) {

    /* "View.MemoryView":1221
 *     result = malloc(size)
 *     if not result:
 *         _err_no_memory()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err_no_memory(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 1221, __pyx_L1_error)

    /* "View.MemoryView":1220
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err_no_memory()
 * 
 */
  }

  /* "View.MemoryView":1224
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1225
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1226
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1227
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1228
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize, ndim, order)
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1230
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize, ndim, order)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1233
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1234
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1);
    if (__pyx_t_2) {

      /* "View.MemoryView":1235
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1234
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1237
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = __pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim);
  if (__pyx_t_2) {

    /* "View.MemoryView":1238
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1237
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1240
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1242
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1205
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1247
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError, f"got differing extents in dimension {i} (got {extent1} and {extent2})"
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1249
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError, f"got differing extents in dimension {i} (got {extent1} and {extent2})"             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_got_differing_extents_in_dimensi);
  __pyx_t_2 += 35;
  __Pyx_GIVEREF(__pyx_kp_u_got_differing_extents_in_dimensi);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_got_differing_extents_in_dimensi);
  __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_i, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_got);
  __pyx_t_2 += 6;
  __Pyx_GIVEREF(__pyx_kp_u_got);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_got);
  __pyx_t_4 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_extent1, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_and);
  __pyx_t_2 += 5;
  __Pyx_GIVEREF(__pyx_kp_u_and);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_and);
  __pyx_t_4 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_extent2, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__7);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__7);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__7);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_builtin_ValueError, __pyx_t_4, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(1, 1249, __pyx_L1_error)

  /* "View.MemoryView":1247
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError, f"got differing extents in dimension {i} (got {extent1} and {extent2})"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1252
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(PyObject *error, str msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg % dim
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, PyObject *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_msg);

  /* "View.MemoryView":1253
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(PyObject *error, str msg, int dim) except -1 with gil:
 *     raise <object>error, msg % dim             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_v_msg, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(((PyObject *)__pyx_v_error), __pyx_t_2, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(1, 1253, __pyx_L1_error)

  /* "View.MemoryView":1252
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(PyObject *error, str msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg % dim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1256
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(PyObject *error, str msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg
 * 
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, PyObject *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_msg);

  /* "View.MemoryView":1257
 * @cname('__pyx_memoryview_err')
 * cdef int _err(PyObject *error, str msg) except -1 with gil:
 *     raise <object>error, msg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_no_memory')
 */
  __Pyx_Raise(((PyObject *)__pyx_v_error), __pyx_v_msg, 0, 0);
  __PYX_ERR(1, 1257, __pyx_L1_error)

  /* "View.MemoryView":1256
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(PyObject *error, str msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1260
 * 
 * @cname('__pyx_memoryview_err_no_memory')
 * cdef int _err_no_memory() except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise MemoryError
 * 
 */

static int __pyx_memoryview_err_no_memory(void) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif

  /* "View.MemoryView":1261
 * @cname('__pyx_memoryview_err_no_memory')
 * cdef int _err_no_memory() except -1 with gil:
 *     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 * 
 */
  PyErr_NoMemory(); __PYX_ERR(1, 1261, __pyx_L1_error)

  /* "View.MemoryView":1260
 * 
 * @cname('__pyx_memoryview_err_no_memory')
 * cdef int _err_no_memory() except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise MemoryError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._err_no_memory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1265
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":1273
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1274
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1276
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1277
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1278
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1281
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = (__pyx_v_src_ndim < __pyx_v_dst_ndim);
  if (__pyx_t_2) {

    /* "View.MemoryView":1282
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1281
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1283
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = (__pyx_v_dst_ndim < __pyx_v_src_ndim);
  if (__pyx_t_2) {

    /* "View.MemoryView":1284
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1283
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1286
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  __pyx_t_2 = (__pyx_t_3 > __pyx_t_4);
  if (__pyx_t_2) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1288
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1289
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = ((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i]));
    if (__pyx_t_2) {

      /* "View.MemoryView":1290
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = ((__pyx_v_src.shape[__pyx_v_i]) == 1);
      if (__pyx_t_2) {

        /* "View.MemoryView":1291
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1292
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1290
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1294
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1294, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1289
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1296
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = ((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1297
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(PyExc_ValueError, __pyx_kp_s_Dimension_d_is_not_direct, __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1297, __pyx_L1_error)

      /* "View.MemoryView":1296
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1299
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = __pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);
  if (__pyx_t_2) {

    /* "View.MemoryView":1301
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = (!__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim));
    if (__pyx_t_2) {

      /* "View.MemoryView":1302
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1301
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1304
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(1, 1304, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1305
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1299
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1307
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (!__pyx_v_broadcasting);
  if (__pyx_t_2) {

    /* "View.MemoryView":1310
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = __pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim);
    if (__pyx_t_2) {

      /* "View.MemoryView":1311
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1310
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1312
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = __pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim);
    if (__pyx_t_2) {

      /* "View.MemoryView":1313
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1312
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1315
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 */
    if (__pyx_v_direct_copy) {

      /* "View.MemoryView":1317
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1318
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1319
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1320
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1321
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1315
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 */
    }

    /* "View.MemoryView":1307
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1323
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  if (__pyx_t_2) {

    /* "View.MemoryView":1326
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 1326, __pyx_L1_error)

    /* "View.MemoryView":1327
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 1327, __pyx_L1_error)

    /* "View.MemoryView":1323
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1329
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1330
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1331
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1333
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1334
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1265
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1337
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) noexcept nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1341
 *                             int ndim_other) noexcept nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1343
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1344
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1345
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1346
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1348
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1349
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1350
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1351
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1337
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) noexcept nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1359
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) noexcept nogil:             # <<<<<<<<<<<<<<
 * 
 *     if dtype_is_object:
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {

  /* "View.MemoryView":1361
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) noexcept nogil:
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape, dst.strides, ndim, inc)
 * 
 */
  if (__pyx_v_dtype_is_object) {

    /* "View.MemoryView":1362
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape, dst.strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1361
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) noexcept nogil:
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape, dst.strides, ndim, inc)
 * 
 */
  }

  /* "View.MemoryView":1359
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) noexcept nogil:             # <<<<<<<<<<<<<<
 * 
 *     if dtype_is_object:
 */

  /* function exit code */
}

/* "View.MemoryView":1365
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) noexcept with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif

  /* "View.MemoryView":1368
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) noexcept with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1365
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) noexcept with gil:
 */

  /* function exit code */
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc) noexcept:
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1374
 *                                     Py_ssize_t *strides, int ndim, bint inc) noexcept:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(shape[0]):
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1376
 *     cdef Py_ssize_t stride = strides[0]
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1377
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = (__pyx_v_ndim == 1);
    if (__pyx_t_4) {

      /* "View.MemoryView":1378
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      if (__pyx_v_inc) {

        /* "View.MemoryView":1379
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1378
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1381
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1, ndim - 1, inc)
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1377
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1383
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1, ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += stride
 */
    /*else*/ {
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1385
 *             refcount_objects_in_slice(data, shape + 1, strides + 1, ndim - 1, inc)
 * 
 *         data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
  }

  /* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc) noexcept:
 *     cdef Py_ssize_t i
 */

  /* function exit code */
}

/* "View.MemoryView":1391
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) noexcept nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1394
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) noexcept nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, inc=False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim, itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, inc=True)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1395
 *                               bint dtype_is_object) noexcept nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, inc=False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim, itemsize, item)             # <<<<<<<<<<<<<<
 *     refcount_copying(dst, dtype_is_object, ndim, inc=True)
 * 
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1396
 *     refcount_copying(dst, dtype_is_object, ndim, inc=False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim, itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, inc=True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1391
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) noexcept nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1400
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) noexcept nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1404
 *                               size_t itemsize, void *item) noexcept nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1405
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1407
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = (__pyx_v_ndim == 1);
  if (__pyx_t_1) {

    /* "View.MemoryView":1408
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1409
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1410
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1407
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1412
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1, ndim - 1, itemsize, item)
 *             data += stride
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1413
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1, ndim - 1, itemsize, item)             # <<<<<<<<<<<<<<
 *             data += stride
 * 
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1414
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1, ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1400
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) noexcept nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__8, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 > 1);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_update); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self):             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self):
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self):             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self):             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self):
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self):             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "time.pxd":20
 * 
 * 
 * cdef inline double time() nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         _PyTime_t tic
 */

static CYTHON_INLINE double __pyx_f_7cpython_4time_time(void) {
  _PyTime_t __pyx_v_tic;
  double __pyx_r;

  /* "time.pxd":24
 *         _PyTime_t tic
 * 
 *     tic = _PyTime_GetSystemClock()             # <<<<<<<<<<<<<<
 *     return _PyTime_AsSecondsDouble(tic)
 * 
 */
  __pyx_v_tic = _PyTime_GetSystemClock();

  /* "time.pxd":25
 * 
 *     tic = _PyTime_GetSystemClock()
 *     return _PyTime_AsSecondsDouble(tic)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = _PyTime_AsSecondsDouble(__pyx_v_tic);
  goto __pyx_L0;

  /* "time.pxd":20
 * 
 * 
 * cdef inline double time() nogil:             # <<<<<<<<<<<<<<
 *     cdef:
 *         _PyTime_t tic
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "time.pxd":28
 * 
 * 
 * cdef inline int _raise_from_errno() except -1 with gil:             # <<<<<<<<<<<<<<
 *     PyErr_SetFromErrno(RuntimeError)
 *     return <int> -1  # Let the C compiler know that this function always raises.
 */

static CYTHON_INLINE int __pyx_f_7cpython_4time__raise_from_errno(void) {
  int __pyx_r;
  PyObject *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif

  /* "time.pxd":29
 * 
 * cdef inline int _raise_from_errno() except -1 with gil:
 *     PyErr_SetFromErrno(RuntimeError)             # <<<<<<<<<<<<<<
 *     return <int> -1  # Let the C compiler know that this function always raises.
 * 
 */
  __pyx_t_1 = PyErr_SetFromErrno(__pyx_builtin_RuntimeError); if (unlikely(__pyx_t_1 == ((PyObject *)NULL))) __PYX_ERR(3, 29, __pyx_L1_error)

  /* "time.pxd":30
 * cdef inline int _raise_from_errno() except -1 with gil:
 *     PyErr_SetFromErrno(RuntimeError)
 *     return <int> -1  # Let the C compiler know that this function always raises.             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((int)-1L);
  goto __pyx_L0;

  /* "time.pxd":28
 * 
 * 
 * cdef inline int _raise_from_errno() except -1 with gil:             # <<<<<<<<<<<<<<
 *     PyErr_SetFromErrno(RuntimeError)
 *     return <int> -1  # Let the C compiler know that this function always raises.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.time._raise_from_errno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "time.pxd":33
 * 
 * 
 * cdef inline tm localtime() except * nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Analogue to the stdlib time.localtime.  The returned struct
 */

static CYTHON_INLINE struct tm __pyx_f_7cpython_4time_localtime(void) {
  time_t __pyx_v_tic;
  struct tm *__pyx_v_result;
  struct tm __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "time.pxd":39
 *     """
 *     cdef:
 *         time_t tic = <time_t>time()             # <<<<<<<<<<<<<<
 *         tm* result
 * 
 */
  __pyx_v_tic = ((time_t)__pyx_f_7cpython_4time_time());

  /* "time.pxd":42
 *         tm* result
 * 
 *     result = libc_localtime(&tic)             # <<<<<<<<<<<<<<
 *     if result is NULL:
 *         _raise_from_errno()
 */
  __pyx_v_result = localtime((&__pyx_v_tic));

  /* "time.pxd":43
 * 
 *     result = libc_localtime(&tic)
 *     if result is NULL:             # <<<<<<<<<<<<<<
 *         _raise_from_errno()
 *     # Fix 0-based date values (and the 1900-based year).
 */
  __pyx_t_1 = (__pyx_v_result == NULL);
  if (__pyx_t_1) {

    /* "time.pxd":44
 *     result = libc_localtime(&tic)
 *     if result is NULL:
 *         _raise_from_errno()             # <<<<<<<<<<<<<<
 *     # Fix 0-based date values (and the 1900-based year).
 *     # See tmtotuple() in https://github.com/python/cpython/blob/master/Modules/timemodule.c
 */
    __pyx_t_2 = __pyx_f_7cpython_4time__raise_from_errno(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(3, 44, __pyx_L1_error)

    /* "time.pxd":43
 * 
 *     result = libc_localtime(&tic)
 *     if result is NULL:             # <<<<<<<<<<<<<<
 *         _raise_from_errno()
 *     # Fix 0-based date values (and the 1900-based year).
 */
  }

  /* "time.pxd":47
 *     # Fix 0-based date values (and the 1900-based year).
 *     # See tmtotuple() in https://github.com/python/cpython/blob/master/Modules/timemodule.c
 *     result.tm_year += 1900             # <<<<<<<<<<<<<<
 *     result.tm_mon += 1
 *     result.tm_wday = (result.tm_wday + 6) % 7
 */
  __pyx_v_result->tm_year = (__pyx_v_result->tm_year + 0x76C);

  /* "time.pxd":48
 *     # See tmtotuple() in https://github.com/python/cpython/blob/master/Modules/timemodule.c
 *     result.tm_year += 1900
 *     result.tm_mon += 1             # <<<<<<<<<<<<<<
 *     result.tm_wday = (result.tm_wday + 6) % 7
 *     result.tm_yday += 1
 */
  __pyx_v_result->tm_mon = (__pyx_v_result->tm_mon + 1);

  /* "time.pxd":49
 *     result.tm_year += 1900
 *     result.tm_mon += 1
 *     result.tm_wday = (result.tm_wday + 6) % 7             # <<<<<<<<<<<<<<
 *     result.tm_yday += 1
 *     return result[0]
 */
  __pyx_v_result->tm_wday = ((__pyx_v_result->tm_wday + 6) % 7);

  /* "time.pxd":50
 *     result.tm_mon += 1
 *     result.tm_wday = (result.tm_wday + 6) % 7
 *     result.tm_yday += 1             # <<<<<<<<<<<<<<
 *     return result[0]
 */
  __pyx_v_result->tm_yday = (__pyx_v_result->tm_yday + 1);

  /* "time.pxd":51
 *     result.tm_wday = (result.tm_wday + 6) % 7
 *     result.tm_yday += 1
 *     return result[0]             # <<<<<<<<<<<<<<
 */
  __pyx_r = (__pyx_v_result[0]);
  goto __pyx_L0;

  /* "time.pxd":33
 * 
 * 
 * cdef inline tm localtime() except * nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Analogue to the stdlib time.localtime.  The returned struct
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("cpython.time.localtime", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/array.pxd":104
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
CYTHON_UNUSED static int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
CYTHON_UNUSED static int __pyx_pw_7cpython_5array_5array_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7cpython_5array_5array___getbuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7cpython_5array_5array___getbuffer__(arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_item_count;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  char __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (unlikely(__pyx_v_info == NULL)) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "cpython/array.pxd":109
 *             # In particular strided access is always provided regardless
 *             # of flags
 *             item_count = Py_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *             info.suboffsets = NULL
 */
  __pyx_v_item_count = Py_SIZE(((PyObject *)__pyx_v_self));

  /* "cpython/array.pxd":111
 *             item_count = Py_SIZE(self)
 * 
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "cpython/array.pxd":112
 * 
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars             # <<<<<<<<<<<<<<
 *             info.readonly = 0
 *             info.ndim = 1
 */
  __pyx_t_1 = __pyx_v_self->data.as_chars;
  __pyx_v_info->buf = __pyx_t_1;

  /* "cpython/array.pxd":113
 *             info.suboffsets = NULL
 *             info.buf = self.data.as_chars
 *             info.readonly = 0             # <<<<<<<<<<<<<<
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 */
  __pyx_v_info->readonly = 0;

  /* "cpython/array.pxd":114
 *             info.buf = self.data.as_chars
 *             info.readonly = 0
 *             info.ndim = 1             # <<<<<<<<<<<<<<
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count
 */
  __pyx_v_info->ndim = 1;

  /* "cpython/array.pxd":115
 *             info.readonly = 0
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)             # <<<<<<<<<<<<<<
 *             info.len = info.itemsize * item_count
 * 
 */
  __pyx_t_2 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_info->itemsize = __pyx_t_2;

  /* "cpython/array.pxd":116
 *             info.ndim = 1
 *             info.itemsize = self.ob_descr.itemsize   # e.g. sizeof(float)
 *             info.len = info.itemsize * item_count             # <<<<<<<<<<<<<<
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 */
  __pyx_v_info->len = (__pyx_v_info->itemsize * __pyx_v_item_count);

  /* "cpython/array.pxd":118
 *             info.len = info.itemsize * item_count
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)             # <<<<<<<<<<<<<<
 *             if not info.shape:
 *                 raise MemoryError()
 */
  __pyx_v_info->shape = ((Py_ssize_t *)PyObject_Malloc(((sizeof(Py_ssize_t)) + 2)));

  /* "cpython/array.pxd":119
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  __pyx_t_3 = (!(__pyx_v_info->shape != 0));
  if (unlikely(__pyx_t_3)) {

    /* "cpython/array.pxd":120
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize
 */
    PyErr_NoMemory(); __PYX_ERR(4, 120, __pyx_L1_error)

    /* "cpython/array.pxd":119
 * 
 *             info.shape = <Py_ssize_t*> PyObject_Malloc(sizeof(Py_ssize_t) + 2)
 *             if not info.shape:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 */
  }

  /* "cpython/array.pxd":121
 *             if not info.shape:
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing             # <<<<<<<<<<<<<<
 *             info.strides = &info.itemsize
 * 
 */
  (__pyx_v_info->shape[0]) = __pyx_v_item_count;

  /* "cpython/array.pxd":122
 *                 raise MemoryError()
 *             info.shape[0] = item_count      # constant regardless of resizing
 *             info.strides = &info.itemsize             # <<<<<<<<<<<<<<
 * 
 *             info.format = <char*> (info.shape + 1)
 */
  __pyx_v_info->strides = (&__pyx_v_info->itemsize);

  /* "cpython/array.pxd":124
 *             info.strides = &info.itemsize
 * 
 *             info.format = <char*> (info.shape + 1)             # <<<<<<<<<<<<<<
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 */
  __pyx_v_info->format = ((char *)(__pyx_v_info->shape + 1));

  /* "cpython/array.pxd":125
 * 
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode             # <<<<<<<<<<<<<<
 *             info.format[1] = 0
 *             info.obj = self
 */
  __pyx_t_4 = __pyx_v_self->ob_descr->typecode;
  (__pyx_v_info->format[0]) = __pyx_t_4;

  /* "cpython/array.pxd":126
 *             info.format = <char*> (info.shape + 1)
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0             # <<<<<<<<<<<<<<
 *             info.obj = self
 * 
 */
  (__pyx_v_info->format[1]) = 0;

  /* "cpython/array.pxd":127
 *             info.format[0] = self.ob_descr.typecode
 *             info.format[1] = 0
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "cpython/array.pxd":104
 *             __data_union data
 * 
 *         def __getbuffer__(self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/array.pxd":129
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

/* Python wrapper */
CYTHON_UNUSED static void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
CYTHON_UNUSED static void __pyx_pw_7cpython_5array_5array_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7cpython_5array_5array_2__releasebuffer__(((arrayobject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7cpython_5array_5array_2__releasebuffer__(CYTHON_UNUSED arrayobject *__pyx_v_self, Py_buffer *__pyx_v_info) {

  /* "cpython/array.pxd":130
 * 
 *         def __releasebuffer__(self, Py_buffer* info):
 *             PyObject_Free(info.shape)             # <<<<<<<<<<<<<<
 * 
 *     array newarrayobject(PyTypeObject* type, Py_ssize_t size, arraydescr *descr)
 */
  PyObject_Free(__pyx_v_info->shape);

  /* "cpython/array.pxd":129
 *             info.obj = self
 * 
 *         def __releasebuffer__(self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             PyObject_Free(info.shape)
 * 
 */

  /* function exit code */
}

/* "cpython/array.pxd":141
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_clone(arrayobject *__pyx_v_template, Py_ssize_t __pyx_v_length, int __pyx_v_zero) {
  arrayobject *__pyx_v_op = 0;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 1);

  /* "cpython/array.pxd":145
 *     type will be same as template.
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)             # <<<<<<<<<<<<<<
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_template)), __pyx_v_length, __pyx_v_template->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cpython/array.pxd":146
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  if (__pyx_v_zero) {
  } else {
    __pyx_t_2 = __pyx_v_zero;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((PyObject *)__pyx_v_op) != Py_None);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "cpython/array.pxd":147
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
    (void)(memset(__pyx_v_op->data.as_chars, 0, (__pyx_v_length * __pyx_v_op->ob_descr->itemsize)));

    /* "cpython/array.pxd":146
 *     if zero is true, new array will be initialized with zeroes."""
 *     cdef array op = newarrayobject(Py_TYPE(template), length, template.ob_descr)
 *     if zero and op is not None:             # <<<<<<<<<<<<<<
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op
 */
  }

  /* "cpython/array.pxd":148
 *     if zero and op is not None:
 *         memset(op.data.as_chars, 0, length * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline array copy(array self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_op);
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "cpython/array.pxd":141
 * 
 * 
 * cdef inline array clone(array template, Py_ssize_t length, bint zero):             # <<<<<<<<<<<<<<
 *     """ fast creation of a new array, given a template array.
 *     type will be same as template.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/array.pxd":150
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

static CYTHON_INLINE arrayobject *__pyx_f_7cpython_5array_copy(arrayobject *__pyx_v_self) {
  arrayobject *__pyx_v_op = 0;
  arrayobject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "cpython/array.pxd":152
 * cdef inline array copy(array self):
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)             # <<<<<<<<<<<<<<
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op
 */
  __pyx_t_1 = ((PyObject *)newarrayobject(Py_TYPE(((PyObject *)__pyx_v_self)), Py_SIZE(((PyObject *)__pyx_v_self)), __pyx_v_self->ob_descr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_op = ((arrayobject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cpython/array.pxd":153
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 *     return op
 * 
 */
  (void)(memcpy(__pyx_v_op->data.as_chars, __pyx_v_self->data.as_chars, (Py_SIZE(((PyObject *)__pyx_v_op)) * __pyx_v_op->ob_descr->itemsize)));

  /* "cpython/array.pxd":154
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 *     memcpy(op.data.as_chars, self.data.as_chars, Py_SIZE(op) * op.ob_descr.itemsize)
 *     return op             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_op);
  __pyx_r = __pyx_v_op;
  goto __pyx_L0;

  /* "cpython/array.pxd":150
 *     return op
 * 
 * cdef inline array copy(array self):             # <<<<<<<<<<<<<<
 *     """ make a copy of an array. """
 *     cdef array op = newarrayobject(Py_TYPE(self), Py_SIZE(self), self.ob_descr)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.array.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_op);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/array.pxd":156
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend_buffer(arrayobject *__pyx_v_self, char *__pyx_v_stuff, Py_ssize_t __pyx_v_n) {
  Py_ssize_t __pyx_v_itemsize;
  Py_ssize_t __pyx_v_origsize;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cpython/array.pxd":160
 *     (e.g. of same array type)
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 */
  __pyx_t_1 = __pyx_v_self->ob_descr->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "cpython/array.pxd":161
 *     n: number of elements (not number of bytes!) """
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)             # <<<<<<<<<<<<<<
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 */
  __pyx_v_origsize = Py_SIZE(((PyObject *)__pyx_v_self));

  /* "cpython/array.pxd":162
 *     cdef Py_ssize_t itemsize = self.ob_descr.itemsize
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)             # <<<<<<<<<<<<<<
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0
 */
  __pyx_t_1 = resize_smart(__pyx_v_self, (__pyx_v_origsize + __pyx_v_n)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(4, 162, __pyx_L1_error)

  /* "cpython/array.pxd":163
 *     cdef Py_ssize_t origsize = Py_SIZE(self)
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  (void)(memcpy((__pyx_v_self->data.as_chars + (__pyx_v_origsize * __pyx_v_itemsize)), __pyx_v_stuff, (__pyx_v_n * __pyx_v_itemsize)));

  /* "cpython/array.pxd":164
 *     resize_smart(self, origsize + n)
 *     memcpy(self.data.as_chars + origsize * itemsize, stuff, n * itemsize)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef inline int extend(array self, array other) except -1:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "cpython/array.pxd":156
 *     return op
 * 
 * cdef inline int extend_buffer(array self, char* stuff, Py_ssize_t n) except -1:             # <<<<<<<<<<<<<<
 *     """ efficient appending of new stuff of same type
 *     (e.g. of same array type)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/array.pxd":166
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

static CYTHON_INLINE int __pyx_f_7cpython_5array_extend(arrayobject *__pyx_v_self, arrayobject *__pyx_v_other) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "cpython/array.pxd":168
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  __pyx_t_1 = (__pyx_v_self->ob_descr->typecode != __pyx_v_other->ob_descr->typecode);
  if (__pyx_t_1) {

    /* "cpython/array.pxd":169
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()             # <<<<<<<<<<<<<<
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 */
    __pyx_t_2 = PyErr_BadArgument(); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(4, 169, __pyx_L1_error)

    /* "cpython/array.pxd":168
 * cdef inline int extend(array self, array other) except -1:
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:             # <<<<<<<<<<<<<<
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 */
  }

  /* "cpython/array.pxd":170
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 *         PyErr_BadArgument()
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))             # <<<<<<<<<<<<<<
 * 
 * cdef inline void zero(array self):
 */
  __pyx_t_2 = __pyx_f_7cpython_5array_extend_buffer(__pyx_v_self, __pyx_v_other->data.as_chars, Py_SIZE(((PyObject *)__pyx_v_other))); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(4, 170, __pyx_L1_error)
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "cpython/array.pxd":166
 *     return 0
 * 
 * cdef inline int extend(array self, array other) except -1:             # <<<<<<<<<<<<<<
 *     """ extend array with data from another array; types must match. """
 *     if self.ob_descr.typecode != other.ob_descr.typecode:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("cpython.array.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/array.pxd":172
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

static CYTHON_INLINE void __pyx_f_7cpython_5array_zero(arrayobject *__pyx_v_self) {

  /* "cpython/array.pxd":174
 * cdef inline void zero(array self):
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)             # <<<<<<<<<<<<<<
 */
  (void)(memset(__pyx_v_self->data.as_chars, 0, (Py_SIZE(((PyObject *)__pyx_v_self)) * __pyx_v_self->ob_descr->itemsize)));

  /* "cpython/array.pxd":172
 *     return extend_buffer(self, other.data.as_chars, Py_SIZE(other))
 * 
 * cdef inline void zero(array self):             # <<<<<<<<<<<<<<
 *     """ set all elements of array to zero. """
 *     memset(self.data.as_chars, 0, Py_SIZE(self) * self.ob_descr.itemsize)
 */

  /* function exit code */
}

/* "pyboy/core/opcodes.py":15
 * 
 * 
 * def NOP_00(cpu): # 00 NOP             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_NOP_00(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":16
 * 
 * def NOP_00(cpu): # 00 NOP
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":17
 * def NOP_00(cpu): # 00 NOP
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":18
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":15
 * 
 * 
 * def NOP_00(cpu): # 00 NOP             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":21
 * 
 * 
 * def LD_01(cpu, v): # 01 LD BC,d16             # <<<<<<<<<<<<<<
 *     cpu.set_bc(v)
 *     cpu.PC += 3
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_01(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":22
 * 
 * def LD_01(cpu, v): # 01 LD BC,d16
 *     cpu.set_bc(v)             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU *)__pyx_v_cpu->__pyx_vtab)->set_bc(__pyx_v_cpu, __pyx_v_v);

  /* "pyboy/core/opcodes.py":23
 * def LD_01(cpu, v): # 01 LD BC,d16
 *     cpu.set_bc(v)
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":24
 *     cpu.set_bc(v)
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":25
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":21
 * 
 * 
 * def LD_01(cpu, v): # 01 LD BC,d16             # <<<<<<<<<<<<<<
 *     cpu.set_bc(v)
 *     cpu.PC += 3
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":28
 * 
 * 
 * def LD_02(cpu): # 02 LD (BC),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(((cpu.B << 8) + cpu.C), cpu.A)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_02(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":29
 * 
 * def LD_02(cpu): # 02 LD (BC),A
 *     cpu.mb.setitem(((cpu.B << 8) + cpu.C), cpu.A)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->B << 8) + __pyx_v_cpu->C), __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":30
 * def LD_02(cpu): # 02 LD (BC),A
 *     cpu.mb.setitem(((cpu.B << 8) + cpu.C), cpu.A)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":31
 *     cpu.mb.setitem(((cpu.B << 8) + cpu.C), cpu.A)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":32
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":28
 * 
 * 
 * def LD_02(cpu): # 02 LD (BC),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(((cpu.B << 8) + cpu.C), cpu.A)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":35
 * 
 * 
 * def INC_03(cpu): # 03 INC BC             # <<<<<<<<<<<<<<
 *     t = ((cpu.B << 8) + cpu.C) + 1
 *     # No flag operations
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_03(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":36
 * 
 * def INC_03(cpu): # 03 INC BC
 *     t = ((cpu.B << 8) + cpu.C) + 1             # <<<<<<<<<<<<<<
 *     # No flag operations
 *     t &= 0xFFFF
 */
  __pyx_v_t = (((__pyx_v_cpu->B << 8) + __pyx_v_cpu->C) + 1);

  /* "pyboy/core/opcodes.py":38
 *     t = ((cpu.B << 8) + cpu.C) + 1
 *     # No flag operations
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.set_bc(t)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":39
 *     # No flag operations
 *     t &= 0xFFFF
 *     cpu.set_bc(t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU *)__pyx_v_cpu->__pyx_vtab)->set_bc(__pyx_v_cpu, __pyx_v_t);

  /* "pyboy/core/opcodes.py":40
 *     t &= 0xFFFF
 *     cpu.set_bc(t)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":41
 *     cpu.set_bc(t)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":42
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":35
 * 
 * 
 * def INC_03(cpu): # 03 INC BC             # <<<<<<<<<<<<<<
 *     t = ((cpu.B << 8) + cpu.C) + 1
 *     # No flag operations
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":45
 * 
 * 
 * def INC_04(cpu): # 04 INC B             # <<<<<<<<<<<<<<
 *     t = cpu.B + 1
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_04(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":46
 * 
 * def INC_04(cpu): # 04 INC B
 *     t = cpu.B + 1             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B + 1);

  /* "pyboy/core/opcodes.py":47
 * def INC_04(cpu): # 04 INC B
 *     t = cpu.B + 1
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.B & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":48
 *     t = cpu.B + 1
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.B & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":49
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.B & 0xF) + (1 & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->B & 0xF) + 0x1) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":50
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.B & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":51
 *     flag += (((cpu.B & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":52
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":53
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":54
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":55
 *     cpu.B = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":56
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":45
 * 
 * 
 * def INC_04(cpu): # 04 INC B             # <<<<<<<<<<<<<<
 *     t = cpu.B + 1
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":59
 * 
 * 
 * def DEC_05(cpu): # 05 DEC B             # <<<<<<<<<<<<<<
 *     t = cpu.B - 1
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_05(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":60
 * 
 * def DEC_05(cpu): # 05 DEC B
 *     t = cpu.B - 1             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B - 1);

  /* "pyboy/core/opcodes.py":61
 * def DEC_05(cpu): # 05 DEC B
 *     t = cpu.B - 1
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.B & 0xF) - (1 & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":62
 *     t = cpu.B - 1
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.B & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":63
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.B & 0xF) - (1 & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->B & 0xF) - 0x1) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":64
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.B & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":65
 *     flag += (((cpu.B & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":66
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":67
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":68
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":69
 *     cpu.B = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":70
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":59
 * 
 * 
 * def DEC_05(cpu): # 05 DEC B             # <<<<<<<<<<<<<<
 *     t = cpu.B - 1
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":73
 * 
 * 
 * def LD_06(cpu, v): # 06 LD B,d8             # <<<<<<<<<<<<<<
 *     cpu.B = v
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_06(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":74
 * 
 * def LD_06(cpu, v): # 06 LD B,d8
 *     cpu.B = v             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_v;

  /* "pyboy/core/opcodes.py":75
 * def LD_06(cpu, v): # 06 LD B,d8
 *     cpu.B = v
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":76
 *     cpu.B = v
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":77
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":73
 * 
 * 
 * def LD_06(cpu, v): # 06 LD B,d8             # <<<<<<<<<<<<<<
 *     cpu.B = v
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":80
 * 
 * 
 * def RLCA_07(cpu): # 07 RLCA             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1) + (cpu.A >> 7)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLCA_07(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":81
 * 
 * def RLCA_07(cpu): # 07 RLCA
 *     t = (cpu.A << 1) + (cpu.A >> 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_t = ((__pyx_v_cpu->A << 1) + (__pyx_v_cpu->A >> 7));

  /* "pyboy/core/opcodes.py":82
 * def RLCA_07(cpu): # 07 RLCA
 *     t = (cpu.A << 1) + (cpu.A >> 7)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":83
 *     t = (cpu.A << 1) + (cpu.A >> 7)
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":84
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":85
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":86
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":87
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":88
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":89
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":90
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":80
 * 
 * 
 * def RLCA_07(cpu): # 07 RLCA             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1) + (cpu.A >> 7)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":93
 * 
 * 
 * def LD_08(cpu, v): # 08 LD (a16),SP             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(v, cpu.SP & 0xFF)
 *     cpu.mb.setitem(v+1, cpu.SP >> 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_08(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":94
 * 
 * def LD_08(cpu, v): # 08 LD (a16),SP
 *     cpu.mb.setitem(v, cpu.SP & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(v+1, cpu.SP >> 8)
 *     cpu.PC += 3
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_v, (__pyx_v_cpu->SP & 0xFF));

  /* "pyboy/core/opcodes.py":95
 * def LD_08(cpu, v): # 08 LD (a16),SP
 *     cpu.mb.setitem(v, cpu.SP & 0xFF)
 *     cpu.mb.setitem(v+1, cpu.SP >> 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, (__pyx_v_v + 1), (__pyx_v_cpu->SP >> 8));

  /* "pyboy/core/opcodes.py":96
 *     cpu.mb.setitem(v, cpu.SP & 0xFF)
 *     cpu.mb.setitem(v+1, cpu.SP >> 8)
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 20
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":97
 *     cpu.mb.setitem(v+1, cpu.SP >> 8)
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 20
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":98
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     return 20             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 20;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":93
 * 
 * 
 * def LD_08(cpu, v): # 08 LD (a16),SP             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(v, cpu.SP & 0xFF)
 *     cpu.mb.setitem(v+1, cpu.SP >> 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":101
 * 
 * 
 * def ADD_09(cpu): # 09 ADD HL,BC             # <<<<<<<<<<<<<<
 *     t = cpu.HL + ((cpu.B << 8) + cpu.C)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_09(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":102
 * 
 * def ADD_09(cpu): # 09 ADD HL,BC
 *     t = cpu.HL + ((cpu.B << 8) + cpu.C)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.B << 8) + cpu.C) & 0xFFF)) > 0xFFF) << FLAGH
 */
  __pyx_v_t = (__pyx_v_cpu->HL + ((__pyx_v_cpu->B << 8) + __pyx_v_cpu->C));

  /* "pyboy/core/opcodes.py":103
 * def ADD_09(cpu): # 09 ADD HL,BC
 *     t = cpu.HL + ((cpu.B << 8) + cpu.C)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.B << 8) + cpu.C) & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":104
 *     t = cpu.HL + ((cpu.B << 8) + cpu.C)
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.B << 8) + cpu.C) & 0xFFF)) > 0xFFF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->HL & 0xFFF) + (((__pyx_v_cpu->B << 8) + __pyx_v_cpu->C) & 0xFFF)) > 0xFFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":105
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.B << 8) + cpu.C) & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFFFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":106
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.B << 8) + cpu.C) & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 128);

  /* "pyboy/core/opcodes.py":107
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFFFF
 *     cpu.HL = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":108
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.HL = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":109
 *     cpu.F |= flag
 *     t &= 0xFFFF
 *     cpu.HL = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = __pyx_v_t;

  /* "pyboy/core/opcodes.py":110
 *     t &= 0xFFFF
 *     cpu.HL = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":111
 *     cpu.HL = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":112
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":101
 * 
 * 
 * def ADD_09(cpu): # 09 ADD HL,BC             # <<<<<<<<<<<<<<
 *     t = cpu.HL + ((cpu.B << 8) + cpu.C)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":115
 * 
 * 
 * def LD_0A(cpu): # 0A LD A,(BC)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(((cpu.B << 8) + cpu.C))
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_0A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":116
 * 
 * def LD_0A(cpu): # 0A LD A,(BC)
 *     cpu.A = cpu.mb.getitem(((cpu.B << 8) + cpu.C))             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->B << 8) + __pyx_v_cpu->C));

  /* "pyboy/core/opcodes.py":117
 * def LD_0A(cpu): # 0A LD A,(BC)
 *     cpu.A = cpu.mb.getitem(((cpu.B << 8) + cpu.C))
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":118
 *     cpu.A = cpu.mb.getitem(((cpu.B << 8) + cpu.C))
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":119
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":115
 * 
 * 
 * def LD_0A(cpu): # 0A LD A,(BC)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(((cpu.B << 8) + cpu.C))
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":122
 * 
 * 
 * def DEC_0B(cpu): # 0B DEC BC             # <<<<<<<<<<<<<<
 *     t = ((cpu.B << 8) + cpu.C) - 1
 *     # No flag operations
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_0B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":123
 * 
 * def DEC_0B(cpu): # 0B DEC BC
 *     t = ((cpu.B << 8) + cpu.C) - 1             # <<<<<<<<<<<<<<
 *     # No flag operations
 *     t &= 0xFFFF
 */
  __pyx_v_t = (((__pyx_v_cpu->B << 8) + __pyx_v_cpu->C) - 1);

  /* "pyboy/core/opcodes.py":125
 *     t = ((cpu.B << 8) + cpu.C) - 1
 *     # No flag operations
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.set_bc(t)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":126
 *     # No flag operations
 *     t &= 0xFFFF
 *     cpu.set_bc(t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU *)__pyx_v_cpu->__pyx_vtab)->set_bc(__pyx_v_cpu, __pyx_v_t);

  /* "pyboy/core/opcodes.py":127
 *     t &= 0xFFFF
 *     cpu.set_bc(t)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":128
 *     cpu.set_bc(t)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":129
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":122
 * 
 * 
 * def DEC_0B(cpu): # 0B DEC BC             # <<<<<<<<<<<<<<
 *     t = ((cpu.B << 8) + cpu.C) - 1
 *     # No flag operations
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":132
 * 
 * 
 * def INC_0C(cpu): # 0C INC C             # <<<<<<<<<<<<<<
 *     t = cpu.C + 1
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_0C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":133
 * 
 * def INC_0C(cpu): # 0C INC C
 *     t = cpu.C + 1             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C + 1);

  /* "pyboy/core/opcodes.py":134
 * def INC_0C(cpu): # 0C INC C
 *     t = cpu.C + 1
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.C & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":135
 *     t = cpu.C + 1
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.C & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":136
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.C & 0xF) + (1 & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->C & 0xF) + 0x1) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":137
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.C & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":138
 *     flag += (((cpu.C & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":139
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":140
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":141
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":142
 *     cpu.C = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":143
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":132
 * 
 * 
 * def INC_0C(cpu): # 0C INC C             # <<<<<<<<<<<<<<
 *     t = cpu.C + 1
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":146
 * 
 * 
 * def DEC_0D(cpu): # 0D DEC C             # <<<<<<<<<<<<<<
 *     t = cpu.C - 1
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_0D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":147
 * 
 * def DEC_0D(cpu): # 0D DEC C
 *     t = cpu.C - 1             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C - 1);

  /* "pyboy/core/opcodes.py":148
 * def DEC_0D(cpu): # 0D DEC C
 *     t = cpu.C - 1
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.C & 0xF) - (1 & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":149
 *     t = cpu.C - 1
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.C & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":150
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.C & 0xF) - (1 & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->C & 0xF) - 0x1) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":151
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.C & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":152
 *     flag += (((cpu.C & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":153
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":154
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":155
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":156
 *     cpu.C = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":157
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":146
 * 
 * 
 * def DEC_0D(cpu): # 0D DEC C             # <<<<<<<<<<<<<<
 *     t = cpu.C - 1
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":160
 * 
 * 
 * def LD_0E(cpu, v): # 0E LD C,d8             # <<<<<<<<<<<<<<
 *     cpu.C = v
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_0E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":161
 * 
 * def LD_0E(cpu, v): # 0E LD C,d8
 *     cpu.C = v             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_v;

  /* "pyboy/core/opcodes.py":162
 * def LD_0E(cpu, v): # 0E LD C,d8
 *     cpu.C = v
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":163
 *     cpu.C = v
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":164
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":160
 * 
 * 
 * def LD_0E(cpu, v): # 0E LD C,d8             # <<<<<<<<<<<<<<
 *     cpu.C = v
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":167
 * 
 * 
 * def RRCA_0F(cpu): # 0F RRCA             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRCA_0F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":168
 * 
 * def RRCA_0F(cpu): # 0F RRCA
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_t = (((__pyx_v_cpu->A >> 1) + ((__pyx_v_cpu->A & 1) << 7)) + ((__pyx_v_cpu->A & 1) << 8));

  /* "pyboy/core/opcodes.py":169
 * def RRCA_0F(cpu): # 0F RRCA
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":170
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":171
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":172
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":173
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":174
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":175
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":176
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":177
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":167
 * 
 * 
 * def RRCA_0F(cpu): # 0F RRCA             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":180
 * 
 * 
 * def STOP_10(cpu, v): # 10 STOP 0             # <<<<<<<<<<<<<<
 *     if cpu.mb.cgb:
 *         cpu.mb.switch_speed()
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_STOP_10(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, CYTHON_UNUSED int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":181
 * 
 * def STOP_10(cpu, v): # 10 STOP 0
 *     if cpu.mb.cgb:             # <<<<<<<<<<<<<<
 *         cpu.mb.switch_speed()
 *         cpu.mb.setitem(0xFF04, 0)
 */
  if (__pyx_v_cpu->mb->cgb) {

    /* "pyboy/core/opcodes.py":182
 * def STOP_10(cpu, v): # 10 STOP 0
 *     if cpu.mb.cgb:
 *         cpu.mb.switch_speed()             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem(0xFF04, 0)
 *     cpu.PC += 2
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->switch_speed(__pyx_v_cpu->mb);

    /* "pyboy/core/opcodes.py":183
 *     if cpu.mb.cgb:
 *         cpu.mb.switch_speed()
 *         cpu.mb.setitem(0xFF04, 0)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, 0xFF04, 0);

    /* "pyboy/core/opcodes.py":181
 * 
 * def STOP_10(cpu, v): # 10 STOP 0
 *     if cpu.mb.cgb:             # <<<<<<<<<<<<<<
 *         cpu.mb.switch_speed()
 *         cpu.mb.setitem(0xFF04, 0)
 */
  }

  /* "pyboy/core/opcodes.py":184
 *         cpu.mb.switch_speed()
 *         cpu.mb.setitem(0xFF04, 0)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":185
 *         cpu.mb.setitem(0xFF04, 0)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":186
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":180
 * 
 * 
 * def STOP_10(cpu, v): # 10 STOP 0             # <<<<<<<<<<<<<<
 *     if cpu.mb.cgb:
 *         cpu.mb.switch_speed()
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":189
 * 
 * 
 * def LD_11(cpu, v): # 11 LD DE,d16             # <<<<<<<<<<<<<<
 *     cpu.set_de(v)
 *     cpu.PC += 3
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_11(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":190
 * 
 * def LD_11(cpu, v): # 11 LD DE,d16
 *     cpu.set_de(v)             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU *)__pyx_v_cpu->__pyx_vtab)->set_de(__pyx_v_cpu, __pyx_v_v);

  /* "pyboy/core/opcodes.py":191
 * def LD_11(cpu, v): # 11 LD DE,d16
 *     cpu.set_de(v)
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":192
 *     cpu.set_de(v)
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":193
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":189
 * 
 * 
 * def LD_11(cpu, v): # 11 LD DE,d16             # <<<<<<<<<<<<<<
 *     cpu.set_de(v)
 *     cpu.PC += 3
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":196
 * 
 * 
 * def LD_12(cpu): # 12 LD (DE),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(((cpu.D << 8) + cpu.E), cpu.A)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_12(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":197
 * 
 * def LD_12(cpu): # 12 LD (DE),A
 *     cpu.mb.setitem(((cpu.D << 8) + cpu.E), cpu.A)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->D << 8) + __pyx_v_cpu->E), __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":198
 * def LD_12(cpu): # 12 LD (DE),A
 *     cpu.mb.setitem(((cpu.D << 8) + cpu.E), cpu.A)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":199
 *     cpu.mb.setitem(((cpu.D << 8) + cpu.E), cpu.A)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":200
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":196
 * 
 * 
 * def LD_12(cpu): # 12 LD (DE),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(((cpu.D << 8) + cpu.E), cpu.A)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":203
 * 
 * 
 * def INC_13(cpu): # 13 INC DE             # <<<<<<<<<<<<<<
 *     t = ((cpu.D << 8) + cpu.E) + 1
 *     # No flag operations
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_13(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":204
 * 
 * def INC_13(cpu): # 13 INC DE
 *     t = ((cpu.D << 8) + cpu.E) + 1             # <<<<<<<<<<<<<<
 *     # No flag operations
 *     t &= 0xFFFF
 */
  __pyx_v_t = (((__pyx_v_cpu->D << 8) + __pyx_v_cpu->E) + 1);

  /* "pyboy/core/opcodes.py":206
 *     t = ((cpu.D << 8) + cpu.E) + 1
 *     # No flag operations
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.set_de(t)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":207
 *     # No flag operations
 *     t &= 0xFFFF
 *     cpu.set_de(t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU *)__pyx_v_cpu->__pyx_vtab)->set_de(__pyx_v_cpu, __pyx_v_t);

  /* "pyboy/core/opcodes.py":208
 *     t &= 0xFFFF
 *     cpu.set_de(t)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":209
 *     cpu.set_de(t)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":210
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":203
 * 
 * 
 * def INC_13(cpu): # 13 INC DE             # <<<<<<<<<<<<<<
 *     t = ((cpu.D << 8) + cpu.E) + 1
 *     # No flag operations
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":213
 * 
 * 
 * def INC_14(cpu): # 14 INC D             # <<<<<<<<<<<<<<
 *     t = cpu.D + 1
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_14(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":214
 * 
 * def INC_14(cpu): # 14 INC D
 *     t = cpu.D + 1             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D + 1);

  /* "pyboy/core/opcodes.py":215
 * def INC_14(cpu): # 14 INC D
 *     t = cpu.D + 1
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.D & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":216
 *     t = cpu.D + 1
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.D & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":217
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.D & 0xF) + (1 & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->D & 0xF) + 0x1) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":218
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.D & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":219
 *     flag += (((cpu.D & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":220
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":221
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":222
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":223
 *     cpu.D = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":224
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":213
 * 
 * 
 * def INC_14(cpu): # 14 INC D             # <<<<<<<<<<<<<<
 *     t = cpu.D + 1
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":227
 * 
 * 
 * def DEC_15(cpu): # 15 DEC D             # <<<<<<<<<<<<<<
 *     t = cpu.D - 1
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_15(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":228
 * 
 * def DEC_15(cpu): # 15 DEC D
 *     t = cpu.D - 1             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D - 1);

  /* "pyboy/core/opcodes.py":229
 * def DEC_15(cpu): # 15 DEC D
 *     t = cpu.D - 1
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.D & 0xF) - (1 & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":230
 *     t = cpu.D - 1
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.D & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":231
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.D & 0xF) - (1 & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->D & 0xF) - 0x1) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":232
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.D & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":233
 *     flag += (((cpu.D & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":234
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":235
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":236
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":237
 *     cpu.D = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":238
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":227
 * 
 * 
 * def DEC_15(cpu): # 15 DEC D             # <<<<<<<<<<<<<<
 *     t = cpu.D - 1
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":241
 * 
 * 
 * def LD_16(cpu, v): # 16 LD D,d8             # <<<<<<<<<<<<<<
 *     cpu.D = v
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_16(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":242
 * 
 * def LD_16(cpu, v): # 16 LD D,d8
 *     cpu.D = v             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_v;

  /* "pyboy/core/opcodes.py":243
 * def LD_16(cpu, v): # 16 LD D,d8
 *     cpu.D = v
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":244
 *     cpu.D = v
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":245
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":241
 * 
 * 
 * def LD_16(cpu, v): # 16 LD D,d8             # <<<<<<<<<<<<<<
 *     cpu.D = v
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":248
 * 
 * 
 * def RLA_17(cpu): # 17 RLA             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLA_17(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":249
 * 
 * def RLA_17(cpu): # 17 RLA
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_t = ((__pyx_v_cpu->A << 1) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":250
 * def RLA_17(cpu): # 17 RLA
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":251
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":252
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":253
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":254
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":255
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":256
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":257
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":258
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":248
 * 
 * 
 * def RLA_17(cpu): # 17 RLA             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":261
 * 
 * 
 * def JR_18(cpu, v): # 18 JR r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2 + ((v ^ 0x80) - 0x80)
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_18(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":262
 * 
 * def JR_18(cpu, v): # 18 JR r8
 *     cpu.PC += 2 + ((v ^ 0x80) - 0x80)             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + (2 + ((__pyx_v_v ^ 0x80) - 0x80)));

  /* "pyboy/core/opcodes.py":263
 * def JR_18(cpu, v): # 18 JR r8
 *     cpu.PC += 2 + ((v ^ 0x80) - 0x80)
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":264
 *     cpu.PC += 2 + ((v ^ 0x80) - 0x80)
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":261
 * 
 * 
 * def JR_18(cpu, v): # 18 JR r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2 + ((v ^ 0x80) - 0x80)
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":267
 * 
 * 
 * def ADD_19(cpu): # 19 ADD HL,DE             # <<<<<<<<<<<<<<
 *     t = cpu.HL + ((cpu.D << 8) + cpu.E)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_19(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":268
 * 
 * def ADD_19(cpu): # 19 ADD HL,DE
 *     t = cpu.HL + ((cpu.D << 8) + cpu.E)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.D << 8) + cpu.E) & 0xFFF)) > 0xFFF) << FLAGH
 */
  __pyx_v_t = (__pyx_v_cpu->HL + ((__pyx_v_cpu->D << 8) + __pyx_v_cpu->E));

  /* "pyboy/core/opcodes.py":269
 * def ADD_19(cpu): # 19 ADD HL,DE
 *     t = cpu.HL + ((cpu.D << 8) + cpu.E)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.D << 8) + cpu.E) & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":270
 *     t = cpu.HL + ((cpu.D << 8) + cpu.E)
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.D << 8) + cpu.E) & 0xFFF)) > 0xFFF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->HL & 0xFFF) + (((__pyx_v_cpu->D << 8) + __pyx_v_cpu->E) & 0xFFF)) > 0xFFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":271
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.D << 8) + cpu.E) & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFFFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":272
 *     flag += (((cpu.HL & 0xFFF) + (((cpu.D << 8) + cpu.E) & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 128);

  /* "pyboy/core/opcodes.py":273
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFFFF
 *     cpu.HL = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":274
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.HL = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":275
 *     cpu.F |= flag
 *     t &= 0xFFFF
 *     cpu.HL = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = __pyx_v_t;

  /* "pyboy/core/opcodes.py":276
 *     t &= 0xFFFF
 *     cpu.HL = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":277
 *     cpu.HL = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":278
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":267
 * 
 * 
 * def ADD_19(cpu): # 19 ADD HL,DE             # <<<<<<<<<<<<<<
 *     t = cpu.HL + ((cpu.D << 8) + cpu.E)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":281
 * 
 * 
 * def LD_1A(cpu): # 1A LD A,(DE)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(((cpu.D << 8) + cpu.E))
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_1A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":282
 * 
 * def LD_1A(cpu): # 1A LD A,(DE)
 *     cpu.A = cpu.mb.getitem(((cpu.D << 8) + cpu.E))             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->D << 8) + __pyx_v_cpu->E));

  /* "pyboy/core/opcodes.py":283
 * def LD_1A(cpu): # 1A LD A,(DE)
 *     cpu.A = cpu.mb.getitem(((cpu.D << 8) + cpu.E))
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":284
 *     cpu.A = cpu.mb.getitem(((cpu.D << 8) + cpu.E))
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":285
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":281
 * 
 * 
 * def LD_1A(cpu): # 1A LD A,(DE)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(((cpu.D << 8) + cpu.E))
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":288
 * 
 * 
 * def DEC_1B(cpu): # 1B DEC DE             # <<<<<<<<<<<<<<
 *     t = ((cpu.D << 8) + cpu.E) - 1
 *     # No flag operations
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_1B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":289
 * 
 * def DEC_1B(cpu): # 1B DEC DE
 *     t = ((cpu.D << 8) + cpu.E) - 1             # <<<<<<<<<<<<<<
 *     # No flag operations
 *     t &= 0xFFFF
 */
  __pyx_v_t = (((__pyx_v_cpu->D << 8) + __pyx_v_cpu->E) - 1);

  /* "pyboy/core/opcodes.py":291
 *     t = ((cpu.D << 8) + cpu.E) - 1
 *     # No flag operations
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.set_de(t)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":292
 *     # No flag operations
 *     t &= 0xFFFF
 *     cpu.set_de(t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU *)__pyx_v_cpu->__pyx_vtab)->set_de(__pyx_v_cpu, __pyx_v_t);

  /* "pyboy/core/opcodes.py":293
 *     t &= 0xFFFF
 *     cpu.set_de(t)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":294
 *     cpu.set_de(t)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":295
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":288
 * 
 * 
 * def DEC_1B(cpu): # 1B DEC DE             # <<<<<<<<<<<<<<
 *     t = ((cpu.D << 8) + cpu.E) - 1
 *     # No flag operations
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":298
 * 
 * 
 * def INC_1C(cpu): # 1C INC E             # <<<<<<<<<<<<<<
 *     t = cpu.E + 1
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_1C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":299
 * 
 * def INC_1C(cpu): # 1C INC E
 *     t = cpu.E + 1             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E + 1);

  /* "pyboy/core/opcodes.py":300
 * def INC_1C(cpu): # 1C INC E
 *     t = cpu.E + 1
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.E & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":301
 *     t = cpu.E + 1
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.E & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":302
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.E & 0xF) + (1 & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->E & 0xF) + 0x1) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":303
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.E & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":304
 *     flag += (((cpu.E & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":305
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":306
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":307
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":308
 *     cpu.E = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":309
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":298
 * 
 * 
 * def INC_1C(cpu): # 1C INC E             # <<<<<<<<<<<<<<
 *     t = cpu.E + 1
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":312
 * 
 * 
 * def DEC_1D(cpu): # 1D DEC E             # <<<<<<<<<<<<<<
 *     t = cpu.E - 1
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_1D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":313
 * 
 * def DEC_1D(cpu): # 1D DEC E
 *     t = cpu.E - 1             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E - 1);

  /* "pyboy/core/opcodes.py":314
 * def DEC_1D(cpu): # 1D DEC E
 *     t = cpu.E - 1
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.E & 0xF) - (1 & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":315
 *     t = cpu.E - 1
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.E & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":316
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.E & 0xF) - (1 & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->E & 0xF) - 0x1) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":317
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.E & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":318
 *     flag += (((cpu.E & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":319
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":320
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":321
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":322
 *     cpu.E = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":323
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":312
 * 
 * 
 * def DEC_1D(cpu): # 1D DEC E             # <<<<<<<<<<<<<<
 *     t = cpu.E - 1
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":326
 * 
 * 
 * def LD_1E(cpu, v): # 1E LD E,d8             # <<<<<<<<<<<<<<
 *     cpu.E = v
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_1E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":327
 * 
 * def LD_1E(cpu, v): # 1E LD E,d8
 *     cpu.E = v             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_v;

  /* "pyboy/core/opcodes.py":328
 * def LD_1E(cpu, v): # 1E LD E,d8
 *     cpu.E = v
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":329
 *     cpu.E = v
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":330
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":326
 * 
 * 
 * def LD_1E(cpu, v): # 1E LD E,d8             # <<<<<<<<<<<<<<
 *     cpu.E = v
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":333
 * 
 * 
 * def RRA_1F(cpu): # 1F RRA             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRA_1F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":334
 * 
 * def RRA_1F(cpu): # 1F RRA
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_t = (((__pyx_v_cpu->A >> 1) + (((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0) << 7)) + ((__pyx_v_cpu->A & 1) << 8));

  /* "pyboy/core/opcodes.py":335
 * def RRA_1F(cpu): # 1F RRA
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":336
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":337
 *     flag = 0b00000000
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":338
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":339
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":340
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":341
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":342
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":343
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":333
 * 
 * 
 * def RRA_1F(cpu): # 1F RRA             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":346
 * 
 * 
 * def JR_20(cpu, v): # 20 JR NZ,r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_20(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":347
 * 
 * def JR_20(cpu, v): # 20 JR NZ,r8
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":348
 * def JR_20(cpu, v): # 20 JR NZ,r8
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGZ)) == 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":349
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 12
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + ((__pyx_v_v ^ 0x80) - 0x80));

    /* "pyboy/core/opcodes.py":350
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 12
 *     else:
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":351
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 *         return 12             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC &= 0xFFFF
 */
    __pyx_r = 12;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":348
 * def JR_20(cpu, v): # 20 JR NZ,r8
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 */
  }

  /* "pyboy/core/opcodes.py":353
 *         return 12
 *     else:
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 8
 * 
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":354
 *     else:
 *         cpu.PC &= 0xFFFF
 *         return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 8;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":346
 * 
 * 
 * def JR_20(cpu, v): # 20 JR NZ,r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":357
 * 
 * 
 * def LD_21(cpu, v): # 21 LD HL,d16             # <<<<<<<<<<<<<<
 *     cpu.HL = v
 *     cpu.PC += 3
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_21(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":358
 * 
 * def LD_21(cpu, v): # 21 LD HL,d16
 *     cpu.HL = v             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = __pyx_v_v;

  /* "pyboy/core/opcodes.py":359
 * def LD_21(cpu, v): # 21 LD HL,d16
 *     cpu.HL = v
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":360
 *     cpu.HL = v
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":361
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":357
 * 
 * 
 * def LD_21(cpu, v): # 21 LD HL,d16             # <<<<<<<<<<<<<<
 *     cpu.HL = v
 *     cpu.PC += 3
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":364
 * 
 * 
 * def LD_22(cpu): # 22 LD (HL+),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.HL += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_22(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":365
 * 
 * def LD_22(cpu): # 22 LD (HL+),A
 *     cpu.mb.setitem(cpu.HL, cpu.A)             # <<<<<<<<<<<<<<
 *     cpu.HL += 1
 *     cpu.HL &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":366
 * def LD_22(cpu): # 22 LD (HL+),A
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.HL += 1             # <<<<<<<<<<<<<<
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->HL + 1);

  /* "pyboy/core/opcodes.py":367
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.HL += 1
 *     cpu.HL &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->HL & 0xFFFF);

  /* "pyboy/core/opcodes.py":368
 *     cpu.HL += 1
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":369
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":370
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":364
 * 
 * 
 * def LD_22(cpu): # 22 LD (HL+),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.HL += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":373
 * 
 * 
 * def INC_23(cpu): # 23 INC HL             # <<<<<<<<<<<<<<
 *     t = cpu.HL + 1
 *     # No flag operations
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_23(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":374
 * 
 * def INC_23(cpu): # 23 INC HL
 *     t = cpu.HL + 1             # <<<<<<<<<<<<<<
 *     # No flag operations
 *     t &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_cpu->HL + 1);

  /* "pyboy/core/opcodes.py":376
 *     t = cpu.HL + 1
 *     # No flag operations
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.HL = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":377
 *     # No flag operations
 *     t &= 0xFFFF
 *     cpu.HL = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = __pyx_v_t;

  /* "pyboy/core/opcodes.py":378
 *     t &= 0xFFFF
 *     cpu.HL = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":379
 *     cpu.HL = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":380
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":373
 * 
 * 
 * def INC_23(cpu): # 23 INC HL             # <<<<<<<<<<<<<<
 *     t = cpu.HL + 1
 *     # No flag operations
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":383
 * 
 * 
 * def INC_24(cpu): # 24 INC H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) + 1
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_24(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":384
 * 
 * def INC_24(cpu): # 24 INC H
 *     t = (cpu.HL >> 8) + 1             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) + 1);

  /* "pyboy/core/opcodes.py":385
 * def INC_24(cpu): # 24 INC H
 *     t = (cpu.HL >> 8) + 1
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL >> 8) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":386
 *     t = (cpu.HL >> 8) + 1
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += ((((cpu.HL >> 8) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":387
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL >> 8) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->HL >> 8) & 0xF) + 0x1) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":388
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL >> 8) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":389
 *     flag += ((((cpu.HL >> 8) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":390
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":391
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":392
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":393
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":394
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":383
 * 
 * 
 * def INC_24(cpu): # 24 INC H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) + 1
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":397
 * 
 * 
 * def DEC_25(cpu): # 25 DEC H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) - 1
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_25(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":398
 * 
 * def DEC_25(cpu): # 25 DEC H
 *     t = (cpu.HL >> 8) - 1             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) - 1);

  /* "pyboy/core/opcodes.py":399
 * def DEC_25(cpu): # 25 DEC H
 *     t = (cpu.HL >> 8) - 1
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL >> 8) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":400
 *     t = (cpu.HL >> 8) - 1
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += ((((cpu.HL >> 8) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":401
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL >> 8) & 0xF) - (1 & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->HL >> 8) & 0xF) - 0x1) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":402
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL >> 8) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":403
 *     flag += ((((cpu.HL >> 8) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":404
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":405
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":406
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":407
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":408
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":397
 * 
 * 
 * def DEC_25(cpu): # 25 DEC H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) - 1
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":411
 * 
 * 
 * def LD_26(cpu, v): # 26 LD H,d8             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (v << 8)
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_26(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":412
 * 
 * def LD_26(cpu, v): # 26 LD H,d8
 *     cpu.HL = (cpu.HL & 0x00FF) | (v << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_v << 8));

  /* "pyboy/core/opcodes.py":413
 * def LD_26(cpu, v): # 26 LD H,d8
 *     cpu.HL = (cpu.HL & 0x00FF) | (v << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":414
 *     cpu.HL = (cpu.HL & 0x00FF) | (v << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":415
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":411
 * 
 * 
 * def LD_26(cpu, v): # 26 LD H,d8             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (v << 8)
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":418
 * 
 * 
 * def DAA_27(cpu): # 27 DAA             # <<<<<<<<<<<<<<
 *     t = cpu.A
 *     corr = 0
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DAA_27(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  unsigned short __pyx_v_corr;
  uint8_t __pyx_r;
  int16_t __pyx_t_1;
  long __pyx_t_2;
  int __pyx_t_3;

  /* "pyboy/core/opcodes.py":419
 * 
 * def DAA_27(cpu): # 27 DAA
 *     t = cpu.A             # <<<<<<<<<<<<<<
 *     corr = 0
 *     corr |= 0x06 if ((cpu.F & (1 << FLAGH)) != 0) else 0x00
 */
  __pyx_t_1 = __pyx_v_cpu->A;
  __pyx_v_t = __pyx_t_1;

  /* "pyboy/core/opcodes.py":420
 * def DAA_27(cpu): # 27 DAA
 *     t = cpu.A
 *     corr = 0             # <<<<<<<<<<<<<<
 *     corr |= 0x06 if ((cpu.F & (1 << FLAGH)) != 0) else 0x00
 *     corr |= 0x60 if ((cpu.F & (1 << FLAGC)) != 0) else 0x00
 */
  __pyx_v_corr = 0;

  /* "pyboy/core/opcodes.py":421
 *     t = cpu.A
 *     corr = 0
 *     corr |= 0x06 if ((cpu.F & (1 << FLAGH)) != 0) else 0x00             # <<<<<<<<<<<<<<
 *     corr |= 0x60 if ((cpu.F & (1 << FLAGC)) != 0) else 0x00
 *     if (cpu.F & (1 << FLAGN)) != 0:
 */
  __pyx_t_3 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGH)) != 0);
  if (__pyx_t_3) {
    __pyx_t_2 = 0x06;
  } else {
    __pyx_t_2 = 0x00;
  }
  __pyx_v_corr = (__pyx_v_corr | __pyx_t_2);

  /* "pyboy/core/opcodes.py":422
 *     corr = 0
 *     corr |= 0x06 if ((cpu.F & (1 << FLAGH)) != 0) else 0x00
 *     corr |= 0x60 if ((cpu.F & (1 << FLAGC)) != 0) else 0x00             # <<<<<<<<<<<<<<
 *     if (cpu.F & (1 << FLAGN)) != 0:
 *         t -= corr
 */
  __pyx_t_3 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0);
  if (__pyx_t_3) {
    __pyx_t_2 = 0x60;
  } else {
    __pyx_t_2 = 0x00;
  }
  __pyx_v_corr = (__pyx_v_corr | __pyx_t_2);

  /* "pyboy/core/opcodes.py":423
 *     corr |= 0x06 if ((cpu.F & (1 << FLAGH)) != 0) else 0x00
 *     corr |= 0x60 if ((cpu.F & (1 << FLAGC)) != 0) else 0x00
 *     if (cpu.F & (1 << FLAGN)) != 0:             # <<<<<<<<<<<<<<
 *         t -= corr
 *     else:
 */
  __pyx_t_3 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGN)) != 0);
  if (__pyx_t_3) {

    /* "pyboy/core/opcodes.py":424
 *     corr |= 0x60 if ((cpu.F & (1 << FLAGC)) != 0) else 0x00
 *     if (cpu.F & (1 << FLAGN)) != 0:
 *         t -= corr             # <<<<<<<<<<<<<<
 *     else:
 *         corr |= 0x06 if (t & 0x0F) > 0x09 else 0x00
 */
    __pyx_v_t = (__pyx_v_t - __pyx_v_corr);

    /* "pyboy/core/opcodes.py":423
 *     corr |= 0x06 if ((cpu.F & (1 << FLAGH)) != 0) else 0x00
 *     corr |= 0x60 if ((cpu.F & (1 << FLAGC)) != 0) else 0x00
 *     if (cpu.F & (1 << FLAGN)) != 0:             # <<<<<<<<<<<<<<
 *         t -= corr
 *     else:
 */
    goto __pyx_L3;
  }

  /* "pyboy/core/opcodes.py":426
 *         t -= corr
 *     else:
 *         corr |= 0x06 if (t & 0x0F) > 0x09 else 0x00             # <<<<<<<<<<<<<<
 *         corr |= 0x60 if t > 0x99 else 0x00
 *         t += corr
 */
  /*else*/ {
    __pyx_t_3 = ((__pyx_v_t & 0x0F) > 0x09);
    if (__pyx_t_3) {
      __pyx_t_2 = 0x06;
    } else {
      __pyx_t_2 = 0x00;
    }
    __pyx_v_corr = (__pyx_v_corr | __pyx_t_2);

    /* "pyboy/core/opcodes.py":427
 *     else:
 *         corr |= 0x06 if (t & 0x0F) > 0x09 else 0x00
 *         corr |= 0x60 if t > 0x99 else 0x00             # <<<<<<<<<<<<<<
 *         t += corr
 *     flag = 0
 */
    __pyx_t_3 = (__pyx_v_t > 0x99);
    if (__pyx_t_3) {
      __pyx_t_2 = 0x60;
    } else {
      __pyx_t_2 = 0x00;
    }
    __pyx_v_corr = (__pyx_v_corr | __pyx_t_2);

    /* "pyboy/core/opcodes.py":428
 *         corr |= 0x06 if (t & 0x0F) > 0x09 else 0x00
 *         corr |= 0x60 if t > 0x99 else 0x00
 *         t += corr             # <<<<<<<<<<<<<<
 *     flag = 0
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
    __pyx_v_t = (__pyx_v_t + __pyx_v_corr);
  }
  __pyx_L3:;

  /* "pyboy/core/opcodes.py":429
 *         corr |= 0x60 if t > 0x99 else 0x00
 *         t += corr
 *     flag = 0             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (corr & 0x60 != 0) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":430
 *         t += corr
 *     flag = 0
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (corr & 0x60 != 0) << FLAGC
 *     cpu.F &= 0b01000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":431
 *     flag = 0
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (corr & 0x60 != 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b01000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_corr & 0x60) != 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":432
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (corr & 0x60 != 0) << FLAGC
 *     cpu.F &= 0b01000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 64);

  /* "pyboy/core/opcodes.py":433
 *     flag += (corr & 0x60 != 0) << FLAGC
 *     cpu.F &= 0b01000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":434
 *     cpu.F &= 0b01000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":435
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":436
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":437
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":438
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":418
 * 
 * 
 * def DAA_27(cpu): # 27 DAA             # <<<<<<<<<<<<<<
 *     t = cpu.A
 *     corr = 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":441
 * 
 * 
 * def JR_28(cpu, v): # 28 JR Z,r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_28(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":442
 * 
 * def JR_28(cpu, v): # 28 JR Z,r8
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":443
 * def JR_28(cpu, v): # 28 JR Z,r8
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGZ)) != 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":444
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 12
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + ((__pyx_v_v ^ 0x80) - 0x80));

    /* "pyboy/core/opcodes.py":445
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 12
 *     else:
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":446
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 *         return 12             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC &= 0xFFFF
 */
    __pyx_r = 12;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":443
 * def JR_28(cpu, v): # 28 JR Z,r8
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 */
  }

  /* "pyboy/core/opcodes.py":448
 *         return 12
 *     else:
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 8
 * 
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":449
 *     else:
 *         cpu.PC &= 0xFFFF
 *         return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 8;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":441
 * 
 * 
 * def JR_28(cpu, v): # 28 JR Z,r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":452
 * 
 * 
 * def ADD_29(cpu): # 29 ADD HL,HL             # <<<<<<<<<<<<<<
 *     t = cpu.HL + cpu.HL
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_29(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":453
 * 
 * def ADD_29(cpu): # 29 ADD HL,HL
 *     t = cpu.HL + cpu.HL             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (cpu.HL & 0xFFF)) > 0xFFF) << FLAGH
 */
  __pyx_v_t = (__pyx_v_cpu->HL + __pyx_v_cpu->HL);

  /* "pyboy/core/opcodes.py":454
 * def ADD_29(cpu): # 29 ADD HL,HL
 *     t = cpu.HL + cpu.HL
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (((cpu.HL & 0xFFF) + (cpu.HL & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":455
 *     t = cpu.HL + cpu.HL
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (cpu.HL & 0xFFF)) > 0xFFF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->HL & 0xFFF) + (__pyx_v_cpu->HL & 0xFFF)) > 0xFFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":456
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (cpu.HL & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFFFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":457
 *     flag += (((cpu.HL & 0xFFF) + (cpu.HL & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 128);

  /* "pyboy/core/opcodes.py":458
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFFFF
 *     cpu.HL = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":459
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.HL = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":460
 *     cpu.F |= flag
 *     t &= 0xFFFF
 *     cpu.HL = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = __pyx_v_t;

  /* "pyboy/core/opcodes.py":461
 *     t &= 0xFFFF
 *     cpu.HL = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":462
 *     cpu.HL = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":463
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":452
 * 
 * 
 * def ADD_29(cpu): # 29 ADD HL,HL             # <<<<<<<<<<<<<<
 *     t = cpu.HL + cpu.HL
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":466
 * 
 * 
 * def LD_2A(cpu): # 2A LD A,(HL+)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.HL += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_2A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":467
 * 
 * def LD_2A(cpu): # 2A LD A,(HL+)
 *     cpu.A = cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     cpu.HL += 1
 *     cpu.HL &= 0xFFFF
 */
  __pyx_v_cpu->A = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL);

  /* "pyboy/core/opcodes.py":468
 * def LD_2A(cpu): # 2A LD A,(HL+)
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.HL += 1             # <<<<<<<<<<<<<<
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->HL + 1);

  /* "pyboy/core/opcodes.py":469
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.HL += 1
 *     cpu.HL &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->HL & 0xFFFF);

  /* "pyboy/core/opcodes.py":470
 *     cpu.HL += 1
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":471
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":472
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":466
 * 
 * 
 * def LD_2A(cpu): # 2A LD A,(HL+)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.HL += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":475
 * 
 * 
 * def DEC_2B(cpu): # 2B DEC HL             # <<<<<<<<<<<<<<
 *     t = cpu.HL - 1
 *     # No flag operations
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_2B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":476
 * 
 * def DEC_2B(cpu): # 2B DEC HL
 *     t = cpu.HL - 1             # <<<<<<<<<<<<<<
 *     # No flag operations
 *     t &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_cpu->HL - 1);

  /* "pyboy/core/opcodes.py":478
 *     t = cpu.HL - 1
 *     # No flag operations
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.HL = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":479
 *     # No flag operations
 *     t &= 0xFFFF
 *     cpu.HL = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = __pyx_v_t;

  /* "pyboy/core/opcodes.py":480
 *     t &= 0xFFFF
 *     cpu.HL = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":481
 *     cpu.HL = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":482
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":475
 * 
 * 
 * def DEC_2B(cpu): # 2B DEC HL             # <<<<<<<<<<<<<<
 *     t = cpu.HL - 1
 *     # No flag operations
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":485
 * 
 * 
 * def INC_2C(cpu): # 2C INC L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) + 1
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_2C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":486
 * 
 * def INC_2C(cpu): # 2C INC L
 *     t = (cpu.HL & 0xFF) + 1             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) + 1);

  /* "pyboy/core/opcodes.py":487
 * def INC_2C(cpu): # 2C INC L
 *     t = (cpu.HL & 0xFF) + 1
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL & 0xFF) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":488
 *     t = (cpu.HL & 0xFF) + 1
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += ((((cpu.HL & 0xFF) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":489
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL & 0xFF) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->HL & 0xFF) & 0xF) + 0x1) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":490
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL & 0xFF) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":491
 *     flag += ((((cpu.HL & 0xFF) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":492
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":493
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":494
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":495
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":496
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":485
 * 
 * 
 * def INC_2C(cpu): # 2C INC L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) + 1
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":499
 * 
 * 
 * def DEC_2D(cpu): # 2D DEC L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) - 1
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_2D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":500
 * 
 * def DEC_2D(cpu): # 2D DEC L
 *     t = (cpu.HL & 0xFF) - 1             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) - 1);

  /* "pyboy/core/opcodes.py":501
 * def DEC_2D(cpu): # 2D DEC L
 *     t = (cpu.HL & 0xFF) - 1
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL & 0xFF) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":502
 *     t = (cpu.HL & 0xFF) - 1
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += ((((cpu.HL & 0xFF) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":503
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL & 0xFF) & 0xF) - (1 & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->HL & 0xFF) & 0xF) - 0x1) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":504
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += ((((cpu.HL & 0xFF) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":505
 *     flag += ((((cpu.HL & 0xFF) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":506
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":507
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":508
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":509
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":510
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":499
 * 
 * 
 * def DEC_2D(cpu): # 2D DEC L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) - 1
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":513
 * 
 * 
 * def LD_2E(cpu, v): # 2E LD L,d8             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (v & 0xFF)
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_2E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":514
 * 
 * def LD_2E(cpu, v): # 2E LD L,d8
 *     cpu.HL = (cpu.HL & 0xFF00) | (v & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_v & 0xFF));

  /* "pyboy/core/opcodes.py":515
 * def LD_2E(cpu, v): # 2E LD L,d8
 *     cpu.HL = (cpu.HL & 0xFF00) | (v & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":516
 *     cpu.HL = (cpu.HL & 0xFF00) | (v & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":517
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":513
 * 
 * 
 * def LD_2E(cpu, v): # 2E LD L,d8             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (v & 0xFF)
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":520
 * 
 * 
 * def CPL_2F(cpu): # 2F CPL             # <<<<<<<<<<<<<<
 *     cpu.A = (~cpu.A) & 0xFF
 *     flag = 0b01100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CPL_2F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":521
 * 
 * def CPL_2F(cpu): # 2F CPL
 *     cpu.A = (~cpu.A) & 0xFF             # <<<<<<<<<<<<<<
 *     flag = 0b01100000
 *     cpu.F &= 0b10010000
 */
  __pyx_v_cpu->A = ((~__pyx_v_cpu->A) & 0xFF);

  /* "pyboy/core/opcodes.py":522
 * def CPL_2F(cpu): # 2F CPL
 *     cpu.A = (~cpu.A) & 0xFF
 *     flag = 0b01100000             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b10010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = 96;

  /* "pyboy/core/opcodes.py":523
 *     cpu.A = (~cpu.A) & 0xFF
 *     flag = 0b01100000
 *     cpu.F &= 0b10010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 144);

  /* "pyboy/core/opcodes.py":524
 *     flag = 0b01100000
 *     cpu.F &= 0b10010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":525
 *     cpu.F &= 0b10010000
 *     cpu.F |= flag
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":526
 *     cpu.F |= flag
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":527
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":520
 * 
 * 
 * def CPL_2F(cpu): # 2F CPL             # <<<<<<<<<<<<<<
 *     cpu.A = (~cpu.A) & 0xFF
 *     flag = 0b01100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":530
 * 
 * 
 * def JR_30(cpu, v): # 30 JR NC,r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_30(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":531
 * 
 * def JR_30(cpu, v): # 30 JR NC,r8
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":532
 * def JR_30(cpu, v): # 30 JR NC,r8
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) == 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":533
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 12
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + ((__pyx_v_v ^ 0x80) - 0x80));

    /* "pyboy/core/opcodes.py":534
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 12
 *     else:
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":535
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 *         return 12             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC &= 0xFFFF
 */
    __pyx_r = 12;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":532
 * def JR_30(cpu, v): # 30 JR NC,r8
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 */
  }

  /* "pyboy/core/opcodes.py":537
 *         return 12
 *     else:
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 8
 * 
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":538
 *     else:
 *         cpu.PC &= 0xFFFF
 *         return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 8;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":530
 * 
 * 
 * def JR_30(cpu, v): # 30 JR NC,r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":541
 * 
 * 
 * def LD_31(cpu, v): # 31 LD SP,d16             # <<<<<<<<<<<<<<
 *     cpu.SP = v
 *     cpu.PC += 3
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_31(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":542
 * 
 * def LD_31(cpu, v): # 31 LD SP,d16
 *     cpu.SP = v             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = __pyx_v_v;

  /* "pyboy/core/opcodes.py":543
 * def LD_31(cpu, v): # 31 LD SP,d16
 *     cpu.SP = v
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":544
 *     cpu.SP = v
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":545
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":541
 * 
 * 
 * def LD_31(cpu, v): # 31 LD SP,d16             # <<<<<<<<<<<<<<
 *     cpu.SP = v
 *     cpu.PC += 3
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":548
 * 
 * 
 * def LD_32(cpu): # 32 LD (HL-),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.HL -= 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_32(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":549
 * 
 * def LD_32(cpu): # 32 LD (HL-),A
 *     cpu.mb.setitem(cpu.HL, cpu.A)             # <<<<<<<<<<<<<<
 *     cpu.HL -= 1
 *     cpu.HL &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":550
 * def LD_32(cpu): # 32 LD (HL-),A
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.HL -= 1             # <<<<<<<<<<<<<<
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->HL - 1);

  /* "pyboy/core/opcodes.py":551
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.HL -= 1
 *     cpu.HL &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->HL & 0xFFFF);

  /* "pyboy/core/opcodes.py":552
 *     cpu.HL -= 1
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":553
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":554
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":548
 * 
 * 
 * def LD_32(cpu): # 32 LD (HL-),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.HL -= 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":557
 * 
 * 
 * def INC_33(cpu): # 33 INC SP             # <<<<<<<<<<<<<<
 *     t = cpu.SP + 1
 *     # No flag operations
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_33(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":558
 * 
 * def INC_33(cpu): # 33 INC SP
 *     t = cpu.SP + 1             # <<<<<<<<<<<<<<
 *     # No flag operations
 *     t &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_cpu->SP + 1);

  /* "pyboy/core/opcodes.py":560
 *     t = cpu.SP + 1
 *     # No flag operations
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.SP = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":561
 *     # No flag operations
 *     t &= 0xFFFF
 *     cpu.SP = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = __pyx_v_t;

  /* "pyboy/core/opcodes.py":562
 *     t &= 0xFFFF
 *     cpu.SP = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":563
 *     cpu.SP = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":564
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":557
 * 
 * 
 * def INC_33(cpu): # 33 INC SP             # <<<<<<<<<<<<<<
 *     t = cpu.SP + 1
 *     # No flag operations
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":567
 * 
 * 
 * def INC_34(cpu): # 34 INC (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) + 1
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_34(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":568
 * 
 * def INC_34(cpu): # 34 INC (HL)
 *     t = cpu.mb.getitem(cpu.HL) + 1             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) + 1);

  /* "pyboy/core/opcodes.py":569
 * def INC_34(cpu): # 34 INC (HL)
 *     t = cpu.mb.getitem(cpu.HL) + 1
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":570
 *     t = cpu.mb.getitem(cpu.HL) + 1
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":571
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0xF) + 0x1) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":572
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":573
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":574
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":575
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":576
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":577
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":578
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":567
 * 
 * 
 * def INC_34(cpu): # 34 INC (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) + 1
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":581
 * 
 * 
 * def DEC_35(cpu): # 35 DEC (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) - 1
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_35(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":582
 * 
 * def DEC_35(cpu): # 35 DEC (HL)
 *     t = cpu.mb.getitem(cpu.HL) - 1             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) - 1);

  /* "pyboy/core/opcodes.py":583
 * def DEC_35(cpu): # 35 DEC (HL)
 *     t = cpu.mb.getitem(cpu.HL) - 1
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":584
 *     t = cpu.mb.getitem(cpu.HL) - 1
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":585
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) - (1 & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0xF) - 0x1) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":586
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":587
 *     flag += (((cpu.mb.getitem(cpu.HL) & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":588
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":589
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":590
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":591
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":592
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":581
 * 
 * 
 * def DEC_35(cpu): # 35 DEC (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) - 1
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":595
 * 
 * 
 * def LD_36(cpu, v): # 36 LD (HL),d8             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, v)
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_36(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":596
 * 
 * def LD_36(cpu, v): # 36 LD (HL),d8
 *     cpu.mb.setitem(cpu.HL, v)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_v);

  /* "pyboy/core/opcodes.py":597
 * def LD_36(cpu, v): # 36 LD (HL),d8
 *     cpu.mb.setitem(cpu.HL, v)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":598
 *     cpu.mb.setitem(cpu.HL, v)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":599
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":595
 * 
 * 
 * def LD_36(cpu, v): # 36 LD (HL),d8             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, v)
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":602
 * 
 * 
 * def SCF_37(cpu): # 37 SCF             # <<<<<<<<<<<<<<
 *     flag = 0b00010000
 *     cpu.F &= 0b10000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SCF_37(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":603
 * 
 * def SCF_37(cpu): # 37 SCF
 *     flag = 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = 16;

  /* "pyboy/core/opcodes.py":604
 * def SCF_37(cpu): # 37 SCF
 *     flag = 0b00010000
 *     cpu.F &= 0b10000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 128);

  /* "pyboy/core/opcodes.py":605
 *     flag = 0b00010000
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":606
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":607
 *     cpu.F |= flag
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":608
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":602
 * 
 * 
 * def SCF_37(cpu): # 37 SCF             # <<<<<<<<<<<<<<
 *     flag = 0b00010000
 *     cpu.F &= 0b10000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":611
 * 
 * 
 * def JR_38(cpu, v): # 38 JR C,r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JR_38(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":612
 * 
 * def JR_38(cpu, v): # 38 JR C,r8
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":613
 * def JR_38(cpu, v): # 38 JR C,r8
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":614
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 12
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + ((__pyx_v_v ^ 0x80) - 0x80));

    /* "pyboy/core/opcodes.py":615
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 12
 *     else:
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":616
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 *         return 12             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC &= 0xFFFF
 */
    __pyx_r = 12;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":613
 * def JR_38(cpu, v): # 38 JR C,r8
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC += ((v ^ 0x80) - 0x80)
 *         cpu.PC &= 0xFFFF
 */
  }

  /* "pyboy/core/opcodes.py":618
 *         return 12
 *     else:
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 8
 * 
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":619
 *     else:
 *         cpu.PC &= 0xFFFF
 *         return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 8;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":611
 * 
 * 
 * def JR_38(cpu, v): # 38 JR C,r8             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":622
 * 
 * 
 * def ADD_39(cpu): # 39 ADD HL,SP             # <<<<<<<<<<<<<<
 *     t = cpu.HL + cpu.SP
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_39(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":623
 * 
 * def ADD_39(cpu): # 39 ADD HL,SP
 *     t = cpu.HL + cpu.SP             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (cpu.SP & 0xFFF)) > 0xFFF) << FLAGH
 */
  __pyx_v_t = (__pyx_v_cpu->HL + __pyx_v_cpu->SP);

  /* "pyboy/core/opcodes.py":624
 * def ADD_39(cpu): # 39 ADD HL,SP
 *     t = cpu.HL + cpu.SP
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (((cpu.HL & 0xFFF) + (cpu.SP & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":625
 *     t = cpu.HL + cpu.SP
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (cpu.SP & 0xFFF)) > 0xFFF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->HL & 0xFFF) + (__pyx_v_cpu->SP & 0xFFF)) > 0xFFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":626
 *     flag = 0b00000000
 *     flag += (((cpu.HL & 0xFFF) + (cpu.SP & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFFFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":627
 *     flag += (((cpu.HL & 0xFFF) + (cpu.SP & 0xFFF)) > 0xFFF) << FLAGH
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 128);

  /* "pyboy/core/opcodes.py":628
 *     flag += (t > 0xFFFF) << FLAGC
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFFFF
 *     cpu.HL = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":629
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.HL = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":630
 *     cpu.F |= flag
 *     t &= 0xFFFF
 *     cpu.HL = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = __pyx_v_t;

  /* "pyboy/core/opcodes.py":631
 *     t &= 0xFFFF
 *     cpu.HL = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":632
 *     cpu.HL = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":633
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":622
 * 
 * 
 * def ADD_39(cpu): # 39 ADD HL,SP             # <<<<<<<<<<<<<<
 *     t = cpu.HL + cpu.SP
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":636
 * 
 * 
 * def LD_3A(cpu): # 3A LD A,(HL-)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.HL -= 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_3A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":637
 * 
 * def LD_3A(cpu): # 3A LD A,(HL-)
 *     cpu.A = cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     cpu.HL -= 1
 *     cpu.HL &= 0xFFFF
 */
  __pyx_v_cpu->A = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL);

  /* "pyboy/core/opcodes.py":638
 * def LD_3A(cpu): # 3A LD A,(HL-)
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.HL -= 1             # <<<<<<<<<<<<<<
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->HL - 1);

  /* "pyboy/core/opcodes.py":639
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.HL -= 1
 *     cpu.HL &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->HL & 0xFFFF);

  /* "pyboy/core/opcodes.py":640
 *     cpu.HL -= 1
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":641
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":642
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":636
 * 
 * 
 * def LD_3A(cpu): # 3A LD A,(HL-)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.HL -= 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":645
 * 
 * 
 * def DEC_3B(cpu): # 3B DEC SP             # <<<<<<<<<<<<<<
 *     t = cpu.SP - 1
 *     # No flag operations
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_3B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":646
 * 
 * def DEC_3B(cpu): # 3B DEC SP
 *     t = cpu.SP - 1             # <<<<<<<<<<<<<<
 *     # No flag operations
 *     t &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_cpu->SP - 1);

  /* "pyboy/core/opcodes.py":648
 *     t = cpu.SP - 1
 *     # No flag operations
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.SP = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":649
 *     # No flag operations
 *     t &= 0xFFFF
 *     cpu.SP = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = __pyx_v_t;

  /* "pyboy/core/opcodes.py":650
 *     t &= 0xFFFF
 *     cpu.SP = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":651
 *     cpu.SP = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":652
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":645
 * 
 * 
 * def DEC_3B(cpu): # 3B DEC SP             # <<<<<<<<<<<<<<
 *     t = cpu.SP - 1
 *     # No flag operations
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":655
 * 
 * 
 * def INC_3C(cpu): # 3C INC A             # <<<<<<<<<<<<<<
 *     t = cpu.A + 1
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_INC_3C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":656
 * 
 * def INC_3C(cpu): # 3C INC A
 *     t = cpu.A + 1             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + 1);

  /* "pyboy/core/opcodes.py":657
 * def INC_3C(cpu): # 3C INC A
 *     t = cpu.A + 1
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":658
 *     t = cpu.A + 1
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":659
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (1 & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + 0x1) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":660
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":661
 *     flag += (((cpu.A & 0xF) + (1 & 0xF)) > 0xF) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":662
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":663
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":664
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":665
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":666
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":655
 * 
 * 
 * def INC_3C(cpu): # 3C INC A             # <<<<<<<<<<<<<<
 *     t = cpu.A + 1
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":669
 * 
 * 
 * def DEC_3D(cpu): # 3D DEC A             # <<<<<<<<<<<<<<
 *     t = cpu.A - 1
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DEC_3D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":670
 * 
 * def DEC_3D(cpu): # 3D DEC A
 *     t = cpu.A - 1             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - 1);

  /* "pyboy/core/opcodes.py":671
 * def DEC_3D(cpu): # 3D DEC A
 *     t = cpu.A - 1
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (1 & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":672
 *     t = cpu.A - 1
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":673
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (1 & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - 0x1) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":674
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":675
 *     flag += (((cpu.A & 0xF) - (1 & 0xF)) < 0) << FLAGH
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":676
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":677
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":678
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":679
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":680
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":669
 * 
 * 
 * def DEC_3D(cpu): # 3D DEC A             # <<<<<<<<<<<<<<
 *     t = cpu.A - 1
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":683
 * 
 * 
 * def LD_3E(cpu, v): # 3E LD A,d8             # <<<<<<<<<<<<<<
 *     cpu.A = v
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_3E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":684
 * 
 * def LD_3E(cpu, v): # 3E LD A,d8
 *     cpu.A = v             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_v;

  /* "pyboy/core/opcodes.py":685
 * def LD_3E(cpu, v): # 3E LD A,d8
 *     cpu.A = v
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":686
 *     cpu.A = v
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":687
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":683
 * 
 * 
 * def LD_3E(cpu, v): # 3E LD A,d8             # <<<<<<<<<<<<<<
 *     cpu.A = v
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":690
 * 
 * 
 * def CCF_3F(cpu): # 3F CCF             # <<<<<<<<<<<<<<
 *     flag = (cpu.F & 0b00010000) ^ 0b00010000
 *     cpu.F &= 0b10000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CCF_3F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":691
 * 
 * def CCF_3F(cpu): # 3F CCF
 *     flag = (cpu.F & 0b00010000) ^ 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = ((__pyx_v_cpu->F & 16) ^ 16);

  /* "pyboy/core/opcodes.py":692
 * def CCF_3F(cpu): # 3F CCF
 *     flag = (cpu.F & 0b00010000) ^ 0b00010000
 *     cpu.F &= 0b10000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 128);

  /* "pyboy/core/opcodes.py":693
 *     flag = (cpu.F & 0b00010000) ^ 0b00010000
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":694
 *     cpu.F &= 0b10000000
 *     cpu.F |= flag
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":695
 *     cpu.F |= flag
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":696
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":690
 * 
 * 
 * def CCF_3F(cpu): # 3F CCF             # <<<<<<<<<<<<<<
 *     flag = (cpu.F & 0b00010000) ^ 0b00010000
 *     cpu.F &= 0b10000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":699
 * 
 * 
 * def LD_40(cpu): # 40 LD B,B             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.B
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_40(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":700
 * 
 * def LD_40(cpu): # 40 LD B,B
 *     cpu.B = cpu.B             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->B;
  __pyx_v_cpu->B = __pyx_t_1;

  /* "pyboy/core/opcodes.py":701
 * def LD_40(cpu): # 40 LD B,B
 *     cpu.B = cpu.B
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":702
 *     cpu.B = cpu.B
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":703
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":699
 * 
 * 
 * def LD_40(cpu): # 40 LD B,B             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.B
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":706
 * 
 * 
 * def LD_41(cpu): # 41 LD B,C             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.C
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_41(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":707
 * 
 * def LD_41(cpu): # 41 LD B,C
 *     cpu.B = cpu.C             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->C;
  __pyx_v_cpu->B = __pyx_t_1;

  /* "pyboy/core/opcodes.py":708
 * def LD_41(cpu): # 41 LD B,C
 *     cpu.B = cpu.C
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":709
 *     cpu.B = cpu.C
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":710
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":706
 * 
 * 
 * def LD_41(cpu): # 41 LD B,C             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.C
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":713
 * 
 * 
 * def LD_42(cpu): # 42 LD B,D             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.D
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_42(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":714
 * 
 * def LD_42(cpu): # 42 LD B,D
 *     cpu.B = cpu.D             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->D;
  __pyx_v_cpu->B = __pyx_t_1;

  /* "pyboy/core/opcodes.py":715
 * def LD_42(cpu): # 42 LD B,D
 *     cpu.B = cpu.D
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":716
 *     cpu.B = cpu.D
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":717
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":713
 * 
 * 
 * def LD_42(cpu): # 42 LD B,D             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.D
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":720
 * 
 * 
 * def LD_43(cpu): # 43 LD B,E             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.E
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_43(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":721
 * 
 * def LD_43(cpu): # 43 LD B,E
 *     cpu.B = cpu.E             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->E;
  __pyx_v_cpu->B = __pyx_t_1;

  /* "pyboy/core/opcodes.py":722
 * def LD_43(cpu): # 43 LD B,E
 *     cpu.B = cpu.E
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":723
 *     cpu.B = cpu.E
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":724
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":720
 * 
 * 
 * def LD_43(cpu): # 43 LD B,E             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.E
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":727
 * 
 * 
 * def LD_44(cpu): # 44 LD B,H             # <<<<<<<<<<<<<<
 *     cpu.B = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_44(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":728
 * 
 * def LD_44(cpu): # 44 LD B,H
 *     cpu.B = (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = (__pyx_v_cpu->HL >> 8);

  /* "pyboy/core/opcodes.py":729
 * def LD_44(cpu): # 44 LD B,H
 *     cpu.B = (cpu.HL >> 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":730
 *     cpu.B = (cpu.HL >> 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":731
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":727
 * 
 * 
 * def LD_44(cpu): # 44 LD B,H             # <<<<<<<<<<<<<<
 *     cpu.B = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":734
 * 
 * 
 * def LD_45(cpu): # 45 LD B,L             # <<<<<<<<<<<<<<
 *     cpu.B = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_45(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":735
 * 
 * def LD_45(cpu): # 45 LD B,L
 *     cpu.B = (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = (__pyx_v_cpu->HL & 0xFF);

  /* "pyboy/core/opcodes.py":736
 * def LD_45(cpu): # 45 LD B,L
 *     cpu.B = (cpu.HL & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":737
 *     cpu.B = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":738
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":734
 * 
 * 
 * def LD_45(cpu): # 45 LD B,L             # <<<<<<<<<<<<<<
 *     cpu.B = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":741
 * 
 * 
 * def LD_46(cpu): # 46 LD B,(HL)             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_46(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":742
 * 
 * def LD_46(cpu): # 46 LD B,(HL)
 *     cpu.B = cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL);

  /* "pyboy/core/opcodes.py":743
 * def LD_46(cpu): # 46 LD B,(HL)
 *     cpu.B = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":744
 *     cpu.B = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":745
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":741
 * 
 * 
 * def LD_46(cpu): # 46 LD B,(HL)             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":748
 * 
 * 
 * def LD_47(cpu): # 47 LD B,A             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.A
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_47(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":749
 * 
 * def LD_47(cpu): # 47 LD B,A
 *     cpu.B = cpu.A             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->A;
  __pyx_v_cpu->B = __pyx_t_1;

  /* "pyboy/core/opcodes.py":750
 * def LD_47(cpu): # 47 LD B,A
 *     cpu.B = cpu.A
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":751
 *     cpu.B = cpu.A
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":752
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":748
 * 
 * 
 * def LD_47(cpu): # 47 LD B,A             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.A
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":755
 * 
 * 
 * def LD_48(cpu): # 48 LD C,B             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.B
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_48(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":756
 * 
 * def LD_48(cpu): # 48 LD C,B
 *     cpu.C = cpu.B             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->B;
  __pyx_v_cpu->C = __pyx_t_1;

  /* "pyboy/core/opcodes.py":757
 * def LD_48(cpu): # 48 LD C,B
 *     cpu.C = cpu.B
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":758
 *     cpu.C = cpu.B
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":759
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":755
 * 
 * 
 * def LD_48(cpu): # 48 LD C,B             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.B
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":762
 * 
 * 
 * def LD_49(cpu): # 49 LD C,C             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.C
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_49(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":763
 * 
 * def LD_49(cpu): # 49 LD C,C
 *     cpu.C = cpu.C             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->C;
  __pyx_v_cpu->C = __pyx_t_1;

  /* "pyboy/core/opcodes.py":764
 * def LD_49(cpu): # 49 LD C,C
 *     cpu.C = cpu.C
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":765
 *     cpu.C = cpu.C
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":766
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":762
 * 
 * 
 * def LD_49(cpu): # 49 LD C,C             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.C
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":769
 * 
 * 
 * def LD_4A(cpu): # 4A LD C,D             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.D
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":770
 * 
 * def LD_4A(cpu): # 4A LD C,D
 *     cpu.C = cpu.D             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->D;
  __pyx_v_cpu->C = __pyx_t_1;

  /* "pyboy/core/opcodes.py":771
 * def LD_4A(cpu): # 4A LD C,D
 *     cpu.C = cpu.D
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":772
 *     cpu.C = cpu.D
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":773
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":769
 * 
 * 
 * def LD_4A(cpu): # 4A LD C,D             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.D
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":776
 * 
 * 
 * def LD_4B(cpu): # 4B LD C,E             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.E
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":777
 * 
 * def LD_4B(cpu): # 4B LD C,E
 *     cpu.C = cpu.E             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->E;
  __pyx_v_cpu->C = __pyx_t_1;

  /* "pyboy/core/opcodes.py":778
 * def LD_4B(cpu): # 4B LD C,E
 *     cpu.C = cpu.E
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":779
 *     cpu.C = cpu.E
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":780
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":776
 * 
 * 
 * def LD_4B(cpu): # 4B LD C,E             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.E
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":783
 * 
 * 
 * def LD_4C(cpu): # 4C LD C,H             # <<<<<<<<<<<<<<
 *     cpu.C = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":784
 * 
 * def LD_4C(cpu): # 4C LD C,H
 *     cpu.C = (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = (__pyx_v_cpu->HL >> 8);

  /* "pyboy/core/opcodes.py":785
 * def LD_4C(cpu): # 4C LD C,H
 *     cpu.C = (cpu.HL >> 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":786
 *     cpu.C = (cpu.HL >> 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":787
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":783
 * 
 * 
 * def LD_4C(cpu): # 4C LD C,H             # <<<<<<<<<<<<<<
 *     cpu.C = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":790
 * 
 * 
 * def LD_4D(cpu): # 4D LD C,L             # <<<<<<<<<<<<<<
 *     cpu.C = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":791
 * 
 * def LD_4D(cpu): # 4D LD C,L
 *     cpu.C = (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = (__pyx_v_cpu->HL & 0xFF);

  /* "pyboy/core/opcodes.py":792
 * def LD_4D(cpu): # 4D LD C,L
 *     cpu.C = (cpu.HL & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":793
 *     cpu.C = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":794
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":790
 * 
 * 
 * def LD_4D(cpu): # 4D LD C,L             # <<<<<<<<<<<<<<
 *     cpu.C = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":797
 * 
 * 
 * def LD_4E(cpu): # 4E LD C,(HL)             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":798
 * 
 * def LD_4E(cpu): # 4E LD C,(HL)
 *     cpu.C = cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL);

  /* "pyboy/core/opcodes.py":799
 * def LD_4E(cpu): # 4E LD C,(HL)
 *     cpu.C = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":800
 *     cpu.C = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":801
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":797
 * 
 * 
 * def LD_4E(cpu): # 4E LD C,(HL)             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":804
 * 
 * 
 * def LD_4F(cpu): # 4F LD C,A             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.A
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_4F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":805
 * 
 * def LD_4F(cpu): # 4F LD C,A
 *     cpu.C = cpu.A             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->A;
  __pyx_v_cpu->C = __pyx_t_1;

  /* "pyboy/core/opcodes.py":806
 * def LD_4F(cpu): # 4F LD C,A
 *     cpu.C = cpu.A
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":807
 *     cpu.C = cpu.A
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":808
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":804
 * 
 * 
 * def LD_4F(cpu): # 4F LD C,A             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.A
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":811
 * 
 * 
 * def LD_50(cpu): # 50 LD D,B             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.B
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_50(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":812
 * 
 * def LD_50(cpu): # 50 LD D,B
 *     cpu.D = cpu.B             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->B;
  __pyx_v_cpu->D = __pyx_t_1;

  /* "pyboy/core/opcodes.py":813
 * def LD_50(cpu): # 50 LD D,B
 *     cpu.D = cpu.B
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":814
 *     cpu.D = cpu.B
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":815
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":811
 * 
 * 
 * def LD_50(cpu): # 50 LD D,B             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.B
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":818
 * 
 * 
 * def LD_51(cpu): # 51 LD D,C             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.C
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_51(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":819
 * 
 * def LD_51(cpu): # 51 LD D,C
 *     cpu.D = cpu.C             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->C;
  __pyx_v_cpu->D = __pyx_t_1;

  /* "pyboy/core/opcodes.py":820
 * def LD_51(cpu): # 51 LD D,C
 *     cpu.D = cpu.C
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":821
 *     cpu.D = cpu.C
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":822
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":818
 * 
 * 
 * def LD_51(cpu): # 51 LD D,C             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.C
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":825
 * 
 * 
 * def LD_52(cpu): # 52 LD D,D             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.D
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_52(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":826
 * 
 * def LD_52(cpu): # 52 LD D,D
 *     cpu.D = cpu.D             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->D;
  __pyx_v_cpu->D = __pyx_t_1;

  /* "pyboy/core/opcodes.py":827
 * def LD_52(cpu): # 52 LD D,D
 *     cpu.D = cpu.D
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":828
 *     cpu.D = cpu.D
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":829
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":825
 * 
 * 
 * def LD_52(cpu): # 52 LD D,D             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.D
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":832
 * 
 * 
 * def LD_53(cpu): # 53 LD D,E             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.E
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_53(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":833
 * 
 * def LD_53(cpu): # 53 LD D,E
 *     cpu.D = cpu.E             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->E;
  __pyx_v_cpu->D = __pyx_t_1;

  /* "pyboy/core/opcodes.py":834
 * def LD_53(cpu): # 53 LD D,E
 *     cpu.D = cpu.E
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":835
 *     cpu.D = cpu.E
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":836
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":832
 * 
 * 
 * def LD_53(cpu): # 53 LD D,E             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.E
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":839
 * 
 * 
 * def LD_54(cpu): # 54 LD D,H             # <<<<<<<<<<<<<<
 *     cpu.D = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_54(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":840
 * 
 * def LD_54(cpu): # 54 LD D,H
 *     cpu.D = (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = (__pyx_v_cpu->HL >> 8);

  /* "pyboy/core/opcodes.py":841
 * def LD_54(cpu): # 54 LD D,H
 *     cpu.D = (cpu.HL >> 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":842
 *     cpu.D = (cpu.HL >> 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":843
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":839
 * 
 * 
 * def LD_54(cpu): # 54 LD D,H             # <<<<<<<<<<<<<<
 *     cpu.D = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":846
 * 
 * 
 * def LD_55(cpu): # 55 LD D,L             # <<<<<<<<<<<<<<
 *     cpu.D = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_55(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":847
 * 
 * def LD_55(cpu): # 55 LD D,L
 *     cpu.D = (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = (__pyx_v_cpu->HL & 0xFF);

  /* "pyboy/core/opcodes.py":848
 * def LD_55(cpu): # 55 LD D,L
 *     cpu.D = (cpu.HL & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":849
 *     cpu.D = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":850
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":846
 * 
 * 
 * def LD_55(cpu): # 55 LD D,L             # <<<<<<<<<<<<<<
 *     cpu.D = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":853
 * 
 * 
 * def LD_56(cpu): # 56 LD D,(HL)             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_56(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":854
 * 
 * def LD_56(cpu): # 56 LD D,(HL)
 *     cpu.D = cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL);

  /* "pyboy/core/opcodes.py":855
 * def LD_56(cpu): # 56 LD D,(HL)
 *     cpu.D = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":856
 *     cpu.D = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":857
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":853
 * 
 * 
 * def LD_56(cpu): # 56 LD D,(HL)             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":860
 * 
 * 
 * def LD_57(cpu): # 57 LD D,A             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.A
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_57(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":861
 * 
 * def LD_57(cpu): # 57 LD D,A
 *     cpu.D = cpu.A             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->A;
  __pyx_v_cpu->D = __pyx_t_1;

  /* "pyboy/core/opcodes.py":862
 * def LD_57(cpu): # 57 LD D,A
 *     cpu.D = cpu.A
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":863
 *     cpu.D = cpu.A
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":864
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":860
 * 
 * 
 * def LD_57(cpu): # 57 LD D,A             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.A
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":867
 * 
 * 
 * def LD_58(cpu): # 58 LD E,B             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.B
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_58(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":868
 * 
 * def LD_58(cpu): # 58 LD E,B
 *     cpu.E = cpu.B             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->B;
  __pyx_v_cpu->E = __pyx_t_1;

  /* "pyboy/core/opcodes.py":869
 * def LD_58(cpu): # 58 LD E,B
 *     cpu.E = cpu.B
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":870
 *     cpu.E = cpu.B
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":871
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":867
 * 
 * 
 * def LD_58(cpu): # 58 LD E,B             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.B
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":874
 * 
 * 
 * def LD_59(cpu): # 59 LD E,C             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.C
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_59(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":875
 * 
 * def LD_59(cpu): # 59 LD E,C
 *     cpu.E = cpu.C             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->C;
  __pyx_v_cpu->E = __pyx_t_1;

  /* "pyboy/core/opcodes.py":876
 * def LD_59(cpu): # 59 LD E,C
 *     cpu.E = cpu.C
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":877
 *     cpu.E = cpu.C
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":878
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":874
 * 
 * 
 * def LD_59(cpu): # 59 LD E,C             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.C
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":881
 * 
 * 
 * def LD_5A(cpu): # 5A LD E,D             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.D
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":882
 * 
 * def LD_5A(cpu): # 5A LD E,D
 *     cpu.E = cpu.D             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->D;
  __pyx_v_cpu->E = __pyx_t_1;

  /* "pyboy/core/opcodes.py":883
 * def LD_5A(cpu): # 5A LD E,D
 *     cpu.E = cpu.D
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":884
 *     cpu.E = cpu.D
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":885
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":881
 * 
 * 
 * def LD_5A(cpu): # 5A LD E,D             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.D
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":888
 * 
 * 
 * def LD_5B(cpu): # 5B LD E,E             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.E
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":889
 * 
 * def LD_5B(cpu): # 5B LD E,E
 *     cpu.E = cpu.E             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->E;
  __pyx_v_cpu->E = __pyx_t_1;

  /* "pyboy/core/opcodes.py":890
 * def LD_5B(cpu): # 5B LD E,E
 *     cpu.E = cpu.E
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":891
 *     cpu.E = cpu.E
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":892
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":888
 * 
 * 
 * def LD_5B(cpu): # 5B LD E,E             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.E
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":895
 * 
 * 
 * def LD_5C(cpu): # 5C LD E,H             # <<<<<<<<<<<<<<
 *     cpu.E = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":896
 * 
 * def LD_5C(cpu): # 5C LD E,H
 *     cpu.E = (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = (__pyx_v_cpu->HL >> 8);

  /* "pyboy/core/opcodes.py":897
 * def LD_5C(cpu): # 5C LD E,H
 *     cpu.E = (cpu.HL >> 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":898
 *     cpu.E = (cpu.HL >> 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":899
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":895
 * 
 * 
 * def LD_5C(cpu): # 5C LD E,H             # <<<<<<<<<<<<<<
 *     cpu.E = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":902
 * 
 * 
 * def LD_5D(cpu): # 5D LD E,L             # <<<<<<<<<<<<<<
 *     cpu.E = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":903
 * 
 * def LD_5D(cpu): # 5D LD E,L
 *     cpu.E = (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = (__pyx_v_cpu->HL & 0xFF);

  /* "pyboy/core/opcodes.py":904
 * def LD_5D(cpu): # 5D LD E,L
 *     cpu.E = (cpu.HL & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":905
 *     cpu.E = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":906
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":902
 * 
 * 
 * def LD_5D(cpu): # 5D LD E,L             # <<<<<<<<<<<<<<
 *     cpu.E = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":909
 * 
 * 
 * def LD_5E(cpu): # 5E LD E,(HL)             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":910
 * 
 * def LD_5E(cpu): # 5E LD E,(HL)
 *     cpu.E = cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL);

  /* "pyboy/core/opcodes.py":911
 * def LD_5E(cpu): # 5E LD E,(HL)
 *     cpu.E = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":912
 *     cpu.E = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":913
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":909
 * 
 * 
 * def LD_5E(cpu): # 5E LD E,(HL)             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":916
 * 
 * 
 * def LD_5F(cpu): # 5F LD E,A             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.A
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_5F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":917
 * 
 * def LD_5F(cpu): # 5F LD E,A
 *     cpu.E = cpu.A             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->A;
  __pyx_v_cpu->E = __pyx_t_1;

  /* "pyboy/core/opcodes.py":918
 * def LD_5F(cpu): # 5F LD E,A
 *     cpu.E = cpu.A
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":919
 *     cpu.E = cpu.A
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":920
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":916
 * 
 * 
 * def LD_5F(cpu): # 5F LD E,A             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.A
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":923
 * 
 * 
 * def LD_60(cpu): # 60 LD H,B             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.B << 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_60(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":924
 * 
 * def LD_60(cpu): # 60 LD H,B
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.B << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_cpu->B << 8));

  /* "pyboy/core/opcodes.py":925
 * def LD_60(cpu): # 60 LD H,B
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.B << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":926
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.B << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":927
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":923
 * 
 * 
 * def LD_60(cpu): # 60 LD H,B             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.B << 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":930
 * 
 * 
 * def LD_61(cpu): # 61 LD H,C             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.C << 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_61(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":931
 * 
 * def LD_61(cpu): # 61 LD H,C
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.C << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_cpu->C << 8));

  /* "pyboy/core/opcodes.py":932
 * def LD_61(cpu): # 61 LD H,C
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.C << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":933
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.C << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":934
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":930
 * 
 * 
 * def LD_61(cpu): # 61 LD H,C             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.C << 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":937
 * 
 * 
 * def LD_62(cpu): # 62 LD H,D             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.D << 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_62(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":938
 * 
 * def LD_62(cpu): # 62 LD H,D
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.D << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_cpu->D << 8));

  /* "pyboy/core/opcodes.py":939
 * def LD_62(cpu): # 62 LD H,D
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.D << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":940
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.D << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":941
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":937
 * 
 * 
 * def LD_62(cpu): # 62 LD H,D             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.D << 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":944
 * 
 * 
 * def LD_63(cpu): # 63 LD H,E             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.E << 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_63(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":945
 * 
 * def LD_63(cpu): # 63 LD H,E
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.E << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_cpu->E << 8));

  /* "pyboy/core/opcodes.py":946
 * def LD_63(cpu): # 63 LD H,E
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.E << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":947
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.E << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":948
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":944
 * 
 * 
 * def LD_63(cpu): # 63 LD H,E             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.E << 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":951
 * 
 * 
 * def LD_64(cpu): # 64 LD H,H             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL >> 8) << 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_64(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":952
 * 
 * def LD_64(cpu): # 64 LD H,H
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL >> 8) << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | ((__pyx_v_cpu->HL >> 8) << 8));

  /* "pyboy/core/opcodes.py":953
 * def LD_64(cpu): # 64 LD H,H
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL >> 8) << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":954
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL >> 8) << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":955
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":951
 * 
 * 
 * def LD_64(cpu): # 64 LD H,H             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL >> 8) << 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":958
 * 
 * 
 * def LD_65(cpu): # 65 LD H,L             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL & 0xFF) << 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_65(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":959
 * 
 * def LD_65(cpu): # 65 LD H,L
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL & 0xFF) << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | ((__pyx_v_cpu->HL & 0xFF) << 8));

  /* "pyboy/core/opcodes.py":960
 * def LD_65(cpu): # 65 LD H,L
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL & 0xFF) << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":961
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL & 0xFF) << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":962
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":958
 * 
 * 
 * def LD_65(cpu): # 65 LD H,L             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | ((cpu.HL & 0xFF) << 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":965
 * 
 * 
 * def LD_66(cpu): # 66 LD H,(HL)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.mb.getitem(cpu.HL) << 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_66(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":966
 * 
 * def LD_66(cpu): # 66 LD H,(HL)
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.mb.getitem(cpu.HL) << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) << 8));

  /* "pyboy/core/opcodes.py":967
 * def LD_66(cpu): # 66 LD H,(HL)
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.mb.getitem(cpu.HL) << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":968
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.mb.getitem(cpu.HL) << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":969
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":965
 * 
 * 
 * def LD_66(cpu): # 66 LD H,(HL)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.mb.getitem(cpu.HL) << 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":972
 * 
 * 
 * def LD_67(cpu): # 67 LD H,A             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.A << 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_67(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":973
 * 
 * def LD_67(cpu): # 67 LD H,A
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.A << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_cpu->A << 8));

  /* "pyboy/core/opcodes.py":974
 * def LD_67(cpu): # 67 LD H,A
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.A << 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":975
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.A << 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":976
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":972
 * 
 * 
 * def LD_67(cpu): # 67 LD H,A             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (cpu.A << 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":979
 * 
 * 
 * def LD_68(cpu): # 68 LD L,B             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.B & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_68(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":980
 * 
 * def LD_68(cpu): # 68 LD L,B
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.B & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_cpu->B & 0xFF));

  /* "pyboy/core/opcodes.py":981
 * def LD_68(cpu): # 68 LD L,B
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.B & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":982
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.B & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":983
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":979
 * 
 * 
 * def LD_68(cpu): # 68 LD L,B             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.B & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":986
 * 
 * 
 * def LD_69(cpu): # 69 LD L,C             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.C & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_69(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":987
 * 
 * def LD_69(cpu): # 69 LD L,C
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.C & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_cpu->C & 0xFF));

  /* "pyboy/core/opcodes.py":988
 * def LD_69(cpu): # 69 LD L,C
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.C & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":989
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.C & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":990
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":986
 * 
 * 
 * def LD_69(cpu): # 69 LD L,C             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.C & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":993
 * 
 * 
 * def LD_6A(cpu): # 6A LD L,D             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.D & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":994
 * 
 * def LD_6A(cpu): # 6A LD L,D
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.D & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_cpu->D & 0xFF));

  /* "pyboy/core/opcodes.py":995
 * def LD_6A(cpu): # 6A LD L,D
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.D & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":996
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.D & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":997
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":993
 * 
 * 
 * def LD_6A(cpu): # 6A LD L,D             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.D & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1000
 * 
 * 
 * def LD_6B(cpu): # 6B LD L,E             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.E & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1001
 * 
 * def LD_6B(cpu): # 6B LD L,E
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.E & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_cpu->E & 0xFF));

  /* "pyboy/core/opcodes.py":1002
 * def LD_6B(cpu): # 6B LD L,E
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.E & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1003
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.E & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1004
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1000
 * 
 * 
 * def LD_6B(cpu): # 6B LD L,E             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.E & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1007
 * 
 * 
 * def LD_6C(cpu): # 6C LD L,H             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL >> 8) & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1008
 * 
 * def LD_6C(cpu): # 6C LD L,H
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL >> 8) & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | ((__pyx_v_cpu->HL >> 8) & 0xFF));

  /* "pyboy/core/opcodes.py":1009
 * def LD_6C(cpu): # 6C LD L,H
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL >> 8) & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1010
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL >> 8) & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1011
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1007
 * 
 * 
 * def LD_6C(cpu): # 6C LD L,H             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL >> 8) & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1014
 * 
 * 
 * def LD_6D(cpu): # 6D LD L,L             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL & 0xFF) & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1015
 * 
 * def LD_6D(cpu): # 6D LD L,L
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL & 0xFF) & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | ((__pyx_v_cpu->HL & 0xFF) & 0xFF));

  /* "pyboy/core/opcodes.py":1016
 * def LD_6D(cpu): # 6D LD L,L
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL & 0xFF) & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1017
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL & 0xFF) & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1018
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1014
 * 
 * 
 * def LD_6D(cpu): # 6D LD L,L             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | ((cpu.HL & 0xFF) & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1021
 * 
 * 
 * def LD_6E(cpu): # 6E LD L,(HL)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.mb.getitem(cpu.HL) & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1022
 * 
 * def LD_6E(cpu): # 6E LD L,(HL)
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.mb.getitem(cpu.HL) & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0xFF));

  /* "pyboy/core/opcodes.py":1023
 * def LD_6E(cpu): # 6E LD L,(HL)
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.mb.getitem(cpu.HL) & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1024
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.mb.getitem(cpu.HL) & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1025
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1021
 * 
 * 
 * def LD_6E(cpu): # 6E LD L,(HL)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.mb.getitem(cpu.HL) & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1028
 * 
 * 
 * def LD_6F(cpu): # 6F LD L,A             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.A & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_6F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1029
 * 
 * def LD_6F(cpu): # 6F LD L,A
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.A & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_cpu->A & 0xFF));

  /* "pyboy/core/opcodes.py":1030
 * def LD_6F(cpu): # 6F LD L,A
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.A & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1031
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.A & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1032
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1028
 * 
 * 
 * def LD_6F(cpu): # 6F LD L,A             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (cpu.A & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1035
 * 
 * 
 * def LD_70(cpu): # 70 LD (HL),B             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.B)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_70(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1036
 * 
 * def LD_70(cpu): # 70 LD (HL),B
 *     cpu.mb.setitem(cpu.HL, cpu.B)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_cpu->B);

  /* "pyboy/core/opcodes.py":1037
 * def LD_70(cpu): # 70 LD (HL),B
 *     cpu.mb.setitem(cpu.HL, cpu.B)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1038
 *     cpu.mb.setitem(cpu.HL, cpu.B)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1039
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1035
 * 
 * 
 * def LD_70(cpu): # 70 LD (HL),B             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.B)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1042
 * 
 * 
 * def LD_71(cpu): # 71 LD (HL),C             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.C)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_71(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1043
 * 
 * def LD_71(cpu): # 71 LD (HL),C
 *     cpu.mb.setitem(cpu.HL, cpu.C)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_cpu->C);

  /* "pyboy/core/opcodes.py":1044
 * def LD_71(cpu): # 71 LD (HL),C
 *     cpu.mb.setitem(cpu.HL, cpu.C)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1045
 *     cpu.mb.setitem(cpu.HL, cpu.C)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1046
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1042
 * 
 * 
 * def LD_71(cpu): # 71 LD (HL),C             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.C)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1049
 * 
 * 
 * def LD_72(cpu): # 72 LD (HL),D             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.D)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_72(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1050
 * 
 * def LD_72(cpu): # 72 LD (HL),D
 *     cpu.mb.setitem(cpu.HL, cpu.D)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_cpu->D);

  /* "pyboy/core/opcodes.py":1051
 * def LD_72(cpu): # 72 LD (HL),D
 *     cpu.mb.setitem(cpu.HL, cpu.D)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1052
 *     cpu.mb.setitem(cpu.HL, cpu.D)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1053
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1049
 * 
 * 
 * def LD_72(cpu): # 72 LD (HL),D             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.D)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1056
 * 
 * 
 * def LD_73(cpu): # 73 LD (HL),E             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.E)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_73(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1057
 * 
 * def LD_73(cpu): # 73 LD (HL),E
 *     cpu.mb.setitem(cpu.HL, cpu.E)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_cpu->E);

  /* "pyboy/core/opcodes.py":1058
 * def LD_73(cpu): # 73 LD (HL),E
 *     cpu.mb.setitem(cpu.HL, cpu.E)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1059
 *     cpu.mb.setitem(cpu.HL, cpu.E)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1060
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1056
 * 
 * 
 * def LD_73(cpu): # 73 LD (HL),E             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.E)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1063
 * 
 * 
 * def LD_74(cpu): # 74 LD (HL),H             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, (cpu.HL >> 8))
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_74(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1064
 * 
 * def LD_74(cpu): # 74 LD (HL),H
 *     cpu.mb.setitem(cpu.HL, (cpu.HL >> 8))             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, (__pyx_v_cpu->HL >> 8));

  /* "pyboy/core/opcodes.py":1065
 * def LD_74(cpu): # 74 LD (HL),H
 *     cpu.mb.setitem(cpu.HL, (cpu.HL >> 8))
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1066
 *     cpu.mb.setitem(cpu.HL, (cpu.HL >> 8))
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1067
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1063
 * 
 * 
 * def LD_74(cpu): # 74 LD (HL),H             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, (cpu.HL >> 8))
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1070
 * 
 * 
 * def LD_75(cpu): # 75 LD (HL),L             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, (cpu.HL & 0xFF))
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_75(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1071
 * 
 * def LD_75(cpu): # 75 LD (HL),L
 *     cpu.mb.setitem(cpu.HL, (cpu.HL & 0xFF))             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, (__pyx_v_cpu->HL & 0xFF));

  /* "pyboy/core/opcodes.py":1072
 * def LD_75(cpu): # 75 LD (HL),L
 *     cpu.mb.setitem(cpu.HL, (cpu.HL & 0xFF))
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1073
 *     cpu.mb.setitem(cpu.HL, (cpu.HL & 0xFF))
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1074
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1070
 * 
 * 
 * def LD_75(cpu): # 75 LD (HL),L             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, (cpu.HL & 0xFF))
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1077
 * 
 * 
 * def HALT_76(cpu): # 76 HALT             # <<<<<<<<<<<<<<
 *     cpu.halted = True
 *     return 4
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_HALT_76(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1078
 * 
 * def HALT_76(cpu): # 76 HALT
 *     cpu.halted = True             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->halted = 1;

  /* "pyboy/core/opcodes.py":1079
 * def HALT_76(cpu): # 76 HALT
 *     cpu.halted = True
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1077
 * 
 * 
 * def HALT_76(cpu): # 76 HALT             # <<<<<<<<<<<<<<
 *     cpu.halted = True
 *     return 4
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1082
 * 
 * 
 * def LD_77(cpu): # 77 LD (HL),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_77(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1083
 * 
 * def LD_77(cpu): # 77 LD (HL),A
 *     cpu.mb.setitem(cpu.HL, cpu.A)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":1084
 * def LD_77(cpu): # 77 LD (HL),A
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1085
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1086
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1082
 * 
 * 
 * def LD_77(cpu): # 77 LD (HL),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, cpu.A)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1089
 * 
 * 
 * def LD_78(cpu): # 78 LD A,B             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.B
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_78(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":1090
 * 
 * def LD_78(cpu): # 78 LD A,B
 *     cpu.A = cpu.B             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->B;
  __pyx_v_cpu->A = __pyx_t_1;

  /* "pyboy/core/opcodes.py":1091
 * def LD_78(cpu): # 78 LD A,B
 *     cpu.A = cpu.B
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1092
 *     cpu.A = cpu.B
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1093
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1089
 * 
 * 
 * def LD_78(cpu): # 78 LD A,B             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.B
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1096
 * 
 * 
 * def LD_79(cpu): # 79 LD A,C             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.C
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_79(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":1097
 * 
 * def LD_79(cpu): # 79 LD A,C
 *     cpu.A = cpu.C             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->C;
  __pyx_v_cpu->A = __pyx_t_1;

  /* "pyboy/core/opcodes.py":1098
 * def LD_79(cpu): # 79 LD A,C
 *     cpu.A = cpu.C
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1099
 *     cpu.A = cpu.C
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1100
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1096
 * 
 * 
 * def LD_79(cpu): # 79 LD A,C             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.C
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1103
 * 
 * 
 * def LD_7A(cpu): # 7A LD A,D             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.D
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":1104
 * 
 * def LD_7A(cpu): # 7A LD A,D
 *     cpu.A = cpu.D             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->D;
  __pyx_v_cpu->A = __pyx_t_1;

  /* "pyboy/core/opcodes.py":1105
 * def LD_7A(cpu): # 7A LD A,D
 *     cpu.A = cpu.D
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1106
 *     cpu.A = cpu.D
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1107
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1103
 * 
 * 
 * def LD_7A(cpu): # 7A LD A,D             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.D
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1110
 * 
 * 
 * def LD_7B(cpu): # 7B LD A,E             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.E
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":1111
 * 
 * def LD_7B(cpu): # 7B LD A,E
 *     cpu.A = cpu.E             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->E;
  __pyx_v_cpu->A = __pyx_t_1;

  /* "pyboy/core/opcodes.py":1112
 * def LD_7B(cpu): # 7B LD A,E
 *     cpu.A = cpu.E
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1113
 *     cpu.A = cpu.E
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1114
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1110
 * 
 * 
 * def LD_7B(cpu): # 7B LD A,E             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.E
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1117
 * 
 * 
 * def LD_7C(cpu): # 7C LD A,H             # <<<<<<<<<<<<<<
 *     cpu.A = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1118
 * 
 * def LD_7C(cpu): # 7C LD A,H
 *     cpu.A = (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = (__pyx_v_cpu->HL >> 8);

  /* "pyboy/core/opcodes.py":1119
 * def LD_7C(cpu): # 7C LD A,H
 *     cpu.A = (cpu.HL >> 8)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1120
 *     cpu.A = (cpu.HL >> 8)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1121
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1117
 * 
 * 
 * def LD_7C(cpu): # 7C LD A,H             # <<<<<<<<<<<<<<
 *     cpu.A = (cpu.HL >> 8)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1124
 * 
 * 
 * def LD_7D(cpu): # 7D LD A,L             # <<<<<<<<<<<<<<
 *     cpu.A = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1125
 * 
 * def LD_7D(cpu): # 7D LD A,L
 *     cpu.A = (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = (__pyx_v_cpu->HL & 0xFF);

  /* "pyboy/core/opcodes.py":1126
 * def LD_7D(cpu): # 7D LD A,L
 *     cpu.A = (cpu.HL & 0xFF)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1127
 *     cpu.A = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1128
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1124
 * 
 * 
 * def LD_7D(cpu): # 7D LD A,L             # <<<<<<<<<<<<<<
 *     cpu.A = (cpu.HL & 0xFF)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1131
 * 
 * 
 * def LD_7E(cpu): # 7E LD A,(HL)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1132
 * 
 * def LD_7E(cpu): # 7E LD A,(HL)
 *     cpu.A = cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL);

  /* "pyboy/core/opcodes.py":1133
 * def LD_7E(cpu): # 7E LD A,(HL)
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1134
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1135
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1131
 * 
 * 
 * def LD_7E(cpu): # 7E LD A,(HL)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(cpu.HL)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1138
 * 
 * 
 * def LD_7F(cpu): # 7F LD A,A             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.A
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_7F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int16_t __pyx_t_1;

  /* "pyboy/core/opcodes.py":1139
 * 
 * def LD_7F(cpu): # 7F LD A,A
 *     cpu.A = cpu.A             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->A;
  __pyx_v_cpu->A = __pyx_t_1;

  /* "pyboy/core/opcodes.py":1140
 * def LD_7F(cpu): # 7F LD A,A
 *     cpu.A = cpu.A
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1141
 *     cpu.A = cpu.A
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1142
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1138
 * 
 * 
 * def LD_7F(cpu): # 7F LD A,A             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.A
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1145
 * 
 * 
 * def ADD_80(cpu): # 80 ADD A,B             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.B
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_80(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1146
 * 
 * def ADD_80(cpu): # 80 ADD A,B
 *     t = cpu.A + cpu.B             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + __pyx_v_cpu->B);

  /* "pyboy/core/opcodes.py":1147
 * def ADD_80(cpu): # 80 ADD A,B
 *     t = cpu.A + cpu.B
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1148
 *     t = cpu.A + cpu.B
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1149
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->B & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1150
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1151
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1152
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1153
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1154
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1155
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1156
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1157
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1145
 * 
 * 
 * def ADD_80(cpu): # 80 ADD A,B             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.B
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1160
 * 
 * 
 * def ADD_81(cpu): # 81 ADD A,C             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.C
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_81(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1161
 * 
 * def ADD_81(cpu): # 81 ADD A,C
 *     t = cpu.A + cpu.C             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + __pyx_v_cpu->C);

  /* "pyboy/core/opcodes.py":1162
 * def ADD_81(cpu): # 81 ADD A,C
 *     t = cpu.A + cpu.C
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1163
 *     t = cpu.A + cpu.C
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1164
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->C & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1165
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1166
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1167
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1168
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1169
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1170
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1171
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1172
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1160
 * 
 * 
 * def ADD_81(cpu): # 81 ADD A,C             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.C
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1175
 * 
 * 
 * def ADD_82(cpu): # 82 ADD A,D             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.D
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_82(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1176
 * 
 * def ADD_82(cpu): # 82 ADD A,D
 *     t = cpu.A + cpu.D             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + __pyx_v_cpu->D);

  /* "pyboy/core/opcodes.py":1177
 * def ADD_82(cpu): # 82 ADD A,D
 *     t = cpu.A + cpu.D
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1178
 *     t = cpu.A + cpu.D
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1179
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->D & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1180
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1181
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1182
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1183
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1184
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1185
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1186
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1187
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1175
 * 
 * 
 * def ADD_82(cpu): # 82 ADD A,D             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.D
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1190
 * 
 * 
 * def ADD_83(cpu): # 83 ADD A,E             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.E
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_83(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1191
 * 
 * def ADD_83(cpu): # 83 ADD A,E
 *     t = cpu.A + cpu.E             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + __pyx_v_cpu->E);

  /* "pyboy/core/opcodes.py":1192
 * def ADD_83(cpu): # 83 ADD A,E
 *     t = cpu.A + cpu.E
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1193
 *     t = cpu.A + cpu.E
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1194
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->E & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1195
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1196
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1197
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1198
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1199
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1200
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1201
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1202
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1190
 * 
 * 
 * def ADD_83(cpu): # 83 ADD A,E             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.E
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1205
 * 
 * 
 * def ADD_84(cpu): # 84 ADD A,H             # <<<<<<<<<<<<<<
 *     t = cpu.A + (cpu.HL >> 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_84(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1206
 * 
 * def ADD_84(cpu): # 84 ADD A,H
 *     t = cpu.A + (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + (__pyx_v_cpu->HL >> 8));

  /* "pyboy/core/opcodes.py":1207
 * def ADD_84(cpu): # 84 ADD A,H
 *     t = cpu.A + (cpu.HL >> 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1208
 *     t = cpu.A + (cpu.HL >> 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1209
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + ((__pyx_v_cpu->HL >> 8) & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1210
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1211
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1212
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1213
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1214
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1215
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1216
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1217
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1205
 * 
 * 
 * def ADD_84(cpu): # 84 ADD A,H             # <<<<<<<<<<<<<<
 *     t = cpu.A + (cpu.HL >> 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1220
 * 
 * 
 * def ADD_85(cpu): # 85 ADD A,L             # <<<<<<<<<<<<<<
 *     t = cpu.A + (cpu.HL & 0xFF)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_85(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1221
 * 
 * def ADD_85(cpu): # 85 ADD A,L
 *     t = cpu.A + (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + (__pyx_v_cpu->HL & 0xFF));

  /* "pyboy/core/opcodes.py":1222
 * def ADD_85(cpu): # 85 ADD A,L
 *     t = cpu.A + (cpu.HL & 0xFF)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1223
 *     t = cpu.A + (cpu.HL & 0xFF)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1224
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + ((__pyx_v_cpu->HL & 0xFF) & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1225
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1226
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1227
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1228
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1229
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1230
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1231
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1232
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1220
 * 
 * 
 * def ADD_85(cpu): # 85 ADD A,L             # <<<<<<<<<<<<<<
 *     t = cpu.A + (cpu.HL & 0xFF)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1235
 * 
 * 
 * def ADD_86(cpu): # 86 ADD A,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_86(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1236
 * 
 * def ADD_86(cpu): # 86 ADD A,(HL)
 *     t = cpu.A + cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL));

  /* "pyboy/core/opcodes.py":1237
 * def ADD_86(cpu): # 86 ADD A,(HL)
 *     t = cpu.A + cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1238
 *     t = cpu.A + cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1239
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1240
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1241
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1242
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1243
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1244
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1245
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1246
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1247
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1235
 * 
 * 
 * def ADD_86(cpu): # 86 ADD A,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1250
 * 
 * 
 * def ADD_87(cpu): # 87 ADD A,A             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.A
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_87(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1251
 * 
 * def ADD_87(cpu): # 87 ADD A,A
 *     t = cpu.A + cpu.A             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":1252
 * def ADD_87(cpu): # 87 ADD A,A
 *     t = cpu.A + cpu.A
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1253
 *     t = cpu.A + cpu.A
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1254
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->A & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1255
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1256
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1257
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1258
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1259
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1260
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1261
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1262
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1250
 * 
 * 
 * def ADD_87(cpu): # 87 ADD A,A             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.A
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1265
 * 
 * 
 * def ADC_88(cpu): # 88 ADC A,B             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.B + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_88(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1266
 * 
 * def ADC_88(cpu): # 88 ADC A,B
 *     t = cpu.A + cpu.B + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A + __pyx_v_cpu->B) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1267
 * def ADC_88(cpu): # 88 ADC A,B
 *     t = cpu.A + cpu.B + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1268
 *     t = cpu.A + cpu.B + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1269
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->B & 0xF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1270
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1271
 *     flag += (((cpu.A & 0xF) + (cpu.B & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1272
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1273
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1274
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1275
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1276
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1277
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1265
 * 
 * 
 * def ADC_88(cpu): # 88 ADC A,B             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.B + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1280
 * 
 * 
 * def ADC_89(cpu): # 89 ADC A,C             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.C + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_89(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1281
 * 
 * def ADC_89(cpu): # 89 ADC A,C
 *     t = cpu.A + cpu.C + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A + __pyx_v_cpu->C) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1282
 * def ADC_89(cpu): # 89 ADC A,C
 *     t = cpu.A + cpu.C + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1283
 *     t = cpu.A + cpu.C + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1284
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->C & 0xF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1285
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1286
 *     flag += (((cpu.A & 0xF) + (cpu.C & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1287
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1288
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1289
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1290
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1291
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1292
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1280
 * 
 * 
 * def ADC_89(cpu): # 89 ADC A,C             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.C + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1295
 * 
 * 
 * def ADC_8A(cpu): # 8A ADC A,D             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.D + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1296
 * 
 * def ADC_8A(cpu): # 8A ADC A,D
 *     t = cpu.A + cpu.D + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A + __pyx_v_cpu->D) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1297
 * def ADC_8A(cpu): # 8A ADC A,D
 *     t = cpu.A + cpu.D + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1298
 *     t = cpu.A + cpu.D + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1299
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->D & 0xF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1300
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1301
 *     flag += (((cpu.A & 0xF) + (cpu.D & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1302
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1303
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1304
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1305
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1306
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1307
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1295
 * 
 * 
 * def ADC_8A(cpu): # 8A ADC A,D             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.D + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1310
 * 
 * 
 * def ADC_8B(cpu): # 8B ADC A,E             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.E + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1311
 * 
 * def ADC_8B(cpu): # 8B ADC A,E
 *     t = cpu.A + cpu.E + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A + __pyx_v_cpu->E) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1312
 * def ADC_8B(cpu): # 8B ADC A,E
 *     t = cpu.A + cpu.E + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1313
 *     t = cpu.A + cpu.E + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1314
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->E & 0xF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1315
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1316
 *     flag += (((cpu.A & 0xF) + (cpu.E & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1317
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1318
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1319
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1320
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1321
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1322
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1310
 * 
 * 
 * def ADC_8B(cpu): # 8B ADC A,E             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.E + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1325
 * 
 * 
 * def ADC_8C(cpu): # 8C ADC A,H             # <<<<<<<<<<<<<<
 *     t = cpu.A + (cpu.HL >> 8) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1326
 * 
 * def ADC_8C(cpu): # 8C ADC A,H
 *     t = cpu.A + (cpu.HL >> 8) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A + (__pyx_v_cpu->HL >> 8)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1327
 * def ADC_8C(cpu): # 8C ADC A,H
 *     t = cpu.A + (cpu.HL >> 8) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1328
 *     t = cpu.A + (cpu.HL >> 8) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1329
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) + ((__pyx_v_cpu->HL >> 8) & 0xF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1330
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1331
 *     flag += (((cpu.A & 0xF) + ((cpu.HL >> 8) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1332
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1333
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1334
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1335
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1336
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1337
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1325
 * 
 * 
 * def ADC_8C(cpu): # 8C ADC A,H             # <<<<<<<<<<<<<<
 *     t = cpu.A + (cpu.HL >> 8) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1340
 * 
 * 
 * def ADC_8D(cpu): # 8D ADC A,L             # <<<<<<<<<<<<<<
 *     t = cpu.A + (cpu.HL & 0xFF) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1341
 * 
 * def ADC_8D(cpu): # 8D ADC A,L
 *     t = cpu.A + (cpu.HL & 0xFF) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A + (__pyx_v_cpu->HL & 0xFF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1342
 * def ADC_8D(cpu): # 8D ADC A,L
 *     t = cpu.A + (cpu.HL & 0xFF) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1343
 *     t = cpu.A + (cpu.HL & 0xFF) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1344
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) + ((__pyx_v_cpu->HL & 0xFF) & 0xF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1345
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1346
 *     flag += (((cpu.A & 0xF) + ((cpu.HL & 0xFF) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1347
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1348
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1349
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1350
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1351
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1352
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1340
 * 
 * 
 * def ADC_8D(cpu): # 8D ADC A,L             # <<<<<<<<<<<<<<
 *     t = cpu.A + (cpu.HL & 0xFF) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1355
 * 
 * 
 * def ADC_8E(cpu): # 8E ADC A,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.mb.getitem(cpu.HL) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1356
 * 
 * def ADC_8E(cpu): # 8E ADC A,(HL)
 *     t = cpu.A + cpu.mb.getitem(cpu.HL) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A + ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1357
 * def ADC_8E(cpu): # 8E ADC A,(HL)
 *     t = cpu.A + cpu.mb.getitem(cpu.HL) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1358
 *     t = cpu.A + cpu.mb.getitem(cpu.HL) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1359
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) + (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0xF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1360
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1361
 *     flag += (((cpu.A & 0xF) + (cpu.mb.getitem(cpu.HL) & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1362
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1363
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1364
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1365
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1366
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1367
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1355
 * 
 * 
 * def ADC_8E(cpu): # 8E ADC A,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.mb.getitem(cpu.HL) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1370
 * 
 * 
 * def ADC_8F(cpu): # 8F ADC A,A             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.A + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_8F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1371
 * 
 * def ADC_8F(cpu): # 8F ADC A,A
 *     t = cpu.A + cpu.A + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A + __pyx_v_cpu->A) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1372
 * def ADC_8F(cpu): # 8F ADC A,A
 *     t = cpu.A + cpu.A + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1373
 *     t = cpu.A + cpu.A + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1374
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) + (__pyx_v_cpu->A & 0xF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1375
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1376
 *     flag += (((cpu.A & 0xF) + (cpu.A & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1377
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1378
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1379
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1380
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1381
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1382
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1370
 * 
 * 
 * def ADC_8F(cpu): # 8F ADC A,A             # <<<<<<<<<<<<<<
 *     t = cpu.A + cpu.A + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1385
 * 
 * 
 * def SUB_90(cpu): # 90 SUB B             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.B
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_90(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1386
 * 
 * def SUB_90(cpu): # 90 SUB B
 *     t = cpu.A - cpu.B             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->B);

  /* "pyboy/core/opcodes.py":1387
 * def SUB_90(cpu): # 90 SUB B
 *     t = cpu.A - cpu.B
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1388
 *     t = cpu.A - cpu.B
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1389
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->B & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1390
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1391
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1392
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1393
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1394
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1395
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1396
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1397
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1385
 * 
 * 
 * def SUB_90(cpu): # 90 SUB B             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.B
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1400
 * 
 * 
 * def SUB_91(cpu): # 91 SUB C             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.C
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_91(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1401
 * 
 * def SUB_91(cpu): # 91 SUB C
 *     t = cpu.A - cpu.C             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->C);

  /* "pyboy/core/opcodes.py":1402
 * def SUB_91(cpu): # 91 SUB C
 *     t = cpu.A - cpu.C
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1403
 *     t = cpu.A - cpu.C
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1404
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->C & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1405
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1406
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1407
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1408
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1409
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1410
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1411
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1412
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1400
 * 
 * 
 * def SUB_91(cpu): # 91 SUB C             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.C
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1415
 * 
 * 
 * def SUB_92(cpu): # 92 SUB D             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.D
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_92(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1416
 * 
 * def SUB_92(cpu): # 92 SUB D
 *     t = cpu.A - cpu.D             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->D);

  /* "pyboy/core/opcodes.py":1417
 * def SUB_92(cpu): # 92 SUB D
 *     t = cpu.A - cpu.D
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1418
 *     t = cpu.A - cpu.D
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1419
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->D & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1420
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1421
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1422
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1423
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1424
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1425
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1426
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1427
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1415
 * 
 * 
 * def SUB_92(cpu): # 92 SUB D             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.D
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1430
 * 
 * 
 * def SUB_93(cpu): # 93 SUB E             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.E
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_93(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1431
 * 
 * def SUB_93(cpu): # 93 SUB E
 *     t = cpu.A - cpu.E             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->E);

  /* "pyboy/core/opcodes.py":1432
 * def SUB_93(cpu): # 93 SUB E
 *     t = cpu.A - cpu.E
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1433
 *     t = cpu.A - cpu.E
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1434
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->E & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1435
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1436
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1437
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1438
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1439
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1440
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1441
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1442
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1430
 * 
 * 
 * def SUB_93(cpu): # 93 SUB E             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.E
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1445
 * 
 * 
 * def SUB_94(cpu): # 94 SUB H             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL >> 8)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_94(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1446
 * 
 * def SUB_94(cpu): # 94 SUB H
 *     t = cpu.A - (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - (__pyx_v_cpu->HL >> 8));

  /* "pyboy/core/opcodes.py":1447
 * def SUB_94(cpu): # 94 SUB H
 *     t = cpu.A - (cpu.HL >> 8)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1448
 *     t = cpu.A - (cpu.HL >> 8)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1449
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - ((__pyx_v_cpu->HL >> 8) & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1450
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1451
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1452
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1453
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1454
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1455
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1456
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1457
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1445
 * 
 * 
 * def SUB_94(cpu): # 94 SUB H             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL >> 8)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1460
 * 
 * 
 * def SUB_95(cpu): # 95 SUB L             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL & 0xFF)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_95(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1461
 * 
 * def SUB_95(cpu): # 95 SUB L
 *     t = cpu.A - (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - (__pyx_v_cpu->HL & 0xFF));

  /* "pyboy/core/opcodes.py":1462
 * def SUB_95(cpu): # 95 SUB L
 *     t = cpu.A - (cpu.HL & 0xFF)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1463
 *     t = cpu.A - (cpu.HL & 0xFF)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1464
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - ((__pyx_v_cpu->HL & 0xFF) & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1465
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1466
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1467
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1468
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1469
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1470
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1471
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1472
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1460
 * 
 * 
 * def SUB_95(cpu): # 95 SUB L             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL & 0xFF)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1475
 * 
 * 
 * def SUB_96(cpu): # 96 SUB (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_96(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1476
 * 
 * def SUB_96(cpu): # 96 SUB (HL)
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL));

  /* "pyboy/core/opcodes.py":1477
 * def SUB_96(cpu): # 96 SUB (HL)
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1478
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1479
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1480
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1481
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1482
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1483
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1484
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1485
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1486
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1487
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1475
 * 
 * 
 * def SUB_96(cpu): # 96 SUB (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1490
 * 
 * 
 * def SUB_97(cpu): # 97 SUB A             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.A
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_97(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1491
 * 
 * def SUB_97(cpu): # 97 SUB A
 *     t = cpu.A - cpu.A             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":1492
 * def SUB_97(cpu): # 97 SUB A
 *     t = cpu.A - cpu.A
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1493
 *     t = cpu.A - cpu.A
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1494
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->A & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1495
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1496
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1497
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1498
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1499
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1500
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1501
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1502
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1490
 * 
 * 
 * def SUB_97(cpu): # 97 SUB A             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.A
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1505
 * 
 * 
 * def SBC_98(cpu): # 98 SBC A,B             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.B - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_98(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1506
 * 
 * def SBC_98(cpu): # 98 SBC A,B
 *     t = cpu.A - cpu.B - ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A - __pyx_v_cpu->B) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1507
 * def SBC_98(cpu): # 98 SBC A,B
 *     t = cpu.A - cpu.B - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1508
 *     t = cpu.A - cpu.B - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1509
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->B & 0xF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1510
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1511
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1512
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1513
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1514
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1515
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1516
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1517
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1505
 * 
 * 
 * def SBC_98(cpu): # 98 SBC A,B             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.B - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1520
 * 
 * 
 * def SBC_99(cpu): # 99 SBC A,C             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.C - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_99(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1521
 * 
 * def SBC_99(cpu): # 99 SBC A,C
 *     t = cpu.A - cpu.C - ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A - __pyx_v_cpu->C) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1522
 * def SBC_99(cpu): # 99 SBC A,C
 *     t = cpu.A - cpu.C - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1523
 *     t = cpu.A - cpu.C - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1524
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->C & 0xF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1525
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1526
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1527
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1528
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1529
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1530
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1531
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1532
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1520
 * 
 * 
 * def SBC_99(cpu): # 99 SBC A,C             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.C - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1535
 * 
 * 
 * def SBC_9A(cpu): # 9A SBC A,D             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.D - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1536
 * 
 * def SBC_9A(cpu): # 9A SBC A,D
 *     t = cpu.A - cpu.D - ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A - __pyx_v_cpu->D) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1537
 * def SBC_9A(cpu): # 9A SBC A,D
 *     t = cpu.A - cpu.D - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1538
 *     t = cpu.A - cpu.D - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1539
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->D & 0xF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1540
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1541
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1542
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1543
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1544
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1545
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1546
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1547
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1535
 * 
 * 
 * def SBC_9A(cpu): # 9A SBC A,D             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.D - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1550
 * 
 * 
 * def SBC_9B(cpu): # 9B SBC A,E             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.E - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1551
 * 
 * def SBC_9B(cpu): # 9B SBC A,E
 *     t = cpu.A - cpu.E - ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A - __pyx_v_cpu->E) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1552
 * def SBC_9B(cpu): # 9B SBC A,E
 *     t = cpu.A - cpu.E - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1553
 *     t = cpu.A - cpu.E - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1554
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->E & 0xF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1555
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1556
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1557
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1558
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1559
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1560
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1561
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1562
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1550
 * 
 * 
 * def SBC_9B(cpu): # 9B SBC A,E             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.E - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1565
 * 
 * 
 * def SBC_9C(cpu): # 9C SBC A,H             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL >> 8) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1566
 * 
 * def SBC_9C(cpu): # 9C SBC A,H
 *     t = cpu.A - (cpu.HL >> 8) - ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A - (__pyx_v_cpu->HL >> 8)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1567
 * def SBC_9C(cpu): # 9C SBC A,H
 *     t = cpu.A - (cpu.HL >> 8) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1568
 *     t = cpu.A - (cpu.HL >> 8) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1569
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) - ((__pyx_v_cpu->HL >> 8) & 0xF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1570
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1571
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1572
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1573
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1574
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1575
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1576
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1577
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1565
 * 
 * 
 * def SBC_9C(cpu): # 9C SBC A,H             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL >> 8) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1580
 * 
 * 
 * def SBC_9D(cpu): # 9D SBC A,L             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL & 0xFF) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1581
 * 
 * def SBC_9D(cpu): # 9D SBC A,L
 *     t = cpu.A - (cpu.HL & 0xFF) - ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A - (__pyx_v_cpu->HL & 0xFF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1582
 * def SBC_9D(cpu): # 9D SBC A,L
 *     t = cpu.A - (cpu.HL & 0xFF) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1583
 *     t = cpu.A - (cpu.HL & 0xFF) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1584
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) - ((__pyx_v_cpu->HL & 0xFF) & 0xF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1585
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1586
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1587
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1588
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1589
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1590
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1591
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1592
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1580
 * 
 * 
 * def SBC_9D(cpu): # 9D SBC A,L             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL & 0xFF) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1595
 * 
 * 
 * def SBC_9E(cpu): # 9E SBC A,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.mb.getitem(cpu.HL) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1596
 * 
 * def SBC_9E(cpu): # 9E SBC A,(HL)
 *     t = cpu.A - cpu.mb.getitem(cpu.HL) - ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A - ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1597
 * def SBC_9E(cpu): # 9E SBC A,(HL)
 *     t = cpu.A - cpu.mb.getitem(cpu.HL) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1598
 *     t = cpu.A - cpu.mb.getitem(cpu.HL) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1599
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) - (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0xF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1600
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1601
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1602
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1603
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1604
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1605
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1606
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1607
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1595
 * 
 * 
 * def SBC_9E(cpu): # 9E SBC A,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.mb.getitem(cpu.HL) - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1610
 * 
 * 
 * def SBC_9F(cpu): # 9F SBC A,A             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.A - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_9F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1611
 * 
 * def SBC_9F(cpu): # 9F SBC A,A
 *     t = cpu.A - cpu.A - ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A - __pyx_v_cpu->A) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":1612
 * def SBC_9F(cpu): # 9F SBC A,A
 *     t = cpu.A - cpu.A - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1613
 *     t = cpu.A - cpu.A - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1614
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->A & 0xF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1615
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1616
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1617
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1618
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1619
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1620
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1621
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1622
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1610
 * 
 * 
 * def SBC_9F(cpu): # 9F SBC A,A             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.A - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1625
 * 
 * 
 * def AND_A0(cpu): # A0 AND B             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.B
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1626
 * 
 * def AND_A0(cpu): # A0 AND B
 *     t = cpu.A & cpu.B             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & __pyx_v_cpu->B);

  /* "pyboy/core/opcodes.py":1627
 * def AND_A0(cpu): # A0 AND B
 *     t = cpu.A & cpu.B
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":1628
 *     t = cpu.A & cpu.B
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1629
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1630
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1631
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1632
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1633
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1634
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1635
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1625
 * 
 * 
 * def AND_A0(cpu): # A0 AND B             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.B
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1638
 * 
 * 
 * def AND_A1(cpu): # A1 AND C             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.C
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1639
 * 
 * def AND_A1(cpu): # A1 AND C
 *     t = cpu.A & cpu.C             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & __pyx_v_cpu->C);

  /* "pyboy/core/opcodes.py":1640
 * def AND_A1(cpu): # A1 AND C
 *     t = cpu.A & cpu.C
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":1641
 *     t = cpu.A & cpu.C
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1642
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1643
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1644
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1645
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1646
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1647
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1648
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1638
 * 
 * 
 * def AND_A1(cpu): # A1 AND C             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.C
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1651
 * 
 * 
 * def AND_A2(cpu): # A2 AND D             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.D
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1652
 * 
 * def AND_A2(cpu): # A2 AND D
 *     t = cpu.A & cpu.D             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & __pyx_v_cpu->D);

  /* "pyboy/core/opcodes.py":1653
 * def AND_A2(cpu): # A2 AND D
 *     t = cpu.A & cpu.D
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":1654
 *     t = cpu.A & cpu.D
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1655
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1656
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1657
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1658
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1659
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1660
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1661
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1651
 * 
 * 
 * def AND_A2(cpu): # A2 AND D             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.D
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1664
 * 
 * 
 * def AND_A3(cpu): # A3 AND E             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.E
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1665
 * 
 * def AND_A3(cpu): # A3 AND E
 *     t = cpu.A & cpu.E             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & __pyx_v_cpu->E);

  /* "pyboy/core/opcodes.py":1666
 * def AND_A3(cpu): # A3 AND E
 *     t = cpu.A & cpu.E
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":1667
 *     t = cpu.A & cpu.E
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1668
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1669
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1670
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1671
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1672
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1673
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1674
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1664
 * 
 * 
 * def AND_A3(cpu): # A3 AND E             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.E
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1677
 * 
 * 
 * def AND_A4(cpu): # A4 AND H             # <<<<<<<<<<<<<<
 *     t = cpu.A & (cpu.HL >> 8)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1678
 * 
 * def AND_A4(cpu): # A4 AND H
 *     t = cpu.A & (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & (__pyx_v_cpu->HL >> 8));

  /* "pyboy/core/opcodes.py":1679
 * def AND_A4(cpu): # A4 AND H
 *     t = cpu.A & (cpu.HL >> 8)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":1680
 *     t = cpu.A & (cpu.HL >> 8)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1681
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1682
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1683
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1684
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1685
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1686
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1687
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1677
 * 
 * 
 * def AND_A4(cpu): # A4 AND H             # <<<<<<<<<<<<<<
 *     t = cpu.A & (cpu.HL >> 8)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1690
 * 
 * 
 * def AND_A5(cpu): # A5 AND L             # <<<<<<<<<<<<<<
 *     t = cpu.A & (cpu.HL & 0xFF)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1691
 * 
 * def AND_A5(cpu): # A5 AND L
 *     t = cpu.A & (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & (__pyx_v_cpu->HL & 0xFF));

  /* "pyboy/core/opcodes.py":1692
 * def AND_A5(cpu): # A5 AND L
 *     t = cpu.A & (cpu.HL & 0xFF)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":1693
 *     t = cpu.A & (cpu.HL & 0xFF)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1694
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1695
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1696
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1697
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1698
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1699
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1700
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1690
 * 
 * 
 * def AND_A5(cpu): # A5 AND L             # <<<<<<<<<<<<<<
 *     t = cpu.A & (cpu.HL & 0xFF)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1703
 * 
 * 
 * def AND_A6(cpu): # A6 AND (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.mb.getitem(cpu.HL)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1704
 * 
 * def AND_A6(cpu): # A6 AND (HL)
 *     t = cpu.A & cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL));

  /* "pyboy/core/opcodes.py":1705
 * def AND_A6(cpu): # A6 AND (HL)
 *     t = cpu.A & cpu.mb.getitem(cpu.HL)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":1706
 *     t = cpu.A & cpu.mb.getitem(cpu.HL)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1707
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1708
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1709
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1710
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1711
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1712
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1713
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1703
 * 
 * 
 * def AND_A6(cpu): # A6 AND (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.mb.getitem(cpu.HL)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1716
 * 
 * 
 * def AND_A7(cpu): # A7 AND A             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.A
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_A7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1717
 * 
 * def AND_A7(cpu): # A7 AND A
 *     t = cpu.A & cpu.A             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":1718
 * def AND_A7(cpu): # A7 AND A
 *     t = cpu.A & cpu.A
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":1719
 *     t = cpu.A & cpu.A
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1720
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1721
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1722
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1723
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1724
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1725
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1726
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1716
 * 
 * 
 * def AND_A7(cpu): # A7 AND A             # <<<<<<<<<<<<<<
 *     t = cpu.A & cpu.A
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1729
 * 
 * 
 * def XOR_A8(cpu): # A8 XOR B             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.B
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_A8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1730
 * 
 * def XOR_A8(cpu): # A8 XOR B
 *     t = cpu.A ^ cpu.B             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A ^ __pyx_v_cpu->B);

  /* "pyboy/core/opcodes.py":1731
 * def XOR_A8(cpu): # A8 XOR B
 *     t = cpu.A ^ cpu.B
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1732
 *     t = cpu.A ^ cpu.B
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1733
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1734
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1735
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1736
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1737
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1738
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1739
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1729
 * 
 * 
 * def XOR_A8(cpu): # A8 XOR B             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.B
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1742
 * 
 * 
 * def XOR_A9(cpu): # A9 XOR C             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.C
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_A9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1743
 * 
 * def XOR_A9(cpu): # A9 XOR C
 *     t = cpu.A ^ cpu.C             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A ^ __pyx_v_cpu->C);

  /* "pyboy/core/opcodes.py":1744
 * def XOR_A9(cpu): # A9 XOR C
 *     t = cpu.A ^ cpu.C
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1745
 *     t = cpu.A ^ cpu.C
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1746
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1747
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1748
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1749
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1750
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1751
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1752
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1742
 * 
 * 
 * def XOR_A9(cpu): # A9 XOR C             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.C
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1755
 * 
 * 
 * def XOR_AA(cpu): # AA XOR D             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.D
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1756
 * 
 * def XOR_AA(cpu): # AA XOR D
 *     t = cpu.A ^ cpu.D             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A ^ __pyx_v_cpu->D);

  /* "pyboy/core/opcodes.py":1757
 * def XOR_AA(cpu): # AA XOR D
 *     t = cpu.A ^ cpu.D
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1758
 *     t = cpu.A ^ cpu.D
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1759
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1760
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1761
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1762
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1763
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1764
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1765
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1755
 * 
 * 
 * def XOR_AA(cpu): # AA XOR D             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.D
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1768
 * 
 * 
 * def XOR_AB(cpu): # AB XOR E             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.E
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1769
 * 
 * def XOR_AB(cpu): # AB XOR E
 *     t = cpu.A ^ cpu.E             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A ^ __pyx_v_cpu->E);

  /* "pyboy/core/opcodes.py":1770
 * def XOR_AB(cpu): # AB XOR E
 *     t = cpu.A ^ cpu.E
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1771
 *     t = cpu.A ^ cpu.E
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1772
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1773
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1774
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1775
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1776
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1777
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1778
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1768
 * 
 * 
 * def XOR_AB(cpu): # AB XOR E             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.E
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1781
 * 
 * 
 * def XOR_AC(cpu): # AC XOR H             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ (cpu.HL >> 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1782
 * 
 * def XOR_AC(cpu): # AC XOR H
 *     t = cpu.A ^ (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A ^ (__pyx_v_cpu->HL >> 8));

  /* "pyboy/core/opcodes.py":1783
 * def XOR_AC(cpu): # AC XOR H
 *     t = cpu.A ^ (cpu.HL >> 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1784
 *     t = cpu.A ^ (cpu.HL >> 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1785
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1786
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1787
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1788
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1789
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1790
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1791
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1781
 * 
 * 
 * def XOR_AC(cpu): # AC XOR H             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ (cpu.HL >> 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1794
 * 
 * 
 * def XOR_AD(cpu): # AD XOR L             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ (cpu.HL & 0xFF)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1795
 * 
 * def XOR_AD(cpu): # AD XOR L
 *     t = cpu.A ^ (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A ^ (__pyx_v_cpu->HL & 0xFF));

  /* "pyboy/core/opcodes.py":1796
 * def XOR_AD(cpu): # AD XOR L
 *     t = cpu.A ^ (cpu.HL & 0xFF)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1797
 *     t = cpu.A ^ (cpu.HL & 0xFF)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1798
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1799
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1800
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1801
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1802
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1803
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1804
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1794
 * 
 * 
 * def XOR_AD(cpu): # AD XOR L             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ (cpu.HL & 0xFF)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1807
 * 
 * 
 * def XOR_AE(cpu): # AE XOR (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1808
 * 
 * def XOR_AE(cpu): # AE XOR (HL)
 *     t = cpu.A ^ cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A ^ ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL));

  /* "pyboy/core/opcodes.py":1809
 * def XOR_AE(cpu): # AE XOR (HL)
 *     t = cpu.A ^ cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1810
 *     t = cpu.A ^ cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1811
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1812
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1813
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1814
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1815
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1816
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1817
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1807
 * 
 * 
 * def XOR_AE(cpu): # AE XOR (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1820
 * 
 * 
 * def XOR_AF(cpu): # AF XOR A             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.A
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_AF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1821
 * 
 * def XOR_AF(cpu): # AF XOR A
 *     t = cpu.A ^ cpu.A             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A ^ __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":1822
 * def XOR_AF(cpu): # AF XOR A
 *     t = cpu.A ^ cpu.A
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1823
 *     t = cpu.A ^ cpu.A
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1824
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1825
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1826
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1827
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1828
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1829
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1830
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1820
 * 
 * 
 * def XOR_AF(cpu): # AF XOR A             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ cpu.A
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1833
 * 
 * 
 * def OR_B0(cpu): # B0 OR B             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.B
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1834
 * 
 * def OR_B0(cpu): # B0 OR B
 *     t = cpu.A | cpu.B             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A | __pyx_v_cpu->B);

  /* "pyboy/core/opcodes.py":1835
 * def OR_B0(cpu): # B0 OR B
 *     t = cpu.A | cpu.B
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1836
 *     t = cpu.A | cpu.B
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1837
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1838
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1839
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1840
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1841
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1842
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1843
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1833
 * 
 * 
 * def OR_B0(cpu): # B0 OR B             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.B
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1846
 * 
 * 
 * def OR_B1(cpu): # B1 OR C             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.C
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1847
 * 
 * def OR_B1(cpu): # B1 OR C
 *     t = cpu.A | cpu.C             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A | __pyx_v_cpu->C);

  /* "pyboy/core/opcodes.py":1848
 * def OR_B1(cpu): # B1 OR C
 *     t = cpu.A | cpu.C
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1849
 *     t = cpu.A | cpu.C
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1850
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1851
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1852
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1853
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1854
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1855
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1856
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1846
 * 
 * 
 * def OR_B1(cpu): # B1 OR C             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.C
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1859
 * 
 * 
 * def OR_B2(cpu): # B2 OR D             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.D
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1860
 * 
 * def OR_B2(cpu): # B2 OR D
 *     t = cpu.A | cpu.D             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A | __pyx_v_cpu->D);

  /* "pyboy/core/opcodes.py":1861
 * def OR_B2(cpu): # B2 OR D
 *     t = cpu.A | cpu.D
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1862
 *     t = cpu.A | cpu.D
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1863
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1864
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1865
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1866
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1867
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1868
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1869
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1859
 * 
 * 
 * def OR_B2(cpu): # B2 OR D             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.D
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1872
 * 
 * 
 * def OR_B3(cpu): # B3 OR E             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.E
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1873
 * 
 * def OR_B3(cpu): # B3 OR E
 *     t = cpu.A | cpu.E             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A | __pyx_v_cpu->E);

  /* "pyboy/core/opcodes.py":1874
 * def OR_B3(cpu): # B3 OR E
 *     t = cpu.A | cpu.E
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1875
 *     t = cpu.A | cpu.E
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1876
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1877
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1878
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1879
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1880
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1881
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1882
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1872
 * 
 * 
 * def OR_B3(cpu): # B3 OR E             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.E
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1885
 * 
 * 
 * def OR_B4(cpu): # B4 OR H             # <<<<<<<<<<<<<<
 *     t = cpu.A | (cpu.HL >> 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1886
 * 
 * def OR_B4(cpu): # B4 OR H
 *     t = cpu.A | (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A | (__pyx_v_cpu->HL >> 8));

  /* "pyboy/core/opcodes.py":1887
 * def OR_B4(cpu): # B4 OR H
 *     t = cpu.A | (cpu.HL >> 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1888
 *     t = cpu.A | (cpu.HL >> 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1889
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1890
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1891
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1892
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1893
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1894
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1895
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1885
 * 
 * 
 * def OR_B4(cpu): # B4 OR H             # <<<<<<<<<<<<<<
 *     t = cpu.A | (cpu.HL >> 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1898
 * 
 * 
 * def OR_B5(cpu): # B5 OR L             # <<<<<<<<<<<<<<
 *     t = cpu.A | (cpu.HL & 0xFF)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1899
 * 
 * def OR_B5(cpu): # B5 OR L
 *     t = cpu.A | (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A | (__pyx_v_cpu->HL & 0xFF));

  /* "pyboy/core/opcodes.py":1900
 * def OR_B5(cpu): # B5 OR L
 *     t = cpu.A | (cpu.HL & 0xFF)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1901
 *     t = cpu.A | (cpu.HL & 0xFF)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1902
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1903
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1904
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1905
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1906
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1907
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1908
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1898
 * 
 * 
 * def OR_B5(cpu): # B5 OR L             # <<<<<<<<<<<<<<
 *     t = cpu.A | (cpu.HL & 0xFF)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1911
 * 
 * 
 * def OR_B6(cpu): # B6 OR (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1912
 * 
 * def OR_B6(cpu): # B6 OR (HL)
 *     t = cpu.A | cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A | ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL));

  /* "pyboy/core/opcodes.py":1913
 * def OR_B6(cpu): # B6 OR (HL)
 *     t = cpu.A | cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1914
 *     t = cpu.A | cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1915
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1916
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1917
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1918
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1919
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1920
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1921
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1911
 * 
 * 
 * def OR_B6(cpu): # B6 OR (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.mb.getitem(cpu.HL)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1924
 * 
 * 
 * def OR_B7(cpu): # B7 OR A             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.A
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_B7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1925
 * 
 * def OR_B7(cpu): # B7 OR A
 *     t = cpu.A | cpu.A             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A | __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":1926
 * def OR_B7(cpu): # B7 OR A
 *     t = cpu.A | cpu.A
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":1927
 *     t = cpu.A | cpu.A
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1928
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1929
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1930
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 1
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1931
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":1932
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1933
 *     cpu.A = t
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1934
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1924
 * 
 * 
 * def OR_B7(cpu): # B7 OR A             # <<<<<<<<<<<<<<
 *     t = cpu.A | cpu.A
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1937
 * 
 * 
 * def CP_B8(cpu): # B8 CP B             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.B
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_B8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1938
 * 
 * def CP_B8(cpu): # B8 CP B
 *     t = cpu.A - cpu.B             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->B);

  /* "pyboy/core/opcodes.py":1939
 * def CP_B8(cpu): # B8 CP B
 *     t = cpu.A - cpu.B
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1940
 *     t = cpu.A - cpu.B
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1941
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->B & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1942
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1943
 *     flag += (((cpu.A & 0xF) - (cpu.B & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1944
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1945
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1946
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1947
 *     t &= 0xFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1948
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1937
 * 
 * 
 * def CP_B8(cpu): # B8 CP B             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.B
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1951
 * 
 * 
 * def CP_B9(cpu): # B9 CP C             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.C
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_B9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1952
 * 
 * def CP_B9(cpu): # B9 CP C
 *     t = cpu.A - cpu.C             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->C);

  /* "pyboy/core/opcodes.py":1953
 * def CP_B9(cpu): # B9 CP C
 *     t = cpu.A - cpu.C
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1954
 *     t = cpu.A - cpu.C
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1955
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->C & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1956
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1957
 *     flag += (((cpu.A & 0xF) - (cpu.C & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1958
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1959
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1960
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1961
 *     t &= 0xFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1962
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1951
 * 
 * 
 * def CP_B9(cpu): # B9 CP C             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.C
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1965
 * 
 * 
 * def CP_BA(cpu): # BA CP D             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.D
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1966
 * 
 * def CP_BA(cpu): # BA CP D
 *     t = cpu.A - cpu.D             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->D);

  /* "pyboy/core/opcodes.py":1967
 * def CP_BA(cpu): # BA CP D
 *     t = cpu.A - cpu.D
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1968
 *     t = cpu.A - cpu.D
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1969
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->D & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1970
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1971
 *     flag += (((cpu.A & 0xF) - (cpu.D & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1972
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1973
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1974
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1975
 *     t &= 0xFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1976
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1965
 * 
 * 
 * def CP_BA(cpu): # BA CP D             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.D
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1979
 * 
 * 
 * def CP_BB(cpu): # BB CP E             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.E
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1980
 * 
 * def CP_BB(cpu): # BB CP E
 *     t = cpu.A - cpu.E             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->E);

  /* "pyboy/core/opcodes.py":1981
 * def CP_BB(cpu): # BB CP E
 *     t = cpu.A - cpu.E
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1982
 *     t = cpu.A - cpu.E
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1983
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->E & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1984
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1985
 *     flag += (((cpu.A & 0xF) - (cpu.E & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":1986
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":1987
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":1988
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":1989
 *     t &= 0xFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":1990
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1979
 * 
 * 
 * def CP_BB(cpu): # BB CP E             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.E
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":1993
 * 
 * 
 * def CP_BC(cpu): # BC CP H             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL >> 8)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":1994
 * 
 * def CP_BC(cpu): # BC CP H
 *     t = cpu.A - (cpu.HL >> 8)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - (__pyx_v_cpu->HL >> 8));

  /* "pyboy/core/opcodes.py":1995
 * def CP_BC(cpu): # BC CP H
 *     t = cpu.A - (cpu.HL >> 8)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":1996
 *     t = cpu.A - (cpu.HL >> 8)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":1997
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - ((__pyx_v_cpu->HL >> 8) & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":1998
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":1999
 *     flag += (((cpu.A & 0xF) - ((cpu.HL >> 8) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2000
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2001
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2002
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2003
 *     t &= 0xFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2004
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":1993
 * 
 * 
 * def CP_BC(cpu): # BC CP H             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL >> 8)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2007
 * 
 * 
 * def CP_BD(cpu): # BD CP L             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL & 0xFF)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2008
 * 
 * def CP_BD(cpu): # BD CP L
 *     t = cpu.A - (cpu.HL & 0xFF)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - (__pyx_v_cpu->HL & 0xFF));

  /* "pyboy/core/opcodes.py":2009
 * def CP_BD(cpu): # BD CP L
 *     t = cpu.A - (cpu.HL & 0xFF)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":2010
 *     t = cpu.A - (cpu.HL & 0xFF)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2011
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - ((__pyx_v_cpu->HL & 0xFF) & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2012
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2013
 *     flag += (((cpu.A & 0xF) - ((cpu.HL & 0xFF) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2014
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2015
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2016
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2017
 *     t &= 0xFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2018
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2007
 * 
 * 
 * def CP_BD(cpu): # BD CP L             # <<<<<<<<<<<<<<
 *     t = cpu.A - (cpu.HL & 0xFF)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2021
 * 
 * 
 * def CP_BE(cpu): # BE CP (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2022
 * 
 * def CP_BE(cpu): # BE CP (HL)
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL));

  /* "pyboy/core/opcodes.py":2023
 * def CP_BE(cpu): # BE CP (HL)
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":2024
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2025
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2026
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2027
 *     flag += (((cpu.A & 0xF) - (cpu.mb.getitem(cpu.HL) & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2028
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2029
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2030
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2031
 *     t &= 0xFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2032
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2021
 * 
 * 
 * def CP_BE(cpu): # BE CP (HL)             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.mb.getitem(cpu.HL)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2035
 * 
 * 
 * def CP_BF(cpu): # BF CP A             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.A
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_BF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2036
 * 
 * def CP_BF(cpu): # BF CP A
 *     t = cpu.A - cpu.A             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":2037
 * def CP_BF(cpu): # BF CP A
 *     t = cpu.A - cpu.A
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":2038
 *     t = cpu.A - cpu.A
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2039
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_cpu->A & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2040
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2041
 *     flag += (((cpu.A & 0xF) - (cpu.A & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2042
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2043
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2044
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2045
 *     t &= 0xFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2046
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2035
 * 
 * 
 * def CP_BF(cpu): # BF CP A             # <<<<<<<<<<<<<<
 *     t = cpu.A - cpu.A
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2049
 * 
 * 
 * def RET_C0(cpu): # C0 RET NZ             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_C0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2050
 * 
 * def RET_C0(cpu): # C0 RET NZ
 *     if ((cpu.F & (1 << FLAGZ)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGZ)) == 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2051
 * def RET_C0(cpu): # C0 RET NZ
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High             # <<<<<<<<<<<<<<
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2
 */
    __pyx_v_cpu->PC = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF)) << 8);

    /* "pyboy/core/opcodes.py":2052
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low             # <<<<<<<<<<<<<<
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC | ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP));

    /* "pyboy/core/opcodes.py":2053
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2             # <<<<<<<<<<<<<<
 *         cpu.SP &= 0xFFFF
 *         return 20
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

    /* "pyboy/core/opcodes.py":2054
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 20
 *     else:
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

    /* "pyboy/core/opcodes.py":2055
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF
 *         return 20             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC += 1
 */
    __pyx_r = 20;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2050
 * 
 * def RET_C0(cpu): # C0 RET NZ
 *     if ((cpu.F & (1 << FLAGZ)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */
  }

  /* "pyboy/core/opcodes.py":2057
 *         return 20
 *     else:
 *         cpu.PC += 1             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 8
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

    /* "pyboy/core/opcodes.py":2058
 *     else:
 *         cpu.PC += 1
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 8
 * 
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":2059
 *         cpu.PC += 1
 *         cpu.PC &= 0xFFFF
 *         return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 8;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2049
 * 
 * 
 * def RET_C0(cpu): # C0 RET NZ             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2062
 * 
 * 
 * def POP_C1(cpu): # C1 POP BC             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.C = cpu.mb.getitem(cpu.SP) # Low
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_POP_C1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2063
 * 
 * def POP_C1(cpu): # C1 POP BC
 *     cpu.B = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High             # <<<<<<<<<<<<<<
 *     cpu.C = cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2
 */
  __pyx_v_cpu->B = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF));

  /* "pyboy/core/opcodes.py":2064
 * def POP_C1(cpu): # C1 POP BC
 *     cpu.B = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.C = cpu.mb.getitem(cpu.SP) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 */
  __pyx_v_cpu->C = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP);

  /* "pyboy/core/opcodes.py":2065
 *     cpu.B = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.C = cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

  /* "pyboy/core/opcodes.py":2066
 *     cpu.C = cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2067
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2068
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2069
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2062
 * 
 * 
 * def POP_C1(cpu): # C1 POP BC             # <<<<<<<<<<<<<<
 *     cpu.B = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.C = cpu.mb.getitem(cpu.SP) # Low
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2072
 * 
 * 
 * def JP_C2(cpu, v): # C2 JP NZ,a16             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC = v
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_C2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2073
 * 
 * def JP_C2(cpu, v): # C2 JP NZ,a16
 *     if ((cpu.F & (1 << FLAGZ)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 16
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGZ)) == 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2074
 * def JP_C2(cpu, v): # C2 JP NZ,a16
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC = v             # <<<<<<<<<<<<<<
 *         return 16
 *     else:
 */
    __pyx_v_cpu->PC = __pyx_v_v;

    /* "pyboy/core/opcodes.py":2075
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC = v
 *         return 16             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC += 3
 */
    __pyx_r = 16;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2073
 * 
 * def JP_C2(cpu, v): # C2 JP NZ,a16
 *     if ((cpu.F & (1 << FLAGZ)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 16
 */
  }

  /* "pyboy/core/opcodes.py":2077
 *         return 16
 *     else:
 *         cpu.PC += 3             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 12
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

    /* "pyboy/core/opcodes.py":2078
 *     else:
 *         cpu.PC += 3
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 12
 * 
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":2079
 *         cpu.PC += 3
 *         cpu.PC &= 0xFFFF
 *         return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 12;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2072
 * 
 * 
 * def JP_C2(cpu, v): # C2 JP NZ,a16             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.PC = v
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2082
 * 
 * 
 * def JP_C3(cpu, v): # C3 JP a16             # <<<<<<<<<<<<<<
 *     cpu.PC = v
 *     return 16
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_C3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2083
 * 
 * def JP_C3(cpu, v): # C3 JP a16
 *     cpu.PC = v             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = __pyx_v_v;

  /* "pyboy/core/opcodes.py":2084
 * def JP_C3(cpu, v): # C3 JP a16
 *     cpu.PC = v
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2082
 * 
 * 
 * def JP_C3(cpu, v): # C3 JP a16             # <<<<<<<<<<<<<<
 *     cpu.PC = v
 *     return 16
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2087
 * 
 * 
 * def CALL_C4(cpu, v): # C4 CALL NZ,a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_C4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2088
 * 
 * def CALL_C4(cpu, v): # C4 CALL NZ,a16
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":2089
 * def CALL_C4(cpu, v): # C4 CALL NZ,a16
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2090
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGZ)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGZ)) == 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2091
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

    /* "pyboy/core/opcodes.py":2092
 *     if ((cpu.F & (1 << FLAGZ)) == 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

    /* "pyboy/core/opcodes.py":2093
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2             # <<<<<<<<<<<<<<
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

    /* "pyboy/core/opcodes.py":2094
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 24
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

    /* "pyboy/core/opcodes.py":2095
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v             # <<<<<<<<<<<<<<
 *         return 24
 *     else:
 */
    __pyx_v_cpu->PC = __pyx_v_v;

    /* "pyboy/core/opcodes.py":2096
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v
 *         return 24             # <<<<<<<<<<<<<<
 *     else:
 *         return 12
 */
    __pyx_r = 24;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2090
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGZ)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  }

  /* "pyboy/core/opcodes.py":2098
 *         return 24
 *     else:
 *         return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = 12;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2087
 * 
 * 
 * def CALL_C4(cpu, v): # C4 CALL NZ,a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2101
 * 
 * 
 * def PUSH_C5(cpu): # C5 PUSH BC             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.B) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.C) # Low
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_PUSH_C5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2102
 * 
 * def PUSH_C5(cpu): # C5 PUSH BC
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.B) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.C) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), __pyx_v_cpu->B);

  /* "pyboy/core/opcodes.py":2103
 * def PUSH_C5(cpu): # C5 PUSH BC
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.B) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.C) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), __pyx_v_cpu->C);

  /* "pyboy/core/opcodes.py":2104
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.B) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.C) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2105
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.C) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2106
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2107
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2108
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2101
 * 
 * 
 * def PUSH_C5(cpu): # C5 PUSH BC             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.B) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.C) # Low
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2111
 * 
 * 
 * def ADD_C6(cpu, v): # C6 ADD A,d8             # <<<<<<<<<<<<<<
 *     t = cpu.A + v
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_C6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2112
 * 
 * def ADD_C6(cpu, v): # C6 ADD A,d8
 *     t = cpu.A + v             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A + __pyx_v_v);

  /* "pyboy/core/opcodes.py":2113
 * def ADD_C6(cpu, v): # C6 ADD A,d8
 *     t = cpu.A + v
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2114
 *     t = cpu.A + v
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2115
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (v & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) + (__pyx_v_v & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2116
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2117
 *     flag += (((cpu.A & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2118
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2119
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2120
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2121
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2122
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2123
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2111
 * 
 * 
 * def ADD_C6(cpu, v): # C6 ADD A,d8             # <<<<<<<<<<<<<<
 *     t = cpu.A + v
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2126
 * 
 * 
 * def RST_C7(cpu): # C7 RST 00H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_C7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2127
 * 
 * def RST_C7(cpu): # C7 RST 00H
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2128
 * def RST_C7(cpu): # C7 RST 00H
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2129
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

  /* "pyboy/core/opcodes.py":2130
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

  /* "pyboy/core/opcodes.py":2131
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 0
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2132
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC = 0
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2133
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 0             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = 0;

  /* "pyboy/core/opcodes.py":2134
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 0
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2126
 * 
 * 
 * def RST_C7(cpu): # C7 RST 00H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2137
 * 
 * 
 * def RET_C8(cpu): # C8 RET Z             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_C8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2138
 * 
 * def RET_C8(cpu): # C8 RET Z
 *     if ((cpu.F & (1 << FLAGZ)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGZ)) != 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2139
 * def RET_C8(cpu): # C8 RET Z
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High             # <<<<<<<<<<<<<<
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2
 */
    __pyx_v_cpu->PC = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF)) << 8);

    /* "pyboy/core/opcodes.py":2140
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low             # <<<<<<<<<<<<<<
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC | ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP));

    /* "pyboy/core/opcodes.py":2141
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2             # <<<<<<<<<<<<<<
 *         cpu.SP &= 0xFFFF
 *         return 20
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

    /* "pyboy/core/opcodes.py":2142
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 20
 *     else:
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

    /* "pyboy/core/opcodes.py":2143
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF
 *         return 20             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC += 1
 */
    __pyx_r = 20;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2138
 * 
 * def RET_C8(cpu): # C8 RET Z
 *     if ((cpu.F & (1 << FLAGZ)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */
  }

  /* "pyboy/core/opcodes.py":2145
 *         return 20
 *     else:
 *         cpu.PC += 1             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 8
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

    /* "pyboy/core/opcodes.py":2146
 *     else:
 *         cpu.PC += 1
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 8
 * 
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":2147
 *         cpu.PC += 1
 *         cpu.PC &= 0xFFFF
 *         return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 8;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2137
 * 
 * 
 * def RET_C8(cpu): # C8 RET Z             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2150
 * 
 * 
 * def RET_C9(cpu): # C9 RET             # <<<<<<<<<<<<<<
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_C9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2151
 * 
 * def RET_C9(cpu): # C9 RET
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High             # <<<<<<<<<<<<<<
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2
 */
  __pyx_v_cpu->PC = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF)) << 8);

  /* "pyboy/core/opcodes.py":2152
 * def RET_C9(cpu): # C9 RET
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC | ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP));

  /* "pyboy/core/opcodes.py":2153
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

  /* "pyboy/core/opcodes.py":2154
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2155
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2150
 * 
 * 
 * def RET_C9(cpu): # C9 RET             # <<<<<<<<<<<<<<
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2158
 * 
 * 
 * def JP_CA(cpu, v): # CA JP Z,a16             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC = v
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_CA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2159
 * 
 * def JP_CA(cpu, v): # CA JP Z,a16
 *     if ((cpu.F & (1 << FLAGZ)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 16
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGZ)) != 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2160
 * def JP_CA(cpu, v): # CA JP Z,a16
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC = v             # <<<<<<<<<<<<<<
 *         return 16
 *     else:
 */
    __pyx_v_cpu->PC = __pyx_v_v;

    /* "pyboy/core/opcodes.py":2161
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC = v
 *         return 16             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC += 3
 */
    __pyx_r = 16;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2159
 * 
 * def JP_CA(cpu, v): # CA JP Z,a16
 *     if ((cpu.F & (1 << FLAGZ)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 16
 */
  }

  /* "pyboy/core/opcodes.py":2163
 *         return 16
 *     else:
 *         cpu.PC += 3             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 12
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

    /* "pyboy/core/opcodes.py":2164
 *     else:
 *         cpu.PC += 3
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 12
 * 
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":2165
 *         cpu.PC += 3
 *         cpu.PC &= 0xFFFF
 *         return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 12;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2158
 * 
 * 
 * def JP_CA(cpu, v): # CA JP Z,a16             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.PC = v
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2168
 * 
 * 
 * def PREFIX_CB(cpu): # CB PREFIX CB             # <<<<<<<<<<<<<<
 *     logger.critical("CB cannot be called!")
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_PREFIX_CB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2169
 * 
 * def PREFIX_CB(cpu): # CB PREFIX CB
 *     logger.critical("CB cannot be called!")             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_7logging_7logging_Logger *)__pyx_v_5pyboy_4core_7opcodes_logger->__pyx_vtab)->critical(__pyx_v_5pyboy_4core_7opcodes_logger, __pyx_kp_u_CB_cannot_be_called, 0);

  /* "pyboy/core/opcodes.py":2170
 * def PREFIX_CB(cpu): # CB PREFIX CB
 *     logger.critical("CB cannot be called!")
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2171
 *     logger.critical("CB cannot be called!")
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2172
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2168
 * 
 * 
 * def PREFIX_CB(cpu): # CB PREFIX CB             # <<<<<<<<<<<<<<
 *     logger.critical("CB cannot be called!")
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2175
 * 
 * 
 * def CALL_CC(cpu, v): # CC CALL Z,a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_CC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2176
 * 
 * def CALL_CC(cpu, v): # CC CALL Z,a16
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":2177
 * def CALL_CC(cpu, v): # CC CALL Z,a16
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2178
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGZ)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGZ)) != 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2179
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

    /* "pyboy/core/opcodes.py":2180
 *     if ((cpu.F & (1 << FLAGZ)) != 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

    /* "pyboy/core/opcodes.py":2181
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2             # <<<<<<<<<<<<<<
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

    /* "pyboy/core/opcodes.py":2182
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 24
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

    /* "pyboy/core/opcodes.py":2183
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v             # <<<<<<<<<<<<<<
 *         return 24
 *     else:
 */
    __pyx_v_cpu->PC = __pyx_v_v;

    /* "pyboy/core/opcodes.py":2184
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v
 *         return 24             # <<<<<<<<<<<<<<
 *     else:
 *         return 12
 */
    __pyx_r = 24;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2178
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGZ)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  }

  /* "pyboy/core/opcodes.py":2186
 *         return 24
 *     else:
 *         return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = 12;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2175
 * 
 * 
 * def CALL_CC(cpu, v): # CC CALL Z,a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2189
 * 
 * 
 * def CALL_CD(cpu, v): # CD CALL a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_CD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2190
 * 
 * def CALL_CD(cpu, v): # CD CALL a16
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":2191
 * def CALL_CD(cpu, v): # CD CALL a16
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2192
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

  /* "pyboy/core/opcodes.py":2193
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

  /* "pyboy/core/opcodes.py":2194
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = v
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2195
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC = v
 *     return 24
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2196
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = v             # <<<<<<<<<<<<<<
 *     return 24
 * 
 */
  __pyx_v_cpu->PC = __pyx_v_v;

  /* "pyboy/core/opcodes.py":2197
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = v
 *     return 24             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 24;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2189
 * 
 * 
 * def CALL_CD(cpu, v): # CD CALL a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2200
 * 
 * 
 * def ADC_CE(cpu, v): # CE ADC A,d8             # <<<<<<<<<<<<<<
 *     t = cpu.A + v + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADC_CE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2201
 * 
 * def ADC_CE(cpu, v): # CE ADC A,d8
 *     t = cpu.A + v + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A + __pyx_v_v) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2202
 * def ADC_CE(cpu, v): # CE ADC A,d8
 *     t = cpu.A + v + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (v & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2203
 *     t = cpu.A + v + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) + (v & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2204
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (v & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) + (__pyx_v_v & 0xF)) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2205
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) + (v & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2206
 *     flag += (((cpu.A & 0xF) + (v & 0xF) + ((cpu.F & (1 << FLAGC)) != 0)) > 0xF) << FLAGH
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2207
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2208
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2209
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2210
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2211
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2212
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2200
 * 
 * 
 * def ADC_CE(cpu, v): # CE ADC A,d8             # <<<<<<<<<<<<<<
 *     t = cpu.A + v + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2215
 * 
 * 
 * def RST_CF(cpu): # CF RST 08H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_CF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2216
 * 
 * def RST_CF(cpu): # CF RST 08H
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2217
 * def RST_CF(cpu): # CF RST 08H
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2218
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

  /* "pyboy/core/opcodes.py":2219
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

  /* "pyboy/core/opcodes.py":2220
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 8
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2221
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC = 8
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2222
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 8             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = 8;

  /* "pyboy/core/opcodes.py":2223
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 8
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2215
 * 
 * 
 * def RST_CF(cpu): # CF RST 08H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2226
 * 
 * 
 * def RET_D0(cpu): # D0 RET NC             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_D0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2227
 * 
 * def RET_D0(cpu): # D0 RET NC
 *     if ((cpu.F & (1 << FLAGC)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) == 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2228
 * def RET_D0(cpu): # D0 RET NC
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High             # <<<<<<<<<<<<<<
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2
 */
    __pyx_v_cpu->PC = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF)) << 8);

    /* "pyboy/core/opcodes.py":2229
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low             # <<<<<<<<<<<<<<
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC | ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP));

    /* "pyboy/core/opcodes.py":2230
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2             # <<<<<<<<<<<<<<
 *         cpu.SP &= 0xFFFF
 *         return 20
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

    /* "pyboy/core/opcodes.py":2231
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 20
 *     else:
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

    /* "pyboy/core/opcodes.py":2232
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF
 *         return 20             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC += 1
 */
    __pyx_r = 20;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2227
 * 
 * def RET_D0(cpu): # D0 RET NC
 *     if ((cpu.F & (1 << FLAGC)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */
  }

  /* "pyboy/core/opcodes.py":2234
 *         return 20
 *     else:
 *         cpu.PC += 1             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 8
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

    /* "pyboy/core/opcodes.py":2235
 *     else:
 *         cpu.PC += 1
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 8
 * 
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":2236
 *         cpu.PC += 1
 *         cpu.PC &= 0xFFFF
 *         return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 8;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2226
 * 
 * 
 * def RET_D0(cpu): # D0 RET NC             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2239
 * 
 * 
 * def POP_D1(cpu): # D1 POP DE             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.E = cpu.mb.getitem(cpu.SP) # Low
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_POP_D1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2240
 * 
 * def POP_D1(cpu): # D1 POP DE
 *     cpu.D = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High             # <<<<<<<<<<<<<<
 *     cpu.E = cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2
 */
  __pyx_v_cpu->D = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF));

  /* "pyboy/core/opcodes.py":2241
 * def POP_D1(cpu): # D1 POP DE
 *     cpu.D = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.E = cpu.mb.getitem(cpu.SP) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 */
  __pyx_v_cpu->E = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP);

  /* "pyboy/core/opcodes.py":2242
 *     cpu.D = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.E = cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

  /* "pyboy/core/opcodes.py":2243
 *     cpu.E = cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2244
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2245
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2246
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2239
 * 
 * 
 * def POP_D1(cpu): # D1 POP DE             # <<<<<<<<<<<<<<
 *     cpu.D = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.E = cpu.mb.getitem(cpu.SP) # Low
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2249
 * 
 * 
 * def JP_D2(cpu, v): # D2 JP NC,a16             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC = v
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_D2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2250
 * 
 * def JP_D2(cpu, v): # D2 JP NC,a16
 *     if ((cpu.F & (1 << FLAGC)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 16
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) == 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2251
 * def JP_D2(cpu, v): # D2 JP NC,a16
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC = v             # <<<<<<<<<<<<<<
 *         return 16
 *     else:
 */
    __pyx_v_cpu->PC = __pyx_v_v;

    /* "pyboy/core/opcodes.py":2252
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC = v
 *         return 16             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC += 3
 */
    __pyx_r = 16;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2250
 * 
 * def JP_D2(cpu, v): # D2 JP NC,a16
 *     if ((cpu.F & (1 << FLAGC)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 16
 */
  }

  /* "pyboy/core/opcodes.py":2254
 *         return 16
 *     else:
 *         cpu.PC += 3             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 12
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

    /* "pyboy/core/opcodes.py":2255
 *     else:
 *         cpu.PC += 3
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 12
 * 
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":2256
 *         cpu.PC += 3
 *         cpu.PC &= 0xFFFF
 *         return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 12;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2249
 * 
 * 
 * def JP_D2(cpu, v): # D2 JP NC,a16             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.PC = v
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2259
 * 
 * 
 * def CALL_D4(cpu, v): # D4 CALL NC,a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_D4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2260
 * 
 * def CALL_D4(cpu, v): # D4 CALL NC,a16
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":2261
 * def CALL_D4(cpu, v): # D4 CALL NC,a16
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2262
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGC)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) == 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2263
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

    /* "pyboy/core/opcodes.py":2264
 *     if ((cpu.F & (1 << FLAGC)) == 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

    /* "pyboy/core/opcodes.py":2265
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2             # <<<<<<<<<<<<<<
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

    /* "pyboy/core/opcodes.py":2266
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 24
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

    /* "pyboy/core/opcodes.py":2267
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v             # <<<<<<<<<<<<<<
 *         return 24
 *     else:
 */
    __pyx_v_cpu->PC = __pyx_v_v;

    /* "pyboy/core/opcodes.py":2268
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v
 *         return 24             # <<<<<<<<<<<<<<
 *     else:
 *         return 12
 */
    __pyx_r = 24;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2262
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGC)) == 0):             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  }

  /* "pyboy/core/opcodes.py":2270
 *         return 24
 *     else:
 *         return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = 12;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2259
 * 
 * 
 * def CALL_D4(cpu, v): # D4 CALL NC,a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2273
 * 
 * 
 * def PUSH_D5(cpu): # D5 PUSH DE             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.D) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.E) # Low
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_PUSH_D5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2274
 * 
 * def PUSH_D5(cpu): # D5 PUSH DE
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.D) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.E) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), __pyx_v_cpu->D);

  /* "pyboy/core/opcodes.py":2275
 * def PUSH_D5(cpu): # D5 PUSH DE
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.D) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.E) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), __pyx_v_cpu->E);

  /* "pyboy/core/opcodes.py":2276
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.D) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.E) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2277
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.E) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2278
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2279
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2280
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2273
 * 
 * 
 * def PUSH_D5(cpu): # D5 PUSH DE             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.D) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.E) # Low
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2283
 * 
 * 
 * def SUB_D6(cpu, v): # D6 SUB d8             # <<<<<<<<<<<<<<
 *     t = cpu.A - v
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SUB_D6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2284
 * 
 * def SUB_D6(cpu, v): # D6 SUB d8
 *     t = cpu.A - v             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_v);

  /* "pyboy/core/opcodes.py":2285
 * def SUB_D6(cpu, v): # D6 SUB d8
 *     t = cpu.A - v
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":2286
 *     t = cpu.A - v
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2287
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_v & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2288
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2289
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2290
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2291
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2292
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2293
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2294
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2295
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2283
 * 
 * 
 * def SUB_D6(cpu, v): # D6 SUB d8             # <<<<<<<<<<<<<<
 *     t = cpu.A - v
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2298
 * 
 * 
 * def RST_D7(cpu): # D7 RST 10H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_D7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2299
 * 
 * def RST_D7(cpu): # D7 RST 10H
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2300
 * def RST_D7(cpu): # D7 RST 10H
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2301
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

  /* "pyboy/core/opcodes.py":2302
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

  /* "pyboy/core/opcodes.py":2303
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2304
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC = 16
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2305
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 16             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = 16;

  /* "pyboy/core/opcodes.py":2306
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 16
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2298
 * 
 * 
 * def RST_D7(cpu): # D7 RST 10H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2309
 * 
 * 
 * def RET_D8(cpu): # D8 RET C             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RET_D8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2310
 * 
 * def RET_D8(cpu): # D8 RET C
 *     if ((cpu.F & (1 << FLAGC)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2311
 * def RET_D8(cpu): # D8 RET C
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High             # <<<<<<<<<<<<<<
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2
 */
    __pyx_v_cpu->PC = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF)) << 8);

    /* "pyboy/core/opcodes.py":2312
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low             # <<<<<<<<<<<<<<
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC | ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP));

    /* "pyboy/core/opcodes.py":2313
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2             # <<<<<<<<<<<<<<
 *         cpu.SP &= 0xFFFF
 *         return 20
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

    /* "pyboy/core/opcodes.py":2314
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 20
 *     else:
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

    /* "pyboy/core/opcodes.py":2315
 *         cpu.SP += 2
 *         cpu.SP &= 0xFFFF
 *         return 20             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC += 1
 */
    __pyx_r = 20;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2310
 * 
 * def RET_D8(cpu): # D8 RET C
 *     if ((cpu.F & (1 << FLAGC)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *         cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */
  }

  /* "pyboy/core/opcodes.py":2317
 *         return 20
 *     else:
 *         cpu.PC += 1             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 8
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

    /* "pyboy/core/opcodes.py":2318
 *     else:
 *         cpu.PC += 1
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 8
 * 
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":2319
 *         cpu.PC += 1
 *         cpu.PC &= 0xFFFF
 *         return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 8;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2309
 * 
 * 
 * def RET_D8(cpu): # D8 RET C             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2322
 * 
 * 
 * def RETI_D9(cpu): # D9 RETI             # <<<<<<<<<<<<<<
 *     cpu.interrupt_master_enable = True
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RETI_D9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2323
 * 
 * def RETI_D9(cpu): # D9 RETI
 *     cpu.interrupt_master_enable = True             # <<<<<<<<<<<<<<
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 */
  __pyx_v_cpu->interrupt_master_enable = 1;

  /* "pyboy/core/opcodes.py":2324
 * def RETI_D9(cpu): # D9 RETI
 *     cpu.interrupt_master_enable = True
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High             # <<<<<<<<<<<<<<
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2
 */
  __pyx_v_cpu->PC = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF)) << 8);

  /* "pyboy/core/opcodes.py":2325
 *     cpu.interrupt_master_enable = True
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC | ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP));

  /* "pyboy/core/opcodes.py":2326
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

  /* "pyboy/core/opcodes.py":2327
 *     cpu.PC |= cpu.mb.getitem(cpu.SP) # Low
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2328
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2322
 * 
 * 
 * def RETI_D9(cpu): # D9 RETI             # <<<<<<<<<<<<<<
 *     cpu.interrupt_master_enable = True
 *     cpu.PC = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8 # High
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2331
 * 
 * 
 * def JP_DA(cpu, v): # DA JP C,a16             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC = v
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_DA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2332
 * 
 * def JP_DA(cpu, v): # DA JP C,a16
 *     if ((cpu.F & (1 << FLAGC)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 16
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2333
 * def JP_DA(cpu, v): # DA JP C,a16
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC = v             # <<<<<<<<<<<<<<
 *         return 16
 *     else:
 */
    __pyx_v_cpu->PC = __pyx_v_v;

    /* "pyboy/core/opcodes.py":2334
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC = v
 *         return 16             # <<<<<<<<<<<<<<
 *     else:
 *         cpu.PC += 3
 */
    __pyx_r = 16;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2332
 * 
 * def JP_DA(cpu, v): # DA JP C,a16
 *     if ((cpu.F & (1 << FLAGC)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 16
 */
  }

  /* "pyboy/core/opcodes.py":2336
 *         return 16
 *     else:
 *         cpu.PC += 3             # <<<<<<<<<<<<<<
 *         cpu.PC &= 0xFFFF
 *         return 12
 */
  /*else*/ {
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

    /* "pyboy/core/opcodes.py":2337
 *     else:
 *         cpu.PC += 3
 *         cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *         return 12
 * 
 */
    __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

    /* "pyboy/core/opcodes.py":2338
 *         cpu.PC += 3
 *         cpu.PC &= 0xFFFF
 *         return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = 12;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2331
 * 
 * 
 * def JP_DA(cpu, v): # DA JP C,a16             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.PC = v
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2341
 * 
 * 
 * def CALL_DC(cpu, v): # DC CALL C,a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CALL_DC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2342
 * 
 * def CALL_DC(cpu, v): # DC CALL C,a16
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":2343
 * def CALL_DC(cpu, v): # DC CALL C,a16
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2344
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGC)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_t_1 = ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0);
  if (__pyx_t_1) {

    /* "pyboy/core/opcodes.py":2345
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

    /* "pyboy/core/opcodes.py":2346
 *     if ((cpu.F & (1 << FLAGC)) != 0):
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF
 */
    ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

    /* "pyboy/core/opcodes.py":2347
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2             # <<<<<<<<<<<<<<
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

    /* "pyboy/core/opcodes.py":2348
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *         cpu.PC = v
 *         return 24
 */
    __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

    /* "pyboy/core/opcodes.py":2349
 *         cpu.SP -= 2
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v             # <<<<<<<<<<<<<<
 *         return 24
 *     else:
 */
    __pyx_v_cpu->PC = __pyx_v_v;

    /* "pyboy/core/opcodes.py":2350
 *         cpu.SP &= 0xFFFF
 *         cpu.PC = v
 *         return 24             # <<<<<<<<<<<<<<
 *     else:
 *         return 12
 */
    __pyx_r = 24;
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":2344
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     if ((cpu.F & (1 << FLAGC)) != 0):             # <<<<<<<<<<<<<<
 *         cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *         cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  }

  /* "pyboy/core/opcodes.py":2352
 *         return 24
 *     else:
 *         return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = 12;
    goto __pyx_L0;
  }

  /* "pyboy/core/opcodes.py":2341
 * 
 * 
 * def CALL_DC(cpu, v): # DC CALL C,a16             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2355
 * 
 * 
 * def SBC_DE(cpu, v): # DE SBC A,d8             # <<<<<<<<<<<<<<
 *     t = cpu.A - v - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SBC_DE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2356
 * 
 * def SBC_DE(cpu, v): # DE SBC A,d8
 *     t = cpu.A - v - ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A - __pyx_v_v) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2357
 * def SBC_DE(cpu, v): # DE SBC A,d8
 *     t = cpu.A - v - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (v & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":2358
 *     t = cpu.A - v - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (v & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2359
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (v & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((((__pyx_v_cpu->A & 0xF) - (__pyx_v_v & 0xF)) - ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2360
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (v & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2361
 *     flag += (((cpu.A & 0xF) - (v & 0xF) - ((cpu.F & (1 << FLAGC)) != 0)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2362
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2363
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2364
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2365
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2366
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2367
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2355
 * 
 * 
 * def SBC_DE(cpu, v): # DE SBC A,d8             # <<<<<<<<<<<<<<
 *     t = cpu.A - v - ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2370
 * 
 * 
 * def RST_DF(cpu): # DF RST 18H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_DF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2371
 * 
 * def RST_DF(cpu): # DF RST 18H
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2372
 * def RST_DF(cpu): # DF RST 18H
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2373
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

  /* "pyboy/core/opcodes.py":2374
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

  /* "pyboy/core/opcodes.py":2375
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 24
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2376
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC = 24
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2377
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 24             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = 24;

  /* "pyboy/core/opcodes.py":2378
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 24
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2370
 * 
 * 
 * def RST_DF(cpu): # DF RST 18H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2381
 * 
 * 
 * def LDH_E0(cpu, v): # E0 LDH (a8),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(v + 0xFF00, cpu.A)
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LDH_E0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2382
 * 
 * def LDH_E0(cpu, v): # E0 LDH (a8),A
 *     cpu.mb.setitem(v + 0xFF00, cpu.A)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, (__pyx_v_v + 0xFF00), __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":2383
 * def LDH_E0(cpu, v): # E0 LDH (a8),A
 *     cpu.mb.setitem(v + 0xFF00, cpu.A)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2384
 *     cpu.mb.setitem(v + 0xFF00, cpu.A)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2385
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2381
 * 
 * 
 * def LDH_E0(cpu, v): # E0 LDH (a8),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(v + 0xFF00, cpu.A)
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2388
 * 
 * 
 * def POP_E1(cpu): # E1 POP HL             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8) + cpu.mb.getitem(cpu.SP) # High
 *     cpu.SP += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_POP_E1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2389
 * 
 * def POP_E1(cpu): # E1 POP HL
 *     cpu.HL = (cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8) + cpu.mb.getitem(cpu.SP) # High             # <<<<<<<<<<<<<<
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF)) << 8) + ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP));

  /* "pyboy/core/opcodes.py":2390
 * def POP_E1(cpu): # E1 POP HL
 *     cpu.HL = (cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8) + cpu.mb.getitem(cpu.SP) # High
 *     cpu.SP += 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

  /* "pyboy/core/opcodes.py":2391
 *     cpu.HL = (cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8) + cpu.mb.getitem(cpu.SP) # High
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2392
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2393
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2394
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2388
 * 
 * 
 * def POP_E1(cpu): # E1 POP HL             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) << 8) + cpu.mb.getitem(cpu.SP) # High
 *     cpu.SP += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2397
 * 
 * 
 * def LD_E2(cpu): # E2 LD (C),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(0xFF00 + cpu.C, cpu.A)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_E2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2398
 * 
 * def LD_E2(cpu): # E2 LD (C),A
 *     cpu.mb.setitem(0xFF00 + cpu.C, cpu.A)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, (0xFF00 + __pyx_v_cpu->C), __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":2399
 * def LD_E2(cpu): # E2 LD (C),A
 *     cpu.mb.setitem(0xFF00 + cpu.C, cpu.A)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2400
 *     cpu.mb.setitem(0xFF00 + cpu.C, cpu.A)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2401
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2397
 * 
 * 
 * def LD_E2(cpu): # E2 LD (C),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(0xFF00 + cpu.C, cpu.A)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2404
 * 
 * 
 * def PUSH_E5(cpu): # E5 PUSH HL             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.HL >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.HL & 0xFF) # Low
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_PUSH_E5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2405
 * 
 * def PUSH_E5(cpu): # E5 PUSH HL
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.HL >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.HL & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->HL >> 8));

  /* "pyboy/core/opcodes.py":2406
 * def PUSH_E5(cpu): # E5 PUSH HL
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.HL >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.HL & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->HL & 0xFF));

  /* "pyboy/core/opcodes.py":2407
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.HL >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.HL & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2408
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.HL & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2409
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2410
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2411
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2404
 * 
 * 
 * def PUSH_E5(cpu): # E5 PUSH HL             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.HL >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.HL & 0xFF) # Low
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2414
 * 
 * 
 * def AND_E6(cpu, v): # E6 AND d8             # <<<<<<<<<<<<<<
 *     t = cpu.A & v
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_AND_E6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2415
 * 
 * def AND_E6(cpu, v): # E6 AND d8
 *     t = cpu.A & v             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & __pyx_v_v);

  /* "pyboy/core/opcodes.py":2416
 * def AND_E6(cpu, v): # E6 AND d8
 *     t = cpu.A & v
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":2417
 *     t = cpu.A & v
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2418
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2419
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2420
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2421
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2422
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2423
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2424
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2414
 * 
 * 
 * def AND_E6(cpu, v): # E6 AND d8             # <<<<<<<<<<<<<<
 *     t = cpu.A & v
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2427
 * 
 * 
 * def RST_E7(cpu): # E7 RST 20H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_E7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2428
 * 
 * def RST_E7(cpu): # E7 RST 20H
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2429
 * def RST_E7(cpu): # E7 RST 20H
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2430
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

  /* "pyboy/core/opcodes.py":2431
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

  /* "pyboy/core/opcodes.py":2432
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 32
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2433
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC = 32
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2434
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 32             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = 32;

  /* "pyboy/core/opcodes.py":2435
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 32
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2427
 * 
 * 
 * def RST_E7(cpu): # E7 RST 20H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2438
 * 
 * 
 * def ADD_E8(cpu, v): # E8 ADD SP,r8             # <<<<<<<<<<<<<<
 *     t = cpu.SP + ((v ^ 0x80) - 0x80)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_ADD_E8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2439
 * 
 * def ADD_E8(cpu, v): # E8 ADD SP,r8
 *     t = cpu.SP + ((v ^ 0x80) - 0x80)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 */
  __pyx_v_t = (__pyx_v_cpu->SP + ((__pyx_v_v ^ 0x80) - 0x80));

  /* "pyboy/core/opcodes.py":2440
 * def ADD_E8(cpu, v): # E8 ADD SP,r8
 *     t = cpu.SP + ((v ^ 0x80) - 0x80)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2441
 *     t = cpu.SP + ((v ^ 0x80) - 0x80)
 *     flag = 0b00000000
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->SP & 0xF) + (__pyx_v_v & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2442
 *     flag = 0b00000000
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->SP & 0xFF) + (__pyx_v_v & 0xFF)) > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2443
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2444
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFFFF
 *     cpu.SP = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2445
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.SP = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFFFF);

  /* "pyboy/core/opcodes.py":2446
 *     cpu.F |= flag
 *     t &= 0xFFFF
 *     cpu.SP = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2447
 *     t &= 0xFFFF
 *     cpu.SP = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2448
 *     cpu.SP = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2449
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2438
 * 
 * 
 * def ADD_E8(cpu, v): # E8 ADD SP,r8             # <<<<<<<<<<<<<<
 *     t = cpu.SP + ((v ^ 0x80) - 0x80)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2452
 * 
 * 
 * def JP_E9(cpu): # E9 JP (HL)             # <<<<<<<<<<<<<<
 *     cpu.PC = cpu.HL
 *     return 4
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_JP_E9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2453
 * 
 * def JP_E9(cpu): # E9 JP (HL)
 *     cpu.PC = cpu.HL             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_t_1 = __pyx_v_cpu->HL;
  __pyx_v_cpu->PC = __pyx_t_1;

  /* "pyboy/core/opcodes.py":2454
 * def JP_E9(cpu): # E9 JP (HL)
 *     cpu.PC = cpu.HL
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2452
 * 
 * 
 * def JP_E9(cpu): # E9 JP (HL)             # <<<<<<<<<<<<<<
 *     cpu.PC = cpu.HL
 *     return 4
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2457
 * 
 * 
 * def LD_EA(cpu, v): # EA LD (a16),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(v, cpu.A)
 *     cpu.PC += 3
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_EA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2458
 * 
 * def LD_EA(cpu, v): # EA LD (a16),A
 *     cpu.mb.setitem(v, cpu.A)             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_v, __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":2459
 * def LD_EA(cpu, v): # EA LD (a16),A
 *     cpu.mb.setitem(v, cpu.A)
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":2460
 *     cpu.mb.setitem(v, cpu.A)
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2461
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2457
 * 
 * 
 * def LD_EA(cpu, v): # EA LD (a16),A             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(v, cpu.A)
 *     cpu.PC += 3
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2464
 * 
 * 
 * def XOR_EE(cpu, v): # EE XOR d8             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ v
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_XOR_EE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2465
 * 
 * def XOR_EE(cpu, v): # EE XOR d8
 *     t = cpu.A ^ v             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A ^ __pyx_v_v);

  /* "pyboy/core/opcodes.py":2466
 * def XOR_EE(cpu, v): # EE XOR d8
 *     t = cpu.A ^ v
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2467
 *     t = cpu.A ^ v
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2468
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2469
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2470
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2471
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2472
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2473
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2474
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2464
 * 
 * 
 * def XOR_EE(cpu, v): # EE XOR d8             # <<<<<<<<<<<<<<
 *     t = cpu.A ^ v
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2477
 * 
 * 
 * def RST_EF(cpu): # EF RST 28H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_EF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2478
 * 
 * def RST_EF(cpu): # EF RST 28H
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2479
 * def RST_EF(cpu): # EF RST 28H
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2480
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

  /* "pyboy/core/opcodes.py":2481
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

  /* "pyboy/core/opcodes.py":2482
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 40
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2483
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC = 40
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2484
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 40             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = 40;

  /* "pyboy/core/opcodes.py":2485
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 40
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2477
 * 
 * 
 * def RST_EF(cpu): # EF RST 28H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2488
 * 
 * 
 * def LDH_F0(cpu, v): # F0 LDH A,(a8)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(v + 0xFF00)
 *     cpu.PC += 2
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LDH_F0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2489
 * 
 * def LDH_F0(cpu, v): # F0 LDH A,(a8)
 *     cpu.A = cpu.mb.getitem(v + 0xFF00)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, (__pyx_v_v + 0xFF00));

  /* "pyboy/core/opcodes.py":2490
 * def LDH_F0(cpu, v): # F0 LDH A,(a8)
 *     cpu.A = cpu.mb.getitem(v + 0xFF00)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2491
 *     cpu.A = cpu.mb.getitem(v + 0xFF00)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2492
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2488
 * 
 * 
 * def LDH_F0(cpu, v): # F0 LDH A,(a8)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(v + 0xFF00)
 *     cpu.PC += 2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2495
 * 
 * 
 * def POP_F1(cpu): # F1 POP AF             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.F = cpu.mb.getitem(cpu.SP) & 0xF0 & 0xF0 # Low
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_POP_F1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2496
 * 
 * def POP_F1(cpu): # F1 POP AF
 *     cpu.A = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High             # <<<<<<<<<<<<<<
 *     cpu.F = cpu.mb.getitem(cpu.SP) & 0xF0 & 0xF0 # Low
 *     cpu.SP += 2
 */
  __pyx_v_cpu->A = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP + 1) & 0xFFFF));

  /* "pyboy/core/opcodes.py":2497
 * def POP_F1(cpu): # F1 POP AF
 *     cpu.A = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.F = cpu.mb.getitem(cpu.SP) & 0xF0 & 0xF0 # Low             # <<<<<<<<<<<<<<
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 */
  __pyx_v_cpu->F = ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->SP) & 0xF0) & 0xF0);

  /* "pyboy/core/opcodes.py":2498
 *     cpu.A = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.F = cpu.mb.getitem(cpu.SP) & 0xF0 & 0xF0 # Low
 *     cpu.SP += 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP + 2);

  /* "pyboy/core/opcodes.py":2499
 *     cpu.F = cpu.mb.getitem(cpu.SP) & 0xF0 & 0xF0 # Low
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2500
 *     cpu.SP += 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2501
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2502
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2495
 * 
 * 
 * def POP_F1(cpu): # F1 POP AF             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem((cpu.SP + 1) & 0xFFFF) # High
 *     cpu.F = cpu.mb.getitem(cpu.SP) & 0xF0 & 0xF0 # Low
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2505
 * 
 * 
 * def LD_F2(cpu): # F2 LD A,(C)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(0xFF00 + cpu.C)
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_F2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2506
 * 
 * def LD_F2(cpu): # F2 LD A,(C)
 *     cpu.A = cpu.mb.getitem(0xFF00 + cpu.C)             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, (0xFF00 + __pyx_v_cpu->C));

  /* "pyboy/core/opcodes.py":2507
 * def LD_F2(cpu): # F2 LD A,(C)
 *     cpu.A = cpu.mb.getitem(0xFF00 + cpu.C)
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2508
 *     cpu.A = cpu.mb.getitem(0xFF00 + cpu.C)
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2509
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2505
 * 
 * 
 * def LD_F2(cpu): # F2 LD A,(C)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(0xFF00 + cpu.C)
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2512
 * 
 * 
 * def DI_F3(cpu): # F3 DI             # <<<<<<<<<<<<<<
 *     cpu.interrupt_master_enable = False
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_DI_F3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2513
 * 
 * def DI_F3(cpu): # F3 DI
 *     cpu.interrupt_master_enable = False             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->interrupt_master_enable = 0;

  /* "pyboy/core/opcodes.py":2514
 * def DI_F3(cpu): # F3 DI
 *     cpu.interrupt_master_enable = False
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2515
 *     cpu.interrupt_master_enable = False
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2516
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2512
 * 
 * 
 * def DI_F3(cpu): # F3 DI             # <<<<<<<<<<<<<<
 *     cpu.interrupt_master_enable = False
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2519
 * 
 * 
 * def PUSH_F5(cpu): # F5 PUSH AF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.A) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.F & 0xF0) # Low
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_PUSH_F5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2520
 * 
 * def PUSH_F5(cpu): # F5 PUSH AF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.A) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.F & 0xF0) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), __pyx_v_cpu->A);

  /* "pyboy/core/opcodes.py":2521
 * def PUSH_F5(cpu): # F5 PUSH AF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.A) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.F & 0xF0) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->F & 0xF0));

  /* "pyboy/core/opcodes.py":2522
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.A) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.F & 0xF0) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2523
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.F & 0xF0) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2524
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2525
 *     cpu.SP &= 0xFFFF
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2526
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2519
 * 
 * 
 * def PUSH_F5(cpu): # F5 PUSH AF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.A) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.F & 0xF0) # Low
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2529
 * 
 * 
 * def OR_F6(cpu, v): # F6 OR d8             # <<<<<<<<<<<<<<
 *     t = cpu.A | v
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_OR_F6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2530
 * 
 * def OR_F6(cpu, v): # F6 OR d8
 *     t = cpu.A | v             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A | __pyx_v_v);

  /* "pyboy/core/opcodes.py":2531
 * def OR_F6(cpu, v): # F6 OR d8
 *     t = cpu.A | v
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2532
 *     t = cpu.A | v
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2533
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2534
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2535
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2536
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2537
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2538
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2539
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2529
 * 
 * 
 * def OR_F6(cpu, v): # F6 OR d8             # <<<<<<<<<<<<<<
 *     t = cpu.A | v
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2542
 * 
 * 
 * def RST_F7(cpu): # F7 RST 30H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_F7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2543
 * 
 * def RST_F7(cpu): # F7 RST 30H
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2544
 * def RST_F7(cpu): # F7 RST 30H
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2545
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

  /* "pyboy/core/opcodes.py":2546
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

  /* "pyboy/core/opcodes.py":2547
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 48
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2548
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC = 48
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2549
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 48             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = 48;

  /* "pyboy/core/opcodes.py":2550
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 48
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2542
 * 
 * 
 * def RST_F7(cpu): # F7 RST 30H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2553
 * 
 * 
 * def LD_F8(cpu, v): # F8 LD HL,SP+r8             # <<<<<<<<<<<<<<
 *     cpu.HL = cpu.SP + ((v ^ 0x80) - 0x80)
 *     t = cpu.HL
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_F8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  CYTHON_UNUSED int __pyx_v_t;
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2554
 * 
 * def LD_F8(cpu, v): # F8 LD HL,SP+r8
 *     cpu.HL = cpu.SP + ((v ^ 0x80) - 0x80)             # <<<<<<<<<<<<<<
 *     t = cpu.HL
 *     flag = 0b00000000
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->SP + ((__pyx_v_v ^ 0x80) - 0x80));

  /* "pyboy/core/opcodes.py":2555
 * def LD_F8(cpu, v): # F8 LD HL,SP+r8
 *     cpu.HL = cpu.SP + ((v ^ 0x80) - 0x80)
 *     t = cpu.HL             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 */
  __pyx_t_1 = __pyx_v_cpu->HL;
  __pyx_v_t = __pyx_t_1;

  /* "pyboy/core/opcodes.py":2556
 *     cpu.HL = cpu.SP + ((v ^ 0x80) - 0x80)
 *     t = cpu.HL
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2557
 *     t = cpu.HL
 *     flag = 0b00000000
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->SP & 0xF) + (__pyx_v_v & 0xF)) > 0xF) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2558
 *     flag = 0b00000000
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->SP & 0xFF) + (__pyx_v_v & 0xFF)) > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2559
 *     flag += (((cpu.SP & 0xF) + (v & 0xF)) > 0xF) << FLAGH
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.HL &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2560
 *     flag += (((cpu.SP & 0xFF) + (v & 0xFF)) > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2561
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     cpu.HL &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = (__pyx_v_cpu->HL & 0xFFFF);

  /* "pyboy/core/opcodes.py":2562
 *     cpu.F |= flag
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 12
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2563
 *     cpu.HL &= 0xFFFF
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 12
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2564
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 12;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2553
 * 
 * 
 * def LD_F8(cpu, v): # F8 LD HL,SP+r8             # <<<<<<<<<<<<<<
 *     cpu.HL = cpu.SP + ((v ^ 0x80) - 0x80)
 *     t = cpu.HL
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2567
 * 
 * 
 * def LD_F9(cpu): # F9 LD SP,HL             # <<<<<<<<<<<<<<
 *     cpu.SP = cpu.HL
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_F9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":2568
 * 
 * def LD_F9(cpu): # F9 LD SP,HL
 *     cpu.SP = cpu.HL             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_t_1 = __pyx_v_cpu->HL;
  __pyx_v_cpu->SP = __pyx_t_1;

  /* "pyboy/core/opcodes.py":2569
 * def LD_F9(cpu): # F9 LD SP,HL
 *     cpu.SP = cpu.HL
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2570
 *     cpu.SP = cpu.HL
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2571
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2567
 * 
 * 
 * def LD_F9(cpu): # F9 LD SP,HL             # <<<<<<<<<<<<<<
 *     cpu.SP = cpu.HL
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2574
 * 
 * 
 * def LD_FA(cpu, v): # FA LD A,(a16)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(v)
 *     cpu.PC += 3
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_LD_FA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2575
 * 
 * def LD_FA(cpu, v): # FA LD A,(a16)
 *     cpu.A = cpu.mb.getitem(v)             # <<<<<<<<<<<<<<
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_v);

  /* "pyboy/core/opcodes.py":2576
 * def LD_FA(cpu, v): # FA LD A,(a16)
 *     cpu.A = cpu.mb.getitem(v)
 *     cpu.PC += 3             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 3);

  /* "pyboy/core/opcodes.py":2577
 *     cpu.A = cpu.mb.getitem(v)
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2578
 *     cpu.PC += 3
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2574
 * 
 * 
 * def LD_FA(cpu, v): # FA LD A,(a16)             # <<<<<<<<<<<<<<
 *     cpu.A = cpu.mb.getitem(v)
 *     cpu.PC += 3
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2581
 * 
 * 
 * def EI_FB(cpu): # FB EI             # <<<<<<<<<<<<<<
 *     cpu.interrupt_master_enable = True
 *     cpu.PC += 1
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_EI_FB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2582
 * 
 * def EI_FB(cpu): # FB EI
 *     cpu.interrupt_master_enable = True             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->interrupt_master_enable = 1;

  /* "pyboy/core/opcodes.py":2583
 * def EI_FB(cpu): # FB EI
 *     cpu.interrupt_master_enable = True
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 4
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2584
 *     cpu.interrupt_master_enable = True
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 4
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2585
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     return 4             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 4;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2581
 * 
 * 
 * def EI_FB(cpu): # FB EI             # <<<<<<<<<<<<<<
 *     cpu.interrupt_master_enable = True
 *     cpu.PC += 1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2588
 * 
 * 
 * def CP_FE(cpu, v): # FE CP d8             # <<<<<<<<<<<<<<
 *     t = cpu.A - v
 *     flag = 0b01000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_CP_FE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, int __pyx_v_v) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2589
 * 
 * def CP_FE(cpu, v): # FE CP d8
 *     t = cpu.A - v             # <<<<<<<<<<<<<<
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A - __pyx_v_v);

  /* "pyboy/core/opcodes.py":2590
 * def CP_FE(cpu, v): # FE CP d8
 *     t = cpu.A - v
 *     flag = 0b01000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH
 */
  __pyx_v_flag = 64;

  /* "pyboy/core/opcodes.py":2591
 *     t = cpu.A - v
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2592
 *     flag = 0b01000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH             # <<<<<<<<<<<<<<
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + ((((__pyx_v_cpu->A & 0xF) - (__pyx_v_v & 0xF)) < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGH));

  /* "pyboy/core/opcodes.py":2593
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t < 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2594
 *     flag += (((cpu.A & 0xF) - (v & 0xF)) < 0) << FLAGH
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2595
 *     flag += (t < 0) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2596
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2597
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2598
 *     t &= 0xFF
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2599
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2588
 * 
 * 
 * def CP_FE(cpu, v): # FE CP d8             # <<<<<<<<<<<<<<
 *     t = cpu.A - v
 *     flag = 0b01000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2602
 * 
 * 
 * def RST_FF(cpu): # FF RST 38H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RST_FF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2603
 * 
 * def RST_FF(cpu): # FF RST 38H
 *     cpu.PC += 1             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 1);

  /* "pyboy/core/opcodes.py":2604
 * def RST_FF(cpu): # FF RST 38H
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2605
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 1) & 0xFFFF), (__pyx_v_cpu->PC >> 8));

  /* "pyboy/core/opcodes.py":2606
 *     cpu.PC &= 0xFFFF
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low             # <<<<<<<<<<<<<<
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, ((__pyx_v_cpu->SP - 2) & 0xFFFF), (__pyx_v_cpu->PC & 0xFF));

  /* "pyboy/core/opcodes.py":2607
 *     cpu.mb.setitem((cpu.SP-1) & 0xFFFF, cpu.PC >> 8) # High
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2             # <<<<<<<<<<<<<<
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 56
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP - 2);

  /* "pyboy/core/opcodes.py":2608
 *     cpu.mb.setitem((cpu.SP-2) & 0xFFFF, cpu.PC & 0xFF) # Low
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF             # <<<<<<<<<<<<<<
 *     cpu.PC = 56
 *     return 16
 */
  __pyx_v_cpu->SP = (__pyx_v_cpu->SP & 0xFFFF);

  /* "pyboy/core/opcodes.py":2609
 *     cpu.SP -= 2
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 56             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = 56;

  /* "pyboy/core/opcodes.py":2610
 *     cpu.SP &= 0xFFFF
 *     cpu.PC = 56
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2602
 * 
 * 
 * def RST_FF(cpu): # FF RST 38H             # <<<<<<<<<<<<<<
 *     cpu.PC += 1
 *     cpu.PC &= 0xFFFF
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2613
 * 
 * 
 * def RLC_100(cpu): # 100 RLC B             # <<<<<<<<<<<<<<
 *     t = (cpu.B << 1) + (cpu.B >> 7)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_100(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2614
 * 
 * def RLC_100(cpu): # 100 RLC B
 *     t = (cpu.B << 1) + (cpu.B >> 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->B << 1) + (__pyx_v_cpu->B >> 7));

  /* "pyboy/core/opcodes.py":2615
 * def RLC_100(cpu): # 100 RLC B
 *     t = (cpu.B << 1) + (cpu.B >> 7)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2616
 *     t = (cpu.B << 1) + (cpu.B >> 7)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2617
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2618
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2619
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2620
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2621
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2622
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2623
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2624
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2613
 * 
 * 
 * def RLC_100(cpu): # 100 RLC B             # <<<<<<<<<<<<<<
 *     t = (cpu.B << 1) + (cpu.B >> 7)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2627
 * 
 * 
 * def RLC_101(cpu): # 101 RLC C             # <<<<<<<<<<<<<<
 *     t = (cpu.C << 1) + (cpu.C >> 7)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_101(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2628
 * 
 * def RLC_101(cpu): # 101 RLC C
 *     t = (cpu.C << 1) + (cpu.C >> 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->C << 1) + (__pyx_v_cpu->C >> 7));

  /* "pyboy/core/opcodes.py":2629
 * def RLC_101(cpu): # 101 RLC C
 *     t = (cpu.C << 1) + (cpu.C >> 7)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2630
 *     t = (cpu.C << 1) + (cpu.C >> 7)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2631
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2632
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2633
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2634
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2635
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2636
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2637
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2638
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2627
 * 
 * 
 * def RLC_101(cpu): # 101 RLC C             # <<<<<<<<<<<<<<
 *     t = (cpu.C << 1) + (cpu.C >> 7)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2641
 * 
 * 
 * def RLC_102(cpu): # 102 RLC D             # <<<<<<<<<<<<<<
 *     t = (cpu.D << 1) + (cpu.D >> 7)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_102(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2642
 * 
 * def RLC_102(cpu): # 102 RLC D
 *     t = (cpu.D << 1) + (cpu.D >> 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->D << 1) + (__pyx_v_cpu->D >> 7));

  /* "pyboy/core/opcodes.py":2643
 * def RLC_102(cpu): # 102 RLC D
 *     t = (cpu.D << 1) + (cpu.D >> 7)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2644
 *     t = (cpu.D << 1) + (cpu.D >> 7)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2645
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2646
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2647
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2648
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2649
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2650
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2651
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2652
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2641
 * 
 * 
 * def RLC_102(cpu): # 102 RLC D             # <<<<<<<<<<<<<<
 *     t = (cpu.D << 1) + (cpu.D >> 7)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2655
 * 
 * 
 * def RLC_103(cpu): # 103 RLC E             # <<<<<<<<<<<<<<
 *     t = (cpu.E << 1) + (cpu.E >> 7)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_103(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2656
 * 
 * def RLC_103(cpu): # 103 RLC E
 *     t = (cpu.E << 1) + (cpu.E >> 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->E << 1) + (__pyx_v_cpu->E >> 7));

  /* "pyboy/core/opcodes.py":2657
 * def RLC_103(cpu): # 103 RLC E
 *     t = (cpu.E << 1) + (cpu.E >> 7)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2658
 *     t = (cpu.E << 1) + (cpu.E >> 7)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2659
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2660
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2661
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2662
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2663
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2664
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2665
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2666
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2655
 * 
 * 
 * def RLC_103(cpu): # 103 RLC E             # <<<<<<<<<<<<<<
 *     t = (cpu.E << 1) + (cpu.E >> 7)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2669
 * 
 * 
 * def RLC_104(cpu): # 104 RLC H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.HL >> 8) >> 7)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_104(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2670
 * 
 * def RLC_104(cpu): # 104 RLC H
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.HL >> 8) >> 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->HL >> 8) << 1) + ((__pyx_v_cpu->HL >> 8) >> 7));

  /* "pyboy/core/opcodes.py":2671
 * def RLC_104(cpu): # 104 RLC H
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.HL >> 8) >> 7)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2672
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.HL >> 8) >> 7)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2673
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2674
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2675
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2676
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2677
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":2678
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2679
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2680
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2669
 * 
 * 
 * def RLC_104(cpu): # 104 RLC H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.HL >> 8) >> 7)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2683
 * 
 * 
 * def RLC_105(cpu): # 105 RLC L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.HL & 0xFF) >> 7)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_105(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2684
 * 
 * def RLC_105(cpu): # 105 RLC L
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.HL & 0xFF) >> 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->HL & 0xFF) << 1) + ((__pyx_v_cpu->HL & 0xFF) >> 7));

  /* "pyboy/core/opcodes.py":2685
 * def RLC_105(cpu): # 105 RLC L
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.HL & 0xFF) >> 7)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2686
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.HL & 0xFF) >> 7)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2687
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2688
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2689
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2690
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2691
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":2692
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2693
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2694
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2683
 * 
 * 
 * def RLC_105(cpu): # 105 RLC L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.HL & 0xFF) >> 7)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2697
 * 
 * 
 * def RLC_106(cpu): # 106 RLC (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + (cpu.mb.getitem(cpu.HL) >> 7)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_106(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2698
 * 
 * def RLC_106(cpu): # 106 RLC (HL)
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + (cpu.mb.getitem(cpu.HL) >> 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) << 1) + (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) >> 7));

  /* "pyboy/core/opcodes.py":2699
 * def RLC_106(cpu): # 106 RLC (HL)
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + (cpu.mb.getitem(cpu.HL) >> 7)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2700
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + (cpu.mb.getitem(cpu.HL) >> 7)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2701
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2702
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2703
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2704
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2705
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":2706
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2707
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2708
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2697
 * 
 * 
 * def RLC_106(cpu): # 106 RLC (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + (cpu.mb.getitem(cpu.HL) >> 7)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2711
 * 
 * 
 * def RLC_107(cpu): # 107 RLC A             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1) + (cpu.A >> 7)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RLC_107(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2712
 * 
 * def RLC_107(cpu): # 107 RLC A
 *     t = (cpu.A << 1) + (cpu.A >> 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A << 1) + (__pyx_v_cpu->A >> 7));

  /* "pyboy/core/opcodes.py":2713
 * def RLC_107(cpu): # 107 RLC A
 *     t = (cpu.A << 1) + (cpu.A >> 7)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2714
 *     t = (cpu.A << 1) + (cpu.A >> 7)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2715
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2716
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2717
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2718
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2719
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2720
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2721
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2722
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2711
 * 
 * 
 * def RLC_107(cpu): # 107 RLC A             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1) + (cpu.A >> 7)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2725
 * 
 * 
 * def RRC_108(cpu): # 108 RRC B             # <<<<<<<<<<<<<<
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 7) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_108(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2726
 * 
 * def RRC_108(cpu): # 108 RRC B
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 7) + ((cpu.B & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->B >> 1) + ((__pyx_v_cpu->B & 1) << 7)) + ((__pyx_v_cpu->B & 1) << 8));

  /* "pyboy/core/opcodes.py":2727
 * def RRC_108(cpu): # 108 RRC B
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 7) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2728
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 7) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2729
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2730
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2731
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2732
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2733
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2734
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2735
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2736
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2725
 * 
 * 
 * def RRC_108(cpu): # 108 RRC B             # <<<<<<<<<<<<<<
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 7) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2739
 * 
 * 
 * def RRC_109(cpu): # 109 RRC C             # <<<<<<<<<<<<<<
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 7) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_109(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2740
 * 
 * def RRC_109(cpu): # 109 RRC C
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 7) + ((cpu.C & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->C >> 1) + ((__pyx_v_cpu->C & 1) << 7)) + ((__pyx_v_cpu->C & 1) << 8));

  /* "pyboy/core/opcodes.py":2741
 * def RRC_109(cpu): # 109 RRC C
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 7) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2742
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 7) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2743
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2744
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2745
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2746
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2747
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2748
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2749
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2750
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2739
 * 
 * 
 * def RRC_109(cpu): # 109 RRC C             # <<<<<<<<<<<<<<
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 7) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2753
 * 
 * 
 * def RRC_10A(cpu): # 10A RRC D             # <<<<<<<<<<<<<<
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 7) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2754
 * 
 * def RRC_10A(cpu): # 10A RRC D
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 7) + ((cpu.D & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->D >> 1) + ((__pyx_v_cpu->D & 1) << 7)) + ((__pyx_v_cpu->D & 1) << 8));

  /* "pyboy/core/opcodes.py":2755
 * def RRC_10A(cpu): # 10A RRC D
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 7) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2756
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 7) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2757
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2758
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2759
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2760
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2761
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2762
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2763
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2764
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2753
 * 
 * 
 * def RRC_10A(cpu): # 10A RRC D             # <<<<<<<<<<<<<<
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 7) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2767
 * 
 * 
 * def RRC_10B(cpu): # 10B RRC E             # <<<<<<<<<<<<<<
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 7) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2768
 * 
 * def RRC_10B(cpu): # 10B RRC E
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 7) + ((cpu.E & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->E >> 1) + ((__pyx_v_cpu->E & 1) << 7)) + ((__pyx_v_cpu->E & 1) << 8));

  /* "pyboy/core/opcodes.py":2769
 * def RRC_10B(cpu): # 10B RRC E
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 7) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2770
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 7) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2771
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2772
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2773
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2774
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2775
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2776
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2777
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2778
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2767
 * 
 * 
 * def RRC_10B(cpu): # 10B RRC E             # <<<<<<<<<<<<<<
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 7) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2781
 * 
 * 
 * def RRC_10C(cpu): # 10C RRC H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 7) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2782
 * 
 * def RRC_10C(cpu): # 10C RRC H
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 7) + (((cpu.HL >> 8) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((__pyx_v_cpu->HL >> 8) >> 1) + (((__pyx_v_cpu->HL >> 8) & 1) << 7)) + (((__pyx_v_cpu->HL >> 8) & 1) << 8));

  /* "pyboy/core/opcodes.py":2783
 * def RRC_10C(cpu): # 10C RRC H
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 7) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2784
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 7) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2785
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2786
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2787
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2788
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2789
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":2790
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2791
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2792
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2781
 * 
 * 
 * def RRC_10C(cpu): # 10C RRC H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 7) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2795
 * 
 * 
 * def RRC_10D(cpu): # 10D RRC L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 7) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2796
 * 
 * def RRC_10D(cpu): # 10D RRC L
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 7) + (((cpu.HL & 0xFF) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((__pyx_v_cpu->HL & 0xFF) >> 1) + (((__pyx_v_cpu->HL & 0xFF) & 1) << 7)) + (((__pyx_v_cpu->HL & 0xFF) & 1) << 8));

  /* "pyboy/core/opcodes.py":2797
 * def RRC_10D(cpu): # 10D RRC L
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 7) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2798
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 7) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2799
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2800
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2801
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2802
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2803
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":2804
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2805
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2806
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2795
 * 
 * 
 * def RRC_10D(cpu): # 10D RRC L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 7) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2809
 * 
 * 
 * def RRC_10E(cpu): # 10E RRC (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2810
 * 
 * def RRC_10E(cpu): # 10E RRC (HL)
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) >> 1) + ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 1) << 7)) + ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 1) << 8));

  /* "pyboy/core/opcodes.py":2811
 * def RRC_10E(cpu): # 10E RRC (HL)
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2812
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2813
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2814
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2815
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2816
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2817
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":2818
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2819
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2820
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2809
 * 
 * 
 * def RRC_10E(cpu): # 10E RRC (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2823
 * 
 * 
 * def RRC_10F(cpu): # 10F RRC A             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RRC_10F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2824
 * 
 * def RRC_10F(cpu): # 10F RRC A
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->A >> 1) + ((__pyx_v_cpu->A & 1) << 7)) + ((__pyx_v_cpu->A & 1) << 8));

  /* "pyboy/core/opcodes.py":2825
 * def RRC_10F(cpu): # 10F RRC A
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2826
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2827
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2828
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2829
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2830
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2831
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2832
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2833
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2834
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2823
 * 
 * 
 * def RRC_10F(cpu): # 10F RRC A             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2837
 * 
 * 
 * def RL_110(cpu): # 110 RL B             # <<<<<<<<<<<<<<
 *     t = (cpu.B << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_110(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2838
 * 
 * def RL_110(cpu): # 110 RL B
 *     t = (cpu.B << 1) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->B << 1) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2839
 * def RL_110(cpu): # 110 RL B
 *     t = (cpu.B << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2840
 *     t = (cpu.B << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2841
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2842
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2843
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2844
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2845
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2846
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2847
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2848
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2837
 * 
 * 
 * def RL_110(cpu): # 110 RL B             # <<<<<<<<<<<<<<
 *     t = (cpu.B << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2851
 * 
 * 
 * def RL_111(cpu): # 111 RL C             # <<<<<<<<<<<<<<
 *     t = (cpu.C << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_111(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2852
 * 
 * def RL_111(cpu): # 111 RL C
 *     t = (cpu.C << 1) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->C << 1) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2853
 * def RL_111(cpu): # 111 RL C
 *     t = (cpu.C << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2854
 *     t = (cpu.C << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2855
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2856
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2857
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2858
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2859
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2860
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2861
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2862
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2851
 * 
 * 
 * def RL_111(cpu): # 111 RL C             # <<<<<<<<<<<<<<
 *     t = (cpu.C << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2865
 * 
 * 
 * def RL_112(cpu): # 112 RL D             # <<<<<<<<<<<<<<
 *     t = (cpu.D << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_112(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2866
 * 
 * def RL_112(cpu): # 112 RL D
 *     t = (cpu.D << 1) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->D << 1) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2867
 * def RL_112(cpu): # 112 RL D
 *     t = (cpu.D << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2868
 *     t = (cpu.D << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2869
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2870
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2871
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2872
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2873
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2874
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2875
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2876
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2865
 * 
 * 
 * def RL_112(cpu): # 112 RL D             # <<<<<<<<<<<<<<
 *     t = (cpu.D << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2879
 * 
 * 
 * def RL_113(cpu): # 113 RL E             # <<<<<<<<<<<<<<
 *     t = (cpu.E << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_113(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2880
 * 
 * def RL_113(cpu): # 113 RL E
 *     t = (cpu.E << 1) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->E << 1) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2881
 * def RL_113(cpu): # 113 RL E
 *     t = (cpu.E << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2882
 *     t = (cpu.E << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2883
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2884
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2885
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2886
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2887
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2888
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2889
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2890
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2879
 * 
 * 
 * def RL_113(cpu): # 113 RL E             # <<<<<<<<<<<<<<
 *     t = (cpu.E << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2893
 * 
 * 
 * def RL_114(cpu): # 114 RL H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_114(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2894
 * 
 * def RL_114(cpu): # 114 RL H
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->HL >> 8) << 1) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2895
 * def RL_114(cpu): # 114 RL H
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2896
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2897
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2898
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2899
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2900
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2901
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":2902
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2903
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2904
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2893
 * 
 * 
 * def RL_114(cpu): # 114 RL H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2907
 * 
 * 
 * def RL_115(cpu): # 115 RL L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_115(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2908
 * 
 * def RL_115(cpu): # 115 RL L
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->HL & 0xFF) << 1) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2909
 * def RL_115(cpu): # 115 RL L
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2910
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2911
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2912
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2913
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2914
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2915
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":2916
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2917
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2918
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2907
 * 
 * 
 * def RL_115(cpu): # 115 RL L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2921
 * 
 * 
 * def RL_116(cpu): # 116 RL (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_116(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2922
 * 
 * def RL_116(cpu): # 116 RL (HL)
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) << 1) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2923
 * def RL_116(cpu): # 116 RL (HL)
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2924
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2925
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2926
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2927
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2928
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2929
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":2930
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2931
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2932
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2921
 * 
 * 
 * def RL_116(cpu): # 116 RL (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2935
 * 
 * 
 * def RL_117(cpu): # 117 RL A             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RL_117(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2936
 * 
 * def RL_117(cpu): # 117 RL A
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A << 1) + ((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0));

  /* "pyboy/core/opcodes.py":2937
 * def RL_117(cpu): # 117 RL A
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2938
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2939
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2940
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2941
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2942
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2943
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2944
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2945
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2946
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2935
 * 
 * 
 * def RL_117(cpu): # 117 RL A             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1) + ((cpu.F & (1 << FLAGC)) != 0)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2949
 * 
 * 
 * def RR_118(cpu): # 118 RR B             # <<<<<<<<<<<<<<
 *     t = (cpu.B >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_118(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2950
 * 
 * def RR_118(cpu): # 118 RR B
 *     t = (cpu.B >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.B & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->B >> 1) + (((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0) << 7)) + ((__pyx_v_cpu->B & 1) << 8));

  /* "pyboy/core/opcodes.py":2951
 * def RR_118(cpu): # 118 RR B
 *     t = (cpu.B >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2952
 *     t = (cpu.B >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2953
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2954
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2955
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2956
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2957
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2958
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2959
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2960
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2949
 * 
 * 
 * def RR_118(cpu): # 118 RR B             # <<<<<<<<<<<<<<
 *     t = (cpu.B >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2963
 * 
 * 
 * def RR_119(cpu): # 119 RR C             # <<<<<<<<<<<<<<
 *     t = (cpu.C >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_119(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2964
 * 
 * def RR_119(cpu): # 119 RR C
 *     t = (cpu.C >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.C & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->C >> 1) + (((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0) << 7)) + ((__pyx_v_cpu->C & 1) << 8));

  /* "pyboy/core/opcodes.py":2965
 * def RR_119(cpu): # 119 RR C
 *     t = (cpu.C >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2966
 *     t = (cpu.C >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2967
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2968
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2969
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2970
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2971
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2972
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2973
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2974
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2963
 * 
 * 
 * def RR_119(cpu): # 119 RR C             # <<<<<<<<<<<<<<
 *     t = (cpu.C >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2977
 * 
 * 
 * def RR_11A(cpu): # 11A RR D             # <<<<<<<<<<<<<<
 *     t = (cpu.D >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2978
 * 
 * def RR_11A(cpu): # 11A RR D
 *     t = (cpu.D >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.D & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->D >> 1) + (((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0) << 7)) + ((__pyx_v_cpu->D & 1) << 8));

  /* "pyboy/core/opcodes.py":2979
 * def RR_11A(cpu): # 11A RR D
 *     t = (cpu.D >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2980
 *     t = (cpu.D >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2981
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2982
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2983
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2984
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2985
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":2986
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":2987
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":2988
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2977
 * 
 * 
 * def RR_11A(cpu): # 11A RR D             # <<<<<<<<<<<<<<
 *     t = (cpu.D >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":2991
 * 
 * 
 * def RR_11B(cpu): # 11B RR E             # <<<<<<<<<<<<<<
 *     t = (cpu.E >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":2992
 * 
 * def RR_11B(cpu): # 11B RR E
 *     t = (cpu.E >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.E & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->E >> 1) + (((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0) << 7)) + ((__pyx_v_cpu->E & 1) << 8));

  /* "pyboy/core/opcodes.py":2993
 * def RR_11B(cpu): # 11B RR E
 *     t = (cpu.E >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":2994
 *     t = (cpu.E >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":2995
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":2996
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":2997
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":2998
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":2999
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3000
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3001
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3002
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":2991
 * 
 * 
 * def RR_11B(cpu): # 11B RR E             # <<<<<<<<<<<<<<
 *     t = (cpu.E >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3005
 * 
 * 
 * def RR_11C(cpu): # 11C RR H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3006
 * 
 * def RR_11C(cpu): # 11C RR H
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL >> 8) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((__pyx_v_cpu->HL >> 8) >> 1) + (((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0) << 7)) + (((__pyx_v_cpu->HL >> 8) & 1) << 8));

  /* "pyboy/core/opcodes.py":3007
 * def RR_11C(cpu): # 11C RR H
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3008
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3009
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3010
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3011
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3012
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3013
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":3014
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3015
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3016
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3005
 * 
 * 
 * def RR_11C(cpu): # 11C RR H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3019
 * 
 * 
 * def RR_11D(cpu): # 11D RR L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3020
 * 
 * def RR_11D(cpu): # 11D RR L
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL & 0xFF) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((__pyx_v_cpu->HL & 0xFF) >> 1) + (((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0) << 7)) + (((__pyx_v_cpu->HL & 0xFF) & 1) << 8));

  /* "pyboy/core/opcodes.py":3021
 * def RR_11D(cpu): # 11D RR L
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3022
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3023
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3024
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3025
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3026
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3027
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":3028
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3029
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3030
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3019
 * 
 * 
 * def RR_11D(cpu): # 11D RR L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3033
 * 
 * 
 * def RR_11E(cpu): # 11E RR (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3034
 * 
 * def RR_11E(cpu): # 11E RR (HL)
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) >> 1) + (((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0) << 7)) + ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 1) << 8));

  /* "pyboy/core/opcodes.py":3035
 * def RR_11E(cpu): # 11E RR (HL)
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3036
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3037
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3038
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3039
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3040
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3041
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":3042
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3043
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3044
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3033
 * 
 * 
 * def RR_11E(cpu): # 11E RR (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3047
 * 
 * 
 * def RR_11F(cpu): # 11F RR A             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RR_11F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3048
 * 
 * def RR_11F(cpu): # 11F RR A
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->A >> 1) + (((__pyx_v_cpu->F & (1 << __pyx_v_5pyboy_4core_7opcodes_FLAGC)) != 0) << 7)) + ((__pyx_v_cpu->A & 1) << 8));

  /* "pyboy/core/opcodes.py":3049
 * def RR_11F(cpu): # 11F RR A
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3050
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3051
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3052
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3053
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3054
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3055
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3056
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3057
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3058
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3047
 * 
 * 
 * def RR_11F(cpu): # 11F RR A             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + (((cpu.F & (1 << FLAGC)) != 0) << 7) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3061
 * 
 * 
 * def SLA_120(cpu): # 120 SLA B             # <<<<<<<<<<<<<<
 *     t = (cpu.B << 1)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_120(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3062
 * 
 * def SLA_120(cpu): # 120 SLA B
 *     t = (cpu.B << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B << 1);

  /* "pyboy/core/opcodes.py":3063
 * def SLA_120(cpu): # 120 SLA B
 *     t = (cpu.B << 1)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3064
 *     t = (cpu.B << 1)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3065
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3066
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3067
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3068
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3069
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3070
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3071
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3072
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3061
 * 
 * 
 * def SLA_120(cpu): # 120 SLA B             # <<<<<<<<<<<<<<
 *     t = (cpu.B << 1)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3075
 * 
 * 
 * def SLA_121(cpu): # 121 SLA C             # <<<<<<<<<<<<<<
 *     t = (cpu.C << 1)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_121(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3076
 * 
 * def SLA_121(cpu): # 121 SLA C
 *     t = (cpu.C << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C << 1);

  /* "pyboy/core/opcodes.py":3077
 * def SLA_121(cpu): # 121 SLA C
 *     t = (cpu.C << 1)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3078
 *     t = (cpu.C << 1)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3079
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3080
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3081
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3082
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3083
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3084
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3085
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3086
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3075
 * 
 * 
 * def SLA_121(cpu): # 121 SLA C             # <<<<<<<<<<<<<<
 *     t = (cpu.C << 1)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3089
 * 
 * 
 * def SLA_122(cpu): # 122 SLA D             # <<<<<<<<<<<<<<
 *     t = (cpu.D << 1)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_122(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3090
 * 
 * def SLA_122(cpu): # 122 SLA D
 *     t = (cpu.D << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D << 1);

  /* "pyboy/core/opcodes.py":3091
 * def SLA_122(cpu): # 122 SLA D
 *     t = (cpu.D << 1)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3092
 *     t = (cpu.D << 1)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3093
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3094
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3095
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3096
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3097
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3098
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3099
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3100
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3089
 * 
 * 
 * def SLA_122(cpu): # 122 SLA D             # <<<<<<<<<<<<<<
 *     t = (cpu.D << 1)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3103
 * 
 * 
 * def SLA_123(cpu): # 123 SLA E             # <<<<<<<<<<<<<<
 *     t = (cpu.E << 1)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_123(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3104
 * 
 * def SLA_123(cpu): # 123 SLA E
 *     t = (cpu.E << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E << 1);

  /* "pyboy/core/opcodes.py":3105
 * def SLA_123(cpu): # 123 SLA E
 *     t = (cpu.E << 1)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3106
 *     t = (cpu.E << 1)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3107
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3108
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3109
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3110
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3111
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3112
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3113
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3114
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3103
 * 
 * 
 * def SLA_123(cpu): # 123 SLA E             # <<<<<<<<<<<<<<
 *     t = (cpu.E << 1)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3117
 * 
 * 
 * def SLA_124(cpu): # 124 SLA H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) << 1)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_124(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3118
 * 
 * def SLA_124(cpu): # 124 SLA H
 *     t = ((cpu.HL >> 8) << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) << 1);

  /* "pyboy/core/opcodes.py":3119
 * def SLA_124(cpu): # 124 SLA H
 *     t = ((cpu.HL >> 8) << 1)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3120
 *     t = ((cpu.HL >> 8) << 1)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3121
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3122
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3123
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3124
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3125
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":3126
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3127
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3128
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3117
 * 
 * 
 * def SLA_124(cpu): # 124 SLA H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) << 1)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3131
 * 
 * 
 * def SLA_125(cpu): # 125 SLA L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) << 1)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_125(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3132
 * 
 * def SLA_125(cpu): # 125 SLA L
 *     t = ((cpu.HL & 0xFF) << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) << 1);

  /* "pyboy/core/opcodes.py":3133
 * def SLA_125(cpu): # 125 SLA L
 *     t = ((cpu.HL & 0xFF) << 1)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3134
 *     t = ((cpu.HL & 0xFF) << 1)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3135
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3136
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3137
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3138
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3139
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":3140
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3141
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3142
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3131
 * 
 * 
 * def SLA_125(cpu): # 125 SLA L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) << 1)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3145
 * 
 * 
 * def SLA_126(cpu): # 126 SLA (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) << 1)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_126(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3146
 * 
 * def SLA_126(cpu): # 126 SLA (HL)
 *     t = (cpu.mb.getitem(cpu.HL) << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) << 1);

  /* "pyboy/core/opcodes.py":3147
 * def SLA_126(cpu): # 126 SLA (HL)
 *     t = (cpu.mb.getitem(cpu.HL) << 1)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3148
 *     t = (cpu.mb.getitem(cpu.HL) << 1)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3149
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3150
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3151
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3152
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3153
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":3154
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3155
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3156
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3145
 * 
 * 
 * def SLA_126(cpu): # 126 SLA (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) << 1)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3159
 * 
 * 
 * def SLA_127(cpu): # 127 SLA A             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SLA_127(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3160
 * 
 * def SLA_127(cpu): # 127 SLA A
 *     t = (cpu.A << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A << 1);

  /* "pyboy/core/opcodes.py":3161
 * def SLA_127(cpu): # 127 SLA A
 *     t = (cpu.A << 1)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3162
 *     t = (cpu.A << 1)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3163
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3164
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3165
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3166
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3167
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3168
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3169
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3170
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3159
 * 
 * 
 * def SLA_127(cpu): # 127 SLA A             # <<<<<<<<<<<<<<
 *     t = (cpu.A << 1)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3173
 * 
 * 
 * def SRA_128(cpu): # 128 SRA B             # <<<<<<<<<<<<<<
 *     t = ((cpu.B >> 1) | (cpu.B & 0x80)) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_128(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3174
 * 
 * def SRA_128(cpu): # 128 SRA B
 *     t = ((cpu.B >> 1) | (cpu.B & 0x80)) + ((cpu.B & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->B >> 1) | (__pyx_v_cpu->B & 0x80)) + ((__pyx_v_cpu->B & 1) << 8));

  /* "pyboy/core/opcodes.py":3175
 * def SRA_128(cpu): # 128 SRA B
 *     t = ((cpu.B >> 1) | (cpu.B & 0x80)) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3176
 *     t = ((cpu.B >> 1) | (cpu.B & 0x80)) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3177
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3178
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3179
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3180
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3181
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3182
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3183
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3184
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3173
 * 
 * 
 * def SRA_128(cpu): # 128 SRA B             # <<<<<<<<<<<<<<
 *     t = ((cpu.B >> 1) | (cpu.B & 0x80)) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3187
 * 
 * 
 * def SRA_129(cpu): # 129 SRA C             # <<<<<<<<<<<<<<
 *     t = ((cpu.C >> 1) | (cpu.C & 0x80)) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_129(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3188
 * 
 * def SRA_129(cpu): # 129 SRA C
 *     t = ((cpu.C >> 1) | (cpu.C & 0x80)) + ((cpu.C & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->C >> 1) | (__pyx_v_cpu->C & 0x80)) + ((__pyx_v_cpu->C & 1) << 8));

  /* "pyboy/core/opcodes.py":3189
 * def SRA_129(cpu): # 129 SRA C
 *     t = ((cpu.C >> 1) | (cpu.C & 0x80)) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3190
 *     t = ((cpu.C >> 1) | (cpu.C & 0x80)) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3191
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3192
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3193
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3194
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3195
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3196
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3197
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3198
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3187
 * 
 * 
 * def SRA_129(cpu): # 129 SRA C             # <<<<<<<<<<<<<<
 *     t = ((cpu.C >> 1) | (cpu.C & 0x80)) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3201
 * 
 * 
 * def SRA_12A(cpu): # 12A SRA D             # <<<<<<<<<<<<<<
 *     t = ((cpu.D >> 1) | (cpu.D & 0x80)) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3202
 * 
 * def SRA_12A(cpu): # 12A SRA D
 *     t = ((cpu.D >> 1) | (cpu.D & 0x80)) + ((cpu.D & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->D >> 1) | (__pyx_v_cpu->D & 0x80)) + ((__pyx_v_cpu->D & 1) << 8));

  /* "pyboy/core/opcodes.py":3203
 * def SRA_12A(cpu): # 12A SRA D
 *     t = ((cpu.D >> 1) | (cpu.D & 0x80)) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3204
 *     t = ((cpu.D >> 1) | (cpu.D & 0x80)) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3205
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3206
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3207
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3208
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3209
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3210
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3211
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3212
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3201
 * 
 * 
 * def SRA_12A(cpu): # 12A SRA D             # <<<<<<<<<<<<<<
 *     t = ((cpu.D >> 1) | (cpu.D & 0x80)) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3215
 * 
 * 
 * def SRA_12B(cpu): # 12B SRA E             # <<<<<<<<<<<<<<
 *     t = ((cpu.E >> 1) | (cpu.E & 0x80)) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3216
 * 
 * def SRA_12B(cpu): # 12B SRA E
 *     t = ((cpu.E >> 1) | (cpu.E & 0x80)) + ((cpu.E & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->E >> 1) | (__pyx_v_cpu->E & 0x80)) + ((__pyx_v_cpu->E & 1) << 8));

  /* "pyboy/core/opcodes.py":3217
 * def SRA_12B(cpu): # 12B SRA E
 *     t = ((cpu.E >> 1) | (cpu.E & 0x80)) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3218
 *     t = ((cpu.E >> 1) | (cpu.E & 0x80)) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3219
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3220
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3221
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3222
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3223
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3224
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3225
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3226
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3215
 * 
 * 
 * def SRA_12B(cpu): # 12B SRA E             # <<<<<<<<<<<<<<
 *     t = ((cpu.E >> 1) | (cpu.E & 0x80)) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3229
 * 
 * 
 * def SRA_12C(cpu): # 12C SRA H             # <<<<<<<<<<<<<<
 *     t = (((cpu.HL >> 8) >> 1) | ((cpu.HL >> 8) & 0x80)) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3230
 * 
 * def SRA_12C(cpu): # 12C SRA H
 *     t = (((cpu.HL >> 8) >> 1) | ((cpu.HL >> 8) & 0x80)) + (((cpu.HL >> 8) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((__pyx_v_cpu->HL >> 8) >> 1) | ((__pyx_v_cpu->HL >> 8) & 0x80)) + (((__pyx_v_cpu->HL >> 8) & 1) << 8));

  /* "pyboy/core/opcodes.py":3231
 * def SRA_12C(cpu): # 12C SRA H
 *     t = (((cpu.HL >> 8) >> 1) | ((cpu.HL >> 8) & 0x80)) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3232
 *     t = (((cpu.HL >> 8) >> 1) | ((cpu.HL >> 8) & 0x80)) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3233
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3234
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3235
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3236
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3237
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":3238
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3239
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3240
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3229
 * 
 * 
 * def SRA_12C(cpu): # 12C SRA H             # <<<<<<<<<<<<<<
 *     t = (((cpu.HL >> 8) >> 1) | ((cpu.HL >> 8) & 0x80)) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3243
 * 
 * 
 * def SRA_12D(cpu): # 12D SRA L             # <<<<<<<<<<<<<<
 *     t = (((cpu.HL & 0xFF) >> 1) | ((cpu.HL & 0xFF) & 0x80)) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3244
 * 
 * def SRA_12D(cpu): # 12D SRA L
 *     t = (((cpu.HL & 0xFF) >> 1) | ((cpu.HL & 0xFF) & 0x80)) + (((cpu.HL & 0xFF) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((__pyx_v_cpu->HL & 0xFF) >> 1) | ((__pyx_v_cpu->HL & 0xFF) & 0x80)) + (((__pyx_v_cpu->HL & 0xFF) & 1) << 8));

  /* "pyboy/core/opcodes.py":3245
 * def SRA_12D(cpu): # 12D SRA L
 *     t = (((cpu.HL & 0xFF) >> 1) | ((cpu.HL & 0xFF) & 0x80)) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3246
 *     t = (((cpu.HL & 0xFF) >> 1) | ((cpu.HL & 0xFF) & 0x80)) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3247
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3248
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3249
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3250
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3251
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":3252
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3253
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3254
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3243
 * 
 * 
 * def SRA_12D(cpu): # 12D SRA L             # <<<<<<<<<<<<<<
 *     t = (((cpu.HL & 0xFF) >> 1) | ((cpu.HL & 0xFF) & 0x80)) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3257
 * 
 * 
 * def SRA_12E(cpu): # 12E SRA (HL)             # <<<<<<<<<<<<<<
 *     t = ((cpu.mb.getitem(cpu.HL) >> 1) | (cpu.mb.getitem(cpu.HL) & 0x80)) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3258
 * 
 * def SRA_12E(cpu): # 12E SRA (HL)
 *     t = ((cpu.mb.getitem(cpu.HL) >> 1) | (cpu.mb.getitem(cpu.HL) & 0x80)) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) >> 1) | (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x80)) + ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 1) << 8));

  /* "pyboy/core/opcodes.py":3259
 * def SRA_12E(cpu): # 12E SRA (HL)
 *     t = ((cpu.mb.getitem(cpu.HL) >> 1) | (cpu.mb.getitem(cpu.HL) & 0x80)) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3260
 *     t = ((cpu.mb.getitem(cpu.HL) >> 1) | (cpu.mb.getitem(cpu.HL) & 0x80)) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3261
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3262
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3263
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3264
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3265
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":3266
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3267
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3268
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3257
 * 
 * 
 * def SRA_12E(cpu): # 12E SRA (HL)             # <<<<<<<<<<<<<<
 *     t = ((cpu.mb.getitem(cpu.HL) >> 1) | (cpu.mb.getitem(cpu.HL) & 0x80)) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3271
 * 
 * 
 * def SRA_12F(cpu): # 12F SRA A             # <<<<<<<<<<<<<<
 *     t = ((cpu.A >> 1) | (cpu.A & 0x80)) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRA_12F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3272
 * 
 * def SRA_12F(cpu): # 12F SRA A
 *     t = ((cpu.A >> 1) | (cpu.A & 0x80)) + ((cpu.A & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->A >> 1) | (__pyx_v_cpu->A & 0x80)) + ((__pyx_v_cpu->A & 1) << 8));

  /* "pyboy/core/opcodes.py":3273
 * def SRA_12F(cpu): # 12F SRA A
 *     t = ((cpu.A >> 1) | (cpu.A & 0x80)) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3274
 *     t = ((cpu.A >> 1) | (cpu.A & 0x80)) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3275
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3276
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3277
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3278
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3279
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3280
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3281
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3282
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3271
 * 
 * 
 * def SRA_12F(cpu): # 12F SRA A             # <<<<<<<<<<<<<<
 *     t = ((cpu.A >> 1) | (cpu.A & 0x80)) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3285
 * 
 * 
 * def SWAP_130(cpu): # 130 SWAP B             # <<<<<<<<<<<<<<
 *     t = ((cpu.B & 0xF0) >> 4) | ((cpu.B & 0x0F) << 4)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_130(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3286
 * 
 * def SWAP_130(cpu): # 130 SWAP B
 *     t = ((cpu.B & 0xF0) >> 4) | ((cpu.B & 0x0F) << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->B & 0xF0) >> 4) | ((__pyx_v_cpu->B & 0x0F) << 4));

  /* "pyboy/core/opcodes.py":3287
 * def SWAP_130(cpu): # 130 SWAP B
 *     t = ((cpu.B & 0xF0) >> 4) | ((cpu.B & 0x0F) << 4)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3288
 *     t = ((cpu.B & 0xF0) >> 4) | ((cpu.B & 0x0F) << 4)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3289
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3290
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3291
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3292
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3293
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3294
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3295
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3285
 * 
 * 
 * def SWAP_130(cpu): # 130 SWAP B             # <<<<<<<<<<<<<<
 *     t = ((cpu.B & 0xF0) >> 4) | ((cpu.B & 0x0F) << 4)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3298
 * 
 * 
 * def SWAP_131(cpu): # 131 SWAP C             # <<<<<<<<<<<<<<
 *     t = ((cpu.C & 0xF0) >> 4) | ((cpu.C & 0x0F) << 4)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_131(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3299
 * 
 * def SWAP_131(cpu): # 131 SWAP C
 *     t = ((cpu.C & 0xF0) >> 4) | ((cpu.C & 0x0F) << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->C & 0xF0) >> 4) | ((__pyx_v_cpu->C & 0x0F) << 4));

  /* "pyboy/core/opcodes.py":3300
 * def SWAP_131(cpu): # 131 SWAP C
 *     t = ((cpu.C & 0xF0) >> 4) | ((cpu.C & 0x0F) << 4)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3301
 *     t = ((cpu.C & 0xF0) >> 4) | ((cpu.C & 0x0F) << 4)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3302
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3303
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3304
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3305
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3306
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3307
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3308
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3298
 * 
 * 
 * def SWAP_131(cpu): # 131 SWAP C             # <<<<<<<<<<<<<<
 *     t = ((cpu.C & 0xF0) >> 4) | ((cpu.C & 0x0F) << 4)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3311
 * 
 * 
 * def SWAP_132(cpu): # 132 SWAP D             # <<<<<<<<<<<<<<
 *     t = ((cpu.D & 0xF0) >> 4) | ((cpu.D & 0x0F) << 4)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_132(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3312
 * 
 * def SWAP_132(cpu): # 132 SWAP D
 *     t = ((cpu.D & 0xF0) >> 4) | ((cpu.D & 0x0F) << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->D & 0xF0) >> 4) | ((__pyx_v_cpu->D & 0x0F) << 4));

  /* "pyboy/core/opcodes.py":3313
 * def SWAP_132(cpu): # 132 SWAP D
 *     t = ((cpu.D & 0xF0) >> 4) | ((cpu.D & 0x0F) << 4)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3314
 *     t = ((cpu.D & 0xF0) >> 4) | ((cpu.D & 0x0F) << 4)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3315
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3316
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3317
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3318
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3319
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3320
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3321
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3311
 * 
 * 
 * def SWAP_132(cpu): # 132 SWAP D             # <<<<<<<<<<<<<<
 *     t = ((cpu.D & 0xF0) >> 4) | ((cpu.D & 0x0F) << 4)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3324
 * 
 * 
 * def SWAP_133(cpu): # 133 SWAP E             # <<<<<<<<<<<<<<
 *     t = ((cpu.E & 0xF0) >> 4) | ((cpu.E & 0x0F) << 4)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_133(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3325
 * 
 * def SWAP_133(cpu): # 133 SWAP E
 *     t = ((cpu.E & 0xF0) >> 4) | ((cpu.E & 0x0F) << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->E & 0xF0) >> 4) | ((__pyx_v_cpu->E & 0x0F) << 4));

  /* "pyboy/core/opcodes.py":3326
 * def SWAP_133(cpu): # 133 SWAP E
 *     t = ((cpu.E & 0xF0) >> 4) | ((cpu.E & 0x0F) << 4)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3327
 *     t = ((cpu.E & 0xF0) >> 4) | ((cpu.E & 0x0F) << 4)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3328
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3329
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3330
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3331
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3332
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3333
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3334
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3324
 * 
 * 
 * def SWAP_133(cpu): # 133 SWAP E             # <<<<<<<<<<<<<<
 *     t = ((cpu.E & 0xF0) >> 4) | ((cpu.E & 0x0F) << 4)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3337
 * 
 * 
 * def SWAP_134(cpu): # 134 SWAP H             # <<<<<<<<<<<<<<
 *     t = (((cpu.HL >> 8) & 0xF0) >> 4) | (((cpu.HL >> 8) & 0x0F) << 4)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_134(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3338
 * 
 * def SWAP_134(cpu): # 134 SWAP H
 *     t = (((cpu.HL >> 8) & 0xF0) >> 4) | (((cpu.HL >> 8) & 0x0F) << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((__pyx_v_cpu->HL >> 8) & 0xF0) >> 4) | (((__pyx_v_cpu->HL >> 8) & 0x0F) << 4));

  /* "pyboy/core/opcodes.py":3339
 * def SWAP_134(cpu): # 134 SWAP H
 *     t = (((cpu.HL >> 8) & 0xF0) >> 4) | (((cpu.HL >> 8) & 0x0F) << 4)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3340
 *     t = (((cpu.HL >> 8) & 0xF0) >> 4) | (((cpu.HL >> 8) & 0x0F) << 4)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3341
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3342
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3343
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3344
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":3345
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3346
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3347
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3337
 * 
 * 
 * def SWAP_134(cpu): # 134 SWAP H             # <<<<<<<<<<<<<<
 *     t = (((cpu.HL >> 8) & 0xF0) >> 4) | (((cpu.HL >> 8) & 0x0F) << 4)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3350
 * 
 * 
 * def SWAP_135(cpu): # 135 SWAP L             # <<<<<<<<<<<<<<
 *     t = (((cpu.HL & 0xFF) & 0xF0) >> 4) | (((cpu.HL & 0xFF) & 0x0F) << 4)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_135(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3351
 * 
 * def SWAP_135(cpu): # 135 SWAP L
 *     t = (((cpu.HL & 0xFF) & 0xF0) >> 4) | (((cpu.HL & 0xFF) & 0x0F) << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((__pyx_v_cpu->HL & 0xFF) & 0xF0) >> 4) | (((__pyx_v_cpu->HL & 0xFF) & 0x0F) << 4));

  /* "pyboy/core/opcodes.py":3352
 * def SWAP_135(cpu): # 135 SWAP L
 *     t = (((cpu.HL & 0xFF) & 0xF0) >> 4) | (((cpu.HL & 0xFF) & 0x0F) << 4)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3353
 *     t = (((cpu.HL & 0xFF) & 0xF0) >> 4) | (((cpu.HL & 0xFF) & 0x0F) << 4)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3354
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3355
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3356
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3357
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":3358
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3359
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3360
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3350
 * 
 * 
 * def SWAP_135(cpu): # 135 SWAP L             # <<<<<<<<<<<<<<
 *     t = (((cpu.HL & 0xFF) & 0xF0) >> 4) | (((cpu.HL & 0xFF) & 0x0F) << 4)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3363
 * 
 * 
 * def SWAP_136(cpu): # 136 SWAP (HL)             # <<<<<<<<<<<<<<
 *     t = ((cpu.mb.getitem(cpu.HL) & 0xF0) >> 4) | ((cpu.mb.getitem(cpu.HL) & 0x0F) << 4)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_136(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3364
 * 
 * def SWAP_136(cpu): # 136 SWAP (HL)
 *     t = ((cpu.mb.getitem(cpu.HL) & 0xF0) >> 4) | ((cpu.mb.getitem(cpu.HL) & 0x0F) << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0xF0) >> 4) | ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x0F) << 4));

  /* "pyboy/core/opcodes.py":3365
 * def SWAP_136(cpu): # 136 SWAP (HL)
 *     t = ((cpu.mb.getitem(cpu.HL) & 0xF0) >> 4) | ((cpu.mb.getitem(cpu.HL) & 0x0F) << 4)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3366
 *     t = ((cpu.mb.getitem(cpu.HL) & 0xF0) >> 4) | ((cpu.mb.getitem(cpu.HL) & 0x0F) << 4)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3367
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3368
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3369
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3370
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":3371
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3372
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3373
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3363
 * 
 * 
 * def SWAP_136(cpu): # 136 SWAP (HL)             # <<<<<<<<<<<<<<
 *     t = ((cpu.mb.getitem(cpu.HL) & 0xF0) >> 4) | ((cpu.mb.getitem(cpu.HL) & 0x0F) << 4)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3376
 * 
 * 
 * def SWAP_137(cpu): # 137 SWAP A             # <<<<<<<<<<<<<<
 *     t = ((cpu.A & 0xF0) >> 4) | ((cpu.A & 0x0F) << 4)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SWAP_137(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3377
 * 
 * def SWAP_137(cpu): # 137 SWAP A
 *     t = ((cpu.A & 0xF0) >> 4) | ((cpu.A & 0x0F) << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->A & 0xF0) >> 4) | ((__pyx_v_cpu->A & 0x0F) << 4));

  /* "pyboy/core/opcodes.py":3378
 * def SWAP_137(cpu): # 137 SWAP A
 *     t = ((cpu.A & 0xF0) >> 4) | ((cpu.A & 0x0F) << 4)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3379
 *     t = ((cpu.A & 0xF0) >> 4) | ((cpu.A & 0x0F) << 4)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3380
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3381
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3382
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3383
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3384
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3385
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3386
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3376
 * 
 * 
 * def SWAP_137(cpu): # 137 SWAP A             # <<<<<<<<<<<<<<
 *     t = ((cpu.A & 0xF0) >> 4) | ((cpu.A & 0x0F) << 4)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3389
 * 
 * 
 * def SRL_138(cpu): # 138 SRL B             # <<<<<<<<<<<<<<
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_138(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3390
 * 
 * def SRL_138(cpu): # 138 SRL B
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->B >> 1) + ((__pyx_v_cpu->B & 1) << 8));

  /* "pyboy/core/opcodes.py":3391
 * def SRL_138(cpu): # 138 SRL B
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3392
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3393
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3394
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3395
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.B = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3396
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3397
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3398
 *     t &= 0xFF
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3399
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3400
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3389
 * 
 * 
 * def SRL_138(cpu): # 138 SRL B             # <<<<<<<<<<<<<<
 *     t = (cpu.B >> 1) + ((cpu.B & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3403
 * 
 * 
 * def SRL_139(cpu): # 139 SRL C             # <<<<<<<<<<<<<<
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_139(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3404
 * 
 * def SRL_139(cpu): # 139 SRL C
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->C >> 1) + ((__pyx_v_cpu->C & 1) << 8));

  /* "pyboy/core/opcodes.py":3405
 * def SRL_139(cpu): # 139 SRL C
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3406
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3407
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3408
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3409
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.C = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3410
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3411
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3412
 *     t &= 0xFF
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3413
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3414
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3403
 * 
 * 
 * def SRL_139(cpu): # 139 SRL C             # <<<<<<<<<<<<<<
 *     t = (cpu.C >> 1) + ((cpu.C & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3417
 * 
 * 
 * def SRL_13A(cpu): # 13A SRL D             # <<<<<<<<<<<<<<
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3418
 * 
 * def SRL_13A(cpu): # 13A SRL D
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->D >> 1) + ((__pyx_v_cpu->D & 1) << 8));

  /* "pyboy/core/opcodes.py":3419
 * def SRL_13A(cpu): # 13A SRL D
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3420
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3421
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3422
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3423
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.D = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3424
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3425
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3426
 *     t &= 0xFF
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3427
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3428
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3417
 * 
 * 
 * def SRL_13A(cpu): # 13A SRL D             # <<<<<<<<<<<<<<
 *     t = (cpu.D >> 1) + ((cpu.D & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3431
 * 
 * 
 * def SRL_13B(cpu): # 13B SRL E             # <<<<<<<<<<<<<<
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3432
 * 
 * def SRL_13B(cpu): # 13B SRL E
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->E >> 1) + ((__pyx_v_cpu->E & 1) << 8));

  /* "pyboy/core/opcodes.py":3433
 * def SRL_13B(cpu): # 13B SRL E
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3434
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3435
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3436
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3437
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.E = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3438
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3439
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3440
 *     t &= 0xFF
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3441
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3442
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3431
 * 
 * 
 * def SRL_13B(cpu): # 13B SRL E             # <<<<<<<<<<<<<<
 *     t = (cpu.E >> 1) + ((cpu.E & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3445
 * 
 * 
 * def SRL_13C(cpu): # 13C SRL H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3446
 * 
 * def SRL_13C(cpu): # 13C SRL H
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->HL >> 8) >> 1) + (((__pyx_v_cpu->HL >> 8) & 1) << 8));

  /* "pyboy/core/opcodes.py":3447
 * def SRL_13C(cpu): # 13C SRL H
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3448
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3449
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3450
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3451
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3452
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3453
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":3454
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3455
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3456
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3445
 * 
 * 
 * def SRL_13C(cpu): # 13C SRL H             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL >> 8) >> 1) + (((cpu.HL >> 8) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3459
 * 
 * 
 * def SRL_13D(cpu): # 13D SRL L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3460
 * 
 * def SRL_13D(cpu): # 13D SRL L
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((__pyx_v_cpu->HL & 0xFF) >> 1) + (((__pyx_v_cpu->HL & 0xFF) & 1) << 8));

  /* "pyboy/core/opcodes.py":3461
 * def SRL_13D(cpu): # 13D SRL L
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3462
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3463
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3464
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3465
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3466
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3467
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":3468
 *     t &= 0xFF
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3469
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3470
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3459
 * 
 * 
 * def SRL_13D(cpu): # 13D SRL L             # <<<<<<<<<<<<<<
 *     t = ((cpu.HL & 0xFF) >> 1) + (((cpu.HL & 0xFF) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3473
 * 
 * 
 * def SRL_13E(cpu): # 13E SRL (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3474
 * 
 * def SRL_13E(cpu): # 13E SRL (HL)
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) >> 1) + ((((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 1) << 8));

  /* "pyboy/core/opcodes.py":3475
 * def SRL_13E(cpu): # 13E SRL (HL)
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3476
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3477
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3478
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3479
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3480
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3481
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":3482
 *     t &= 0xFF
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3483
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3484
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3473
 * 
 * 
 * def SRL_13E(cpu): # 13E SRL (HL)             # <<<<<<<<<<<<<<
 *     t = (cpu.mb.getitem(cpu.HL) >> 1) + ((cpu.mb.getitem(cpu.HL) & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3487
 * 
 * 
 * def SRL_13F(cpu): # 13F SRL A             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SRL_13F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3488
 * 
 * def SRL_13F(cpu): # 13F SRL A
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 8)             # <<<<<<<<<<<<<<
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->A >> 1) + ((__pyx_v_cpu->A & 1) << 8));

  /* "pyboy/core/opcodes.py":3489
 * def SRL_13F(cpu): # 13F SRL A
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 */
  __pyx_v_flag = 0;

  /* "pyboy/core/opcodes.py":3490
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3491
 *     flag = 0b00000000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + ((__pyx_v_t > 0xFF) << __pyx_v_5pyboy_4core_7opcodes_FLAGC));

  /* "pyboy/core/opcodes.py":3492
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     t &= 0xFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 0);

  /* "pyboy/core/opcodes.py":3493
 *     flag += (t > 0xFF) << FLAGC
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     t &= 0xFF
 *     cpu.A = t
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3494
 *     cpu.F &= 0b00000000
 *     cpu.F |= flag
 *     t &= 0xFF             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_t & 0xFF);

  /* "pyboy/core/opcodes.py":3495
 *     cpu.F |= flag
 *     t &= 0xFF
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":3496
 *     t &= 0xFF
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3497
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3498
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3487
 * 
 * 
 * def SRL_13F(cpu): # 13F SRL A             # <<<<<<<<<<<<<<
 *     t = (cpu.A >> 1) + ((cpu.A & 1) << 8)
 *     flag = 0b00000000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3501
 * 
 * 
 * def BIT_140(cpu): # 140 BIT 0,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 0)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_140(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3502
 * 
 * def BIT_140(cpu): # 140 BIT 0,B
 *     t = cpu.B & (1 << 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B & 0x1);

  /* "pyboy/core/opcodes.py":3503
 * def BIT_140(cpu): # 140 BIT 0,B
 *     t = cpu.B & (1 << 0)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3504
 *     t = cpu.B & (1 << 0)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3505
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3506
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3507
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3508
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3509
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3501
 * 
 * 
 * def BIT_140(cpu): # 140 BIT 0,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 0)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3512
 * 
 * 
 * def BIT_141(cpu): # 141 BIT 0,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 0)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_141(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3513
 * 
 * def BIT_141(cpu): # 141 BIT 0,C
 *     t = cpu.C & (1 << 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C & 0x1);

  /* "pyboy/core/opcodes.py":3514
 * def BIT_141(cpu): # 141 BIT 0,C
 *     t = cpu.C & (1 << 0)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3515
 *     t = cpu.C & (1 << 0)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3516
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3517
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3518
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3519
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3520
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3512
 * 
 * 
 * def BIT_141(cpu): # 141 BIT 0,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 0)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3523
 * 
 * 
 * def BIT_142(cpu): # 142 BIT 0,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 0)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_142(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3524
 * 
 * def BIT_142(cpu): # 142 BIT 0,D
 *     t = cpu.D & (1 << 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D & 0x1);

  /* "pyboy/core/opcodes.py":3525
 * def BIT_142(cpu): # 142 BIT 0,D
 *     t = cpu.D & (1 << 0)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3526
 *     t = cpu.D & (1 << 0)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3527
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3528
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3529
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3530
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3531
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3523
 * 
 * 
 * def BIT_142(cpu): # 142 BIT 0,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 0)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3534
 * 
 * 
 * def BIT_143(cpu): # 143 BIT 0,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 0)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_143(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3535
 * 
 * def BIT_143(cpu): # 143 BIT 0,E
 *     t = cpu.E & (1 << 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E & 0x1);

  /* "pyboy/core/opcodes.py":3536
 * def BIT_143(cpu): # 143 BIT 0,E
 *     t = cpu.E & (1 << 0)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3537
 *     t = cpu.E & (1 << 0)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3538
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3539
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3540
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3541
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3542
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3534
 * 
 * 
 * def BIT_143(cpu): # 143 BIT 0,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 0)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3545
 * 
 * 
 * def BIT_144(cpu): # 144 BIT 0,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 0)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_144(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3546
 * 
 * def BIT_144(cpu): # 144 BIT 0,H
 *     t = (cpu.HL >> 8) & (1 << 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & 0x1);

  /* "pyboy/core/opcodes.py":3547
 * def BIT_144(cpu): # 144 BIT 0,H
 *     t = (cpu.HL >> 8) & (1 << 0)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3548
 *     t = (cpu.HL >> 8) & (1 << 0)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3549
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3550
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3551
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3552
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3553
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3545
 * 
 * 
 * def BIT_144(cpu): # 144 BIT 0,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 0)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3556
 * 
 * 
 * def BIT_145(cpu): # 145 BIT 0,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 0)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_145(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3557
 * 
 * def BIT_145(cpu): # 145 BIT 0,L
 *     t = (cpu.HL & 0xFF) & (1 << 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & 0x1);

  /* "pyboy/core/opcodes.py":3558
 * def BIT_145(cpu): # 145 BIT 0,L
 *     t = (cpu.HL & 0xFF) & (1 << 0)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3559
 *     t = (cpu.HL & 0xFF) & (1 << 0)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3560
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3561
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3562
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3563
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3564
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3556
 * 
 * 
 * def BIT_145(cpu): # 145 BIT 0,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 0)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3567
 * 
 * 
 * def BIT_146(cpu): # 146 BIT 0,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 0)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_146(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3568
 * 
 * def BIT_146(cpu): # 146 BIT 0,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x1);

  /* "pyboy/core/opcodes.py":3569
 * def BIT_146(cpu): # 146 BIT 0,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 0)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3570
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 0)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3571
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3572
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3573
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3574
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3575
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3567
 * 
 * 
 * def BIT_146(cpu): # 146 BIT 0,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 0)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3578
 * 
 * 
 * def BIT_147(cpu): # 147 BIT 0,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 0)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_147(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3579
 * 
 * def BIT_147(cpu): # 147 BIT 0,A
 *     t = cpu.A & (1 << 0)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & 0x1);

  /* "pyboy/core/opcodes.py":3580
 * def BIT_147(cpu): # 147 BIT 0,A
 *     t = cpu.A & (1 << 0)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3581
 *     t = cpu.A & (1 << 0)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3582
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3583
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3584
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3585
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3586
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3578
 * 
 * 
 * def BIT_147(cpu): # 147 BIT 0,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 0)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3589
 * 
 * 
 * def BIT_148(cpu): # 148 BIT 1,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 1)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_148(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3590
 * 
 * def BIT_148(cpu): # 148 BIT 1,B
 *     t = cpu.B & (1 << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B & 0x2);

  /* "pyboy/core/opcodes.py":3591
 * def BIT_148(cpu): # 148 BIT 1,B
 *     t = cpu.B & (1 << 1)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3592
 *     t = cpu.B & (1 << 1)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3593
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3594
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3595
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3596
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3597
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3589
 * 
 * 
 * def BIT_148(cpu): # 148 BIT 1,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 1)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3600
 * 
 * 
 * def BIT_149(cpu): # 149 BIT 1,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 1)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_149(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3601
 * 
 * def BIT_149(cpu): # 149 BIT 1,C
 *     t = cpu.C & (1 << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C & 0x2);

  /* "pyboy/core/opcodes.py":3602
 * def BIT_149(cpu): # 149 BIT 1,C
 *     t = cpu.C & (1 << 1)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3603
 *     t = cpu.C & (1 << 1)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3604
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3605
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3606
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3607
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3608
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3600
 * 
 * 
 * def BIT_149(cpu): # 149 BIT 1,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 1)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3611
 * 
 * 
 * def BIT_14A(cpu): # 14A BIT 1,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 1)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3612
 * 
 * def BIT_14A(cpu): # 14A BIT 1,D
 *     t = cpu.D & (1 << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D & 0x2);

  /* "pyboy/core/opcodes.py":3613
 * def BIT_14A(cpu): # 14A BIT 1,D
 *     t = cpu.D & (1 << 1)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3614
 *     t = cpu.D & (1 << 1)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3615
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3616
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3617
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3618
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3619
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3611
 * 
 * 
 * def BIT_14A(cpu): # 14A BIT 1,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 1)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3622
 * 
 * 
 * def BIT_14B(cpu): # 14B BIT 1,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 1)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3623
 * 
 * def BIT_14B(cpu): # 14B BIT 1,E
 *     t = cpu.E & (1 << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E & 0x2);

  /* "pyboy/core/opcodes.py":3624
 * def BIT_14B(cpu): # 14B BIT 1,E
 *     t = cpu.E & (1 << 1)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3625
 *     t = cpu.E & (1 << 1)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3626
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3627
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3628
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3629
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3630
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3622
 * 
 * 
 * def BIT_14B(cpu): # 14B BIT 1,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 1)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3633
 * 
 * 
 * def BIT_14C(cpu): # 14C BIT 1,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 1)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3634
 * 
 * def BIT_14C(cpu): # 14C BIT 1,H
 *     t = (cpu.HL >> 8) & (1 << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & 0x2);

  /* "pyboy/core/opcodes.py":3635
 * def BIT_14C(cpu): # 14C BIT 1,H
 *     t = (cpu.HL >> 8) & (1 << 1)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3636
 *     t = (cpu.HL >> 8) & (1 << 1)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3637
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3638
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3639
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3640
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3641
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3633
 * 
 * 
 * def BIT_14C(cpu): # 14C BIT 1,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 1)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3644
 * 
 * 
 * def BIT_14D(cpu): # 14D BIT 1,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 1)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3645
 * 
 * def BIT_14D(cpu): # 14D BIT 1,L
 *     t = (cpu.HL & 0xFF) & (1 << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & 0x2);

  /* "pyboy/core/opcodes.py":3646
 * def BIT_14D(cpu): # 14D BIT 1,L
 *     t = (cpu.HL & 0xFF) & (1 << 1)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3647
 *     t = (cpu.HL & 0xFF) & (1 << 1)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3648
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3649
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3650
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3651
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3652
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3644
 * 
 * 
 * def BIT_14D(cpu): # 14D BIT 1,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 1)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3655
 * 
 * 
 * def BIT_14E(cpu): # 14E BIT 1,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 1)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3656
 * 
 * def BIT_14E(cpu): # 14E BIT 1,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x2);

  /* "pyboy/core/opcodes.py":3657
 * def BIT_14E(cpu): # 14E BIT 1,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 1)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3658
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 1)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3659
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3660
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3661
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3662
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3663
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3655
 * 
 * 
 * def BIT_14E(cpu): # 14E BIT 1,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 1)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3666
 * 
 * 
 * def BIT_14F(cpu): # 14F BIT 1,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 1)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_14F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3667
 * 
 * def BIT_14F(cpu): # 14F BIT 1,A
 *     t = cpu.A & (1 << 1)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & 0x2);

  /* "pyboy/core/opcodes.py":3668
 * def BIT_14F(cpu): # 14F BIT 1,A
 *     t = cpu.A & (1 << 1)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3669
 *     t = cpu.A & (1 << 1)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3670
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3671
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3672
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3673
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3674
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3666
 * 
 * 
 * def BIT_14F(cpu): # 14F BIT 1,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 1)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3677
 * 
 * 
 * def BIT_150(cpu): # 150 BIT 2,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 2)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_150(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3678
 * 
 * def BIT_150(cpu): # 150 BIT 2,B
 *     t = cpu.B & (1 << 2)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B & 0x4);

  /* "pyboy/core/opcodes.py":3679
 * def BIT_150(cpu): # 150 BIT 2,B
 *     t = cpu.B & (1 << 2)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3680
 *     t = cpu.B & (1 << 2)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3681
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3682
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3683
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3684
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3685
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3677
 * 
 * 
 * def BIT_150(cpu): # 150 BIT 2,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 2)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3688
 * 
 * 
 * def BIT_151(cpu): # 151 BIT 2,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 2)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_151(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3689
 * 
 * def BIT_151(cpu): # 151 BIT 2,C
 *     t = cpu.C & (1 << 2)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C & 0x4);

  /* "pyboy/core/opcodes.py":3690
 * def BIT_151(cpu): # 151 BIT 2,C
 *     t = cpu.C & (1 << 2)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3691
 *     t = cpu.C & (1 << 2)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3692
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3693
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3694
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3695
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3696
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3688
 * 
 * 
 * def BIT_151(cpu): # 151 BIT 2,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 2)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3699
 * 
 * 
 * def BIT_152(cpu): # 152 BIT 2,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 2)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_152(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3700
 * 
 * def BIT_152(cpu): # 152 BIT 2,D
 *     t = cpu.D & (1 << 2)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D & 0x4);

  /* "pyboy/core/opcodes.py":3701
 * def BIT_152(cpu): # 152 BIT 2,D
 *     t = cpu.D & (1 << 2)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3702
 *     t = cpu.D & (1 << 2)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3703
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3704
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3705
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3706
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3707
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3699
 * 
 * 
 * def BIT_152(cpu): # 152 BIT 2,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 2)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3710
 * 
 * 
 * def BIT_153(cpu): # 153 BIT 2,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 2)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_153(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3711
 * 
 * def BIT_153(cpu): # 153 BIT 2,E
 *     t = cpu.E & (1 << 2)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E & 0x4);

  /* "pyboy/core/opcodes.py":3712
 * def BIT_153(cpu): # 153 BIT 2,E
 *     t = cpu.E & (1 << 2)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3713
 *     t = cpu.E & (1 << 2)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3714
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3715
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3716
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3717
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3718
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3710
 * 
 * 
 * def BIT_153(cpu): # 153 BIT 2,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 2)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3721
 * 
 * 
 * def BIT_154(cpu): # 154 BIT 2,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 2)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_154(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3722
 * 
 * def BIT_154(cpu): # 154 BIT 2,H
 *     t = (cpu.HL >> 8) & (1 << 2)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & 0x4);

  /* "pyboy/core/opcodes.py":3723
 * def BIT_154(cpu): # 154 BIT 2,H
 *     t = (cpu.HL >> 8) & (1 << 2)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3724
 *     t = (cpu.HL >> 8) & (1 << 2)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3725
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3726
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3727
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3728
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3729
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3721
 * 
 * 
 * def BIT_154(cpu): # 154 BIT 2,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 2)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3732
 * 
 * 
 * def BIT_155(cpu): # 155 BIT 2,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 2)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_155(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3733
 * 
 * def BIT_155(cpu): # 155 BIT 2,L
 *     t = (cpu.HL & 0xFF) & (1 << 2)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & 0x4);

  /* "pyboy/core/opcodes.py":3734
 * def BIT_155(cpu): # 155 BIT 2,L
 *     t = (cpu.HL & 0xFF) & (1 << 2)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3735
 *     t = (cpu.HL & 0xFF) & (1 << 2)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3736
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3737
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3738
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3739
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3740
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3732
 * 
 * 
 * def BIT_155(cpu): # 155 BIT 2,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 2)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3743
 * 
 * 
 * def BIT_156(cpu): # 156 BIT 2,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 2)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_156(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3744
 * 
 * def BIT_156(cpu): # 156 BIT 2,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 2)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x4);

  /* "pyboy/core/opcodes.py":3745
 * def BIT_156(cpu): # 156 BIT 2,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 2)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3746
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 2)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3747
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3748
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3749
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3750
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3751
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3743
 * 
 * 
 * def BIT_156(cpu): # 156 BIT 2,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 2)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3754
 * 
 * 
 * def BIT_157(cpu): # 157 BIT 2,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 2)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_157(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3755
 * 
 * def BIT_157(cpu): # 157 BIT 2,A
 *     t = cpu.A & (1 << 2)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & 0x4);

  /* "pyboy/core/opcodes.py":3756
 * def BIT_157(cpu): # 157 BIT 2,A
 *     t = cpu.A & (1 << 2)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3757
 *     t = cpu.A & (1 << 2)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3758
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3759
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3760
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3761
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3762
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3754
 * 
 * 
 * def BIT_157(cpu): # 157 BIT 2,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 2)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3765
 * 
 * 
 * def BIT_158(cpu): # 158 BIT 3,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 3)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_158(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3766
 * 
 * def BIT_158(cpu): # 158 BIT 3,B
 *     t = cpu.B & (1 << 3)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B & 0x8);

  /* "pyboy/core/opcodes.py":3767
 * def BIT_158(cpu): # 158 BIT 3,B
 *     t = cpu.B & (1 << 3)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3768
 *     t = cpu.B & (1 << 3)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3769
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3770
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3771
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3772
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3773
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3765
 * 
 * 
 * def BIT_158(cpu): # 158 BIT 3,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 3)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3776
 * 
 * 
 * def BIT_159(cpu): # 159 BIT 3,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 3)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_159(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3777
 * 
 * def BIT_159(cpu): # 159 BIT 3,C
 *     t = cpu.C & (1 << 3)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C & 0x8);

  /* "pyboy/core/opcodes.py":3778
 * def BIT_159(cpu): # 159 BIT 3,C
 *     t = cpu.C & (1 << 3)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3779
 *     t = cpu.C & (1 << 3)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3780
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3781
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3782
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3783
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3784
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3776
 * 
 * 
 * def BIT_159(cpu): # 159 BIT 3,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 3)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3787
 * 
 * 
 * def BIT_15A(cpu): # 15A BIT 3,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 3)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3788
 * 
 * def BIT_15A(cpu): # 15A BIT 3,D
 *     t = cpu.D & (1 << 3)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D & 0x8);

  /* "pyboy/core/opcodes.py":3789
 * def BIT_15A(cpu): # 15A BIT 3,D
 *     t = cpu.D & (1 << 3)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3790
 *     t = cpu.D & (1 << 3)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3791
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3792
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3793
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3794
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3795
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3787
 * 
 * 
 * def BIT_15A(cpu): # 15A BIT 3,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 3)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3798
 * 
 * 
 * def BIT_15B(cpu): # 15B BIT 3,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 3)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3799
 * 
 * def BIT_15B(cpu): # 15B BIT 3,E
 *     t = cpu.E & (1 << 3)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E & 0x8);

  /* "pyboy/core/opcodes.py":3800
 * def BIT_15B(cpu): # 15B BIT 3,E
 *     t = cpu.E & (1 << 3)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3801
 *     t = cpu.E & (1 << 3)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3802
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3803
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3804
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3805
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3806
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3798
 * 
 * 
 * def BIT_15B(cpu): # 15B BIT 3,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 3)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3809
 * 
 * 
 * def BIT_15C(cpu): # 15C BIT 3,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 3)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3810
 * 
 * def BIT_15C(cpu): # 15C BIT 3,H
 *     t = (cpu.HL >> 8) & (1 << 3)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & 0x8);

  /* "pyboy/core/opcodes.py":3811
 * def BIT_15C(cpu): # 15C BIT 3,H
 *     t = (cpu.HL >> 8) & (1 << 3)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3812
 *     t = (cpu.HL >> 8) & (1 << 3)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3813
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3814
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3815
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3816
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3817
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3809
 * 
 * 
 * def BIT_15C(cpu): # 15C BIT 3,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 3)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3820
 * 
 * 
 * def BIT_15D(cpu): # 15D BIT 3,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 3)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3821
 * 
 * def BIT_15D(cpu): # 15D BIT 3,L
 *     t = (cpu.HL & 0xFF) & (1 << 3)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & 0x8);

  /* "pyboy/core/opcodes.py":3822
 * def BIT_15D(cpu): # 15D BIT 3,L
 *     t = (cpu.HL & 0xFF) & (1 << 3)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3823
 *     t = (cpu.HL & 0xFF) & (1 << 3)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3824
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3825
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3826
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3827
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3828
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3820
 * 
 * 
 * def BIT_15D(cpu): # 15D BIT 3,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 3)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3831
 * 
 * 
 * def BIT_15E(cpu): # 15E BIT 3,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 3)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3832
 * 
 * def BIT_15E(cpu): # 15E BIT 3,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 3)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x8);

  /* "pyboy/core/opcodes.py":3833
 * def BIT_15E(cpu): # 15E BIT 3,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 3)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3834
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 3)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3835
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3836
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3837
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3838
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3839
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3831
 * 
 * 
 * def BIT_15E(cpu): # 15E BIT 3,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 3)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3842
 * 
 * 
 * def BIT_15F(cpu): # 15F BIT 3,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 3)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_15F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3843
 * 
 * def BIT_15F(cpu): # 15F BIT 3,A
 *     t = cpu.A & (1 << 3)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & 0x8);

  /* "pyboy/core/opcodes.py":3844
 * def BIT_15F(cpu): # 15F BIT 3,A
 *     t = cpu.A & (1 << 3)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3845
 *     t = cpu.A & (1 << 3)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3846
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3847
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3848
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3849
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3850
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3842
 * 
 * 
 * def BIT_15F(cpu): # 15F BIT 3,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 3)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3853
 * 
 * 
 * def BIT_160(cpu): # 160 BIT 4,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 4)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_160(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3854
 * 
 * def BIT_160(cpu): # 160 BIT 4,B
 *     t = cpu.B & (1 << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B & 0x10);

  /* "pyboy/core/opcodes.py":3855
 * def BIT_160(cpu): # 160 BIT 4,B
 *     t = cpu.B & (1 << 4)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3856
 *     t = cpu.B & (1 << 4)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3857
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3858
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3859
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3860
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3861
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3853
 * 
 * 
 * def BIT_160(cpu): # 160 BIT 4,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 4)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3864
 * 
 * 
 * def BIT_161(cpu): # 161 BIT 4,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 4)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_161(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3865
 * 
 * def BIT_161(cpu): # 161 BIT 4,C
 *     t = cpu.C & (1 << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C & 0x10);

  /* "pyboy/core/opcodes.py":3866
 * def BIT_161(cpu): # 161 BIT 4,C
 *     t = cpu.C & (1 << 4)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3867
 *     t = cpu.C & (1 << 4)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3868
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3869
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3870
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3871
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3872
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3864
 * 
 * 
 * def BIT_161(cpu): # 161 BIT 4,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 4)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3875
 * 
 * 
 * def BIT_162(cpu): # 162 BIT 4,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 4)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_162(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3876
 * 
 * def BIT_162(cpu): # 162 BIT 4,D
 *     t = cpu.D & (1 << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D & 0x10);

  /* "pyboy/core/opcodes.py":3877
 * def BIT_162(cpu): # 162 BIT 4,D
 *     t = cpu.D & (1 << 4)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3878
 *     t = cpu.D & (1 << 4)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3879
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3880
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3881
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3882
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3883
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3875
 * 
 * 
 * def BIT_162(cpu): # 162 BIT 4,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 4)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3886
 * 
 * 
 * def BIT_163(cpu): # 163 BIT 4,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 4)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_163(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3887
 * 
 * def BIT_163(cpu): # 163 BIT 4,E
 *     t = cpu.E & (1 << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E & 0x10);

  /* "pyboy/core/opcodes.py":3888
 * def BIT_163(cpu): # 163 BIT 4,E
 *     t = cpu.E & (1 << 4)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3889
 *     t = cpu.E & (1 << 4)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3890
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3891
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3892
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3893
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3894
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3886
 * 
 * 
 * def BIT_163(cpu): # 163 BIT 4,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 4)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3897
 * 
 * 
 * def BIT_164(cpu): # 164 BIT 4,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 4)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_164(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3898
 * 
 * def BIT_164(cpu): # 164 BIT 4,H
 *     t = (cpu.HL >> 8) & (1 << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & 0x10);

  /* "pyboy/core/opcodes.py":3899
 * def BIT_164(cpu): # 164 BIT 4,H
 *     t = (cpu.HL >> 8) & (1 << 4)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3900
 *     t = (cpu.HL >> 8) & (1 << 4)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3901
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3902
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3903
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3904
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3905
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3897
 * 
 * 
 * def BIT_164(cpu): # 164 BIT 4,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 4)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3908
 * 
 * 
 * def BIT_165(cpu): # 165 BIT 4,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 4)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_165(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3909
 * 
 * def BIT_165(cpu): # 165 BIT 4,L
 *     t = (cpu.HL & 0xFF) & (1 << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & 0x10);

  /* "pyboy/core/opcodes.py":3910
 * def BIT_165(cpu): # 165 BIT 4,L
 *     t = (cpu.HL & 0xFF) & (1 << 4)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3911
 *     t = (cpu.HL & 0xFF) & (1 << 4)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3912
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3913
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3914
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3915
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3916
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3908
 * 
 * 
 * def BIT_165(cpu): # 165 BIT 4,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 4)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3919
 * 
 * 
 * def BIT_166(cpu): # 166 BIT 4,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 4)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_166(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3920
 * 
 * def BIT_166(cpu): # 166 BIT 4,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x10);

  /* "pyboy/core/opcodes.py":3921
 * def BIT_166(cpu): # 166 BIT 4,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 4)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3922
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 4)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3923
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3924
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3925
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3926
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3927
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3919
 * 
 * 
 * def BIT_166(cpu): # 166 BIT 4,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 4)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3930
 * 
 * 
 * def BIT_167(cpu): # 167 BIT 4,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 4)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_167(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3931
 * 
 * def BIT_167(cpu): # 167 BIT 4,A
 *     t = cpu.A & (1 << 4)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & 0x10);

  /* "pyboy/core/opcodes.py":3932
 * def BIT_167(cpu): # 167 BIT 4,A
 *     t = cpu.A & (1 << 4)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3933
 *     t = cpu.A & (1 << 4)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3934
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3935
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3936
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3937
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3938
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3930
 * 
 * 
 * def BIT_167(cpu): # 167 BIT 4,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 4)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3941
 * 
 * 
 * def BIT_168(cpu): # 168 BIT 5,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 5)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_168(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3942
 * 
 * def BIT_168(cpu): # 168 BIT 5,B
 *     t = cpu.B & (1 << 5)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B & 0x20);

  /* "pyboy/core/opcodes.py":3943
 * def BIT_168(cpu): # 168 BIT 5,B
 *     t = cpu.B & (1 << 5)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3944
 *     t = cpu.B & (1 << 5)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3945
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3946
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3947
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3948
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3949
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3941
 * 
 * 
 * def BIT_168(cpu): # 168 BIT 5,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 5)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3952
 * 
 * 
 * def BIT_169(cpu): # 169 BIT 5,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 5)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_169(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3953
 * 
 * def BIT_169(cpu): # 169 BIT 5,C
 *     t = cpu.C & (1 << 5)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C & 0x20);

  /* "pyboy/core/opcodes.py":3954
 * def BIT_169(cpu): # 169 BIT 5,C
 *     t = cpu.C & (1 << 5)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3955
 *     t = cpu.C & (1 << 5)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3956
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3957
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3958
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3959
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3960
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3952
 * 
 * 
 * def BIT_169(cpu): # 169 BIT 5,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 5)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3963
 * 
 * 
 * def BIT_16A(cpu): # 16A BIT 5,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 5)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3964
 * 
 * def BIT_16A(cpu): # 16A BIT 5,D
 *     t = cpu.D & (1 << 5)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D & 0x20);

  /* "pyboy/core/opcodes.py":3965
 * def BIT_16A(cpu): # 16A BIT 5,D
 *     t = cpu.D & (1 << 5)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3966
 *     t = cpu.D & (1 << 5)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3967
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3968
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3969
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3970
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3971
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3963
 * 
 * 
 * def BIT_16A(cpu): # 16A BIT 5,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 5)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3974
 * 
 * 
 * def BIT_16B(cpu): # 16B BIT 5,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 5)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3975
 * 
 * def BIT_16B(cpu): # 16B BIT 5,E
 *     t = cpu.E & (1 << 5)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E & 0x20);

  /* "pyboy/core/opcodes.py":3976
 * def BIT_16B(cpu): # 16B BIT 5,E
 *     t = cpu.E & (1 << 5)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3977
 *     t = cpu.E & (1 << 5)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3978
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3979
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3980
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3981
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3982
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3974
 * 
 * 
 * def BIT_16B(cpu): # 16B BIT 5,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 5)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3985
 * 
 * 
 * def BIT_16C(cpu): # 16C BIT 5,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 5)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3986
 * 
 * def BIT_16C(cpu): # 16C BIT 5,H
 *     t = (cpu.HL >> 8) & (1 << 5)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & 0x20);

  /* "pyboy/core/opcodes.py":3987
 * def BIT_16C(cpu): # 16C BIT 5,H
 *     t = (cpu.HL >> 8) & (1 << 5)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3988
 *     t = (cpu.HL >> 8) & (1 << 5)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":3989
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":3990
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":3991
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":3992
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":3993
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3985
 * 
 * 
 * def BIT_16C(cpu): # 16C BIT 5,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 5)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":3996
 * 
 * 
 * def BIT_16D(cpu): # 16D BIT 5,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 5)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":3997
 * 
 * def BIT_16D(cpu): # 16D BIT 5,L
 *     t = (cpu.HL & 0xFF) & (1 << 5)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & 0x20);

  /* "pyboy/core/opcodes.py":3998
 * def BIT_16D(cpu): # 16D BIT 5,L
 *     t = (cpu.HL & 0xFF) & (1 << 5)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":3999
 *     t = (cpu.HL & 0xFF) & (1 << 5)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4000
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4001
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4002
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4003
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4004
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":3996
 * 
 * 
 * def BIT_16D(cpu): # 16D BIT 5,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 5)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4007
 * 
 * 
 * def BIT_16E(cpu): # 16E BIT 5,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 5)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4008
 * 
 * def BIT_16E(cpu): # 16E BIT 5,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 5)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x20);

  /* "pyboy/core/opcodes.py":4009
 * def BIT_16E(cpu): # 16E BIT 5,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 5)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4010
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 5)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4011
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4012
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4013
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4014
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4015
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4007
 * 
 * 
 * def BIT_16E(cpu): # 16E BIT 5,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 5)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4018
 * 
 * 
 * def BIT_16F(cpu): # 16F BIT 5,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 5)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_16F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4019
 * 
 * def BIT_16F(cpu): # 16F BIT 5,A
 *     t = cpu.A & (1 << 5)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & 0x20);

  /* "pyboy/core/opcodes.py":4020
 * def BIT_16F(cpu): # 16F BIT 5,A
 *     t = cpu.A & (1 << 5)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4021
 *     t = cpu.A & (1 << 5)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4022
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4023
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4024
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4025
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4026
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4018
 * 
 * 
 * def BIT_16F(cpu): # 16F BIT 5,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 5)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4029
 * 
 * 
 * def BIT_170(cpu): # 170 BIT 6,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 6)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_170(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4030
 * 
 * def BIT_170(cpu): # 170 BIT 6,B
 *     t = cpu.B & (1 << 6)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B & 0x40);

  /* "pyboy/core/opcodes.py":4031
 * def BIT_170(cpu): # 170 BIT 6,B
 *     t = cpu.B & (1 << 6)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4032
 *     t = cpu.B & (1 << 6)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4033
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4034
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4035
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4036
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4037
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4029
 * 
 * 
 * def BIT_170(cpu): # 170 BIT 6,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 6)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4040
 * 
 * 
 * def BIT_171(cpu): # 171 BIT 6,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 6)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_171(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4041
 * 
 * def BIT_171(cpu): # 171 BIT 6,C
 *     t = cpu.C & (1 << 6)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C & 0x40);

  /* "pyboy/core/opcodes.py":4042
 * def BIT_171(cpu): # 171 BIT 6,C
 *     t = cpu.C & (1 << 6)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4043
 *     t = cpu.C & (1 << 6)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4044
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4045
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4046
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4047
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4048
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4040
 * 
 * 
 * def BIT_171(cpu): # 171 BIT 6,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 6)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4051
 * 
 * 
 * def BIT_172(cpu): # 172 BIT 6,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 6)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_172(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4052
 * 
 * def BIT_172(cpu): # 172 BIT 6,D
 *     t = cpu.D & (1 << 6)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D & 0x40);

  /* "pyboy/core/opcodes.py":4053
 * def BIT_172(cpu): # 172 BIT 6,D
 *     t = cpu.D & (1 << 6)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4054
 *     t = cpu.D & (1 << 6)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4055
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4056
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4057
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4058
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4059
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4051
 * 
 * 
 * def BIT_172(cpu): # 172 BIT 6,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 6)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4062
 * 
 * 
 * def BIT_173(cpu): # 173 BIT 6,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 6)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_173(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4063
 * 
 * def BIT_173(cpu): # 173 BIT 6,E
 *     t = cpu.E & (1 << 6)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E & 0x40);

  /* "pyboy/core/opcodes.py":4064
 * def BIT_173(cpu): # 173 BIT 6,E
 *     t = cpu.E & (1 << 6)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4065
 *     t = cpu.E & (1 << 6)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4066
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4067
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4068
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4069
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4070
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4062
 * 
 * 
 * def BIT_173(cpu): # 173 BIT 6,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 6)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4073
 * 
 * 
 * def BIT_174(cpu): # 174 BIT 6,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 6)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_174(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4074
 * 
 * def BIT_174(cpu): # 174 BIT 6,H
 *     t = (cpu.HL >> 8) & (1 << 6)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & 0x40);

  /* "pyboy/core/opcodes.py":4075
 * def BIT_174(cpu): # 174 BIT 6,H
 *     t = (cpu.HL >> 8) & (1 << 6)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4076
 *     t = (cpu.HL >> 8) & (1 << 6)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4077
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4078
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4079
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4080
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4081
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4073
 * 
 * 
 * def BIT_174(cpu): # 174 BIT 6,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 6)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4084
 * 
 * 
 * def BIT_175(cpu): # 175 BIT 6,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 6)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_175(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4085
 * 
 * def BIT_175(cpu): # 175 BIT 6,L
 *     t = (cpu.HL & 0xFF) & (1 << 6)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & 0x40);

  /* "pyboy/core/opcodes.py":4086
 * def BIT_175(cpu): # 175 BIT 6,L
 *     t = (cpu.HL & 0xFF) & (1 << 6)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4087
 *     t = (cpu.HL & 0xFF) & (1 << 6)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4088
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4089
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4090
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4091
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4092
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4084
 * 
 * 
 * def BIT_175(cpu): # 175 BIT 6,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 6)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4095
 * 
 * 
 * def BIT_176(cpu): # 176 BIT 6,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 6)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_176(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4096
 * 
 * def BIT_176(cpu): # 176 BIT 6,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 6)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x40);

  /* "pyboy/core/opcodes.py":4097
 * def BIT_176(cpu): # 176 BIT 6,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 6)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4098
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 6)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4099
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4100
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4101
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4102
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4103
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4095
 * 
 * 
 * def BIT_176(cpu): # 176 BIT 6,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 6)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4106
 * 
 * 
 * def BIT_177(cpu): # 177 BIT 6,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 6)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_177(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4107
 * 
 * def BIT_177(cpu): # 177 BIT 6,A
 *     t = cpu.A & (1 << 6)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & 0x40);

  /* "pyboy/core/opcodes.py":4108
 * def BIT_177(cpu): # 177 BIT 6,A
 *     t = cpu.A & (1 << 6)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4109
 *     t = cpu.A & (1 << 6)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4110
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4111
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4112
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4113
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4114
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4106
 * 
 * 
 * def BIT_177(cpu): # 177 BIT 6,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 6)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4117
 * 
 * 
 * def BIT_178(cpu): # 178 BIT 7,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 7)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_178(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4118
 * 
 * def BIT_178(cpu): # 178 BIT 7,B
 *     t = cpu.B & (1 << 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->B & 0x80);

  /* "pyboy/core/opcodes.py":4119
 * def BIT_178(cpu): # 178 BIT 7,B
 *     t = cpu.B & (1 << 7)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4120
 *     t = cpu.B & (1 << 7)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4121
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4122
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4123
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4124
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4125
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4117
 * 
 * 
 * def BIT_178(cpu): # 178 BIT 7,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & (1 << 7)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4128
 * 
 * 
 * def BIT_179(cpu): # 179 BIT 7,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 7)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_179(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4129
 * 
 * def BIT_179(cpu): # 179 BIT 7,C
 *     t = cpu.C & (1 << 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->C & 0x80);

  /* "pyboy/core/opcodes.py":4130
 * def BIT_179(cpu): # 179 BIT 7,C
 *     t = cpu.C & (1 << 7)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4131
 *     t = cpu.C & (1 << 7)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4132
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4133
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4134
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4135
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4136
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4128
 * 
 * 
 * def BIT_179(cpu): # 179 BIT 7,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & (1 << 7)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4139
 * 
 * 
 * def BIT_17A(cpu): # 17A BIT 7,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 7)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4140
 * 
 * def BIT_17A(cpu): # 17A BIT 7,D
 *     t = cpu.D & (1 << 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->D & 0x80);

  /* "pyboy/core/opcodes.py":4141
 * def BIT_17A(cpu): # 17A BIT 7,D
 *     t = cpu.D & (1 << 7)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4142
 *     t = cpu.D & (1 << 7)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4143
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4144
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4145
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4146
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4147
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4139
 * 
 * 
 * def BIT_17A(cpu): # 17A BIT 7,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & (1 << 7)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4150
 * 
 * 
 * def BIT_17B(cpu): # 17B BIT 7,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 7)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4151
 * 
 * def BIT_17B(cpu): # 17B BIT 7,E
 *     t = cpu.E & (1 << 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->E & 0x80);

  /* "pyboy/core/opcodes.py":4152
 * def BIT_17B(cpu): # 17B BIT 7,E
 *     t = cpu.E & (1 << 7)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4153
 *     t = cpu.E & (1 << 7)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4154
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4155
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4156
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4157
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4158
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4150
 * 
 * 
 * def BIT_17B(cpu): # 17B BIT 7,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & (1 << 7)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4161
 * 
 * 
 * def BIT_17C(cpu): # 17C BIT 7,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 7)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4162
 * 
 * def BIT_17C(cpu): # 17C BIT 7,H
 *     t = (cpu.HL >> 8) & (1 << 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & 0x80);

  /* "pyboy/core/opcodes.py":4163
 * def BIT_17C(cpu): # 17C BIT 7,H
 *     t = (cpu.HL >> 8) & (1 << 7)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4164
 *     t = (cpu.HL >> 8) & (1 << 7)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4165
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4166
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4167
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4168
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4169
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4161
 * 
 * 
 * def BIT_17C(cpu): # 17C BIT 7,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & (1 << 7)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4172
 * 
 * 
 * def BIT_17D(cpu): # 17D BIT 7,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 7)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4173
 * 
 * def BIT_17D(cpu): # 17D BIT 7,L
 *     t = (cpu.HL & 0xFF) & (1 << 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & 0x80);

  /* "pyboy/core/opcodes.py":4174
 * def BIT_17D(cpu): # 17D BIT 7,L
 *     t = (cpu.HL & 0xFF) & (1 << 7)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4175
 *     t = (cpu.HL & 0xFF) & (1 << 7)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4176
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4177
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4178
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4179
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4180
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4172
 * 
 * 
 * def BIT_17D(cpu): # 17D BIT 7,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & (1 << 7)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4183
 * 
 * 
 * def BIT_17E(cpu): # 17E BIT 7,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 7)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4184
 * 
 * def BIT_17E(cpu): # 17E BIT 7,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & 0x80);

  /* "pyboy/core/opcodes.py":4185
 * def BIT_17E(cpu): # 17E BIT 7,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 7)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4186
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 7)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4187
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4188
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4189
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4190
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4191
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4183
 * 
 * 
 * def BIT_17E(cpu): # 17E BIT 7,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & (1 << 7)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4194
 * 
 * 
 * def BIT_17F(cpu): # 17F BIT 7,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 7)
 *     flag = 0b00100000
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_BIT_17F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_v_flag;
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4195
 * 
 * def BIT_17F(cpu): # 17F BIT 7,A
 *     t = cpu.A & (1 << 7)             # <<<<<<<<<<<<<<
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 */
  __pyx_v_t = (__pyx_v_cpu->A & 0x80);

  /* "pyboy/core/opcodes.py":4196
 * def BIT_17F(cpu): # 17F BIT 7,A
 *     t = cpu.A & (1 << 7)
 *     flag = 0b00100000             # <<<<<<<<<<<<<<
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 */
  __pyx_v_flag = 32;

  /* "pyboy/core/opcodes.py":4197
 *     t = cpu.A & (1 << 7)
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ             # <<<<<<<<<<<<<<
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 */
  __pyx_v_flag = (__pyx_v_flag + (((__pyx_v_t & 0xFF) == 0) << __pyx_v_5pyboy_4core_7opcodes_FLAGZ));

  /* "pyboy/core/opcodes.py":4198
 *     flag = 0b00100000
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000             # <<<<<<<<<<<<<<
 *     cpu.F |= flag
 *     cpu.PC += 2
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F & 16);

  /* "pyboy/core/opcodes.py":4199
 *     flag += ((t & 0xFF) == 0) << FLAGZ
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->F = (__pyx_v_cpu->F | __pyx_v_flag);

  /* "pyboy/core/opcodes.py":4200
 *     cpu.F &= 0b00010000
 *     cpu.F |= flag
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4201
 *     cpu.F |= flag
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4202
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4194
 * 
 * 
 * def BIT_17F(cpu): # 17F BIT 7,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & (1 << 7)
 *     flag = 0b00100000
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4205
 * 
 * 
 * def RES_180(cpu): # 180 RES 0,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 0)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_180(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4206
 * 
 * def RES_180(cpu): # 180 RES 0,B
 *     t = cpu.B & ~(1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B & (~0x1));

  /* "pyboy/core/opcodes.py":4207
 * def RES_180(cpu): # 180 RES 0,B
 *     t = cpu.B & ~(1 << 0)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4208
 *     t = cpu.B & ~(1 << 0)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4209
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4210
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4205
 * 
 * 
 * def RES_180(cpu): # 180 RES 0,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 0)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4213
 * 
 * 
 * def RES_181(cpu): # 181 RES 0,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 0)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_181(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4214
 * 
 * def RES_181(cpu): # 181 RES 0,C
 *     t = cpu.C & ~(1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C & (~0x1));

  /* "pyboy/core/opcodes.py":4215
 * def RES_181(cpu): # 181 RES 0,C
 *     t = cpu.C & ~(1 << 0)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4216
 *     t = cpu.C & ~(1 << 0)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4217
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4218
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4213
 * 
 * 
 * def RES_181(cpu): # 181 RES 0,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 0)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4221
 * 
 * 
 * def RES_182(cpu): # 182 RES 0,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 0)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_182(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4222
 * 
 * def RES_182(cpu): # 182 RES 0,D
 *     t = cpu.D & ~(1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D & (~0x1));

  /* "pyboy/core/opcodes.py":4223
 * def RES_182(cpu): # 182 RES 0,D
 *     t = cpu.D & ~(1 << 0)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4224
 *     t = cpu.D & ~(1 << 0)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4225
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4226
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4221
 * 
 * 
 * def RES_182(cpu): # 182 RES 0,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 0)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4229
 * 
 * 
 * def RES_183(cpu): # 183 RES 0,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 0)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_183(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4230
 * 
 * def RES_183(cpu): # 183 RES 0,E
 *     t = cpu.E & ~(1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E & (~0x1));

  /* "pyboy/core/opcodes.py":4231
 * def RES_183(cpu): # 183 RES 0,E
 *     t = cpu.E & ~(1 << 0)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4232
 *     t = cpu.E & ~(1 << 0)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4233
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4234
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4229
 * 
 * 
 * def RES_183(cpu): # 183 RES 0,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 0)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4237
 * 
 * 
 * def RES_184(cpu): # 184 RES 0,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 0)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_184(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4238
 * 
 * def RES_184(cpu): # 184 RES 0,H
 *     t = (cpu.HL >> 8) & ~(1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & (~0x1));

  /* "pyboy/core/opcodes.py":4239
 * def RES_184(cpu): # 184 RES 0,H
 *     t = (cpu.HL >> 8) & ~(1 << 0)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4240
 *     t = (cpu.HL >> 8) & ~(1 << 0)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4241
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4242
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4237
 * 
 * 
 * def RES_184(cpu): # 184 RES 0,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 0)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4245
 * 
 * 
 * def RES_185(cpu): # 185 RES 0,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 0)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_185(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4246
 * 
 * def RES_185(cpu): # 185 RES 0,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & (~0x1));

  /* "pyboy/core/opcodes.py":4247
 * def RES_185(cpu): # 185 RES 0,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 0)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4248
 *     t = (cpu.HL & 0xFF) & ~(1 << 0)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4249
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4250
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4245
 * 
 * 
 * def RES_185(cpu): # 185 RES 0,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 0)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4253
 * 
 * 
 * def RES_186(cpu): # 186 RES 0,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 0)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_186(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4254
 * 
 * def RES_186(cpu): # 186 RES 0,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & (~0x1));

  /* "pyboy/core/opcodes.py":4255
 * def RES_186(cpu): # 186 RES 0,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 0)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4256
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 0)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4257
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4258
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4253
 * 
 * 
 * def RES_186(cpu): # 186 RES 0,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 0)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4261
 * 
 * 
 * def RES_187(cpu): # 187 RES 0,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 0)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_187(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4262
 * 
 * def RES_187(cpu): # 187 RES 0,A
 *     t = cpu.A & ~(1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A & (~0x1));

  /* "pyboy/core/opcodes.py":4263
 * def RES_187(cpu): # 187 RES 0,A
 *     t = cpu.A & ~(1 << 0)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4264
 *     t = cpu.A & ~(1 << 0)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4265
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4266
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4261
 * 
 * 
 * def RES_187(cpu): # 187 RES 0,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 0)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4269
 * 
 * 
 * def RES_188(cpu): # 188 RES 1,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 1)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_188(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4270
 * 
 * def RES_188(cpu): # 188 RES 1,B
 *     t = cpu.B & ~(1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B & (~0x2));

  /* "pyboy/core/opcodes.py":4271
 * def RES_188(cpu): # 188 RES 1,B
 *     t = cpu.B & ~(1 << 1)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4272
 *     t = cpu.B & ~(1 << 1)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4273
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4274
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4269
 * 
 * 
 * def RES_188(cpu): # 188 RES 1,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 1)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4277
 * 
 * 
 * def RES_189(cpu): # 189 RES 1,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 1)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_189(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4278
 * 
 * def RES_189(cpu): # 189 RES 1,C
 *     t = cpu.C & ~(1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C & (~0x2));

  /* "pyboy/core/opcodes.py":4279
 * def RES_189(cpu): # 189 RES 1,C
 *     t = cpu.C & ~(1 << 1)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4280
 *     t = cpu.C & ~(1 << 1)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4281
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4282
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4277
 * 
 * 
 * def RES_189(cpu): # 189 RES 1,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 1)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4285
 * 
 * 
 * def RES_18A(cpu): # 18A RES 1,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 1)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4286
 * 
 * def RES_18A(cpu): # 18A RES 1,D
 *     t = cpu.D & ~(1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D & (~0x2));

  /* "pyboy/core/opcodes.py":4287
 * def RES_18A(cpu): # 18A RES 1,D
 *     t = cpu.D & ~(1 << 1)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4288
 *     t = cpu.D & ~(1 << 1)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4289
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4290
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4285
 * 
 * 
 * def RES_18A(cpu): # 18A RES 1,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 1)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4293
 * 
 * 
 * def RES_18B(cpu): # 18B RES 1,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 1)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4294
 * 
 * def RES_18B(cpu): # 18B RES 1,E
 *     t = cpu.E & ~(1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E & (~0x2));

  /* "pyboy/core/opcodes.py":4295
 * def RES_18B(cpu): # 18B RES 1,E
 *     t = cpu.E & ~(1 << 1)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4296
 *     t = cpu.E & ~(1 << 1)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4297
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4298
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4293
 * 
 * 
 * def RES_18B(cpu): # 18B RES 1,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 1)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4301
 * 
 * 
 * def RES_18C(cpu): # 18C RES 1,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 1)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4302
 * 
 * def RES_18C(cpu): # 18C RES 1,H
 *     t = (cpu.HL >> 8) & ~(1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & (~0x2));

  /* "pyboy/core/opcodes.py":4303
 * def RES_18C(cpu): # 18C RES 1,H
 *     t = (cpu.HL >> 8) & ~(1 << 1)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4304
 *     t = (cpu.HL >> 8) & ~(1 << 1)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4305
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4306
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4301
 * 
 * 
 * def RES_18C(cpu): # 18C RES 1,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 1)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4309
 * 
 * 
 * def RES_18D(cpu): # 18D RES 1,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 1)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4310
 * 
 * def RES_18D(cpu): # 18D RES 1,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & (~0x2));

  /* "pyboy/core/opcodes.py":4311
 * def RES_18D(cpu): # 18D RES 1,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 1)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4312
 *     t = (cpu.HL & 0xFF) & ~(1 << 1)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4313
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4314
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4309
 * 
 * 
 * def RES_18D(cpu): # 18D RES 1,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 1)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4317
 * 
 * 
 * def RES_18E(cpu): # 18E RES 1,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 1)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4318
 * 
 * def RES_18E(cpu): # 18E RES 1,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & (~0x2));

  /* "pyboy/core/opcodes.py":4319
 * def RES_18E(cpu): # 18E RES 1,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 1)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4320
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 1)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4321
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4322
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4317
 * 
 * 
 * def RES_18E(cpu): # 18E RES 1,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 1)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4325
 * 
 * 
 * def RES_18F(cpu): # 18F RES 1,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 1)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_18F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4326
 * 
 * def RES_18F(cpu): # 18F RES 1,A
 *     t = cpu.A & ~(1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A & (~0x2));

  /* "pyboy/core/opcodes.py":4327
 * def RES_18F(cpu): # 18F RES 1,A
 *     t = cpu.A & ~(1 << 1)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4328
 *     t = cpu.A & ~(1 << 1)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4329
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4330
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4325
 * 
 * 
 * def RES_18F(cpu): # 18F RES 1,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 1)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4333
 * 
 * 
 * def RES_190(cpu): # 190 RES 2,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 2)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_190(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4334
 * 
 * def RES_190(cpu): # 190 RES 2,B
 *     t = cpu.B & ~(1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B & (~0x4));

  /* "pyboy/core/opcodes.py":4335
 * def RES_190(cpu): # 190 RES 2,B
 *     t = cpu.B & ~(1 << 2)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4336
 *     t = cpu.B & ~(1 << 2)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4337
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4338
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4333
 * 
 * 
 * def RES_190(cpu): # 190 RES 2,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 2)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4341
 * 
 * 
 * def RES_191(cpu): # 191 RES 2,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 2)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_191(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4342
 * 
 * def RES_191(cpu): # 191 RES 2,C
 *     t = cpu.C & ~(1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C & (~0x4));

  /* "pyboy/core/opcodes.py":4343
 * def RES_191(cpu): # 191 RES 2,C
 *     t = cpu.C & ~(1 << 2)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4344
 *     t = cpu.C & ~(1 << 2)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4345
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4346
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4341
 * 
 * 
 * def RES_191(cpu): # 191 RES 2,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 2)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4349
 * 
 * 
 * def RES_192(cpu): # 192 RES 2,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 2)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_192(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4350
 * 
 * def RES_192(cpu): # 192 RES 2,D
 *     t = cpu.D & ~(1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D & (~0x4));

  /* "pyboy/core/opcodes.py":4351
 * def RES_192(cpu): # 192 RES 2,D
 *     t = cpu.D & ~(1 << 2)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4352
 *     t = cpu.D & ~(1 << 2)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4353
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4354
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4349
 * 
 * 
 * def RES_192(cpu): # 192 RES 2,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 2)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4357
 * 
 * 
 * def RES_193(cpu): # 193 RES 2,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 2)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_193(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4358
 * 
 * def RES_193(cpu): # 193 RES 2,E
 *     t = cpu.E & ~(1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E & (~0x4));

  /* "pyboy/core/opcodes.py":4359
 * def RES_193(cpu): # 193 RES 2,E
 *     t = cpu.E & ~(1 << 2)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4360
 *     t = cpu.E & ~(1 << 2)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4361
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4362
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4357
 * 
 * 
 * def RES_193(cpu): # 193 RES 2,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 2)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4365
 * 
 * 
 * def RES_194(cpu): # 194 RES 2,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 2)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_194(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4366
 * 
 * def RES_194(cpu): # 194 RES 2,H
 *     t = (cpu.HL >> 8) & ~(1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & (~0x4));

  /* "pyboy/core/opcodes.py":4367
 * def RES_194(cpu): # 194 RES 2,H
 *     t = (cpu.HL >> 8) & ~(1 << 2)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4368
 *     t = (cpu.HL >> 8) & ~(1 << 2)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4369
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4370
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4365
 * 
 * 
 * def RES_194(cpu): # 194 RES 2,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 2)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4373
 * 
 * 
 * def RES_195(cpu): # 195 RES 2,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 2)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_195(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4374
 * 
 * def RES_195(cpu): # 195 RES 2,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & (~0x4));

  /* "pyboy/core/opcodes.py":4375
 * def RES_195(cpu): # 195 RES 2,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 2)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4376
 *     t = (cpu.HL & 0xFF) & ~(1 << 2)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4377
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4378
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4373
 * 
 * 
 * def RES_195(cpu): # 195 RES 2,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 2)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4381
 * 
 * 
 * def RES_196(cpu): # 196 RES 2,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 2)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_196(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4382
 * 
 * def RES_196(cpu): # 196 RES 2,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & (~0x4));

  /* "pyboy/core/opcodes.py":4383
 * def RES_196(cpu): # 196 RES 2,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 2)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4384
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 2)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4385
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4386
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4381
 * 
 * 
 * def RES_196(cpu): # 196 RES 2,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 2)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4389
 * 
 * 
 * def RES_197(cpu): # 197 RES 2,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 2)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_197(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4390
 * 
 * def RES_197(cpu): # 197 RES 2,A
 *     t = cpu.A & ~(1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A & (~0x4));

  /* "pyboy/core/opcodes.py":4391
 * def RES_197(cpu): # 197 RES 2,A
 *     t = cpu.A & ~(1 << 2)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4392
 *     t = cpu.A & ~(1 << 2)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4393
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4394
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4389
 * 
 * 
 * def RES_197(cpu): # 197 RES 2,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 2)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4397
 * 
 * 
 * def RES_198(cpu): # 198 RES 3,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 3)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_198(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4398
 * 
 * def RES_198(cpu): # 198 RES 3,B
 *     t = cpu.B & ~(1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B & (~0x8));

  /* "pyboy/core/opcodes.py":4399
 * def RES_198(cpu): # 198 RES 3,B
 *     t = cpu.B & ~(1 << 3)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4400
 *     t = cpu.B & ~(1 << 3)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4401
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4402
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4397
 * 
 * 
 * def RES_198(cpu): # 198 RES 3,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 3)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4405
 * 
 * 
 * def RES_199(cpu): # 199 RES 3,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 3)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_199(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4406
 * 
 * def RES_199(cpu): # 199 RES 3,C
 *     t = cpu.C & ~(1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C & (~0x8));

  /* "pyboy/core/opcodes.py":4407
 * def RES_199(cpu): # 199 RES 3,C
 *     t = cpu.C & ~(1 << 3)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4408
 *     t = cpu.C & ~(1 << 3)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4409
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4410
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4405
 * 
 * 
 * def RES_199(cpu): # 199 RES 3,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 3)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4413
 * 
 * 
 * def RES_19A(cpu): # 19A RES 3,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 3)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19A(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4414
 * 
 * def RES_19A(cpu): # 19A RES 3,D
 *     t = cpu.D & ~(1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D & (~0x8));

  /* "pyboy/core/opcodes.py":4415
 * def RES_19A(cpu): # 19A RES 3,D
 *     t = cpu.D & ~(1 << 3)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4416
 *     t = cpu.D & ~(1 << 3)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4417
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4418
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4413
 * 
 * 
 * def RES_19A(cpu): # 19A RES 3,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 3)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4421
 * 
 * 
 * def RES_19B(cpu): # 19B RES 3,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 3)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19B(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4422
 * 
 * def RES_19B(cpu): # 19B RES 3,E
 *     t = cpu.E & ~(1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E & (~0x8));

  /* "pyboy/core/opcodes.py":4423
 * def RES_19B(cpu): # 19B RES 3,E
 *     t = cpu.E & ~(1 << 3)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4424
 *     t = cpu.E & ~(1 << 3)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4425
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4426
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4421
 * 
 * 
 * def RES_19B(cpu): # 19B RES 3,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 3)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4429
 * 
 * 
 * def RES_19C(cpu): # 19C RES 3,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 3)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19C(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4430
 * 
 * def RES_19C(cpu): # 19C RES 3,H
 *     t = (cpu.HL >> 8) & ~(1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & (~0x8));

  /* "pyboy/core/opcodes.py":4431
 * def RES_19C(cpu): # 19C RES 3,H
 *     t = (cpu.HL >> 8) & ~(1 << 3)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4432
 *     t = (cpu.HL >> 8) & ~(1 << 3)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4433
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4434
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4429
 * 
 * 
 * def RES_19C(cpu): # 19C RES 3,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 3)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4437
 * 
 * 
 * def RES_19D(cpu): # 19D RES 3,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 3)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19D(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4438
 * 
 * def RES_19D(cpu): # 19D RES 3,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & (~0x8));

  /* "pyboy/core/opcodes.py":4439
 * def RES_19D(cpu): # 19D RES 3,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 3)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4440
 *     t = (cpu.HL & 0xFF) & ~(1 << 3)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4441
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4442
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4437
 * 
 * 
 * def RES_19D(cpu): # 19D RES 3,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 3)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4445
 * 
 * 
 * def RES_19E(cpu): # 19E RES 3,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 3)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19E(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4446
 * 
 * def RES_19E(cpu): # 19E RES 3,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & (~0x8));

  /* "pyboy/core/opcodes.py":4447
 * def RES_19E(cpu): # 19E RES 3,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 3)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4448
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 3)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4449
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4450
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4445
 * 
 * 
 * def RES_19E(cpu): # 19E RES 3,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 3)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4453
 * 
 * 
 * def RES_19F(cpu): # 19F RES 3,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 3)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_19F(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4454
 * 
 * def RES_19F(cpu): # 19F RES 3,A
 *     t = cpu.A & ~(1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A & (~0x8));

  /* "pyboy/core/opcodes.py":4455
 * def RES_19F(cpu): # 19F RES 3,A
 *     t = cpu.A & ~(1 << 3)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4456
 *     t = cpu.A & ~(1 << 3)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4457
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4458
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4453
 * 
 * 
 * def RES_19F(cpu): # 19F RES 3,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 3)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4461
 * 
 * 
 * def RES_1A0(cpu): # 1A0 RES 4,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 4)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4462
 * 
 * def RES_1A0(cpu): # 1A0 RES 4,B
 *     t = cpu.B & ~(1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B & (~0x10));

  /* "pyboy/core/opcodes.py":4463
 * def RES_1A0(cpu): # 1A0 RES 4,B
 *     t = cpu.B & ~(1 << 4)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4464
 *     t = cpu.B & ~(1 << 4)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4465
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4466
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4461
 * 
 * 
 * def RES_1A0(cpu): # 1A0 RES 4,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 4)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4469
 * 
 * 
 * def RES_1A1(cpu): # 1A1 RES 4,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 4)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4470
 * 
 * def RES_1A1(cpu): # 1A1 RES 4,C
 *     t = cpu.C & ~(1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C & (~0x10));

  /* "pyboy/core/opcodes.py":4471
 * def RES_1A1(cpu): # 1A1 RES 4,C
 *     t = cpu.C & ~(1 << 4)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4472
 *     t = cpu.C & ~(1 << 4)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4473
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4474
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4469
 * 
 * 
 * def RES_1A1(cpu): # 1A1 RES 4,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 4)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4477
 * 
 * 
 * def RES_1A2(cpu): # 1A2 RES 4,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 4)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4478
 * 
 * def RES_1A2(cpu): # 1A2 RES 4,D
 *     t = cpu.D & ~(1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D & (~0x10));

  /* "pyboy/core/opcodes.py":4479
 * def RES_1A2(cpu): # 1A2 RES 4,D
 *     t = cpu.D & ~(1 << 4)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4480
 *     t = cpu.D & ~(1 << 4)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4481
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4482
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4477
 * 
 * 
 * def RES_1A2(cpu): # 1A2 RES 4,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 4)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4485
 * 
 * 
 * def RES_1A3(cpu): # 1A3 RES 4,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 4)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4486
 * 
 * def RES_1A3(cpu): # 1A3 RES 4,E
 *     t = cpu.E & ~(1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E & (~0x10));

  /* "pyboy/core/opcodes.py":4487
 * def RES_1A3(cpu): # 1A3 RES 4,E
 *     t = cpu.E & ~(1 << 4)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4488
 *     t = cpu.E & ~(1 << 4)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4489
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4490
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4485
 * 
 * 
 * def RES_1A3(cpu): # 1A3 RES 4,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 4)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4493
 * 
 * 
 * def RES_1A4(cpu): # 1A4 RES 4,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 4)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4494
 * 
 * def RES_1A4(cpu): # 1A4 RES 4,H
 *     t = (cpu.HL >> 8) & ~(1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & (~0x10));

  /* "pyboy/core/opcodes.py":4495
 * def RES_1A4(cpu): # 1A4 RES 4,H
 *     t = (cpu.HL >> 8) & ~(1 << 4)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4496
 *     t = (cpu.HL >> 8) & ~(1 << 4)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4497
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4498
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4493
 * 
 * 
 * def RES_1A4(cpu): # 1A4 RES 4,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 4)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4501
 * 
 * 
 * def RES_1A5(cpu): # 1A5 RES 4,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 4)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4502
 * 
 * def RES_1A5(cpu): # 1A5 RES 4,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & (~0x10));

  /* "pyboy/core/opcodes.py":4503
 * def RES_1A5(cpu): # 1A5 RES 4,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 4)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4504
 *     t = (cpu.HL & 0xFF) & ~(1 << 4)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4505
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4506
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4501
 * 
 * 
 * def RES_1A5(cpu): # 1A5 RES 4,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 4)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4509
 * 
 * 
 * def RES_1A6(cpu): # 1A6 RES 4,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 4)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4510
 * 
 * def RES_1A6(cpu): # 1A6 RES 4,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & (~0x10));

  /* "pyboy/core/opcodes.py":4511
 * def RES_1A6(cpu): # 1A6 RES 4,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 4)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4512
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 4)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4513
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4514
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4509
 * 
 * 
 * def RES_1A6(cpu): # 1A6 RES 4,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 4)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4517
 * 
 * 
 * def RES_1A7(cpu): # 1A7 RES 4,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 4)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4518
 * 
 * def RES_1A7(cpu): # 1A7 RES 4,A
 *     t = cpu.A & ~(1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A & (~0x10));

  /* "pyboy/core/opcodes.py":4519
 * def RES_1A7(cpu): # 1A7 RES 4,A
 *     t = cpu.A & ~(1 << 4)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4520
 *     t = cpu.A & ~(1 << 4)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4521
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4522
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4517
 * 
 * 
 * def RES_1A7(cpu): # 1A7 RES 4,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 4)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4525
 * 
 * 
 * def RES_1A8(cpu): # 1A8 RES 5,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 5)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4526
 * 
 * def RES_1A8(cpu): # 1A8 RES 5,B
 *     t = cpu.B & ~(1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B & (~0x20));

  /* "pyboy/core/opcodes.py":4527
 * def RES_1A8(cpu): # 1A8 RES 5,B
 *     t = cpu.B & ~(1 << 5)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4528
 *     t = cpu.B & ~(1 << 5)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4529
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4530
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4525
 * 
 * 
 * def RES_1A8(cpu): # 1A8 RES 5,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 5)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4533
 * 
 * 
 * def RES_1A9(cpu): # 1A9 RES 5,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 5)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1A9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4534
 * 
 * def RES_1A9(cpu): # 1A9 RES 5,C
 *     t = cpu.C & ~(1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C & (~0x20));

  /* "pyboy/core/opcodes.py":4535
 * def RES_1A9(cpu): # 1A9 RES 5,C
 *     t = cpu.C & ~(1 << 5)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4536
 *     t = cpu.C & ~(1 << 5)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4537
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4538
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4533
 * 
 * 
 * def RES_1A9(cpu): # 1A9 RES 5,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 5)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4541
 * 
 * 
 * def RES_1AA(cpu): # 1AA RES 5,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 5)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4542
 * 
 * def RES_1AA(cpu): # 1AA RES 5,D
 *     t = cpu.D & ~(1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D & (~0x20));

  /* "pyboy/core/opcodes.py":4543
 * def RES_1AA(cpu): # 1AA RES 5,D
 *     t = cpu.D & ~(1 << 5)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4544
 *     t = cpu.D & ~(1 << 5)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4545
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4546
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4541
 * 
 * 
 * def RES_1AA(cpu): # 1AA RES 5,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 5)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4549
 * 
 * 
 * def RES_1AB(cpu): # 1AB RES 5,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 5)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4550
 * 
 * def RES_1AB(cpu): # 1AB RES 5,E
 *     t = cpu.E & ~(1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E & (~0x20));

  /* "pyboy/core/opcodes.py":4551
 * def RES_1AB(cpu): # 1AB RES 5,E
 *     t = cpu.E & ~(1 << 5)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4552
 *     t = cpu.E & ~(1 << 5)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4553
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4554
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4549
 * 
 * 
 * def RES_1AB(cpu): # 1AB RES 5,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 5)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4557
 * 
 * 
 * def RES_1AC(cpu): # 1AC RES 5,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 5)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4558
 * 
 * def RES_1AC(cpu): # 1AC RES 5,H
 *     t = (cpu.HL >> 8) & ~(1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & (~0x20));

  /* "pyboy/core/opcodes.py":4559
 * def RES_1AC(cpu): # 1AC RES 5,H
 *     t = (cpu.HL >> 8) & ~(1 << 5)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4560
 *     t = (cpu.HL >> 8) & ~(1 << 5)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4561
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4562
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4557
 * 
 * 
 * def RES_1AC(cpu): # 1AC RES 5,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 5)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4565
 * 
 * 
 * def RES_1AD(cpu): # 1AD RES 5,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 5)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4566
 * 
 * def RES_1AD(cpu): # 1AD RES 5,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & (~0x20));

  /* "pyboy/core/opcodes.py":4567
 * def RES_1AD(cpu): # 1AD RES 5,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 5)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4568
 *     t = (cpu.HL & 0xFF) & ~(1 << 5)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4569
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4570
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4565
 * 
 * 
 * def RES_1AD(cpu): # 1AD RES 5,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 5)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4573
 * 
 * 
 * def RES_1AE(cpu): # 1AE RES 5,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 5)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4574
 * 
 * def RES_1AE(cpu): # 1AE RES 5,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & (~0x20));

  /* "pyboy/core/opcodes.py":4575
 * def RES_1AE(cpu): # 1AE RES 5,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 5)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4576
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 5)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4577
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4578
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4573
 * 
 * 
 * def RES_1AE(cpu): # 1AE RES 5,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 5)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4581
 * 
 * 
 * def RES_1AF(cpu): # 1AF RES 5,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 5)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1AF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4582
 * 
 * def RES_1AF(cpu): # 1AF RES 5,A
 *     t = cpu.A & ~(1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A & (~0x20));

  /* "pyboy/core/opcodes.py":4583
 * def RES_1AF(cpu): # 1AF RES 5,A
 *     t = cpu.A & ~(1 << 5)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4584
 *     t = cpu.A & ~(1 << 5)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4585
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4586
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4581
 * 
 * 
 * def RES_1AF(cpu): # 1AF RES 5,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 5)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4589
 * 
 * 
 * def RES_1B0(cpu): # 1B0 RES 6,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 6)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4590
 * 
 * def RES_1B0(cpu): # 1B0 RES 6,B
 *     t = cpu.B & ~(1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B & (~0x40));

  /* "pyboy/core/opcodes.py":4591
 * def RES_1B0(cpu): # 1B0 RES 6,B
 *     t = cpu.B & ~(1 << 6)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4592
 *     t = cpu.B & ~(1 << 6)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4593
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4594
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4589
 * 
 * 
 * def RES_1B0(cpu): # 1B0 RES 6,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 6)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4597
 * 
 * 
 * def RES_1B1(cpu): # 1B1 RES 6,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 6)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4598
 * 
 * def RES_1B1(cpu): # 1B1 RES 6,C
 *     t = cpu.C & ~(1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C & (~0x40));

  /* "pyboy/core/opcodes.py":4599
 * def RES_1B1(cpu): # 1B1 RES 6,C
 *     t = cpu.C & ~(1 << 6)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4600
 *     t = cpu.C & ~(1 << 6)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4601
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4602
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4597
 * 
 * 
 * def RES_1B1(cpu): # 1B1 RES 6,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 6)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4605
 * 
 * 
 * def RES_1B2(cpu): # 1B2 RES 6,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 6)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4606
 * 
 * def RES_1B2(cpu): # 1B2 RES 6,D
 *     t = cpu.D & ~(1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D & (~0x40));

  /* "pyboy/core/opcodes.py":4607
 * def RES_1B2(cpu): # 1B2 RES 6,D
 *     t = cpu.D & ~(1 << 6)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4608
 *     t = cpu.D & ~(1 << 6)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4609
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4610
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4605
 * 
 * 
 * def RES_1B2(cpu): # 1B2 RES 6,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 6)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4613
 * 
 * 
 * def RES_1B3(cpu): # 1B3 RES 6,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 6)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4614
 * 
 * def RES_1B3(cpu): # 1B3 RES 6,E
 *     t = cpu.E & ~(1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E & (~0x40));

  /* "pyboy/core/opcodes.py":4615
 * def RES_1B3(cpu): # 1B3 RES 6,E
 *     t = cpu.E & ~(1 << 6)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4616
 *     t = cpu.E & ~(1 << 6)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4617
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4618
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4613
 * 
 * 
 * def RES_1B3(cpu): # 1B3 RES 6,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 6)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4621
 * 
 * 
 * def RES_1B4(cpu): # 1B4 RES 6,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 6)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4622
 * 
 * def RES_1B4(cpu): # 1B4 RES 6,H
 *     t = (cpu.HL >> 8) & ~(1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & (~0x40));

  /* "pyboy/core/opcodes.py":4623
 * def RES_1B4(cpu): # 1B4 RES 6,H
 *     t = (cpu.HL >> 8) & ~(1 << 6)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4624
 *     t = (cpu.HL >> 8) & ~(1 << 6)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4625
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4626
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4621
 * 
 * 
 * def RES_1B4(cpu): # 1B4 RES 6,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 6)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4629
 * 
 * 
 * def RES_1B5(cpu): # 1B5 RES 6,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 6)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4630
 * 
 * def RES_1B5(cpu): # 1B5 RES 6,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & (~0x40));

  /* "pyboy/core/opcodes.py":4631
 * def RES_1B5(cpu): # 1B5 RES 6,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 6)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4632
 *     t = (cpu.HL & 0xFF) & ~(1 << 6)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4633
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4634
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4629
 * 
 * 
 * def RES_1B5(cpu): # 1B5 RES 6,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 6)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4637
 * 
 * 
 * def RES_1B6(cpu): # 1B6 RES 6,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 6)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4638
 * 
 * def RES_1B6(cpu): # 1B6 RES 6,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & (~0x40));

  /* "pyboy/core/opcodes.py":4639
 * def RES_1B6(cpu): # 1B6 RES 6,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 6)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4640
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 6)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4641
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4642
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4637
 * 
 * 
 * def RES_1B6(cpu): # 1B6 RES 6,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 6)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4645
 * 
 * 
 * def RES_1B7(cpu): # 1B7 RES 6,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 6)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4646
 * 
 * def RES_1B7(cpu): # 1B7 RES 6,A
 *     t = cpu.A & ~(1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A & (~0x40));

  /* "pyboy/core/opcodes.py":4647
 * def RES_1B7(cpu): # 1B7 RES 6,A
 *     t = cpu.A & ~(1 << 6)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4648
 *     t = cpu.A & ~(1 << 6)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4649
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4650
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4645
 * 
 * 
 * def RES_1B7(cpu): # 1B7 RES 6,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 6)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4653
 * 
 * 
 * def RES_1B8(cpu): # 1B8 RES 7,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 7)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4654
 * 
 * def RES_1B8(cpu): # 1B8 RES 7,B
 *     t = cpu.B & ~(1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B & (~0x80));

  /* "pyboy/core/opcodes.py":4655
 * def RES_1B8(cpu): # 1B8 RES 7,B
 *     t = cpu.B & ~(1 << 7)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4656
 *     t = cpu.B & ~(1 << 7)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4657
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4658
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4653
 * 
 * 
 * def RES_1B8(cpu): # 1B8 RES 7,B             # <<<<<<<<<<<<<<
 *     t = cpu.B & ~(1 << 7)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4661
 * 
 * 
 * def RES_1B9(cpu): # 1B9 RES 7,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 7)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1B9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4662
 * 
 * def RES_1B9(cpu): # 1B9 RES 7,C
 *     t = cpu.C & ~(1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C & (~0x80));

  /* "pyboy/core/opcodes.py":4663
 * def RES_1B9(cpu): # 1B9 RES 7,C
 *     t = cpu.C & ~(1 << 7)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4664
 *     t = cpu.C & ~(1 << 7)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4665
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4666
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4661
 * 
 * 
 * def RES_1B9(cpu): # 1B9 RES 7,C             # <<<<<<<<<<<<<<
 *     t = cpu.C & ~(1 << 7)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4669
 * 
 * 
 * def RES_1BA(cpu): # 1BA RES 7,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 7)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4670
 * 
 * def RES_1BA(cpu): # 1BA RES 7,D
 *     t = cpu.D & ~(1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D & (~0x80));

  /* "pyboy/core/opcodes.py":4671
 * def RES_1BA(cpu): # 1BA RES 7,D
 *     t = cpu.D & ~(1 << 7)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4672
 *     t = cpu.D & ~(1 << 7)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4673
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4674
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4669
 * 
 * 
 * def RES_1BA(cpu): # 1BA RES 7,D             # <<<<<<<<<<<<<<
 *     t = cpu.D & ~(1 << 7)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4677
 * 
 * 
 * def RES_1BB(cpu): # 1BB RES 7,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 7)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4678
 * 
 * def RES_1BB(cpu): # 1BB RES 7,E
 *     t = cpu.E & ~(1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E & (~0x80));

  /* "pyboy/core/opcodes.py":4679
 * def RES_1BB(cpu): # 1BB RES 7,E
 *     t = cpu.E & ~(1 << 7)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4680
 *     t = cpu.E & ~(1 << 7)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4681
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4682
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4677
 * 
 * 
 * def RES_1BB(cpu): # 1BB RES 7,E             # <<<<<<<<<<<<<<
 *     t = cpu.E & ~(1 << 7)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4685
 * 
 * 
 * def RES_1BC(cpu): # 1BC RES 7,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 7)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4686
 * 
 * def RES_1BC(cpu): # 1BC RES 7,H
 *     t = (cpu.HL >> 8) & ~(1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) & (~0x80));

  /* "pyboy/core/opcodes.py":4687
 * def RES_1BC(cpu): # 1BC RES 7,H
 *     t = (cpu.HL >> 8) & ~(1 << 7)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4688
 *     t = (cpu.HL >> 8) & ~(1 << 7)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4689
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4690
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4685
 * 
 * 
 * def RES_1BC(cpu): # 1BC RES 7,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) & ~(1 << 7)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4693
 * 
 * 
 * def RES_1BD(cpu): # 1BD RES 7,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 7)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4694
 * 
 * def RES_1BD(cpu): # 1BD RES 7,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) & (~0x80));

  /* "pyboy/core/opcodes.py":4695
 * def RES_1BD(cpu): # 1BD RES 7,L
 *     t = (cpu.HL & 0xFF) & ~(1 << 7)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4696
 *     t = (cpu.HL & 0xFF) & ~(1 << 7)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4697
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4698
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4693
 * 
 * 
 * def RES_1BD(cpu): # 1BD RES 7,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) & ~(1 << 7)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4701
 * 
 * 
 * def RES_1BE(cpu): # 1BE RES 7,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 7)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4702
 * 
 * def RES_1BE(cpu): # 1BE RES 7,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) & (~0x80));

  /* "pyboy/core/opcodes.py":4703
 * def RES_1BE(cpu): # 1BE RES 7,(HL)
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 7)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4704
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 7)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4705
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4706
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4701
 * 
 * 
 * def RES_1BE(cpu): # 1BE RES 7,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) & ~(1 << 7)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4709
 * 
 * 
 * def RES_1BF(cpu): # 1BF RES 7,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 7)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_RES_1BF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4710
 * 
 * def RES_1BF(cpu): # 1BF RES 7,A
 *     t = cpu.A & ~(1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A & (~0x80));

  /* "pyboy/core/opcodes.py":4711
 * def RES_1BF(cpu): # 1BF RES 7,A
 *     t = cpu.A & ~(1 << 7)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4712
 *     t = cpu.A & ~(1 << 7)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4713
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4714
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4709
 * 
 * 
 * def RES_1BF(cpu): # 1BF RES 7,A             # <<<<<<<<<<<<<<
 *     t = cpu.A & ~(1 << 7)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4717
 * 
 * 
 * def SET_1C0(cpu): # 1C0 SET 0,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 0)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4718
 * 
 * def SET_1C0(cpu): # 1C0 SET 0,B
 *     t = cpu.B | (1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B | 0x1);

  /* "pyboy/core/opcodes.py":4719
 * def SET_1C0(cpu): # 1C0 SET 0,B
 *     t = cpu.B | (1 << 0)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4720
 *     t = cpu.B | (1 << 0)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4721
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4722
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4717
 * 
 * 
 * def SET_1C0(cpu): # 1C0 SET 0,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 0)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4725
 * 
 * 
 * def SET_1C1(cpu): # 1C1 SET 0,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 0)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4726
 * 
 * def SET_1C1(cpu): # 1C1 SET 0,C
 *     t = cpu.C | (1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C | 0x1);

  /* "pyboy/core/opcodes.py":4727
 * def SET_1C1(cpu): # 1C1 SET 0,C
 *     t = cpu.C | (1 << 0)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4728
 *     t = cpu.C | (1 << 0)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4729
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4730
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4725
 * 
 * 
 * def SET_1C1(cpu): # 1C1 SET 0,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 0)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4733
 * 
 * 
 * def SET_1C2(cpu): # 1C2 SET 0,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 0)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4734
 * 
 * def SET_1C2(cpu): # 1C2 SET 0,D
 *     t = cpu.D | (1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D | 0x1);

  /* "pyboy/core/opcodes.py":4735
 * def SET_1C2(cpu): # 1C2 SET 0,D
 *     t = cpu.D | (1 << 0)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4736
 *     t = cpu.D | (1 << 0)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4737
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4738
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4733
 * 
 * 
 * def SET_1C2(cpu): # 1C2 SET 0,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 0)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4741
 * 
 * 
 * def SET_1C3(cpu): # 1C3 SET 0,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 0)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4742
 * 
 * def SET_1C3(cpu): # 1C3 SET 0,E
 *     t = cpu.E | (1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E | 0x1);

  /* "pyboy/core/opcodes.py":4743
 * def SET_1C3(cpu): # 1C3 SET 0,E
 *     t = cpu.E | (1 << 0)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4744
 *     t = cpu.E | (1 << 0)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4745
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4746
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4741
 * 
 * 
 * def SET_1C3(cpu): # 1C3 SET 0,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 0)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4749
 * 
 * 
 * def SET_1C4(cpu): # 1C4 SET 0,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 0)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4750
 * 
 * def SET_1C4(cpu): # 1C4 SET 0,H
 *     t = (cpu.HL >> 8) | (1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) | 0x1);

  /* "pyboy/core/opcodes.py":4751
 * def SET_1C4(cpu): # 1C4 SET 0,H
 *     t = (cpu.HL >> 8) | (1 << 0)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4752
 *     t = (cpu.HL >> 8) | (1 << 0)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4753
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4754
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4749
 * 
 * 
 * def SET_1C4(cpu): # 1C4 SET 0,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 0)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4757
 * 
 * 
 * def SET_1C5(cpu): # 1C5 SET 0,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 0)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4758
 * 
 * def SET_1C5(cpu): # 1C5 SET 0,L
 *     t = (cpu.HL & 0xFF) | (1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) | 0x1);

  /* "pyboy/core/opcodes.py":4759
 * def SET_1C5(cpu): # 1C5 SET 0,L
 *     t = (cpu.HL & 0xFF) | (1 << 0)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4760
 *     t = (cpu.HL & 0xFF) | (1 << 0)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4761
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4762
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4757
 * 
 * 
 * def SET_1C5(cpu): # 1C5 SET 0,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 0)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4765
 * 
 * 
 * def SET_1C6(cpu): # 1C6 SET 0,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 0)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4766
 * 
 * def SET_1C6(cpu): # 1C6 SET 0,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) | 0x1);

  /* "pyboy/core/opcodes.py":4767
 * def SET_1C6(cpu): # 1C6 SET 0,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 0)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4768
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 0)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4769
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4770
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4765
 * 
 * 
 * def SET_1C6(cpu): # 1C6 SET 0,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 0)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4773
 * 
 * 
 * def SET_1C7(cpu): # 1C7 SET 0,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 0)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4774
 * 
 * def SET_1C7(cpu): # 1C7 SET 0,A
 *     t = cpu.A | (1 << 0)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A | 0x1);

  /* "pyboy/core/opcodes.py":4775
 * def SET_1C7(cpu): # 1C7 SET 0,A
 *     t = cpu.A | (1 << 0)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4776
 *     t = cpu.A | (1 << 0)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4777
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4778
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4773
 * 
 * 
 * def SET_1C7(cpu): # 1C7 SET 0,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 0)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4781
 * 
 * 
 * def SET_1C8(cpu): # 1C8 SET 1,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 1)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4782
 * 
 * def SET_1C8(cpu): # 1C8 SET 1,B
 *     t = cpu.B | (1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B | 0x2);

  /* "pyboy/core/opcodes.py":4783
 * def SET_1C8(cpu): # 1C8 SET 1,B
 *     t = cpu.B | (1 << 1)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4784
 *     t = cpu.B | (1 << 1)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4785
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4786
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4781
 * 
 * 
 * def SET_1C8(cpu): # 1C8 SET 1,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 1)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4789
 * 
 * 
 * def SET_1C9(cpu): # 1C9 SET 1,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 1)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1C9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4790
 * 
 * def SET_1C9(cpu): # 1C9 SET 1,C
 *     t = cpu.C | (1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C | 0x2);

  /* "pyboy/core/opcodes.py":4791
 * def SET_1C9(cpu): # 1C9 SET 1,C
 *     t = cpu.C | (1 << 1)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4792
 *     t = cpu.C | (1 << 1)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4793
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4794
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4789
 * 
 * 
 * def SET_1C9(cpu): # 1C9 SET 1,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 1)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4797
 * 
 * 
 * def SET_1CA(cpu): # 1CA SET 1,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 1)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4798
 * 
 * def SET_1CA(cpu): # 1CA SET 1,D
 *     t = cpu.D | (1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D | 0x2);

  /* "pyboy/core/opcodes.py":4799
 * def SET_1CA(cpu): # 1CA SET 1,D
 *     t = cpu.D | (1 << 1)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4800
 *     t = cpu.D | (1 << 1)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4801
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4802
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4797
 * 
 * 
 * def SET_1CA(cpu): # 1CA SET 1,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 1)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4805
 * 
 * 
 * def SET_1CB(cpu): # 1CB SET 1,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 1)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4806
 * 
 * def SET_1CB(cpu): # 1CB SET 1,E
 *     t = cpu.E | (1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E | 0x2);

  /* "pyboy/core/opcodes.py":4807
 * def SET_1CB(cpu): # 1CB SET 1,E
 *     t = cpu.E | (1 << 1)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4808
 *     t = cpu.E | (1 << 1)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4809
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4810
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4805
 * 
 * 
 * def SET_1CB(cpu): # 1CB SET 1,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 1)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4813
 * 
 * 
 * def SET_1CC(cpu): # 1CC SET 1,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 1)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4814
 * 
 * def SET_1CC(cpu): # 1CC SET 1,H
 *     t = (cpu.HL >> 8) | (1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) | 0x2);

  /* "pyboy/core/opcodes.py":4815
 * def SET_1CC(cpu): # 1CC SET 1,H
 *     t = (cpu.HL >> 8) | (1 << 1)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4816
 *     t = (cpu.HL >> 8) | (1 << 1)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4817
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4818
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4813
 * 
 * 
 * def SET_1CC(cpu): # 1CC SET 1,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 1)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4821
 * 
 * 
 * def SET_1CD(cpu): # 1CD SET 1,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 1)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4822
 * 
 * def SET_1CD(cpu): # 1CD SET 1,L
 *     t = (cpu.HL & 0xFF) | (1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) | 0x2);

  /* "pyboy/core/opcodes.py":4823
 * def SET_1CD(cpu): # 1CD SET 1,L
 *     t = (cpu.HL & 0xFF) | (1 << 1)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4824
 *     t = (cpu.HL & 0xFF) | (1 << 1)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4825
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4826
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4821
 * 
 * 
 * def SET_1CD(cpu): # 1CD SET 1,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 1)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4829
 * 
 * 
 * def SET_1CE(cpu): # 1CE SET 1,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 1)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4830
 * 
 * def SET_1CE(cpu): # 1CE SET 1,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) | 0x2);

  /* "pyboy/core/opcodes.py":4831
 * def SET_1CE(cpu): # 1CE SET 1,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 1)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4832
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 1)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4833
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4834
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4829
 * 
 * 
 * def SET_1CE(cpu): # 1CE SET 1,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 1)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4837
 * 
 * 
 * def SET_1CF(cpu): # 1CF SET 1,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 1)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1CF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4838
 * 
 * def SET_1CF(cpu): # 1CF SET 1,A
 *     t = cpu.A | (1 << 1)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A | 0x2);

  /* "pyboy/core/opcodes.py":4839
 * def SET_1CF(cpu): # 1CF SET 1,A
 *     t = cpu.A | (1 << 1)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4840
 *     t = cpu.A | (1 << 1)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4841
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4842
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4837
 * 
 * 
 * def SET_1CF(cpu): # 1CF SET 1,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 1)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4845
 * 
 * 
 * def SET_1D0(cpu): # 1D0 SET 2,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 2)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4846
 * 
 * def SET_1D0(cpu): # 1D0 SET 2,B
 *     t = cpu.B | (1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B | 0x4);

  /* "pyboy/core/opcodes.py":4847
 * def SET_1D0(cpu): # 1D0 SET 2,B
 *     t = cpu.B | (1 << 2)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4848
 *     t = cpu.B | (1 << 2)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4849
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4850
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4845
 * 
 * 
 * def SET_1D0(cpu): # 1D0 SET 2,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 2)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4853
 * 
 * 
 * def SET_1D1(cpu): # 1D1 SET 2,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 2)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4854
 * 
 * def SET_1D1(cpu): # 1D1 SET 2,C
 *     t = cpu.C | (1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C | 0x4);

  /* "pyboy/core/opcodes.py":4855
 * def SET_1D1(cpu): # 1D1 SET 2,C
 *     t = cpu.C | (1 << 2)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4856
 *     t = cpu.C | (1 << 2)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4857
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4858
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4853
 * 
 * 
 * def SET_1D1(cpu): # 1D1 SET 2,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 2)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4861
 * 
 * 
 * def SET_1D2(cpu): # 1D2 SET 2,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 2)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4862
 * 
 * def SET_1D2(cpu): # 1D2 SET 2,D
 *     t = cpu.D | (1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D | 0x4);

  /* "pyboy/core/opcodes.py":4863
 * def SET_1D2(cpu): # 1D2 SET 2,D
 *     t = cpu.D | (1 << 2)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4864
 *     t = cpu.D | (1 << 2)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4865
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4866
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4861
 * 
 * 
 * def SET_1D2(cpu): # 1D2 SET 2,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 2)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4869
 * 
 * 
 * def SET_1D3(cpu): # 1D3 SET 2,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 2)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4870
 * 
 * def SET_1D3(cpu): # 1D3 SET 2,E
 *     t = cpu.E | (1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E | 0x4);

  /* "pyboy/core/opcodes.py":4871
 * def SET_1D3(cpu): # 1D3 SET 2,E
 *     t = cpu.E | (1 << 2)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4872
 *     t = cpu.E | (1 << 2)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4873
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4874
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4869
 * 
 * 
 * def SET_1D3(cpu): # 1D3 SET 2,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 2)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4877
 * 
 * 
 * def SET_1D4(cpu): # 1D4 SET 2,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 2)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4878
 * 
 * def SET_1D4(cpu): # 1D4 SET 2,H
 *     t = (cpu.HL >> 8) | (1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) | 0x4);

  /* "pyboy/core/opcodes.py":4879
 * def SET_1D4(cpu): # 1D4 SET 2,H
 *     t = (cpu.HL >> 8) | (1 << 2)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4880
 *     t = (cpu.HL >> 8) | (1 << 2)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4881
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4882
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4877
 * 
 * 
 * def SET_1D4(cpu): # 1D4 SET 2,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 2)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4885
 * 
 * 
 * def SET_1D5(cpu): # 1D5 SET 2,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 2)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4886
 * 
 * def SET_1D5(cpu): # 1D5 SET 2,L
 *     t = (cpu.HL & 0xFF) | (1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) | 0x4);

  /* "pyboy/core/opcodes.py":4887
 * def SET_1D5(cpu): # 1D5 SET 2,L
 *     t = (cpu.HL & 0xFF) | (1 << 2)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4888
 *     t = (cpu.HL & 0xFF) | (1 << 2)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4889
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4890
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4885
 * 
 * 
 * def SET_1D5(cpu): # 1D5 SET 2,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 2)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4893
 * 
 * 
 * def SET_1D6(cpu): # 1D6 SET 2,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 2)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4894
 * 
 * def SET_1D6(cpu): # 1D6 SET 2,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) | 0x4);

  /* "pyboy/core/opcodes.py":4895
 * def SET_1D6(cpu): # 1D6 SET 2,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 2)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4896
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 2)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4897
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4898
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4893
 * 
 * 
 * def SET_1D6(cpu): # 1D6 SET 2,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 2)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4901
 * 
 * 
 * def SET_1D7(cpu): # 1D7 SET 2,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 2)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4902
 * 
 * def SET_1D7(cpu): # 1D7 SET 2,A
 *     t = cpu.A | (1 << 2)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A | 0x4);

  /* "pyboy/core/opcodes.py":4903
 * def SET_1D7(cpu): # 1D7 SET 2,A
 *     t = cpu.A | (1 << 2)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4904
 *     t = cpu.A | (1 << 2)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4905
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4906
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4901
 * 
 * 
 * def SET_1D7(cpu): # 1D7 SET 2,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 2)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4909
 * 
 * 
 * def SET_1D8(cpu): # 1D8 SET 3,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 3)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4910
 * 
 * def SET_1D8(cpu): # 1D8 SET 3,B
 *     t = cpu.B | (1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B | 0x8);

  /* "pyboy/core/opcodes.py":4911
 * def SET_1D8(cpu): # 1D8 SET 3,B
 *     t = cpu.B | (1 << 3)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4912
 *     t = cpu.B | (1 << 3)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4913
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4914
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4909
 * 
 * 
 * def SET_1D8(cpu): # 1D8 SET 3,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 3)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4917
 * 
 * 
 * def SET_1D9(cpu): # 1D9 SET 3,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 3)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1D9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4918
 * 
 * def SET_1D9(cpu): # 1D9 SET 3,C
 *     t = cpu.C | (1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C | 0x8);

  /* "pyboy/core/opcodes.py":4919
 * def SET_1D9(cpu): # 1D9 SET 3,C
 *     t = cpu.C | (1 << 3)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4920
 *     t = cpu.C | (1 << 3)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4921
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4922
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4917
 * 
 * 
 * def SET_1D9(cpu): # 1D9 SET 3,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 3)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4925
 * 
 * 
 * def SET_1DA(cpu): # 1DA SET 3,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 3)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4926
 * 
 * def SET_1DA(cpu): # 1DA SET 3,D
 *     t = cpu.D | (1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D | 0x8);

  /* "pyboy/core/opcodes.py":4927
 * def SET_1DA(cpu): # 1DA SET 3,D
 *     t = cpu.D | (1 << 3)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4928
 *     t = cpu.D | (1 << 3)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4929
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4930
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4925
 * 
 * 
 * def SET_1DA(cpu): # 1DA SET 3,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 3)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4933
 * 
 * 
 * def SET_1DB(cpu): # 1DB SET 3,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 3)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4934
 * 
 * def SET_1DB(cpu): # 1DB SET 3,E
 *     t = cpu.E | (1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E | 0x8);

  /* "pyboy/core/opcodes.py":4935
 * def SET_1DB(cpu): # 1DB SET 3,E
 *     t = cpu.E | (1 << 3)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4936
 *     t = cpu.E | (1 << 3)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4937
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4938
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4933
 * 
 * 
 * def SET_1DB(cpu): # 1DB SET 3,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 3)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4941
 * 
 * 
 * def SET_1DC(cpu): # 1DC SET 3,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 3)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4942
 * 
 * def SET_1DC(cpu): # 1DC SET 3,H
 *     t = (cpu.HL >> 8) | (1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) | 0x8);

  /* "pyboy/core/opcodes.py":4943
 * def SET_1DC(cpu): # 1DC SET 3,H
 *     t = (cpu.HL >> 8) | (1 << 3)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":4944
 *     t = (cpu.HL >> 8) | (1 << 3)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4945
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4946
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4941
 * 
 * 
 * def SET_1DC(cpu): # 1DC SET 3,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 3)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4949
 * 
 * 
 * def SET_1DD(cpu): # 1DD SET 3,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 3)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4950
 * 
 * def SET_1DD(cpu): # 1DD SET 3,L
 *     t = (cpu.HL & 0xFF) | (1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) | 0x8);

  /* "pyboy/core/opcodes.py":4951
 * def SET_1DD(cpu): # 1DD SET 3,L
 *     t = (cpu.HL & 0xFF) | (1 << 3)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":4952
 *     t = (cpu.HL & 0xFF) | (1 << 3)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4953
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4954
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4949
 * 
 * 
 * def SET_1DD(cpu): # 1DD SET 3,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 3)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4957
 * 
 * 
 * def SET_1DE(cpu): # 1DE SET 3,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 3)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4958
 * 
 * def SET_1DE(cpu): # 1DE SET 3,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) | 0x8);

  /* "pyboy/core/opcodes.py":4959
 * def SET_1DE(cpu): # 1DE SET 3,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 3)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":4960
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 3)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4961
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4962
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4957
 * 
 * 
 * def SET_1DE(cpu): # 1DE SET 3,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 3)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4965
 * 
 * 
 * def SET_1DF(cpu): # 1DF SET 3,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 3)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1DF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4966
 * 
 * def SET_1DF(cpu): # 1DF SET 3,A
 *     t = cpu.A | (1 << 3)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A | 0x8);

  /* "pyboy/core/opcodes.py":4967
 * def SET_1DF(cpu): # 1DF SET 3,A
 *     t = cpu.A | (1 << 3)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4968
 *     t = cpu.A | (1 << 3)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4969
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4970
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4965
 * 
 * 
 * def SET_1DF(cpu): # 1DF SET 3,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 3)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4973
 * 
 * 
 * def SET_1E0(cpu): # 1E0 SET 4,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 4)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4974
 * 
 * def SET_1E0(cpu): # 1E0 SET 4,B
 *     t = cpu.B | (1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B | 0x10);

  /* "pyboy/core/opcodes.py":4975
 * def SET_1E0(cpu): # 1E0 SET 4,B
 *     t = cpu.B | (1 << 4)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4976
 *     t = cpu.B | (1 << 4)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4977
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4978
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4973
 * 
 * 
 * def SET_1E0(cpu): # 1E0 SET 4,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 4)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4981
 * 
 * 
 * def SET_1E1(cpu): # 1E1 SET 4,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 4)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4982
 * 
 * def SET_1E1(cpu): # 1E1 SET 4,C
 *     t = cpu.C | (1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C | 0x10);

  /* "pyboy/core/opcodes.py":4983
 * def SET_1E1(cpu): # 1E1 SET 4,C
 *     t = cpu.C | (1 << 4)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4984
 *     t = cpu.C | (1 << 4)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4985
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4986
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4981
 * 
 * 
 * def SET_1E1(cpu): # 1E1 SET 4,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 4)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4989
 * 
 * 
 * def SET_1E2(cpu): # 1E2 SET 4,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 4)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4990
 * 
 * def SET_1E2(cpu): # 1E2 SET 4,D
 *     t = cpu.D | (1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D | 0x10);

  /* "pyboy/core/opcodes.py":4991
 * def SET_1E2(cpu): # 1E2 SET 4,D
 *     t = cpu.D | (1 << 4)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":4992
 *     t = cpu.D | (1 << 4)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":4993
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":4994
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4989
 * 
 * 
 * def SET_1E2(cpu): # 1E2 SET 4,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 4)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":4997
 * 
 * 
 * def SET_1E3(cpu): # 1E3 SET 4,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 4)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":4998
 * 
 * def SET_1E3(cpu): # 1E3 SET 4,E
 *     t = cpu.E | (1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E | 0x10);

  /* "pyboy/core/opcodes.py":4999
 * def SET_1E3(cpu): # 1E3 SET 4,E
 *     t = cpu.E | (1 << 4)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5000
 *     t = cpu.E | (1 << 4)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5001
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5002
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":4997
 * 
 * 
 * def SET_1E3(cpu): # 1E3 SET 4,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 4)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5005
 * 
 * 
 * def SET_1E4(cpu): # 1E4 SET 4,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 4)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5006
 * 
 * def SET_1E4(cpu): # 1E4 SET 4,H
 *     t = (cpu.HL >> 8) | (1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) | 0x10);

  /* "pyboy/core/opcodes.py":5007
 * def SET_1E4(cpu): # 1E4 SET 4,H
 *     t = (cpu.HL >> 8) | (1 << 4)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":5008
 *     t = (cpu.HL >> 8) | (1 << 4)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5009
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5010
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5005
 * 
 * 
 * def SET_1E4(cpu): # 1E4 SET 4,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 4)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5013
 * 
 * 
 * def SET_1E5(cpu): # 1E5 SET 4,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 4)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5014
 * 
 * def SET_1E5(cpu): # 1E5 SET 4,L
 *     t = (cpu.HL & 0xFF) | (1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) | 0x10);

  /* "pyboy/core/opcodes.py":5015
 * def SET_1E5(cpu): # 1E5 SET 4,L
 *     t = (cpu.HL & 0xFF) | (1 << 4)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":5016
 *     t = (cpu.HL & 0xFF) | (1 << 4)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5017
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5018
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5013
 * 
 * 
 * def SET_1E5(cpu): # 1E5 SET 4,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 4)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5021
 * 
 * 
 * def SET_1E6(cpu): # 1E6 SET 4,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 4)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5022
 * 
 * def SET_1E6(cpu): # 1E6 SET 4,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) | 0x10);

  /* "pyboy/core/opcodes.py":5023
 * def SET_1E6(cpu): # 1E6 SET 4,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 4)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":5024
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 4)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5025
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5026
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5021
 * 
 * 
 * def SET_1E6(cpu): # 1E6 SET 4,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 4)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5029
 * 
 * 
 * def SET_1E7(cpu): # 1E7 SET 4,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 4)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5030
 * 
 * def SET_1E7(cpu): # 1E7 SET 4,A
 *     t = cpu.A | (1 << 4)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A | 0x10);

  /* "pyboy/core/opcodes.py":5031
 * def SET_1E7(cpu): # 1E7 SET 4,A
 *     t = cpu.A | (1 << 4)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5032
 *     t = cpu.A | (1 << 4)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5033
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5034
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5029
 * 
 * 
 * def SET_1E7(cpu): # 1E7 SET 4,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 4)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5037
 * 
 * 
 * def SET_1E8(cpu): # 1E8 SET 5,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 5)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5038
 * 
 * def SET_1E8(cpu): # 1E8 SET 5,B
 *     t = cpu.B | (1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B | 0x20);

  /* "pyboy/core/opcodes.py":5039
 * def SET_1E8(cpu): # 1E8 SET 5,B
 *     t = cpu.B | (1 << 5)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5040
 *     t = cpu.B | (1 << 5)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5041
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5042
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5037
 * 
 * 
 * def SET_1E8(cpu): # 1E8 SET 5,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 5)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5045
 * 
 * 
 * def SET_1E9(cpu): # 1E9 SET 5,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 5)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1E9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5046
 * 
 * def SET_1E9(cpu): # 1E9 SET 5,C
 *     t = cpu.C | (1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C | 0x20);

  /* "pyboy/core/opcodes.py":5047
 * def SET_1E9(cpu): # 1E9 SET 5,C
 *     t = cpu.C | (1 << 5)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5048
 *     t = cpu.C | (1 << 5)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5049
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5050
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5045
 * 
 * 
 * def SET_1E9(cpu): # 1E9 SET 5,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 5)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5053
 * 
 * 
 * def SET_1EA(cpu): # 1EA SET 5,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 5)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5054
 * 
 * def SET_1EA(cpu): # 1EA SET 5,D
 *     t = cpu.D | (1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D | 0x20);

  /* "pyboy/core/opcodes.py":5055
 * def SET_1EA(cpu): # 1EA SET 5,D
 *     t = cpu.D | (1 << 5)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5056
 *     t = cpu.D | (1 << 5)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5057
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5058
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5053
 * 
 * 
 * def SET_1EA(cpu): # 1EA SET 5,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 5)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5061
 * 
 * 
 * def SET_1EB(cpu): # 1EB SET 5,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 5)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5062
 * 
 * def SET_1EB(cpu): # 1EB SET 5,E
 *     t = cpu.E | (1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E | 0x20);

  /* "pyboy/core/opcodes.py":5063
 * def SET_1EB(cpu): # 1EB SET 5,E
 *     t = cpu.E | (1 << 5)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5064
 *     t = cpu.E | (1 << 5)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5065
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5066
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5061
 * 
 * 
 * def SET_1EB(cpu): # 1EB SET 5,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 5)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5069
 * 
 * 
 * def SET_1EC(cpu): # 1EC SET 5,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 5)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5070
 * 
 * def SET_1EC(cpu): # 1EC SET 5,H
 *     t = (cpu.HL >> 8) | (1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) | 0x20);

  /* "pyboy/core/opcodes.py":5071
 * def SET_1EC(cpu): # 1EC SET 5,H
 *     t = (cpu.HL >> 8) | (1 << 5)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":5072
 *     t = (cpu.HL >> 8) | (1 << 5)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5073
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5074
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5069
 * 
 * 
 * def SET_1EC(cpu): # 1EC SET 5,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 5)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5077
 * 
 * 
 * def SET_1ED(cpu): # 1ED SET 5,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 5)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1ED(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5078
 * 
 * def SET_1ED(cpu): # 1ED SET 5,L
 *     t = (cpu.HL & 0xFF) | (1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) | 0x20);

  /* "pyboy/core/opcodes.py":5079
 * def SET_1ED(cpu): # 1ED SET 5,L
 *     t = (cpu.HL & 0xFF) | (1 << 5)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":5080
 *     t = (cpu.HL & 0xFF) | (1 << 5)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5081
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5082
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5077
 * 
 * 
 * def SET_1ED(cpu): # 1ED SET 5,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 5)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5085
 * 
 * 
 * def SET_1EE(cpu): # 1EE SET 5,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 5)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5086
 * 
 * def SET_1EE(cpu): # 1EE SET 5,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) | 0x20);

  /* "pyboy/core/opcodes.py":5087
 * def SET_1EE(cpu): # 1EE SET 5,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 5)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":5088
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 5)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5089
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5090
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5085
 * 
 * 
 * def SET_1EE(cpu): # 1EE SET 5,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 5)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5093
 * 
 * 
 * def SET_1EF(cpu): # 1EF SET 5,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 5)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1EF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5094
 * 
 * def SET_1EF(cpu): # 1EF SET 5,A
 *     t = cpu.A | (1 << 5)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A | 0x20);

  /* "pyboy/core/opcodes.py":5095
 * def SET_1EF(cpu): # 1EF SET 5,A
 *     t = cpu.A | (1 << 5)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5096
 *     t = cpu.A | (1 << 5)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5097
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5098
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5093
 * 
 * 
 * def SET_1EF(cpu): # 1EF SET 5,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 5)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5101
 * 
 * 
 * def SET_1F0(cpu): # 1F0 SET 6,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 6)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F0(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5102
 * 
 * def SET_1F0(cpu): # 1F0 SET 6,B
 *     t = cpu.B | (1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B | 0x40);

  /* "pyboy/core/opcodes.py":5103
 * def SET_1F0(cpu): # 1F0 SET 6,B
 *     t = cpu.B | (1 << 6)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5104
 *     t = cpu.B | (1 << 6)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5105
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5106
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5101
 * 
 * 
 * def SET_1F0(cpu): # 1F0 SET 6,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 6)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5109
 * 
 * 
 * def SET_1F1(cpu): # 1F1 SET 6,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 6)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F1(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5110
 * 
 * def SET_1F1(cpu): # 1F1 SET 6,C
 *     t = cpu.C | (1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C | 0x40);

  /* "pyboy/core/opcodes.py":5111
 * def SET_1F1(cpu): # 1F1 SET 6,C
 *     t = cpu.C | (1 << 6)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5112
 *     t = cpu.C | (1 << 6)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5113
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5114
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5109
 * 
 * 
 * def SET_1F1(cpu): # 1F1 SET 6,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 6)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5117
 * 
 * 
 * def SET_1F2(cpu): # 1F2 SET 6,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 6)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F2(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5118
 * 
 * def SET_1F2(cpu): # 1F2 SET 6,D
 *     t = cpu.D | (1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D | 0x40);

  /* "pyboy/core/opcodes.py":5119
 * def SET_1F2(cpu): # 1F2 SET 6,D
 *     t = cpu.D | (1 << 6)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5120
 *     t = cpu.D | (1 << 6)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5121
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5122
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5117
 * 
 * 
 * def SET_1F2(cpu): # 1F2 SET 6,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 6)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5125
 * 
 * 
 * def SET_1F3(cpu): # 1F3 SET 6,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 6)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F3(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5126
 * 
 * def SET_1F3(cpu): # 1F3 SET 6,E
 *     t = cpu.E | (1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E | 0x40);

  /* "pyboy/core/opcodes.py":5127
 * def SET_1F3(cpu): # 1F3 SET 6,E
 *     t = cpu.E | (1 << 6)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5128
 *     t = cpu.E | (1 << 6)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5129
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5130
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5125
 * 
 * 
 * def SET_1F3(cpu): # 1F3 SET 6,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 6)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5133
 * 
 * 
 * def SET_1F4(cpu): # 1F4 SET 6,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 6)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F4(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5134
 * 
 * def SET_1F4(cpu): # 1F4 SET 6,H
 *     t = (cpu.HL >> 8) | (1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) | 0x40);

  /* "pyboy/core/opcodes.py":5135
 * def SET_1F4(cpu): # 1F4 SET 6,H
 *     t = (cpu.HL >> 8) | (1 << 6)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":5136
 *     t = (cpu.HL >> 8) | (1 << 6)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5137
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5138
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5133
 * 
 * 
 * def SET_1F4(cpu): # 1F4 SET 6,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 6)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5141
 * 
 * 
 * def SET_1F5(cpu): # 1F5 SET 6,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 6)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F5(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5142
 * 
 * def SET_1F5(cpu): # 1F5 SET 6,L
 *     t = (cpu.HL & 0xFF) | (1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) | 0x40);

  /* "pyboy/core/opcodes.py":5143
 * def SET_1F5(cpu): # 1F5 SET 6,L
 *     t = (cpu.HL & 0xFF) | (1 << 6)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":5144
 *     t = (cpu.HL & 0xFF) | (1 << 6)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5145
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5146
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5141
 * 
 * 
 * def SET_1F5(cpu): # 1F5 SET 6,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 6)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5149
 * 
 * 
 * def SET_1F6(cpu): # 1F6 SET 6,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 6)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F6(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5150
 * 
 * def SET_1F6(cpu): # 1F6 SET 6,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) | 0x40);

  /* "pyboy/core/opcodes.py":5151
 * def SET_1F6(cpu): # 1F6 SET 6,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 6)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":5152
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 6)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5153
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5154
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5149
 * 
 * 
 * def SET_1F6(cpu): # 1F6 SET 6,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 6)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5157
 * 
 * 
 * def SET_1F7(cpu): # 1F7 SET 6,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 6)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F7(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5158
 * 
 * def SET_1F7(cpu): # 1F7 SET 6,A
 *     t = cpu.A | (1 << 6)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A | 0x40);

  /* "pyboy/core/opcodes.py":5159
 * def SET_1F7(cpu): # 1F7 SET 6,A
 *     t = cpu.A | (1 << 6)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5160
 *     t = cpu.A | (1 << 6)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5161
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5162
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5157
 * 
 * 
 * def SET_1F7(cpu): # 1F7 SET 6,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 6)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5165
 * 
 * 
 * def SET_1F8(cpu): # 1F8 SET 7,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 7)
 *     cpu.B = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F8(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5166
 * 
 * def SET_1F8(cpu): # 1F8 SET 7,B
 *     t = cpu.B | (1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.B = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->B | 0x80);

  /* "pyboy/core/opcodes.py":5167
 * def SET_1F8(cpu): # 1F8 SET 7,B
 *     t = cpu.B | (1 << 7)
 *     cpu.B = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->B = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5168
 *     t = cpu.B | (1 << 7)
 *     cpu.B = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5169
 *     cpu.B = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5170
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5165
 * 
 * 
 * def SET_1F8(cpu): # 1F8 SET 7,B             # <<<<<<<<<<<<<<
 *     t = cpu.B | (1 << 7)
 *     cpu.B = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5173
 * 
 * 
 * def SET_1F9(cpu): # 1F9 SET 7,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 7)
 *     cpu.C = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1F9(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5174
 * 
 * def SET_1F9(cpu): # 1F9 SET 7,C
 *     t = cpu.C | (1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.C = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->C | 0x80);

  /* "pyboy/core/opcodes.py":5175
 * def SET_1F9(cpu): # 1F9 SET 7,C
 *     t = cpu.C | (1 << 7)
 *     cpu.C = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->C = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5176
 *     t = cpu.C | (1 << 7)
 *     cpu.C = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5177
 *     cpu.C = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5178
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5173
 * 
 * 
 * def SET_1F9(cpu): # 1F9 SET 7,C             # <<<<<<<<<<<<<<
 *     t = cpu.C | (1 << 7)
 *     cpu.C = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5181
 * 
 * 
 * def SET_1FA(cpu): # 1FA SET 7,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 7)
 *     cpu.D = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FA(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5182
 * 
 * def SET_1FA(cpu): # 1FA SET 7,D
 *     t = cpu.D | (1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.D = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->D | 0x80);

  /* "pyboy/core/opcodes.py":5183
 * def SET_1FA(cpu): # 1FA SET 7,D
 *     t = cpu.D | (1 << 7)
 *     cpu.D = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->D = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5184
 *     t = cpu.D | (1 << 7)
 *     cpu.D = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5185
 *     cpu.D = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5186
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5181
 * 
 * 
 * def SET_1FA(cpu): # 1FA SET 7,D             # <<<<<<<<<<<<<<
 *     t = cpu.D | (1 << 7)
 *     cpu.D = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5189
 * 
 * 
 * def SET_1FB(cpu): # 1FB SET 7,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 7)
 *     cpu.E = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FB(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5190
 * 
 * def SET_1FB(cpu): # 1FB SET 7,E
 *     t = cpu.E | (1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.E = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->E | 0x80);

  /* "pyboy/core/opcodes.py":5191
 * def SET_1FB(cpu): # 1FB SET 7,E
 *     t = cpu.E | (1 << 7)
 *     cpu.E = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->E = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5192
 *     t = cpu.E | (1 << 7)
 *     cpu.E = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5193
 *     cpu.E = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5194
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5189
 * 
 * 
 * def SET_1FB(cpu): # 1FB SET 7,E             # <<<<<<<<<<<<<<
 *     t = cpu.E | (1 << 7)
 *     cpu.E = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5197
 * 
 * 
 * def SET_1FC(cpu): # 1FC SET 7,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 7)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FC(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5198
 * 
 * def SET_1FC(cpu): # 1FC SET 7,H
 *     t = (cpu.HL >> 8) | (1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL >> 8) | 0x80);

  /* "pyboy/core/opcodes.py":5199
 * def SET_1FC(cpu): # 1FC SET 7,H
 *     t = (cpu.HL >> 8) | (1 << 7)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0x00FF) | (__pyx_v_t << 8));

  /* "pyboy/core/opcodes.py":5200
 *     t = (cpu.HL >> 8) | (1 << 7)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5201
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5202
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5197
 * 
 * 
 * def SET_1FC(cpu): # 1FC SET 7,H             # <<<<<<<<<<<<<<
 *     t = (cpu.HL >> 8) | (1 << 7)
 *     cpu.HL = (cpu.HL & 0x00FF) | (t << 8)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5205
 * 
 * 
 * def SET_1FD(cpu): # 1FD SET 7,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 7)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FD(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5206
 * 
 * def SET_1FD(cpu): # 1FD SET 7,L
 *     t = (cpu.HL & 0xFF) | (1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 */
  __pyx_v_t = ((__pyx_v_cpu->HL & 0xFF) | 0x80);

  /* "pyboy/core/opcodes.py":5207
 * def SET_1FD(cpu): # 1FD SET 7,L
 *     t = (cpu.HL & 0xFF) | (1 << 7)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->HL = ((__pyx_v_cpu->HL & 0xFF00) | (__pyx_v_t & 0xFF));

  /* "pyboy/core/opcodes.py":5208
 *     t = (cpu.HL & 0xFF) | (1 << 7)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5209
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5210
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5205
 * 
 * 
 * def SET_1FD(cpu): # 1FD SET 7,L             # <<<<<<<<<<<<<<
 *     t = (cpu.HL & 0xFF) | (1 << 7)
 *     cpu.HL = (cpu.HL & 0xFF00) | (t & 0xFF)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5213
 * 
 * 
 * def SET_1FE(cpu): # 1FE SET 7,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 7)
 *     cpu.mb.setitem(cpu.HL, t)
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FE(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5214
 * 
 * def SET_1FE(cpu): # 1FE SET 7,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 */
  __pyx_v_t = (((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL) | 0x80);

  /* "pyboy/core/opcodes.py":5215
 * def SET_1FE(cpu): # 1FE SET 7,(HL)
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 7)
 *     cpu.mb.setitem(cpu.HL, t)             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->setitem(__pyx_v_cpu->mb, __pyx_v_cpu->HL, __pyx_v_t);

  /* "pyboy/core/opcodes.py":5216
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 7)
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 16
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5217
 *     cpu.mb.setitem(cpu.HL, t)
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 16
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5218
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 16             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 16;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5213
 * 
 * 
 * def SET_1FE(cpu): # 1FE SET 7,(HL)             # <<<<<<<<<<<<<<
 *     t = cpu.mb.getitem(cpu.HL) | (1 << 7)
 *     cpu.mb.setitem(cpu.HL, t)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5221
 * 
 * 
 * def SET_1FF(cpu): # 1FF SET 7,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 7)
 *     cpu.A = t
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_SET_1FF(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  int __pyx_v_t;
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5222
 * 
 * def SET_1FF(cpu): # 1FF SET 7,A
 *     t = cpu.A | (1 << 7)             # <<<<<<<<<<<<<<
 *     cpu.A = t
 *     cpu.PC += 2
 */
  __pyx_v_t = (__pyx_v_cpu->A | 0x80);

  /* "pyboy/core/opcodes.py":5223
 * def SET_1FF(cpu): # 1FF SET 7,A
 *     t = cpu.A | (1 << 7)
 *     cpu.A = t             # <<<<<<<<<<<<<<
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 */
  __pyx_v_cpu->A = __pyx_v_t;

  /* "pyboy/core/opcodes.py":5224
 *     t = cpu.A | (1 << 7)
 *     cpu.A = t
 *     cpu.PC += 2             # <<<<<<<<<<<<<<
 *     cpu.PC &= 0xFFFF
 *     return 8
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC + 2);

  /* "pyboy/core/opcodes.py":5225
 *     cpu.A = t
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF             # <<<<<<<<<<<<<<
 *     return 8
 * 
 */
  __pyx_v_cpu->PC = (__pyx_v_cpu->PC & 0xFFFF);

  /* "pyboy/core/opcodes.py":5226
 *     cpu.PC += 2
 *     cpu.PC &= 0xFFFF
 *     return 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 8;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5221
 * 
 * 
 * def SET_1FF(cpu): # 1FF SET 7,A             # <<<<<<<<<<<<<<
 *     t = cpu.A | (1 << 7)
 *     cpu.A = t
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5229
 * 
 * 
 * def no_opcode(cpu):             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

static uint8_t __pyx_f_5pyboy_4core_7opcodes_no_opcode(CYTHON_UNUSED struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu) {
  uint8_t __pyx_r;

  /* "pyboy/core/opcodes.py":5230
 * 
 * def no_opcode(cpu):
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyboy/core/opcodes.py":5229
 * 
 * 
 * def no_opcode(cpu):             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pyboy/core/opcodes.py":5234
 * 
 * 
 * def execute_opcode(cpu, opcode):             # <<<<<<<<<<<<<<
 *     oplen = OPCODE_LENGTHS[opcode]
 *     v = 0
 */

static int __pyx_f_5pyboy_4core_7opcodes_execute_opcode(struct __pyx_obj_5pyboy_4core_3cpu_CPU *__pyx_v_cpu, uint16_t __pyx_v_opcode) {
  int __pyx_v_v;
  int __pyx_v_a;
  int __pyx_v_b;
  unsigned short __pyx_v_pc;
  uint8_t __pyx_v_oplen;
  int __pyx_r;
  int __pyx_t_1;

  /* "pyboy/core/opcodes.py":5235
 * 
 * def execute_opcode(cpu, opcode):
 *     oplen = OPCODE_LENGTHS[opcode]             # <<<<<<<<<<<<<<
 *     v = 0
 *     pc = cpu.PC
 */
  __pyx_v_oplen = (__pyx_v_5pyboy_4core_7opcodes_OPCODE_LENGTHS[__pyx_v_opcode]);

  /* "pyboy/core/opcodes.py":5236
 * def execute_opcode(cpu, opcode):
 *     oplen = OPCODE_LENGTHS[opcode]
 *     v = 0             # <<<<<<<<<<<<<<
 *     pc = cpu.PC
 *     if oplen == 2:
 */
  __pyx_v_v = 0;

  /* "pyboy/core/opcodes.py":5237
 *     oplen = OPCODE_LENGTHS[opcode]
 *     v = 0
 *     pc = cpu.PC             # <<<<<<<<<<<<<<
 *     if oplen == 2:
 *         # 8-bit immediate
 */
  __pyx_t_1 = __pyx_v_cpu->PC;
  __pyx_v_pc = __pyx_t_1;

  /* "pyboy/core/opcodes.py":5238
 *     v = 0
 *     pc = cpu.PC
 *     if oplen == 2:             # <<<<<<<<<<<<<<
 *         # 8-bit immediate
 *         v = cpu.mb.getitem(pc+1)
 */
  switch (__pyx_v_oplen) {
    case 2:

    /* "pyboy/core/opcodes.py":5240
 *     if oplen == 2:
 *         # 8-bit immediate
 *         v = cpu.mb.getitem(pc+1)             # <<<<<<<<<<<<<<
 *     elif oplen == 3:
 *         # 16-bit immediate
 */
    __pyx_v_v = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, (__pyx_v_pc + 1));

    /* "pyboy/core/opcodes.py":5238
 *     v = 0
 *     pc = cpu.PC
 *     if oplen == 2:             # <<<<<<<<<<<<<<
 *         # 8-bit immediate
 *         v = cpu.mb.getitem(pc+1)
 */
    break;
    case 3:

    /* "pyboy/core/opcodes.py":5244
 *         # 16-bit immediate
 *         # Flips order of values due to big-endian
 *         a = cpu.mb.getitem(pc+2)             # <<<<<<<<<<<<<<
 *         b = cpu.mb.getitem(pc+1)
 *         v = (a << 8) + b
 */
    __pyx_v_a = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, (__pyx_v_pc + 2));

    /* "pyboy/core/opcodes.py":5245
 *         # Flips order of values due to big-endian
 *         a = cpu.mb.getitem(pc+2)
 *         b = cpu.mb.getitem(pc+1)             # <<<<<<<<<<<<<<
 *         v = (a << 8) + b
 * 
 */
    __pyx_v_b = ((struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard *)__pyx_v_cpu->mb->__pyx_vtab)->getitem(__pyx_v_cpu->mb, (__pyx_v_pc + 1));

    /* "pyboy/core/opcodes.py":5246
 *         a = cpu.mb.getitem(pc+2)
 *         b = cpu.mb.getitem(pc+1)
 *         v = (a << 8) + b             # <<<<<<<<<<<<<<
 * 
 *     if opcode == 0x00:
 */
    __pyx_v_v = ((__pyx_v_a << 8) + __pyx_v_b);

    /* "pyboy/core/opcodes.py":5241
 *         # 8-bit immediate
 *         v = cpu.mb.getitem(pc+1)
 *     elif oplen == 3:             # <<<<<<<<<<<<<<
 *         # 16-bit immediate
 *         # Flips order of values due to big-endian
 */
    break;
    default: break;
  }

  /* "pyboy/core/opcodes.py":5248
 *         v = (a << 8) + b
 * 
 *     if opcode == 0x00:             # <<<<<<<<<<<<<<
 *         return NOP_00(cpu)
 *     elif opcode == 0x01:
 */
  switch (__pyx_v_opcode) {
    case 0x00:

    /* "pyboy/core/opcodes.py":5249
 * 
 *     if opcode == 0x00:
 *         return NOP_00(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x01:
 *         return LD_01(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_NOP_00(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5248
 *         v = (a << 8) + b
 * 
 *     if opcode == 0x00:             # <<<<<<<<<<<<<<
 *         return NOP_00(cpu)
 *     elif opcode == 0x01:
 */
    break;
    case 0x01:

    /* "pyboy/core/opcodes.py":5251
 *         return NOP_00(cpu)
 *     elif opcode == 0x01:
 *         return LD_01(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x02:
 *         return LD_02(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_01(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5250
 *     if opcode == 0x00:
 *         return NOP_00(cpu)
 *     elif opcode == 0x01:             # <<<<<<<<<<<<<<
 *         return LD_01(cpu, v)
 *     elif opcode == 0x02:
 */
    break;
    case 0x02:

    /* "pyboy/core/opcodes.py":5253
 *         return LD_01(cpu, v)
 *     elif opcode == 0x02:
 *         return LD_02(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x03:
 *         return INC_03(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_02(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5252
 *     elif opcode == 0x01:
 *         return LD_01(cpu, v)
 *     elif opcode == 0x02:             # <<<<<<<<<<<<<<
 *         return LD_02(cpu)
 *     elif opcode == 0x03:
 */
    break;
    case 0x03:

    /* "pyboy/core/opcodes.py":5255
 *         return LD_02(cpu)
 *     elif opcode == 0x03:
 *         return INC_03(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x04:
 *         return INC_04(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_03(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5254
 *     elif opcode == 0x02:
 *         return LD_02(cpu)
 *     elif opcode == 0x03:             # <<<<<<<<<<<<<<
 *         return INC_03(cpu)
 *     elif opcode == 0x04:
 */
    break;
    case 0x04:

    /* "pyboy/core/opcodes.py":5257
 *         return INC_03(cpu)
 *     elif opcode == 0x04:
 *         return INC_04(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x05:
 *         return DEC_05(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_04(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5256
 *     elif opcode == 0x03:
 *         return INC_03(cpu)
 *     elif opcode == 0x04:             # <<<<<<<<<<<<<<
 *         return INC_04(cpu)
 *     elif opcode == 0x05:
 */
    break;
    case 0x05:

    /* "pyboy/core/opcodes.py":5259
 *         return INC_04(cpu)
 *     elif opcode == 0x05:
 *         return DEC_05(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x06:
 *         return LD_06(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_05(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5258
 *     elif opcode == 0x04:
 *         return INC_04(cpu)
 *     elif opcode == 0x05:             # <<<<<<<<<<<<<<
 *         return DEC_05(cpu)
 *     elif opcode == 0x06:
 */
    break;
    case 0x06:

    /* "pyboy/core/opcodes.py":5261
 *         return DEC_05(cpu)
 *     elif opcode == 0x06:
 *         return LD_06(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x07:
 *         return RLCA_07(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_06(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5260
 *     elif opcode == 0x05:
 *         return DEC_05(cpu)
 *     elif opcode == 0x06:             # <<<<<<<<<<<<<<
 *         return LD_06(cpu, v)
 *     elif opcode == 0x07:
 */
    break;
    case 0x07:

    /* "pyboy/core/opcodes.py":5263
 *         return LD_06(cpu, v)
 *     elif opcode == 0x07:
 *         return RLCA_07(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x08:
 *         return LD_08(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLCA_07(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5262
 *     elif opcode == 0x06:
 *         return LD_06(cpu, v)
 *     elif opcode == 0x07:             # <<<<<<<<<<<<<<
 *         return RLCA_07(cpu)
 *     elif opcode == 0x08:
 */
    break;
    case 0x08:

    /* "pyboy/core/opcodes.py":5265
 *         return RLCA_07(cpu)
 *     elif opcode == 0x08:
 *         return LD_08(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x09:
 *         return ADD_09(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_08(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5264
 *     elif opcode == 0x07:
 *         return RLCA_07(cpu)
 *     elif opcode == 0x08:             # <<<<<<<<<<<<<<
 *         return LD_08(cpu, v)
 *     elif opcode == 0x09:
 */
    break;
    case 0x09:

    /* "pyboy/core/opcodes.py":5267
 *         return LD_08(cpu, v)
 *     elif opcode == 0x09:
 *         return ADD_09(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x0A:
 *         return LD_0A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_09(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5266
 *     elif opcode == 0x08:
 *         return LD_08(cpu, v)
 *     elif opcode == 0x09:             # <<<<<<<<<<<<<<
 *         return ADD_09(cpu)
 *     elif opcode == 0x0A:
 */
    break;
    case 0x0A:

    /* "pyboy/core/opcodes.py":5269
 *         return ADD_09(cpu)
 *     elif opcode == 0x0A:
 *         return LD_0A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x0B:
 *         return DEC_0B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_0A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5268
 *     elif opcode == 0x09:
 *         return ADD_09(cpu)
 *     elif opcode == 0x0A:             # <<<<<<<<<<<<<<
 *         return LD_0A(cpu)
 *     elif opcode == 0x0B:
 */
    break;
    case 0x0B:

    /* "pyboy/core/opcodes.py":5271
 *         return LD_0A(cpu)
 *     elif opcode == 0x0B:
 *         return DEC_0B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x0C:
 *         return INC_0C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_0B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5270
 *     elif opcode == 0x0A:
 *         return LD_0A(cpu)
 *     elif opcode == 0x0B:             # <<<<<<<<<<<<<<
 *         return DEC_0B(cpu)
 *     elif opcode == 0x0C:
 */
    break;
    case 0x0C:

    /* "pyboy/core/opcodes.py":5273
 *         return DEC_0B(cpu)
 *     elif opcode == 0x0C:
 *         return INC_0C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x0D:
 *         return DEC_0D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_0C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5272
 *     elif opcode == 0x0B:
 *         return DEC_0B(cpu)
 *     elif opcode == 0x0C:             # <<<<<<<<<<<<<<
 *         return INC_0C(cpu)
 *     elif opcode == 0x0D:
 */
    break;
    case 0x0D:

    /* "pyboy/core/opcodes.py":5275
 *         return INC_0C(cpu)
 *     elif opcode == 0x0D:
 *         return DEC_0D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x0E:
 *         return LD_0E(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_0D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5274
 *     elif opcode == 0x0C:
 *         return INC_0C(cpu)
 *     elif opcode == 0x0D:             # <<<<<<<<<<<<<<
 *         return DEC_0D(cpu)
 *     elif opcode == 0x0E:
 */
    break;
    case 0x0E:

    /* "pyboy/core/opcodes.py":5277
 *         return DEC_0D(cpu)
 *     elif opcode == 0x0E:
 *         return LD_0E(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x0F:
 *         return RRCA_0F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_0E(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5276
 *     elif opcode == 0x0D:
 *         return DEC_0D(cpu)
 *     elif opcode == 0x0E:             # <<<<<<<<<<<<<<
 *         return LD_0E(cpu, v)
 *     elif opcode == 0x0F:
 */
    break;
    case 0x0F:

    /* "pyboy/core/opcodes.py":5279
 *         return LD_0E(cpu, v)
 *     elif opcode == 0x0F:
 *         return RRCA_0F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x10:
 *         return STOP_10(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRCA_0F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5278
 *     elif opcode == 0x0E:
 *         return LD_0E(cpu, v)
 *     elif opcode == 0x0F:             # <<<<<<<<<<<<<<
 *         return RRCA_0F(cpu)
 *     elif opcode == 0x10:
 */
    break;
    case 0x10:

    /* "pyboy/core/opcodes.py":5281
 *         return RRCA_0F(cpu)
 *     elif opcode == 0x10:
 *         return STOP_10(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x11:
 *         return LD_11(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_STOP_10(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5280
 *     elif opcode == 0x0F:
 *         return RRCA_0F(cpu)
 *     elif opcode == 0x10:             # <<<<<<<<<<<<<<
 *         return STOP_10(cpu, v)
 *     elif opcode == 0x11:
 */
    break;
    case 0x11:

    /* "pyboy/core/opcodes.py":5283
 *         return STOP_10(cpu, v)
 *     elif opcode == 0x11:
 *         return LD_11(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x12:
 *         return LD_12(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_11(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5282
 *     elif opcode == 0x10:
 *         return STOP_10(cpu, v)
 *     elif opcode == 0x11:             # <<<<<<<<<<<<<<
 *         return LD_11(cpu, v)
 *     elif opcode == 0x12:
 */
    break;
    case 0x12:

    /* "pyboy/core/opcodes.py":5285
 *         return LD_11(cpu, v)
 *     elif opcode == 0x12:
 *         return LD_12(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x13:
 *         return INC_13(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_12(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5284
 *     elif opcode == 0x11:
 *         return LD_11(cpu, v)
 *     elif opcode == 0x12:             # <<<<<<<<<<<<<<
 *         return LD_12(cpu)
 *     elif opcode == 0x13:
 */
    break;
    case 0x13:

    /* "pyboy/core/opcodes.py":5287
 *         return LD_12(cpu)
 *     elif opcode == 0x13:
 *         return INC_13(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x14:
 *         return INC_14(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_13(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5286
 *     elif opcode == 0x12:
 *         return LD_12(cpu)
 *     elif opcode == 0x13:             # <<<<<<<<<<<<<<
 *         return INC_13(cpu)
 *     elif opcode == 0x14:
 */
    break;
    case 0x14:

    /* "pyboy/core/opcodes.py":5289
 *         return INC_13(cpu)
 *     elif opcode == 0x14:
 *         return INC_14(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x15:
 *         return DEC_15(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_14(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5288
 *     elif opcode == 0x13:
 *         return INC_13(cpu)
 *     elif opcode == 0x14:             # <<<<<<<<<<<<<<
 *         return INC_14(cpu)
 *     elif opcode == 0x15:
 */
    break;
    case 0x15:

    /* "pyboy/core/opcodes.py":5291
 *         return INC_14(cpu)
 *     elif opcode == 0x15:
 *         return DEC_15(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x16:
 *         return LD_16(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_15(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5290
 *     elif opcode == 0x14:
 *         return INC_14(cpu)
 *     elif opcode == 0x15:             # <<<<<<<<<<<<<<
 *         return DEC_15(cpu)
 *     elif opcode == 0x16:
 */
    break;
    case 0x16:

    /* "pyboy/core/opcodes.py":5293
 *         return DEC_15(cpu)
 *     elif opcode == 0x16:
 *         return LD_16(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x17:
 *         return RLA_17(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_16(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5292
 *     elif opcode == 0x15:
 *         return DEC_15(cpu)
 *     elif opcode == 0x16:             # <<<<<<<<<<<<<<
 *         return LD_16(cpu, v)
 *     elif opcode == 0x17:
 */
    break;
    case 0x17:

    /* "pyboy/core/opcodes.py":5295
 *         return LD_16(cpu, v)
 *     elif opcode == 0x17:
 *         return RLA_17(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x18:
 *         return JR_18(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLA_17(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5294
 *     elif opcode == 0x16:
 *         return LD_16(cpu, v)
 *     elif opcode == 0x17:             # <<<<<<<<<<<<<<
 *         return RLA_17(cpu)
 *     elif opcode == 0x18:
 */
    break;
    case 0x18:

    /* "pyboy/core/opcodes.py":5297
 *         return RLA_17(cpu)
 *     elif opcode == 0x18:
 *         return JR_18(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x19:
 *         return ADD_19(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JR_18(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5296
 *     elif opcode == 0x17:
 *         return RLA_17(cpu)
 *     elif opcode == 0x18:             # <<<<<<<<<<<<<<
 *         return JR_18(cpu, v)
 *     elif opcode == 0x19:
 */
    break;
    case 0x19:

    /* "pyboy/core/opcodes.py":5299
 *         return JR_18(cpu, v)
 *     elif opcode == 0x19:
 *         return ADD_19(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A:
 *         return LD_1A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_19(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5298
 *     elif opcode == 0x18:
 *         return JR_18(cpu, v)
 *     elif opcode == 0x19:             # <<<<<<<<<<<<<<
 *         return ADD_19(cpu)
 *     elif opcode == 0x1A:
 */
    break;
    case 0x1A:

    /* "pyboy/core/opcodes.py":5301
 *         return ADD_19(cpu)
 *     elif opcode == 0x1A:
 *         return LD_1A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B:
 *         return DEC_1B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_1A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5300
 *     elif opcode == 0x19:
 *         return ADD_19(cpu)
 *     elif opcode == 0x1A:             # <<<<<<<<<<<<<<
 *         return LD_1A(cpu)
 *     elif opcode == 0x1B:
 */
    break;
    case 0x1B:

    /* "pyboy/core/opcodes.py":5303
 *         return LD_1A(cpu)
 *     elif opcode == 0x1B:
 *         return DEC_1B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C:
 *         return INC_1C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_1B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5302
 *     elif opcode == 0x1A:
 *         return LD_1A(cpu)
 *     elif opcode == 0x1B:             # <<<<<<<<<<<<<<
 *         return DEC_1B(cpu)
 *     elif opcode == 0x1C:
 */
    break;
    case 0x1C:

    /* "pyboy/core/opcodes.py":5305
 *         return DEC_1B(cpu)
 *     elif opcode == 0x1C:
 *         return INC_1C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D:
 *         return DEC_1D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_1C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5304
 *     elif opcode == 0x1B:
 *         return DEC_1B(cpu)
 *     elif opcode == 0x1C:             # <<<<<<<<<<<<<<
 *         return INC_1C(cpu)
 *     elif opcode == 0x1D:
 */
    break;
    case 0x1D:

    /* "pyboy/core/opcodes.py":5307
 *         return INC_1C(cpu)
 *     elif opcode == 0x1D:
 *         return DEC_1D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E:
 *         return LD_1E(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_1D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5306
 *     elif opcode == 0x1C:
 *         return INC_1C(cpu)
 *     elif opcode == 0x1D:             # <<<<<<<<<<<<<<
 *         return DEC_1D(cpu)
 *     elif opcode == 0x1E:
 */
    break;
    case 0x1E:

    /* "pyboy/core/opcodes.py":5309
 *         return DEC_1D(cpu)
 *     elif opcode == 0x1E:
 *         return LD_1E(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F:
 *         return RRA_1F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_1E(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5308
 *     elif opcode == 0x1D:
 *         return DEC_1D(cpu)
 *     elif opcode == 0x1E:             # <<<<<<<<<<<<<<
 *         return LD_1E(cpu, v)
 *     elif opcode == 0x1F:
 */
    break;
    case 0x1F:

    /* "pyboy/core/opcodes.py":5311
 *         return LD_1E(cpu, v)
 *     elif opcode == 0x1F:
 *         return RRA_1F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x20:
 *         return JR_20(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRA_1F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5310
 *     elif opcode == 0x1E:
 *         return LD_1E(cpu, v)
 *     elif opcode == 0x1F:             # <<<<<<<<<<<<<<
 *         return RRA_1F(cpu)
 *     elif opcode == 0x20:
 */
    break;
    case 0x20:

    /* "pyboy/core/opcodes.py":5313
 *         return RRA_1F(cpu)
 *     elif opcode == 0x20:
 *         return JR_20(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x21:
 *         return LD_21(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JR_20(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5312
 *     elif opcode == 0x1F:
 *         return RRA_1F(cpu)
 *     elif opcode == 0x20:             # <<<<<<<<<<<<<<
 *         return JR_20(cpu, v)
 *     elif opcode == 0x21:
 */
    break;
    case 0x21:

    /* "pyboy/core/opcodes.py":5315
 *         return JR_20(cpu, v)
 *     elif opcode == 0x21:
 *         return LD_21(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x22:
 *         return LD_22(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_21(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5314
 *     elif opcode == 0x20:
 *         return JR_20(cpu, v)
 *     elif opcode == 0x21:             # <<<<<<<<<<<<<<
 *         return LD_21(cpu, v)
 *     elif opcode == 0x22:
 */
    break;
    case 0x22:

    /* "pyboy/core/opcodes.py":5317
 *         return LD_21(cpu, v)
 *     elif opcode == 0x22:
 *         return LD_22(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x23:
 *         return INC_23(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_22(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5316
 *     elif opcode == 0x21:
 *         return LD_21(cpu, v)
 *     elif opcode == 0x22:             # <<<<<<<<<<<<<<
 *         return LD_22(cpu)
 *     elif opcode == 0x23:
 */
    break;
    case 0x23:

    /* "pyboy/core/opcodes.py":5319
 *         return LD_22(cpu)
 *     elif opcode == 0x23:
 *         return INC_23(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x24:
 *         return INC_24(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_23(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5318
 *     elif opcode == 0x22:
 *         return LD_22(cpu)
 *     elif opcode == 0x23:             # <<<<<<<<<<<<<<
 *         return INC_23(cpu)
 *     elif opcode == 0x24:
 */
    break;
    case 0x24:

    /* "pyboy/core/opcodes.py":5321
 *         return INC_23(cpu)
 *     elif opcode == 0x24:
 *         return INC_24(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x25:
 *         return DEC_25(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_24(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5320
 *     elif opcode == 0x23:
 *         return INC_23(cpu)
 *     elif opcode == 0x24:             # <<<<<<<<<<<<<<
 *         return INC_24(cpu)
 *     elif opcode == 0x25:
 */
    break;
    case 0x25:

    /* "pyboy/core/opcodes.py":5323
 *         return INC_24(cpu)
 *     elif opcode == 0x25:
 *         return DEC_25(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x26:
 *         return LD_26(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_25(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5322
 *     elif opcode == 0x24:
 *         return INC_24(cpu)
 *     elif opcode == 0x25:             # <<<<<<<<<<<<<<
 *         return DEC_25(cpu)
 *     elif opcode == 0x26:
 */
    break;
    case 0x26:

    /* "pyboy/core/opcodes.py":5325
 *         return DEC_25(cpu)
 *     elif opcode == 0x26:
 *         return LD_26(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x27:
 *         return DAA_27(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_26(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5324
 *     elif opcode == 0x25:
 *         return DEC_25(cpu)
 *     elif opcode == 0x26:             # <<<<<<<<<<<<<<
 *         return LD_26(cpu, v)
 *     elif opcode == 0x27:
 */
    break;
    case 0x27:

    /* "pyboy/core/opcodes.py":5327
 *         return LD_26(cpu, v)
 *     elif opcode == 0x27:
 *         return DAA_27(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x28:
 *         return JR_28(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DAA_27(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5326
 *     elif opcode == 0x26:
 *         return LD_26(cpu, v)
 *     elif opcode == 0x27:             # <<<<<<<<<<<<<<
 *         return DAA_27(cpu)
 *     elif opcode == 0x28:
 */
    break;
    case 0x28:

    /* "pyboy/core/opcodes.py":5329
 *         return DAA_27(cpu)
 *     elif opcode == 0x28:
 *         return JR_28(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x29:
 *         return ADD_29(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JR_28(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5328
 *     elif opcode == 0x27:
 *         return DAA_27(cpu)
 *     elif opcode == 0x28:             # <<<<<<<<<<<<<<
 *         return JR_28(cpu, v)
 *     elif opcode == 0x29:
 */
    break;
    case 0x29:

    /* "pyboy/core/opcodes.py":5331
 *         return JR_28(cpu, v)
 *     elif opcode == 0x29:
 *         return ADD_29(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x2A:
 *         return LD_2A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_29(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5330
 *     elif opcode == 0x28:
 *         return JR_28(cpu, v)
 *     elif opcode == 0x29:             # <<<<<<<<<<<<<<
 *         return ADD_29(cpu)
 *     elif opcode == 0x2A:
 */
    break;
    case 0x2A:

    /* "pyboy/core/opcodes.py":5333
 *         return ADD_29(cpu)
 *     elif opcode == 0x2A:
 *         return LD_2A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x2B:
 *         return DEC_2B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_2A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5332
 *     elif opcode == 0x29:
 *         return ADD_29(cpu)
 *     elif opcode == 0x2A:             # <<<<<<<<<<<<<<
 *         return LD_2A(cpu)
 *     elif opcode == 0x2B:
 */
    break;
    case 0x2B:

    /* "pyboy/core/opcodes.py":5335
 *         return LD_2A(cpu)
 *     elif opcode == 0x2B:
 *         return DEC_2B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x2C:
 *         return INC_2C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_2B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5334
 *     elif opcode == 0x2A:
 *         return LD_2A(cpu)
 *     elif opcode == 0x2B:             # <<<<<<<<<<<<<<
 *         return DEC_2B(cpu)
 *     elif opcode == 0x2C:
 */
    break;
    case 0x2C:

    /* "pyboy/core/opcodes.py":5337
 *         return DEC_2B(cpu)
 *     elif opcode == 0x2C:
 *         return INC_2C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x2D:
 *         return DEC_2D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_2C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5336
 *     elif opcode == 0x2B:
 *         return DEC_2B(cpu)
 *     elif opcode == 0x2C:             # <<<<<<<<<<<<<<
 *         return INC_2C(cpu)
 *     elif opcode == 0x2D:
 */
    break;
    case 0x2D:

    /* "pyboy/core/opcodes.py":5339
 *         return INC_2C(cpu)
 *     elif opcode == 0x2D:
 *         return DEC_2D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x2E:
 *         return LD_2E(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_2D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5338
 *     elif opcode == 0x2C:
 *         return INC_2C(cpu)
 *     elif opcode == 0x2D:             # <<<<<<<<<<<<<<
 *         return DEC_2D(cpu)
 *     elif opcode == 0x2E:
 */
    break;
    case 0x2E:

    /* "pyboy/core/opcodes.py":5341
 *         return DEC_2D(cpu)
 *     elif opcode == 0x2E:
 *         return LD_2E(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x2F:
 *         return CPL_2F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_2E(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5340
 *     elif opcode == 0x2D:
 *         return DEC_2D(cpu)
 *     elif opcode == 0x2E:             # <<<<<<<<<<<<<<
 *         return LD_2E(cpu, v)
 *     elif opcode == 0x2F:
 */
    break;
    case 0x2F:

    /* "pyboy/core/opcodes.py":5343
 *         return LD_2E(cpu, v)
 *     elif opcode == 0x2F:
 *         return CPL_2F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x30:
 *         return JR_30(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CPL_2F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5342
 *     elif opcode == 0x2E:
 *         return LD_2E(cpu, v)
 *     elif opcode == 0x2F:             # <<<<<<<<<<<<<<
 *         return CPL_2F(cpu)
 *     elif opcode == 0x30:
 */
    break;
    case 0x30:

    /* "pyboy/core/opcodes.py":5345
 *         return CPL_2F(cpu)
 *     elif opcode == 0x30:
 *         return JR_30(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x31:
 *         return LD_31(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JR_30(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5344
 *     elif opcode == 0x2F:
 *         return CPL_2F(cpu)
 *     elif opcode == 0x30:             # <<<<<<<<<<<<<<
 *         return JR_30(cpu, v)
 *     elif opcode == 0x31:
 */
    break;
    case 0x31:

    /* "pyboy/core/opcodes.py":5347
 *         return JR_30(cpu, v)
 *     elif opcode == 0x31:
 *         return LD_31(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x32:
 *         return LD_32(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_31(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5346
 *     elif opcode == 0x30:
 *         return JR_30(cpu, v)
 *     elif opcode == 0x31:             # <<<<<<<<<<<<<<
 *         return LD_31(cpu, v)
 *     elif opcode == 0x32:
 */
    break;
    case 0x32:

    /* "pyboy/core/opcodes.py":5349
 *         return LD_31(cpu, v)
 *     elif opcode == 0x32:
 *         return LD_32(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x33:
 *         return INC_33(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_32(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5348
 *     elif opcode == 0x31:
 *         return LD_31(cpu, v)
 *     elif opcode == 0x32:             # <<<<<<<<<<<<<<
 *         return LD_32(cpu)
 *     elif opcode == 0x33:
 */
    break;
    case 0x33:

    /* "pyboy/core/opcodes.py":5351
 *         return LD_32(cpu)
 *     elif opcode == 0x33:
 *         return INC_33(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x34:
 *         return INC_34(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_33(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5350
 *     elif opcode == 0x32:
 *         return LD_32(cpu)
 *     elif opcode == 0x33:             # <<<<<<<<<<<<<<
 *         return INC_33(cpu)
 *     elif opcode == 0x34:
 */
    break;
    case 0x34:

    /* "pyboy/core/opcodes.py":5353
 *         return INC_33(cpu)
 *     elif opcode == 0x34:
 *         return INC_34(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x35:
 *         return DEC_35(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_34(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5352
 *     elif opcode == 0x33:
 *         return INC_33(cpu)
 *     elif opcode == 0x34:             # <<<<<<<<<<<<<<
 *         return INC_34(cpu)
 *     elif opcode == 0x35:
 */
    break;
    case 0x35:

    /* "pyboy/core/opcodes.py":5355
 *         return INC_34(cpu)
 *     elif opcode == 0x35:
 *         return DEC_35(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x36:
 *         return LD_36(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_35(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5354
 *     elif opcode == 0x34:
 *         return INC_34(cpu)
 *     elif opcode == 0x35:             # <<<<<<<<<<<<<<
 *         return DEC_35(cpu)
 *     elif opcode == 0x36:
 */
    break;
    case 0x36:

    /* "pyboy/core/opcodes.py":5357
 *         return DEC_35(cpu)
 *     elif opcode == 0x36:
 *         return LD_36(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x37:
 *         return SCF_37(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_36(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5356
 *     elif opcode == 0x35:
 *         return DEC_35(cpu)
 *     elif opcode == 0x36:             # <<<<<<<<<<<<<<
 *         return LD_36(cpu, v)
 *     elif opcode == 0x37:
 */
    break;
    case 0x37:

    /* "pyboy/core/opcodes.py":5359
 *         return LD_36(cpu, v)
 *     elif opcode == 0x37:
 *         return SCF_37(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x38:
 *         return JR_38(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SCF_37(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5358
 *     elif opcode == 0x36:
 *         return LD_36(cpu, v)
 *     elif opcode == 0x37:             # <<<<<<<<<<<<<<
 *         return SCF_37(cpu)
 *     elif opcode == 0x38:
 */
    break;
    case 0x38:

    /* "pyboy/core/opcodes.py":5361
 *         return SCF_37(cpu)
 *     elif opcode == 0x38:
 *         return JR_38(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x39:
 *         return ADD_39(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JR_38(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5360
 *     elif opcode == 0x37:
 *         return SCF_37(cpu)
 *     elif opcode == 0x38:             # <<<<<<<<<<<<<<
 *         return JR_38(cpu, v)
 *     elif opcode == 0x39:
 */
    break;
    case 0x39:

    /* "pyboy/core/opcodes.py":5363
 *         return JR_38(cpu, v)
 *     elif opcode == 0x39:
 *         return ADD_39(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x3A:
 *         return LD_3A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_39(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5362
 *     elif opcode == 0x38:
 *         return JR_38(cpu, v)
 *     elif opcode == 0x39:             # <<<<<<<<<<<<<<
 *         return ADD_39(cpu)
 *     elif opcode == 0x3A:
 */
    break;
    case 0x3A:

    /* "pyboy/core/opcodes.py":5365
 *         return ADD_39(cpu)
 *     elif opcode == 0x3A:
 *         return LD_3A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x3B:
 *         return DEC_3B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_3A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5364
 *     elif opcode == 0x39:
 *         return ADD_39(cpu)
 *     elif opcode == 0x3A:             # <<<<<<<<<<<<<<
 *         return LD_3A(cpu)
 *     elif opcode == 0x3B:
 */
    break;
    case 0x3B:

    /* "pyboy/core/opcodes.py":5367
 *         return LD_3A(cpu)
 *     elif opcode == 0x3B:
 *         return DEC_3B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x3C:
 *         return INC_3C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_3B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5366
 *     elif opcode == 0x3A:
 *         return LD_3A(cpu)
 *     elif opcode == 0x3B:             # <<<<<<<<<<<<<<
 *         return DEC_3B(cpu)
 *     elif opcode == 0x3C:
 */
    break;
    case 0x3C:

    /* "pyboy/core/opcodes.py":5369
 *         return DEC_3B(cpu)
 *     elif opcode == 0x3C:
 *         return INC_3C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x3D:
 *         return DEC_3D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_INC_3C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5368
 *     elif opcode == 0x3B:
 *         return DEC_3B(cpu)
 *     elif opcode == 0x3C:             # <<<<<<<<<<<<<<
 *         return INC_3C(cpu)
 *     elif opcode == 0x3D:
 */
    break;
    case 0x3D:

    /* "pyboy/core/opcodes.py":5371
 *         return INC_3C(cpu)
 *     elif opcode == 0x3D:
 *         return DEC_3D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x3E:
 *         return LD_3E(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DEC_3D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5370
 *     elif opcode == 0x3C:
 *         return INC_3C(cpu)
 *     elif opcode == 0x3D:             # <<<<<<<<<<<<<<
 *         return DEC_3D(cpu)
 *     elif opcode == 0x3E:
 */
    break;
    case 0x3E:

    /* "pyboy/core/opcodes.py":5373
 *         return DEC_3D(cpu)
 *     elif opcode == 0x3E:
 *         return LD_3E(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x3F:
 *         return CCF_3F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_3E(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5372
 *     elif opcode == 0x3D:
 *         return DEC_3D(cpu)
 *     elif opcode == 0x3E:             # <<<<<<<<<<<<<<
 *         return LD_3E(cpu, v)
 *     elif opcode == 0x3F:
 */
    break;
    case 0x3F:

    /* "pyboy/core/opcodes.py":5375
 *         return LD_3E(cpu, v)
 *     elif opcode == 0x3F:
 *         return CCF_3F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x40:
 *         return LD_40(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CCF_3F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5374
 *     elif opcode == 0x3E:
 *         return LD_3E(cpu, v)
 *     elif opcode == 0x3F:             # <<<<<<<<<<<<<<
 *         return CCF_3F(cpu)
 *     elif opcode == 0x40:
 */
    break;
    case 0x40:

    /* "pyboy/core/opcodes.py":5377
 *         return CCF_3F(cpu)
 *     elif opcode == 0x40:
 *         return LD_40(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x41:
 *         return LD_41(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_40(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5376
 *     elif opcode == 0x3F:
 *         return CCF_3F(cpu)
 *     elif opcode == 0x40:             # <<<<<<<<<<<<<<
 *         return LD_40(cpu)
 *     elif opcode == 0x41:
 */
    break;
    case 0x41:

    /* "pyboy/core/opcodes.py":5379
 *         return LD_40(cpu)
 *     elif opcode == 0x41:
 *         return LD_41(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x42:
 *         return LD_42(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_41(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5378
 *     elif opcode == 0x40:
 *         return LD_40(cpu)
 *     elif opcode == 0x41:             # <<<<<<<<<<<<<<
 *         return LD_41(cpu)
 *     elif opcode == 0x42:
 */
    break;
    case 0x42:

    /* "pyboy/core/opcodes.py":5381
 *         return LD_41(cpu)
 *     elif opcode == 0x42:
 *         return LD_42(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x43:
 *         return LD_43(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_42(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5380
 *     elif opcode == 0x41:
 *         return LD_41(cpu)
 *     elif opcode == 0x42:             # <<<<<<<<<<<<<<
 *         return LD_42(cpu)
 *     elif opcode == 0x43:
 */
    break;
    case 0x43:

    /* "pyboy/core/opcodes.py":5383
 *         return LD_42(cpu)
 *     elif opcode == 0x43:
 *         return LD_43(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x44:
 *         return LD_44(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_43(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5382
 *     elif opcode == 0x42:
 *         return LD_42(cpu)
 *     elif opcode == 0x43:             # <<<<<<<<<<<<<<
 *         return LD_43(cpu)
 *     elif opcode == 0x44:
 */
    break;
    case 0x44:

    /* "pyboy/core/opcodes.py":5385
 *         return LD_43(cpu)
 *     elif opcode == 0x44:
 *         return LD_44(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x45:
 *         return LD_45(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_44(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5384
 *     elif opcode == 0x43:
 *         return LD_43(cpu)
 *     elif opcode == 0x44:             # <<<<<<<<<<<<<<
 *         return LD_44(cpu)
 *     elif opcode == 0x45:
 */
    break;
    case 0x45:

    /* "pyboy/core/opcodes.py":5387
 *         return LD_44(cpu)
 *     elif opcode == 0x45:
 *         return LD_45(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x46:
 *         return LD_46(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_45(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5386
 *     elif opcode == 0x44:
 *         return LD_44(cpu)
 *     elif opcode == 0x45:             # <<<<<<<<<<<<<<
 *         return LD_45(cpu)
 *     elif opcode == 0x46:
 */
    break;
    case 0x46:

    /* "pyboy/core/opcodes.py":5389
 *         return LD_45(cpu)
 *     elif opcode == 0x46:
 *         return LD_46(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x47:
 *         return LD_47(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_46(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5388
 *     elif opcode == 0x45:
 *         return LD_45(cpu)
 *     elif opcode == 0x46:             # <<<<<<<<<<<<<<
 *         return LD_46(cpu)
 *     elif opcode == 0x47:
 */
    break;
    case 0x47:

    /* "pyboy/core/opcodes.py":5391
 *         return LD_46(cpu)
 *     elif opcode == 0x47:
 *         return LD_47(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x48:
 *         return LD_48(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_47(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5390
 *     elif opcode == 0x46:
 *         return LD_46(cpu)
 *     elif opcode == 0x47:             # <<<<<<<<<<<<<<
 *         return LD_47(cpu)
 *     elif opcode == 0x48:
 */
    break;
    case 0x48:

    /* "pyboy/core/opcodes.py":5393
 *         return LD_47(cpu)
 *     elif opcode == 0x48:
 *         return LD_48(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x49:
 *         return LD_49(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_48(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5392
 *     elif opcode == 0x47:
 *         return LD_47(cpu)
 *     elif opcode == 0x48:             # <<<<<<<<<<<<<<
 *         return LD_48(cpu)
 *     elif opcode == 0x49:
 */
    break;
    case 0x49:

    /* "pyboy/core/opcodes.py":5395
 *         return LD_48(cpu)
 *     elif opcode == 0x49:
 *         return LD_49(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x4A:
 *         return LD_4A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_49(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5394
 *     elif opcode == 0x48:
 *         return LD_48(cpu)
 *     elif opcode == 0x49:             # <<<<<<<<<<<<<<
 *         return LD_49(cpu)
 *     elif opcode == 0x4A:
 */
    break;
    case 0x4A:

    /* "pyboy/core/opcodes.py":5397
 *         return LD_49(cpu)
 *     elif opcode == 0x4A:
 *         return LD_4A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x4B:
 *         return LD_4B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_4A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5396
 *     elif opcode == 0x49:
 *         return LD_49(cpu)
 *     elif opcode == 0x4A:             # <<<<<<<<<<<<<<
 *         return LD_4A(cpu)
 *     elif opcode == 0x4B:
 */
    break;
    case 0x4B:

    /* "pyboy/core/opcodes.py":5399
 *         return LD_4A(cpu)
 *     elif opcode == 0x4B:
 *         return LD_4B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x4C:
 *         return LD_4C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_4B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5398
 *     elif opcode == 0x4A:
 *         return LD_4A(cpu)
 *     elif opcode == 0x4B:             # <<<<<<<<<<<<<<
 *         return LD_4B(cpu)
 *     elif opcode == 0x4C:
 */
    break;
    case 0x4C:

    /* "pyboy/core/opcodes.py":5401
 *         return LD_4B(cpu)
 *     elif opcode == 0x4C:
 *         return LD_4C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x4D:
 *         return LD_4D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_4C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5400
 *     elif opcode == 0x4B:
 *         return LD_4B(cpu)
 *     elif opcode == 0x4C:             # <<<<<<<<<<<<<<
 *         return LD_4C(cpu)
 *     elif opcode == 0x4D:
 */
    break;
    case 0x4D:

    /* "pyboy/core/opcodes.py":5403
 *         return LD_4C(cpu)
 *     elif opcode == 0x4D:
 *         return LD_4D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x4E:
 *         return LD_4E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_4D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5402
 *     elif opcode == 0x4C:
 *         return LD_4C(cpu)
 *     elif opcode == 0x4D:             # <<<<<<<<<<<<<<
 *         return LD_4D(cpu)
 *     elif opcode == 0x4E:
 */
    break;
    case 0x4E:

    /* "pyboy/core/opcodes.py":5405
 *         return LD_4D(cpu)
 *     elif opcode == 0x4E:
 *         return LD_4E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x4F:
 *         return LD_4F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_4E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5404
 *     elif opcode == 0x4D:
 *         return LD_4D(cpu)
 *     elif opcode == 0x4E:             # <<<<<<<<<<<<<<
 *         return LD_4E(cpu)
 *     elif opcode == 0x4F:
 */
    break;
    case 0x4F:

    /* "pyboy/core/opcodes.py":5407
 *         return LD_4E(cpu)
 *     elif opcode == 0x4F:
 *         return LD_4F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x50:
 *         return LD_50(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_4F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5406
 *     elif opcode == 0x4E:
 *         return LD_4E(cpu)
 *     elif opcode == 0x4F:             # <<<<<<<<<<<<<<
 *         return LD_4F(cpu)
 *     elif opcode == 0x50:
 */
    break;
    case 0x50:

    /* "pyboy/core/opcodes.py":5409
 *         return LD_4F(cpu)
 *     elif opcode == 0x50:
 *         return LD_50(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x51:
 *         return LD_51(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_50(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5408
 *     elif opcode == 0x4F:
 *         return LD_4F(cpu)
 *     elif opcode == 0x50:             # <<<<<<<<<<<<<<
 *         return LD_50(cpu)
 *     elif opcode == 0x51:
 */
    break;
    case 0x51:

    /* "pyboy/core/opcodes.py":5411
 *         return LD_50(cpu)
 *     elif opcode == 0x51:
 *         return LD_51(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x52:
 *         return LD_52(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_51(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5410
 *     elif opcode == 0x50:
 *         return LD_50(cpu)
 *     elif opcode == 0x51:             # <<<<<<<<<<<<<<
 *         return LD_51(cpu)
 *     elif opcode == 0x52:
 */
    break;
    case 0x52:

    /* "pyboy/core/opcodes.py":5413
 *         return LD_51(cpu)
 *     elif opcode == 0x52:
 *         return LD_52(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x53:
 *         return LD_53(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_52(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5412
 *     elif opcode == 0x51:
 *         return LD_51(cpu)
 *     elif opcode == 0x52:             # <<<<<<<<<<<<<<
 *         return LD_52(cpu)
 *     elif opcode == 0x53:
 */
    break;
    case 0x53:

    /* "pyboy/core/opcodes.py":5415
 *         return LD_52(cpu)
 *     elif opcode == 0x53:
 *         return LD_53(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x54:
 *         return LD_54(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_53(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5414
 *     elif opcode == 0x52:
 *         return LD_52(cpu)
 *     elif opcode == 0x53:             # <<<<<<<<<<<<<<
 *         return LD_53(cpu)
 *     elif opcode == 0x54:
 */
    break;
    case 0x54:

    /* "pyboy/core/opcodes.py":5417
 *         return LD_53(cpu)
 *     elif opcode == 0x54:
 *         return LD_54(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x55:
 *         return LD_55(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_54(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5416
 *     elif opcode == 0x53:
 *         return LD_53(cpu)
 *     elif opcode == 0x54:             # <<<<<<<<<<<<<<
 *         return LD_54(cpu)
 *     elif opcode == 0x55:
 */
    break;
    case 0x55:

    /* "pyboy/core/opcodes.py":5419
 *         return LD_54(cpu)
 *     elif opcode == 0x55:
 *         return LD_55(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x56:
 *         return LD_56(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_55(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5418
 *     elif opcode == 0x54:
 *         return LD_54(cpu)
 *     elif opcode == 0x55:             # <<<<<<<<<<<<<<
 *         return LD_55(cpu)
 *     elif opcode == 0x56:
 */
    break;
    case 0x56:

    /* "pyboy/core/opcodes.py":5421
 *         return LD_55(cpu)
 *     elif opcode == 0x56:
 *         return LD_56(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x57:
 *         return LD_57(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_56(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5420
 *     elif opcode == 0x55:
 *         return LD_55(cpu)
 *     elif opcode == 0x56:             # <<<<<<<<<<<<<<
 *         return LD_56(cpu)
 *     elif opcode == 0x57:
 */
    break;
    case 0x57:

    /* "pyboy/core/opcodes.py":5423
 *         return LD_56(cpu)
 *     elif opcode == 0x57:
 *         return LD_57(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x58:
 *         return LD_58(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_57(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5422
 *     elif opcode == 0x56:
 *         return LD_56(cpu)
 *     elif opcode == 0x57:             # <<<<<<<<<<<<<<
 *         return LD_57(cpu)
 *     elif opcode == 0x58:
 */
    break;
    case 0x58:

    /* "pyboy/core/opcodes.py":5425
 *         return LD_57(cpu)
 *     elif opcode == 0x58:
 *         return LD_58(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x59:
 *         return LD_59(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_58(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5424
 *     elif opcode == 0x57:
 *         return LD_57(cpu)
 *     elif opcode == 0x58:             # <<<<<<<<<<<<<<
 *         return LD_58(cpu)
 *     elif opcode == 0x59:
 */
    break;
    case 0x59:

    /* "pyboy/core/opcodes.py":5427
 *         return LD_58(cpu)
 *     elif opcode == 0x59:
 *         return LD_59(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x5A:
 *         return LD_5A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_59(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5426
 *     elif opcode == 0x58:
 *         return LD_58(cpu)
 *     elif opcode == 0x59:             # <<<<<<<<<<<<<<
 *         return LD_59(cpu)
 *     elif opcode == 0x5A:
 */
    break;
    case 0x5A:

    /* "pyboy/core/opcodes.py":5429
 *         return LD_59(cpu)
 *     elif opcode == 0x5A:
 *         return LD_5A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x5B:
 *         return LD_5B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_5A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5428
 *     elif opcode == 0x59:
 *         return LD_59(cpu)
 *     elif opcode == 0x5A:             # <<<<<<<<<<<<<<
 *         return LD_5A(cpu)
 *     elif opcode == 0x5B:
 */
    break;
    case 0x5B:

    /* "pyboy/core/opcodes.py":5431
 *         return LD_5A(cpu)
 *     elif opcode == 0x5B:
 *         return LD_5B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x5C:
 *         return LD_5C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_5B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5430
 *     elif opcode == 0x5A:
 *         return LD_5A(cpu)
 *     elif opcode == 0x5B:             # <<<<<<<<<<<<<<
 *         return LD_5B(cpu)
 *     elif opcode == 0x5C:
 */
    break;
    case 0x5C:

    /* "pyboy/core/opcodes.py":5433
 *         return LD_5B(cpu)
 *     elif opcode == 0x5C:
 *         return LD_5C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x5D:
 *         return LD_5D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_5C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5432
 *     elif opcode == 0x5B:
 *         return LD_5B(cpu)
 *     elif opcode == 0x5C:             # <<<<<<<<<<<<<<
 *         return LD_5C(cpu)
 *     elif opcode == 0x5D:
 */
    break;
    case 0x5D:

    /* "pyboy/core/opcodes.py":5435
 *         return LD_5C(cpu)
 *     elif opcode == 0x5D:
 *         return LD_5D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x5E:
 *         return LD_5E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_5D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5434
 *     elif opcode == 0x5C:
 *         return LD_5C(cpu)
 *     elif opcode == 0x5D:             # <<<<<<<<<<<<<<
 *         return LD_5D(cpu)
 *     elif opcode == 0x5E:
 */
    break;
    case 0x5E:

    /* "pyboy/core/opcodes.py":5437
 *         return LD_5D(cpu)
 *     elif opcode == 0x5E:
 *         return LD_5E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x5F:
 *         return LD_5F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_5E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5436
 *     elif opcode == 0x5D:
 *         return LD_5D(cpu)
 *     elif opcode == 0x5E:             # <<<<<<<<<<<<<<
 *         return LD_5E(cpu)
 *     elif opcode == 0x5F:
 */
    break;
    case 0x5F:

    /* "pyboy/core/opcodes.py":5439
 *         return LD_5E(cpu)
 *     elif opcode == 0x5F:
 *         return LD_5F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x60:
 *         return LD_60(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_5F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5438
 *     elif opcode == 0x5E:
 *         return LD_5E(cpu)
 *     elif opcode == 0x5F:             # <<<<<<<<<<<<<<
 *         return LD_5F(cpu)
 *     elif opcode == 0x60:
 */
    break;
    case 0x60:

    /* "pyboy/core/opcodes.py":5441
 *         return LD_5F(cpu)
 *     elif opcode == 0x60:
 *         return LD_60(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x61:
 *         return LD_61(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_60(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5440
 *     elif opcode == 0x5F:
 *         return LD_5F(cpu)
 *     elif opcode == 0x60:             # <<<<<<<<<<<<<<
 *         return LD_60(cpu)
 *     elif opcode == 0x61:
 */
    break;
    case 0x61:

    /* "pyboy/core/opcodes.py":5443
 *         return LD_60(cpu)
 *     elif opcode == 0x61:
 *         return LD_61(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x62:
 *         return LD_62(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_61(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5442
 *     elif opcode == 0x60:
 *         return LD_60(cpu)
 *     elif opcode == 0x61:             # <<<<<<<<<<<<<<
 *         return LD_61(cpu)
 *     elif opcode == 0x62:
 */
    break;
    case 0x62:

    /* "pyboy/core/opcodes.py":5445
 *         return LD_61(cpu)
 *     elif opcode == 0x62:
 *         return LD_62(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x63:
 *         return LD_63(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_62(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5444
 *     elif opcode == 0x61:
 *         return LD_61(cpu)
 *     elif opcode == 0x62:             # <<<<<<<<<<<<<<
 *         return LD_62(cpu)
 *     elif opcode == 0x63:
 */
    break;
    case 0x63:

    /* "pyboy/core/opcodes.py":5447
 *         return LD_62(cpu)
 *     elif opcode == 0x63:
 *         return LD_63(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x64:
 *         return LD_64(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_63(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5446
 *     elif opcode == 0x62:
 *         return LD_62(cpu)
 *     elif opcode == 0x63:             # <<<<<<<<<<<<<<
 *         return LD_63(cpu)
 *     elif opcode == 0x64:
 */
    break;
    case 0x64:

    /* "pyboy/core/opcodes.py":5449
 *         return LD_63(cpu)
 *     elif opcode == 0x64:
 *         return LD_64(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x65:
 *         return LD_65(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_64(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5448
 *     elif opcode == 0x63:
 *         return LD_63(cpu)
 *     elif opcode == 0x64:             # <<<<<<<<<<<<<<
 *         return LD_64(cpu)
 *     elif opcode == 0x65:
 */
    break;
    case 0x65:

    /* "pyboy/core/opcodes.py":5451
 *         return LD_64(cpu)
 *     elif opcode == 0x65:
 *         return LD_65(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x66:
 *         return LD_66(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_65(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5450
 *     elif opcode == 0x64:
 *         return LD_64(cpu)
 *     elif opcode == 0x65:             # <<<<<<<<<<<<<<
 *         return LD_65(cpu)
 *     elif opcode == 0x66:
 */
    break;
    case 0x66:

    /* "pyboy/core/opcodes.py":5453
 *         return LD_65(cpu)
 *     elif opcode == 0x66:
 *         return LD_66(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x67:
 *         return LD_67(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_66(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5452
 *     elif opcode == 0x65:
 *         return LD_65(cpu)
 *     elif opcode == 0x66:             # <<<<<<<<<<<<<<
 *         return LD_66(cpu)
 *     elif opcode == 0x67:
 */
    break;
    case 0x67:

    /* "pyboy/core/opcodes.py":5455
 *         return LD_66(cpu)
 *     elif opcode == 0x67:
 *         return LD_67(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x68:
 *         return LD_68(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_67(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5454
 *     elif opcode == 0x66:
 *         return LD_66(cpu)
 *     elif opcode == 0x67:             # <<<<<<<<<<<<<<
 *         return LD_67(cpu)
 *     elif opcode == 0x68:
 */
    break;
    case 0x68:

    /* "pyboy/core/opcodes.py":5457
 *         return LD_67(cpu)
 *     elif opcode == 0x68:
 *         return LD_68(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x69:
 *         return LD_69(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_68(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5456
 *     elif opcode == 0x67:
 *         return LD_67(cpu)
 *     elif opcode == 0x68:             # <<<<<<<<<<<<<<
 *         return LD_68(cpu)
 *     elif opcode == 0x69:
 */
    break;
    case 0x69:

    /* "pyboy/core/opcodes.py":5459
 *         return LD_68(cpu)
 *     elif opcode == 0x69:
 *         return LD_69(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x6A:
 *         return LD_6A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_69(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5458
 *     elif opcode == 0x68:
 *         return LD_68(cpu)
 *     elif opcode == 0x69:             # <<<<<<<<<<<<<<
 *         return LD_69(cpu)
 *     elif opcode == 0x6A:
 */
    break;
    case 0x6A:

    /* "pyboy/core/opcodes.py":5461
 *         return LD_69(cpu)
 *     elif opcode == 0x6A:
 *         return LD_6A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x6B:
 *         return LD_6B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_6A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5460
 *     elif opcode == 0x69:
 *         return LD_69(cpu)
 *     elif opcode == 0x6A:             # <<<<<<<<<<<<<<
 *         return LD_6A(cpu)
 *     elif opcode == 0x6B:
 */
    break;
    case 0x6B:

    /* "pyboy/core/opcodes.py":5463
 *         return LD_6A(cpu)
 *     elif opcode == 0x6B:
 *         return LD_6B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x6C:
 *         return LD_6C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_6B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5462
 *     elif opcode == 0x6A:
 *         return LD_6A(cpu)
 *     elif opcode == 0x6B:             # <<<<<<<<<<<<<<
 *         return LD_6B(cpu)
 *     elif opcode == 0x6C:
 */
    break;
    case 0x6C:

    /* "pyboy/core/opcodes.py":5465
 *         return LD_6B(cpu)
 *     elif opcode == 0x6C:
 *         return LD_6C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x6D:
 *         return LD_6D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_6C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5464
 *     elif opcode == 0x6B:
 *         return LD_6B(cpu)
 *     elif opcode == 0x6C:             # <<<<<<<<<<<<<<
 *         return LD_6C(cpu)
 *     elif opcode == 0x6D:
 */
    break;
    case 0x6D:

    /* "pyboy/core/opcodes.py":5467
 *         return LD_6C(cpu)
 *     elif opcode == 0x6D:
 *         return LD_6D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x6E:
 *         return LD_6E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_6D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5466
 *     elif opcode == 0x6C:
 *         return LD_6C(cpu)
 *     elif opcode == 0x6D:             # <<<<<<<<<<<<<<
 *         return LD_6D(cpu)
 *     elif opcode == 0x6E:
 */
    break;
    case 0x6E:

    /* "pyboy/core/opcodes.py":5469
 *         return LD_6D(cpu)
 *     elif opcode == 0x6E:
 *         return LD_6E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x6F:
 *         return LD_6F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_6E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5468
 *     elif opcode == 0x6D:
 *         return LD_6D(cpu)
 *     elif opcode == 0x6E:             # <<<<<<<<<<<<<<
 *         return LD_6E(cpu)
 *     elif opcode == 0x6F:
 */
    break;
    case 0x6F:

    /* "pyboy/core/opcodes.py":5471
 *         return LD_6E(cpu)
 *     elif opcode == 0x6F:
 *         return LD_6F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x70:
 *         return LD_70(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_6F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5470
 *     elif opcode == 0x6E:
 *         return LD_6E(cpu)
 *     elif opcode == 0x6F:             # <<<<<<<<<<<<<<
 *         return LD_6F(cpu)
 *     elif opcode == 0x70:
 */
    break;
    case 0x70:

    /* "pyboy/core/opcodes.py":5473
 *         return LD_6F(cpu)
 *     elif opcode == 0x70:
 *         return LD_70(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x71:
 *         return LD_71(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_70(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5472
 *     elif opcode == 0x6F:
 *         return LD_6F(cpu)
 *     elif opcode == 0x70:             # <<<<<<<<<<<<<<
 *         return LD_70(cpu)
 *     elif opcode == 0x71:
 */
    break;
    case 0x71:

    /* "pyboy/core/opcodes.py":5475
 *         return LD_70(cpu)
 *     elif opcode == 0x71:
 *         return LD_71(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x72:
 *         return LD_72(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_71(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5474
 *     elif opcode == 0x70:
 *         return LD_70(cpu)
 *     elif opcode == 0x71:             # <<<<<<<<<<<<<<
 *         return LD_71(cpu)
 *     elif opcode == 0x72:
 */
    break;
    case 0x72:

    /* "pyboy/core/opcodes.py":5477
 *         return LD_71(cpu)
 *     elif opcode == 0x72:
 *         return LD_72(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x73:
 *         return LD_73(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_72(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5476
 *     elif opcode == 0x71:
 *         return LD_71(cpu)
 *     elif opcode == 0x72:             # <<<<<<<<<<<<<<
 *         return LD_72(cpu)
 *     elif opcode == 0x73:
 */
    break;
    case 0x73:

    /* "pyboy/core/opcodes.py":5479
 *         return LD_72(cpu)
 *     elif opcode == 0x73:
 *         return LD_73(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x74:
 *         return LD_74(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_73(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5478
 *     elif opcode == 0x72:
 *         return LD_72(cpu)
 *     elif opcode == 0x73:             # <<<<<<<<<<<<<<
 *         return LD_73(cpu)
 *     elif opcode == 0x74:
 */
    break;
    case 0x74:

    /* "pyboy/core/opcodes.py":5481
 *         return LD_73(cpu)
 *     elif opcode == 0x74:
 *         return LD_74(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x75:
 *         return LD_75(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_74(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5480
 *     elif opcode == 0x73:
 *         return LD_73(cpu)
 *     elif opcode == 0x74:             # <<<<<<<<<<<<<<
 *         return LD_74(cpu)
 *     elif opcode == 0x75:
 */
    break;
    case 0x75:

    /* "pyboy/core/opcodes.py":5483
 *         return LD_74(cpu)
 *     elif opcode == 0x75:
 *         return LD_75(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x76:
 *         return HALT_76(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_75(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5482
 *     elif opcode == 0x74:
 *         return LD_74(cpu)
 *     elif opcode == 0x75:             # <<<<<<<<<<<<<<
 *         return LD_75(cpu)
 *     elif opcode == 0x76:
 */
    break;
    case 0x76:

    /* "pyboy/core/opcodes.py":5485
 *         return LD_75(cpu)
 *     elif opcode == 0x76:
 *         return HALT_76(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x77:
 *         return LD_77(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_HALT_76(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5484
 *     elif opcode == 0x75:
 *         return LD_75(cpu)
 *     elif opcode == 0x76:             # <<<<<<<<<<<<<<
 *         return HALT_76(cpu)
 *     elif opcode == 0x77:
 */
    break;
    case 0x77:

    /* "pyboy/core/opcodes.py":5487
 *         return HALT_76(cpu)
 *     elif opcode == 0x77:
 *         return LD_77(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x78:
 *         return LD_78(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_77(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5486
 *     elif opcode == 0x76:
 *         return HALT_76(cpu)
 *     elif opcode == 0x77:             # <<<<<<<<<<<<<<
 *         return LD_77(cpu)
 *     elif opcode == 0x78:
 */
    break;
    case 0x78:

    /* "pyboy/core/opcodes.py":5489
 *         return LD_77(cpu)
 *     elif opcode == 0x78:
 *         return LD_78(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x79:
 *         return LD_79(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_78(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5488
 *     elif opcode == 0x77:
 *         return LD_77(cpu)
 *     elif opcode == 0x78:             # <<<<<<<<<<<<<<
 *         return LD_78(cpu)
 *     elif opcode == 0x79:
 */
    break;
    case 0x79:

    /* "pyboy/core/opcodes.py":5491
 *         return LD_78(cpu)
 *     elif opcode == 0x79:
 *         return LD_79(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x7A:
 *         return LD_7A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_79(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5490
 *     elif opcode == 0x78:
 *         return LD_78(cpu)
 *     elif opcode == 0x79:             # <<<<<<<<<<<<<<
 *         return LD_79(cpu)
 *     elif opcode == 0x7A:
 */
    break;
    case 0x7A:

    /* "pyboy/core/opcodes.py":5493
 *         return LD_79(cpu)
 *     elif opcode == 0x7A:
 *         return LD_7A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x7B:
 *         return LD_7B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_7A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5492
 *     elif opcode == 0x79:
 *         return LD_79(cpu)
 *     elif opcode == 0x7A:             # <<<<<<<<<<<<<<
 *         return LD_7A(cpu)
 *     elif opcode == 0x7B:
 */
    break;
    case 0x7B:

    /* "pyboy/core/opcodes.py":5495
 *         return LD_7A(cpu)
 *     elif opcode == 0x7B:
 *         return LD_7B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x7C:
 *         return LD_7C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_7B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5494
 *     elif opcode == 0x7A:
 *         return LD_7A(cpu)
 *     elif opcode == 0x7B:             # <<<<<<<<<<<<<<
 *         return LD_7B(cpu)
 *     elif opcode == 0x7C:
 */
    break;
    case 0x7C:

    /* "pyboy/core/opcodes.py":5497
 *         return LD_7B(cpu)
 *     elif opcode == 0x7C:
 *         return LD_7C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x7D:
 *         return LD_7D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_7C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5496
 *     elif opcode == 0x7B:
 *         return LD_7B(cpu)
 *     elif opcode == 0x7C:             # <<<<<<<<<<<<<<
 *         return LD_7C(cpu)
 *     elif opcode == 0x7D:
 */
    break;
    case 0x7D:

    /* "pyboy/core/opcodes.py":5499
 *         return LD_7C(cpu)
 *     elif opcode == 0x7D:
 *         return LD_7D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x7E:
 *         return LD_7E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_7D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5498
 *     elif opcode == 0x7C:
 *         return LD_7C(cpu)
 *     elif opcode == 0x7D:             # <<<<<<<<<<<<<<
 *         return LD_7D(cpu)
 *     elif opcode == 0x7E:
 */
    break;
    case 0x7E:

    /* "pyboy/core/opcodes.py":5501
 *         return LD_7D(cpu)
 *     elif opcode == 0x7E:
 *         return LD_7E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x7F:
 *         return LD_7F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_7E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5500
 *     elif opcode == 0x7D:
 *         return LD_7D(cpu)
 *     elif opcode == 0x7E:             # <<<<<<<<<<<<<<
 *         return LD_7E(cpu)
 *     elif opcode == 0x7F:
 */
    break;
    case 0x7F:

    /* "pyboy/core/opcodes.py":5503
 *         return LD_7E(cpu)
 *     elif opcode == 0x7F:
 *         return LD_7F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x80:
 *         return ADD_80(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_7F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5502
 *     elif opcode == 0x7E:
 *         return LD_7E(cpu)
 *     elif opcode == 0x7F:             # <<<<<<<<<<<<<<
 *         return LD_7F(cpu)
 *     elif opcode == 0x80:
 */
    break;
    case 0x80:

    /* "pyboy/core/opcodes.py":5505
 *         return LD_7F(cpu)
 *     elif opcode == 0x80:
 *         return ADD_80(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x81:
 *         return ADD_81(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_80(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5504
 *     elif opcode == 0x7F:
 *         return LD_7F(cpu)
 *     elif opcode == 0x80:             # <<<<<<<<<<<<<<
 *         return ADD_80(cpu)
 *     elif opcode == 0x81:
 */
    break;
    case 0x81:

    /* "pyboy/core/opcodes.py":5507
 *         return ADD_80(cpu)
 *     elif opcode == 0x81:
 *         return ADD_81(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x82:
 *         return ADD_82(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_81(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5506
 *     elif opcode == 0x80:
 *         return ADD_80(cpu)
 *     elif opcode == 0x81:             # <<<<<<<<<<<<<<
 *         return ADD_81(cpu)
 *     elif opcode == 0x82:
 */
    break;
    case 0x82:

    /* "pyboy/core/opcodes.py":5509
 *         return ADD_81(cpu)
 *     elif opcode == 0x82:
 *         return ADD_82(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x83:
 *         return ADD_83(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_82(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5508
 *     elif opcode == 0x81:
 *         return ADD_81(cpu)
 *     elif opcode == 0x82:             # <<<<<<<<<<<<<<
 *         return ADD_82(cpu)
 *     elif opcode == 0x83:
 */
    break;
    case 0x83:

    /* "pyboy/core/opcodes.py":5511
 *         return ADD_82(cpu)
 *     elif opcode == 0x83:
 *         return ADD_83(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x84:
 *         return ADD_84(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_83(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5510
 *     elif opcode == 0x82:
 *         return ADD_82(cpu)
 *     elif opcode == 0x83:             # <<<<<<<<<<<<<<
 *         return ADD_83(cpu)
 *     elif opcode == 0x84:
 */
    break;
    case 0x84:

    /* "pyboy/core/opcodes.py":5513
 *         return ADD_83(cpu)
 *     elif opcode == 0x84:
 *         return ADD_84(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x85:
 *         return ADD_85(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_84(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5512
 *     elif opcode == 0x83:
 *         return ADD_83(cpu)
 *     elif opcode == 0x84:             # <<<<<<<<<<<<<<
 *         return ADD_84(cpu)
 *     elif opcode == 0x85:
 */
    break;
    case 0x85:

    /* "pyboy/core/opcodes.py":5515
 *         return ADD_84(cpu)
 *     elif opcode == 0x85:
 *         return ADD_85(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x86:
 *         return ADD_86(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_85(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5514
 *     elif opcode == 0x84:
 *         return ADD_84(cpu)
 *     elif opcode == 0x85:             # <<<<<<<<<<<<<<
 *         return ADD_85(cpu)
 *     elif opcode == 0x86:
 */
    break;
    case 0x86:

    /* "pyboy/core/opcodes.py":5517
 *         return ADD_85(cpu)
 *     elif opcode == 0x86:
 *         return ADD_86(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x87:
 *         return ADD_87(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_86(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5516
 *     elif opcode == 0x85:
 *         return ADD_85(cpu)
 *     elif opcode == 0x86:             # <<<<<<<<<<<<<<
 *         return ADD_86(cpu)
 *     elif opcode == 0x87:
 */
    break;
    case 0x87:

    /* "pyboy/core/opcodes.py":5519
 *         return ADD_86(cpu)
 *     elif opcode == 0x87:
 *         return ADD_87(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x88:
 *         return ADC_88(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_87(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5518
 *     elif opcode == 0x86:
 *         return ADD_86(cpu)
 *     elif opcode == 0x87:             # <<<<<<<<<<<<<<
 *         return ADD_87(cpu)
 *     elif opcode == 0x88:
 */
    break;
    case 0x88:

    /* "pyboy/core/opcodes.py":5521
 *         return ADD_87(cpu)
 *     elif opcode == 0x88:
 *         return ADC_88(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x89:
 *         return ADC_89(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADC_88(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5520
 *     elif opcode == 0x87:
 *         return ADD_87(cpu)
 *     elif opcode == 0x88:             # <<<<<<<<<<<<<<
 *         return ADC_88(cpu)
 *     elif opcode == 0x89:
 */
    break;
    case 0x89:

    /* "pyboy/core/opcodes.py":5523
 *         return ADC_88(cpu)
 *     elif opcode == 0x89:
 *         return ADC_89(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x8A:
 *         return ADC_8A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADC_89(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5522
 *     elif opcode == 0x88:
 *         return ADC_88(cpu)
 *     elif opcode == 0x89:             # <<<<<<<<<<<<<<
 *         return ADC_89(cpu)
 *     elif opcode == 0x8A:
 */
    break;
    case 0x8A:

    /* "pyboy/core/opcodes.py":5525
 *         return ADC_89(cpu)
 *     elif opcode == 0x8A:
 *         return ADC_8A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x8B:
 *         return ADC_8B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADC_8A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5524
 *     elif opcode == 0x89:
 *         return ADC_89(cpu)
 *     elif opcode == 0x8A:             # <<<<<<<<<<<<<<
 *         return ADC_8A(cpu)
 *     elif opcode == 0x8B:
 */
    break;
    case 0x8B:

    /* "pyboy/core/opcodes.py":5527
 *         return ADC_8A(cpu)
 *     elif opcode == 0x8B:
 *         return ADC_8B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x8C:
 *         return ADC_8C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADC_8B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5526
 *     elif opcode == 0x8A:
 *         return ADC_8A(cpu)
 *     elif opcode == 0x8B:             # <<<<<<<<<<<<<<
 *         return ADC_8B(cpu)
 *     elif opcode == 0x8C:
 */
    break;
    case 0x8C:

    /* "pyboy/core/opcodes.py":5529
 *         return ADC_8B(cpu)
 *     elif opcode == 0x8C:
 *         return ADC_8C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x8D:
 *         return ADC_8D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADC_8C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5528
 *     elif opcode == 0x8B:
 *         return ADC_8B(cpu)
 *     elif opcode == 0x8C:             # <<<<<<<<<<<<<<
 *         return ADC_8C(cpu)
 *     elif opcode == 0x8D:
 */
    break;
    case 0x8D:

    /* "pyboy/core/opcodes.py":5531
 *         return ADC_8C(cpu)
 *     elif opcode == 0x8D:
 *         return ADC_8D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x8E:
 *         return ADC_8E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADC_8D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5530
 *     elif opcode == 0x8C:
 *         return ADC_8C(cpu)
 *     elif opcode == 0x8D:             # <<<<<<<<<<<<<<
 *         return ADC_8D(cpu)
 *     elif opcode == 0x8E:
 */
    break;
    case 0x8E:

    /* "pyboy/core/opcodes.py":5533
 *         return ADC_8D(cpu)
 *     elif opcode == 0x8E:
 *         return ADC_8E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x8F:
 *         return ADC_8F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADC_8E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5532
 *     elif opcode == 0x8D:
 *         return ADC_8D(cpu)
 *     elif opcode == 0x8E:             # <<<<<<<<<<<<<<
 *         return ADC_8E(cpu)
 *     elif opcode == 0x8F:
 */
    break;
    case 0x8F:

    /* "pyboy/core/opcodes.py":5535
 *         return ADC_8E(cpu)
 *     elif opcode == 0x8F:
 *         return ADC_8F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x90:
 *         return SUB_90(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADC_8F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5534
 *     elif opcode == 0x8E:
 *         return ADC_8E(cpu)
 *     elif opcode == 0x8F:             # <<<<<<<<<<<<<<
 *         return ADC_8F(cpu)
 *     elif opcode == 0x90:
 */
    break;
    case 0x90:

    /* "pyboy/core/opcodes.py":5537
 *         return ADC_8F(cpu)
 *     elif opcode == 0x90:
 *         return SUB_90(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x91:
 *         return SUB_91(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SUB_90(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5536
 *     elif opcode == 0x8F:
 *         return ADC_8F(cpu)
 *     elif opcode == 0x90:             # <<<<<<<<<<<<<<
 *         return SUB_90(cpu)
 *     elif opcode == 0x91:
 */
    break;
    case 0x91:

    /* "pyboy/core/opcodes.py":5539
 *         return SUB_90(cpu)
 *     elif opcode == 0x91:
 *         return SUB_91(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x92:
 *         return SUB_92(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SUB_91(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5538
 *     elif opcode == 0x90:
 *         return SUB_90(cpu)
 *     elif opcode == 0x91:             # <<<<<<<<<<<<<<
 *         return SUB_91(cpu)
 *     elif opcode == 0x92:
 */
    break;
    case 0x92:

    /* "pyboy/core/opcodes.py":5541
 *         return SUB_91(cpu)
 *     elif opcode == 0x92:
 *         return SUB_92(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x93:
 *         return SUB_93(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SUB_92(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5540
 *     elif opcode == 0x91:
 *         return SUB_91(cpu)
 *     elif opcode == 0x92:             # <<<<<<<<<<<<<<
 *         return SUB_92(cpu)
 *     elif opcode == 0x93:
 */
    break;
    case 0x93:

    /* "pyboy/core/opcodes.py":5543
 *         return SUB_92(cpu)
 *     elif opcode == 0x93:
 *         return SUB_93(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x94:
 *         return SUB_94(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SUB_93(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5542
 *     elif opcode == 0x92:
 *         return SUB_92(cpu)
 *     elif opcode == 0x93:             # <<<<<<<<<<<<<<
 *         return SUB_93(cpu)
 *     elif opcode == 0x94:
 */
    break;
    case 0x94:

    /* "pyboy/core/opcodes.py":5545
 *         return SUB_93(cpu)
 *     elif opcode == 0x94:
 *         return SUB_94(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x95:
 *         return SUB_95(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SUB_94(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5544
 *     elif opcode == 0x93:
 *         return SUB_93(cpu)
 *     elif opcode == 0x94:             # <<<<<<<<<<<<<<
 *         return SUB_94(cpu)
 *     elif opcode == 0x95:
 */
    break;
    case 0x95:

    /* "pyboy/core/opcodes.py":5547
 *         return SUB_94(cpu)
 *     elif opcode == 0x95:
 *         return SUB_95(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x96:
 *         return SUB_96(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SUB_95(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5546
 *     elif opcode == 0x94:
 *         return SUB_94(cpu)
 *     elif opcode == 0x95:             # <<<<<<<<<<<<<<
 *         return SUB_95(cpu)
 *     elif opcode == 0x96:
 */
    break;
    case 0x96:

    /* "pyboy/core/opcodes.py":5549
 *         return SUB_95(cpu)
 *     elif opcode == 0x96:
 *         return SUB_96(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x97:
 *         return SUB_97(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SUB_96(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5548
 *     elif opcode == 0x95:
 *         return SUB_95(cpu)
 *     elif opcode == 0x96:             # <<<<<<<<<<<<<<
 *         return SUB_96(cpu)
 *     elif opcode == 0x97:
 */
    break;
    case 0x97:

    /* "pyboy/core/opcodes.py":5551
 *         return SUB_96(cpu)
 *     elif opcode == 0x97:
 *         return SUB_97(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x98:
 *         return SBC_98(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SUB_97(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5550
 *     elif opcode == 0x96:
 *         return SUB_96(cpu)
 *     elif opcode == 0x97:             # <<<<<<<<<<<<<<
 *         return SUB_97(cpu)
 *     elif opcode == 0x98:
 */
    break;
    case 0x98:

    /* "pyboy/core/opcodes.py":5553
 *         return SUB_97(cpu)
 *     elif opcode == 0x98:
 *         return SBC_98(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x99:
 *         return SBC_99(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SBC_98(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5552
 *     elif opcode == 0x97:
 *         return SUB_97(cpu)
 *     elif opcode == 0x98:             # <<<<<<<<<<<<<<
 *         return SBC_98(cpu)
 *     elif opcode == 0x99:
 */
    break;
    case 0x99:

    /* "pyboy/core/opcodes.py":5555
 *         return SBC_98(cpu)
 *     elif opcode == 0x99:
 *         return SBC_99(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x9A:
 *         return SBC_9A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SBC_99(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5554
 *     elif opcode == 0x98:
 *         return SBC_98(cpu)
 *     elif opcode == 0x99:             # <<<<<<<<<<<<<<
 *         return SBC_99(cpu)
 *     elif opcode == 0x9A:
 */
    break;
    case 0x9A:

    /* "pyboy/core/opcodes.py":5557
 *         return SBC_99(cpu)
 *     elif opcode == 0x9A:
 *         return SBC_9A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x9B:
 *         return SBC_9B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SBC_9A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5556
 *     elif opcode == 0x99:
 *         return SBC_99(cpu)
 *     elif opcode == 0x9A:             # <<<<<<<<<<<<<<
 *         return SBC_9A(cpu)
 *     elif opcode == 0x9B:
 */
    break;
    case 0x9B:

    /* "pyboy/core/opcodes.py":5559
 *         return SBC_9A(cpu)
 *     elif opcode == 0x9B:
 *         return SBC_9B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x9C:
 *         return SBC_9C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SBC_9B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5558
 *     elif opcode == 0x9A:
 *         return SBC_9A(cpu)
 *     elif opcode == 0x9B:             # <<<<<<<<<<<<<<
 *         return SBC_9B(cpu)
 *     elif opcode == 0x9C:
 */
    break;
    case 0x9C:

    /* "pyboy/core/opcodes.py":5561
 *         return SBC_9B(cpu)
 *     elif opcode == 0x9C:
 *         return SBC_9C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x9D:
 *         return SBC_9D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SBC_9C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5560
 *     elif opcode == 0x9B:
 *         return SBC_9B(cpu)
 *     elif opcode == 0x9C:             # <<<<<<<<<<<<<<
 *         return SBC_9C(cpu)
 *     elif opcode == 0x9D:
 */
    break;
    case 0x9D:

    /* "pyboy/core/opcodes.py":5563
 *         return SBC_9C(cpu)
 *     elif opcode == 0x9D:
 *         return SBC_9D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x9E:
 *         return SBC_9E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SBC_9D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5562
 *     elif opcode == 0x9C:
 *         return SBC_9C(cpu)
 *     elif opcode == 0x9D:             # <<<<<<<<<<<<<<
 *         return SBC_9D(cpu)
 *     elif opcode == 0x9E:
 */
    break;
    case 0x9E:

    /* "pyboy/core/opcodes.py":5565
 *         return SBC_9D(cpu)
 *     elif opcode == 0x9E:
 *         return SBC_9E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x9F:
 *         return SBC_9F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SBC_9E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5564
 *     elif opcode == 0x9D:
 *         return SBC_9D(cpu)
 *     elif opcode == 0x9E:             # <<<<<<<<<<<<<<
 *         return SBC_9E(cpu)
 *     elif opcode == 0x9F:
 */
    break;
    case 0x9F:

    /* "pyboy/core/opcodes.py":5567
 *         return SBC_9E(cpu)
 *     elif opcode == 0x9F:
 *         return SBC_9F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA0:
 *         return AND_A0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SBC_9F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5566
 *     elif opcode == 0x9E:
 *         return SBC_9E(cpu)
 *     elif opcode == 0x9F:             # <<<<<<<<<<<<<<
 *         return SBC_9F(cpu)
 *     elif opcode == 0xA0:
 */
    break;
    case 0xA0:

    /* "pyboy/core/opcodes.py":5569
 *         return SBC_9F(cpu)
 *     elif opcode == 0xA0:
 *         return AND_A0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA1:
 *         return AND_A1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_AND_A0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5568
 *     elif opcode == 0x9F:
 *         return SBC_9F(cpu)
 *     elif opcode == 0xA0:             # <<<<<<<<<<<<<<
 *         return AND_A0(cpu)
 *     elif opcode == 0xA1:
 */
    break;
    case 0xA1:

    /* "pyboy/core/opcodes.py":5571
 *         return AND_A0(cpu)
 *     elif opcode == 0xA1:
 *         return AND_A1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA2:
 *         return AND_A2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_AND_A1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5570
 *     elif opcode == 0xA0:
 *         return AND_A0(cpu)
 *     elif opcode == 0xA1:             # <<<<<<<<<<<<<<
 *         return AND_A1(cpu)
 *     elif opcode == 0xA2:
 */
    break;
    case 0xA2:

    /* "pyboy/core/opcodes.py":5573
 *         return AND_A1(cpu)
 *     elif opcode == 0xA2:
 *         return AND_A2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA3:
 *         return AND_A3(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_AND_A2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5572
 *     elif opcode == 0xA1:
 *         return AND_A1(cpu)
 *     elif opcode == 0xA2:             # <<<<<<<<<<<<<<
 *         return AND_A2(cpu)
 *     elif opcode == 0xA3:
 */
    break;
    case 0xA3:

    /* "pyboy/core/opcodes.py":5575
 *         return AND_A2(cpu)
 *     elif opcode == 0xA3:
 *         return AND_A3(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA4:
 *         return AND_A4(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_AND_A3(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5574
 *     elif opcode == 0xA2:
 *         return AND_A2(cpu)
 *     elif opcode == 0xA3:             # <<<<<<<<<<<<<<
 *         return AND_A3(cpu)
 *     elif opcode == 0xA4:
 */
    break;
    case 0xA4:

    /* "pyboy/core/opcodes.py":5577
 *         return AND_A3(cpu)
 *     elif opcode == 0xA4:
 *         return AND_A4(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA5:
 *         return AND_A5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_AND_A4(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5576
 *     elif opcode == 0xA3:
 *         return AND_A3(cpu)
 *     elif opcode == 0xA4:             # <<<<<<<<<<<<<<
 *         return AND_A4(cpu)
 *     elif opcode == 0xA5:
 */
    break;
    case 0xA5:

    /* "pyboy/core/opcodes.py":5579
 *         return AND_A4(cpu)
 *     elif opcode == 0xA5:
 *         return AND_A5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA6:
 *         return AND_A6(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_AND_A5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5578
 *     elif opcode == 0xA4:
 *         return AND_A4(cpu)
 *     elif opcode == 0xA5:             # <<<<<<<<<<<<<<
 *         return AND_A5(cpu)
 *     elif opcode == 0xA6:
 */
    break;
    case 0xA6:

    /* "pyboy/core/opcodes.py":5581
 *         return AND_A5(cpu)
 *     elif opcode == 0xA6:
 *         return AND_A6(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA7:
 *         return AND_A7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_AND_A6(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5580
 *     elif opcode == 0xA5:
 *         return AND_A5(cpu)
 *     elif opcode == 0xA6:             # <<<<<<<<<<<<<<
 *         return AND_A6(cpu)
 *     elif opcode == 0xA7:
 */
    break;
    case 0xA7:

    /* "pyboy/core/opcodes.py":5583
 *         return AND_A6(cpu)
 *     elif opcode == 0xA7:
 *         return AND_A7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA8:
 *         return XOR_A8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_AND_A7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5582
 *     elif opcode == 0xA6:
 *         return AND_A6(cpu)
 *     elif opcode == 0xA7:             # <<<<<<<<<<<<<<
 *         return AND_A7(cpu)
 *     elif opcode == 0xA8:
 */
    break;
    case 0xA8:

    /* "pyboy/core/opcodes.py":5585
 *         return AND_A7(cpu)
 *     elif opcode == 0xA8:
 *         return XOR_A8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xA9:
 *         return XOR_A9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_XOR_A8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5584
 *     elif opcode == 0xA7:
 *         return AND_A7(cpu)
 *     elif opcode == 0xA8:             # <<<<<<<<<<<<<<
 *         return XOR_A8(cpu)
 *     elif opcode == 0xA9:
 */
    break;
    case 0xA9:

    /* "pyboy/core/opcodes.py":5587
 *         return XOR_A8(cpu)
 *     elif opcode == 0xA9:
 *         return XOR_A9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xAA:
 *         return XOR_AA(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_XOR_A9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5586
 *     elif opcode == 0xA8:
 *         return XOR_A8(cpu)
 *     elif opcode == 0xA9:             # <<<<<<<<<<<<<<
 *         return XOR_A9(cpu)
 *     elif opcode == 0xAA:
 */
    break;
    case 0xAA:

    /* "pyboy/core/opcodes.py":5589
 *         return XOR_A9(cpu)
 *     elif opcode == 0xAA:
 *         return XOR_AA(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xAB:
 *         return XOR_AB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_XOR_AA(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5588
 *     elif opcode == 0xA9:
 *         return XOR_A9(cpu)
 *     elif opcode == 0xAA:             # <<<<<<<<<<<<<<
 *         return XOR_AA(cpu)
 *     elif opcode == 0xAB:
 */
    break;
    case 0xAB:

    /* "pyboy/core/opcodes.py":5591
 *         return XOR_AA(cpu)
 *     elif opcode == 0xAB:
 *         return XOR_AB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xAC:
 *         return XOR_AC(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_XOR_AB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5590
 *     elif opcode == 0xAA:
 *         return XOR_AA(cpu)
 *     elif opcode == 0xAB:             # <<<<<<<<<<<<<<
 *         return XOR_AB(cpu)
 *     elif opcode == 0xAC:
 */
    break;
    case 0xAC:

    /* "pyboy/core/opcodes.py":5593
 *         return XOR_AB(cpu)
 *     elif opcode == 0xAC:
 *         return XOR_AC(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xAD:
 *         return XOR_AD(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_XOR_AC(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5592
 *     elif opcode == 0xAB:
 *         return XOR_AB(cpu)
 *     elif opcode == 0xAC:             # <<<<<<<<<<<<<<
 *         return XOR_AC(cpu)
 *     elif opcode == 0xAD:
 */
    break;
    case 0xAD:

    /* "pyboy/core/opcodes.py":5595
 *         return XOR_AC(cpu)
 *     elif opcode == 0xAD:
 *         return XOR_AD(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xAE:
 *         return XOR_AE(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_XOR_AD(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5594
 *     elif opcode == 0xAC:
 *         return XOR_AC(cpu)
 *     elif opcode == 0xAD:             # <<<<<<<<<<<<<<
 *         return XOR_AD(cpu)
 *     elif opcode == 0xAE:
 */
    break;
    case 0xAE:

    /* "pyboy/core/opcodes.py":5597
 *         return XOR_AD(cpu)
 *     elif opcode == 0xAE:
 *         return XOR_AE(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xAF:
 *         return XOR_AF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_XOR_AE(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5596
 *     elif opcode == 0xAD:
 *         return XOR_AD(cpu)
 *     elif opcode == 0xAE:             # <<<<<<<<<<<<<<
 *         return XOR_AE(cpu)
 *     elif opcode == 0xAF:
 */
    break;
    case 0xAF:

    /* "pyboy/core/opcodes.py":5599
 *         return XOR_AE(cpu)
 *     elif opcode == 0xAF:
 *         return XOR_AF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB0:
 *         return OR_B0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_XOR_AF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5598
 *     elif opcode == 0xAE:
 *         return XOR_AE(cpu)
 *     elif opcode == 0xAF:             # <<<<<<<<<<<<<<
 *         return XOR_AF(cpu)
 *     elif opcode == 0xB0:
 */
    break;
    case 0xB0:

    /* "pyboy/core/opcodes.py":5601
 *         return XOR_AF(cpu)
 *     elif opcode == 0xB0:
 *         return OR_B0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB1:
 *         return OR_B1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_OR_B0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5600
 *     elif opcode == 0xAF:
 *         return XOR_AF(cpu)
 *     elif opcode == 0xB0:             # <<<<<<<<<<<<<<
 *         return OR_B0(cpu)
 *     elif opcode == 0xB1:
 */
    break;
    case 0xB1:

    /* "pyboy/core/opcodes.py":5603
 *         return OR_B0(cpu)
 *     elif opcode == 0xB1:
 *         return OR_B1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB2:
 *         return OR_B2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_OR_B1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5602
 *     elif opcode == 0xB0:
 *         return OR_B0(cpu)
 *     elif opcode == 0xB1:             # <<<<<<<<<<<<<<
 *         return OR_B1(cpu)
 *     elif opcode == 0xB2:
 */
    break;
    case 0xB2:

    /* "pyboy/core/opcodes.py":5605
 *         return OR_B1(cpu)
 *     elif opcode == 0xB2:
 *         return OR_B2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB3:
 *         return OR_B3(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_OR_B2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5604
 *     elif opcode == 0xB1:
 *         return OR_B1(cpu)
 *     elif opcode == 0xB2:             # <<<<<<<<<<<<<<
 *         return OR_B2(cpu)
 *     elif opcode == 0xB3:
 */
    break;
    case 0xB3:

    /* "pyboy/core/opcodes.py":5607
 *         return OR_B2(cpu)
 *     elif opcode == 0xB3:
 *         return OR_B3(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB4:
 *         return OR_B4(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_OR_B3(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5606
 *     elif opcode == 0xB2:
 *         return OR_B2(cpu)
 *     elif opcode == 0xB3:             # <<<<<<<<<<<<<<
 *         return OR_B3(cpu)
 *     elif opcode == 0xB4:
 */
    break;
    case 0xB4:

    /* "pyboy/core/opcodes.py":5609
 *         return OR_B3(cpu)
 *     elif opcode == 0xB4:
 *         return OR_B4(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB5:
 *         return OR_B5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_OR_B4(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5608
 *     elif opcode == 0xB3:
 *         return OR_B3(cpu)
 *     elif opcode == 0xB4:             # <<<<<<<<<<<<<<
 *         return OR_B4(cpu)
 *     elif opcode == 0xB5:
 */
    break;
    case 0xB5:

    /* "pyboy/core/opcodes.py":5611
 *         return OR_B4(cpu)
 *     elif opcode == 0xB5:
 *         return OR_B5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB6:
 *         return OR_B6(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_OR_B5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5610
 *     elif opcode == 0xB4:
 *         return OR_B4(cpu)
 *     elif opcode == 0xB5:             # <<<<<<<<<<<<<<
 *         return OR_B5(cpu)
 *     elif opcode == 0xB6:
 */
    break;
    case 0xB6:

    /* "pyboy/core/opcodes.py":5613
 *         return OR_B5(cpu)
 *     elif opcode == 0xB6:
 *         return OR_B6(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB7:
 *         return OR_B7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_OR_B6(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5612
 *     elif opcode == 0xB5:
 *         return OR_B5(cpu)
 *     elif opcode == 0xB6:             # <<<<<<<<<<<<<<
 *         return OR_B6(cpu)
 *     elif opcode == 0xB7:
 */
    break;
    case 0xB7:

    /* "pyboy/core/opcodes.py":5615
 *         return OR_B6(cpu)
 *     elif opcode == 0xB7:
 *         return OR_B7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB8:
 *         return CP_B8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_OR_B7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5614
 *     elif opcode == 0xB6:
 *         return OR_B6(cpu)
 *     elif opcode == 0xB7:             # <<<<<<<<<<<<<<
 *         return OR_B7(cpu)
 *     elif opcode == 0xB8:
 */
    break;
    case 0xB8:

    /* "pyboy/core/opcodes.py":5617
 *         return OR_B7(cpu)
 *     elif opcode == 0xB8:
 *         return CP_B8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xB9:
 *         return CP_B9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CP_B8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5616
 *     elif opcode == 0xB7:
 *         return OR_B7(cpu)
 *     elif opcode == 0xB8:             # <<<<<<<<<<<<<<
 *         return CP_B8(cpu)
 *     elif opcode == 0xB9:
 */
    break;
    case 0xB9:

    /* "pyboy/core/opcodes.py":5619
 *         return CP_B8(cpu)
 *     elif opcode == 0xB9:
 *         return CP_B9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xBA:
 *         return CP_BA(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CP_B9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5618
 *     elif opcode == 0xB8:
 *         return CP_B8(cpu)
 *     elif opcode == 0xB9:             # <<<<<<<<<<<<<<
 *         return CP_B9(cpu)
 *     elif opcode == 0xBA:
 */
    break;
    case 0xBA:

    /* "pyboy/core/opcodes.py":5621
 *         return CP_B9(cpu)
 *     elif opcode == 0xBA:
 *         return CP_BA(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xBB:
 *         return CP_BB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CP_BA(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5620
 *     elif opcode == 0xB9:
 *         return CP_B9(cpu)
 *     elif opcode == 0xBA:             # <<<<<<<<<<<<<<
 *         return CP_BA(cpu)
 *     elif opcode == 0xBB:
 */
    break;
    case 0xBB:

    /* "pyboy/core/opcodes.py":5623
 *         return CP_BA(cpu)
 *     elif opcode == 0xBB:
 *         return CP_BB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xBC:
 *         return CP_BC(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CP_BB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5622
 *     elif opcode == 0xBA:
 *         return CP_BA(cpu)
 *     elif opcode == 0xBB:             # <<<<<<<<<<<<<<
 *         return CP_BB(cpu)
 *     elif opcode == 0xBC:
 */
    break;
    case 0xBC:

    /* "pyboy/core/opcodes.py":5625
 *         return CP_BB(cpu)
 *     elif opcode == 0xBC:
 *         return CP_BC(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xBD:
 *         return CP_BD(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CP_BC(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5624
 *     elif opcode == 0xBB:
 *         return CP_BB(cpu)
 *     elif opcode == 0xBC:             # <<<<<<<<<<<<<<
 *         return CP_BC(cpu)
 *     elif opcode == 0xBD:
 */
    break;
    case 0xBD:

    /* "pyboy/core/opcodes.py":5627
 *         return CP_BC(cpu)
 *     elif opcode == 0xBD:
 *         return CP_BD(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xBE:
 *         return CP_BE(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CP_BD(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5626
 *     elif opcode == 0xBC:
 *         return CP_BC(cpu)
 *     elif opcode == 0xBD:             # <<<<<<<<<<<<<<
 *         return CP_BD(cpu)
 *     elif opcode == 0xBE:
 */
    break;
    case 0xBE:

    /* "pyboy/core/opcodes.py":5629
 *         return CP_BD(cpu)
 *     elif opcode == 0xBE:
 *         return CP_BE(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xBF:
 *         return CP_BF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CP_BE(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5628
 *     elif opcode == 0xBD:
 *         return CP_BD(cpu)
 *     elif opcode == 0xBE:             # <<<<<<<<<<<<<<
 *         return CP_BE(cpu)
 *     elif opcode == 0xBF:
 */
    break;
    case 0xBF:

    /* "pyboy/core/opcodes.py":5631
 *         return CP_BE(cpu)
 *     elif opcode == 0xBF:
 *         return CP_BF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC0:
 *         return RET_C0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CP_BF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5630
 *     elif opcode == 0xBE:
 *         return CP_BE(cpu)
 *     elif opcode == 0xBF:             # <<<<<<<<<<<<<<
 *         return CP_BF(cpu)
 *     elif opcode == 0xC0:
 */
    break;
    case 0xC0:

    /* "pyboy/core/opcodes.py":5633
 *         return CP_BF(cpu)
 *     elif opcode == 0xC0:
 *         return RET_C0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC1:
 *         return POP_C1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RET_C0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5632
 *     elif opcode == 0xBF:
 *         return CP_BF(cpu)
 *     elif opcode == 0xC0:             # <<<<<<<<<<<<<<
 *         return RET_C0(cpu)
 *     elif opcode == 0xC1:
 */
    break;
    case 0xC1:

    /* "pyboy/core/opcodes.py":5635
 *         return RET_C0(cpu)
 *     elif opcode == 0xC1:
 *         return POP_C1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC2:
 *         return JP_C2(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_POP_C1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5634
 *     elif opcode == 0xC0:
 *         return RET_C0(cpu)
 *     elif opcode == 0xC1:             # <<<<<<<<<<<<<<
 *         return POP_C1(cpu)
 *     elif opcode == 0xC2:
 */
    break;
    case 0xC2:

    /* "pyboy/core/opcodes.py":5637
 *         return POP_C1(cpu)
 *     elif opcode == 0xC2:
 *         return JP_C2(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC3:
 *         return JP_C3(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JP_C2(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5636
 *     elif opcode == 0xC1:
 *         return POP_C1(cpu)
 *     elif opcode == 0xC2:             # <<<<<<<<<<<<<<
 *         return JP_C2(cpu, v)
 *     elif opcode == 0xC3:
 */
    break;
    case 0xC3:

    /* "pyboy/core/opcodes.py":5639
 *         return JP_C2(cpu, v)
 *     elif opcode == 0xC3:
 *         return JP_C3(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC4:
 *         return CALL_C4(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JP_C3(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5638
 *     elif opcode == 0xC2:
 *         return JP_C2(cpu, v)
 *     elif opcode == 0xC3:             # <<<<<<<<<<<<<<
 *         return JP_C3(cpu, v)
 *     elif opcode == 0xC4:
 */
    break;
    case 0xC4:

    /* "pyboy/core/opcodes.py":5641
 *         return JP_C3(cpu, v)
 *     elif opcode == 0xC4:
 *         return CALL_C4(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC5:
 *         return PUSH_C5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CALL_C4(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5640
 *     elif opcode == 0xC3:
 *         return JP_C3(cpu, v)
 *     elif opcode == 0xC4:             # <<<<<<<<<<<<<<
 *         return CALL_C4(cpu, v)
 *     elif opcode == 0xC5:
 */
    break;
    case 0xC5:

    /* "pyboy/core/opcodes.py":5643
 *         return CALL_C4(cpu, v)
 *     elif opcode == 0xC5:
 *         return PUSH_C5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC6:
 *         return ADD_C6(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_PUSH_C5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5642
 *     elif opcode == 0xC4:
 *         return CALL_C4(cpu, v)
 *     elif opcode == 0xC5:             # <<<<<<<<<<<<<<
 *         return PUSH_C5(cpu)
 *     elif opcode == 0xC6:
 */
    break;
    case 0xC6:

    /* "pyboy/core/opcodes.py":5645
 *         return PUSH_C5(cpu)
 *     elif opcode == 0xC6:
 *         return ADD_C6(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC7:
 *         return RST_C7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_C6(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5644
 *     elif opcode == 0xC5:
 *         return PUSH_C5(cpu)
 *     elif opcode == 0xC6:             # <<<<<<<<<<<<<<
 *         return ADD_C6(cpu, v)
 *     elif opcode == 0xC7:
 */
    break;
    case 0xC7:

    /* "pyboy/core/opcodes.py":5647
 *         return ADD_C6(cpu, v)
 *     elif opcode == 0xC7:
 *         return RST_C7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC8:
 *         return RET_C8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RST_C7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5646
 *     elif opcode == 0xC6:
 *         return ADD_C6(cpu, v)
 *     elif opcode == 0xC7:             # <<<<<<<<<<<<<<
 *         return RST_C7(cpu)
 *     elif opcode == 0xC8:
 */
    break;
    case 0xC8:

    /* "pyboy/core/opcodes.py":5649
 *         return RST_C7(cpu)
 *     elif opcode == 0xC8:
 *         return RET_C8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xC9:
 *         return RET_C9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RET_C8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5648
 *     elif opcode == 0xC7:
 *         return RST_C7(cpu)
 *     elif opcode == 0xC8:             # <<<<<<<<<<<<<<
 *         return RET_C8(cpu)
 *     elif opcode == 0xC9:
 */
    break;
    case 0xC9:

    /* "pyboy/core/opcodes.py":5651
 *         return RET_C8(cpu)
 *     elif opcode == 0xC9:
 *         return RET_C9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xCA:
 *         return JP_CA(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RET_C9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5650
 *     elif opcode == 0xC8:
 *         return RET_C8(cpu)
 *     elif opcode == 0xC9:             # <<<<<<<<<<<<<<
 *         return RET_C9(cpu)
 *     elif opcode == 0xCA:
 */
    break;
    case 0xCA:

    /* "pyboy/core/opcodes.py":5653
 *         return RET_C9(cpu)
 *     elif opcode == 0xCA:
 *         return JP_CA(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xCB:
 *         return PREFIX_CB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JP_CA(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5652
 *     elif opcode == 0xC9:
 *         return RET_C9(cpu)
 *     elif opcode == 0xCA:             # <<<<<<<<<<<<<<
 *         return JP_CA(cpu, v)
 *     elif opcode == 0xCB:
 */
    break;
    case 0xCB:

    /* "pyboy/core/opcodes.py":5655
 *         return JP_CA(cpu, v)
 *     elif opcode == 0xCB:
 *         return PREFIX_CB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xCC:
 *         return CALL_CC(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_PREFIX_CB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5654
 *     elif opcode == 0xCA:
 *         return JP_CA(cpu, v)
 *     elif opcode == 0xCB:             # <<<<<<<<<<<<<<
 *         return PREFIX_CB(cpu)
 *     elif opcode == 0xCC:
 */
    break;
    case 0xCC:

    /* "pyboy/core/opcodes.py":5657
 *         return PREFIX_CB(cpu)
 *     elif opcode == 0xCC:
 *         return CALL_CC(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xCD:
 *         return CALL_CD(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CALL_CC(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5656
 *     elif opcode == 0xCB:
 *         return PREFIX_CB(cpu)
 *     elif opcode == 0xCC:             # <<<<<<<<<<<<<<
 *         return CALL_CC(cpu, v)
 *     elif opcode == 0xCD:
 */
    break;
    case 0xCD:

    /* "pyboy/core/opcodes.py":5659
 *         return CALL_CC(cpu, v)
 *     elif opcode == 0xCD:
 *         return CALL_CD(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xCE:
 *         return ADC_CE(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CALL_CD(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5658
 *     elif opcode == 0xCC:
 *         return CALL_CC(cpu, v)
 *     elif opcode == 0xCD:             # <<<<<<<<<<<<<<
 *         return CALL_CD(cpu, v)
 *     elif opcode == 0xCE:
 */
    break;
    case 0xCE:

    /* "pyboy/core/opcodes.py":5661
 *         return CALL_CD(cpu, v)
 *     elif opcode == 0xCE:
 *         return ADC_CE(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xCF:
 *         return RST_CF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADC_CE(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5660
 *     elif opcode == 0xCD:
 *         return CALL_CD(cpu, v)
 *     elif opcode == 0xCE:             # <<<<<<<<<<<<<<
 *         return ADC_CE(cpu, v)
 *     elif opcode == 0xCF:
 */
    break;
    case 0xCF:

    /* "pyboy/core/opcodes.py":5663
 *         return ADC_CE(cpu, v)
 *     elif opcode == 0xCF:
 *         return RST_CF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD0:
 *         return RET_D0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RST_CF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5662
 *     elif opcode == 0xCE:
 *         return ADC_CE(cpu, v)
 *     elif opcode == 0xCF:             # <<<<<<<<<<<<<<
 *         return RST_CF(cpu)
 *     elif opcode == 0xD0:
 */
    break;
    case 0xD0:

    /* "pyboy/core/opcodes.py":5665
 *         return RST_CF(cpu)
 *     elif opcode == 0xD0:
 *         return RET_D0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD1:
 *         return POP_D1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RET_D0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5664
 *     elif opcode == 0xCF:
 *         return RST_CF(cpu)
 *     elif opcode == 0xD0:             # <<<<<<<<<<<<<<
 *         return RET_D0(cpu)
 *     elif opcode == 0xD1:
 */
    break;
    case 0xD1:

    /* "pyboy/core/opcodes.py":5667
 *         return RET_D0(cpu)
 *     elif opcode == 0xD1:
 *         return POP_D1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD2:
 *         return JP_D2(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_POP_D1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5666
 *     elif opcode == 0xD0:
 *         return RET_D0(cpu)
 *     elif opcode == 0xD1:             # <<<<<<<<<<<<<<
 *         return POP_D1(cpu)
 *     elif opcode == 0xD2:
 */
    break;
    case 0xD2:

    /* "pyboy/core/opcodes.py":5669
 *         return POP_D1(cpu)
 *     elif opcode == 0xD2:
 *         return JP_D2(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD3:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JP_D2(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5668
 *     elif opcode == 0xD1:
 *         return POP_D1(cpu)
 *     elif opcode == 0xD2:             # <<<<<<<<<<<<<<
 *         return JP_D2(cpu, v)
 *     elif opcode == 0xD3:
 */
    break;
    case 0xD3:

    /* "pyboy/core/opcodes.py":5671
 *         return JP_D2(cpu, v)
 *     elif opcode == 0xD3:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD4:
 *         return CALL_D4(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5670
 *     elif opcode == 0xD2:
 *         return JP_D2(cpu, v)
 *     elif opcode == 0xD3:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xD4:
 */
    break;
    case 0xD4:

    /* "pyboy/core/opcodes.py":5673
 *         return no_opcode(cpu)
 *     elif opcode == 0xD4:
 *         return CALL_D4(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD5:
 *         return PUSH_D5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CALL_D4(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5672
 *     elif opcode == 0xD3:
 *         return no_opcode(cpu)
 *     elif opcode == 0xD4:             # <<<<<<<<<<<<<<
 *         return CALL_D4(cpu, v)
 *     elif opcode == 0xD5:
 */
    break;
    case 0xD5:

    /* "pyboy/core/opcodes.py":5675
 *         return CALL_D4(cpu, v)
 *     elif opcode == 0xD5:
 *         return PUSH_D5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD6:
 *         return SUB_D6(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_PUSH_D5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5674
 *     elif opcode == 0xD4:
 *         return CALL_D4(cpu, v)
 *     elif opcode == 0xD5:             # <<<<<<<<<<<<<<
 *         return PUSH_D5(cpu)
 *     elif opcode == 0xD6:
 */
    break;
    case 0xD6:

    /* "pyboy/core/opcodes.py":5677
 *         return PUSH_D5(cpu)
 *     elif opcode == 0xD6:
 *         return SUB_D6(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD7:
 *         return RST_D7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SUB_D6(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5676
 *     elif opcode == 0xD5:
 *         return PUSH_D5(cpu)
 *     elif opcode == 0xD6:             # <<<<<<<<<<<<<<
 *         return SUB_D6(cpu, v)
 *     elif opcode == 0xD7:
 */
    break;
    case 0xD7:

    /* "pyboy/core/opcodes.py":5679
 *         return SUB_D6(cpu, v)
 *     elif opcode == 0xD7:
 *         return RST_D7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD8:
 *         return RET_D8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RST_D7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5678
 *     elif opcode == 0xD6:
 *         return SUB_D6(cpu, v)
 *     elif opcode == 0xD7:             # <<<<<<<<<<<<<<
 *         return RST_D7(cpu)
 *     elif opcode == 0xD8:
 */
    break;
    case 0xD8:

    /* "pyboy/core/opcodes.py":5681
 *         return RST_D7(cpu)
 *     elif opcode == 0xD8:
 *         return RET_D8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xD9:
 *         return RETI_D9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RET_D8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5680
 *     elif opcode == 0xD7:
 *         return RST_D7(cpu)
 *     elif opcode == 0xD8:             # <<<<<<<<<<<<<<
 *         return RET_D8(cpu)
 *     elif opcode == 0xD9:
 */
    break;
    case 0xD9:

    /* "pyboy/core/opcodes.py":5683
 *         return RET_D8(cpu)
 *     elif opcode == 0xD9:
 *         return RETI_D9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xDA:
 *         return JP_DA(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RETI_D9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5682
 *     elif opcode == 0xD8:
 *         return RET_D8(cpu)
 *     elif opcode == 0xD9:             # <<<<<<<<<<<<<<
 *         return RETI_D9(cpu)
 *     elif opcode == 0xDA:
 */
    break;
    case 0xDA:

    /* "pyboy/core/opcodes.py":5685
 *         return RETI_D9(cpu)
 *     elif opcode == 0xDA:
 *         return JP_DA(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xDB:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JP_DA(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5684
 *     elif opcode == 0xD9:
 *         return RETI_D9(cpu)
 *     elif opcode == 0xDA:             # <<<<<<<<<<<<<<
 *         return JP_DA(cpu, v)
 *     elif opcode == 0xDB:
 */
    break;
    case 0xDB:

    /* "pyboy/core/opcodes.py":5687
 *         return JP_DA(cpu, v)
 *     elif opcode == 0xDB:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xDC:
 *         return CALL_DC(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5686
 *     elif opcode == 0xDA:
 *         return JP_DA(cpu, v)
 *     elif opcode == 0xDB:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xDC:
 */
    break;
    case 0xDC:

    /* "pyboy/core/opcodes.py":5689
 *         return no_opcode(cpu)
 *     elif opcode == 0xDC:
 *         return CALL_DC(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xDD:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CALL_DC(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5688
 *     elif opcode == 0xDB:
 *         return no_opcode(cpu)
 *     elif opcode == 0xDC:             # <<<<<<<<<<<<<<
 *         return CALL_DC(cpu, v)
 *     elif opcode == 0xDD:
 */
    break;
    case 0xDD:

    /* "pyboy/core/opcodes.py":5691
 *         return CALL_DC(cpu, v)
 *     elif opcode == 0xDD:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xDE:
 *         return SBC_DE(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5690
 *     elif opcode == 0xDC:
 *         return CALL_DC(cpu, v)
 *     elif opcode == 0xDD:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xDE:
 */
    break;
    case 0xDE:

    /* "pyboy/core/opcodes.py":5693
 *         return no_opcode(cpu)
 *     elif opcode == 0xDE:
 *         return SBC_DE(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xDF:
 *         return RST_DF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SBC_DE(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5692
 *     elif opcode == 0xDD:
 *         return no_opcode(cpu)
 *     elif opcode == 0xDE:             # <<<<<<<<<<<<<<
 *         return SBC_DE(cpu, v)
 *     elif opcode == 0xDF:
 */
    break;
    case 0xDF:

    /* "pyboy/core/opcodes.py":5695
 *         return SBC_DE(cpu, v)
 *     elif opcode == 0xDF:
 *         return RST_DF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE0:
 *         return LDH_E0(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RST_DF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5694
 *     elif opcode == 0xDE:
 *         return SBC_DE(cpu, v)
 *     elif opcode == 0xDF:             # <<<<<<<<<<<<<<
 *         return RST_DF(cpu)
 *     elif opcode == 0xE0:
 */
    break;
    case 0xE0:

    /* "pyboy/core/opcodes.py":5697
 *         return RST_DF(cpu)
 *     elif opcode == 0xE0:
 *         return LDH_E0(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE1:
 *         return POP_E1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LDH_E0(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5696
 *     elif opcode == 0xDF:
 *         return RST_DF(cpu)
 *     elif opcode == 0xE0:             # <<<<<<<<<<<<<<
 *         return LDH_E0(cpu, v)
 *     elif opcode == 0xE1:
 */
    break;
    case 0xE1:

    /* "pyboy/core/opcodes.py":5699
 *         return LDH_E0(cpu, v)
 *     elif opcode == 0xE1:
 *         return POP_E1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE2:
 *         return LD_E2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_POP_E1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5698
 *     elif opcode == 0xE0:
 *         return LDH_E0(cpu, v)
 *     elif opcode == 0xE1:             # <<<<<<<<<<<<<<
 *         return POP_E1(cpu)
 *     elif opcode == 0xE2:
 */
    break;
    case 0xE2:

    /* "pyboy/core/opcodes.py":5701
 *         return POP_E1(cpu)
 *     elif opcode == 0xE2:
 *         return LD_E2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE3:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_E2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5700
 *     elif opcode == 0xE1:
 *         return POP_E1(cpu)
 *     elif opcode == 0xE2:             # <<<<<<<<<<<<<<
 *         return LD_E2(cpu)
 *     elif opcode == 0xE3:
 */
    break;
    case 0xE3:

    /* "pyboy/core/opcodes.py":5703
 *         return LD_E2(cpu)
 *     elif opcode == 0xE3:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE4:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5702
 *     elif opcode == 0xE2:
 *         return LD_E2(cpu)
 *     elif opcode == 0xE3:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xE4:
 */
    break;
    case 0xE4:

    /* "pyboy/core/opcodes.py":5705
 *         return no_opcode(cpu)
 *     elif opcode == 0xE4:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE5:
 *         return PUSH_E5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5704
 *     elif opcode == 0xE3:
 *         return no_opcode(cpu)
 *     elif opcode == 0xE4:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xE5:
 */
    break;
    case 0xE5:

    /* "pyboy/core/opcodes.py":5707
 *         return no_opcode(cpu)
 *     elif opcode == 0xE5:
 *         return PUSH_E5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE6:
 *         return AND_E6(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_PUSH_E5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5706
 *     elif opcode == 0xE4:
 *         return no_opcode(cpu)
 *     elif opcode == 0xE5:             # <<<<<<<<<<<<<<
 *         return PUSH_E5(cpu)
 *     elif opcode == 0xE6:
 */
    break;
    case 0xE6:

    /* "pyboy/core/opcodes.py":5709
 *         return PUSH_E5(cpu)
 *     elif opcode == 0xE6:
 *         return AND_E6(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE7:
 *         return RST_E7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_AND_E6(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5708
 *     elif opcode == 0xE5:
 *         return PUSH_E5(cpu)
 *     elif opcode == 0xE6:             # <<<<<<<<<<<<<<
 *         return AND_E6(cpu, v)
 *     elif opcode == 0xE7:
 */
    break;
    case 0xE7:

    /* "pyboy/core/opcodes.py":5711
 *         return AND_E6(cpu, v)
 *     elif opcode == 0xE7:
 *         return RST_E7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE8:
 *         return ADD_E8(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RST_E7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5710
 *     elif opcode == 0xE6:
 *         return AND_E6(cpu, v)
 *     elif opcode == 0xE7:             # <<<<<<<<<<<<<<
 *         return RST_E7(cpu)
 *     elif opcode == 0xE8:
 */
    break;
    case 0xE8:

    /* "pyboy/core/opcodes.py":5713
 *         return RST_E7(cpu)
 *     elif opcode == 0xE8:
 *         return ADD_E8(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xE9:
 *         return JP_E9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_ADD_E8(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5712
 *     elif opcode == 0xE7:
 *         return RST_E7(cpu)
 *     elif opcode == 0xE8:             # <<<<<<<<<<<<<<
 *         return ADD_E8(cpu, v)
 *     elif opcode == 0xE9:
 */
    break;
    case 0xE9:

    /* "pyboy/core/opcodes.py":5715
 *         return ADD_E8(cpu, v)
 *     elif opcode == 0xE9:
 *         return JP_E9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xEA:
 *         return LD_EA(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_JP_E9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5714
 *     elif opcode == 0xE8:
 *         return ADD_E8(cpu, v)
 *     elif opcode == 0xE9:             # <<<<<<<<<<<<<<
 *         return JP_E9(cpu)
 *     elif opcode == 0xEA:
 */
    break;
    case 0xEA:

    /* "pyboy/core/opcodes.py":5717
 *         return JP_E9(cpu)
 *     elif opcode == 0xEA:
 *         return LD_EA(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xEB:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_EA(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5716
 *     elif opcode == 0xE9:
 *         return JP_E9(cpu)
 *     elif opcode == 0xEA:             # <<<<<<<<<<<<<<
 *         return LD_EA(cpu, v)
 *     elif opcode == 0xEB:
 */
    break;
    case 0xEB:

    /* "pyboy/core/opcodes.py":5719
 *         return LD_EA(cpu, v)
 *     elif opcode == 0xEB:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xEC:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5718
 *     elif opcode == 0xEA:
 *         return LD_EA(cpu, v)
 *     elif opcode == 0xEB:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xEC:
 */
    break;
    case 0xEC:

    /* "pyboy/core/opcodes.py":5721
 *         return no_opcode(cpu)
 *     elif opcode == 0xEC:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xED:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5720
 *     elif opcode == 0xEB:
 *         return no_opcode(cpu)
 *     elif opcode == 0xEC:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xED:
 */
    break;
    case 0xED:

    /* "pyboy/core/opcodes.py":5723
 *         return no_opcode(cpu)
 *     elif opcode == 0xED:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xEE:
 *         return XOR_EE(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5722
 *     elif opcode == 0xEC:
 *         return no_opcode(cpu)
 *     elif opcode == 0xED:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xEE:
 */
    break;
    case 0xEE:

    /* "pyboy/core/opcodes.py":5725
 *         return no_opcode(cpu)
 *     elif opcode == 0xEE:
 *         return XOR_EE(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xEF:
 *         return RST_EF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_XOR_EE(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5724
 *     elif opcode == 0xED:
 *         return no_opcode(cpu)
 *     elif opcode == 0xEE:             # <<<<<<<<<<<<<<
 *         return XOR_EE(cpu, v)
 *     elif opcode == 0xEF:
 */
    break;
    case 0xEF:

    /* "pyboy/core/opcodes.py":5727
 *         return XOR_EE(cpu, v)
 *     elif opcode == 0xEF:
 *         return RST_EF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF0:
 *         return LDH_F0(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RST_EF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5726
 *     elif opcode == 0xEE:
 *         return XOR_EE(cpu, v)
 *     elif opcode == 0xEF:             # <<<<<<<<<<<<<<
 *         return RST_EF(cpu)
 *     elif opcode == 0xF0:
 */
    break;
    case 0xF0:

    /* "pyboy/core/opcodes.py":5729
 *         return RST_EF(cpu)
 *     elif opcode == 0xF0:
 *         return LDH_F0(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF1:
 *         return POP_F1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LDH_F0(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5728
 *     elif opcode == 0xEF:
 *         return RST_EF(cpu)
 *     elif opcode == 0xF0:             # <<<<<<<<<<<<<<
 *         return LDH_F0(cpu, v)
 *     elif opcode == 0xF1:
 */
    break;
    case 0xF1:

    /* "pyboy/core/opcodes.py":5731
 *         return LDH_F0(cpu, v)
 *     elif opcode == 0xF1:
 *         return POP_F1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF2:
 *         return LD_F2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_POP_F1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5730
 *     elif opcode == 0xF0:
 *         return LDH_F0(cpu, v)
 *     elif opcode == 0xF1:             # <<<<<<<<<<<<<<
 *         return POP_F1(cpu)
 *     elif opcode == 0xF2:
 */
    break;
    case 0xF2:

    /* "pyboy/core/opcodes.py":5733
 *         return POP_F1(cpu)
 *     elif opcode == 0xF2:
 *         return LD_F2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF3:
 *         return DI_F3(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_F2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5732
 *     elif opcode == 0xF1:
 *         return POP_F1(cpu)
 *     elif opcode == 0xF2:             # <<<<<<<<<<<<<<
 *         return LD_F2(cpu)
 *     elif opcode == 0xF3:
 */
    break;
    case 0xF3:

    /* "pyboy/core/opcodes.py":5735
 *         return LD_F2(cpu)
 *     elif opcode == 0xF3:
 *         return DI_F3(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF4:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_DI_F3(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5734
 *     elif opcode == 0xF2:
 *         return LD_F2(cpu)
 *     elif opcode == 0xF3:             # <<<<<<<<<<<<<<
 *         return DI_F3(cpu)
 *     elif opcode == 0xF4:
 */
    break;
    case 0xF4:

    /* "pyboy/core/opcodes.py":5737
 *         return DI_F3(cpu)
 *     elif opcode == 0xF4:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF5:
 *         return PUSH_F5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5736
 *     elif opcode == 0xF3:
 *         return DI_F3(cpu)
 *     elif opcode == 0xF4:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xF5:
 */
    break;
    case 0xF5:

    /* "pyboy/core/opcodes.py":5739
 *         return no_opcode(cpu)
 *     elif opcode == 0xF5:
 *         return PUSH_F5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF6:
 *         return OR_F6(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_PUSH_F5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5738
 *     elif opcode == 0xF4:
 *         return no_opcode(cpu)
 *     elif opcode == 0xF5:             # <<<<<<<<<<<<<<
 *         return PUSH_F5(cpu)
 *     elif opcode == 0xF6:
 */
    break;
    case 0xF6:

    /* "pyboy/core/opcodes.py":5741
 *         return PUSH_F5(cpu)
 *     elif opcode == 0xF6:
 *         return OR_F6(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF7:
 *         return RST_F7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_OR_F6(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5740
 *     elif opcode == 0xF5:
 *         return PUSH_F5(cpu)
 *     elif opcode == 0xF6:             # <<<<<<<<<<<<<<
 *         return OR_F6(cpu, v)
 *     elif opcode == 0xF7:
 */
    break;
    case 0xF7:

    /* "pyboy/core/opcodes.py":5743
 *         return OR_F6(cpu, v)
 *     elif opcode == 0xF7:
 *         return RST_F7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF8:
 *         return LD_F8(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RST_F7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5742
 *     elif opcode == 0xF6:
 *         return OR_F6(cpu, v)
 *     elif opcode == 0xF7:             # <<<<<<<<<<<<<<
 *         return RST_F7(cpu)
 *     elif opcode == 0xF8:
 */
    break;
    case 0xF8:

    /* "pyboy/core/opcodes.py":5745
 *         return RST_F7(cpu)
 *     elif opcode == 0xF8:
 *         return LD_F8(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xF9:
 *         return LD_F9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_F8(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5744
 *     elif opcode == 0xF7:
 *         return RST_F7(cpu)
 *     elif opcode == 0xF8:             # <<<<<<<<<<<<<<
 *         return LD_F8(cpu, v)
 *     elif opcode == 0xF9:
 */
    break;
    case 0xF9:

    /* "pyboy/core/opcodes.py":5747
 *         return LD_F8(cpu, v)
 *     elif opcode == 0xF9:
 *         return LD_F9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xFA:
 *         return LD_FA(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_F9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5746
 *     elif opcode == 0xF8:
 *         return LD_F8(cpu, v)
 *     elif opcode == 0xF9:             # <<<<<<<<<<<<<<
 *         return LD_F9(cpu)
 *     elif opcode == 0xFA:
 */
    break;
    case 0xFA:

    /* "pyboy/core/opcodes.py":5749
 *         return LD_F9(cpu)
 *     elif opcode == 0xFA:
 *         return LD_FA(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xFB:
 *         return EI_FB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_LD_FA(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5748
 *     elif opcode == 0xF9:
 *         return LD_F9(cpu)
 *     elif opcode == 0xFA:             # <<<<<<<<<<<<<<
 *         return LD_FA(cpu, v)
 *     elif opcode == 0xFB:
 */
    break;
    case 0xFB:

    /* "pyboy/core/opcodes.py":5751
 *         return LD_FA(cpu, v)
 *     elif opcode == 0xFB:
 *         return EI_FB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xFC:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_EI_FB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5750
 *     elif opcode == 0xFA:
 *         return LD_FA(cpu, v)
 *     elif opcode == 0xFB:             # <<<<<<<<<<<<<<
 *         return EI_FB(cpu)
 *     elif opcode == 0xFC:
 */
    break;
    case 0xFC:

    /* "pyboy/core/opcodes.py":5753
 *         return EI_FB(cpu)
 *     elif opcode == 0xFC:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xFD:
 *         return no_opcode(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5752
 *     elif opcode == 0xFB:
 *         return EI_FB(cpu)
 *     elif opcode == 0xFC:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xFD:
 */
    break;
    case 0xFD:

    /* "pyboy/core/opcodes.py":5755
 *         return no_opcode(cpu)
 *     elif opcode == 0xFD:
 *         return no_opcode(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xFE:
 *         return CP_FE(cpu, v)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_no_opcode(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5754
 *     elif opcode == 0xFC:
 *         return no_opcode(cpu)
 *     elif opcode == 0xFD:             # <<<<<<<<<<<<<<
 *         return no_opcode(cpu)
 *     elif opcode == 0xFE:
 */
    break;
    case 0xFE:

    /* "pyboy/core/opcodes.py":5757
 *         return no_opcode(cpu)
 *     elif opcode == 0xFE:
 *         return CP_FE(cpu, v)             # <<<<<<<<<<<<<<
 *     elif opcode == 0xFF:
 *         return RST_FF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_CP_FE(__pyx_v_cpu, __pyx_v_v);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5756
 *     elif opcode == 0xFD:
 *         return no_opcode(cpu)
 *     elif opcode == 0xFE:             # <<<<<<<<<<<<<<
 *         return CP_FE(cpu, v)
 *     elif opcode == 0xFF:
 */
    break;
    case 0xFF:

    /* "pyboy/core/opcodes.py":5759
 *         return CP_FE(cpu, v)
 *     elif opcode == 0xFF:
 *         return RST_FF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x100:
 *         return RLC_100(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RST_FF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5758
 *     elif opcode == 0xFE:
 *         return CP_FE(cpu, v)
 *     elif opcode == 0xFF:             # <<<<<<<<<<<<<<
 *         return RST_FF(cpu)
 *     elif opcode == 0x100:
 */
    break;
    case 0x100:

    /* "pyboy/core/opcodes.py":5761
 *         return RST_FF(cpu)
 *     elif opcode == 0x100:
 *         return RLC_100(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x101:
 *         return RLC_101(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLC_100(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5760
 *     elif opcode == 0xFF:
 *         return RST_FF(cpu)
 *     elif opcode == 0x100:             # <<<<<<<<<<<<<<
 *         return RLC_100(cpu)
 *     elif opcode == 0x101:
 */
    break;
    case 0x101:

    /* "pyboy/core/opcodes.py":5763
 *         return RLC_100(cpu)
 *     elif opcode == 0x101:
 *         return RLC_101(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x102:
 *         return RLC_102(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLC_101(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5762
 *     elif opcode == 0x100:
 *         return RLC_100(cpu)
 *     elif opcode == 0x101:             # <<<<<<<<<<<<<<
 *         return RLC_101(cpu)
 *     elif opcode == 0x102:
 */
    break;
    case 0x102:

    /* "pyboy/core/opcodes.py":5765
 *         return RLC_101(cpu)
 *     elif opcode == 0x102:
 *         return RLC_102(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x103:
 *         return RLC_103(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLC_102(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5764
 *     elif opcode == 0x101:
 *         return RLC_101(cpu)
 *     elif opcode == 0x102:             # <<<<<<<<<<<<<<
 *         return RLC_102(cpu)
 *     elif opcode == 0x103:
 */
    break;
    case 0x103:

    /* "pyboy/core/opcodes.py":5767
 *         return RLC_102(cpu)
 *     elif opcode == 0x103:
 *         return RLC_103(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x104:
 *         return RLC_104(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLC_103(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5766
 *     elif opcode == 0x102:
 *         return RLC_102(cpu)
 *     elif opcode == 0x103:             # <<<<<<<<<<<<<<
 *         return RLC_103(cpu)
 *     elif opcode == 0x104:
 */
    break;
    case 0x104:

    /* "pyboy/core/opcodes.py":5769
 *         return RLC_103(cpu)
 *     elif opcode == 0x104:
 *         return RLC_104(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x105:
 *         return RLC_105(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLC_104(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5768
 *     elif opcode == 0x103:
 *         return RLC_103(cpu)
 *     elif opcode == 0x104:             # <<<<<<<<<<<<<<
 *         return RLC_104(cpu)
 *     elif opcode == 0x105:
 */
    break;
    case 0x105:

    /* "pyboy/core/opcodes.py":5771
 *         return RLC_104(cpu)
 *     elif opcode == 0x105:
 *         return RLC_105(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x106:
 *         return RLC_106(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLC_105(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5770
 *     elif opcode == 0x104:
 *         return RLC_104(cpu)
 *     elif opcode == 0x105:             # <<<<<<<<<<<<<<
 *         return RLC_105(cpu)
 *     elif opcode == 0x106:
 */
    break;
    case 0x106:

    /* "pyboy/core/opcodes.py":5773
 *         return RLC_105(cpu)
 *     elif opcode == 0x106:
 *         return RLC_106(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x107:
 *         return RLC_107(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLC_106(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5772
 *     elif opcode == 0x105:
 *         return RLC_105(cpu)
 *     elif opcode == 0x106:             # <<<<<<<<<<<<<<
 *         return RLC_106(cpu)
 *     elif opcode == 0x107:
 */
    break;
    case 0x107:

    /* "pyboy/core/opcodes.py":5775
 *         return RLC_106(cpu)
 *     elif opcode == 0x107:
 *         return RLC_107(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x108:
 *         return RRC_108(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RLC_107(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5774
 *     elif opcode == 0x106:
 *         return RLC_106(cpu)
 *     elif opcode == 0x107:             # <<<<<<<<<<<<<<
 *         return RLC_107(cpu)
 *     elif opcode == 0x108:
 */
    break;
    case 0x108:

    /* "pyboy/core/opcodes.py":5777
 *         return RLC_107(cpu)
 *     elif opcode == 0x108:
 *         return RRC_108(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x109:
 *         return RRC_109(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRC_108(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5776
 *     elif opcode == 0x107:
 *         return RLC_107(cpu)
 *     elif opcode == 0x108:             # <<<<<<<<<<<<<<
 *         return RRC_108(cpu)
 *     elif opcode == 0x109:
 */
    break;
    case 0x109:

    /* "pyboy/core/opcodes.py":5779
 *         return RRC_108(cpu)
 *     elif opcode == 0x109:
 *         return RRC_109(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x10A:
 *         return RRC_10A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRC_109(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5778
 *     elif opcode == 0x108:
 *         return RRC_108(cpu)
 *     elif opcode == 0x109:             # <<<<<<<<<<<<<<
 *         return RRC_109(cpu)
 *     elif opcode == 0x10A:
 */
    break;
    case 0x10A:

    /* "pyboy/core/opcodes.py":5781
 *         return RRC_109(cpu)
 *     elif opcode == 0x10A:
 *         return RRC_10A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x10B:
 *         return RRC_10B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRC_10A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5780
 *     elif opcode == 0x109:
 *         return RRC_109(cpu)
 *     elif opcode == 0x10A:             # <<<<<<<<<<<<<<
 *         return RRC_10A(cpu)
 *     elif opcode == 0x10B:
 */
    break;
    case 0x10B:

    /* "pyboy/core/opcodes.py":5783
 *         return RRC_10A(cpu)
 *     elif opcode == 0x10B:
 *         return RRC_10B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x10C:
 *         return RRC_10C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRC_10B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5782
 *     elif opcode == 0x10A:
 *         return RRC_10A(cpu)
 *     elif opcode == 0x10B:             # <<<<<<<<<<<<<<
 *         return RRC_10B(cpu)
 *     elif opcode == 0x10C:
 */
    break;
    case 0x10C:

    /* "pyboy/core/opcodes.py":5785
 *         return RRC_10B(cpu)
 *     elif opcode == 0x10C:
 *         return RRC_10C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x10D:
 *         return RRC_10D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRC_10C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5784
 *     elif opcode == 0x10B:
 *         return RRC_10B(cpu)
 *     elif opcode == 0x10C:             # <<<<<<<<<<<<<<
 *         return RRC_10C(cpu)
 *     elif opcode == 0x10D:
 */
    break;
    case 0x10D:

    /* "pyboy/core/opcodes.py":5787
 *         return RRC_10C(cpu)
 *     elif opcode == 0x10D:
 *         return RRC_10D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x10E:
 *         return RRC_10E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRC_10D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5786
 *     elif opcode == 0x10C:
 *         return RRC_10C(cpu)
 *     elif opcode == 0x10D:             # <<<<<<<<<<<<<<
 *         return RRC_10D(cpu)
 *     elif opcode == 0x10E:
 */
    break;
    case 0x10E:

    /* "pyboy/core/opcodes.py":5789
 *         return RRC_10D(cpu)
 *     elif opcode == 0x10E:
 *         return RRC_10E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x10F:
 *         return RRC_10F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRC_10E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5788
 *     elif opcode == 0x10D:
 *         return RRC_10D(cpu)
 *     elif opcode == 0x10E:             # <<<<<<<<<<<<<<
 *         return RRC_10E(cpu)
 *     elif opcode == 0x10F:
 */
    break;
    case 0x10F:

    /* "pyboy/core/opcodes.py":5791
 *         return RRC_10E(cpu)
 *     elif opcode == 0x10F:
 *         return RRC_10F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x110:
 *         return RL_110(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RRC_10F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5790
 *     elif opcode == 0x10E:
 *         return RRC_10E(cpu)
 *     elif opcode == 0x10F:             # <<<<<<<<<<<<<<
 *         return RRC_10F(cpu)
 *     elif opcode == 0x110:
 */
    break;
    case 0x110:

    /* "pyboy/core/opcodes.py":5793
 *         return RRC_10F(cpu)
 *     elif opcode == 0x110:
 *         return RL_110(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x111:
 *         return RL_111(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RL_110(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5792
 *     elif opcode == 0x10F:
 *         return RRC_10F(cpu)
 *     elif opcode == 0x110:             # <<<<<<<<<<<<<<
 *         return RL_110(cpu)
 *     elif opcode == 0x111:
 */
    break;
    case 0x111:

    /* "pyboy/core/opcodes.py":5795
 *         return RL_110(cpu)
 *     elif opcode == 0x111:
 *         return RL_111(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x112:
 *         return RL_112(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RL_111(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5794
 *     elif opcode == 0x110:
 *         return RL_110(cpu)
 *     elif opcode == 0x111:             # <<<<<<<<<<<<<<
 *         return RL_111(cpu)
 *     elif opcode == 0x112:
 */
    break;
    case 0x112:

    /* "pyboy/core/opcodes.py":5797
 *         return RL_111(cpu)
 *     elif opcode == 0x112:
 *         return RL_112(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x113:
 *         return RL_113(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RL_112(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5796
 *     elif opcode == 0x111:
 *         return RL_111(cpu)
 *     elif opcode == 0x112:             # <<<<<<<<<<<<<<
 *         return RL_112(cpu)
 *     elif opcode == 0x113:
 */
    break;
    case 0x113:

    /* "pyboy/core/opcodes.py":5799
 *         return RL_112(cpu)
 *     elif opcode == 0x113:
 *         return RL_113(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x114:
 *         return RL_114(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RL_113(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5798
 *     elif opcode == 0x112:
 *         return RL_112(cpu)
 *     elif opcode == 0x113:             # <<<<<<<<<<<<<<
 *         return RL_113(cpu)
 *     elif opcode == 0x114:
 */
    break;
    case 0x114:

    /* "pyboy/core/opcodes.py":5801
 *         return RL_113(cpu)
 *     elif opcode == 0x114:
 *         return RL_114(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x115:
 *         return RL_115(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RL_114(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5800
 *     elif opcode == 0x113:
 *         return RL_113(cpu)
 *     elif opcode == 0x114:             # <<<<<<<<<<<<<<
 *         return RL_114(cpu)
 *     elif opcode == 0x115:
 */
    break;
    case 0x115:

    /* "pyboy/core/opcodes.py":5803
 *         return RL_114(cpu)
 *     elif opcode == 0x115:
 *         return RL_115(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x116:
 *         return RL_116(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RL_115(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5802
 *     elif opcode == 0x114:
 *         return RL_114(cpu)
 *     elif opcode == 0x115:             # <<<<<<<<<<<<<<
 *         return RL_115(cpu)
 *     elif opcode == 0x116:
 */
    break;
    case 0x116:

    /* "pyboy/core/opcodes.py":5805
 *         return RL_115(cpu)
 *     elif opcode == 0x116:
 *         return RL_116(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x117:
 *         return RL_117(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RL_116(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5804
 *     elif opcode == 0x115:
 *         return RL_115(cpu)
 *     elif opcode == 0x116:             # <<<<<<<<<<<<<<
 *         return RL_116(cpu)
 *     elif opcode == 0x117:
 */
    break;
    case 0x117:

    /* "pyboy/core/opcodes.py":5807
 *         return RL_116(cpu)
 *     elif opcode == 0x117:
 *         return RL_117(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x118:
 *         return RR_118(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RL_117(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5806
 *     elif opcode == 0x116:
 *         return RL_116(cpu)
 *     elif opcode == 0x117:             # <<<<<<<<<<<<<<
 *         return RL_117(cpu)
 *     elif opcode == 0x118:
 */
    break;
    case 0x118:

    /* "pyboy/core/opcodes.py":5809
 *         return RL_117(cpu)
 *     elif opcode == 0x118:
 *         return RR_118(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x119:
 *         return RR_119(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RR_118(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5808
 *     elif opcode == 0x117:
 *         return RL_117(cpu)
 *     elif opcode == 0x118:             # <<<<<<<<<<<<<<
 *         return RR_118(cpu)
 *     elif opcode == 0x119:
 */
    break;
    case 0x119:

    /* "pyboy/core/opcodes.py":5811
 *         return RR_118(cpu)
 *     elif opcode == 0x119:
 *         return RR_119(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x11A:
 *         return RR_11A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RR_119(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5810
 *     elif opcode == 0x118:
 *         return RR_118(cpu)
 *     elif opcode == 0x119:             # <<<<<<<<<<<<<<
 *         return RR_119(cpu)
 *     elif opcode == 0x11A:
 */
    break;
    case 0x11A:

    /* "pyboy/core/opcodes.py":5813
 *         return RR_119(cpu)
 *     elif opcode == 0x11A:
 *         return RR_11A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x11B:
 *         return RR_11B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RR_11A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5812
 *     elif opcode == 0x119:
 *         return RR_119(cpu)
 *     elif opcode == 0x11A:             # <<<<<<<<<<<<<<
 *         return RR_11A(cpu)
 *     elif opcode == 0x11B:
 */
    break;
    case 0x11B:

    /* "pyboy/core/opcodes.py":5815
 *         return RR_11A(cpu)
 *     elif opcode == 0x11B:
 *         return RR_11B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x11C:
 *         return RR_11C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RR_11B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5814
 *     elif opcode == 0x11A:
 *         return RR_11A(cpu)
 *     elif opcode == 0x11B:             # <<<<<<<<<<<<<<
 *         return RR_11B(cpu)
 *     elif opcode == 0x11C:
 */
    break;
    case 0x11C:

    /* "pyboy/core/opcodes.py":5817
 *         return RR_11B(cpu)
 *     elif opcode == 0x11C:
 *         return RR_11C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x11D:
 *         return RR_11D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RR_11C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5816
 *     elif opcode == 0x11B:
 *         return RR_11B(cpu)
 *     elif opcode == 0x11C:             # <<<<<<<<<<<<<<
 *         return RR_11C(cpu)
 *     elif opcode == 0x11D:
 */
    break;
    case 0x11D:

    /* "pyboy/core/opcodes.py":5819
 *         return RR_11C(cpu)
 *     elif opcode == 0x11D:
 *         return RR_11D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x11E:
 *         return RR_11E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RR_11D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5818
 *     elif opcode == 0x11C:
 *         return RR_11C(cpu)
 *     elif opcode == 0x11D:             # <<<<<<<<<<<<<<
 *         return RR_11D(cpu)
 *     elif opcode == 0x11E:
 */
    break;
    case 0x11E:

    /* "pyboy/core/opcodes.py":5821
 *         return RR_11D(cpu)
 *     elif opcode == 0x11E:
 *         return RR_11E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x11F:
 *         return RR_11F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RR_11E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5820
 *     elif opcode == 0x11D:
 *         return RR_11D(cpu)
 *     elif opcode == 0x11E:             # <<<<<<<<<<<<<<
 *         return RR_11E(cpu)
 *     elif opcode == 0x11F:
 */
    break;
    case 0x11F:

    /* "pyboy/core/opcodes.py":5823
 *         return RR_11E(cpu)
 *     elif opcode == 0x11F:
 *         return RR_11F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x120:
 *         return SLA_120(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RR_11F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5822
 *     elif opcode == 0x11E:
 *         return RR_11E(cpu)
 *     elif opcode == 0x11F:             # <<<<<<<<<<<<<<
 *         return RR_11F(cpu)
 *     elif opcode == 0x120:
 */
    break;
    case 0x120:

    /* "pyboy/core/opcodes.py":5825
 *         return RR_11F(cpu)
 *     elif opcode == 0x120:
 *         return SLA_120(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x121:
 *         return SLA_121(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SLA_120(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5824
 *     elif opcode == 0x11F:
 *         return RR_11F(cpu)
 *     elif opcode == 0x120:             # <<<<<<<<<<<<<<
 *         return SLA_120(cpu)
 *     elif opcode == 0x121:
 */
    break;
    case 0x121:

    /* "pyboy/core/opcodes.py":5827
 *         return SLA_120(cpu)
 *     elif opcode == 0x121:
 *         return SLA_121(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x122:
 *         return SLA_122(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SLA_121(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5826
 *     elif opcode == 0x120:
 *         return SLA_120(cpu)
 *     elif opcode == 0x121:             # <<<<<<<<<<<<<<
 *         return SLA_121(cpu)
 *     elif opcode == 0x122:
 */
    break;
    case 0x122:

    /* "pyboy/core/opcodes.py":5829
 *         return SLA_121(cpu)
 *     elif opcode == 0x122:
 *         return SLA_122(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x123:
 *         return SLA_123(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SLA_122(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5828
 *     elif opcode == 0x121:
 *         return SLA_121(cpu)
 *     elif opcode == 0x122:             # <<<<<<<<<<<<<<
 *         return SLA_122(cpu)
 *     elif opcode == 0x123:
 */
    break;
    case 0x123:

    /* "pyboy/core/opcodes.py":5831
 *         return SLA_122(cpu)
 *     elif opcode == 0x123:
 *         return SLA_123(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x124:
 *         return SLA_124(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SLA_123(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5830
 *     elif opcode == 0x122:
 *         return SLA_122(cpu)
 *     elif opcode == 0x123:             # <<<<<<<<<<<<<<
 *         return SLA_123(cpu)
 *     elif opcode == 0x124:
 */
    break;
    case 0x124:

    /* "pyboy/core/opcodes.py":5833
 *         return SLA_123(cpu)
 *     elif opcode == 0x124:
 *         return SLA_124(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x125:
 *         return SLA_125(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SLA_124(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5832
 *     elif opcode == 0x123:
 *         return SLA_123(cpu)
 *     elif opcode == 0x124:             # <<<<<<<<<<<<<<
 *         return SLA_124(cpu)
 *     elif opcode == 0x125:
 */
    break;
    case 0x125:

    /* "pyboy/core/opcodes.py":5835
 *         return SLA_124(cpu)
 *     elif opcode == 0x125:
 *         return SLA_125(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x126:
 *         return SLA_126(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SLA_125(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5834
 *     elif opcode == 0x124:
 *         return SLA_124(cpu)
 *     elif opcode == 0x125:             # <<<<<<<<<<<<<<
 *         return SLA_125(cpu)
 *     elif opcode == 0x126:
 */
    break;
    case 0x126:

    /* "pyboy/core/opcodes.py":5837
 *         return SLA_125(cpu)
 *     elif opcode == 0x126:
 *         return SLA_126(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x127:
 *         return SLA_127(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SLA_126(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5836
 *     elif opcode == 0x125:
 *         return SLA_125(cpu)
 *     elif opcode == 0x126:             # <<<<<<<<<<<<<<
 *         return SLA_126(cpu)
 *     elif opcode == 0x127:
 */
    break;
    case 0x127:

    /* "pyboy/core/opcodes.py":5839
 *         return SLA_126(cpu)
 *     elif opcode == 0x127:
 *         return SLA_127(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x128:
 *         return SRA_128(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SLA_127(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5838
 *     elif opcode == 0x126:
 *         return SLA_126(cpu)
 *     elif opcode == 0x127:             # <<<<<<<<<<<<<<
 *         return SLA_127(cpu)
 *     elif opcode == 0x128:
 */
    break;
    case 0x128:

    /* "pyboy/core/opcodes.py":5841
 *         return SLA_127(cpu)
 *     elif opcode == 0x128:
 *         return SRA_128(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x129:
 *         return SRA_129(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRA_128(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5840
 *     elif opcode == 0x127:
 *         return SLA_127(cpu)
 *     elif opcode == 0x128:             # <<<<<<<<<<<<<<
 *         return SRA_128(cpu)
 *     elif opcode == 0x129:
 */
    break;
    case 0x129:

    /* "pyboy/core/opcodes.py":5843
 *         return SRA_128(cpu)
 *     elif opcode == 0x129:
 *         return SRA_129(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x12A:
 *         return SRA_12A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRA_129(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5842
 *     elif opcode == 0x128:
 *         return SRA_128(cpu)
 *     elif opcode == 0x129:             # <<<<<<<<<<<<<<
 *         return SRA_129(cpu)
 *     elif opcode == 0x12A:
 */
    break;
    case 0x12A:

    /* "pyboy/core/opcodes.py":5845
 *         return SRA_129(cpu)
 *     elif opcode == 0x12A:
 *         return SRA_12A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x12B:
 *         return SRA_12B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRA_12A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5844
 *     elif opcode == 0x129:
 *         return SRA_129(cpu)
 *     elif opcode == 0x12A:             # <<<<<<<<<<<<<<
 *         return SRA_12A(cpu)
 *     elif opcode == 0x12B:
 */
    break;
    case 0x12B:

    /* "pyboy/core/opcodes.py":5847
 *         return SRA_12A(cpu)
 *     elif opcode == 0x12B:
 *         return SRA_12B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x12C:
 *         return SRA_12C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRA_12B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5846
 *     elif opcode == 0x12A:
 *         return SRA_12A(cpu)
 *     elif opcode == 0x12B:             # <<<<<<<<<<<<<<
 *         return SRA_12B(cpu)
 *     elif opcode == 0x12C:
 */
    break;
    case 0x12C:

    /* "pyboy/core/opcodes.py":5849
 *         return SRA_12B(cpu)
 *     elif opcode == 0x12C:
 *         return SRA_12C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x12D:
 *         return SRA_12D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRA_12C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5848
 *     elif opcode == 0x12B:
 *         return SRA_12B(cpu)
 *     elif opcode == 0x12C:             # <<<<<<<<<<<<<<
 *         return SRA_12C(cpu)
 *     elif opcode == 0x12D:
 */
    break;
    case 0x12D:

    /* "pyboy/core/opcodes.py":5851
 *         return SRA_12C(cpu)
 *     elif opcode == 0x12D:
 *         return SRA_12D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x12E:
 *         return SRA_12E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRA_12D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5850
 *     elif opcode == 0x12C:
 *         return SRA_12C(cpu)
 *     elif opcode == 0x12D:             # <<<<<<<<<<<<<<
 *         return SRA_12D(cpu)
 *     elif opcode == 0x12E:
 */
    break;
    case 0x12E:

    /* "pyboy/core/opcodes.py":5853
 *         return SRA_12D(cpu)
 *     elif opcode == 0x12E:
 *         return SRA_12E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x12F:
 *         return SRA_12F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRA_12E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5852
 *     elif opcode == 0x12D:
 *         return SRA_12D(cpu)
 *     elif opcode == 0x12E:             # <<<<<<<<<<<<<<
 *         return SRA_12E(cpu)
 *     elif opcode == 0x12F:
 */
    break;
    case 0x12F:

    /* "pyboy/core/opcodes.py":5855
 *         return SRA_12E(cpu)
 *     elif opcode == 0x12F:
 *         return SRA_12F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x130:
 *         return SWAP_130(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRA_12F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5854
 *     elif opcode == 0x12E:
 *         return SRA_12E(cpu)
 *     elif opcode == 0x12F:             # <<<<<<<<<<<<<<
 *         return SRA_12F(cpu)
 *     elif opcode == 0x130:
 */
    break;
    case 0x130:

    /* "pyboy/core/opcodes.py":5857
 *         return SRA_12F(cpu)
 *     elif opcode == 0x130:
 *         return SWAP_130(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x131:
 *         return SWAP_131(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SWAP_130(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5856
 *     elif opcode == 0x12F:
 *         return SRA_12F(cpu)
 *     elif opcode == 0x130:             # <<<<<<<<<<<<<<
 *         return SWAP_130(cpu)
 *     elif opcode == 0x131:
 */
    break;
    case 0x131:

    /* "pyboy/core/opcodes.py":5859
 *         return SWAP_130(cpu)
 *     elif opcode == 0x131:
 *         return SWAP_131(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x132:
 *         return SWAP_132(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SWAP_131(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5858
 *     elif opcode == 0x130:
 *         return SWAP_130(cpu)
 *     elif opcode == 0x131:             # <<<<<<<<<<<<<<
 *         return SWAP_131(cpu)
 *     elif opcode == 0x132:
 */
    break;
    case 0x132:

    /* "pyboy/core/opcodes.py":5861
 *         return SWAP_131(cpu)
 *     elif opcode == 0x132:
 *         return SWAP_132(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x133:
 *         return SWAP_133(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SWAP_132(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5860
 *     elif opcode == 0x131:
 *         return SWAP_131(cpu)
 *     elif opcode == 0x132:             # <<<<<<<<<<<<<<
 *         return SWAP_132(cpu)
 *     elif opcode == 0x133:
 */
    break;
    case 0x133:

    /* "pyboy/core/opcodes.py":5863
 *         return SWAP_132(cpu)
 *     elif opcode == 0x133:
 *         return SWAP_133(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x134:
 *         return SWAP_134(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SWAP_133(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5862
 *     elif opcode == 0x132:
 *         return SWAP_132(cpu)
 *     elif opcode == 0x133:             # <<<<<<<<<<<<<<
 *         return SWAP_133(cpu)
 *     elif opcode == 0x134:
 */
    break;
    case 0x134:

    /* "pyboy/core/opcodes.py":5865
 *         return SWAP_133(cpu)
 *     elif opcode == 0x134:
 *         return SWAP_134(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x135:
 *         return SWAP_135(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SWAP_134(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5864
 *     elif opcode == 0x133:
 *         return SWAP_133(cpu)
 *     elif opcode == 0x134:             # <<<<<<<<<<<<<<
 *         return SWAP_134(cpu)
 *     elif opcode == 0x135:
 */
    break;
    case 0x135:

    /* "pyboy/core/opcodes.py":5867
 *         return SWAP_134(cpu)
 *     elif opcode == 0x135:
 *         return SWAP_135(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x136:
 *         return SWAP_136(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SWAP_135(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5866
 *     elif opcode == 0x134:
 *         return SWAP_134(cpu)
 *     elif opcode == 0x135:             # <<<<<<<<<<<<<<
 *         return SWAP_135(cpu)
 *     elif opcode == 0x136:
 */
    break;
    case 0x136:

    /* "pyboy/core/opcodes.py":5869
 *         return SWAP_135(cpu)
 *     elif opcode == 0x136:
 *         return SWAP_136(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x137:
 *         return SWAP_137(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SWAP_136(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5868
 *     elif opcode == 0x135:
 *         return SWAP_135(cpu)
 *     elif opcode == 0x136:             # <<<<<<<<<<<<<<
 *         return SWAP_136(cpu)
 *     elif opcode == 0x137:
 */
    break;
    case 0x137:

    /* "pyboy/core/opcodes.py":5871
 *         return SWAP_136(cpu)
 *     elif opcode == 0x137:
 *         return SWAP_137(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x138:
 *         return SRL_138(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SWAP_137(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5870
 *     elif opcode == 0x136:
 *         return SWAP_136(cpu)
 *     elif opcode == 0x137:             # <<<<<<<<<<<<<<
 *         return SWAP_137(cpu)
 *     elif opcode == 0x138:
 */
    break;
    case 0x138:

    /* "pyboy/core/opcodes.py":5873
 *         return SWAP_137(cpu)
 *     elif opcode == 0x138:
 *         return SRL_138(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x139:
 *         return SRL_139(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRL_138(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5872
 *     elif opcode == 0x137:
 *         return SWAP_137(cpu)
 *     elif opcode == 0x138:             # <<<<<<<<<<<<<<
 *         return SRL_138(cpu)
 *     elif opcode == 0x139:
 */
    break;
    case 0x139:

    /* "pyboy/core/opcodes.py":5875
 *         return SRL_138(cpu)
 *     elif opcode == 0x139:
 *         return SRL_139(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x13A:
 *         return SRL_13A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRL_139(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5874
 *     elif opcode == 0x138:
 *         return SRL_138(cpu)
 *     elif opcode == 0x139:             # <<<<<<<<<<<<<<
 *         return SRL_139(cpu)
 *     elif opcode == 0x13A:
 */
    break;
    case 0x13A:

    /* "pyboy/core/opcodes.py":5877
 *         return SRL_139(cpu)
 *     elif opcode == 0x13A:
 *         return SRL_13A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x13B:
 *         return SRL_13B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRL_13A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5876
 *     elif opcode == 0x139:
 *         return SRL_139(cpu)
 *     elif opcode == 0x13A:             # <<<<<<<<<<<<<<
 *         return SRL_13A(cpu)
 *     elif opcode == 0x13B:
 */
    break;
    case 0x13B:

    /* "pyboy/core/opcodes.py":5879
 *         return SRL_13A(cpu)
 *     elif opcode == 0x13B:
 *         return SRL_13B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x13C:
 *         return SRL_13C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRL_13B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5878
 *     elif opcode == 0x13A:
 *         return SRL_13A(cpu)
 *     elif opcode == 0x13B:             # <<<<<<<<<<<<<<
 *         return SRL_13B(cpu)
 *     elif opcode == 0x13C:
 */
    break;
    case 0x13C:

    /* "pyboy/core/opcodes.py":5881
 *         return SRL_13B(cpu)
 *     elif opcode == 0x13C:
 *         return SRL_13C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x13D:
 *         return SRL_13D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRL_13C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5880
 *     elif opcode == 0x13B:
 *         return SRL_13B(cpu)
 *     elif opcode == 0x13C:             # <<<<<<<<<<<<<<
 *         return SRL_13C(cpu)
 *     elif opcode == 0x13D:
 */
    break;
    case 0x13D:

    /* "pyboy/core/opcodes.py":5883
 *         return SRL_13C(cpu)
 *     elif opcode == 0x13D:
 *         return SRL_13D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x13E:
 *         return SRL_13E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRL_13D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5882
 *     elif opcode == 0x13C:
 *         return SRL_13C(cpu)
 *     elif opcode == 0x13D:             # <<<<<<<<<<<<<<
 *         return SRL_13D(cpu)
 *     elif opcode == 0x13E:
 */
    break;
    case 0x13E:

    /* "pyboy/core/opcodes.py":5885
 *         return SRL_13D(cpu)
 *     elif opcode == 0x13E:
 *         return SRL_13E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x13F:
 *         return SRL_13F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRL_13E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5884
 *     elif opcode == 0x13D:
 *         return SRL_13D(cpu)
 *     elif opcode == 0x13E:             # <<<<<<<<<<<<<<
 *         return SRL_13E(cpu)
 *     elif opcode == 0x13F:
 */
    break;
    case 0x13F:

    /* "pyboy/core/opcodes.py":5887
 *         return SRL_13E(cpu)
 *     elif opcode == 0x13F:
 *         return SRL_13F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x140:
 *         return BIT_140(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SRL_13F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5886
 *     elif opcode == 0x13E:
 *         return SRL_13E(cpu)
 *     elif opcode == 0x13F:             # <<<<<<<<<<<<<<
 *         return SRL_13F(cpu)
 *     elif opcode == 0x140:
 */
    break;
    case 0x140:

    /* "pyboy/core/opcodes.py":5889
 *         return SRL_13F(cpu)
 *     elif opcode == 0x140:
 *         return BIT_140(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x141:
 *         return BIT_141(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_140(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5888
 *     elif opcode == 0x13F:
 *         return SRL_13F(cpu)
 *     elif opcode == 0x140:             # <<<<<<<<<<<<<<
 *         return BIT_140(cpu)
 *     elif opcode == 0x141:
 */
    break;
    case 0x141:

    /* "pyboy/core/opcodes.py":5891
 *         return BIT_140(cpu)
 *     elif opcode == 0x141:
 *         return BIT_141(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x142:
 *         return BIT_142(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_141(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5890
 *     elif opcode == 0x140:
 *         return BIT_140(cpu)
 *     elif opcode == 0x141:             # <<<<<<<<<<<<<<
 *         return BIT_141(cpu)
 *     elif opcode == 0x142:
 */
    break;
    case 0x142:

    /* "pyboy/core/opcodes.py":5893
 *         return BIT_141(cpu)
 *     elif opcode == 0x142:
 *         return BIT_142(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x143:
 *         return BIT_143(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_142(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5892
 *     elif opcode == 0x141:
 *         return BIT_141(cpu)
 *     elif opcode == 0x142:             # <<<<<<<<<<<<<<
 *         return BIT_142(cpu)
 *     elif opcode == 0x143:
 */
    break;
    case 0x143:

    /* "pyboy/core/opcodes.py":5895
 *         return BIT_142(cpu)
 *     elif opcode == 0x143:
 *         return BIT_143(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x144:
 *         return BIT_144(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_143(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5894
 *     elif opcode == 0x142:
 *         return BIT_142(cpu)
 *     elif opcode == 0x143:             # <<<<<<<<<<<<<<
 *         return BIT_143(cpu)
 *     elif opcode == 0x144:
 */
    break;
    case 0x144:

    /* "pyboy/core/opcodes.py":5897
 *         return BIT_143(cpu)
 *     elif opcode == 0x144:
 *         return BIT_144(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x145:
 *         return BIT_145(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_144(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5896
 *     elif opcode == 0x143:
 *         return BIT_143(cpu)
 *     elif opcode == 0x144:             # <<<<<<<<<<<<<<
 *         return BIT_144(cpu)
 *     elif opcode == 0x145:
 */
    break;
    case 0x145:

    /* "pyboy/core/opcodes.py":5899
 *         return BIT_144(cpu)
 *     elif opcode == 0x145:
 *         return BIT_145(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x146:
 *         return BIT_146(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_145(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5898
 *     elif opcode == 0x144:
 *         return BIT_144(cpu)
 *     elif opcode == 0x145:             # <<<<<<<<<<<<<<
 *         return BIT_145(cpu)
 *     elif opcode == 0x146:
 */
    break;
    case 0x146:

    /* "pyboy/core/opcodes.py":5901
 *         return BIT_145(cpu)
 *     elif opcode == 0x146:
 *         return BIT_146(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x147:
 *         return BIT_147(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_146(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5900
 *     elif opcode == 0x145:
 *         return BIT_145(cpu)
 *     elif opcode == 0x146:             # <<<<<<<<<<<<<<
 *         return BIT_146(cpu)
 *     elif opcode == 0x147:
 */
    break;
    case 0x147:

    /* "pyboy/core/opcodes.py":5903
 *         return BIT_146(cpu)
 *     elif opcode == 0x147:
 *         return BIT_147(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x148:
 *         return BIT_148(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_147(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5902
 *     elif opcode == 0x146:
 *         return BIT_146(cpu)
 *     elif opcode == 0x147:             # <<<<<<<<<<<<<<
 *         return BIT_147(cpu)
 *     elif opcode == 0x148:
 */
    break;
    case 0x148:

    /* "pyboy/core/opcodes.py":5905
 *         return BIT_147(cpu)
 *     elif opcode == 0x148:
 *         return BIT_148(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x149:
 *         return BIT_149(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_148(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5904
 *     elif opcode == 0x147:
 *         return BIT_147(cpu)
 *     elif opcode == 0x148:             # <<<<<<<<<<<<<<
 *         return BIT_148(cpu)
 *     elif opcode == 0x149:
 */
    break;
    case 0x149:

    /* "pyboy/core/opcodes.py":5907
 *         return BIT_148(cpu)
 *     elif opcode == 0x149:
 *         return BIT_149(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x14A:
 *         return BIT_14A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_149(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5906
 *     elif opcode == 0x148:
 *         return BIT_148(cpu)
 *     elif opcode == 0x149:             # <<<<<<<<<<<<<<
 *         return BIT_149(cpu)
 *     elif opcode == 0x14A:
 */
    break;
    case 0x14A:

    /* "pyboy/core/opcodes.py":5909
 *         return BIT_149(cpu)
 *     elif opcode == 0x14A:
 *         return BIT_14A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x14B:
 *         return BIT_14B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_14A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5908
 *     elif opcode == 0x149:
 *         return BIT_149(cpu)
 *     elif opcode == 0x14A:             # <<<<<<<<<<<<<<
 *         return BIT_14A(cpu)
 *     elif opcode == 0x14B:
 */
    break;
    case 0x14B:

    /* "pyboy/core/opcodes.py":5911
 *         return BIT_14A(cpu)
 *     elif opcode == 0x14B:
 *         return BIT_14B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x14C:
 *         return BIT_14C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_14B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5910
 *     elif opcode == 0x14A:
 *         return BIT_14A(cpu)
 *     elif opcode == 0x14B:             # <<<<<<<<<<<<<<
 *         return BIT_14B(cpu)
 *     elif opcode == 0x14C:
 */
    break;
    case 0x14C:

    /* "pyboy/core/opcodes.py":5913
 *         return BIT_14B(cpu)
 *     elif opcode == 0x14C:
 *         return BIT_14C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x14D:
 *         return BIT_14D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_14C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5912
 *     elif opcode == 0x14B:
 *         return BIT_14B(cpu)
 *     elif opcode == 0x14C:             # <<<<<<<<<<<<<<
 *         return BIT_14C(cpu)
 *     elif opcode == 0x14D:
 */
    break;
    case 0x14D:

    /* "pyboy/core/opcodes.py":5915
 *         return BIT_14C(cpu)
 *     elif opcode == 0x14D:
 *         return BIT_14D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x14E:
 *         return BIT_14E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_14D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5914
 *     elif opcode == 0x14C:
 *         return BIT_14C(cpu)
 *     elif opcode == 0x14D:             # <<<<<<<<<<<<<<
 *         return BIT_14D(cpu)
 *     elif opcode == 0x14E:
 */
    break;
    case 0x14E:

    /* "pyboy/core/opcodes.py":5917
 *         return BIT_14D(cpu)
 *     elif opcode == 0x14E:
 *         return BIT_14E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x14F:
 *         return BIT_14F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_14E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5916
 *     elif opcode == 0x14D:
 *         return BIT_14D(cpu)
 *     elif opcode == 0x14E:             # <<<<<<<<<<<<<<
 *         return BIT_14E(cpu)
 *     elif opcode == 0x14F:
 */
    break;
    case 0x14F:

    /* "pyboy/core/opcodes.py":5919
 *         return BIT_14E(cpu)
 *     elif opcode == 0x14F:
 *         return BIT_14F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x150:
 *         return BIT_150(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_14F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5918
 *     elif opcode == 0x14E:
 *         return BIT_14E(cpu)
 *     elif opcode == 0x14F:             # <<<<<<<<<<<<<<
 *         return BIT_14F(cpu)
 *     elif opcode == 0x150:
 */
    break;
    case 0x150:

    /* "pyboy/core/opcodes.py":5921
 *         return BIT_14F(cpu)
 *     elif opcode == 0x150:
 *         return BIT_150(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x151:
 *         return BIT_151(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_150(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5920
 *     elif opcode == 0x14F:
 *         return BIT_14F(cpu)
 *     elif opcode == 0x150:             # <<<<<<<<<<<<<<
 *         return BIT_150(cpu)
 *     elif opcode == 0x151:
 */
    break;
    case 0x151:

    /* "pyboy/core/opcodes.py":5923
 *         return BIT_150(cpu)
 *     elif opcode == 0x151:
 *         return BIT_151(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x152:
 *         return BIT_152(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_151(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5922
 *     elif opcode == 0x150:
 *         return BIT_150(cpu)
 *     elif opcode == 0x151:             # <<<<<<<<<<<<<<
 *         return BIT_151(cpu)
 *     elif opcode == 0x152:
 */
    break;
    case 0x152:

    /* "pyboy/core/opcodes.py":5925
 *         return BIT_151(cpu)
 *     elif opcode == 0x152:
 *         return BIT_152(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x153:
 *         return BIT_153(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_152(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5924
 *     elif opcode == 0x151:
 *         return BIT_151(cpu)
 *     elif opcode == 0x152:             # <<<<<<<<<<<<<<
 *         return BIT_152(cpu)
 *     elif opcode == 0x153:
 */
    break;
    case 0x153:

    /* "pyboy/core/opcodes.py":5927
 *         return BIT_152(cpu)
 *     elif opcode == 0x153:
 *         return BIT_153(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x154:
 *         return BIT_154(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_153(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5926
 *     elif opcode == 0x152:
 *         return BIT_152(cpu)
 *     elif opcode == 0x153:             # <<<<<<<<<<<<<<
 *         return BIT_153(cpu)
 *     elif opcode == 0x154:
 */
    break;
    case 0x154:

    /* "pyboy/core/opcodes.py":5929
 *         return BIT_153(cpu)
 *     elif opcode == 0x154:
 *         return BIT_154(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x155:
 *         return BIT_155(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_154(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5928
 *     elif opcode == 0x153:
 *         return BIT_153(cpu)
 *     elif opcode == 0x154:             # <<<<<<<<<<<<<<
 *         return BIT_154(cpu)
 *     elif opcode == 0x155:
 */
    break;
    case 0x155:

    /* "pyboy/core/opcodes.py":5931
 *         return BIT_154(cpu)
 *     elif opcode == 0x155:
 *         return BIT_155(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x156:
 *         return BIT_156(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_155(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5930
 *     elif opcode == 0x154:
 *         return BIT_154(cpu)
 *     elif opcode == 0x155:             # <<<<<<<<<<<<<<
 *         return BIT_155(cpu)
 *     elif opcode == 0x156:
 */
    break;
    case 0x156:

    /* "pyboy/core/opcodes.py":5933
 *         return BIT_155(cpu)
 *     elif opcode == 0x156:
 *         return BIT_156(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x157:
 *         return BIT_157(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_156(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5932
 *     elif opcode == 0x155:
 *         return BIT_155(cpu)
 *     elif opcode == 0x156:             # <<<<<<<<<<<<<<
 *         return BIT_156(cpu)
 *     elif opcode == 0x157:
 */
    break;
    case 0x157:

    /* "pyboy/core/opcodes.py":5935
 *         return BIT_156(cpu)
 *     elif opcode == 0x157:
 *         return BIT_157(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x158:
 *         return BIT_158(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_157(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5934
 *     elif opcode == 0x156:
 *         return BIT_156(cpu)
 *     elif opcode == 0x157:             # <<<<<<<<<<<<<<
 *         return BIT_157(cpu)
 *     elif opcode == 0x158:
 */
    break;
    case 0x158:

    /* "pyboy/core/opcodes.py":5937
 *         return BIT_157(cpu)
 *     elif opcode == 0x158:
 *         return BIT_158(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x159:
 *         return BIT_159(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_158(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5936
 *     elif opcode == 0x157:
 *         return BIT_157(cpu)
 *     elif opcode == 0x158:             # <<<<<<<<<<<<<<
 *         return BIT_158(cpu)
 *     elif opcode == 0x159:
 */
    break;
    case 0x159:

    /* "pyboy/core/opcodes.py":5939
 *         return BIT_158(cpu)
 *     elif opcode == 0x159:
 *         return BIT_159(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x15A:
 *         return BIT_15A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_159(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5938
 *     elif opcode == 0x158:
 *         return BIT_158(cpu)
 *     elif opcode == 0x159:             # <<<<<<<<<<<<<<
 *         return BIT_159(cpu)
 *     elif opcode == 0x15A:
 */
    break;
    case 0x15A:

    /* "pyboy/core/opcodes.py":5941
 *         return BIT_159(cpu)
 *     elif opcode == 0x15A:
 *         return BIT_15A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x15B:
 *         return BIT_15B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_15A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5940
 *     elif opcode == 0x159:
 *         return BIT_159(cpu)
 *     elif opcode == 0x15A:             # <<<<<<<<<<<<<<
 *         return BIT_15A(cpu)
 *     elif opcode == 0x15B:
 */
    break;
    case 0x15B:

    /* "pyboy/core/opcodes.py":5943
 *         return BIT_15A(cpu)
 *     elif opcode == 0x15B:
 *         return BIT_15B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x15C:
 *         return BIT_15C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_15B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5942
 *     elif opcode == 0x15A:
 *         return BIT_15A(cpu)
 *     elif opcode == 0x15B:             # <<<<<<<<<<<<<<
 *         return BIT_15B(cpu)
 *     elif opcode == 0x15C:
 */
    break;
    case 0x15C:

    /* "pyboy/core/opcodes.py":5945
 *         return BIT_15B(cpu)
 *     elif opcode == 0x15C:
 *         return BIT_15C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x15D:
 *         return BIT_15D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_15C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5944
 *     elif opcode == 0x15B:
 *         return BIT_15B(cpu)
 *     elif opcode == 0x15C:             # <<<<<<<<<<<<<<
 *         return BIT_15C(cpu)
 *     elif opcode == 0x15D:
 */
    break;
    case 0x15D:

    /* "pyboy/core/opcodes.py":5947
 *         return BIT_15C(cpu)
 *     elif opcode == 0x15D:
 *         return BIT_15D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x15E:
 *         return BIT_15E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_15D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5946
 *     elif opcode == 0x15C:
 *         return BIT_15C(cpu)
 *     elif opcode == 0x15D:             # <<<<<<<<<<<<<<
 *         return BIT_15D(cpu)
 *     elif opcode == 0x15E:
 */
    break;
    case 0x15E:

    /* "pyboy/core/opcodes.py":5949
 *         return BIT_15D(cpu)
 *     elif opcode == 0x15E:
 *         return BIT_15E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x15F:
 *         return BIT_15F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_15E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5948
 *     elif opcode == 0x15D:
 *         return BIT_15D(cpu)
 *     elif opcode == 0x15E:             # <<<<<<<<<<<<<<
 *         return BIT_15E(cpu)
 *     elif opcode == 0x15F:
 */
    break;
    case 0x15F:

    /* "pyboy/core/opcodes.py":5951
 *         return BIT_15E(cpu)
 *     elif opcode == 0x15F:
 *         return BIT_15F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x160:
 *         return BIT_160(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_15F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5950
 *     elif opcode == 0x15E:
 *         return BIT_15E(cpu)
 *     elif opcode == 0x15F:             # <<<<<<<<<<<<<<
 *         return BIT_15F(cpu)
 *     elif opcode == 0x160:
 */
    break;
    case 0x160:

    /* "pyboy/core/opcodes.py":5953
 *         return BIT_15F(cpu)
 *     elif opcode == 0x160:
 *         return BIT_160(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x161:
 *         return BIT_161(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_160(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5952
 *     elif opcode == 0x15F:
 *         return BIT_15F(cpu)
 *     elif opcode == 0x160:             # <<<<<<<<<<<<<<
 *         return BIT_160(cpu)
 *     elif opcode == 0x161:
 */
    break;
    case 0x161:

    /* "pyboy/core/opcodes.py":5955
 *         return BIT_160(cpu)
 *     elif opcode == 0x161:
 *         return BIT_161(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x162:
 *         return BIT_162(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_161(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5954
 *     elif opcode == 0x160:
 *         return BIT_160(cpu)
 *     elif opcode == 0x161:             # <<<<<<<<<<<<<<
 *         return BIT_161(cpu)
 *     elif opcode == 0x162:
 */
    break;
    case 0x162:

    /* "pyboy/core/opcodes.py":5957
 *         return BIT_161(cpu)
 *     elif opcode == 0x162:
 *         return BIT_162(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x163:
 *         return BIT_163(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_162(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5956
 *     elif opcode == 0x161:
 *         return BIT_161(cpu)
 *     elif opcode == 0x162:             # <<<<<<<<<<<<<<
 *         return BIT_162(cpu)
 *     elif opcode == 0x163:
 */
    break;
    case 0x163:

    /* "pyboy/core/opcodes.py":5959
 *         return BIT_162(cpu)
 *     elif opcode == 0x163:
 *         return BIT_163(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x164:
 *         return BIT_164(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_163(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5958
 *     elif opcode == 0x162:
 *         return BIT_162(cpu)
 *     elif opcode == 0x163:             # <<<<<<<<<<<<<<
 *         return BIT_163(cpu)
 *     elif opcode == 0x164:
 */
    break;
    case 0x164:

    /* "pyboy/core/opcodes.py":5961
 *         return BIT_163(cpu)
 *     elif opcode == 0x164:
 *         return BIT_164(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x165:
 *         return BIT_165(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_164(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5960
 *     elif opcode == 0x163:
 *         return BIT_163(cpu)
 *     elif opcode == 0x164:             # <<<<<<<<<<<<<<
 *         return BIT_164(cpu)
 *     elif opcode == 0x165:
 */
    break;
    case 0x165:

    /* "pyboy/core/opcodes.py":5963
 *         return BIT_164(cpu)
 *     elif opcode == 0x165:
 *         return BIT_165(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x166:
 *         return BIT_166(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_165(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5962
 *     elif opcode == 0x164:
 *         return BIT_164(cpu)
 *     elif opcode == 0x165:             # <<<<<<<<<<<<<<
 *         return BIT_165(cpu)
 *     elif opcode == 0x166:
 */
    break;
    case 0x166:

    /* "pyboy/core/opcodes.py":5965
 *         return BIT_165(cpu)
 *     elif opcode == 0x166:
 *         return BIT_166(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x167:
 *         return BIT_167(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_166(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5964
 *     elif opcode == 0x165:
 *         return BIT_165(cpu)
 *     elif opcode == 0x166:             # <<<<<<<<<<<<<<
 *         return BIT_166(cpu)
 *     elif opcode == 0x167:
 */
    break;
    case 0x167:

    /* "pyboy/core/opcodes.py":5967
 *         return BIT_166(cpu)
 *     elif opcode == 0x167:
 *         return BIT_167(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x168:
 *         return BIT_168(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_167(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5966
 *     elif opcode == 0x166:
 *         return BIT_166(cpu)
 *     elif opcode == 0x167:             # <<<<<<<<<<<<<<
 *         return BIT_167(cpu)
 *     elif opcode == 0x168:
 */
    break;
    case 0x168:

    /* "pyboy/core/opcodes.py":5969
 *         return BIT_167(cpu)
 *     elif opcode == 0x168:
 *         return BIT_168(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x169:
 *         return BIT_169(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_168(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5968
 *     elif opcode == 0x167:
 *         return BIT_167(cpu)
 *     elif opcode == 0x168:             # <<<<<<<<<<<<<<
 *         return BIT_168(cpu)
 *     elif opcode == 0x169:
 */
    break;
    case 0x169:

    /* "pyboy/core/opcodes.py":5971
 *         return BIT_168(cpu)
 *     elif opcode == 0x169:
 *         return BIT_169(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x16A:
 *         return BIT_16A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_169(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5970
 *     elif opcode == 0x168:
 *         return BIT_168(cpu)
 *     elif opcode == 0x169:             # <<<<<<<<<<<<<<
 *         return BIT_169(cpu)
 *     elif opcode == 0x16A:
 */
    break;
    case 0x16A:

    /* "pyboy/core/opcodes.py":5973
 *         return BIT_169(cpu)
 *     elif opcode == 0x16A:
 *         return BIT_16A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x16B:
 *         return BIT_16B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_16A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5972
 *     elif opcode == 0x169:
 *         return BIT_169(cpu)
 *     elif opcode == 0x16A:             # <<<<<<<<<<<<<<
 *         return BIT_16A(cpu)
 *     elif opcode == 0x16B:
 */
    break;
    case 0x16B:

    /* "pyboy/core/opcodes.py":5975
 *         return BIT_16A(cpu)
 *     elif opcode == 0x16B:
 *         return BIT_16B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x16C:
 *         return BIT_16C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_16B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5974
 *     elif opcode == 0x16A:
 *         return BIT_16A(cpu)
 *     elif opcode == 0x16B:             # <<<<<<<<<<<<<<
 *         return BIT_16B(cpu)
 *     elif opcode == 0x16C:
 */
    break;
    case 0x16C:

    /* "pyboy/core/opcodes.py":5977
 *         return BIT_16B(cpu)
 *     elif opcode == 0x16C:
 *         return BIT_16C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x16D:
 *         return BIT_16D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_16C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5976
 *     elif opcode == 0x16B:
 *         return BIT_16B(cpu)
 *     elif opcode == 0x16C:             # <<<<<<<<<<<<<<
 *         return BIT_16C(cpu)
 *     elif opcode == 0x16D:
 */
    break;
    case 0x16D:

    /* "pyboy/core/opcodes.py":5979
 *         return BIT_16C(cpu)
 *     elif opcode == 0x16D:
 *         return BIT_16D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x16E:
 *         return BIT_16E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_16D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5978
 *     elif opcode == 0x16C:
 *         return BIT_16C(cpu)
 *     elif opcode == 0x16D:             # <<<<<<<<<<<<<<
 *         return BIT_16D(cpu)
 *     elif opcode == 0x16E:
 */
    break;
    case 0x16E:

    /* "pyboy/core/opcodes.py":5981
 *         return BIT_16D(cpu)
 *     elif opcode == 0x16E:
 *         return BIT_16E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x16F:
 *         return BIT_16F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_16E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5980
 *     elif opcode == 0x16D:
 *         return BIT_16D(cpu)
 *     elif opcode == 0x16E:             # <<<<<<<<<<<<<<
 *         return BIT_16E(cpu)
 *     elif opcode == 0x16F:
 */
    break;
    case 0x16F:

    /* "pyboy/core/opcodes.py":5983
 *         return BIT_16E(cpu)
 *     elif opcode == 0x16F:
 *         return BIT_16F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x170:
 *         return BIT_170(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_16F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5982
 *     elif opcode == 0x16E:
 *         return BIT_16E(cpu)
 *     elif opcode == 0x16F:             # <<<<<<<<<<<<<<
 *         return BIT_16F(cpu)
 *     elif opcode == 0x170:
 */
    break;
    case 0x170:

    /* "pyboy/core/opcodes.py":5985
 *         return BIT_16F(cpu)
 *     elif opcode == 0x170:
 *         return BIT_170(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x171:
 *         return BIT_171(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_170(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5984
 *     elif opcode == 0x16F:
 *         return BIT_16F(cpu)
 *     elif opcode == 0x170:             # <<<<<<<<<<<<<<
 *         return BIT_170(cpu)
 *     elif opcode == 0x171:
 */
    break;
    case 0x171:

    /* "pyboy/core/opcodes.py":5987
 *         return BIT_170(cpu)
 *     elif opcode == 0x171:
 *         return BIT_171(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x172:
 *         return BIT_172(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_171(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5986
 *     elif opcode == 0x170:
 *         return BIT_170(cpu)
 *     elif opcode == 0x171:             # <<<<<<<<<<<<<<
 *         return BIT_171(cpu)
 *     elif opcode == 0x172:
 */
    break;
    case 0x172:

    /* "pyboy/core/opcodes.py":5989
 *         return BIT_171(cpu)
 *     elif opcode == 0x172:
 *         return BIT_172(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x173:
 *         return BIT_173(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_172(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5988
 *     elif opcode == 0x171:
 *         return BIT_171(cpu)
 *     elif opcode == 0x172:             # <<<<<<<<<<<<<<
 *         return BIT_172(cpu)
 *     elif opcode == 0x173:
 */
    break;
    case 0x173:

    /* "pyboy/core/opcodes.py":5991
 *         return BIT_172(cpu)
 *     elif opcode == 0x173:
 *         return BIT_173(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x174:
 *         return BIT_174(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_173(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5990
 *     elif opcode == 0x172:
 *         return BIT_172(cpu)
 *     elif opcode == 0x173:             # <<<<<<<<<<<<<<
 *         return BIT_173(cpu)
 *     elif opcode == 0x174:
 */
    break;
    case 0x174:

    /* "pyboy/core/opcodes.py":5993
 *         return BIT_173(cpu)
 *     elif opcode == 0x174:
 *         return BIT_174(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x175:
 *         return BIT_175(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_174(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5992
 *     elif opcode == 0x173:
 *         return BIT_173(cpu)
 *     elif opcode == 0x174:             # <<<<<<<<<<<<<<
 *         return BIT_174(cpu)
 *     elif opcode == 0x175:
 */
    break;
    case 0x175:

    /* "pyboy/core/opcodes.py":5995
 *         return BIT_174(cpu)
 *     elif opcode == 0x175:
 *         return BIT_175(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x176:
 *         return BIT_176(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_175(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5994
 *     elif opcode == 0x174:
 *         return BIT_174(cpu)
 *     elif opcode == 0x175:             # <<<<<<<<<<<<<<
 *         return BIT_175(cpu)
 *     elif opcode == 0x176:
 */
    break;
    case 0x176:

    /* "pyboy/core/opcodes.py":5997
 *         return BIT_175(cpu)
 *     elif opcode == 0x176:
 *         return BIT_176(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x177:
 *         return BIT_177(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_176(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5996
 *     elif opcode == 0x175:
 *         return BIT_175(cpu)
 *     elif opcode == 0x176:             # <<<<<<<<<<<<<<
 *         return BIT_176(cpu)
 *     elif opcode == 0x177:
 */
    break;
    case 0x177:

    /* "pyboy/core/opcodes.py":5999
 *         return BIT_176(cpu)
 *     elif opcode == 0x177:
 *         return BIT_177(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x178:
 *         return BIT_178(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_177(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":5998
 *     elif opcode == 0x176:
 *         return BIT_176(cpu)
 *     elif opcode == 0x177:             # <<<<<<<<<<<<<<
 *         return BIT_177(cpu)
 *     elif opcode == 0x178:
 */
    break;
    case 0x178:

    /* "pyboy/core/opcodes.py":6001
 *         return BIT_177(cpu)
 *     elif opcode == 0x178:
 *         return BIT_178(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x179:
 *         return BIT_179(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_178(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6000
 *     elif opcode == 0x177:
 *         return BIT_177(cpu)
 *     elif opcode == 0x178:             # <<<<<<<<<<<<<<
 *         return BIT_178(cpu)
 *     elif opcode == 0x179:
 */
    break;
    case 0x179:

    /* "pyboy/core/opcodes.py":6003
 *         return BIT_178(cpu)
 *     elif opcode == 0x179:
 *         return BIT_179(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x17A:
 *         return BIT_17A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_179(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6002
 *     elif opcode == 0x178:
 *         return BIT_178(cpu)
 *     elif opcode == 0x179:             # <<<<<<<<<<<<<<
 *         return BIT_179(cpu)
 *     elif opcode == 0x17A:
 */
    break;
    case 0x17A:

    /* "pyboy/core/opcodes.py":6005
 *         return BIT_179(cpu)
 *     elif opcode == 0x17A:
 *         return BIT_17A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x17B:
 *         return BIT_17B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_17A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6004
 *     elif opcode == 0x179:
 *         return BIT_179(cpu)
 *     elif opcode == 0x17A:             # <<<<<<<<<<<<<<
 *         return BIT_17A(cpu)
 *     elif opcode == 0x17B:
 */
    break;
    case 0x17B:

    /* "pyboy/core/opcodes.py":6007
 *         return BIT_17A(cpu)
 *     elif opcode == 0x17B:
 *         return BIT_17B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x17C:
 *         return BIT_17C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_17B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6006
 *     elif opcode == 0x17A:
 *         return BIT_17A(cpu)
 *     elif opcode == 0x17B:             # <<<<<<<<<<<<<<
 *         return BIT_17B(cpu)
 *     elif opcode == 0x17C:
 */
    break;
    case 0x17C:

    /* "pyboy/core/opcodes.py":6009
 *         return BIT_17B(cpu)
 *     elif opcode == 0x17C:
 *         return BIT_17C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x17D:
 *         return BIT_17D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_17C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6008
 *     elif opcode == 0x17B:
 *         return BIT_17B(cpu)
 *     elif opcode == 0x17C:             # <<<<<<<<<<<<<<
 *         return BIT_17C(cpu)
 *     elif opcode == 0x17D:
 */
    break;
    case 0x17D:

    /* "pyboy/core/opcodes.py":6011
 *         return BIT_17C(cpu)
 *     elif opcode == 0x17D:
 *         return BIT_17D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x17E:
 *         return BIT_17E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_17D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6010
 *     elif opcode == 0x17C:
 *         return BIT_17C(cpu)
 *     elif opcode == 0x17D:             # <<<<<<<<<<<<<<
 *         return BIT_17D(cpu)
 *     elif opcode == 0x17E:
 */
    break;
    case 0x17E:

    /* "pyboy/core/opcodes.py":6013
 *         return BIT_17D(cpu)
 *     elif opcode == 0x17E:
 *         return BIT_17E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x17F:
 *         return BIT_17F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_17E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6012
 *     elif opcode == 0x17D:
 *         return BIT_17D(cpu)
 *     elif opcode == 0x17E:             # <<<<<<<<<<<<<<
 *         return BIT_17E(cpu)
 *     elif opcode == 0x17F:
 */
    break;
    case 0x17F:

    /* "pyboy/core/opcodes.py":6015
 *         return BIT_17E(cpu)
 *     elif opcode == 0x17F:
 *         return BIT_17F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x180:
 *         return RES_180(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_BIT_17F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6014
 *     elif opcode == 0x17E:
 *         return BIT_17E(cpu)
 *     elif opcode == 0x17F:             # <<<<<<<<<<<<<<
 *         return BIT_17F(cpu)
 *     elif opcode == 0x180:
 */
    break;
    case 0x180:

    /* "pyboy/core/opcodes.py":6017
 *         return BIT_17F(cpu)
 *     elif opcode == 0x180:
 *         return RES_180(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x181:
 *         return RES_181(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_180(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6016
 *     elif opcode == 0x17F:
 *         return BIT_17F(cpu)
 *     elif opcode == 0x180:             # <<<<<<<<<<<<<<
 *         return RES_180(cpu)
 *     elif opcode == 0x181:
 */
    break;
    case 0x181:

    /* "pyboy/core/opcodes.py":6019
 *         return RES_180(cpu)
 *     elif opcode == 0x181:
 *         return RES_181(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x182:
 *         return RES_182(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_181(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6018
 *     elif opcode == 0x180:
 *         return RES_180(cpu)
 *     elif opcode == 0x181:             # <<<<<<<<<<<<<<
 *         return RES_181(cpu)
 *     elif opcode == 0x182:
 */
    break;
    case 0x182:

    /* "pyboy/core/opcodes.py":6021
 *         return RES_181(cpu)
 *     elif opcode == 0x182:
 *         return RES_182(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x183:
 *         return RES_183(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_182(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6020
 *     elif opcode == 0x181:
 *         return RES_181(cpu)
 *     elif opcode == 0x182:             # <<<<<<<<<<<<<<
 *         return RES_182(cpu)
 *     elif opcode == 0x183:
 */
    break;
    case 0x183:

    /* "pyboy/core/opcodes.py":6023
 *         return RES_182(cpu)
 *     elif opcode == 0x183:
 *         return RES_183(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x184:
 *         return RES_184(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_183(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6022
 *     elif opcode == 0x182:
 *         return RES_182(cpu)
 *     elif opcode == 0x183:             # <<<<<<<<<<<<<<
 *         return RES_183(cpu)
 *     elif opcode == 0x184:
 */
    break;
    case 0x184:

    /* "pyboy/core/opcodes.py":6025
 *         return RES_183(cpu)
 *     elif opcode == 0x184:
 *         return RES_184(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x185:
 *         return RES_185(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_184(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6024
 *     elif opcode == 0x183:
 *         return RES_183(cpu)
 *     elif opcode == 0x184:             # <<<<<<<<<<<<<<
 *         return RES_184(cpu)
 *     elif opcode == 0x185:
 */
    break;
    case 0x185:

    /* "pyboy/core/opcodes.py":6027
 *         return RES_184(cpu)
 *     elif opcode == 0x185:
 *         return RES_185(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x186:
 *         return RES_186(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_185(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6026
 *     elif opcode == 0x184:
 *         return RES_184(cpu)
 *     elif opcode == 0x185:             # <<<<<<<<<<<<<<
 *         return RES_185(cpu)
 *     elif opcode == 0x186:
 */
    break;
    case 0x186:

    /* "pyboy/core/opcodes.py":6029
 *         return RES_185(cpu)
 *     elif opcode == 0x186:
 *         return RES_186(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x187:
 *         return RES_187(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_186(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6028
 *     elif opcode == 0x185:
 *         return RES_185(cpu)
 *     elif opcode == 0x186:             # <<<<<<<<<<<<<<
 *         return RES_186(cpu)
 *     elif opcode == 0x187:
 */
    break;
    case 0x187:

    /* "pyboy/core/opcodes.py":6031
 *         return RES_186(cpu)
 *     elif opcode == 0x187:
 *         return RES_187(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x188:
 *         return RES_188(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_187(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6030
 *     elif opcode == 0x186:
 *         return RES_186(cpu)
 *     elif opcode == 0x187:             # <<<<<<<<<<<<<<
 *         return RES_187(cpu)
 *     elif opcode == 0x188:
 */
    break;
    case 0x188:

    /* "pyboy/core/opcodes.py":6033
 *         return RES_187(cpu)
 *     elif opcode == 0x188:
 *         return RES_188(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x189:
 *         return RES_189(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_188(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6032
 *     elif opcode == 0x187:
 *         return RES_187(cpu)
 *     elif opcode == 0x188:             # <<<<<<<<<<<<<<
 *         return RES_188(cpu)
 *     elif opcode == 0x189:
 */
    break;
    case 0x189:

    /* "pyboy/core/opcodes.py":6035
 *         return RES_188(cpu)
 *     elif opcode == 0x189:
 *         return RES_189(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x18A:
 *         return RES_18A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_189(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6034
 *     elif opcode == 0x188:
 *         return RES_188(cpu)
 *     elif opcode == 0x189:             # <<<<<<<<<<<<<<
 *         return RES_189(cpu)
 *     elif opcode == 0x18A:
 */
    break;
    case 0x18A:

    /* "pyboy/core/opcodes.py":6037
 *         return RES_189(cpu)
 *     elif opcode == 0x18A:
 *         return RES_18A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x18B:
 *         return RES_18B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_18A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6036
 *     elif opcode == 0x189:
 *         return RES_189(cpu)
 *     elif opcode == 0x18A:             # <<<<<<<<<<<<<<
 *         return RES_18A(cpu)
 *     elif opcode == 0x18B:
 */
    break;
    case 0x18B:

    /* "pyboy/core/opcodes.py":6039
 *         return RES_18A(cpu)
 *     elif opcode == 0x18B:
 *         return RES_18B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x18C:
 *         return RES_18C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_18B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6038
 *     elif opcode == 0x18A:
 *         return RES_18A(cpu)
 *     elif opcode == 0x18B:             # <<<<<<<<<<<<<<
 *         return RES_18B(cpu)
 *     elif opcode == 0x18C:
 */
    break;
    case 0x18C:

    /* "pyboy/core/opcodes.py":6041
 *         return RES_18B(cpu)
 *     elif opcode == 0x18C:
 *         return RES_18C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x18D:
 *         return RES_18D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_18C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6040
 *     elif opcode == 0x18B:
 *         return RES_18B(cpu)
 *     elif opcode == 0x18C:             # <<<<<<<<<<<<<<
 *         return RES_18C(cpu)
 *     elif opcode == 0x18D:
 */
    break;
    case 0x18D:

    /* "pyboy/core/opcodes.py":6043
 *         return RES_18C(cpu)
 *     elif opcode == 0x18D:
 *         return RES_18D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x18E:
 *         return RES_18E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_18D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6042
 *     elif opcode == 0x18C:
 *         return RES_18C(cpu)
 *     elif opcode == 0x18D:             # <<<<<<<<<<<<<<
 *         return RES_18D(cpu)
 *     elif opcode == 0x18E:
 */
    break;
    case 0x18E:

    /* "pyboy/core/opcodes.py":6045
 *         return RES_18D(cpu)
 *     elif opcode == 0x18E:
 *         return RES_18E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x18F:
 *         return RES_18F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_18E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6044
 *     elif opcode == 0x18D:
 *         return RES_18D(cpu)
 *     elif opcode == 0x18E:             # <<<<<<<<<<<<<<
 *         return RES_18E(cpu)
 *     elif opcode == 0x18F:
 */
    break;
    case 0x18F:

    /* "pyboy/core/opcodes.py":6047
 *         return RES_18E(cpu)
 *     elif opcode == 0x18F:
 *         return RES_18F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x190:
 *         return RES_190(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_18F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6046
 *     elif opcode == 0x18E:
 *         return RES_18E(cpu)
 *     elif opcode == 0x18F:             # <<<<<<<<<<<<<<
 *         return RES_18F(cpu)
 *     elif opcode == 0x190:
 */
    break;
    case 0x190:

    /* "pyboy/core/opcodes.py":6049
 *         return RES_18F(cpu)
 *     elif opcode == 0x190:
 *         return RES_190(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x191:
 *         return RES_191(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_190(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6048
 *     elif opcode == 0x18F:
 *         return RES_18F(cpu)
 *     elif opcode == 0x190:             # <<<<<<<<<<<<<<
 *         return RES_190(cpu)
 *     elif opcode == 0x191:
 */
    break;
    case 0x191:

    /* "pyboy/core/opcodes.py":6051
 *         return RES_190(cpu)
 *     elif opcode == 0x191:
 *         return RES_191(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x192:
 *         return RES_192(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_191(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6050
 *     elif opcode == 0x190:
 *         return RES_190(cpu)
 *     elif opcode == 0x191:             # <<<<<<<<<<<<<<
 *         return RES_191(cpu)
 *     elif opcode == 0x192:
 */
    break;
    case 0x192:

    /* "pyboy/core/opcodes.py":6053
 *         return RES_191(cpu)
 *     elif opcode == 0x192:
 *         return RES_192(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x193:
 *         return RES_193(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_192(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6052
 *     elif opcode == 0x191:
 *         return RES_191(cpu)
 *     elif opcode == 0x192:             # <<<<<<<<<<<<<<
 *         return RES_192(cpu)
 *     elif opcode == 0x193:
 */
    break;
    case 0x193:

    /* "pyboy/core/opcodes.py":6055
 *         return RES_192(cpu)
 *     elif opcode == 0x193:
 *         return RES_193(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x194:
 *         return RES_194(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_193(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6054
 *     elif opcode == 0x192:
 *         return RES_192(cpu)
 *     elif opcode == 0x193:             # <<<<<<<<<<<<<<
 *         return RES_193(cpu)
 *     elif opcode == 0x194:
 */
    break;
    case 0x194:

    /* "pyboy/core/opcodes.py":6057
 *         return RES_193(cpu)
 *     elif opcode == 0x194:
 *         return RES_194(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x195:
 *         return RES_195(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_194(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6056
 *     elif opcode == 0x193:
 *         return RES_193(cpu)
 *     elif opcode == 0x194:             # <<<<<<<<<<<<<<
 *         return RES_194(cpu)
 *     elif opcode == 0x195:
 */
    break;
    case 0x195:

    /* "pyboy/core/opcodes.py":6059
 *         return RES_194(cpu)
 *     elif opcode == 0x195:
 *         return RES_195(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x196:
 *         return RES_196(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_195(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6058
 *     elif opcode == 0x194:
 *         return RES_194(cpu)
 *     elif opcode == 0x195:             # <<<<<<<<<<<<<<
 *         return RES_195(cpu)
 *     elif opcode == 0x196:
 */
    break;
    case 0x196:

    /* "pyboy/core/opcodes.py":6061
 *         return RES_195(cpu)
 *     elif opcode == 0x196:
 *         return RES_196(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x197:
 *         return RES_197(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_196(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6060
 *     elif opcode == 0x195:
 *         return RES_195(cpu)
 *     elif opcode == 0x196:             # <<<<<<<<<<<<<<
 *         return RES_196(cpu)
 *     elif opcode == 0x197:
 */
    break;
    case 0x197:

    /* "pyboy/core/opcodes.py":6063
 *         return RES_196(cpu)
 *     elif opcode == 0x197:
 *         return RES_197(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x198:
 *         return RES_198(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_197(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6062
 *     elif opcode == 0x196:
 *         return RES_196(cpu)
 *     elif opcode == 0x197:             # <<<<<<<<<<<<<<
 *         return RES_197(cpu)
 *     elif opcode == 0x198:
 */
    break;
    case 0x198:

    /* "pyboy/core/opcodes.py":6065
 *         return RES_197(cpu)
 *     elif opcode == 0x198:
 *         return RES_198(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x199:
 *         return RES_199(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_198(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6064
 *     elif opcode == 0x197:
 *         return RES_197(cpu)
 *     elif opcode == 0x198:             # <<<<<<<<<<<<<<
 *         return RES_198(cpu)
 *     elif opcode == 0x199:
 */
    break;
    case 0x199:

    /* "pyboy/core/opcodes.py":6067
 *         return RES_198(cpu)
 *     elif opcode == 0x199:
 *         return RES_199(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x19A:
 *         return RES_19A(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_199(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6066
 *     elif opcode == 0x198:
 *         return RES_198(cpu)
 *     elif opcode == 0x199:             # <<<<<<<<<<<<<<
 *         return RES_199(cpu)
 *     elif opcode == 0x19A:
 */
    break;
    case 0x19A:

    /* "pyboy/core/opcodes.py":6069
 *         return RES_199(cpu)
 *     elif opcode == 0x19A:
 *         return RES_19A(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x19B:
 *         return RES_19B(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_19A(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6068
 *     elif opcode == 0x199:
 *         return RES_199(cpu)
 *     elif opcode == 0x19A:             # <<<<<<<<<<<<<<
 *         return RES_19A(cpu)
 *     elif opcode == 0x19B:
 */
    break;
    case 0x19B:

    /* "pyboy/core/opcodes.py":6071
 *         return RES_19A(cpu)
 *     elif opcode == 0x19B:
 *         return RES_19B(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x19C:
 *         return RES_19C(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_19B(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6070
 *     elif opcode == 0x19A:
 *         return RES_19A(cpu)
 *     elif opcode == 0x19B:             # <<<<<<<<<<<<<<
 *         return RES_19B(cpu)
 *     elif opcode == 0x19C:
 */
    break;
    case 0x19C:

    /* "pyboy/core/opcodes.py":6073
 *         return RES_19B(cpu)
 *     elif opcode == 0x19C:
 *         return RES_19C(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x19D:
 *         return RES_19D(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_19C(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6072
 *     elif opcode == 0x19B:
 *         return RES_19B(cpu)
 *     elif opcode == 0x19C:             # <<<<<<<<<<<<<<
 *         return RES_19C(cpu)
 *     elif opcode == 0x19D:
 */
    break;
    case 0x19D:

    /* "pyboy/core/opcodes.py":6075
 *         return RES_19C(cpu)
 *     elif opcode == 0x19D:
 *         return RES_19D(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x19E:
 *         return RES_19E(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_19D(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6074
 *     elif opcode == 0x19C:
 *         return RES_19C(cpu)
 *     elif opcode == 0x19D:             # <<<<<<<<<<<<<<
 *         return RES_19D(cpu)
 *     elif opcode == 0x19E:
 */
    break;
    case 0x19E:

    /* "pyboy/core/opcodes.py":6077
 *         return RES_19D(cpu)
 *     elif opcode == 0x19E:
 *         return RES_19E(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x19F:
 *         return RES_19F(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_19E(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6076
 *     elif opcode == 0x19D:
 *         return RES_19D(cpu)
 *     elif opcode == 0x19E:             # <<<<<<<<<<<<<<
 *         return RES_19E(cpu)
 *     elif opcode == 0x19F:
 */
    break;
    case 0x19F:

    /* "pyboy/core/opcodes.py":6079
 *         return RES_19E(cpu)
 *     elif opcode == 0x19F:
 *         return RES_19F(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A0:
 *         return RES_1A0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_19F(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6078
 *     elif opcode == 0x19E:
 *         return RES_19E(cpu)
 *     elif opcode == 0x19F:             # <<<<<<<<<<<<<<
 *         return RES_19F(cpu)
 *     elif opcode == 0x1A0:
 */
    break;
    case 0x1A0:

    /* "pyboy/core/opcodes.py":6081
 *         return RES_19F(cpu)
 *     elif opcode == 0x1A0:
 *         return RES_1A0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A1:
 *         return RES_1A1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6080
 *     elif opcode == 0x19F:
 *         return RES_19F(cpu)
 *     elif opcode == 0x1A0:             # <<<<<<<<<<<<<<
 *         return RES_1A0(cpu)
 *     elif opcode == 0x1A1:
 */
    break;
    case 0x1A1:

    /* "pyboy/core/opcodes.py":6083
 *         return RES_1A0(cpu)
 *     elif opcode == 0x1A1:
 *         return RES_1A1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A2:
 *         return RES_1A2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6082
 *     elif opcode == 0x1A0:
 *         return RES_1A0(cpu)
 *     elif opcode == 0x1A1:             # <<<<<<<<<<<<<<
 *         return RES_1A1(cpu)
 *     elif opcode == 0x1A2:
 */
    break;
    case 0x1A2:

    /* "pyboy/core/opcodes.py":6085
 *         return RES_1A1(cpu)
 *     elif opcode == 0x1A2:
 *         return RES_1A2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A3:
 *         return RES_1A3(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6084
 *     elif opcode == 0x1A1:
 *         return RES_1A1(cpu)
 *     elif opcode == 0x1A2:             # <<<<<<<<<<<<<<
 *         return RES_1A2(cpu)
 *     elif opcode == 0x1A3:
 */
    break;
    case 0x1A3:

    /* "pyboy/core/opcodes.py":6087
 *         return RES_1A2(cpu)
 *     elif opcode == 0x1A3:
 *         return RES_1A3(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A4:
 *         return RES_1A4(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A3(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6086
 *     elif opcode == 0x1A2:
 *         return RES_1A2(cpu)
 *     elif opcode == 0x1A3:             # <<<<<<<<<<<<<<
 *         return RES_1A3(cpu)
 *     elif opcode == 0x1A4:
 */
    break;
    case 0x1A4:

    /* "pyboy/core/opcodes.py":6089
 *         return RES_1A3(cpu)
 *     elif opcode == 0x1A4:
 *         return RES_1A4(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A5:
 *         return RES_1A5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A4(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6088
 *     elif opcode == 0x1A3:
 *         return RES_1A3(cpu)
 *     elif opcode == 0x1A4:             # <<<<<<<<<<<<<<
 *         return RES_1A4(cpu)
 *     elif opcode == 0x1A5:
 */
    break;
    case 0x1A5:

    /* "pyboy/core/opcodes.py":6091
 *         return RES_1A4(cpu)
 *     elif opcode == 0x1A5:
 *         return RES_1A5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A6:
 *         return RES_1A6(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6090
 *     elif opcode == 0x1A4:
 *         return RES_1A4(cpu)
 *     elif opcode == 0x1A5:             # <<<<<<<<<<<<<<
 *         return RES_1A5(cpu)
 *     elif opcode == 0x1A6:
 */
    break;
    case 0x1A6:

    /* "pyboy/core/opcodes.py":6093
 *         return RES_1A5(cpu)
 *     elif opcode == 0x1A6:
 *         return RES_1A6(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A7:
 *         return RES_1A7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A6(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6092
 *     elif opcode == 0x1A5:
 *         return RES_1A5(cpu)
 *     elif opcode == 0x1A6:             # <<<<<<<<<<<<<<
 *         return RES_1A6(cpu)
 *     elif opcode == 0x1A7:
 */
    break;
    case 0x1A7:

    /* "pyboy/core/opcodes.py":6095
 *         return RES_1A6(cpu)
 *     elif opcode == 0x1A7:
 *         return RES_1A7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A8:
 *         return RES_1A8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6094
 *     elif opcode == 0x1A6:
 *         return RES_1A6(cpu)
 *     elif opcode == 0x1A7:             # <<<<<<<<<<<<<<
 *         return RES_1A7(cpu)
 *     elif opcode == 0x1A8:
 */
    break;
    case 0x1A8:

    /* "pyboy/core/opcodes.py":6097
 *         return RES_1A7(cpu)
 *     elif opcode == 0x1A8:
 *         return RES_1A8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1A9:
 *         return RES_1A9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6096
 *     elif opcode == 0x1A7:
 *         return RES_1A7(cpu)
 *     elif opcode == 0x1A8:             # <<<<<<<<<<<<<<
 *         return RES_1A8(cpu)
 *     elif opcode == 0x1A9:
 */
    break;
    case 0x1A9:

    /* "pyboy/core/opcodes.py":6099
 *         return RES_1A8(cpu)
 *     elif opcode == 0x1A9:
 *         return RES_1A9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1AA:
 *         return RES_1AA(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1A9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6098
 *     elif opcode == 0x1A8:
 *         return RES_1A8(cpu)
 *     elif opcode == 0x1A9:             # <<<<<<<<<<<<<<
 *         return RES_1A9(cpu)
 *     elif opcode == 0x1AA:
 */
    break;
    case 0x1AA:

    /* "pyboy/core/opcodes.py":6101
 *         return RES_1A9(cpu)
 *     elif opcode == 0x1AA:
 *         return RES_1AA(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1AB:
 *         return RES_1AB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1AA(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6100
 *     elif opcode == 0x1A9:
 *         return RES_1A9(cpu)
 *     elif opcode == 0x1AA:             # <<<<<<<<<<<<<<
 *         return RES_1AA(cpu)
 *     elif opcode == 0x1AB:
 */
    break;
    case 0x1AB:

    /* "pyboy/core/opcodes.py":6103
 *         return RES_1AA(cpu)
 *     elif opcode == 0x1AB:
 *         return RES_1AB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1AC:
 *         return RES_1AC(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1AB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6102
 *     elif opcode == 0x1AA:
 *         return RES_1AA(cpu)
 *     elif opcode == 0x1AB:             # <<<<<<<<<<<<<<
 *         return RES_1AB(cpu)
 *     elif opcode == 0x1AC:
 */
    break;
    case 0x1AC:

    /* "pyboy/core/opcodes.py":6105
 *         return RES_1AB(cpu)
 *     elif opcode == 0x1AC:
 *         return RES_1AC(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1AD:
 *         return RES_1AD(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1AC(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6104
 *     elif opcode == 0x1AB:
 *         return RES_1AB(cpu)
 *     elif opcode == 0x1AC:             # <<<<<<<<<<<<<<
 *         return RES_1AC(cpu)
 *     elif opcode == 0x1AD:
 */
    break;
    case 0x1AD:

    /* "pyboy/core/opcodes.py":6107
 *         return RES_1AC(cpu)
 *     elif opcode == 0x1AD:
 *         return RES_1AD(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1AE:
 *         return RES_1AE(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1AD(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6106
 *     elif opcode == 0x1AC:
 *         return RES_1AC(cpu)
 *     elif opcode == 0x1AD:             # <<<<<<<<<<<<<<
 *         return RES_1AD(cpu)
 *     elif opcode == 0x1AE:
 */
    break;
    case 0x1AE:

    /* "pyboy/core/opcodes.py":6109
 *         return RES_1AD(cpu)
 *     elif opcode == 0x1AE:
 *         return RES_1AE(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1AF:
 *         return RES_1AF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1AE(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6108
 *     elif opcode == 0x1AD:
 *         return RES_1AD(cpu)
 *     elif opcode == 0x1AE:             # <<<<<<<<<<<<<<
 *         return RES_1AE(cpu)
 *     elif opcode == 0x1AF:
 */
    break;
    case 0x1AF:

    /* "pyboy/core/opcodes.py":6111
 *         return RES_1AE(cpu)
 *     elif opcode == 0x1AF:
 *         return RES_1AF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B0:
 *         return RES_1B0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1AF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6110
 *     elif opcode == 0x1AE:
 *         return RES_1AE(cpu)
 *     elif opcode == 0x1AF:             # <<<<<<<<<<<<<<
 *         return RES_1AF(cpu)
 *     elif opcode == 0x1B0:
 */
    break;
    case 0x1B0:

    /* "pyboy/core/opcodes.py":6113
 *         return RES_1AF(cpu)
 *     elif opcode == 0x1B0:
 *         return RES_1B0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B1:
 *         return RES_1B1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6112
 *     elif opcode == 0x1AF:
 *         return RES_1AF(cpu)
 *     elif opcode == 0x1B0:             # <<<<<<<<<<<<<<
 *         return RES_1B0(cpu)
 *     elif opcode == 0x1B1:
 */
    break;
    case 0x1B1:

    /* "pyboy/core/opcodes.py":6115
 *         return RES_1B0(cpu)
 *     elif opcode == 0x1B1:
 *         return RES_1B1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B2:
 *         return RES_1B2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6114
 *     elif opcode == 0x1B0:
 *         return RES_1B0(cpu)
 *     elif opcode == 0x1B1:             # <<<<<<<<<<<<<<
 *         return RES_1B1(cpu)
 *     elif opcode == 0x1B2:
 */
    break;
    case 0x1B2:

    /* "pyboy/core/opcodes.py":6117
 *         return RES_1B1(cpu)
 *     elif opcode == 0x1B2:
 *         return RES_1B2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B3:
 *         return RES_1B3(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6116
 *     elif opcode == 0x1B1:
 *         return RES_1B1(cpu)
 *     elif opcode == 0x1B2:             # <<<<<<<<<<<<<<
 *         return RES_1B2(cpu)
 *     elif opcode == 0x1B3:
 */
    break;
    case 0x1B3:

    /* "pyboy/core/opcodes.py":6119
 *         return RES_1B2(cpu)
 *     elif opcode == 0x1B3:
 *         return RES_1B3(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B4:
 *         return RES_1B4(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B3(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6118
 *     elif opcode == 0x1B2:
 *         return RES_1B2(cpu)
 *     elif opcode == 0x1B3:             # <<<<<<<<<<<<<<
 *         return RES_1B3(cpu)
 *     elif opcode == 0x1B4:
 */
    break;
    case 0x1B4:

    /* "pyboy/core/opcodes.py":6121
 *         return RES_1B3(cpu)
 *     elif opcode == 0x1B4:
 *         return RES_1B4(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B5:
 *         return RES_1B5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B4(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6120
 *     elif opcode == 0x1B3:
 *         return RES_1B3(cpu)
 *     elif opcode == 0x1B4:             # <<<<<<<<<<<<<<
 *         return RES_1B4(cpu)
 *     elif opcode == 0x1B5:
 */
    break;
    case 0x1B5:

    /* "pyboy/core/opcodes.py":6123
 *         return RES_1B4(cpu)
 *     elif opcode == 0x1B5:
 *         return RES_1B5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B6:
 *         return RES_1B6(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6122
 *     elif opcode == 0x1B4:
 *         return RES_1B4(cpu)
 *     elif opcode == 0x1B5:             # <<<<<<<<<<<<<<
 *         return RES_1B5(cpu)
 *     elif opcode == 0x1B6:
 */
    break;
    case 0x1B6:

    /* "pyboy/core/opcodes.py":6125
 *         return RES_1B5(cpu)
 *     elif opcode == 0x1B6:
 *         return RES_1B6(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B7:
 *         return RES_1B7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B6(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6124
 *     elif opcode == 0x1B5:
 *         return RES_1B5(cpu)
 *     elif opcode == 0x1B6:             # <<<<<<<<<<<<<<
 *         return RES_1B6(cpu)
 *     elif opcode == 0x1B7:
 */
    break;
    case 0x1B7:

    /* "pyboy/core/opcodes.py":6127
 *         return RES_1B6(cpu)
 *     elif opcode == 0x1B7:
 *         return RES_1B7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B8:
 *         return RES_1B8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6126
 *     elif opcode == 0x1B6:
 *         return RES_1B6(cpu)
 *     elif opcode == 0x1B7:             # <<<<<<<<<<<<<<
 *         return RES_1B7(cpu)
 *     elif opcode == 0x1B8:
 */
    break;
    case 0x1B8:

    /* "pyboy/core/opcodes.py":6129
 *         return RES_1B7(cpu)
 *     elif opcode == 0x1B8:
 *         return RES_1B8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1B9:
 *         return RES_1B9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6128
 *     elif opcode == 0x1B7:
 *         return RES_1B7(cpu)
 *     elif opcode == 0x1B8:             # <<<<<<<<<<<<<<
 *         return RES_1B8(cpu)
 *     elif opcode == 0x1B9:
 */
    break;
    case 0x1B9:

    /* "pyboy/core/opcodes.py":6131
 *         return RES_1B8(cpu)
 *     elif opcode == 0x1B9:
 *         return RES_1B9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1BA:
 *         return RES_1BA(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1B9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6130
 *     elif opcode == 0x1B8:
 *         return RES_1B8(cpu)
 *     elif opcode == 0x1B9:             # <<<<<<<<<<<<<<
 *         return RES_1B9(cpu)
 *     elif opcode == 0x1BA:
 */
    break;
    case 0x1BA:

    /* "pyboy/core/opcodes.py":6133
 *         return RES_1B9(cpu)
 *     elif opcode == 0x1BA:
 *         return RES_1BA(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1BB:
 *         return RES_1BB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1BA(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6132
 *     elif opcode == 0x1B9:
 *         return RES_1B9(cpu)
 *     elif opcode == 0x1BA:             # <<<<<<<<<<<<<<
 *         return RES_1BA(cpu)
 *     elif opcode == 0x1BB:
 */
    break;
    case 0x1BB:

    /* "pyboy/core/opcodes.py":6135
 *         return RES_1BA(cpu)
 *     elif opcode == 0x1BB:
 *         return RES_1BB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1BC:
 *         return RES_1BC(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1BB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6134
 *     elif opcode == 0x1BA:
 *         return RES_1BA(cpu)
 *     elif opcode == 0x1BB:             # <<<<<<<<<<<<<<
 *         return RES_1BB(cpu)
 *     elif opcode == 0x1BC:
 */
    break;
    case 0x1BC:

    /* "pyboy/core/opcodes.py":6137
 *         return RES_1BB(cpu)
 *     elif opcode == 0x1BC:
 *         return RES_1BC(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1BD:
 *         return RES_1BD(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1BC(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6136
 *     elif opcode == 0x1BB:
 *         return RES_1BB(cpu)
 *     elif opcode == 0x1BC:             # <<<<<<<<<<<<<<
 *         return RES_1BC(cpu)
 *     elif opcode == 0x1BD:
 */
    break;
    case 0x1BD:

    /* "pyboy/core/opcodes.py":6139
 *         return RES_1BC(cpu)
 *     elif opcode == 0x1BD:
 *         return RES_1BD(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1BE:
 *         return RES_1BE(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1BD(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6138
 *     elif opcode == 0x1BC:
 *         return RES_1BC(cpu)
 *     elif opcode == 0x1BD:             # <<<<<<<<<<<<<<
 *         return RES_1BD(cpu)
 *     elif opcode == 0x1BE:
 */
    break;
    case 0x1BE:

    /* "pyboy/core/opcodes.py":6141
 *         return RES_1BD(cpu)
 *     elif opcode == 0x1BE:
 *         return RES_1BE(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1BF:
 *         return RES_1BF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1BE(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6140
 *     elif opcode == 0x1BD:
 *         return RES_1BD(cpu)
 *     elif opcode == 0x1BE:             # <<<<<<<<<<<<<<
 *         return RES_1BE(cpu)
 *     elif opcode == 0x1BF:
 */
    break;
    case 0x1BF:

    /* "pyboy/core/opcodes.py":6143
 *         return RES_1BE(cpu)
 *     elif opcode == 0x1BF:
 *         return RES_1BF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C0:
 *         return SET_1C0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_RES_1BF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6142
 *     elif opcode == 0x1BE:
 *         return RES_1BE(cpu)
 *     elif opcode == 0x1BF:             # <<<<<<<<<<<<<<
 *         return RES_1BF(cpu)
 *     elif opcode == 0x1C0:
 */
    break;
    case 0x1C0:

    /* "pyboy/core/opcodes.py":6145
 *         return RES_1BF(cpu)
 *     elif opcode == 0x1C0:
 *         return SET_1C0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C1:
 *         return SET_1C1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6144
 *     elif opcode == 0x1BF:
 *         return RES_1BF(cpu)
 *     elif opcode == 0x1C0:             # <<<<<<<<<<<<<<
 *         return SET_1C0(cpu)
 *     elif opcode == 0x1C1:
 */
    break;
    case 0x1C1:

    /* "pyboy/core/opcodes.py":6147
 *         return SET_1C0(cpu)
 *     elif opcode == 0x1C1:
 *         return SET_1C1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C2:
 *         return SET_1C2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6146
 *     elif opcode == 0x1C0:
 *         return SET_1C0(cpu)
 *     elif opcode == 0x1C1:             # <<<<<<<<<<<<<<
 *         return SET_1C1(cpu)
 *     elif opcode == 0x1C2:
 */
    break;
    case 0x1C2:

    /* "pyboy/core/opcodes.py":6149
 *         return SET_1C1(cpu)
 *     elif opcode == 0x1C2:
 *         return SET_1C2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C3:
 *         return SET_1C3(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6148
 *     elif opcode == 0x1C1:
 *         return SET_1C1(cpu)
 *     elif opcode == 0x1C2:             # <<<<<<<<<<<<<<
 *         return SET_1C2(cpu)
 *     elif opcode == 0x1C3:
 */
    break;
    case 0x1C3:

    /* "pyboy/core/opcodes.py":6151
 *         return SET_1C2(cpu)
 *     elif opcode == 0x1C3:
 *         return SET_1C3(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C4:
 *         return SET_1C4(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C3(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6150
 *     elif opcode == 0x1C2:
 *         return SET_1C2(cpu)
 *     elif opcode == 0x1C3:             # <<<<<<<<<<<<<<
 *         return SET_1C3(cpu)
 *     elif opcode == 0x1C4:
 */
    break;
    case 0x1C4:

    /* "pyboy/core/opcodes.py":6153
 *         return SET_1C3(cpu)
 *     elif opcode == 0x1C4:
 *         return SET_1C4(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C5:
 *         return SET_1C5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C4(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6152
 *     elif opcode == 0x1C3:
 *         return SET_1C3(cpu)
 *     elif opcode == 0x1C4:             # <<<<<<<<<<<<<<
 *         return SET_1C4(cpu)
 *     elif opcode == 0x1C5:
 */
    break;
    case 0x1C5:

    /* "pyboy/core/opcodes.py":6155
 *         return SET_1C4(cpu)
 *     elif opcode == 0x1C5:
 *         return SET_1C5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C6:
 *         return SET_1C6(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6154
 *     elif opcode == 0x1C4:
 *         return SET_1C4(cpu)
 *     elif opcode == 0x1C5:             # <<<<<<<<<<<<<<
 *         return SET_1C5(cpu)
 *     elif opcode == 0x1C6:
 */
    break;
    case 0x1C6:

    /* "pyboy/core/opcodes.py":6157
 *         return SET_1C5(cpu)
 *     elif opcode == 0x1C6:
 *         return SET_1C6(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C7:
 *         return SET_1C7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C6(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6156
 *     elif opcode == 0x1C5:
 *         return SET_1C5(cpu)
 *     elif opcode == 0x1C6:             # <<<<<<<<<<<<<<
 *         return SET_1C6(cpu)
 *     elif opcode == 0x1C7:
 */
    break;
    case 0x1C7:

    /* "pyboy/core/opcodes.py":6159
 *         return SET_1C6(cpu)
 *     elif opcode == 0x1C7:
 *         return SET_1C7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C8:
 *         return SET_1C8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6158
 *     elif opcode == 0x1C6:
 *         return SET_1C6(cpu)
 *     elif opcode == 0x1C7:             # <<<<<<<<<<<<<<
 *         return SET_1C7(cpu)
 *     elif opcode == 0x1C8:
 */
    break;
    case 0x1C8:

    /* "pyboy/core/opcodes.py":6161
 *         return SET_1C7(cpu)
 *     elif opcode == 0x1C8:
 *         return SET_1C8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1C9:
 *         return SET_1C9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6160
 *     elif opcode == 0x1C7:
 *         return SET_1C7(cpu)
 *     elif opcode == 0x1C8:             # <<<<<<<<<<<<<<
 *         return SET_1C8(cpu)
 *     elif opcode == 0x1C9:
 */
    break;
    case 0x1C9:

    /* "pyboy/core/opcodes.py":6163
 *         return SET_1C8(cpu)
 *     elif opcode == 0x1C9:
 *         return SET_1C9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1CA:
 *         return SET_1CA(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1C9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6162
 *     elif opcode == 0x1C8:
 *         return SET_1C8(cpu)
 *     elif opcode == 0x1C9:             # <<<<<<<<<<<<<<
 *         return SET_1C9(cpu)
 *     elif opcode == 0x1CA:
 */
    break;
    case 0x1CA:

    /* "pyboy/core/opcodes.py":6165
 *         return SET_1C9(cpu)
 *     elif opcode == 0x1CA:
 *         return SET_1CA(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1CB:
 *         return SET_1CB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1CA(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6164
 *     elif opcode == 0x1C9:
 *         return SET_1C9(cpu)
 *     elif opcode == 0x1CA:             # <<<<<<<<<<<<<<
 *         return SET_1CA(cpu)
 *     elif opcode == 0x1CB:
 */
    break;
    case 0x1CB:

    /* "pyboy/core/opcodes.py":6167
 *         return SET_1CA(cpu)
 *     elif opcode == 0x1CB:
 *         return SET_1CB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1CC:
 *         return SET_1CC(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1CB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6166
 *     elif opcode == 0x1CA:
 *         return SET_1CA(cpu)
 *     elif opcode == 0x1CB:             # <<<<<<<<<<<<<<
 *         return SET_1CB(cpu)
 *     elif opcode == 0x1CC:
 */
    break;
    case 0x1CC:

    /* "pyboy/core/opcodes.py":6169
 *         return SET_1CB(cpu)
 *     elif opcode == 0x1CC:
 *         return SET_1CC(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1CD:
 *         return SET_1CD(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1CC(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6168
 *     elif opcode == 0x1CB:
 *         return SET_1CB(cpu)
 *     elif opcode == 0x1CC:             # <<<<<<<<<<<<<<
 *         return SET_1CC(cpu)
 *     elif opcode == 0x1CD:
 */
    break;
    case 0x1CD:

    /* "pyboy/core/opcodes.py":6171
 *         return SET_1CC(cpu)
 *     elif opcode == 0x1CD:
 *         return SET_1CD(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1CE:
 *         return SET_1CE(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1CD(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6170
 *     elif opcode == 0x1CC:
 *         return SET_1CC(cpu)
 *     elif opcode == 0x1CD:             # <<<<<<<<<<<<<<
 *         return SET_1CD(cpu)
 *     elif opcode == 0x1CE:
 */
    break;
    case 0x1CE:

    /* "pyboy/core/opcodes.py":6173
 *         return SET_1CD(cpu)
 *     elif opcode == 0x1CE:
 *         return SET_1CE(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1CF:
 *         return SET_1CF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1CE(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6172
 *     elif opcode == 0x1CD:
 *         return SET_1CD(cpu)
 *     elif opcode == 0x1CE:             # <<<<<<<<<<<<<<
 *         return SET_1CE(cpu)
 *     elif opcode == 0x1CF:
 */
    break;
    case 0x1CF:

    /* "pyboy/core/opcodes.py":6175
 *         return SET_1CE(cpu)
 *     elif opcode == 0x1CF:
 *         return SET_1CF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D0:
 *         return SET_1D0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1CF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6174
 *     elif opcode == 0x1CE:
 *         return SET_1CE(cpu)
 *     elif opcode == 0x1CF:             # <<<<<<<<<<<<<<
 *         return SET_1CF(cpu)
 *     elif opcode == 0x1D0:
 */
    break;
    case 0x1D0:

    /* "pyboy/core/opcodes.py":6177
 *         return SET_1CF(cpu)
 *     elif opcode == 0x1D0:
 *         return SET_1D0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D1:
 *         return SET_1D1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6176
 *     elif opcode == 0x1CF:
 *         return SET_1CF(cpu)
 *     elif opcode == 0x1D0:             # <<<<<<<<<<<<<<
 *         return SET_1D0(cpu)
 *     elif opcode == 0x1D1:
 */
    break;
    case 0x1D1:

    /* "pyboy/core/opcodes.py":6179
 *         return SET_1D0(cpu)
 *     elif opcode == 0x1D1:
 *         return SET_1D1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D2:
 *         return SET_1D2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6178
 *     elif opcode == 0x1D0:
 *         return SET_1D0(cpu)
 *     elif opcode == 0x1D1:             # <<<<<<<<<<<<<<
 *         return SET_1D1(cpu)
 *     elif opcode == 0x1D2:
 */
    break;
    case 0x1D2:

    /* "pyboy/core/opcodes.py":6181
 *         return SET_1D1(cpu)
 *     elif opcode == 0x1D2:
 *         return SET_1D2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D3:
 *         return SET_1D3(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6180
 *     elif opcode == 0x1D1:
 *         return SET_1D1(cpu)
 *     elif opcode == 0x1D2:             # <<<<<<<<<<<<<<
 *         return SET_1D2(cpu)
 *     elif opcode == 0x1D3:
 */
    break;
    case 0x1D3:

    /* "pyboy/core/opcodes.py":6183
 *         return SET_1D2(cpu)
 *     elif opcode == 0x1D3:
 *         return SET_1D3(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D4:
 *         return SET_1D4(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D3(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6182
 *     elif opcode == 0x1D2:
 *         return SET_1D2(cpu)
 *     elif opcode == 0x1D3:             # <<<<<<<<<<<<<<
 *         return SET_1D3(cpu)
 *     elif opcode == 0x1D4:
 */
    break;
    case 0x1D4:

    /* "pyboy/core/opcodes.py":6185
 *         return SET_1D3(cpu)
 *     elif opcode == 0x1D4:
 *         return SET_1D4(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D5:
 *         return SET_1D5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D4(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6184
 *     elif opcode == 0x1D3:
 *         return SET_1D3(cpu)
 *     elif opcode == 0x1D4:             # <<<<<<<<<<<<<<
 *         return SET_1D4(cpu)
 *     elif opcode == 0x1D5:
 */
    break;
    case 0x1D5:

    /* "pyboy/core/opcodes.py":6187
 *         return SET_1D4(cpu)
 *     elif opcode == 0x1D5:
 *         return SET_1D5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D6:
 *         return SET_1D6(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6186
 *     elif opcode == 0x1D4:
 *         return SET_1D4(cpu)
 *     elif opcode == 0x1D5:             # <<<<<<<<<<<<<<
 *         return SET_1D5(cpu)
 *     elif opcode == 0x1D6:
 */
    break;
    case 0x1D6:

    /* "pyboy/core/opcodes.py":6189
 *         return SET_1D5(cpu)
 *     elif opcode == 0x1D6:
 *         return SET_1D6(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D7:
 *         return SET_1D7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D6(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6188
 *     elif opcode == 0x1D5:
 *         return SET_1D5(cpu)
 *     elif opcode == 0x1D6:             # <<<<<<<<<<<<<<
 *         return SET_1D6(cpu)
 *     elif opcode == 0x1D7:
 */
    break;
    case 0x1D7:

    /* "pyboy/core/opcodes.py":6191
 *         return SET_1D6(cpu)
 *     elif opcode == 0x1D7:
 *         return SET_1D7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D8:
 *         return SET_1D8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6190
 *     elif opcode == 0x1D6:
 *         return SET_1D6(cpu)
 *     elif opcode == 0x1D7:             # <<<<<<<<<<<<<<
 *         return SET_1D7(cpu)
 *     elif opcode == 0x1D8:
 */
    break;
    case 0x1D8:

    /* "pyboy/core/opcodes.py":6193
 *         return SET_1D7(cpu)
 *     elif opcode == 0x1D8:
 *         return SET_1D8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1D9:
 *         return SET_1D9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6192
 *     elif opcode == 0x1D7:
 *         return SET_1D7(cpu)
 *     elif opcode == 0x1D8:             # <<<<<<<<<<<<<<
 *         return SET_1D8(cpu)
 *     elif opcode == 0x1D9:
 */
    break;
    case 0x1D9:

    /* "pyboy/core/opcodes.py":6195
 *         return SET_1D8(cpu)
 *     elif opcode == 0x1D9:
 *         return SET_1D9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1DA:
 *         return SET_1DA(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1D9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6194
 *     elif opcode == 0x1D8:
 *         return SET_1D8(cpu)
 *     elif opcode == 0x1D9:             # <<<<<<<<<<<<<<
 *         return SET_1D9(cpu)
 *     elif opcode == 0x1DA:
 */
    break;
    case 0x1DA:

    /* "pyboy/core/opcodes.py":6197
 *         return SET_1D9(cpu)
 *     elif opcode == 0x1DA:
 *         return SET_1DA(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1DB:
 *         return SET_1DB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1DA(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6196
 *     elif opcode == 0x1D9:
 *         return SET_1D9(cpu)
 *     elif opcode == 0x1DA:             # <<<<<<<<<<<<<<
 *         return SET_1DA(cpu)
 *     elif opcode == 0x1DB:
 */
    break;
    case 0x1DB:

    /* "pyboy/core/opcodes.py":6199
 *         return SET_1DA(cpu)
 *     elif opcode == 0x1DB:
 *         return SET_1DB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1DC:
 *         return SET_1DC(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1DB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6198
 *     elif opcode == 0x1DA:
 *         return SET_1DA(cpu)
 *     elif opcode == 0x1DB:             # <<<<<<<<<<<<<<
 *         return SET_1DB(cpu)
 *     elif opcode == 0x1DC:
 */
    break;
    case 0x1DC:

    /* "pyboy/core/opcodes.py":6201
 *         return SET_1DB(cpu)
 *     elif opcode == 0x1DC:
 *         return SET_1DC(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1DD:
 *         return SET_1DD(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1DC(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6200
 *     elif opcode == 0x1DB:
 *         return SET_1DB(cpu)
 *     elif opcode == 0x1DC:             # <<<<<<<<<<<<<<
 *         return SET_1DC(cpu)
 *     elif opcode == 0x1DD:
 */
    break;
    case 0x1DD:

    /* "pyboy/core/opcodes.py":6203
 *         return SET_1DC(cpu)
 *     elif opcode == 0x1DD:
 *         return SET_1DD(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1DE:
 *         return SET_1DE(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1DD(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6202
 *     elif opcode == 0x1DC:
 *         return SET_1DC(cpu)
 *     elif opcode == 0x1DD:             # <<<<<<<<<<<<<<
 *         return SET_1DD(cpu)
 *     elif opcode == 0x1DE:
 */
    break;
    case 0x1DE:

    /* "pyboy/core/opcodes.py":6205
 *         return SET_1DD(cpu)
 *     elif opcode == 0x1DE:
 *         return SET_1DE(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1DF:
 *         return SET_1DF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1DE(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6204
 *     elif opcode == 0x1DD:
 *         return SET_1DD(cpu)
 *     elif opcode == 0x1DE:             # <<<<<<<<<<<<<<
 *         return SET_1DE(cpu)
 *     elif opcode == 0x1DF:
 */
    break;
    case 0x1DF:

    /* "pyboy/core/opcodes.py":6207
 *         return SET_1DE(cpu)
 *     elif opcode == 0x1DF:
 *         return SET_1DF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E0:
 *         return SET_1E0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1DF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6206
 *     elif opcode == 0x1DE:
 *         return SET_1DE(cpu)
 *     elif opcode == 0x1DF:             # <<<<<<<<<<<<<<
 *         return SET_1DF(cpu)
 *     elif opcode == 0x1E0:
 */
    break;
    case 0x1E0:

    /* "pyboy/core/opcodes.py":6209
 *         return SET_1DF(cpu)
 *     elif opcode == 0x1E0:
 *         return SET_1E0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E1:
 *         return SET_1E1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6208
 *     elif opcode == 0x1DF:
 *         return SET_1DF(cpu)
 *     elif opcode == 0x1E0:             # <<<<<<<<<<<<<<
 *         return SET_1E0(cpu)
 *     elif opcode == 0x1E1:
 */
    break;
    case 0x1E1:

    /* "pyboy/core/opcodes.py":6211
 *         return SET_1E0(cpu)
 *     elif opcode == 0x1E1:
 *         return SET_1E1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E2:
 *         return SET_1E2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6210
 *     elif opcode == 0x1E0:
 *         return SET_1E0(cpu)
 *     elif opcode == 0x1E1:             # <<<<<<<<<<<<<<
 *         return SET_1E1(cpu)
 *     elif opcode == 0x1E2:
 */
    break;
    case 0x1E2:

    /* "pyboy/core/opcodes.py":6213
 *         return SET_1E1(cpu)
 *     elif opcode == 0x1E2:
 *         return SET_1E2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E3:
 *         return SET_1E3(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6212
 *     elif opcode == 0x1E1:
 *         return SET_1E1(cpu)
 *     elif opcode == 0x1E2:             # <<<<<<<<<<<<<<
 *         return SET_1E2(cpu)
 *     elif opcode == 0x1E3:
 */
    break;
    case 0x1E3:

    /* "pyboy/core/opcodes.py":6215
 *         return SET_1E2(cpu)
 *     elif opcode == 0x1E3:
 *         return SET_1E3(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E4:
 *         return SET_1E4(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E3(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6214
 *     elif opcode == 0x1E2:
 *         return SET_1E2(cpu)
 *     elif opcode == 0x1E3:             # <<<<<<<<<<<<<<
 *         return SET_1E3(cpu)
 *     elif opcode == 0x1E4:
 */
    break;
    case 0x1E4:

    /* "pyboy/core/opcodes.py":6217
 *         return SET_1E3(cpu)
 *     elif opcode == 0x1E4:
 *         return SET_1E4(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E5:
 *         return SET_1E5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E4(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6216
 *     elif opcode == 0x1E3:
 *         return SET_1E3(cpu)
 *     elif opcode == 0x1E4:             # <<<<<<<<<<<<<<
 *         return SET_1E4(cpu)
 *     elif opcode == 0x1E5:
 */
    break;
    case 0x1E5:

    /* "pyboy/core/opcodes.py":6219
 *         return SET_1E4(cpu)
 *     elif opcode == 0x1E5:
 *         return SET_1E5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E6:
 *         return SET_1E6(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6218
 *     elif opcode == 0x1E4:
 *         return SET_1E4(cpu)
 *     elif opcode == 0x1E5:             # <<<<<<<<<<<<<<
 *         return SET_1E5(cpu)
 *     elif opcode == 0x1E6:
 */
    break;
    case 0x1E6:

    /* "pyboy/core/opcodes.py":6221
 *         return SET_1E5(cpu)
 *     elif opcode == 0x1E6:
 *         return SET_1E6(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E7:
 *         return SET_1E7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E6(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6220
 *     elif opcode == 0x1E5:
 *         return SET_1E5(cpu)
 *     elif opcode == 0x1E6:             # <<<<<<<<<<<<<<
 *         return SET_1E6(cpu)
 *     elif opcode == 0x1E7:
 */
    break;
    case 0x1E7:

    /* "pyboy/core/opcodes.py":6223
 *         return SET_1E6(cpu)
 *     elif opcode == 0x1E7:
 *         return SET_1E7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E8:
 *         return SET_1E8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6222
 *     elif opcode == 0x1E6:
 *         return SET_1E6(cpu)
 *     elif opcode == 0x1E7:             # <<<<<<<<<<<<<<
 *         return SET_1E7(cpu)
 *     elif opcode == 0x1E8:
 */
    break;
    case 0x1E8:

    /* "pyboy/core/opcodes.py":6225
 *         return SET_1E7(cpu)
 *     elif opcode == 0x1E8:
 *         return SET_1E8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1E9:
 *         return SET_1E9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6224
 *     elif opcode == 0x1E7:
 *         return SET_1E7(cpu)
 *     elif opcode == 0x1E8:             # <<<<<<<<<<<<<<
 *         return SET_1E8(cpu)
 *     elif opcode == 0x1E9:
 */
    break;
    case 0x1E9:

    /* "pyboy/core/opcodes.py":6227
 *         return SET_1E8(cpu)
 *     elif opcode == 0x1E9:
 *         return SET_1E9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1EA:
 *         return SET_1EA(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1E9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6226
 *     elif opcode == 0x1E8:
 *         return SET_1E8(cpu)
 *     elif opcode == 0x1E9:             # <<<<<<<<<<<<<<
 *         return SET_1E9(cpu)
 *     elif opcode == 0x1EA:
 */
    break;
    case 0x1EA:

    /* "pyboy/core/opcodes.py":6229
 *         return SET_1E9(cpu)
 *     elif opcode == 0x1EA:
 *         return SET_1EA(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1EB:
 *         return SET_1EB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1EA(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6228
 *     elif opcode == 0x1E9:
 *         return SET_1E9(cpu)
 *     elif opcode == 0x1EA:             # <<<<<<<<<<<<<<
 *         return SET_1EA(cpu)
 *     elif opcode == 0x1EB:
 */
    break;
    case 0x1EB:

    /* "pyboy/core/opcodes.py":6231
 *         return SET_1EA(cpu)
 *     elif opcode == 0x1EB:
 *         return SET_1EB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1EC:
 *         return SET_1EC(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1EB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6230
 *     elif opcode == 0x1EA:
 *         return SET_1EA(cpu)
 *     elif opcode == 0x1EB:             # <<<<<<<<<<<<<<
 *         return SET_1EB(cpu)
 *     elif opcode == 0x1EC:
 */
    break;
    case 0x1EC:

    /* "pyboy/core/opcodes.py":6233
 *         return SET_1EB(cpu)
 *     elif opcode == 0x1EC:
 *         return SET_1EC(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1ED:
 *         return SET_1ED(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1EC(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6232
 *     elif opcode == 0x1EB:
 *         return SET_1EB(cpu)
 *     elif opcode == 0x1EC:             # <<<<<<<<<<<<<<
 *         return SET_1EC(cpu)
 *     elif opcode == 0x1ED:
 */
    break;
    case 0x1ED:

    /* "pyboy/core/opcodes.py":6235
 *         return SET_1EC(cpu)
 *     elif opcode == 0x1ED:
 *         return SET_1ED(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1EE:
 *         return SET_1EE(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1ED(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6234
 *     elif opcode == 0x1EC:
 *         return SET_1EC(cpu)
 *     elif opcode == 0x1ED:             # <<<<<<<<<<<<<<
 *         return SET_1ED(cpu)
 *     elif opcode == 0x1EE:
 */
    break;
    case 0x1EE:

    /* "pyboy/core/opcodes.py":6237
 *         return SET_1ED(cpu)
 *     elif opcode == 0x1EE:
 *         return SET_1EE(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1EF:
 *         return SET_1EF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1EE(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6236
 *     elif opcode == 0x1ED:
 *         return SET_1ED(cpu)
 *     elif opcode == 0x1EE:             # <<<<<<<<<<<<<<
 *         return SET_1EE(cpu)
 *     elif opcode == 0x1EF:
 */
    break;
    case 0x1EF:

    /* "pyboy/core/opcodes.py":6239
 *         return SET_1EE(cpu)
 *     elif opcode == 0x1EF:
 *         return SET_1EF(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F0:
 *         return SET_1F0(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1EF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6238
 *     elif opcode == 0x1EE:
 *         return SET_1EE(cpu)
 *     elif opcode == 0x1EF:             # <<<<<<<<<<<<<<
 *         return SET_1EF(cpu)
 *     elif opcode == 0x1F0:
 */
    break;
    case 0x1F0:

    /* "pyboy/core/opcodes.py":6241
 *         return SET_1EF(cpu)
 *     elif opcode == 0x1F0:
 *         return SET_1F0(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F1:
 *         return SET_1F1(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F0(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6240
 *     elif opcode == 0x1EF:
 *         return SET_1EF(cpu)
 *     elif opcode == 0x1F0:             # <<<<<<<<<<<<<<
 *         return SET_1F0(cpu)
 *     elif opcode == 0x1F1:
 */
    break;
    case 0x1F1:

    /* "pyboy/core/opcodes.py":6243
 *         return SET_1F0(cpu)
 *     elif opcode == 0x1F1:
 *         return SET_1F1(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F2:
 *         return SET_1F2(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F1(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6242
 *     elif opcode == 0x1F0:
 *         return SET_1F0(cpu)
 *     elif opcode == 0x1F1:             # <<<<<<<<<<<<<<
 *         return SET_1F1(cpu)
 *     elif opcode == 0x1F2:
 */
    break;
    case 0x1F2:

    /* "pyboy/core/opcodes.py":6245
 *         return SET_1F1(cpu)
 *     elif opcode == 0x1F2:
 *         return SET_1F2(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F3:
 *         return SET_1F3(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F2(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6244
 *     elif opcode == 0x1F1:
 *         return SET_1F1(cpu)
 *     elif opcode == 0x1F2:             # <<<<<<<<<<<<<<
 *         return SET_1F2(cpu)
 *     elif opcode == 0x1F3:
 */
    break;
    case 0x1F3:

    /* "pyboy/core/opcodes.py":6247
 *         return SET_1F2(cpu)
 *     elif opcode == 0x1F3:
 *         return SET_1F3(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F4:
 *         return SET_1F4(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F3(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6246
 *     elif opcode == 0x1F2:
 *         return SET_1F2(cpu)
 *     elif opcode == 0x1F3:             # <<<<<<<<<<<<<<
 *         return SET_1F3(cpu)
 *     elif opcode == 0x1F4:
 */
    break;
    case 0x1F4:

    /* "pyboy/core/opcodes.py":6249
 *         return SET_1F3(cpu)
 *     elif opcode == 0x1F4:
 *         return SET_1F4(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F5:
 *         return SET_1F5(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F4(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6248
 *     elif opcode == 0x1F3:
 *         return SET_1F3(cpu)
 *     elif opcode == 0x1F4:             # <<<<<<<<<<<<<<
 *         return SET_1F4(cpu)
 *     elif opcode == 0x1F5:
 */
    break;
    case 0x1F5:

    /* "pyboy/core/opcodes.py":6251
 *         return SET_1F4(cpu)
 *     elif opcode == 0x1F5:
 *         return SET_1F5(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F6:
 *         return SET_1F6(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F5(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6250
 *     elif opcode == 0x1F4:
 *         return SET_1F4(cpu)
 *     elif opcode == 0x1F5:             # <<<<<<<<<<<<<<
 *         return SET_1F5(cpu)
 *     elif opcode == 0x1F6:
 */
    break;
    case 0x1F6:

    /* "pyboy/core/opcodes.py":6253
 *         return SET_1F5(cpu)
 *     elif opcode == 0x1F6:
 *         return SET_1F6(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F7:
 *         return SET_1F7(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F6(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6252
 *     elif opcode == 0x1F5:
 *         return SET_1F5(cpu)
 *     elif opcode == 0x1F6:             # <<<<<<<<<<<<<<
 *         return SET_1F6(cpu)
 *     elif opcode == 0x1F7:
 */
    break;
    case 0x1F7:

    /* "pyboy/core/opcodes.py":6255
 *         return SET_1F6(cpu)
 *     elif opcode == 0x1F7:
 *         return SET_1F7(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F8:
 *         return SET_1F8(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F7(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6254
 *     elif opcode == 0x1F6:
 *         return SET_1F6(cpu)
 *     elif opcode == 0x1F7:             # <<<<<<<<<<<<<<
 *         return SET_1F7(cpu)
 *     elif opcode == 0x1F8:
 */
    break;
    case 0x1F8:

    /* "pyboy/core/opcodes.py":6257
 *         return SET_1F7(cpu)
 *     elif opcode == 0x1F8:
 *         return SET_1F8(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1F9:
 *         return SET_1F9(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F8(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6256
 *     elif opcode == 0x1F7:
 *         return SET_1F7(cpu)
 *     elif opcode == 0x1F8:             # <<<<<<<<<<<<<<
 *         return SET_1F8(cpu)
 *     elif opcode == 0x1F9:
 */
    break;
    case 0x1F9:

    /* "pyboy/core/opcodes.py":6259
 *         return SET_1F8(cpu)
 *     elif opcode == 0x1F9:
 *         return SET_1F9(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1FA:
 *         return SET_1FA(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1F9(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6258
 *     elif opcode == 0x1F8:
 *         return SET_1F8(cpu)
 *     elif opcode == 0x1F9:             # <<<<<<<<<<<<<<
 *         return SET_1F9(cpu)
 *     elif opcode == 0x1FA:
 */
    break;
    case 0x1FA:

    /* "pyboy/core/opcodes.py":6261
 *         return SET_1F9(cpu)
 *     elif opcode == 0x1FA:
 *         return SET_1FA(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1FB:
 *         return SET_1FB(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1FA(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6260
 *     elif opcode == 0x1F9:
 *         return SET_1F9(cpu)
 *     elif opcode == 0x1FA:             # <<<<<<<<<<<<<<
 *         return SET_1FA(cpu)
 *     elif opcode == 0x1FB:
 */
    break;
    case 0x1FB:

    /* "pyboy/core/opcodes.py":6263
 *         return SET_1FA(cpu)
 *     elif opcode == 0x1FB:
 *         return SET_1FB(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1FC:
 *         return SET_1FC(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1FB(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6262
 *     elif opcode == 0x1FA:
 *         return SET_1FA(cpu)
 *     elif opcode == 0x1FB:             # <<<<<<<<<<<<<<
 *         return SET_1FB(cpu)
 *     elif opcode == 0x1FC:
 */
    break;
    case 0x1FC:

    /* "pyboy/core/opcodes.py":6265
 *         return SET_1FB(cpu)
 *     elif opcode == 0x1FC:
 *         return SET_1FC(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1FD:
 *         return SET_1FD(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1FC(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6264
 *     elif opcode == 0x1FB:
 *         return SET_1FB(cpu)
 *     elif opcode == 0x1FC:             # <<<<<<<<<<<<<<
 *         return SET_1FC(cpu)
 *     elif opcode == 0x1FD:
 */
    break;
    case 0x1FD:

    /* "pyboy/core/opcodes.py":6267
 *         return SET_1FC(cpu)
 *     elif opcode == 0x1FD:
 *         return SET_1FD(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1FE:
 *         return SET_1FE(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1FD(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6266
 *     elif opcode == 0x1FC:
 *         return SET_1FC(cpu)
 *     elif opcode == 0x1FD:             # <<<<<<<<<<<<<<
 *         return SET_1FD(cpu)
 *     elif opcode == 0x1FE:
 */
    break;
    case 0x1FE:

    /* "pyboy/core/opcodes.py":6269
 *         return SET_1FD(cpu)
 *     elif opcode == 0x1FE:
 *         return SET_1FE(cpu)             # <<<<<<<<<<<<<<
 *     elif opcode == 0x1FF:
 *         return SET_1FF(cpu)
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1FE(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6268
 *     elif opcode == 0x1FD:
 *         return SET_1FD(cpu)
 *     elif opcode == 0x1FE:             # <<<<<<<<<<<<<<
 *         return SET_1FE(cpu)
 *     elif opcode == 0x1FF:
 */
    break;
    case 0x1FF:

    /* "pyboy/core/opcodes.py":6271
 *         return SET_1FE(cpu)
 *     elif opcode == 0x1FF:
 *         return SET_1FF(cpu)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = __pyx_f_5pyboy_4core_7opcodes_SET_1FF(__pyx_v_cpu);
    goto __pyx_L0;

    /* "pyboy/core/opcodes.py":6270
 *     elif opcode == 0x1FE:
 *         return SET_1FE(cpu)
 *     elif opcode == 0x1FF:             # <<<<<<<<<<<<<<
 *         return SET_1FF(cpu)
 * 
 */
    break;
    default: break;
  }

  /* "pyboy/core/opcodes.py":5234
 * 
 * 
 * def execute_opcode(cpu, opcode):             # <<<<<<<<<<<<<<
 *     oplen = OPCODE_LENGTHS[opcode]
 *     v = 0
 */

  /* function exit code */
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_array) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_array___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    __Pyx_TypeName o_type_name;
    o_type_name = __Pyx_PyType_GetName(Py_TYPE(o));
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by " __Pyx_FMT_TYPENAME, o_type_name);
    __Pyx_DECREF_TypeName(o_type_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_array_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_array_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
#if !CYTHON_COMPILING_IN_LIMITED_API

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};
#endif
static PyType_Slot __pyx_type___pyx_array_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_array},
  {Py_sq_length, (void *)__pyx_array___len__},
  {Py_sq_item, (void *)__pyx_sq_item_array},
  {Py_mp_length, (void *)__pyx_array___len__},
  {Py_mp_subscript, (void *)__pyx_array___getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_array},
  {Py_tp_getattro, (void *)__pyx_tp_getattro_array},
  #if defined(Py_bf_getbuffer)
  {Py_bf_getbuffer, (void *)__pyx_array_getbuffer},
  #endif
  {Py_tp_methods, (void *)__pyx_methods_array},
  {Py_tp_getset, (void *)__pyx_getsets_array},
  {Py_tp_new, (void *)__pyx_tp_new_array},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_array_spec = {
  "pyboy.core.opcodes.array",
  sizeof(struct __pyx_array_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_SEQUENCE,
  __pyx_type___pyx_array_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyboy.core.opcodes.""array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_SEQUENCE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_Enum) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_specialmethod___pyx_MemviewEnum___repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_MemviewEnum___repr__(self);
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_MemviewEnum___repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type___pyx_MemviewEnum_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_Enum},
  {Py_tp_repr, (void *)__pyx_MemviewEnum___repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_Enum},
  {Py_tp_clear, (void *)__pyx_tp_clear_Enum},
  {Py_tp_methods, (void *)__pyx_methods_Enum},
  {Py_tp_init, (void *)__pyx_MemviewEnum___init__},
  {Py_tp_new, (void *)__pyx_tp_new_Enum},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_MemviewEnum_spec = {
  "pyboy.core.opcodes.Enum",
  sizeof(struct __pyx_MemviewEnum_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type___pyx_MemviewEnum_slots,
};
#else

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyboy.core.opcodes.""Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_memoryview) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryview___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    __Pyx_TypeName o_type_name;
    o_type_name = __Pyx_PyType_GetName(Py_TYPE(o));
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by " __Pyx_FMT_TYPENAME, o_type_name);
    __Pyx_DECREF_TypeName(o_type_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_memoryview___repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_memoryview___repr__(self);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_memoryview___repr__, METH_NOARGS|METH_COEXIST, 0},
  {"is_c_contig", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_memoryview_is_c_contig, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"is_f_contig", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_memoryview_is_f_contig, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_memoryview_copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"copy_fortran", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_memoryview_copy_fortran, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_memoryview_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_memoryview_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
#if !CYTHON_COMPILING_IN_LIMITED_API

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};
#endif
static PyType_Slot __pyx_type___pyx_memoryview_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_memoryview},
  {Py_tp_repr, (void *)__pyx_memoryview___repr__},
  {Py_sq_length, (void *)__pyx_memoryview___len__},
  {Py_sq_item, (void *)__pyx_sq_item_memoryview},
  {Py_mp_length, (void *)__pyx_memoryview___len__},
  {Py_mp_subscript, (void *)__pyx_memoryview___getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_memoryview},
  {Py_tp_str, (void *)__pyx_memoryview___str__},
  #if defined(Py_bf_getbuffer)
  {Py_bf_getbuffer, (void *)__pyx_memoryview_getbuffer},
  #endif
  {Py_tp_traverse, (void *)__pyx_tp_traverse_memoryview},
  {Py_tp_clear, (void *)__pyx_tp_clear_memoryview},
  {Py_tp_methods, (void *)__pyx_methods_memoryview},
  {Py_tp_getset, (void *)__pyx_getsets_memoryview},
  {Py_tp_new, (void *)__pyx_tp_new_memoryview},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_memoryview_spec = {
  "pyboy.core.opcodes.memoryview",
  sizeof(struct __pyx_memoryview_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type___pyx_memoryview_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyboy.core.opcodes.""memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc__memoryviewslice) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryviewslice___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XCLEAR_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type___pyx_memoryviewslice_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc__memoryviewslice},
  {Py_tp_doc, (void *)PyDoc_STR("Internal class for passing memoryview slices to Python")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse__memoryviewslice},
  {Py_tp_clear, (void *)__pyx_tp_clear__memoryviewslice},
  {Py_tp_methods, (void *)__pyx_methods__memoryviewslice},
  {Py_tp_new, (void *)__pyx_tp_new__memoryviewslice},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_memoryviewslice_spec = {
  "pyboy.core.opcodes._memoryviewslice",
  sizeof(struct __pyx_memoryviewslice_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_SEQUENCE,
  __pyx_type___pyx_memoryviewslice_slots,
};
#else

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyboy.core.opcodes.""_memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_SEQUENCE, /*tp_flags*/
  PyDoc_STR("Internal class for passing memoryview slices to Python"), /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
    {&__pyx_kp_u_ADC_A_A, __pyx_k_ADC_A_A, sizeof(__pyx_k_ADC_A_A), 0, 1, 0, 0},
    {&__pyx_kp_u_ADC_A_B, __pyx_k_ADC_A_B, sizeof(__pyx_k_ADC_A_B), 0, 1, 0, 0},
    {&__pyx_kp_u_ADC_A_C, __pyx_k_ADC_A_C, sizeof(__pyx_k_ADC_A_C), 0, 1, 0, 0},
    {&__pyx_kp_u_ADC_A_D, __pyx_k_ADC_A_D, sizeof(__pyx_k_ADC_A_D), 0, 1, 0, 0},
    {&__pyx_kp_u_ADC_A_E, __pyx_k_ADC_A_E, sizeof(__pyx_k_ADC_A_E), 0, 1, 0, 0},
    {&__pyx_kp_u_ADC_A_H, __pyx_k_ADC_A_H, sizeof(__pyx_k_ADC_A_H), 0, 1, 0, 0},
    {&__pyx_kp_u_ADC_A_HL, __pyx_k_ADC_A_HL, sizeof(__pyx_k_ADC_A_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_ADC_A_L, __pyx_k_ADC_A_L, sizeof(__pyx_k_ADC_A_L), 0, 1, 0, 0},
    {&__pyx_kp_u_ADC_A_d8, __pyx_k_ADC_A_d8, sizeof(__pyx_k_ADC_A_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_A_A, __pyx_k_ADD_A_A, sizeof(__pyx_k_ADD_A_A), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_A_B, __pyx_k_ADD_A_B, sizeof(__pyx_k_ADD_A_B), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_A_C, __pyx_k_ADD_A_C, sizeof(__pyx_k_ADD_A_C), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_A_D, __pyx_k_ADD_A_D, sizeof(__pyx_k_ADD_A_D), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_A_E, __pyx_k_ADD_A_E, sizeof(__pyx_k_ADD_A_E), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_A_H, __pyx_k_ADD_A_H, sizeof(__pyx_k_ADD_A_H), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_A_HL, __pyx_k_ADD_A_HL, sizeof(__pyx_k_ADD_A_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_A_L, __pyx_k_ADD_A_L, sizeof(__pyx_k_ADD_A_L), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_A_d8, __pyx_k_ADD_A_d8, sizeof(__pyx_k_ADD_A_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_HL_BC, __pyx_k_ADD_HL_BC, sizeof(__pyx_k_ADD_HL_BC), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_HL_DE, __pyx_k_ADD_HL_DE, sizeof(__pyx_k_ADD_HL_DE), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_HL_HL, __pyx_k_ADD_HL_HL, sizeof(__pyx_k_ADD_HL_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_HL_SP, __pyx_k_ADD_HL_SP, sizeof(__pyx_k_ADD_HL_SP), 0, 1, 0, 0},
    {&__pyx_kp_u_ADD_SP_r8, __pyx_k_ADD_SP_r8, sizeof(__pyx_k_ADD_SP_r8), 0, 1, 0, 0},
    {&__pyx_kp_u_AND_A, __pyx_k_AND_A, sizeof(__pyx_k_AND_A), 0, 1, 0, 0},
    {&__pyx_kp_u_AND_B, __pyx_k_AND_B, sizeof(__pyx_k_AND_B), 0, 1, 0, 0},
    {&__pyx_kp_u_AND_C, __pyx_k_AND_C, sizeof(__pyx_k_AND_C), 0, 1, 0, 0},
    {&__pyx_kp_u_AND_D, __pyx_k_AND_D, sizeof(__pyx_k_AND_D), 0, 1, 0, 0},
    {&__pyx_kp_u_AND_E, __pyx_k_AND_E, sizeof(__pyx_k_AND_E), 0, 1, 0, 0},
    {&__pyx_kp_u_AND_H, __pyx_k_AND_H, sizeof(__pyx_k_AND_H), 0, 1, 0, 0},
    {&__pyx_kp_u_AND_HL, __pyx_k_AND_HL, sizeof(__pyx_k_AND_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_AND_L, __pyx_k_AND_L, sizeof(__pyx_k_AND_L), 0, 1, 0, 0},
    {&__pyx_kp_u_AND_d8, __pyx_k_AND_d8, sizeof(__pyx_k_AND_d8), 0, 1, 0, 0},
    {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
    {&__pyx_kp_s_All_dimensions_preceding_dimensi, __pyx_k_All_dimensions_preceding_dimensi, sizeof(__pyx_k_All_dimensions_preceding_dimensi), 0, 0, 1, 0},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_u_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 1, 0, 1},
    {&__pyx_kp_u_BIT_0_A, __pyx_k_BIT_0_A, sizeof(__pyx_k_BIT_0_A), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_0_B, __pyx_k_BIT_0_B, sizeof(__pyx_k_BIT_0_B), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_0_C, __pyx_k_BIT_0_C, sizeof(__pyx_k_BIT_0_C), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_0_D, __pyx_k_BIT_0_D, sizeof(__pyx_k_BIT_0_D), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_0_E, __pyx_k_BIT_0_E, sizeof(__pyx_k_BIT_0_E), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_0_H, __pyx_k_BIT_0_H, sizeof(__pyx_k_BIT_0_H), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_0_HL, __pyx_k_BIT_0_HL, sizeof(__pyx_k_BIT_0_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_0_L, __pyx_k_BIT_0_L, sizeof(__pyx_k_BIT_0_L), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_1_A, __pyx_k_BIT_1_A, sizeof(__pyx_k_BIT_1_A), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_1_B, __pyx_k_BIT_1_B, sizeof(__pyx_k_BIT_1_B), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_1_C, __pyx_k_BIT_1_C, sizeof(__pyx_k_BIT_1_C), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_1_D, __pyx_k_BIT_1_D, sizeof(__pyx_k_BIT_1_D), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_1_E, __pyx_k_BIT_1_E, sizeof(__pyx_k_BIT_1_E), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_1_H, __pyx_k_BIT_1_H, sizeof(__pyx_k_BIT_1_H), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_1_HL, __pyx_k_BIT_1_HL, sizeof(__pyx_k_BIT_1_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_1_L, __pyx_k_BIT_1_L, sizeof(__pyx_k_BIT_1_L), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_2_A, __pyx_k_BIT_2_A, sizeof(__pyx_k_BIT_2_A), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_2_B, __pyx_k_BIT_2_B, sizeof(__pyx_k_BIT_2_B), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_2_C, __pyx_k_BIT_2_C, sizeof(__pyx_k_BIT_2_C), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_2_D, __pyx_k_BIT_2_D, sizeof(__pyx_k_BIT_2_D), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_2_E, __pyx_k_BIT_2_E, sizeof(__pyx_k_BIT_2_E), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_2_H, __pyx_k_BIT_2_H, sizeof(__pyx_k_BIT_2_H), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_2_HL, __pyx_k_BIT_2_HL, sizeof(__pyx_k_BIT_2_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_2_L, __pyx_k_BIT_2_L, sizeof(__pyx_k_BIT_2_L), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_3_A, __pyx_k_BIT_3_A, sizeof(__pyx_k_BIT_3_A), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_3_B, __pyx_k_BIT_3_B, sizeof(__pyx_k_BIT_3_B), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_3_C, __pyx_k_BIT_3_C, sizeof(__pyx_k_BIT_3_C), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_3_D, __pyx_k_BIT_3_D, sizeof(__pyx_k_BIT_3_D), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_3_E, __pyx_k_BIT_3_E, sizeof(__pyx_k_BIT_3_E), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_3_H, __pyx_k_BIT_3_H, sizeof(__pyx_k_BIT_3_H), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_3_HL, __pyx_k_BIT_3_HL, sizeof(__pyx_k_BIT_3_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_3_L, __pyx_k_BIT_3_L, sizeof(__pyx_k_BIT_3_L), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_4_A, __pyx_k_BIT_4_A, sizeof(__pyx_k_BIT_4_A), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_4_B, __pyx_k_BIT_4_B, sizeof(__pyx_k_BIT_4_B), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_4_C, __pyx_k_BIT_4_C, sizeof(__pyx_k_BIT_4_C), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_4_D, __pyx_k_BIT_4_D, sizeof(__pyx_k_BIT_4_D), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_4_E, __pyx_k_BIT_4_E, sizeof(__pyx_k_BIT_4_E), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_4_H, __pyx_k_BIT_4_H, sizeof(__pyx_k_BIT_4_H), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_4_HL, __pyx_k_BIT_4_HL, sizeof(__pyx_k_BIT_4_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_4_L, __pyx_k_BIT_4_L, sizeof(__pyx_k_BIT_4_L), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_5_A, __pyx_k_BIT_5_A, sizeof(__pyx_k_BIT_5_A), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_5_B, __pyx_k_BIT_5_B, sizeof(__pyx_k_BIT_5_B), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_5_C, __pyx_k_BIT_5_C, sizeof(__pyx_k_BIT_5_C), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_5_D, __pyx_k_BIT_5_D, sizeof(__pyx_k_BIT_5_D), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_5_E, __pyx_k_BIT_5_E, sizeof(__pyx_k_BIT_5_E), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_5_H, __pyx_k_BIT_5_H, sizeof(__pyx_k_BIT_5_H), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_5_HL, __pyx_k_BIT_5_HL, sizeof(__pyx_k_BIT_5_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_5_L, __pyx_k_BIT_5_L, sizeof(__pyx_k_BIT_5_L), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_6_A, __pyx_k_BIT_6_A, sizeof(__pyx_k_BIT_6_A), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_6_B, __pyx_k_BIT_6_B, sizeof(__pyx_k_BIT_6_B), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_6_C, __pyx_k_BIT_6_C, sizeof(__pyx_k_BIT_6_C), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_6_D, __pyx_k_BIT_6_D, sizeof(__pyx_k_BIT_6_D), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_6_E, __pyx_k_BIT_6_E, sizeof(__pyx_k_BIT_6_E), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_6_H, __pyx_k_BIT_6_H, sizeof(__pyx_k_BIT_6_H), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_6_HL, __pyx_k_BIT_6_HL, sizeof(__pyx_k_BIT_6_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_6_L, __pyx_k_BIT_6_L, sizeof(__pyx_k_BIT_6_L), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_7_A, __pyx_k_BIT_7_A, sizeof(__pyx_k_BIT_7_A), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_7_B, __pyx_k_BIT_7_B, sizeof(__pyx_k_BIT_7_B), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_7_C, __pyx_k_BIT_7_C, sizeof(__pyx_k_BIT_7_C), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_7_D, __pyx_k_BIT_7_D, sizeof(__pyx_k_BIT_7_D), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_7_E, __pyx_k_BIT_7_E, sizeof(__pyx_k_BIT_7_E), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_7_H, __pyx_k_BIT_7_H, sizeof(__pyx_k_BIT_7_H), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_7_HL, __pyx_k_BIT_7_HL, sizeof(__pyx_k_BIT_7_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_BIT_7_L, __pyx_k_BIT_7_L, sizeof(__pyx_k_BIT_7_L), 0, 1, 0, 0},
    {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
    {&__pyx_kp_u_CALL_C_a16, __pyx_k_CALL_C_a16, sizeof(__pyx_k_CALL_C_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_CALL_NC_a16, __pyx_k_CALL_NC_a16, sizeof(__pyx_k_CALL_NC_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_CALL_NZ_a16, __pyx_k_CALL_NZ_a16, sizeof(__pyx_k_CALL_NZ_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_CALL_Z_a16, __pyx_k_CALL_Z_a16, sizeof(__pyx_k_CALL_Z_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_CALL_a16, __pyx_k_CALL_a16, sizeof(__pyx_k_CALL_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_CB_cannot_be_called, __pyx_k_CB_cannot_be_called, sizeof(__pyx_k_CB_cannot_be_called), 0, 1, 0, 0},
    {&__pyx_n_u_CCF, __pyx_k_CCF, sizeof(__pyx_k_CCF), 0, 1, 0, 1},
    {&__pyx_n_u_CPL, __pyx_k_CPL, sizeof(__pyx_k_CPL), 0, 1, 0, 1},
    {&__pyx_n_s_CPU_COMMANDS, __pyx_k_CPU_COMMANDS, sizeof(__pyx_k_CPU_COMMANDS), 0, 0, 1, 1},
    {&__pyx_kp_u_CP_A, __pyx_k_CP_A, sizeof(__pyx_k_CP_A), 0, 1, 0, 0},
    {&__pyx_kp_u_CP_B, __pyx_k_CP_B, sizeof(__pyx_k_CP_B), 0, 1, 0, 0},
    {&__pyx_kp_u_CP_C, __pyx_k_CP_C, sizeof(__pyx_k_CP_C), 0, 1, 0, 0},
    {&__pyx_kp_u_CP_D, __pyx_k_CP_D, sizeof(__pyx_k_CP_D), 0, 1, 0, 0},
    {&__pyx_kp_u_CP_E, __pyx_k_CP_E, sizeof(__pyx_k_CP_E), 0, 1, 0, 0},
    {&__pyx_kp_u_CP_H, __pyx_k_CP_H, sizeof(__pyx_k_CP_H), 0, 1, 0, 0},
    {&__pyx_kp_u_CP_HL, __pyx_k_CP_HL, sizeof(__pyx_k_CP_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_CP_L, __pyx_k_CP_L, sizeof(__pyx_k_CP_L), 0, 1, 0, 0},
    {&__pyx_kp_u_CP_d8, __pyx_k_CP_d8, sizeof(__pyx_k_CP_d8), 0, 1, 0, 0},
    {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
    {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
    {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
    {&__pyx_kp_u_Cannot_index_with_type, __pyx_k_Cannot_index_with_type, sizeof(__pyx_k_Cannot_index_with_type), 0, 1, 0, 0},
    {&__pyx_kp_s_Cannot_transpose_memoryview_with, __pyx_k_Cannot_transpose_memoryview_with, sizeof(__pyx_k_Cannot_transpose_memoryview_with), 0, 0, 1, 0},
    {&__pyx_n_u_DAA, __pyx_k_DAA, sizeof(__pyx_k_DAA), 0, 1, 0, 1},
    {&__pyx_kp_u_DEC_A, __pyx_k_DEC_A, sizeof(__pyx_k_DEC_A), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_B, __pyx_k_DEC_B, sizeof(__pyx_k_DEC_B), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_BC, __pyx_k_DEC_BC, sizeof(__pyx_k_DEC_BC), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_C, __pyx_k_DEC_C, sizeof(__pyx_k_DEC_C), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_D, __pyx_k_DEC_D, sizeof(__pyx_k_DEC_D), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_DE, __pyx_k_DEC_DE, sizeof(__pyx_k_DEC_DE), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_E, __pyx_k_DEC_E, sizeof(__pyx_k_DEC_E), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_H, __pyx_k_DEC_H, sizeof(__pyx_k_DEC_H), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_HL, __pyx_k_DEC_HL, sizeof(__pyx_k_DEC_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_HL_2, __pyx_k_DEC_HL_2, sizeof(__pyx_k_DEC_HL_2), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_L, __pyx_k_DEC_L, sizeof(__pyx_k_DEC_L), 0, 1, 0, 0},
    {&__pyx_kp_u_DEC_SP, __pyx_k_DEC_SP, sizeof(__pyx_k_DEC_SP), 0, 1, 0, 0},
    {&__pyx_n_u_DI, __pyx_k_DI, sizeof(__pyx_k_DI), 0, 1, 0, 1},
    {&__pyx_kp_s_Dimension_d_is_not_direct, __pyx_k_Dimension_d_is_not_direct, sizeof(__pyx_k_Dimension_d_is_not_direct), 0, 0, 1, 0},
    {&__pyx_n_u_EI, __pyx_k_EI, sizeof(__pyx_k_EI), 0, 1, 0, 1},
    {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
    {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
    {&__pyx_n_s_FLAGC, __pyx_k_FLAGC, sizeof(__pyx_k_FLAGC), 0, 0, 1, 1},
    {&__pyx_n_s_FLAGH, __pyx_k_FLAGH, sizeof(__pyx_k_FLAGH), 0, 0, 1, 1},
    {&__pyx_n_s_FLAGN, __pyx_k_FLAGN, sizeof(__pyx_k_FLAGN), 0, 0, 1, 1},
    {&__pyx_n_s_FLAGZ, __pyx_k_FLAGZ, sizeof(__pyx_k_FLAGZ), 0, 0, 1, 1},
    {&__pyx_n_u_HALT, __pyx_k_HALT, sizeof(__pyx_k_HALT), 0, 1, 0, 1},
    {&__pyx_kp_u_INC_A, __pyx_k_INC_A, sizeof(__pyx_k_INC_A), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_B, __pyx_k_INC_B, sizeof(__pyx_k_INC_B), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_BC, __pyx_k_INC_BC, sizeof(__pyx_k_INC_BC), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_C, __pyx_k_INC_C, sizeof(__pyx_k_INC_C), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_D, __pyx_k_INC_D, sizeof(__pyx_k_INC_D), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_DE, __pyx_k_INC_DE, sizeof(__pyx_k_INC_DE), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_E, __pyx_k_INC_E, sizeof(__pyx_k_INC_E), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_H, __pyx_k_INC_H, sizeof(__pyx_k_INC_H), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_HL, __pyx_k_INC_HL, sizeof(__pyx_k_INC_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_HL_2, __pyx_k_INC_HL_2, sizeof(__pyx_k_INC_HL_2), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_L, __pyx_k_INC_L, sizeof(__pyx_k_INC_L), 0, 1, 0, 0},
    {&__pyx_kp_u_INC_SP, __pyx_k_INC_SP, sizeof(__pyx_k_INC_SP), 0, 1, 0, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
    {&__pyx_kp_s_Index_out_of_bounds_axis_d, __pyx_k_Index_out_of_bounds_axis_d, sizeof(__pyx_k_Index_out_of_bounds_axis_d), 0, 0, 1, 0},
    {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
    {&__pyx_kp_u_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_shape_in_axis, __pyx_k_Invalid_shape_in_axis, sizeof(__pyx_k_Invalid_shape_in_axis), 0, 1, 0, 0},
    {&__pyx_kp_u_JP_C_a16, __pyx_k_JP_C_a16, sizeof(__pyx_k_JP_C_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_JP_HL, __pyx_k_JP_HL, sizeof(__pyx_k_JP_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_JP_NC_a16, __pyx_k_JP_NC_a16, sizeof(__pyx_k_JP_NC_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_JP_NZ_a16, __pyx_k_JP_NZ_a16, sizeof(__pyx_k_JP_NZ_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_JP_Z_a16, __pyx_k_JP_Z_a16, sizeof(__pyx_k_JP_Z_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_JP_a16, __pyx_k_JP_a16, sizeof(__pyx_k_JP_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_JR_C_r8, __pyx_k_JR_C_r8, sizeof(__pyx_k_JR_C_r8), 0, 1, 0, 0},
    {&__pyx_kp_u_JR_NC_r8, __pyx_k_JR_NC_r8, sizeof(__pyx_k_JR_NC_r8), 0, 1, 0, 0},
    {&__pyx_kp_u_JR_NZ_r8, __pyx_k_JR_NZ_r8, sizeof(__pyx_k_JR_NZ_r8), 0, 1, 0, 0},
    {&__pyx_kp_u_JR_Z_r8, __pyx_k_JR_Z_r8, sizeof(__pyx_k_JR_Z_r8), 0, 1, 0, 0},
    {&__pyx_kp_u_JR_r8, __pyx_k_JR_r8, sizeof(__pyx_k_JR_r8), 0, 1, 0, 0},
    {&__pyx_kp_u_LDH_A_a8, __pyx_k_LDH_A_a8, sizeof(__pyx_k_LDH_A_a8), 0, 1, 0, 0},
    {&__pyx_kp_u_LDH_a8_A, __pyx_k_LDH_a8_A, sizeof(__pyx_k_LDH_a8_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_A, __pyx_k_LD_A_A, sizeof(__pyx_k_LD_A_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_B, __pyx_k_LD_A_B, sizeof(__pyx_k_LD_A_B), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_BC, __pyx_k_LD_A_BC, sizeof(__pyx_k_LD_A_BC), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_C, __pyx_k_LD_A_C, sizeof(__pyx_k_LD_A_C), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_C_2, __pyx_k_LD_A_C_2, sizeof(__pyx_k_LD_A_C_2), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_D, __pyx_k_LD_A_D, sizeof(__pyx_k_LD_A_D), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_DE, __pyx_k_LD_A_DE, sizeof(__pyx_k_LD_A_DE), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_E, __pyx_k_LD_A_E, sizeof(__pyx_k_LD_A_E), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_H, __pyx_k_LD_A_H, sizeof(__pyx_k_LD_A_H), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_HL, __pyx_k_LD_A_HL, sizeof(__pyx_k_LD_A_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_HL_2, __pyx_k_LD_A_HL_2, sizeof(__pyx_k_LD_A_HL_2), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_HL_3, __pyx_k_LD_A_HL_3, sizeof(__pyx_k_LD_A_HL_3), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_L, __pyx_k_LD_A_L, sizeof(__pyx_k_LD_A_L), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_a16, __pyx_k_LD_A_a16, sizeof(__pyx_k_LD_A_a16), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_A_d8, __pyx_k_LD_A_d8, sizeof(__pyx_k_LD_A_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_BC_A, __pyx_k_LD_BC_A, sizeof(__pyx_k_LD_BC_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_BC_d16, __pyx_k_LD_BC_d16, sizeof(__pyx_k_LD_BC_d16), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_B_A, __pyx_k_LD_B_A, sizeof(__pyx_k_LD_B_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_B_B, __pyx_k_LD_B_B, sizeof(__pyx_k_LD_B_B), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_B_C, __pyx_k_LD_B_C, sizeof(__pyx_k_LD_B_C), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_B_D, __pyx_k_LD_B_D, sizeof(__pyx_k_LD_B_D), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_B_E, __pyx_k_LD_B_E, sizeof(__pyx_k_LD_B_E), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_B_H, __pyx_k_LD_B_H, sizeof(__pyx_k_LD_B_H), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_B_HL, __pyx_k_LD_B_HL, sizeof(__pyx_k_LD_B_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_B_L, __pyx_k_LD_B_L, sizeof(__pyx_k_LD_B_L), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_B_d8, __pyx_k_LD_B_d8, sizeof(__pyx_k_LD_B_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_A, __pyx_k_LD_C_A, sizeof(__pyx_k_LD_C_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_A_2, __pyx_k_LD_C_A_2, sizeof(__pyx_k_LD_C_A_2), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_B, __pyx_k_LD_C_B, sizeof(__pyx_k_LD_C_B), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_C, __pyx_k_LD_C_C, sizeof(__pyx_k_LD_C_C), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_D, __pyx_k_LD_C_D, sizeof(__pyx_k_LD_C_D), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_E, __pyx_k_LD_C_E, sizeof(__pyx_k_LD_C_E), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_H, __pyx_k_LD_C_H, sizeof(__pyx_k_LD_C_H), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_HL, __pyx_k_LD_C_HL, sizeof(__pyx_k_LD_C_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_L, __pyx_k_LD_C_L, sizeof(__pyx_k_LD_C_L), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_C_d8, __pyx_k_LD_C_d8, sizeof(__pyx_k_LD_C_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_DE_A, __pyx_k_LD_DE_A, sizeof(__pyx_k_LD_DE_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_DE_d16, __pyx_k_LD_DE_d16, sizeof(__pyx_k_LD_DE_d16), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_D_A, __pyx_k_LD_D_A, sizeof(__pyx_k_LD_D_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_D_B, __pyx_k_LD_D_B, sizeof(__pyx_k_LD_D_B), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_D_C, __pyx_k_LD_D_C, sizeof(__pyx_k_LD_D_C), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_D_D, __pyx_k_LD_D_D, sizeof(__pyx_k_LD_D_D), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_D_E, __pyx_k_LD_D_E, sizeof(__pyx_k_LD_D_E), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_D_H, __pyx_k_LD_D_H, sizeof(__pyx_k_LD_D_H), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_D_HL, __pyx_k_LD_D_HL, sizeof(__pyx_k_LD_D_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_D_L, __pyx_k_LD_D_L, sizeof(__pyx_k_LD_D_L), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_D_d8, __pyx_k_LD_D_d8, sizeof(__pyx_k_LD_D_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_E_A, __pyx_k_LD_E_A, sizeof(__pyx_k_LD_E_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_E_B, __pyx_k_LD_E_B, sizeof(__pyx_k_LD_E_B), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_E_C, __pyx_k_LD_E_C, sizeof(__pyx_k_LD_E_C), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_E_D, __pyx_k_LD_E_D, sizeof(__pyx_k_LD_E_D), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_E_E, __pyx_k_LD_E_E, sizeof(__pyx_k_LD_E_E), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_E_H, __pyx_k_LD_E_H, sizeof(__pyx_k_LD_E_H), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_E_HL, __pyx_k_LD_E_HL, sizeof(__pyx_k_LD_E_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_E_L, __pyx_k_LD_E_L, sizeof(__pyx_k_LD_E_L), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_E_d8, __pyx_k_LD_E_d8, sizeof(__pyx_k_LD_E_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_A, __pyx_k_LD_HL_A, sizeof(__pyx_k_LD_HL_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_A_2, __pyx_k_LD_HL_A_2, sizeof(__pyx_k_LD_HL_A_2), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_A_3, __pyx_k_LD_HL_A_3, sizeof(__pyx_k_LD_HL_A_3), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_B, __pyx_k_LD_HL_B, sizeof(__pyx_k_LD_HL_B), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_C, __pyx_k_LD_HL_C, sizeof(__pyx_k_LD_HL_C), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_D, __pyx_k_LD_HL_D, sizeof(__pyx_k_LD_HL_D), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_E, __pyx_k_LD_HL_E, sizeof(__pyx_k_LD_HL_E), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_H, __pyx_k_LD_HL_H, sizeof(__pyx_k_LD_HL_H), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_L, __pyx_k_LD_HL_L, sizeof(__pyx_k_LD_HL_L), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_SP_r8, __pyx_k_LD_HL_SP_r8, sizeof(__pyx_k_LD_HL_SP_r8), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_d16, __pyx_k_LD_HL_d16, sizeof(__pyx_k_LD_HL_d16), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_HL_d8, __pyx_k_LD_HL_d8, sizeof(__pyx_k_LD_HL_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_H_A, __pyx_k_LD_H_A, sizeof(__pyx_k_LD_H_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_H_B, __pyx_k_LD_H_B, sizeof(__pyx_k_LD_H_B), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_H_C, __pyx_k_LD_H_C, sizeof(__pyx_k_LD_H_C), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_H_D, __pyx_k_LD_H_D, sizeof(__pyx_k_LD_H_D), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_H_E, __pyx_k_LD_H_E, sizeof(__pyx_k_LD_H_E), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_H_H, __pyx_k_LD_H_H, sizeof(__pyx_k_LD_H_H), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_H_HL, __pyx_k_LD_H_HL, sizeof(__pyx_k_LD_H_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_H_L, __pyx_k_LD_H_L, sizeof(__pyx_k_LD_H_L), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_H_d8, __pyx_k_LD_H_d8, sizeof(__pyx_k_LD_H_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_L_A, __pyx_k_LD_L_A, sizeof(__pyx_k_LD_L_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_L_B, __pyx_k_LD_L_B, sizeof(__pyx_k_LD_L_B), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_L_C, __pyx_k_LD_L_C, sizeof(__pyx_k_LD_L_C), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_L_D, __pyx_k_LD_L_D, sizeof(__pyx_k_LD_L_D), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_L_E, __pyx_k_LD_L_E, sizeof(__pyx_k_LD_L_E), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_L_H, __pyx_k_LD_L_H, sizeof(__pyx_k_LD_L_H), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_L_HL, __pyx_k_LD_L_HL, sizeof(__pyx_k_LD_L_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_L_L, __pyx_k_LD_L_L, sizeof(__pyx_k_LD_L_L), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_L_d8, __pyx_k_LD_L_d8, sizeof(__pyx_k_LD_L_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_SP_HL, __pyx_k_LD_SP_HL, sizeof(__pyx_k_LD_SP_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_SP_d16, __pyx_k_LD_SP_d16, sizeof(__pyx_k_LD_SP_d16), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_a16_A, __pyx_k_LD_a16_A, sizeof(__pyx_k_LD_a16_A), 0, 1, 0, 0},
    {&__pyx_kp_u_LD_a16_SP, __pyx_k_LD_a16_SP, sizeof(__pyx_k_LD_a16_SP), 0, 1, 0, 0},
    {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
    {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
    {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
    {&__pyx_n_u_NOP, __pyx_k_NOP, sizeof(__pyx_k_NOP), 0, 1, 0, 1},
    {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
    {&__pyx_n_s_OPCODE_LENGTHS, __pyx_k_OPCODE_LENGTHS, sizeof(__pyx_k_OPCODE_LENGTHS), 0, 0, 1, 1},
    {&__pyx_kp_u_OR_A, __pyx_k_OR_A, sizeof(__pyx_k_OR_A), 0, 1, 0, 0},
    {&__pyx_kp_u_OR_B, __pyx_k_OR_B, sizeof(__pyx_k_OR_B), 0, 1, 0, 0},
    {&__pyx_kp_u_OR_C, __pyx_k_OR_C, sizeof(__pyx_k_OR_C), 0, 1, 0, 0},
    {&__pyx_kp_u_OR_D, __pyx_k_OR_D, sizeof(__pyx_k_OR_D), 0, 1, 0, 0},
    {&__pyx_kp_u_OR_E, __pyx_k_OR_E, sizeof(__pyx_k_OR_E), 0, 1, 0, 0},
    {&__pyx_kp_u_OR_H, __pyx_k_OR_H, sizeof(__pyx_k_OR_H), 0, 1, 0, 0},
    {&__pyx_kp_u_OR_HL, __pyx_k_OR_HL, sizeof(__pyx_k_OR_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_OR_L, __pyx_k_OR_L, sizeof(__pyx_k_OR_L), 0, 1, 0, 0},
    {&__pyx_kp_u_OR_d8, __pyx_k_OR_d8, sizeof(__pyx_k_OR_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 1, 0, 0},
    {&__pyx_n_s_OverflowError, __pyx_k_OverflowError, sizeof(__pyx_k_OverflowError), 0, 0, 1, 1},
    {&__pyx_kp_u_POP_AF, __pyx_k_POP_AF, sizeof(__pyx_k_POP_AF), 0, 1, 0, 0},
    {&__pyx_kp_u_POP_BC, __pyx_k_POP_BC, sizeof(__pyx_k_POP_BC), 0, 1, 0, 0},
    {&__pyx_kp_u_POP_DE, __pyx_k_POP_DE, sizeof(__pyx_k_POP_DE), 0, 1, 0, 0},
    {&__pyx_kp_u_POP_HL, __pyx_k_POP_HL, sizeof(__pyx_k_POP_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_PREFIX_CB, __pyx_k_PREFIX_CB, sizeof(__pyx_k_PREFIX_CB), 0, 1, 0, 0},
    {&__pyx_kp_u_PUSH_AF, __pyx_k_PUSH_AF, sizeof(__pyx_k_PUSH_AF), 0, 1, 0, 0},
    {&__pyx_kp_u_PUSH_BC, __pyx_k_PUSH_BC, sizeof(__pyx_k_PUSH_BC), 0, 1, 0, 0},
    {&__pyx_kp_u_PUSH_DE, __pyx_k_PUSH_DE, sizeof(__pyx_k_PUSH_DE), 0, 1, 0, 0},
    {&__pyx_kp_u_PUSH_HL, __pyx_k_PUSH_HL, sizeof(__pyx_k_PUSH_HL), 0, 1, 0, 0},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_kp_u_RES_0_A, __pyx_k_RES_0_A, sizeof(__pyx_k_RES_0_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_0_B, __pyx_k_RES_0_B, sizeof(__pyx_k_RES_0_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_0_C, __pyx_k_RES_0_C, sizeof(__pyx_k_RES_0_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_0_D, __pyx_k_RES_0_D, sizeof(__pyx_k_RES_0_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_0_E, __pyx_k_RES_0_E, sizeof(__pyx_k_RES_0_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_0_H, __pyx_k_RES_0_H, sizeof(__pyx_k_RES_0_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_0_HL, __pyx_k_RES_0_HL, sizeof(__pyx_k_RES_0_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_0_L, __pyx_k_RES_0_L, sizeof(__pyx_k_RES_0_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_1_A, __pyx_k_RES_1_A, sizeof(__pyx_k_RES_1_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_1_B, __pyx_k_RES_1_B, sizeof(__pyx_k_RES_1_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_1_C, __pyx_k_RES_1_C, sizeof(__pyx_k_RES_1_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_1_D, __pyx_k_RES_1_D, sizeof(__pyx_k_RES_1_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_1_E, __pyx_k_RES_1_E, sizeof(__pyx_k_RES_1_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_1_H, __pyx_k_RES_1_H, sizeof(__pyx_k_RES_1_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_1_HL, __pyx_k_RES_1_HL, sizeof(__pyx_k_RES_1_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_1_L, __pyx_k_RES_1_L, sizeof(__pyx_k_RES_1_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_2_A, __pyx_k_RES_2_A, sizeof(__pyx_k_RES_2_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_2_B, __pyx_k_RES_2_B, sizeof(__pyx_k_RES_2_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_2_C, __pyx_k_RES_2_C, sizeof(__pyx_k_RES_2_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_2_D, __pyx_k_RES_2_D, sizeof(__pyx_k_RES_2_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_2_E, __pyx_k_RES_2_E, sizeof(__pyx_k_RES_2_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_2_H, __pyx_k_RES_2_H, sizeof(__pyx_k_RES_2_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_2_HL, __pyx_k_RES_2_HL, sizeof(__pyx_k_RES_2_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_2_L, __pyx_k_RES_2_L, sizeof(__pyx_k_RES_2_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_3_A, __pyx_k_RES_3_A, sizeof(__pyx_k_RES_3_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_3_B, __pyx_k_RES_3_B, sizeof(__pyx_k_RES_3_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_3_C, __pyx_k_RES_3_C, sizeof(__pyx_k_RES_3_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_3_D, __pyx_k_RES_3_D, sizeof(__pyx_k_RES_3_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_3_E, __pyx_k_RES_3_E, sizeof(__pyx_k_RES_3_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_3_H, __pyx_k_RES_3_H, sizeof(__pyx_k_RES_3_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_3_HL, __pyx_k_RES_3_HL, sizeof(__pyx_k_RES_3_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_3_L, __pyx_k_RES_3_L, sizeof(__pyx_k_RES_3_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_4_A, __pyx_k_RES_4_A, sizeof(__pyx_k_RES_4_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_4_B, __pyx_k_RES_4_B, sizeof(__pyx_k_RES_4_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_4_C, __pyx_k_RES_4_C, sizeof(__pyx_k_RES_4_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_4_D, __pyx_k_RES_4_D, sizeof(__pyx_k_RES_4_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_4_E, __pyx_k_RES_4_E, sizeof(__pyx_k_RES_4_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_4_H, __pyx_k_RES_4_H, sizeof(__pyx_k_RES_4_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_4_HL, __pyx_k_RES_4_HL, sizeof(__pyx_k_RES_4_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_4_L, __pyx_k_RES_4_L, sizeof(__pyx_k_RES_4_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_5_A, __pyx_k_RES_5_A, sizeof(__pyx_k_RES_5_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_5_B, __pyx_k_RES_5_B, sizeof(__pyx_k_RES_5_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_5_C, __pyx_k_RES_5_C, sizeof(__pyx_k_RES_5_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_5_D, __pyx_k_RES_5_D, sizeof(__pyx_k_RES_5_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_5_E, __pyx_k_RES_5_E, sizeof(__pyx_k_RES_5_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_5_H, __pyx_k_RES_5_H, sizeof(__pyx_k_RES_5_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_5_HL, __pyx_k_RES_5_HL, sizeof(__pyx_k_RES_5_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_5_L, __pyx_k_RES_5_L, sizeof(__pyx_k_RES_5_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_6_A, __pyx_k_RES_6_A, sizeof(__pyx_k_RES_6_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_6_B, __pyx_k_RES_6_B, sizeof(__pyx_k_RES_6_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_6_C, __pyx_k_RES_6_C, sizeof(__pyx_k_RES_6_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_6_D, __pyx_k_RES_6_D, sizeof(__pyx_k_RES_6_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_6_E, __pyx_k_RES_6_E, sizeof(__pyx_k_RES_6_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_6_H, __pyx_k_RES_6_H, sizeof(__pyx_k_RES_6_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_6_HL, __pyx_k_RES_6_HL, sizeof(__pyx_k_RES_6_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_6_L, __pyx_k_RES_6_L, sizeof(__pyx_k_RES_6_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_7_A, __pyx_k_RES_7_A, sizeof(__pyx_k_RES_7_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_7_B, __pyx_k_RES_7_B, sizeof(__pyx_k_RES_7_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_7_C, __pyx_k_RES_7_C, sizeof(__pyx_k_RES_7_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_7_D, __pyx_k_RES_7_D, sizeof(__pyx_k_RES_7_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_7_E, __pyx_k_RES_7_E, sizeof(__pyx_k_RES_7_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_7_H, __pyx_k_RES_7_H, sizeof(__pyx_k_RES_7_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_7_HL, __pyx_k_RES_7_HL, sizeof(__pyx_k_RES_7_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RES_7_L, __pyx_k_RES_7_L, sizeof(__pyx_k_RES_7_L), 0, 1, 0, 0},
    {&__pyx_n_u_RET, __pyx_k_RET, sizeof(__pyx_k_RET), 0, 1, 0, 1},
    {&__pyx_n_u_RETI, __pyx_k_RETI, sizeof(__pyx_k_RETI), 0, 1, 0, 1},
    {&__pyx_kp_u_RET_C, __pyx_k_RET_C, sizeof(__pyx_k_RET_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RET_NC, __pyx_k_RET_NC, sizeof(__pyx_k_RET_NC), 0, 1, 0, 0},
    {&__pyx_kp_u_RET_NZ, __pyx_k_RET_NZ, sizeof(__pyx_k_RET_NZ), 0, 1, 0, 0},
    {&__pyx_kp_u_RET_Z, __pyx_k_RET_Z, sizeof(__pyx_k_RET_Z), 0, 1, 0, 0},
    {&__pyx_n_u_RLA, __pyx_k_RLA, sizeof(__pyx_k_RLA), 0, 1, 0, 1},
    {&__pyx_n_u_RLCA, __pyx_k_RLCA, sizeof(__pyx_k_RLCA), 0, 1, 0, 1},
    {&__pyx_kp_u_RLC_A, __pyx_k_RLC_A, sizeof(__pyx_k_RLC_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RLC_B, __pyx_k_RLC_B, sizeof(__pyx_k_RLC_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RLC_C, __pyx_k_RLC_C, sizeof(__pyx_k_RLC_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RLC_D, __pyx_k_RLC_D, sizeof(__pyx_k_RLC_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RLC_E, __pyx_k_RLC_E, sizeof(__pyx_k_RLC_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RLC_H, __pyx_k_RLC_H, sizeof(__pyx_k_RLC_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RLC_HL, __pyx_k_RLC_HL, sizeof(__pyx_k_RLC_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RLC_L, __pyx_k_RLC_L, sizeof(__pyx_k_RLC_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RL_A, __pyx_k_RL_A, sizeof(__pyx_k_RL_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RL_B, __pyx_k_RL_B, sizeof(__pyx_k_RL_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RL_C, __pyx_k_RL_C, sizeof(__pyx_k_RL_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RL_D, __pyx_k_RL_D, sizeof(__pyx_k_RL_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RL_E, __pyx_k_RL_E, sizeof(__pyx_k_RL_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RL_H, __pyx_k_RL_H, sizeof(__pyx_k_RL_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RL_HL, __pyx_k_RL_HL, sizeof(__pyx_k_RL_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RL_L, __pyx_k_RL_L, sizeof(__pyx_k_RL_L), 0, 1, 0, 0},
    {&__pyx_n_u_RRA, __pyx_k_RRA, sizeof(__pyx_k_RRA), 0, 1, 0, 1},
    {&__pyx_n_u_RRCA, __pyx_k_RRCA, sizeof(__pyx_k_RRCA), 0, 1, 0, 1},
    {&__pyx_kp_u_RRC_A, __pyx_k_RRC_A, sizeof(__pyx_k_RRC_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RRC_B, __pyx_k_RRC_B, sizeof(__pyx_k_RRC_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RRC_C, __pyx_k_RRC_C, sizeof(__pyx_k_RRC_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RRC_D, __pyx_k_RRC_D, sizeof(__pyx_k_RRC_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RRC_E, __pyx_k_RRC_E, sizeof(__pyx_k_RRC_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RRC_H, __pyx_k_RRC_H, sizeof(__pyx_k_RRC_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RRC_HL, __pyx_k_RRC_HL, sizeof(__pyx_k_RRC_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RRC_L, __pyx_k_RRC_L, sizeof(__pyx_k_RRC_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RR_A, __pyx_k_RR_A, sizeof(__pyx_k_RR_A), 0, 1, 0, 0},
    {&__pyx_kp_u_RR_B, __pyx_k_RR_B, sizeof(__pyx_k_RR_B), 0, 1, 0, 0},
    {&__pyx_kp_u_RR_C, __pyx_k_RR_C, sizeof(__pyx_k_RR_C), 0, 1, 0, 0},
    {&__pyx_kp_u_RR_D, __pyx_k_RR_D, sizeof(__pyx_k_RR_D), 0, 1, 0, 0},
    {&__pyx_kp_u_RR_E, __pyx_k_RR_E, sizeof(__pyx_k_RR_E), 0, 1, 0, 0},
    {&__pyx_kp_u_RR_H, __pyx_k_RR_H, sizeof(__pyx_k_RR_H), 0, 1, 0, 0},
    {&__pyx_kp_u_RR_HL, __pyx_k_RR_HL, sizeof(__pyx_k_RR_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_RR_L, __pyx_k_RR_L, sizeof(__pyx_k_RR_L), 0, 1, 0, 0},
    {&__pyx_kp_u_RST_00H, __pyx_k_RST_00H, sizeof(__pyx_k_RST_00H), 0, 1, 0, 0},
    {&__pyx_kp_u_RST_08H, __pyx_k_RST_08H, sizeof(__pyx_k_RST_08H), 0, 1, 0, 0},
    {&__pyx_kp_u_RST_10H, __pyx_k_RST_10H, sizeof(__pyx_k_RST_10H), 0, 1, 0, 0},
    {&__pyx_kp_u_RST_18H, __pyx_k_RST_18H, sizeof(__pyx_k_RST_18H), 0, 1, 0, 0},
    {&__pyx_kp_u_RST_20H, __pyx_k_RST_20H, sizeof(__pyx_k_RST_20H), 0, 1, 0, 0},
    {&__pyx_kp_u_RST_28H, __pyx_k_RST_28H, sizeof(__pyx_k_RST_28H), 0, 1, 0, 0},
    {&__pyx_kp_u_RST_30H, __pyx_k_RST_30H, sizeof(__pyx_k_RST_30H), 0, 1, 0, 0},
    {&__pyx_kp_u_RST_38H, __pyx_k_RST_38H, sizeof(__pyx_k_RST_38H), 0, 1, 0, 0},
    {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
    {&__pyx_kp_u_SBC_A_A, __pyx_k_SBC_A_A, sizeof(__pyx_k_SBC_A_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SBC_A_B, __pyx_k_SBC_A_B, sizeof(__pyx_k_SBC_A_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SBC_A_C, __pyx_k_SBC_A_C, sizeof(__pyx_k_SBC_A_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SBC_A_D, __pyx_k_SBC_A_D, sizeof(__pyx_k_SBC_A_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SBC_A_E, __pyx_k_SBC_A_E, sizeof(__pyx_k_SBC_A_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SBC_A_H, __pyx_k_SBC_A_H, sizeof(__pyx_k_SBC_A_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SBC_A_HL, __pyx_k_SBC_A_HL, sizeof(__pyx_k_SBC_A_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SBC_A_L, __pyx_k_SBC_A_L, sizeof(__pyx_k_SBC_A_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SBC_A_d8, __pyx_k_SBC_A_d8, sizeof(__pyx_k_SBC_A_d8), 0, 1, 0, 0},
    {&__pyx_n_u_SCF, __pyx_k_SCF, sizeof(__pyx_k_SCF), 0, 1, 0, 1},
    {&__pyx_kp_u_SET_0_A, __pyx_k_SET_0_A, sizeof(__pyx_k_SET_0_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_0_B, __pyx_k_SET_0_B, sizeof(__pyx_k_SET_0_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_0_C, __pyx_k_SET_0_C, sizeof(__pyx_k_SET_0_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_0_D, __pyx_k_SET_0_D, sizeof(__pyx_k_SET_0_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_0_E, __pyx_k_SET_0_E, sizeof(__pyx_k_SET_0_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_0_H, __pyx_k_SET_0_H, sizeof(__pyx_k_SET_0_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_0_HL, __pyx_k_SET_0_HL, sizeof(__pyx_k_SET_0_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_0_L, __pyx_k_SET_0_L, sizeof(__pyx_k_SET_0_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_1_A, __pyx_k_SET_1_A, sizeof(__pyx_k_SET_1_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_1_B, __pyx_k_SET_1_B, sizeof(__pyx_k_SET_1_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_1_C, __pyx_k_SET_1_C, sizeof(__pyx_k_SET_1_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_1_D, __pyx_k_SET_1_D, sizeof(__pyx_k_SET_1_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_1_E, __pyx_k_SET_1_E, sizeof(__pyx_k_SET_1_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_1_H, __pyx_k_SET_1_H, sizeof(__pyx_k_SET_1_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_1_HL, __pyx_k_SET_1_HL, sizeof(__pyx_k_SET_1_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_1_L, __pyx_k_SET_1_L, sizeof(__pyx_k_SET_1_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_2_A, __pyx_k_SET_2_A, sizeof(__pyx_k_SET_2_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_2_B, __pyx_k_SET_2_B, sizeof(__pyx_k_SET_2_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_2_C, __pyx_k_SET_2_C, sizeof(__pyx_k_SET_2_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_2_D, __pyx_k_SET_2_D, sizeof(__pyx_k_SET_2_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_2_E, __pyx_k_SET_2_E, sizeof(__pyx_k_SET_2_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_2_H, __pyx_k_SET_2_H, sizeof(__pyx_k_SET_2_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_2_HL, __pyx_k_SET_2_HL, sizeof(__pyx_k_SET_2_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_2_L, __pyx_k_SET_2_L, sizeof(__pyx_k_SET_2_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_3_A, __pyx_k_SET_3_A, sizeof(__pyx_k_SET_3_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_3_B, __pyx_k_SET_3_B, sizeof(__pyx_k_SET_3_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_3_C, __pyx_k_SET_3_C, sizeof(__pyx_k_SET_3_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_3_D, __pyx_k_SET_3_D, sizeof(__pyx_k_SET_3_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_3_E, __pyx_k_SET_3_E, sizeof(__pyx_k_SET_3_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_3_H, __pyx_k_SET_3_H, sizeof(__pyx_k_SET_3_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_3_HL, __pyx_k_SET_3_HL, sizeof(__pyx_k_SET_3_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_3_L, __pyx_k_SET_3_L, sizeof(__pyx_k_SET_3_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_4_A, __pyx_k_SET_4_A, sizeof(__pyx_k_SET_4_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_4_B, __pyx_k_SET_4_B, sizeof(__pyx_k_SET_4_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_4_C, __pyx_k_SET_4_C, sizeof(__pyx_k_SET_4_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_4_D, __pyx_k_SET_4_D, sizeof(__pyx_k_SET_4_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_4_E, __pyx_k_SET_4_E, sizeof(__pyx_k_SET_4_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_4_H, __pyx_k_SET_4_H, sizeof(__pyx_k_SET_4_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_4_HL, __pyx_k_SET_4_HL, sizeof(__pyx_k_SET_4_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_4_L, __pyx_k_SET_4_L, sizeof(__pyx_k_SET_4_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_5_A, __pyx_k_SET_5_A, sizeof(__pyx_k_SET_5_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_5_B, __pyx_k_SET_5_B, sizeof(__pyx_k_SET_5_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_5_C, __pyx_k_SET_5_C, sizeof(__pyx_k_SET_5_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_5_D, __pyx_k_SET_5_D, sizeof(__pyx_k_SET_5_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_5_E, __pyx_k_SET_5_E, sizeof(__pyx_k_SET_5_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_5_H, __pyx_k_SET_5_H, sizeof(__pyx_k_SET_5_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_5_HL, __pyx_k_SET_5_HL, sizeof(__pyx_k_SET_5_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_5_L, __pyx_k_SET_5_L, sizeof(__pyx_k_SET_5_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_6_A, __pyx_k_SET_6_A, sizeof(__pyx_k_SET_6_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_6_B, __pyx_k_SET_6_B, sizeof(__pyx_k_SET_6_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_6_C, __pyx_k_SET_6_C, sizeof(__pyx_k_SET_6_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_6_D, __pyx_k_SET_6_D, sizeof(__pyx_k_SET_6_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_6_E, __pyx_k_SET_6_E, sizeof(__pyx_k_SET_6_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_6_H, __pyx_k_SET_6_H, sizeof(__pyx_k_SET_6_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_6_HL, __pyx_k_SET_6_HL, sizeof(__pyx_k_SET_6_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_6_L, __pyx_k_SET_6_L, sizeof(__pyx_k_SET_6_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_7_A, __pyx_k_SET_7_A, sizeof(__pyx_k_SET_7_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_7_B, __pyx_k_SET_7_B, sizeof(__pyx_k_SET_7_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_7_C, __pyx_k_SET_7_C, sizeof(__pyx_k_SET_7_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_7_D, __pyx_k_SET_7_D, sizeof(__pyx_k_SET_7_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_7_E, __pyx_k_SET_7_E, sizeof(__pyx_k_SET_7_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_7_H, __pyx_k_SET_7_H, sizeof(__pyx_k_SET_7_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_7_HL, __pyx_k_SET_7_HL, sizeof(__pyx_k_SET_7_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SET_7_L, __pyx_k_SET_7_L, sizeof(__pyx_k_SET_7_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SLA_A, __pyx_k_SLA_A, sizeof(__pyx_k_SLA_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SLA_B, __pyx_k_SLA_B, sizeof(__pyx_k_SLA_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SLA_C, __pyx_k_SLA_C, sizeof(__pyx_k_SLA_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SLA_D, __pyx_k_SLA_D, sizeof(__pyx_k_SLA_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SLA_E, __pyx_k_SLA_E, sizeof(__pyx_k_SLA_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SLA_H, __pyx_k_SLA_H, sizeof(__pyx_k_SLA_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SLA_HL, __pyx_k_SLA_HL, sizeof(__pyx_k_SLA_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SLA_L, __pyx_k_SLA_L, sizeof(__pyx_k_SLA_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SRA_A, __pyx_k_SRA_A, sizeof(__pyx_k_SRA_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SRA_B, __pyx_k_SRA_B, sizeof(__pyx_k_SRA_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SRA_C, __pyx_k_SRA_C, sizeof(__pyx_k_SRA_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SRA_D, __pyx_k_SRA_D, sizeof(__pyx_k_SRA_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SRA_E, __pyx_k_SRA_E, sizeof(__pyx_k_SRA_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SRA_H, __pyx_k_SRA_H, sizeof(__pyx_k_SRA_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SRA_HL, __pyx_k_SRA_HL, sizeof(__pyx_k_SRA_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SRA_L, __pyx_k_SRA_L, sizeof(__pyx_k_SRA_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SRL_A, __pyx_k_SRL_A, sizeof(__pyx_k_SRL_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SRL_B, __pyx_k_SRL_B, sizeof(__pyx_k_SRL_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SRL_C, __pyx_k_SRL_C, sizeof(__pyx_k_SRL_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SRL_D, __pyx_k_SRL_D, sizeof(__pyx_k_SRL_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SRL_E, __pyx_k_SRL_E, sizeof(__pyx_k_SRL_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SRL_H, __pyx_k_SRL_H, sizeof(__pyx_k_SRL_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SRL_HL, __pyx_k_SRL_HL, sizeof(__pyx_k_SRL_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SRL_L, __pyx_k_SRL_L, sizeof(__pyx_k_SRL_L), 0, 1, 0, 0},
    {&__pyx_kp_u_STOP_0, __pyx_k_STOP_0, sizeof(__pyx_k_STOP_0), 0, 1, 0, 0},
    {&__pyx_kp_u_SUB_A, __pyx_k_SUB_A, sizeof(__pyx_k_SUB_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SUB_B, __pyx_k_SUB_B, sizeof(__pyx_k_SUB_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SUB_C, __pyx_k_SUB_C, sizeof(__pyx_k_SUB_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SUB_D, __pyx_k_SUB_D, sizeof(__pyx_k_SUB_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SUB_E, __pyx_k_SUB_E, sizeof(__pyx_k_SUB_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SUB_H, __pyx_k_SUB_H, sizeof(__pyx_k_SUB_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SUB_HL, __pyx_k_SUB_HL, sizeof(__pyx_k_SUB_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SUB_L, __pyx_k_SUB_L, sizeof(__pyx_k_SUB_L), 0, 1, 0, 0},
    {&__pyx_kp_u_SUB_d8, __pyx_k_SUB_d8, sizeof(__pyx_k_SUB_d8), 0, 1, 0, 0},
    {&__pyx_kp_u_SWAP_A, __pyx_k_SWAP_A, sizeof(__pyx_k_SWAP_A), 0, 1, 0, 0},
    {&__pyx_kp_u_SWAP_B, __pyx_k_SWAP_B, sizeof(__pyx_k_SWAP_B), 0, 1, 0, 0},
    {&__pyx_kp_u_SWAP_C, __pyx_k_SWAP_C, sizeof(__pyx_k_SWAP_C), 0, 1, 0, 0},
    {&__pyx_kp_u_SWAP_D, __pyx_k_SWAP_D, sizeof(__pyx_k_SWAP_D), 0, 1, 0, 0},
    {&__pyx_kp_u_SWAP_E, __pyx_k_SWAP_E, sizeof(__pyx_k_SWAP_E), 0, 1, 0, 0},
    {&__pyx_kp_u_SWAP_H, __pyx_k_SWAP_H, sizeof(__pyx_k_SWAP_H), 0, 1, 0, 0},
    {&__pyx_kp_u_SWAP_HL, __pyx_k_SWAP_HL, sizeof(__pyx_k_SWAP_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_SWAP_L, __pyx_k_SWAP_L, sizeof(__pyx_k_SWAP_L), 0, 1, 0, 0},
    {&__pyx_n_s_Sequence, __pyx_k_Sequence, sizeof(__pyx_k_Sequence), 0, 0, 1, 1},
    {&__pyx_kp_s_Step_may_not_be_zero_axis_d, __pyx_k_Step_may_not_be_zero_axis_d, sizeof(__pyx_k_Step_may_not_be_zero_axis_d), 0, 0, 1, 0},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
    {&__pyx_kp_u_XOR_A, __pyx_k_XOR_A, sizeof(__pyx_k_XOR_A), 0, 1, 0, 0},
    {&__pyx_kp_u_XOR_B, __pyx_k_XOR_B, sizeof(__pyx_k_XOR_B), 0, 1, 0, 0},
    {&__pyx_kp_u_XOR_C, __pyx_k_XOR_C, sizeof(__pyx_k_XOR_C), 0, 1, 0, 0},
    {&__pyx_kp_u_XOR_D, __pyx_k_XOR_D, sizeof(__pyx_k_XOR_D), 0, 1, 0, 0},
    {&__pyx_kp_u_XOR_E, __pyx_k_XOR_E, sizeof(__pyx_k_XOR_E), 0, 1, 0, 0},
    {&__pyx_kp_u_XOR_H, __pyx_k_XOR_H, sizeof(__pyx_k_XOR_H), 0, 1, 0, 0},
    {&__pyx_kp_u_XOR_HL, __pyx_k_XOR_HL, sizeof(__pyx_k_XOR_HL), 0, 1, 0, 0},
    {&__pyx_kp_u_XOR_L, __pyx_k_XOR_L, sizeof(__pyx_k_XOR_L), 0, 1, 0, 0},
    {&__pyx_kp_u_XOR_d8, __pyx_k_XOR_d8, sizeof(__pyx_k_XOR_d8), 0, 1, 0, 0},
    {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
    {&__pyx_kp_u__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 1, 0, 0},
    {&__pyx_n_s__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 0, 1, 1},
    {&__pyx_n_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 1},
    {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
    {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
    {&__pyx_n_s_abc, __pyx_k_abc, sizeof(__pyx_k_abc), 0, 0, 1, 1},
    {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
    {&__pyx_kp_u_and, __pyx_k_and, sizeof(__pyx_k_and), 0, 1, 0, 0},
    {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
    {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
    {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
    {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
    {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
    {&__pyx_kp_s_collections_abc, __pyx_k_collections_abc, sizeof(__pyx_k_collections_abc), 0, 0, 1, 0},
    {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
    {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
    {&__pyx_n_s_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
    {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
    {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
    {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
    {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_get_logger, __pyx_k_get_logger, sizeof(__pyx_k_get_logger), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_kp_u_got, __pyx_k_got, sizeof(__pyx_k_got), 0, 1, 0, 0},
    {&__pyx_kp_u_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 1, 0, 0},
    {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
    {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
    {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
    {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
    {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
    {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
    {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_s_pyboy, __pyx_k_pyboy, sizeof(__pyx_k_pyboy), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_register, __pyx_k_register, sizeof(__pyx_k_register), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
    {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
    {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
    {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
    {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
    {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
    {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
    {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
    {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_utils, __pyx_k_utils, sizeof(__pyx_k_utils), 0, 0, 1, 1},
    {&__pyx_n_s_version_info, __pyx_k_version_info, sizeof(__pyx_k_version_info), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 83, __pyx_L1_error)
  __pyx_builtin_OverflowError = __Pyx_GetBuiltinName(__pyx_n_s_OverflowError); if (!__pyx_builtin_OverflowError) __PYX_ERR(1, 83, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(1, 86, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(1, 96, __pyx_L1_error)
  __pyx_builtin___import__ = __Pyx_GetBuiltinName(__pyx_n_s_import); if (!__pyx_builtin___import__) __PYX_ERR(1, 100, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(1, 141, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(1, 156, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(1, 373, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(1, 408, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(1, 618, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(3, 29, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "View.MemoryView":582
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__4 = PyTuple_New(1); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_tuple__4, 0, __pyx_int_neg_1)) __PYX_ERR(1, 582, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "View.MemoryView":679
 *     tup = <tuple>index if isinstance(index, tuple) else (index,)
 * 
 *     result = [slice(None)] * ndim             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_slice__5 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__5)) __PYX_ERR(1, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__5);
  __Pyx_GIVEREF(__pyx_slice__5);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x82a3537, 0x6ae9995, 0xb068931):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x82a3537, 0x6ae9995, 0xb068931) = (name))" % __pyx_checksum
 */
  __pyx_tuple__8 = PyTuple_Pack(3, __pyx_int_136983863, __pyx_int_112105877, __pyx_int_184977713); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "View.MemoryView":100
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:
 *     if __import__("sys").version_info >= (3, 3):             # <<<<<<<<<<<<<<
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_n_s_sys); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_int_3, __pyx_int_3); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "View.MemoryView":101
 * try:
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence             # <<<<<<<<<<<<<<
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_collections_abc); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "View.MemoryView":103
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence             # <<<<<<<<<<<<<<
 * except:
 * 
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_n_s_collections); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(1, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "View.MemoryView":309
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "View.MemoryView":310
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "View.MemoryView":311
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "View.MemoryView":314
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "View.MemoryView":315
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__18 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "pyboy/core/opcodes.py":12
 * logger = pyboy.logging.get_logger(__name__)
 * 
 * FLAGC, FLAGH, FLAGN, FLAGZ = range(4, 8)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__20 = PyTuple_Pack(2, __pyx_int_4, __pyx_int_8); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 2, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_136983863 = PyInt_FromLong(136983863L); if (unlikely(!__pyx_int_136983863)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_5pyboy_4core_7opcodes_logger = ((struct __pyx_obj_5pyboy_7logging_7logging_Logger *)Py_None); Py_INCREF(Py_None);
  __pyx_collections_abc_Sequence = Py_None; Py_INCREF(Py_None);
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_logger, (void *)&__pyx_v_5pyboy_4core_7opcodes_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_FLAGC, (void *)&__pyx_v_5pyboy_4core_7opcodes_FLAGC, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_FLAGH, (void *)&__pyx_v_5pyboy_4core_7opcodes_FLAGH, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_FLAGN, (void *)&__pyx_v_5pyboy_4core_7opcodes_FLAGN, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_FLAGZ, (void *)&__pyx_v_5pyboy_4core_7opcodes_FLAGZ, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_OPCODE_LENGTHS, (void *)&__pyx_v_5pyboy_4core_7opcodes_OPCODE_LENGTHS, "uint8_t [0x200]") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("execute_opcode", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_execute_opcode, "int (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, uint16_t)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("no_opcode", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_no_opcode, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("NOP_00", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_NOP_00, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_01", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_01, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_02", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_02, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_03", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_03, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_04", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_04, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_05", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_05, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_06", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_06, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLCA_07", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLCA_07, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_08", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_08, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_09", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_09, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_0A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_0A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_0B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_0B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_0C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_0C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_0D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_0D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_0E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_0E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRCA_0F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRCA_0F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("STOP_10", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_STOP_10, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_11", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_11, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_12", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_12, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_13", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_13, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_14", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_14, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_15", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_15, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_16", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_16, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLA_17", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLA_17, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JR_18", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JR_18, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_19", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_19, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_1A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_1A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_1B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_1B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_1C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_1C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_1D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_1D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_1E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_1E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRA_1F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRA_1F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JR_20", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JR_20, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_21", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_21, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_22", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_22, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_23", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_23, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_24", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_24, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_25", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_25, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_26", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_26, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DAA_27", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DAA_27, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JR_28", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JR_28, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_29", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_29, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_2A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_2A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_2B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_2B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_2C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_2C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_2D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_2D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_2E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_2E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CPL_2F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CPL_2F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JR_30", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JR_30, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_31", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_31, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_32", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_32, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_33", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_33, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_34", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_34, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_35", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_35, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_36", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_36, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SCF_37", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SCF_37, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JR_38", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JR_38, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_39", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_39, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_3A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_3A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_3B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_3B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("INC_3C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_INC_3C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DEC_3D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DEC_3D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_3E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_3E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CCF_3F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CCF_3F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_40", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_40, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_41", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_41, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_42", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_42, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_43", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_43, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_44", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_44, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_45", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_45, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_46", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_46, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_47", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_47, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_48", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_48, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_49", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_49, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_4A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_4A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_4B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_4B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_4C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_4C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_4D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_4D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_4E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_4E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_4F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_4F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_50", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_50, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_51", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_51, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_52", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_52, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_53", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_53, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_54", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_54, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_55", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_55, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_56", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_56, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_57", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_57, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_58", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_58, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_59", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_59, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_5A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_5A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_5B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_5B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_5C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_5C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_5D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_5D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_5E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_5E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_5F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_5F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_60", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_60, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_61", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_61, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_62", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_62, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_63", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_63, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_64", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_64, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_65", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_65, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_66", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_66, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_67", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_67, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_68", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_68, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_69", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_69, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_6A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_6A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_6B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_6B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_6C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_6C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_6D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_6D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_6E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_6E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_6F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_6F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_70", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_70, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_71", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_71, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_72", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_72, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_73", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_73, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_74", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_74, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_75", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_75, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("HALT_76", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_HALT_76, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_77", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_77, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_78", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_78, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_79", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_79, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_7A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_7A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_7B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_7B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_7C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_7C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_7D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_7D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_7E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_7E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_7F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_7F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_80", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_80, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_81", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_81, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_82", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_82, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_83", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_83, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_84", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_84, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_85", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_85, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_86", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_86, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_87", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_87, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADC_88", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADC_88, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADC_89", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADC_89, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADC_8A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADC_8A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADC_8B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADC_8B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADC_8C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADC_8C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADC_8D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADC_8D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADC_8E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADC_8E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADC_8F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADC_8F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SUB_90", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SUB_90, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SUB_91", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SUB_91, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SUB_92", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SUB_92, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SUB_93", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SUB_93, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SUB_94", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SUB_94, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SUB_95", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SUB_95, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SUB_96", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SUB_96, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SUB_97", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SUB_97, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SBC_98", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SBC_98, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SBC_99", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SBC_99, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SBC_9A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SBC_9A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SBC_9B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SBC_9B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SBC_9C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SBC_9C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SBC_9D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SBC_9D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SBC_9E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SBC_9E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SBC_9F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SBC_9F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("AND_A0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_AND_A0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("AND_A1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_AND_A1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("AND_A2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_AND_A2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("AND_A3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_AND_A3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("AND_A4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_AND_A4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("AND_A5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_AND_A5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("AND_A6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_AND_A6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("AND_A7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_AND_A7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("XOR_A8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_XOR_A8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("XOR_A9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_XOR_A9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("XOR_AA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_XOR_AA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("XOR_AB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_XOR_AB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("XOR_AC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_XOR_AC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("XOR_AD", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_XOR_AD, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("XOR_AE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_XOR_AE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("XOR_AF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_XOR_AF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("OR_B0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_OR_B0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("OR_B1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_OR_B1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("OR_B2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_OR_B2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("OR_B3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_OR_B3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("OR_B4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_OR_B4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("OR_B5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_OR_B5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("OR_B6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_OR_B6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("OR_B7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_OR_B7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CP_B8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CP_B8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CP_B9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CP_B9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CP_BA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CP_BA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CP_BB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CP_BB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CP_BC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CP_BC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CP_BD", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CP_BD, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CP_BE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CP_BE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CP_BF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CP_BF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RET_C0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RET_C0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("POP_C1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_POP_C1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JP_C2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JP_C2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JP_C3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JP_C3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CALL_C4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CALL_C4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("PUSH_C5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_PUSH_C5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_C6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_C6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RST_C7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RST_C7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RET_C8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RET_C8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RET_C9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RET_C9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JP_CA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JP_CA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("PREFIX_CB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_PREFIX_CB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CALL_CC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CALL_CC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CALL_CD", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CALL_CD, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADC_CE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADC_CE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RST_CF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RST_CF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RET_D0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RET_D0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("POP_D1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_POP_D1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JP_D2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JP_D2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CALL_D4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CALL_D4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("PUSH_D5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_PUSH_D5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SUB_D6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SUB_D6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RST_D7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RST_D7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RET_D8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RET_D8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RETI_D9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RETI_D9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JP_DA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JP_DA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CALL_DC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CALL_DC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SBC_DE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SBC_DE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RST_DF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RST_DF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LDH_E0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LDH_E0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("POP_E1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_POP_E1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_E2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_E2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("PUSH_E5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_PUSH_E5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("AND_E6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_AND_E6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RST_E7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RST_E7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("ADD_E8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_ADD_E8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("JP_E9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_JP_E9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_EA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_EA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("XOR_EE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_XOR_EE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RST_EF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RST_EF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LDH_F0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LDH_F0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("POP_F1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_POP_F1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_F2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_F2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("DI_F3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_DI_F3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("PUSH_F5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_PUSH_F5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("OR_F6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_OR_F6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RST_F7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RST_F7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_F8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_F8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_F9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_F9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("LD_FA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_LD_FA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("EI_FB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_EI_FB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("CP_FE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_CP_FE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *, int)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RST_FF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RST_FF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLC_100", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLC_100, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLC_101", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLC_101, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLC_102", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLC_102, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLC_103", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLC_103, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLC_104", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLC_104, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLC_105", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLC_105, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLC_106", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLC_106, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RLC_107", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RLC_107, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRC_108", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRC_108, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRC_109", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRC_109, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRC_10A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRC_10A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRC_10B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRC_10B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRC_10C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRC_10C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRC_10D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRC_10D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRC_10E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRC_10E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RRC_10F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RRC_10F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RL_110", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RL_110, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RL_111", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RL_111, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RL_112", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RL_112, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RL_113", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RL_113, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RL_114", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RL_114, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RL_115", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RL_115, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RL_116", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RL_116, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RL_117", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RL_117, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RR_118", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RR_118, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RR_119", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RR_119, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RR_11A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RR_11A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RR_11B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RR_11B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RR_11C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RR_11C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RR_11D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RR_11D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RR_11E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RR_11E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RR_11F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RR_11F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SLA_120", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SLA_120, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SLA_121", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SLA_121, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SLA_122", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SLA_122, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SLA_123", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SLA_123, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SLA_124", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SLA_124, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SLA_125", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SLA_125, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SLA_126", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SLA_126, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SLA_127", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SLA_127, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRA_128", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRA_128, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRA_129", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRA_129, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRA_12A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRA_12A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRA_12B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRA_12B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRA_12C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRA_12C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRA_12D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRA_12D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRA_12E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRA_12E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRA_12F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRA_12F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SWAP_130", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SWAP_130, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SWAP_131", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SWAP_131, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SWAP_132", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SWAP_132, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SWAP_133", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SWAP_133, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SWAP_134", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SWAP_134, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SWAP_135", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SWAP_135, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SWAP_136", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SWAP_136, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SWAP_137", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SWAP_137, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRL_138", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRL_138, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRL_139", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRL_139, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRL_13A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRL_13A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRL_13B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRL_13B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRL_13C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRL_13C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRL_13D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRL_13D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRL_13E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRL_13E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SRL_13F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SRL_13F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_140", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_140, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_141", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_141, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_142", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_142, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_143", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_143, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_144", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_144, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_145", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_145, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_146", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_146, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_147", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_147, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_148", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_148, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_149", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_149, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_14A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_14A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_14B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_14B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_14C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_14C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_14D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_14D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_14E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_14E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_14F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_14F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_150", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_150, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_151", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_151, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_152", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_152, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_153", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_153, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_154", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_154, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_155", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_155, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_156", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_156, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_157", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_157, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_158", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_158, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_159", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_159, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_15A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_15A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_15B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_15B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_15C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_15C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_15D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_15D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_15E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_15E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_15F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_15F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_160", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_160, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_161", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_161, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_162", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_162, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_163", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_163, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_164", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_164, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_165", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_165, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_166", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_166, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_167", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_167, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_168", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_168, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_169", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_169, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_16A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_16A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_16B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_16B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_16C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_16C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_16D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_16D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_16E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_16E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_16F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_16F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_170", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_170, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_171", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_171, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_172", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_172, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_173", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_173, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_174", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_174, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_175", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_175, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_176", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_176, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_177", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_177, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_178", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_178, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_179", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_179, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_17A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_17A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_17B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_17B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_17C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_17C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_17D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_17D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_17E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_17E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("BIT_17F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_BIT_17F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_180", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_180, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_181", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_181, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_182", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_182, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_183", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_183, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_184", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_184, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_185", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_185, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_186", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_186, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_187", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_187, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_188", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_188, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_189", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_189, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_18A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_18A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_18B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_18B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_18C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_18C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_18D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_18D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_18E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_18E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_18F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_18F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_190", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_190, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_191", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_191, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_192", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_192, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_193", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_193, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_194", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_194, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_195", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_195, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_196", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_196, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_197", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_197, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_198", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_198, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_199", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_199, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_19A", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_19A, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_19B", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_19B, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_19C", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_19C, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_19D", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_19D, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_19E", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_19E, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_19F", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_19F, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1A9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1A9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1AA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1AA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1AB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1AB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1AC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1AC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1AD", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1AD, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1AE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1AE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1AF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1AF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1B9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1B9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1BA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1BA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1BB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1BB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1BC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1BC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1BD", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1BD, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1BE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1BE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("RES_1BF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_RES_1BF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1C9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1C9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1CA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1CA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1CB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1CB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1CC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1CC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1CD", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1CD, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1CE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1CE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1CF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1CF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1D9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1D9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1DA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1DA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1DB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1DB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1DC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1DC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1DD", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1DD, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1DE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1DE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1DF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1DF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1E9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1E9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1EA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1EA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1EB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1EB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1EC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1EC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1ED", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1ED, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1EE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1EE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1EF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1EF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F0", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F0, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F1", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F1, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F2", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F2, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F3", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F3, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F4", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F4, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F5", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F5, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F6", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F6, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F7", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F7, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F8", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F8, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1F9", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1F9, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1FA", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1FA, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1FB", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1FB, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1FC", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1FC, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1FD", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1FD, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1FE", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1FE, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ExportFunction("SET_1FF", (void (*)(void))__pyx_f_5pyboy_4core_7opcodes_SET_1FF, "uint8_t (struct __pyx_obj_5pyboy_4core_3cpu_CPU *)") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_array_type = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type___pyx_array_spec, NULL); if (unlikely(!__pyx_array_type)) __PYX_ERR(1, 114, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_array_type->tp_as_buffer = &__pyx_tp_as_buffer_array;
  if (!__pyx_array_type->tp_as_buffer->bf_releasebuffer && __pyx_array_type->tp_base->tp_as_buffer && __pyx_array_type->tp_base->tp_as_buffer->bf_releasebuffer) {
    __pyx_array_type->tp_as_buffer->bf_releasebuffer = __pyx_array_type->tp_base->tp_as_buffer->bf_releasebuffer;
  }
  #elif defined(Py_bf_getbuffer) && defined(Py_bf_releasebuffer)
  /* PY_VERSION_HEX >= 0x03090000 || Py_LIMITED_API >= 0x030B0000 */
  #elif defined(_MSC_VER)
  #pragma message ("The buffer protocol is not supported in the Limited C-API < 3.11.")
  #else
  #warning "The buffer protocol is not supported in the Limited C-API < 3.11."
  #endif
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type___pyx_array_spec, __pyx_array_type) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #else
  __pyx_array_type = &__pyx_type___pyx_array;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_array_type) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_array_type->tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_array_type, __pyx_vtabptr_array) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_array_type) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_array_type) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_MemviewEnum_type = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type___pyx_MemviewEnum_spec, NULL); if (unlikely(!__pyx_MemviewEnum_type)) __PYX_ERR(1, 302, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type___pyx_MemviewEnum_spec, __pyx_MemviewEnum_type) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #else
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_MemviewEnum_type) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_MemviewEnum_type->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_MemviewEnum_type->tp_dictoffset && __pyx_MemviewEnum_type->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_MemviewEnum_type->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_MemviewEnum_type) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #endif
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  __pyx_vtable_memoryview._get_base = (PyObject *(*)(struct __pyx_memoryview_obj *))__pyx_memoryview__get_base;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_memoryview_type = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type___pyx_memoryview_spec, NULL); if (unlikely(!__pyx_memoryview_type)) __PYX_ERR(1, 337, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_memoryview_type->tp_as_buffer = &__pyx_tp_as_buffer_memoryview;
  if (!__pyx_memoryview_type->tp_as_buffer->bf_releasebuffer && __pyx_memoryview_type->tp_base->tp_as_buffer && __pyx_memoryview_type->tp_base->tp_as_buffer->bf_releasebuffer) {
    __pyx_memoryview_type->tp_as_buffer->bf_releasebuffer = __pyx_memoryview_type->tp_base->tp_as_buffer->bf_releasebuffer;
  }
  #elif defined(Py_bf_getbuffer) && defined(Py_bf_releasebuffer)
  /* PY_VERSION_HEX >= 0x03090000 || Py_LIMITED_API >= 0x030B0000 */
  #elif defined(_MSC_VER)
  #pragma message ("The buffer protocol is not supported in the Limited C-API < 3.11.")
  #else
  #warning "The buffer protocol is not supported in the Limited C-API < 3.11."
  #endif
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type___pyx_memoryview_spec, __pyx_memoryview_type) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #else
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_memoryview_type) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_memoryview_type->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_memoryview_type->tp_dictoffset && __pyx_memoryview_type->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_memoryview_type->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_memoryview_type, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_memoryview_type) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_memoryview_type) < 0) __PYX_ERR(1, 337, __pyx_L1_error)
  #endif
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_vtable__memoryviewslice.__pyx_base._get_base = (PyObject *(*)(struct __pyx_memoryview_obj *))__pyx_memoryviewslice__get_base;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_memoryview_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_memoryviewslice_type = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type___pyx_memoryviewslice_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_memoryviewslice_type)) __PYX_ERR(1, 952, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type___pyx_memoryviewslice_spec, __pyx_memoryviewslice_type) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #else
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_memoryviewslice_type->tp_base = __pyx_memoryview_type;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_memoryviewslice_type) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_memoryviewslice_type->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_memoryviewslice_type->tp_dictoffset && __pyx_memoryviewslice_type->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_memoryviewslice_type->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_memoryviewslice_type, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_memoryviewslice_type) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_memoryviewslice_type) < 0) __PYX_ERR(1, 952, __pyx_L1_error)
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyboy.logging.logging"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_7logging_7logging_Logger = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.logging.logging", "Logger", sizeof(struct __pyx_obj_5pyboy_7logging_7logging_Logger), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_7logging_7logging_Logger),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_7logging_7logging_Logger) __PYX_ERR(5, 6, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_7logging_7logging_Logger = (struct __pyx_vtabstruct_5pyboy_7logging_7logging_Logger*)__Pyx_GetVtable(__pyx_ptype_5pyboy_7logging_7logging_Logger); if (unlikely(!__pyx_vtabptr_5pyboy_7logging_7logging_Logger)) __PYX_ERR(5, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.bootrom"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_7bootrom_BootROM = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.bootrom", "BootROM", sizeof(struct __pyx_obj_5pyboy_4core_7bootrom_BootROM), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_7bootrom_BootROM),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_7bootrom_BootROM) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_7bootrom_BootROM = (struct __pyx_vtabstruct_5pyboy_4core_7bootrom_BootROM*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_7bootrom_BootROM); if (unlikely(!__pyx_vtabptr_5pyboy_4core_7bootrom_BootROM)) __PYX_ERR(6, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_8(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(7, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_8(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(8, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_8(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(9, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.utils"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_5utils_IntIOInterface = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.utils", "IntIOInterface", sizeof(struct __pyx_obj_5pyboy_5utils_IntIOInterface), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_5utils_IntIOInterface),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_5utils_IntIOInterface) __PYX_ERR(10, 12, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_5utils_IntIOInterface = (struct __pyx_vtabstruct_5pyboy_5utils_IntIOInterface*)__Pyx_GetVtable(__pyx_ptype_5pyboy_5utils_IntIOInterface); if (unlikely(!__pyx_vtabptr_5pyboy_5utils_IntIOInterface)) __PYX_ERR(10, 12, __pyx_L1_error)
  __pyx_ptype_5pyboy_5utils_IntIOWrapper = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.utils", "IntIOWrapper", sizeof(struct __pyx_obj_5pyboy_5utils_IntIOWrapper), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_5utils_IntIOWrapper),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_5utils_IntIOWrapper) __PYX_ERR(10, 28, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_5utils_IntIOWrapper = (struct __pyx_vtabstruct_5pyboy_5utils_IntIOWrapper*)__Pyx_GetVtable(__pyx_ptype_5pyboy_5utils_IntIOWrapper); if (unlikely(!__pyx_vtabptr_5pyboy_5utils_IntIOWrapper)) __PYX_ERR(10, 28, __pyx_L1_error)
  __pyx_ptype_5pyboy_5utils_WindowEvent = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.utils", "WindowEvent", sizeof(struct __pyx_obj_5pyboy_5utils_WindowEvent), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_5utils_WindowEvent),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_5utils_WindowEvent) __PYX_ERR(10, 40, __pyx_L1_error)
  __pyx_ptype_5pyboy_5utils_WindowEventMouse = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.utils", "WindowEventMouse", sizeof(struct __pyx_obj_5pyboy_5utils_WindowEventMouse), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_5utils_WindowEventMouse),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_5utils_WindowEventMouse) __PYX_ERR(10, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.cartridge.rtc"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_9cartridge_3rtc_RTC = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.cartridge.rtc", "RTC", sizeof(struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_9cartridge_3rtc_RTC),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_9cartridge_3rtc_RTC) __PYX_ERR(11, 18, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_9cartridge_3rtc_RTC = (struct __pyx_vtabstruct_5pyboy_4core_9cartridge_3rtc_RTC*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_9cartridge_3rtc_RTC); if (unlikely(!__pyx_vtabptr_5pyboy_4core_9cartridge_3rtc_RTC)) __PYX_ERR(11, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.cartridge.base_mbc"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_BaseMBC = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.cartridge.base_mbc", "BaseMBC", sizeof(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_BaseMBC),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_BaseMBC) __PYX_ERR(12, 15, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_9cartridge_8base_mbc_BaseMBC = (struct __pyx_vtabstruct_5pyboy_4core_9cartridge_8base_mbc_BaseMBC*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_BaseMBC); if (unlikely(!__pyx_vtabptr_5pyboy_4core_9cartridge_8base_mbc_BaseMBC)) __PYX_ERR(12, 15, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_ROMOnly = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.cartridge.base_mbc", "ROMOnly", sizeof(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_ROMOnly), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_9cartridge_8base_mbc_ROMOnly),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_ROMOnly) __PYX_ERR(12, 44, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_9cartridge_8base_mbc_ROMOnly = (struct __pyx_vtabstruct_5pyboy_4core_9cartridge_8base_mbc_ROMOnly*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_9cartridge_8base_mbc_ROMOnly); if (unlikely(!__pyx_vtabptr_5pyboy_4core_9cartridge_8base_mbc_ROMOnly)) __PYX_ERR(12, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.interaction"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_11interaction_Interaction = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.interaction", "Interaction", sizeof(struct __pyx_obj_5pyboy_4core_11interaction_Interaction), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_11interaction_Interaction),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_11interaction_Interaction) __PYX_ERR(13, 19, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_11interaction_Interaction = (struct __pyx_vtabstruct_5pyboy_4core_11interaction_Interaction*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_11interaction_Interaction); if (unlikely(!__pyx_vtabptr_5pyboy_4core_11interaction_Interaction)) __PYX_ERR(13, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("array"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_5array_array = __Pyx_ImportType_3_0_8(__pyx_t_1, "array", "array", sizeof(arrayobject), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(arrayobject),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_7cpython_5array_array) __PYX_ERR(4, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.lcd"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_3lcd_LCD = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "LCD", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_LCD), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_LCD),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_LCD) __PYX_ERR(14, 26, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_LCD = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_LCD*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_LCD); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_LCD)) __PYX_ERR(14, 26, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_3lcd_PaletteRegister = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "PaletteRegister", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_PaletteRegister),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_PaletteRegister) __PYX_ERR(14, 77, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_PaletteRegister = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteRegister*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_PaletteRegister); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_PaletteRegister)) __PYX_ERR(14, 77, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_3lcd_STATRegister = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "STATRegister", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_STATRegister), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_STATRegister),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_STATRegister) __PYX_ERR(14, 87, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_STATRegister = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_STATRegister*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_STATRegister); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_STATRegister)) __PYX_ERR(14, 87, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_3lcd_LCDCRegister = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "LCDCRegister", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_LCDCRegister), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_LCDCRegister),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_LCDCRegister) __PYX_ERR(14, 94, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_LCDCRegister = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_LCDCRegister*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_LCDCRegister); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_LCDCRegister)) __PYX_ERR(14, 94, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_3lcd_Renderer = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "Renderer", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_Renderer), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_Renderer),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_Renderer) __PYX_ERR(14, 110, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_Renderer = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_Renderer*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_Renderer); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_Renderer)) __PYX_ERR(14, 110, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_3lcd_CGBLCD = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "CGBLCD", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_CGBLCD), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_CGBLCD),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_CGBLCD) __PYX_ERR(14, 238, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_CGBLCD = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_CGBLCD*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_CGBLCD); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_CGBLCD)) __PYX_ERR(14, 238, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_3lcd_CGBRenderer = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "CGBRenderer", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_CGBRenderer), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_CGBRenderer),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_CGBRenderer) __PYX_ERR(14, 247, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_CGBRenderer = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_CGBRenderer*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_CGBRenderer); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_CGBRenderer)) __PYX_ERR(14, 247, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_3lcd_VBKregister = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "VBKregister", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_VBKregister), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_VBKregister),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_VBKregister) __PYX_ERR(14, 250, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_VBKregister = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_VBKregister*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_VBKregister); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_VBKregister)) __PYX_ERR(14, 250, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_3lcd_PaletteIndexRegister = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "PaletteIndexRegister", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_PaletteIndexRegister),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_PaletteIndexRegister) __PYX_ERR(14, 256, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_PaletteIndexRegister = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteIndexRegister*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_PaletteIndexRegister); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_PaletteIndexRegister)) __PYX_ERR(14, 256, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_3lcd_PaletteColorRegister = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.lcd", "PaletteColorRegister", sizeof(struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3lcd_PaletteColorRegister),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3lcd_PaletteColorRegister) __PYX_ERR(14, 270, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3lcd_PaletteColorRegister = (struct __pyx_vtabstruct_5pyboy_4core_3lcd_PaletteColorRegister*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3lcd_PaletteColorRegister); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3lcd_PaletteColorRegister)) __PYX_ERR(14, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.ram"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_3ram_RAM = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.ram", "RAM", sizeof(struct __pyx_obj_5pyboy_4core_3ram_RAM), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3ram_RAM),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3ram_RAM) __PYX_ERR(15, 15, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3ram_RAM = (struct __pyx_vtabstruct_5pyboy_4core_3ram_RAM*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3ram_RAM); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3ram_RAM)) __PYX_ERR(15, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.sound"); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_5sound_Sound = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.sound", "Sound", sizeof(struct __pyx_obj_5pyboy_4core_5sound_Sound), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_5sound_Sound),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_5sound_Sound) __PYX_ERR(16, 18, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_5sound_Sound = (struct __pyx_vtabstruct_5pyboy_4core_5sound_Sound*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_5sound_Sound); if (unlikely(!__pyx_vtabptr_5pyboy_4core_5sound_Sound)) __PYX_ERR(16, 18, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_5sound_ToneChannel = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.sound", "ToneChannel", sizeof(struct __pyx_obj_5pyboy_4core_5sound_ToneChannel), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_5sound_ToneChannel),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_5sound_ToneChannel) __PYX_ERR(16, 52, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_5sound_ToneChannel = (struct __pyx_vtabstruct_5pyboy_4core_5sound_ToneChannel*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_5sound_ToneChannel); if (unlikely(!__pyx_vtabptr_5pyboy_4core_5sound_ToneChannel)) __PYX_ERR(16, 52, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_5sound_SweepChannel = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.sound", "SweepChannel", sizeof(struct __pyx_obj_5pyboy_4core_5sound_SweepChannel), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_5sound_SweepChannel),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_5sound_SweepChannel) __PYX_ERR(16, 83, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_5sound_SweepChannel = (struct __pyx_vtabstruct_5pyboy_4core_5sound_SweepChannel*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_5sound_SweepChannel); if (unlikely(!__pyx_vtabptr_5pyboy_4core_5sound_SweepChannel)) __PYX_ERR(16, 83, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_5sound_WaveChannel = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.sound", "WaveChannel", sizeof(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_5sound_WaveChannel),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_5sound_WaveChannel) __PYX_ERR(16, 101, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_5sound_WaveChannel = (struct __pyx_vtabstruct_5pyboy_4core_5sound_WaveChannel*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_5sound_WaveChannel); if (unlikely(!__pyx_vtabptr_5pyboy_4core_5sound_WaveChannel)) __PYX_ERR(16, 101, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_5sound_NoiseChannel = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.sound", "NoiseChannel", sizeof(struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_5sound_NoiseChannel),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_5sound_NoiseChannel) __PYX_ERR(16, 134, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_5sound_NoiseChannel = (struct __pyx_vtabstruct_5pyboy_4core_5sound_NoiseChannel*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_5sound_NoiseChannel); if (unlikely(!__pyx_vtabptr_5pyboy_4core_5sound_NoiseChannel)) __PYX_ERR(16, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.timer"); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_5timer_Timer = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.timer", "Timer", sizeof(struct __pyx_obj_5pyboy_4core_5timer_Timer), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_5timer_Timer),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_5timer_Timer) __PYX_ERR(17, 17, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_5timer_Timer = (struct __pyx_vtabstruct_5pyboy_4core_5timer_Timer*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_5timer_Timer); if (unlikely(!__pyx_vtabptr_5pyboy_4core_5timer_Timer)) __PYX_ERR(17, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.mb"); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_2mb_Motherboard = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.mb", "Motherboard", sizeof(struct __pyx_obj_5pyboy_4core_2mb_Motherboard), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_2mb_Motherboard),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_2mb_Motherboard) __PYX_ERR(18, 30, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_2mb_Motherboard = (struct __pyx_vtabstruct_5pyboy_4core_2mb_Motherboard*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_2mb_Motherboard); if (unlikely(!__pyx_vtabptr_5pyboy_4core_2mb_Motherboard)) __PYX_ERR(18, 30, __pyx_L1_error)
  __pyx_ptype_5pyboy_4core_2mb_HDMA = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.mb", "HDMA", sizeof(struct __pyx_obj_5pyboy_4core_2mb_HDMA), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_2mb_HDMA),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_2mb_HDMA) __PYX_ERR(18, 73, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_2mb_HDMA = (struct __pyx_vtabstruct_5pyboy_4core_2mb_HDMA*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_2mb_HDMA); if (unlikely(!__pyx_vtabptr_5pyboy_4core_2mb_HDMA)) __PYX_ERR(18, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.cpu"); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5pyboy_4core_3cpu_CPU = __Pyx_ImportType_3_0_8(__pyx_t_1, "pyboy.core.cpu", "CPU", sizeof(struct __pyx_obj_5pyboy_4core_3cpu_CPU), __PYX_GET_STRUCT_ALIGNMENT_3_0_8(struct __pyx_obj_5pyboy_4core_3cpu_CPU),__Pyx_ImportType_CheckSize_Warn_3_0_8); if (!__pyx_ptype_5pyboy_4core_3cpu_CPU) __PYX_ERR(19, 27, __pyx_L1_error)
  __pyx_vtabptr_5pyboy_4core_3cpu_CPU = (struct __pyx_vtabstruct_5pyboy_4core_3cpu_CPU*)__Pyx_GetVtable(__pyx_ptype_5pyboy_4core_3cpu_CPU); if (unlikely(!__pyx_vtabptr_5pyboy_4core_3cpu_CPU)) __PYX_ERR(19, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyboy.logging.logging"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "CRITICAL", (void **)&__pyx_vp_5pyboy_7logging_7logging_CRITICAL, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "FATAL", (void **)&__pyx_vp_5pyboy_7logging_7logging_FATAL, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "ERROR", (void **)&__pyx_vp_5pyboy_7logging_7logging_ERROR, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "WARNING", (void **)&__pyx_vp_5pyboy_7logging_7logging_WARNING, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INFO", (void **)&__pyx_vp_5pyboy_7logging_7logging_INFO, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "DEBUG", (void **)&__pyx_vp_5pyboy_7logging_7logging_DEBUG, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.bootrom"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_7bootrom_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.cartridge.rtc"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_9cartridge_3rtc_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.cartridge.base_mbc"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_9cartridge_8base_mbc_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.interaction"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_11interaction_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "P10", (void **)&__pyx_vp_5pyboy_4core_11interaction_P10, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "P11", (void **)&__pyx_vp_5pyboy_4core_11interaction_P11, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "P12", (void **)&__pyx_vp_5pyboy_4core_11interaction_P12, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "P13", (void **)&__pyx_vp_5pyboy_4core_11interaction_P13, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "P14", (void **)&__pyx_vp_5pyboy_4core_11interaction_P14, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "P15", (void **)&__pyx_vp_5pyboy_4core_11interaction_P15, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.lcd"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_VBLANK", (void **)&__pyx_vp_5pyboy_4core_3lcd_INTR_VBLANK, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_LCDC", (void **)&__pyx_vp_5pyboy_4core_3lcd_INTR_LCDC, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_TIMER", (void **)&__pyx_vp_5pyboy_4core_3lcd_INTR_TIMER, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_SERIAL", (void **)&__pyx_vp_5pyboy_4core_3lcd_INTR_SERIAL, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_HIGHTOLOW", (void **)&__pyx_vp_5pyboy_4core_3lcd_INTR_HIGHTOLOW, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "LCDC", (void **)&__pyx_vp_5pyboy_4core_3lcd_LCDC, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "STAT", (void **)&__pyx_vp_5pyboy_4core_3lcd_STAT, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "SCY", (void **)&__pyx_vp_5pyboy_4core_3lcd_SCY, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "SCX", (void **)&__pyx_vp_5pyboy_4core_3lcd_SCX, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "LY", (void **)&__pyx_vp_5pyboy_4core_3lcd_LY, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "LYC", (void **)&__pyx_vp_5pyboy_4core_3lcd_LYC, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "DMA", (void **)&__pyx_vp_5pyboy_4core_3lcd_DMA, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "BGP", (void **)&__pyx_vp_5pyboy_4core_3lcd_BGP, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "OBP0", (void **)&__pyx_vp_5pyboy_4core_3lcd_OBP0, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "OBP1", (void **)&__pyx_vp_5pyboy_4core_3lcd_OBP1, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "WY", (void **)&__pyx_vp_5pyboy_4core_3lcd_WY, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "WX", (void **)&__pyx_vp_5pyboy_4core_3lcd_WX, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "ROWS", (void **)&__pyx_vp_5pyboy_4core_3lcd_ROWS, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "COLS", (void **)&__pyx_vp_5pyboy_4core_3lcd_COLS, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "TILES", (void **)&__pyx_vp_5pyboy_4core_3lcd_TILES, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "FRAME_CYCLES", (void **)&__pyx_vp_5pyboy_4core_3lcd_FRAME_CYCLES, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "VIDEO_RAM", (void **)&__pyx_vp_5pyboy_4core_3lcd_VIDEO_RAM, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "OBJECT_ATTRIBUTE_MEMORY", (void **)&__pyx_vp_5pyboy_4core_3lcd_OBJECT_ATTRIBUTE_MEMORY, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "COL0_FLAG", (void **)&__pyx_vp_5pyboy_4core_3lcd_COL0_FLAG, "uint32_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "BG_PRIORITY_FLAG", (void **)&__pyx_vp_5pyboy_4core_3lcd_BG_PRIORITY_FLAG, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "CGB_NUM_PALETTES", (void **)&__pyx_vp_5pyboy_4core_3lcd_CGB_NUM_PALETTES, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_3lcd_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.ram"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_3ram_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.sound"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_5sound_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "SOUND_DESYNC_THRESHOLD", (void **)&__pyx_vp_5pyboy_4core_5sound_SOUND_DESYNC_THRESHOLD, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "CPU_FREQ", (void **)&__pyx_vp_5pyboy_4core_5sound_CPU_FREQ, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.timer"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_5timer_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.mb"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_2mb_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "STAT", (void **)&__pyx_vp_5pyboy_4core_2mb_STAT, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "LY", (void **)&__pyx_vp_5pyboy_4core_2mb_LY, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "LYC", (void **)&__pyx_vp_5pyboy_4core_2mb_LYC, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "VBLANK", (void **)&__pyx_vp_5pyboy_4core_2mb_VBLANK, "short") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "LCDC", (void **)&__pyx_vp_5pyboy_4core_2mb_LCDC, "short") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "TIMER", (void **)&__pyx_vp_5pyboy_4core_2mb_TIMER, "short") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "SERIAL", (void **)&__pyx_vp_5pyboy_4core_2mb_SERIAL, "short") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "HIGHTOLOW", (void **)&__pyx_vp_5pyboy_4core_2mb_HIGHTOLOW, "short") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_VBLANK", (void **)&__pyx_vp_5pyboy_4core_2mb_INTR_VBLANK, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_LCDC", (void **)&__pyx_vp_5pyboy_4core_2mb_INTR_LCDC, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_TIMER", (void **)&__pyx_vp_5pyboy_4core_2mb_INTR_TIMER, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_SERIAL", (void **)&__pyx_vp_5pyboy_4core_2mb_INTR_SERIAL, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_HIGHTOLOW", (void **)&__pyx_vp_5pyboy_4core_2mb_INTR_HIGHTOLOW, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "STATE_VERSION", (void **)&__pyx_vp_5pyboy_4core_2mb_STATE_VERSION, "int") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyboy.core.cpu"); if (!__pyx_t_1) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "logger", (void **)&__pyx_vp_5pyboy_4core_3cpu_logger, "struct __pyx_obj_5pyboy_7logging_7logging_Logger *") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "IF_ADDRESS", (void **)&__pyx_vp_5pyboy_4core_3cpu_IF_ADDRESS, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "IE_ADDRESS", (void **)&__pyx_vp_5pyboy_4core_3cpu_IE_ADDRESS, "uint16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "FLAGC", (void **)&__pyx_vp_5pyboy_4core_3cpu_FLAGC, "int16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "FLAGH", (void **)&__pyx_vp_5pyboy_4core_3cpu_FLAGH, "int16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "FLAGN", (void **)&__pyx_vp_5pyboy_4core_3cpu_FLAGN, "int16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "FLAGZ", (void **)&__pyx_vp_5pyboy_4core_3cpu_FLAGZ, "int16_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_VBLANK", (void **)&__pyx_vp_5pyboy_4core_3cpu_INTR_VBLANK, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_LCDC", (void **)&__pyx_vp_5pyboy_4core_3cpu_INTR_LCDC, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_TIMER", (void **)&__pyx_vp_5pyboy_4core_3cpu_INTR_TIMER, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_SERIAL", (void **)&__pyx_vp_5pyboy_4core_3cpu_INTR_SERIAL, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_8(__pyx_t_1, "INTR_HIGHTOLOW", (void **)&__pyx_vp_5pyboy_4core_3cpu_INTR_HIGHTOLOW, "uint8_t") < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_opcodes(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_opcodes},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "opcodes",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initopcodes(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initopcodes(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_opcodes(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_opcodes(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_opcodes(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  static PyThread_type_lock __pyx_t_8[8];
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  uint16_t __pyx_t_13;
  uint16_t __pyx_t_14;
  uint16_t __pyx_t_15;
  uint16_t __pyx_t_16;
  uint8_t __pyx_t_17[0x200];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'opcodes' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("opcodes", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 2, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "opcodes" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 2, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 2, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_opcodes(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 2, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pyboy__core__opcodes) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 2, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyboy.core.opcodes")) {
      if (unlikely((PyDict_SetItemString(modules, "pyboy.core.opcodes", __pyx_m) < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely((__Pyx_modinit_variable_export_code() < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_function_export_code() < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_variable_import_code() < 0))) __PYX_ERR(0, 2, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif

  /* "View.MemoryView":99
 * 
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:             # <<<<<<<<<<<<<<
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "View.MemoryView":100
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:
 *     if __import__("sys").version_info >= (3, 3):             # <<<<<<<<<<<<<<
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 100, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_version_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 100, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_5, __pyx_tuple__10, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 100, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(1, 100, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {

        /* "View.MemoryView":101
 * try:
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence             # <<<<<<<<<<<<<<
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 101, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_abc); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 101, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_Sequence); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 101, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XGOTREF(__pyx_collections_abc_Sequence);
        __Pyx_DECREF_SET(__pyx_collections_abc_Sequence, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __pyx_t_4 = 0;

        /* "View.MemoryView":100
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:
 *     if __import__("sys").version_info >= (3, 3):             # <<<<<<<<<<<<<<
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 */
        goto __pyx_L8;
      }

      /* "View.MemoryView":103
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence             # <<<<<<<<<<<<<<
 * except:
 * 
 */
      /*else*/ {
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin___import__, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 103, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_Sequence); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 103, __pyx_L2_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XGOTREF(__pyx_collections_abc_Sequence);
        __Pyx_DECREF_SET(__pyx_collections_abc_Sequence, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_5);
        __pyx_t_5 = 0;
      }
      __pyx_L8:;

      /* "View.MemoryView":99
 * 
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:             # <<<<<<<<<<<<<<
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L7_try_end;
    __pyx_L2_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "View.MemoryView":104
 *     else:
 *         __pyx_collections_abc_Sequence = __import__("collections").Sequence
 * except:             # <<<<<<<<<<<<<<
 * 
 *     __pyx_collections_abc_Sequence = None
 */
    /*except:*/ {
      __Pyx_AddTraceback("View.MemoryView", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(1, 104, __pyx_L4_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "View.MemoryView":106
 * except:
 * 
 *     __pyx_collections_abc_Sequence = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XGOTREF(__pyx_collections_abc_Sequence);
      __Pyx_DECREF_SET(__pyx_collections_abc_Sequence, Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L3_exception_handled;
    }

    /* "View.MemoryView":99
 * 
 * cdef object __pyx_collections_abc_Sequence "__pyx_collections_abc_Sequence"
 * try:             # <<<<<<<<<<<<<<
 *     if __import__("sys").version_info >= (3, 3):
 *         __pyx_collections_abc_Sequence = __import__("collections.abc").abc.Sequence
 */
    __pyx_L4_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L3_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L7_try_end:;
  }

  /* "View.MemoryView":241
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "View.MemoryView":242
 * 
 *     try:
 *         count = __pyx_collections_abc_Sequence.count             # <<<<<<<<<<<<<<
 *         index = __pyx_collections_abc_Sequence.index
 *     except:
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_count); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 242, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_SetItemOnTypeDict(__pyx_array_type, __pyx_n_s_count, __pyx_t_7) < 0) __PYX_ERR(1, 242, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      PyType_Modified(__pyx_array_type);

      /* "View.MemoryView":243
 *     try:
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index             # <<<<<<<<<<<<<<
 *     except:
 *         pass
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_index); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 243, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_SetItemOnTypeDict(__pyx_array_type, __pyx_n_s_index, __pyx_t_7) < 0) __PYX_ERR(1, 243, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      PyType_Modified(__pyx_array_type);

      /* "View.MemoryView":241
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L16_try_end;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "View.MemoryView":244
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 *     except:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L12_exception_handled;
    }
    __pyx_L12_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L16_try_end:;
  }

  /* "View.MemoryView":309
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":310
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":311
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":314
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":315
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "View.MemoryView":323
 * 
 * 
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[8] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":324
 * 
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[8] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_8[0] = PyThread_allocate_lock();
  __pyx_t_8[1] = PyThread_allocate_lock();
  __pyx_t_8[2] = PyThread_allocate_lock();
  __pyx_t_8[3] = PyThread_allocate_lock();
  __pyx_t_8[4] = PyThread_allocate_lock();
  __pyx_t_8[5] = PyThread_allocate_lock();
  __pyx_t_8[6] = PyThread_allocate_lock();
  __pyx_t_8[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_8, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":982
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "View.MemoryView":983
 * 
 *     try:
 *         count = __pyx_collections_abc_Sequence.count             # <<<<<<<<<<<<<<
 *         index = __pyx_collections_abc_Sequence.index
 *     except:
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_count); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 983, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_SetItemOnTypeDict(__pyx_memoryviewslice_type, __pyx_n_s_count, __pyx_t_7) < 0) __PYX_ERR(1, 983, __pyx_L17_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      PyType_Modified(__pyx_memoryviewslice_type);

      /* "View.MemoryView":984
 *     try:
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index             # <<<<<<<<<<<<<<
 *     except:
 *         pass
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_index); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 984, __pyx_L17_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_SetItemOnTypeDict(__pyx_memoryviewslice_type, __pyx_n_s_index, __pyx_t_7) < 0) __PYX_ERR(1, 984, __pyx_L17_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      PyType_Modified(__pyx_memoryviewslice_type);

      /* "View.MemoryView":982
 * 
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L22_try_end;
    __pyx_L17_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "View.MemoryView":985
 *         count = __pyx_collections_abc_Sequence.count
 *         index = __pyx_collections_abc_Sequence.index
 *     except:             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L18_exception_handled;
    }
    __pyx_L18_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L22_try_end:;
  }

  /* "View.MemoryView":988
 *         pass
 * 
 * try:             # <<<<<<<<<<<<<<
 *     if __pyx_collections_abc_Sequence:
 * 
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "View.MemoryView":989
 * 
 * try:
 *     if __pyx_collections_abc_Sequence:             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_collections_abc_Sequence); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(1, 989, __pyx_L23_error)
      if (__pyx_t_6) {

        /* "View.MemoryView":993
 * 
 * 
 *         __pyx_collections_abc_Sequence.register(_memoryviewslice)             # <<<<<<<<<<<<<<
 *         __pyx_collections_abc_Sequence.register(array)
 * except:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_register); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 993, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_memoryviewslice_type)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 993, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "View.MemoryView":994
 * 
 *         __pyx_collections_abc_Sequence.register(_memoryviewslice)
 *         __pyx_collections_abc_Sequence.register(array)             # <<<<<<<<<<<<<<
 * except:
 *     pass  # ignore failure, it's a minor issue
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_collections_abc_Sequence, __pyx_n_s_register); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 994, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_array_type)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 994, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":989
 * 
 * try:
 *     if __pyx_collections_abc_Sequence:             # <<<<<<<<<<<<<<
 * 
 * 
 */
      }

      /* "View.MemoryView":988
 *         pass
 * 
 * try:             # <<<<<<<<<<<<<<
 *     if __pyx_collections_abc_Sequence:
 * 
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L28_try_end;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "View.MemoryView":995
 *         __pyx_collections_abc_Sequence.register(_memoryviewslice)
 *         __pyx_collections_abc_Sequence.register(array)
 * except:             # <<<<<<<<<<<<<<
 *     pass  # ignore failure, it's a minor issue
 * 
 */
    /*except:*/ {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L24_exception_handled;
    }
    __pyx_L24_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L28_try_end:;
  }

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_7 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_7) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pyboy/core/opcodes.py":6
 * # CHANGES TO THE CODE SHOULD BE MADE IN 'opcodes_gen.py'.
 * 
 * from pyboy import utils             # <<<<<<<<<<<<<<
 * import array
 * 
 */
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_utils);
  __Pyx_GIVEREF(__pyx_n_s_utils);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_utils)) __PYX_ERR(0, 6, __pyx_L1_error);
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_pyboy, __pyx_t_7, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_utils); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_utils, __pyx_t_7) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyboy/core/opcodes.py":7
 * 
 * from pyboy import utils
 * import array             # <<<<<<<<<<<<<<
 * 
 * import pyboy
 */
  __pyx_t_4 = __Pyx_ImportDottedModule(__pyx_n_s_array, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_array, __pyx_t_4) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyboy/core/opcodes.py":9
 * import array
 * 
 * import pyboy             # <<<<<<<<<<<<<<
 * logger = pyboy.logging.get_logger(__name__)
 * 
 */
  __pyx_t_4 = __Pyx_ImportDottedModule(__pyx_n_s_pyboy, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyboy, __pyx_t_4) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyboy/core/opcodes.py":10
 * 
 * import pyboy
 * logger = pyboy.logging.get_logger(__name__)             # <<<<<<<<<<<<<<
 * 
 * FLAGC, FLAGH, FLAGN, FLAGZ = range(4, 8)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyboy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_logging); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_get_logger); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_name_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5pyboy_7logging_7logging_Logger))))) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_XGOTREF((PyObject *)__pyx_v_5pyboy_4core_7opcodes_logger);
  __Pyx_DECREF_SET(__pyx_v_5pyboy_4core_7opcodes_logger, ((struct __pyx_obj_5pyboy_7logging_7logging_Logger *)__pyx_t_5));
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "pyboy/core/opcodes.py":12
 * logger = pyboy.logging.get_logger(__name__)
 * 
 * FLAGC, FLAGH, FLAGN, FLAGZ = range(4, 8)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
    PyObject* sequence = __pyx_t_5;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 12, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_10 = PyTuple_GET_ITEM(sequence, 3); 
    } else {
      __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_10 = PyList_GET_ITEM(sequence, 3); 
    }
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_10);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_7,&__pyx_t_4,&__pyx_t_9,&__pyx_t_10};
      for (i=0; i < 4; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 12, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_7,&__pyx_t_4,&__pyx_t_9,&__pyx_t_10};
    __pyx_t_11 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_12 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_11);
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_12(__pyx_t_11); if (unlikely(!item)) goto __pyx_L30_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 4) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
    __pyx_t_12 = NULL;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L31_unpacking_done;
    __pyx_L30_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 12, __pyx_L1_error)
    __pyx_L31_unpacking_done:;
  }
  __pyx_t_13 = __Pyx_PyInt_As_uint16_t(__pyx_t_7); if (unlikely((__pyx_t_13 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_14 = __Pyx_PyInt_As_uint16_t(__pyx_t_4); if (unlikely((__pyx_t_14 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_15 = __Pyx_PyInt_As_uint16_t(__pyx_t_9); if (unlikely((__pyx_t_15 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_16 = __Pyx_PyInt_As_uint16_t(__pyx_t_10); if (unlikely((__pyx_t_16 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_5pyboy_4core_7opcodes_FLAGC = __pyx_t_13;
  __pyx_v_5pyboy_4core_7opcodes_FLAGH = __pyx_t_14;
  __pyx_v_5pyboy_4core_7opcodes_FLAGN = __pyx_t_15;
  __pyx_v_5pyboy_4core_7opcodes_FLAGZ = __pyx_t_16;

  /* "pyboy/core/opcodes.py":6274
 * 
 * 
 * OPCODE_LENGTHS = array.array("B", [             # <<<<<<<<<<<<<<
 *     1, 3, 1, 1, 1, 1, 2, 1, 3, 1, 1, 1, 1, 1, 2, 1,
 *     2, 3, 1, 1, 1, 1, 2, 1, 2, 1, 1, 1, 1, 1, 2, 1,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 6274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(512); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 2, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 3, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 4, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 5, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 6, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 7, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 8, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 9, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 10, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 11, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 12, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 13, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 14, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 15, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 16, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 17, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 18, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 19, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 20, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 21, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 22, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 23, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 24, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 25, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 26, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 27, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 28, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 29, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 30, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 31, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 32, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 33, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 34, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 35, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 36, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 37, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 38, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 39, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 40, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 41, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 42, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 43, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 44, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 45, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 46, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 47, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 48, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 49, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 50, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 51, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 52, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 53, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 54, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 55, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 56, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 57, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 58, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 59, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 60, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 61, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 62, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 63, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 64, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 65, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 66, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 67, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 68, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 69, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 70, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 71, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 72, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 73, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 74, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 75, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 76, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 77, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 78, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 79, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 80, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 81, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 82, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 83, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 84, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 85, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 86, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 87, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 88, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 89, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 90, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 91, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 92, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 93, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 94, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 95, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 96, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 97, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 98, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 99, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 100, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 101, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 102, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 103, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 104, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 105, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 106, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 107, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 108, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 109, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 110, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 111, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 112, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 113, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 114, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 115, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 116, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 117, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 118, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 119, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 120, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 121, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 122, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 123, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 124, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 125, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 126, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 127, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 128, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 129, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 130, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 131, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 132, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 133, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 134, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 135, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 136, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 137, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 138, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 139, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 140, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 141, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 142, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 143, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 144, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 145, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 146, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 147, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 148, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 149, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 150, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 151, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 152, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 153, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 154, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 155, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 156, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 157, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 158, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 159, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 160, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 161, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 162, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 163, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 164, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 165, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 166, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 167, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 168, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 169, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 170, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 171, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 172, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 173, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 174, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 175, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 176, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 177, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 178, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 179, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 180, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 181, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 182, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 183, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 184, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 185, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 186, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 187, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 188, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 189, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 190, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 191, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 192, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 193, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 194, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 195, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 196, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 197, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 198, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 199, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 200, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 201, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 202, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 203, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 204, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 205, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 206, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 207, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 208, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 209, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 210, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 211, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 212, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 213, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 214, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 215, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 216, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 217, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 218, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 219, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 220, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 221, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 222, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 223, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 224, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 225, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 226, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 227, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 228, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 229, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 230, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 231, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 232, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 233, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 234, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 235, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 236, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 237, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 238, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 239, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 240, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 241, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 242, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 243, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 244, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 245, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 246, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 247, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 248, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 249, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 250, __pyx_int_3)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 251, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 252, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 253, __pyx_int_0)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 254, __pyx_int_2)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 255, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 256, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 257, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 258, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 259, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 260, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 261, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 262, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 263, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 264, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 265, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 266, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 267, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 268, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 269, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 270, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 271, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 272, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 273, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 274, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 275, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 276, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 277, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 278, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 279, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 280, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 281, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 282, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 283, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 284, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 285, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 286, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 287, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 288, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 289, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 290, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 291, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 292, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 293, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 294, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 295, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 296, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 297, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 298, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 299, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 300, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 301, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 302, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 303, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 304, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 305, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 306, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 307, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 308, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 309, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 310, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 311, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 312, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 313, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 314, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 315, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 316, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 317, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 318, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 319, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 320, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 321, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 322, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 323, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 324, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 325, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 326, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 327, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 328, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 329, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 330, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 331, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 332, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 333, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 334, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 335, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 336, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 337, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 338, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 339, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 340, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 341, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 342, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 343, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 344, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 345, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 346, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 347, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 348, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 349, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 350, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 351, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 352, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 353, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 354, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 355, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 356, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 357, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 358, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 359, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 360, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 361, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 362, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 363, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 364, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 365, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 366, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 367, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 368, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 369, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 370, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 371, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 372, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 373, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 374, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 375, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 376, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 377, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 378, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 379, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 380, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 381, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 382, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 383, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 384, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 385, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 386, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 387, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 388, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 389, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 390, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 391, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 392, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 393, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 394, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 395, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 396, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 397, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 398, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 399, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 400, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 401, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 402, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 403, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 404, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 405, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 406, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 407, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 408, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 409, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 410, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 411, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 412, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 413, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 414, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 415, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 416, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 417, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 418, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 419, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 420, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 421, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 422, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 423, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 424, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 425, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 426, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 427, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 428, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 429, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 430, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 431, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 432, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 433, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 434, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 435, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 436, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 437, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 438, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 439, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 440, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 441, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 442, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 443, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 444, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 445, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 446, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 447, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 448, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 449, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 450, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 451, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 452, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 453, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 454, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 455, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 456, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 457, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 458, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 459, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 460, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 461, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 462, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 463, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 464, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 465, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 466, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 467, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 468, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 469, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 470, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 471, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 472, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 473, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 474, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 475, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 476, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 477, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 478, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 479, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 480, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 481, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 482, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 483, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 484, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 485, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 486, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 487, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 488, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 489, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 490, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 491, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 492, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 493, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 494, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 495, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 496, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 497, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 498, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 499, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 500, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 501, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 502, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 503, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 504, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 505, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 506, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 507, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 508, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 509, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 510, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 511, __pyx_int_1)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 6274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_n_u_B);
  __Pyx_GIVEREF(__pyx_n_u_B);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_u_B)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_5)) __PYX_ERR(0, 6274, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely((__Pyx_carray_from_py_uint8_t(__pyx_t_5, __pyx_t_17, 0x200) < 0))) __PYX_ERR(0, 6274, __pyx_L1_error)
  if (unlikely((0x200) != (0x200))) {
    PyErr_Format(PyExc_ValueError, "Assignment to slice of wrong length, expected %" CYTHON_FORMAT_SSIZE_T "d, got %" CYTHON_FORMAT_SSIZE_T "d", (Py_ssize_t)(0x200), (Py_ssize_t)(0x200));
    __PYX_ERR(0, 6274, __pyx_L1_error)
  }
  memcpy(&(__pyx_v_5pyboy_4core_7opcodes_OPCODE_LENGTHS[0]), __pyx_t_17, sizeof(__pyx_v_5pyboy_4core_7opcodes_OPCODE_LENGTHS[0]) * (0x200));
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyboy/core/opcodes.py":6310
 * 
 * 
 * CPU_COMMANDS = [             # <<<<<<<<<<<<<<
 *     "NOP",
 *     "LD BC,d16",
 */
  __pyx_t_5 = PyList_New(512); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_NOP);
  __Pyx_GIVEREF(__pyx_n_u_NOP);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_NOP)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_BC_d16);
  __Pyx_GIVEREF(__pyx_kp_u_LD_BC_d16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_kp_u_LD_BC_d16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_BC_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_BC_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u_LD_BC_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_BC);
  __Pyx_GIVEREF(__pyx_kp_u_INC_BC);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 3, __pyx_kp_u_INC_BC)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_B);
  __Pyx_GIVEREF(__pyx_kp_u_INC_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 4, __pyx_kp_u_INC_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_B);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 5, __pyx_kp_u_DEC_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_B_d8);
  __Pyx_GIVEREF(__pyx_kp_u_LD_B_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 6, __pyx_kp_u_LD_B_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RLCA);
  __Pyx_GIVEREF(__pyx_n_u_RLCA);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 7, __pyx_n_u_RLCA)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_a16_SP);
  __Pyx_GIVEREF(__pyx_kp_u_LD_a16_SP);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 8, __pyx_kp_u_LD_a16_SP)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_HL_BC);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_HL_BC);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 9, __pyx_kp_u_ADD_HL_BC)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_BC);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_BC);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 10, __pyx_kp_u_LD_A_BC)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_BC);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_BC);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 11, __pyx_kp_u_DEC_BC)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_C);
  __Pyx_GIVEREF(__pyx_kp_u_INC_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 12, __pyx_kp_u_INC_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_C);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 13, __pyx_kp_u_DEC_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_d8);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 14, __pyx_kp_u_LD_C_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RRCA);
  __Pyx_GIVEREF(__pyx_n_u_RRCA);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 15, __pyx_n_u_RRCA)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_STOP_0);
  __Pyx_GIVEREF(__pyx_kp_u_STOP_0);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 16, __pyx_kp_u_STOP_0)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_DE_d16);
  __Pyx_GIVEREF(__pyx_kp_u_LD_DE_d16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 17, __pyx_kp_u_LD_DE_d16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_DE_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_DE_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 18, __pyx_kp_u_LD_DE_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_DE);
  __Pyx_GIVEREF(__pyx_kp_u_INC_DE);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 19, __pyx_kp_u_INC_DE)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_D);
  __Pyx_GIVEREF(__pyx_kp_u_INC_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 20, __pyx_kp_u_INC_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_D);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 21, __pyx_kp_u_DEC_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_D_d8);
  __Pyx_GIVEREF(__pyx_kp_u_LD_D_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 22, __pyx_kp_u_LD_D_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RLA);
  __Pyx_GIVEREF(__pyx_n_u_RLA);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 23, __pyx_n_u_RLA)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JR_r8);
  __Pyx_GIVEREF(__pyx_kp_u_JR_r8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 24, __pyx_kp_u_JR_r8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_HL_DE);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_HL_DE);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 25, __pyx_kp_u_ADD_HL_DE)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_DE);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_DE);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 26, __pyx_kp_u_LD_A_DE)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_DE);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_DE);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 27, __pyx_kp_u_DEC_DE)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_E);
  __Pyx_GIVEREF(__pyx_kp_u_INC_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 28, __pyx_kp_u_INC_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_E);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 29, __pyx_kp_u_DEC_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_E_d8);
  __Pyx_GIVEREF(__pyx_kp_u_LD_E_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 30, __pyx_kp_u_LD_E_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RRA);
  __Pyx_GIVEREF(__pyx_n_u_RRA);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 31, __pyx_n_u_RRA)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JR_NZ_r8);
  __Pyx_GIVEREF(__pyx_kp_u_JR_NZ_r8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 32, __pyx_kp_u_JR_NZ_r8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_d16);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_d16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 33, __pyx_kp_u_LD_HL_d16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 34, __pyx_kp_u_LD_HL_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_HL);
  __Pyx_GIVEREF(__pyx_kp_u_INC_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 35, __pyx_kp_u_INC_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_H);
  __Pyx_GIVEREF(__pyx_kp_u_INC_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 36, __pyx_kp_u_INC_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_H);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 37, __pyx_kp_u_DEC_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_H_d8);
  __Pyx_GIVEREF(__pyx_kp_u_LD_H_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 38, __pyx_kp_u_LD_H_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DAA);
  __Pyx_GIVEREF(__pyx_n_u_DAA);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 39, __pyx_n_u_DAA)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JR_Z_r8);
  __Pyx_GIVEREF(__pyx_kp_u_JR_Z_r8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 40, __pyx_kp_u_JR_Z_r8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_HL_HL);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_HL_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 41, __pyx_kp_u_ADD_HL_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_HL);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 42, __pyx_kp_u_LD_A_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_HL);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 43, __pyx_kp_u_DEC_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_L);
  __Pyx_GIVEREF(__pyx_kp_u_INC_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 44, __pyx_kp_u_INC_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_L);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 45, __pyx_kp_u_DEC_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_L_d8);
  __Pyx_GIVEREF(__pyx_kp_u_LD_L_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 46, __pyx_kp_u_LD_L_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_CPL);
  __Pyx_GIVEREF(__pyx_n_u_CPL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 47, __pyx_n_u_CPL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JR_NC_r8);
  __Pyx_GIVEREF(__pyx_kp_u_JR_NC_r8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 48, __pyx_kp_u_JR_NC_r8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_SP_d16);
  __Pyx_GIVEREF(__pyx_kp_u_LD_SP_d16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 49, __pyx_kp_u_LD_SP_d16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_A_2);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_A_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 50, __pyx_kp_u_LD_HL_A_2)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_SP);
  __Pyx_GIVEREF(__pyx_kp_u_INC_SP);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 51, __pyx_kp_u_INC_SP)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_HL_2);
  __Pyx_GIVEREF(__pyx_kp_u_INC_HL_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 52, __pyx_kp_u_INC_HL_2)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_HL_2);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_HL_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 53, __pyx_kp_u_DEC_HL_2)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_d8);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 54, __pyx_kp_u_LD_HL_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_SCF);
  __Pyx_GIVEREF(__pyx_n_u_SCF);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 55, __pyx_n_u_SCF)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JR_C_r8);
  __Pyx_GIVEREF(__pyx_kp_u_JR_C_r8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 56, __pyx_kp_u_JR_C_r8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_HL_SP);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_HL_SP);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 57, __pyx_kp_u_ADD_HL_SP)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_HL_2);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_HL_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 58, __pyx_kp_u_LD_A_HL_2)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_SP);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_SP);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 59, __pyx_kp_u_DEC_SP)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_INC_A);
  __Pyx_GIVEREF(__pyx_kp_u_INC_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 60, __pyx_kp_u_INC_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_DEC_A);
  __Pyx_GIVEREF(__pyx_kp_u_DEC_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 61, __pyx_kp_u_DEC_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_d8);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 62, __pyx_kp_u_LD_A_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_CCF);
  __Pyx_GIVEREF(__pyx_n_u_CCF);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 63, __pyx_n_u_CCF)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_B_B);
  __Pyx_GIVEREF(__pyx_kp_u_LD_B_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 64, __pyx_kp_u_LD_B_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_B_C);
  __Pyx_GIVEREF(__pyx_kp_u_LD_B_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 65, __pyx_kp_u_LD_B_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_B_D);
  __Pyx_GIVEREF(__pyx_kp_u_LD_B_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 66, __pyx_kp_u_LD_B_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_B_E);
  __Pyx_GIVEREF(__pyx_kp_u_LD_B_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 67, __pyx_kp_u_LD_B_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_B_H);
  __Pyx_GIVEREF(__pyx_kp_u_LD_B_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 68, __pyx_kp_u_LD_B_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_B_L);
  __Pyx_GIVEREF(__pyx_kp_u_LD_B_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 69, __pyx_kp_u_LD_B_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_B_HL);
  __Pyx_GIVEREF(__pyx_kp_u_LD_B_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 70, __pyx_kp_u_LD_B_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_B_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_B_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 71, __pyx_kp_u_LD_B_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_B);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 72, __pyx_kp_u_LD_C_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_C);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 73, __pyx_kp_u_LD_C_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_D);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 74, __pyx_kp_u_LD_C_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_E);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 75, __pyx_kp_u_LD_C_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_H);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 76, __pyx_kp_u_LD_C_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_L);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 77, __pyx_kp_u_LD_C_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_HL);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 78, __pyx_kp_u_LD_C_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 79, __pyx_kp_u_LD_C_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_D_B);
  __Pyx_GIVEREF(__pyx_kp_u_LD_D_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 80, __pyx_kp_u_LD_D_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_D_C);
  __Pyx_GIVEREF(__pyx_kp_u_LD_D_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 81, __pyx_kp_u_LD_D_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_D_D);
  __Pyx_GIVEREF(__pyx_kp_u_LD_D_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 82, __pyx_kp_u_LD_D_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_D_E);
  __Pyx_GIVEREF(__pyx_kp_u_LD_D_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 83, __pyx_kp_u_LD_D_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_D_H);
  __Pyx_GIVEREF(__pyx_kp_u_LD_D_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 84, __pyx_kp_u_LD_D_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_D_L);
  __Pyx_GIVEREF(__pyx_kp_u_LD_D_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 85, __pyx_kp_u_LD_D_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_D_HL);
  __Pyx_GIVEREF(__pyx_kp_u_LD_D_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 86, __pyx_kp_u_LD_D_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_D_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_D_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 87, __pyx_kp_u_LD_D_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_E_B);
  __Pyx_GIVEREF(__pyx_kp_u_LD_E_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 88, __pyx_kp_u_LD_E_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_E_C);
  __Pyx_GIVEREF(__pyx_kp_u_LD_E_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 89, __pyx_kp_u_LD_E_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_E_D);
  __Pyx_GIVEREF(__pyx_kp_u_LD_E_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 90, __pyx_kp_u_LD_E_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_E_E);
  __Pyx_GIVEREF(__pyx_kp_u_LD_E_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 91, __pyx_kp_u_LD_E_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_E_H);
  __Pyx_GIVEREF(__pyx_kp_u_LD_E_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 92, __pyx_kp_u_LD_E_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_E_L);
  __Pyx_GIVEREF(__pyx_kp_u_LD_E_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 93, __pyx_kp_u_LD_E_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_E_HL);
  __Pyx_GIVEREF(__pyx_kp_u_LD_E_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 94, __pyx_kp_u_LD_E_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_E_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_E_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 95, __pyx_kp_u_LD_E_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_H_B);
  __Pyx_GIVEREF(__pyx_kp_u_LD_H_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 96, __pyx_kp_u_LD_H_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_H_C);
  __Pyx_GIVEREF(__pyx_kp_u_LD_H_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 97, __pyx_kp_u_LD_H_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_H_D);
  __Pyx_GIVEREF(__pyx_kp_u_LD_H_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 98, __pyx_kp_u_LD_H_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_H_E);
  __Pyx_GIVEREF(__pyx_kp_u_LD_H_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 99, __pyx_kp_u_LD_H_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_H_H);
  __Pyx_GIVEREF(__pyx_kp_u_LD_H_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 100, __pyx_kp_u_LD_H_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_H_L);
  __Pyx_GIVEREF(__pyx_kp_u_LD_H_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 101, __pyx_kp_u_LD_H_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_H_HL);
  __Pyx_GIVEREF(__pyx_kp_u_LD_H_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 102, __pyx_kp_u_LD_H_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_H_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_H_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 103, __pyx_kp_u_LD_H_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_L_B);
  __Pyx_GIVEREF(__pyx_kp_u_LD_L_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 104, __pyx_kp_u_LD_L_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_L_C);
  __Pyx_GIVEREF(__pyx_kp_u_LD_L_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 105, __pyx_kp_u_LD_L_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_L_D);
  __Pyx_GIVEREF(__pyx_kp_u_LD_L_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 106, __pyx_kp_u_LD_L_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_L_E);
  __Pyx_GIVEREF(__pyx_kp_u_LD_L_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 107, __pyx_kp_u_LD_L_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_L_H);
  __Pyx_GIVEREF(__pyx_kp_u_LD_L_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 108, __pyx_kp_u_LD_L_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_L_L);
  __Pyx_GIVEREF(__pyx_kp_u_LD_L_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 109, __pyx_kp_u_LD_L_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_L_HL);
  __Pyx_GIVEREF(__pyx_kp_u_LD_L_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 110, __pyx_kp_u_LD_L_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_L_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_L_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 111, __pyx_kp_u_LD_L_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_B);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 112, __pyx_kp_u_LD_HL_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_C);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 113, __pyx_kp_u_LD_HL_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_D);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 114, __pyx_kp_u_LD_HL_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_E);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 115, __pyx_kp_u_LD_HL_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_H);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 116, __pyx_kp_u_LD_HL_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_L);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 117, __pyx_kp_u_LD_HL_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_HALT);
  __Pyx_GIVEREF(__pyx_n_u_HALT);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 118, __pyx_n_u_HALT)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_A_3);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_A_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 119, __pyx_kp_u_LD_HL_A_3)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_B);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 120, __pyx_kp_u_LD_A_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_C);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 121, __pyx_kp_u_LD_A_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_D);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 122, __pyx_kp_u_LD_A_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_E);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 123, __pyx_kp_u_LD_A_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_H);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 124, __pyx_kp_u_LD_A_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_L);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 125, __pyx_kp_u_LD_A_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_HL_3);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_HL_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 126, __pyx_kp_u_LD_A_HL_3)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 127, __pyx_kp_u_LD_A_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_A_B);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_A_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 128, __pyx_kp_u_ADD_A_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_A_C);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_A_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 129, __pyx_kp_u_ADD_A_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_A_D);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_A_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 130, __pyx_kp_u_ADD_A_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_A_E);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_A_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 131, __pyx_kp_u_ADD_A_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_A_H);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_A_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 132, __pyx_kp_u_ADD_A_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_A_L);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_A_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 133, __pyx_kp_u_ADD_A_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_A_HL);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_A_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 134, __pyx_kp_u_ADD_A_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_A_A);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_A_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 135, __pyx_kp_u_ADD_A_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADC_A_B);
  __Pyx_GIVEREF(__pyx_kp_u_ADC_A_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 136, __pyx_kp_u_ADC_A_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADC_A_C);
  __Pyx_GIVEREF(__pyx_kp_u_ADC_A_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 137, __pyx_kp_u_ADC_A_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADC_A_D);
  __Pyx_GIVEREF(__pyx_kp_u_ADC_A_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 138, __pyx_kp_u_ADC_A_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADC_A_E);
  __Pyx_GIVEREF(__pyx_kp_u_ADC_A_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 139, __pyx_kp_u_ADC_A_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADC_A_H);
  __Pyx_GIVEREF(__pyx_kp_u_ADC_A_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 140, __pyx_kp_u_ADC_A_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADC_A_L);
  __Pyx_GIVEREF(__pyx_kp_u_ADC_A_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 141, __pyx_kp_u_ADC_A_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADC_A_HL);
  __Pyx_GIVEREF(__pyx_kp_u_ADC_A_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 142, __pyx_kp_u_ADC_A_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADC_A_A);
  __Pyx_GIVEREF(__pyx_kp_u_ADC_A_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 143, __pyx_kp_u_ADC_A_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SUB_B);
  __Pyx_GIVEREF(__pyx_kp_u_SUB_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 144, __pyx_kp_u_SUB_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SUB_C);
  __Pyx_GIVEREF(__pyx_kp_u_SUB_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 145, __pyx_kp_u_SUB_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SUB_D);
  __Pyx_GIVEREF(__pyx_kp_u_SUB_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 146, __pyx_kp_u_SUB_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SUB_E);
  __Pyx_GIVEREF(__pyx_kp_u_SUB_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 147, __pyx_kp_u_SUB_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SUB_H);
  __Pyx_GIVEREF(__pyx_kp_u_SUB_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 148, __pyx_kp_u_SUB_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SUB_L);
  __Pyx_GIVEREF(__pyx_kp_u_SUB_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 149, __pyx_kp_u_SUB_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SUB_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SUB_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 150, __pyx_kp_u_SUB_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SUB_A);
  __Pyx_GIVEREF(__pyx_kp_u_SUB_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 151, __pyx_kp_u_SUB_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SBC_A_B);
  __Pyx_GIVEREF(__pyx_kp_u_SBC_A_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 152, __pyx_kp_u_SBC_A_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SBC_A_C);
  __Pyx_GIVEREF(__pyx_kp_u_SBC_A_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 153, __pyx_kp_u_SBC_A_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SBC_A_D);
  __Pyx_GIVEREF(__pyx_kp_u_SBC_A_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 154, __pyx_kp_u_SBC_A_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SBC_A_E);
  __Pyx_GIVEREF(__pyx_kp_u_SBC_A_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 155, __pyx_kp_u_SBC_A_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SBC_A_H);
  __Pyx_GIVEREF(__pyx_kp_u_SBC_A_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 156, __pyx_kp_u_SBC_A_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SBC_A_L);
  __Pyx_GIVEREF(__pyx_kp_u_SBC_A_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 157, __pyx_kp_u_SBC_A_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SBC_A_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SBC_A_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 158, __pyx_kp_u_SBC_A_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SBC_A_A);
  __Pyx_GIVEREF(__pyx_kp_u_SBC_A_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 159, __pyx_kp_u_SBC_A_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_AND_B);
  __Pyx_GIVEREF(__pyx_kp_u_AND_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 160, __pyx_kp_u_AND_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_AND_C);
  __Pyx_GIVEREF(__pyx_kp_u_AND_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 161, __pyx_kp_u_AND_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_AND_D);
  __Pyx_GIVEREF(__pyx_kp_u_AND_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 162, __pyx_kp_u_AND_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_AND_E);
  __Pyx_GIVEREF(__pyx_kp_u_AND_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 163, __pyx_kp_u_AND_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_AND_H);
  __Pyx_GIVEREF(__pyx_kp_u_AND_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 164, __pyx_kp_u_AND_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_AND_L);
  __Pyx_GIVEREF(__pyx_kp_u_AND_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 165, __pyx_kp_u_AND_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_AND_HL);
  __Pyx_GIVEREF(__pyx_kp_u_AND_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 166, __pyx_kp_u_AND_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_AND_A);
  __Pyx_GIVEREF(__pyx_kp_u_AND_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 167, __pyx_kp_u_AND_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_XOR_B);
  __Pyx_GIVEREF(__pyx_kp_u_XOR_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 168, __pyx_kp_u_XOR_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_XOR_C);
  __Pyx_GIVEREF(__pyx_kp_u_XOR_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 169, __pyx_kp_u_XOR_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_XOR_D);
  __Pyx_GIVEREF(__pyx_kp_u_XOR_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 170, __pyx_kp_u_XOR_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_XOR_E);
  __Pyx_GIVEREF(__pyx_kp_u_XOR_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 171, __pyx_kp_u_XOR_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_XOR_H);
  __Pyx_GIVEREF(__pyx_kp_u_XOR_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 172, __pyx_kp_u_XOR_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_XOR_L);
  __Pyx_GIVEREF(__pyx_kp_u_XOR_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 173, __pyx_kp_u_XOR_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_XOR_HL);
  __Pyx_GIVEREF(__pyx_kp_u_XOR_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 174, __pyx_kp_u_XOR_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_XOR_A);
  __Pyx_GIVEREF(__pyx_kp_u_XOR_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 175, __pyx_kp_u_XOR_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_OR_B);
  __Pyx_GIVEREF(__pyx_kp_u_OR_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 176, __pyx_kp_u_OR_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_OR_C);
  __Pyx_GIVEREF(__pyx_kp_u_OR_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 177, __pyx_kp_u_OR_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_OR_D);
  __Pyx_GIVEREF(__pyx_kp_u_OR_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 178, __pyx_kp_u_OR_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_OR_E);
  __Pyx_GIVEREF(__pyx_kp_u_OR_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 179, __pyx_kp_u_OR_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_OR_H);
  __Pyx_GIVEREF(__pyx_kp_u_OR_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 180, __pyx_kp_u_OR_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_OR_L);
  __Pyx_GIVEREF(__pyx_kp_u_OR_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 181, __pyx_kp_u_OR_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_OR_HL);
  __Pyx_GIVEREF(__pyx_kp_u_OR_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 182, __pyx_kp_u_OR_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_OR_A);
  __Pyx_GIVEREF(__pyx_kp_u_OR_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 183, __pyx_kp_u_OR_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CP_B);
  __Pyx_GIVEREF(__pyx_kp_u_CP_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 184, __pyx_kp_u_CP_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CP_C);
  __Pyx_GIVEREF(__pyx_kp_u_CP_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 185, __pyx_kp_u_CP_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CP_D);
  __Pyx_GIVEREF(__pyx_kp_u_CP_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 186, __pyx_kp_u_CP_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CP_E);
  __Pyx_GIVEREF(__pyx_kp_u_CP_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 187, __pyx_kp_u_CP_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CP_H);
  __Pyx_GIVEREF(__pyx_kp_u_CP_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 188, __pyx_kp_u_CP_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CP_L);
  __Pyx_GIVEREF(__pyx_kp_u_CP_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 189, __pyx_kp_u_CP_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CP_HL);
  __Pyx_GIVEREF(__pyx_kp_u_CP_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 190, __pyx_kp_u_CP_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CP_A);
  __Pyx_GIVEREF(__pyx_kp_u_CP_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 191, __pyx_kp_u_CP_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RET_NZ);
  __Pyx_GIVEREF(__pyx_kp_u_RET_NZ);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 192, __pyx_kp_u_RET_NZ)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_POP_BC);
  __Pyx_GIVEREF(__pyx_kp_u_POP_BC);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 193, __pyx_kp_u_POP_BC)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JP_NZ_a16);
  __Pyx_GIVEREF(__pyx_kp_u_JP_NZ_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 194, __pyx_kp_u_JP_NZ_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JP_a16);
  __Pyx_GIVEREF(__pyx_kp_u_JP_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 195, __pyx_kp_u_JP_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CALL_NZ_a16);
  __Pyx_GIVEREF(__pyx_kp_u_CALL_NZ_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 196, __pyx_kp_u_CALL_NZ_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_PUSH_BC);
  __Pyx_GIVEREF(__pyx_kp_u_PUSH_BC);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 197, __pyx_kp_u_PUSH_BC)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_A_d8);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_A_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 198, __pyx_kp_u_ADD_A_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RST_00H);
  __Pyx_GIVEREF(__pyx_kp_u_RST_00H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 199, __pyx_kp_u_RST_00H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RET_Z);
  __Pyx_GIVEREF(__pyx_kp_u_RET_Z);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 200, __pyx_kp_u_RET_Z)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RET);
  __Pyx_GIVEREF(__pyx_n_u_RET);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 201, __pyx_n_u_RET)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JP_Z_a16);
  __Pyx_GIVEREF(__pyx_kp_u_JP_Z_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 202, __pyx_kp_u_JP_Z_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_PREFIX_CB);
  __Pyx_GIVEREF(__pyx_kp_u_PREFIX_CB);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 203, __pyx_kp_u_PREFIX_CB)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CALL_Z_a16);
  __Pyx_GIVEREF(__pyx_kp_u_CALL_Z_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 204, __pyx_kp_u_CALL_Z_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CALL_a16);
  __Pyx_GIVEREF(__pyx_kp_u_CALL_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 205, __pyx_kp_u_CALL_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADC_A_d8);
  __Pyx_GIVEREF(__pyx_kp_u_ADC_A_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 206, __pyx_kp_u_ADC_A_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RST_08H);
  __Pyx_GIVEREF(__pyx_kp_u_RST_08H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 207, __pyx_kp_u_RST_08H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RET_NC);
  __Pyx_GIVEREF(__pyx_kp_u_RET_NC);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 208, __pyx_kp_u_RET_NC)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_POP_DE);
  __Pyx_GIVEREF(__pyx_kp_u_POP_DE);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 209, __pyx_kp_u_POP_DE)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JP_NC_a16);
  __Pyx_GIVEREF(__pyx_kp_u_JP_NC_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 210, __pyx_kp_u_JP_NC_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 211, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CALL_NC_a16);
  __Pyx_GIVEREF(__pyx_kp_u_CALL_NC_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 212, __pyx_kp_u_CALL_NC_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_PUSH_DE);
  __Pyx_GIVEREF(__pyx_kp_u_PUSH_DE);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 213, __pyx_kp_u_PUSH_DE)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SUB_d8);
  __Pyx_GIVEREF(__pyx_kp_u_SUB_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 214, __pyx_kp_u_SUB_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RST_10H);
  __Pyx_GIVEREF(__pyx_kp_u_RST_10H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 215, __pyx_kp_u_RST_10H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RET_C);
  __Pyx_GIVEREF(__pyx_kp_u_RET_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 216, __pyx_kp_u_RET_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_RETI);
  __Pyx_GIVEREF(__pyx_n_u_RETI);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 217, __pyx_n_u_RETI)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JP_C_a16);
  __Pyx_GIVEREF(__pyx_kp_u_JP_C_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 218, __pyx_kp_u_JP_C_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 219, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CALL_C_a16);
  __Pyx_GIVEREF(__pyx_kp_u_CALL_C_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 220, __pyx_kp_u_CALL_C_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 221, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SBC_A_d8);
  __Pyx_GIVEREF(__pyx_kp_u_SBC_A_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 222, __pyx_kp_u_SBC_A_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RST_18H);
  __Pyx_GIVEREF(__pyx_kp_u_RST_18H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 223, __pyx_kp_u_RST_18H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LDH_a8_A);
  __Pyx_GIVEREF(__pyx_kp_u_LDH_a8_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 224, __pyx_kp_u_LDH_a8_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_POP_HL);
  __Pyx_GIVEREF(__pyx_kp_u_POP_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 225, __pyx_kp_u_POP_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_C_A_2);
  __Pyx_GIVEREF(__pyx_kp_u_LD_C_A_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 226, __pyx_kp_u_LD_C_A_2)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 227, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 228, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_PUSH_HL);
  __Pyx_GIVEREF(__pyx_kp_u_PUSH_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 229, __pyx_kp_u_PUSH_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_AND_d8);
  __Pyx_GIVEREF(__pyx_kp_u_AND_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 230, __pyx_kp_u_AND_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RST_20H);
  __Pyx_GIVEREF(__pyx_kp_u_RST_20H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 231, __pyx_kp_u_RST_20H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_ADD_SP_r8);
  __Pyx_GIVEREF(__pyx_kp_u_ADD_SP_r8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 232, __pyx_kp_u_ADD_SP_r8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_JP_HL);
  __Pyx_GIVEREF(__pyx_kp_u_JP_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 233, __pyx_kp_u_JP_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_a16_A);
  __Pyx_GIVEREF(__pyx_kp_u_LD_a16_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 234, __pyx_kp_u_LD_a16_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 235, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 236, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 237, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_XOR_d8);
  __Pyx_GIVEREF(__pyx_kp_u_XOR_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 238, __pyx_kp_u_XOR_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RST_28H);
  __Pyx_GIVEREF(__pyx_kp_u_RST_28H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 239, __pyx_kp_u_RST_28H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LDH_A_a8);
  __Pyx_GIVEREF(__pyx_kp_u_LDH_A_a8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 240, __pyx_kp_u_LDH_A_a8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_POP_AF);
  __Pyx_GIVEREF(__pyx_kp_u_POP_AF);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 241, __pyx_kp_u_POP_AF)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_C_2);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_C_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 242, __pyx_kp_u_LD_A_C_2)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_DI);
  __Pyx_GIVEREF(__pyx_n_u_DI);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 243, __pyx_n_u_DI)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 244, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_PUSH_AF);
  __Pyx_GIVEREF(__pyx_kp_u_PUSH_AF);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 245, __pyx_kp_u_PUSH_AF)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_OR_d8);
  __Pyx_GIVEREF(__pyx_kp_u_OR_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 246, __pyx_kp_u_OR_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RST_30H);
  __Pyx_GIVEREF(__pyx_kp_u_RST_30H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 247, __pyx_kp_u_RST_30H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_HL_SP_r8);
  __Pyx_GIVEREF(__pyx_kp_u_LD_HL_SP_r8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 248, __pyx_kp_u_LD_HL_SP_r8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_SP_HL);
  __Pyx_GIVEREF(__pyx_kp_u_LD_SP_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 249, __pyx_kp_u_LD_SP_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_LD_A_a16);
  __Pyx_GIVEREF(__pyx_kp_u_LD_A_a16);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 250, __pyx_kp_u_LD_A_a16)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_EI);
  __Pyx_GIVEREF(__pyx_n_u_EI);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 251, __pyx_n_u_EI)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 252, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__21);
  __Pyx_GIVEREF(__pyx_kp_u__21);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 253, __pyx_kp_u__21)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_CP_d8);
  __Pyx_GIVEREF(__pyx_kp_u_CP_d8);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 254, __pyx_kp_u_CP_d8)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RST_38H);
  __Pyx_GIVEREF(__pyx_kp_u_RST_38H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 255, __pyx_kp_u_RST_38H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RLC_B);
  __Pyx_GIVEREF(__pyx_kp_u_RLC_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 256, __pyx_kp_u_RLC_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RLC_C);
  __Pyx_GIVEREF(__pyx_kp_u_RLC_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 257, __pyx_kp_u_RLC_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RLC_D);
  __Pyx_GIVEREF(__pyx_kp_u_RLC_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 258, __pyx_kp_u_RLC_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RLC_E);
  __Pyx_GIVEREF(__pyx_kp_u_RLC_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 259, __pyx_kp_u_RLC_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RLC_H);
  __Pyx_GIVEREF(__pyx_kp_u_RLC_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 260, __pyx_kp_u_RLC_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RLC_L);
  __Pyx_GIVEREF(__pyx_kp_u_RLC_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 261, __pyx_kp_u_RLC_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RLC_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RLC_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 262, __pyx_kp_u_RLC_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RLC_A);
  __Pyx_GIVEREF(__pyx_kp_u_RLC_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 263, __pyx_kp_u_RLC_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RRC_B);
  __Pyx_GIVEREF(__pyx_kp_u_RRC_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 264, __pyx_kp_u_RRC_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RRC_C);
  __Pyx_GIVEREF(__pyx_kp_u_RRC_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 265, __pyx_kp_u_RRC_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RRC_D);
  __Pyx_GIVEREF(__pyx_kp_u_RRC_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 266, __pyx_kp_u_RRC_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RRC_E);
  __Pyx_GIVEREF(__pyx_kp_u_RRC_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 267, __pyx_kp_u_RRC_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RRC_H);
  __Pyx_GIVEREF(__pyx_kp_u_RRC_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 268, __pyx_kp_u_RRC_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RRC_L);
  __Pyx_GIVEREF(__pyx_kp_u_RRC_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 269, __pyx_kp_u_RRC_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RRC_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RRC_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 270, __pyx_kp_u_RRC_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RRC_A);
  __Pyx_GIVEREF(__pyx_kp_u_RRC_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 271, __pyx_kp_u_RRC_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RL_B);
  __Pyx_GIVEREF(__pyx_kp_u_RL_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 272, __pyx_kp_u_RL_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RL_C);
  __Pyx_GIVEREF(__pyx_kp_u_RL_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 273, __pyx_kp_u_RL_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RL_D);
  __Pyx_GIVEREF(__pyx_kp_u_RL_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 274, __pyx_kp_u_RL_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RL_E);
  __Pyx_GIVEREF(__pyx_kp_u_RL_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 275, __pyx_kp_u_RL_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RL_H);
  __Pyx_GIVEREF(__pyx_kp_u_RL_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 276, __pyx_kp_u_RL_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RL_L);
  __Pyx_GIVEREF(__pyx_kp_u_RL_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 277, __pyx_kp_u_RL_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RL_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RL_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 278, __pyx_kp_u_RL_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RL_A);
  __Pyx_GIVEREF(__pyx_kp_u_RL_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 279, __pyx_kp_u_RL_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RR_B);
  __Pyx_GIVEREF(__pyx_kp_u_RR_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 280, __pyx_kp_u_RR_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RR_C);
  __Pyx_GIVEREF(__pyx_kp_u_RR_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 281, __pyx_kp_u_RR_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RR_D);
  __Pyx_GIVEREF(__pyx_kp_u_RR_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 282, __pyx_kp_u_RR_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RR_E);
  __Pyx_GIVEREF(__pyx_kp_u_RR_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 283, __pyx_kp_u_RR_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RR_H);
  __Pyx_GIVEREF(__pyx_kp_u_RR_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 284, __pyx_kp_u_RR_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RR_L);
  __Pyx_GIVEREF(__pyx_kp_u_RR_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 285, __pyx_kp_u_RR_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RR_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RR_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 286, __pyx_kp_u_RR_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RR_A);
  __Pyx_GIVEREF(__pyx_kp_u_RR_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 287, __pyx_kp_u_RR_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SLA_B);
  __Pyx_GIVEREF(__pyx_kp_u_SLA_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 288, __pyx_kp_u_SLA_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SLA_C);
  __Pyx_GIVEREF(__pyx_kp_u_SLA_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 289, __pyx_kp_u_SLA_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SLA_D);
  __Pyx_GIVEREF(__pyx_kp_u_SLA_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 290, __pyx_kp_u_SLA_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SLA_E);
  __Pyx_GIVEREF(__pyx_kp_u_SLA_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 291, __pyx_kp_u_SLA_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SLA_H);
  __Pyx_GIVEREF(__pyx_kp_u_SLA_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 292, __pyx_kp_u_SLA_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SLA_L);
  __Pyx_GIVEREF(__pyx_kp_u_SLA_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 293, __pyx_kp_u_SLA_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SLA_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SLA_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 294, __pyx_kp_u_SLA_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SLA_A);
  __Pyx_GIVEREF(__pyx_kp_u_SLA_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 295, __pyx_kp_u_SLA_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRA_B);
  __Pyx_GIVEREF(__pyx_kp_u_SRA_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 296, __pyx_kp_u_SRA_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRA_C);
  __Pyx_GIVEREF(__pyx_kp_u_SRA_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 297, __pyx_kp_u_SRA_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRA_D);
  __Pyx_GIVEREF(__pyx_kp_u_SRA_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 298, __pyx_kp_u_SRA_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRA_E);
  __Pyx_GIVEREF(__pyx_kp_u_SRA_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 299, __pyx_kp_u_SRA_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRA_H);
  __Pyx_GIVEREF(__pyx_kp_u_SRA_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 300, __pyx_kp_u_SRA_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRA_L);
  __Pyx_GIVEREF(__pyx_kp_u_SRA_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 301, __pyx_kp_u_SRA_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRA_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SRA_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 302, __pyx_kp_u_SRA_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRA_A);
  __Pyx_GIVEREF(__pyx_kp_u_SRA_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 303, __pyx_kp_u_SRA_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SWAP_B);
  __Pyx_GIVEREF(__pyx_kp_u_SWAP_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 304, __pyx_kp_u_SWAP_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SWAP_C);
  __Pyx_GIVEREF(__pyx_kp_u_SWAP_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 305, __pyx_kp_u_SWAP_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SWAP_D);
  __Pyx_GIVEREF(__pyx_kp_u_SWAP_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 306, __pyx_kp_u_SWAP_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SWAP_E);
  __Pyx_GIVEREF(__pyx_kp_u_SWAP_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 307, __pyx_kp_u_SWAP_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SWAP_H);
  __Pyx_GIVEREF(__pyx_kp_u_SWAP_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 308, __pyx_kp_u_SWAP_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SWAP_L);
  __Pyx_GIVEREF(__pyx_kp_u_SWAP_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 309, __pyx_kp_u_SWAP_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SWAP_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SWAP_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 310, __pyx_kp_u_SWAP_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SWAP_A);
  __Pyx_GIVEREF(__pyx_kp_u_SWAP_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 311, __pyx_kp_u_SWAP_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRL_B);
  __Pyx_GIVEREF(__pyx_kp_u_SRL_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 312, __pyx_kp_u_SRL_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRL_C);
  __Pyx_GIVEREF(__pyx_kp_u_SRL_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 313, __pyx_kp_u_SRL_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRL_D);
  __Pyx_GIVEREF(__pyx_kp_u_SRL_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 314, __pyx_kp_u_SRL_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRL_E);
  __Pyx_GIVEREF(__pyx_kp_u_SRL_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 315, __pyx_kp_u_SRL_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRL_H);
  __Pyx_GIVEREF(__pyx_kp_u_SRL_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 316, __pyx_kp_u_SRL_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRL_L);
  __Pyx_GIVEREF(__pyx_kp_u_SRL_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 317, __pyx_kp_u_SRL_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRL_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SRL_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 318, __pyx_kp_u_SRL_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SRL_A);
  __Pyx_GIVEREF(__pyx_kp_u_SRL_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 319, __pyx_kp_u_SRL_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_0_B);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_0_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 320, __pyx_kp_u_BIT_0_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_0_C);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_0_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 321, __pyx_kp_u_BIT_0_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_0_D);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_0_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 322, __pyx_kp_u_BIT_0_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_0_E);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_0_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 323, __pyx_kp_u_BIT_0_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_0_H);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_0_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 324, __pyx_kp_u_BIT_0_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_0_L);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_0_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 325, __pyx_kp_u_BIT_0_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_0_HL);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_0_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 326, __pyx_kp_u_BIT_0_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_0_A);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_0_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 327, __pyx_kp_u_BIT_0_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_1_B);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_1_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 328, __pyx_kp_u_BIT_1_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_1_C);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_1_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 329, __pyx_kp_u_BIT_1_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_1_D);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_1_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 330, __pyx_kp_u_BIT_1_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_1_E);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_1_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 331, __pyx_kp_u_BIT_1_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_1_H);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_1_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 332, __pyx_kp_u_BIT_1_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_1_L);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_1_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 333, __pyx_kp_u_BIT_1_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_1_HL);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_1_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 334, __pyx_kp_u_BIT_1_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_1_A);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_1_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 335, __pyx_kp_u_BIT_1_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_2_B);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_2_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 336, __pyx_kp_u_BIT_2_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_2_C);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_2_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 337, __pyx_kp_u_BIT_2_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_2_D);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_2_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 338, __pyx_kp_u_BIT_2_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_2_E);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_2_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 339, __pyx_kp_u_BIT_2_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_2_H);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_2_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 340, __pyx_kp_u_BIT_2_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_2_L);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_2_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 341, __pyx_kp_u_BIT_2_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_2_HL);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_2_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 342, __pyx_kp_u_BIT_2_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_2_A);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_2_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 343, __pyx_kp_u_BIT_2_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_3_B);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_3_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 344, __pyx_kp_u_BIT_3_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_3_C);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_3_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 345, __pyx_kp_u_BIT_3_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_3_D);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_3_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 346, __pyx_kp_u_BIT_3_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_3_E);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_3_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 347, __pyx_kp_u_BIT_3_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_3_H);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_3_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 348, __pyx_kp_u_BIT_3_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_3_L);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_3_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 349, __pyx_kp_u_BIT_3_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_3_HL);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_3_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 350, __pyx_kp_u_BIT_3_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_3_A);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_3_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 351, __pyx_kp_u_BIT_3_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_4_B);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_4_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 352, __pyx_kp_u_BIT_4_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_4_C);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_4_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 353, __pyx_kp_u_BIT_4_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_4_D);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_4_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 354, __pyx_kp_u_BIT_4_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_4_E);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_4_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 355, __pyx_kp_u_BIT_4_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_4_H);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_4_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 356, __pyx_kp_u_BIT_4_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_4_L);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_4_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 357, __pyx_kp_u_BIT_4_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_4_HL);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_4_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 358, __pyx_kp_u_BIT_4_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_4_A);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_4_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 359, __pyx_kp_u_BIT_4_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_5_B);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_5_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 360, __pyx_kp_u_BIT_5_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_5_C);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_5_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 361, __pyx_kp_u_BIT_5_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_5_D);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_5_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 362, __pyx_kp_u_BIT_5_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_5_E);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_5_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 363, __pyx_kp_u_BIT_5_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_5_H);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_5_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 364, __pyx_kp_u_BIT_5_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_5_L);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_5_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 365, __pyx_kp_u_BIT_5_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_5_HL);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_5_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 366, __pyx_kp_u_BIT_5_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_5_A);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_5_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 367, __pyx_kp_u_BIT_5_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_6_B);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_6_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 368, __pyx_kp_u_BIT_6_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_6_C);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_6_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 369, __pyx_kp_u_BIT_6_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_6_D);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_6_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 370, __pyx_kp_u_BIT_6_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_6_E);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_6_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 371, __pyx_kp_u_BIT_6_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_6_H);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_6_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 372, __pyx_kp_u_BIT_6_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_6_L);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_6_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 373, __pyx_kp_u_BIT_6_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_6_HL);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_6_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 374, __pyx_kp_u_BIT_6_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_6_A);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_6_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 375, __pyx_kp_u_BIT_6_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_7_B);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_7_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 376, __pyx_kp_u_BIT_7_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_7_C);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_7_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 377, __pyx_kp_u_BIT_7_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_7_D);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_7_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 378, __pyx_kp_u_BIT_7_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_7_E);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_7_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 379, __pyx_kp_u_BIT_7_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_7_H);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_7_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 380, __pyx_kp_u_BIT_7_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_7_L);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_7_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 381, __pyx_kp_u_BIT_7_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_7_HL);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_7_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 382, __pyx_kp_u_BIT_7_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_BIT_7_A);
  __Pyx_GIVEREF(__pyx_kp_u_BIT_7_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 383, __pyx_kp_u_BIT_7_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_0_B);
  __Pyx_GIVEREF(__pyx_kp_u_RES_0_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 384, __pyx_kp_u_RES_0_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_0_C);
  __Pyx_GIVEREF(__pyx_kp_u_RES_0_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 385, __pyx_kp_u_RES_0_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_0_D);
  __Pyx_GIVEREF(__pyx_kp_u_RES_0_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 386, __pyx_kp_u_RES_0_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_0_E);
  __Pyx_GIVEREF(__pyx_kp_u_RES_0_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 387, __pyx_kp_u_RES_0_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_0_H);
  __Pyx_GIVEREF(__pyx_kp_u_RES_0_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 388, __pyx_kp_u_RES_0_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_0_L);
  __Pyx_GIVEREF(__pyx_kp_u_RES_0_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 389, __pyx_kp_u_RES_0_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_0_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RES_0_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 390, __pyx_kp_u_RES_0_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_0_A);
  __Pyx_GIVEREF(__pyx_kp_u_RES_0_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 391, __pyx_kp_u_RES_0_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_1_B);
  __Pyx_GIVEREF(__pyx_kp_u_RES_1_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 392, __pyx_kp_u_RES_1_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_1_C);
  __Pyx_GIVEREF(__pyx_kp_u_RES_1_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 393, __pyx_kp_u_RES_1_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_1_D);
  __Pyx_GIVEREF(__pyx_kp_u_RES_1_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 394, __pyx_kp_u_RES_1_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_1_E);
  __Pyx_GIVEREF(__pyx_kp_u_RES_1_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 395, __pyx_kp_u_RES_1_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_1_H);
  __Pyx_GIVEREF(__pyx_kp_u_RES_1_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 396, __pyx_kp_u_RES_1_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_1_L);
  __Pyx_GIVEREF(__pyx_kp_u_RES_1_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 397, __pyx_kp_u_RES_1_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_1_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RES_1_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 398, __pyx_kp_u_RES_1_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_1_A);
  __Pyx_GIVEREF(__pyx_kp_u_RES_1_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 399, __pyx_kp_u_RES_1_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_2_B);
  __Pyx_GIVEREF(__pyx_kp_u_RES_2_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 400, __pyx_kp_u_RES_2_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_2_C);
  __Pyx_GIVEREF(__pyx_kp_u_RES_2_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 401, __pyx_kp_u_RES_2_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_2_D);
  __Pyx_GIVEREF(__pyx_kp_u_RES_2_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 402, __pyx_kp_u_RES_2_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_2_E);
  __Pyx_GIVEREF(__pyx_kp_u_RES_2_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 403, __pyx_kp_u_RES_2_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_2_H);
  __Pyx_GIVEREF(__pyx_kp_u_RES_2_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 404, __pyx_kp_u_RES_2_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_2_L);
  __Pyx_GIVEREF(__pyx_kp_u_RES_2_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 405, __pyx_kp_u_RES_2_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_2_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RES_2_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 406, __pyx_kp_u_RES_2_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_2_A);
  __Pyx_GIVEREF(__pyx_kp_u_RES_2_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 407, __pyx_kp_u_RES_2_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_3_B);
  __Pyx_GIVEREF(__pyx_kp_u_RES_3_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 408, __pyx_kp_u_RES_3_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_3_C);
  __Pyx_GIVEREF(__pyx_kp_u_RES_3_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 409, __pyx_kp_u_RES_3_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_3_D);
  __Pyx_GIVEREF(__pyx_kp_u_RES_3_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 410, __pyx_kp_u_RES_3_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_3_E);
  __Pyx_GIVEREF(__pyx_kp_u_RES_3_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 411, __pyx_kp_u_RES_3_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_3_H);
  __Pyx_GIVEREF(__pyx_kp_u_RES_3_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 412, __pyx_kp_u_RES_3_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_3_L);
  __Pyx_GIVEREF(__pyx_kp_u_RES_3_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 413, __pyx_kp_u_RES_3_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_3_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RES_3_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 414, __pyx_kp_u_RES_3_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_3_A);
  __Pyx_GIVEREF(__pyx_kp_u_RES_3_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 415, __pyx_kp_u_RES_3_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_4_B);
  __Pyx_GIVEREF(__pyx_kp_u_RES_4_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 416, __pyx_kp_u_RES_4_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_4_C);
  __Pyx_GIVEREF(__pyx_kp_u_RES_4_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 417, __pyx_kp_u_RES_4_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_4_D);
  __Pyx_GIVEREF(__pyx_kp_u_RES_4_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 418, __pyx_kp_u_RES_4_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_4_E);
  __Pyx_GIVEREF(__pyx_kp_u_RES_4_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 419, __pyx_kp_u_RES_4_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_4_H);
  __Pyx_GIVEREF(__pyx_kp_u_RES_4_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 420, __pyx_kp_u_RES_4_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_4_L);
  __Pyx_GIVEREF(__pyx_kp_u_RES_4_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 421, __pyx_kp_u_RES_4_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_4_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RES_4_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 422, __pyx_kp_u_RES_4_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_4_A);
  __Pyx_GIVEREF(__pyx_kp_u_RES_4_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 423, __pyx_kp_u_RES_4_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_5_B);
  __Pyx_GIVEREF(__pyx_kp_u_RES_5_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 424, __pyx_kp_u_RES_5_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_5_C);
  __Pyx_GIVEREF(__pyx_kp_u_RES_5_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 425, __pyx_kp_u_RES_5_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_5_D);
  __Pyx_GIVEREF(__pyx_kp_u_RES_5_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 426, __pyx_kp_u_RES_5_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_5_E);
  __Pyx_GIVEREF(__pyx_kp_u_RES_5_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 427, __pyx_kp_u_RES_5_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_5_H);
  __Pyx_GIVEREF(__pyx_kp_u_RES_5_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 428, __pyx_kp_u_RES_5_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_5_L);
  __Pyx_GIVEREF(__pyx_kp_u_RES_5_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 429, __pyx_kp_u_RES_5_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_5_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RES_5_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 430, __pyx_kp_u_RES_5_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_5_A);
  __Pyx_GIVEREF(__pyx_kp_u_RES_5_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 431, __pyx_kp_u_RES_5_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_6_B);
  __Pyx_GIVEREF(__pyx_kp_u_RES_6_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 432, __pyx_kp_u_RES_6_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_6_C);
  __Pyx_GIVEREF(__pyx_kp_u_RES_6_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 433, __pyx_kp_u_RES_6_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_6_D);
  __Pyx_GIVEREF(__pyx_kp_u_RES_6_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 434, __pyx_kp_u_RES_6_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_6_E);
  __Pyx_GIVEREF(__pyx_kp_u_RES_6_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 435, __pyx_kp_u_RES_6_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_6_H);
  __Pyx_GIVEREF(__pyx_kp_u_RES_6_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 436, __pyx_kp_u_RES_6_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_6_L);
  __Pyx_GIVEREF(__pyx_kp_u_RES_6_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 437, __pyx_kp_u_RES_6_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_6_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RES_6_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 438, __pyx_kp_u_RES_6_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_6_A);
  __Pyx_GIVEREF(__pyx_kp_u_RES_6_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 439, __pyx_kp_u_RES_6_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_7_B);
  __Pyx_GIVEREF(__pyx_kp_u_RES_7_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 440, __pyx_kp_u_RES_7_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_7_C);
  __Pyx_GIVEREF(__pyx_kp_u_RES_7_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 441, __pyx_kp_u_RES_7_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_7_D);
  __Pyx_GIVEREF(__pyx_kp_u_RES_7_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 442, __pyx_kp_u_RES_7_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_7_E);
  __Pyx_GIVEREF(__pyx_kp_u_RES_7_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 443, __pyx_kp_u_RES_7_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_7_H);
  __Pyx_GIVEREF(__pyx_kp_u_RES_7_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 444, __pyx_kp_u_RES_7_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_7_L);
  __Pyx_GIVEREF(__pyx_kp_u_RES_7_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 445, __pyx_kp_u_RES_7_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_7_HL);
  __Pyx_GIVEREF(__pyx_kp_u_RES_7_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 446, __pyx_kp_u_RES_7_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_RES_7_A);
  __Pyx_GIVEREF(__pyx_kp_u_RES_7_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 447, __pyx_kp_u_RES_7_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_0_B);
  __Pyx_GIVEREF(__pyx_kp_u_SET_0_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 448, __pyx_kp_u_SET_0_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_0_C);
  __Pyx_GIVEREF(__pyx_kp_u_SET_0_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 449, __pyx_kp_u_SET_0_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_0_D);
  __Pyx_GIVEREF(__pyx_kp_u_SET_0_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 450, __pyx_kp_u_SET_0_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_0_E);
  __Pyx_GIVEREF(__pyx_kp_u_SET_0_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 451, __pyx_kp_u_SET_0_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_0_H);
  __Pyx_GIVEREF(__pyx_kp_u_SET_0_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 452, __pyx_kp_u_SET_0_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_0_L);
  __Pyx_GIVEREF(__pyx_kp_u_SET_0_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 453, __pyx_kp_u_SET_0_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_0_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SET_0_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 454, __pyx_kp_u_SET_0_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_0_A);
  __Pyx_GIVEREF(__pyx_kp_u_SET_0_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 455, __pyx_kp_u_SET_0_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_1_B);
  __Pyx_GIVEREF(__pyx_kp_u_SET_1_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 456, __pyx_kp_u_SET_1_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_1_C);
  __Pyx_GIVEREF(__pyx_kp_u_SET_1_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 457, __pyx_kp_u_SET_1_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_1_D);
  __Pyx_GIVEREF(__pyx_kp_u_SET_1_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 458, __pyx_kp_u_SET_1_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_1_E);
  __Pyx_GIVEREF(__pyx_kp_u_SET_1_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 459, __pyx_kp_u_SET_1_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_1_H);
  __Pyx_GIVEREF(__pyx_kp_u_SET_1_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 460, __pyx_kp_u_SET_1_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_1_L);
  __Pyx_GIVEREF(__pyx_kp_u_SET_1_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 461, __pyx_kp_u_SET_1_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_1_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SET_1_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 462, __pyx_kp_u_SET_1_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_1_A);
  __Pyx_GIVEREF(__pyx_kp_u_SET_1_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 463, __pyx_kp_u_SET_1_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_2_B);
  __Pyx_GIVEREF(__pyx_kp_u_SET_2_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 464, __pyx_kp_u_SET_2_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_2_C);
  __Pyx_GIVEREF(__pyx_kp_u_SET_2_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 465, __pyx_kp_u_SET_2_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_2_D);
  __Pyx_GIVEREF(__pyx_kp_u_SET_2_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 466, __pyx_kp_u_SET_2_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_2_E);
  __Pyx_GIVEREF(__pyx_kp_u_SET_2_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 467, __pyx_kp_u_SET_2_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_2_H);
  __Pyx_GIVEREF(__pyx_kp_u_SET_2_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 468, __pyx_kp_u_SET_2_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_2_L);
  __Pyx_GIVEREF(__pyx_kp_u_SET_2_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 469, __pyx_kp_u_SET_2_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_2_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SET_2_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 470, __pyx_kp_u_SET_2_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_2_A);
  __Pyx_GIVEREF(__pyx_kp_u_SET_2_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 471, __pyx_kp_u_SET_2_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_3_B);
  __Pyx_GIVEREF(__pyx_kp_u_SET_3_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 472, __pyx_kp_u_SET_3_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_3_C);
  __Pyx_GIVEREF(__pyx_kp_u_SET_3_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 473, __pyx_kp_u_SET_3_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_3_D);
  __Pyx_GIVEREF(__pyx_kp_u_SET_3_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 474, __pyx_kp_u_SET_3_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_3_E);
  __Pyx_GIVEREF(__pyx_kp_u_SET_3_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 475, __pyx_kp_u_SET_3_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_3_H);
  __Pyx_GIVEREF(__pyx_kp_u_SET_3_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 476, __pyx_kp_u_SET_3_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_3_L);
  __Pyx_GIVEREF(__pyx_kp_u_SET_3_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 477, __pyx_kp_u_SET_3_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_3_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SET_3_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 478, __pyx_kp_u_SET_3_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_3_A);
  __Pyx_GIVEREF(__pyx_kp_u_SET_3_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 479, __pyx_kp_u_SET_3_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_4_B);
  __Pyx_GIVEREF(__pyx_kp_u_SET_4_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 480, __pyx_kp_u_SET_4_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_4_C);
  __Pyx_GIVEREF(__pyx_kp_u_SET_4_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 481, __pyx_kp_u_SET_4_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_4_D);
  __Pyx_GIVEREF(__pyx_kp_u_SET_4_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 482, __pyx_kp_u_SET_4_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_4_E);
  __Pyx_GIVEREF(__pyx_kp_u_SET_4_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 483, __pyx_kp_u_SET_4_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_4_H);
  __Pyx_GIVEREF(__pyx_kp_u_SET_4_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 484, __pyx_kp_u_SET_4_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_4_L);
  __Pyx_GIVEREF(__pyx_kp_u_SET_4_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 485, __pyx_kp_u_SET_4_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_4_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SET_4_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 486, __pyx_kp_u_SET_4_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_4_A);
  __Pyx_GIVEREF(__pyx_kp_u_SET_4_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 487, __pyx_kp_u_SET_4_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_5_B);
  __Pyx_GIVEREF(__pyx_kp_u_SET_5_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 488, __pyx_kp_u_SET_5_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_5_C);
  __Pyx_GIVEREF(__pyx_kp_u_SET_5_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 489, __pyx_kp_u_SET_5_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_5_D);
  __Pyx_GIVEREF(__pyx_kp_u_SET_5_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 490, __pyx_kp_u_SET_5_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_5_E);
  __Pyx_GIVEREF(__pyx_kp_u_SET_5_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 491, __pyx_kp_u_SET_5_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_5_H);
  __Pyx_GIVEREF(__pyx_kp_u_SET_5_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 492, __pyx_kp_u_SET_5_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_5_L);
  __Pyx_GIVEREF(__pyx_kp_u_SET_5_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 493, __pyx_kp_u_SET_5_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_5_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SET_5_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 494, __pyx_kp_u_SET_5_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_5_A);
  __Pyx_GIVEREF(__pyx_kp_u_SET_5_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 495, __pyx_kp_u_SET_5_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_6_B);
  __Pyx_GIVEREF(__pyx_kp_u_SET_6_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 496, __pyx_kp_u_SET_6_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_6_C);
  __Pyx_GIVEREF(__pyx_kp_u_SET_6_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 497, __pyx_kp_u_SET_6_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_6_D);
  __Pyx_GIVEREF(__pyx_kp_u_SET_6_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 498, __pyx_kp_u_SET_6_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_6_E);
  __Pyx_GIVEREF(__pyx_kp_u_SET_6_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 499, __pyx_kp_u_SET_6_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_6_H);
  __Pyx_GIVEREF(__pyx_kp_u_SET_6_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 500, __pyx_kp_u_SET_6_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_6_L);
  __Pyx_GIVEREF(__pyx_kp_u_SET_6_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 501, __pyx_kp_u_SET_6_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_6_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SET_6_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 502, __pyx_kp_u_SET_6_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_6_A);
  __Pyx_GIVEREF(__pyx_kp_u_SET_6_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 503, __pyx_kp_u_SET_6_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_7_B);
  __Pyx_GIVEREF(__pyx_kp_u_SET_7_B);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 504, __pyx_kp_u_SET_7_B)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_7_C);
  __Pyx_GIVEREF(__pyx_kp_u_SET_7_C);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 505, __pyx_kp_u_SET_7_C)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_7_D);
  __Pyx_GIVEREF(__pyx_kp_u_SET_7_D);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 506, __pyx_kp_u_SET_7_D)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_7_E);
  __Pyx_GIVEREF(__pyx_kp_u_SET_7_E);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 507, __pyx_kp_u_SET_7_E)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_7_H);
  __Pyx_GIVEREF(__pyx_kp_u_SET_7_H);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 508, __pyx_kp_u_SET_7_H)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_7_L);
  __Pyx_GIVEREF(__pyx_kp_u_SET_7_L);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 509, __pyx_kp_u_SET_7_L)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_7_HL);
  __Pyx_GIVEREF(__pyx_kp_u_SET_7_HL);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 510, __pyx_kp_u_SET_7_HL)) __PYX_ERR(0, 6310, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_SET_7_A);
  __Pyx_GIVEREF(__pyx_kp_u_SET_7_A);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 511, __pyx_kp_u_SET_7_A)) __PYX_ERR(0, 6310, __pyx_L1_error);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CPU_COMMANDS, __pyx_t_5) < 0) __PYX_ERR(0, 6310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyboy/core/opcodes.py":2
 * 
 * # THIS FILE IS AUTO-GENERATED!!!             # <<<<<<<<<<<<<<
 * # DO NOT MODIFY THIS FILE.
 * # CHANGES TO THE CODE SHOULD BE MADE IN 'opcodes_gen.py'.
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_5) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyboy.core.opcodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyboy.core.opcodes");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* DivInt[Py_ssize_t] */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__3;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* PySequenceMultiply */
static PyObject* __Pyx_PySequence_Multiply_Generic(PyObject *seq, Py_ssize_t mul) {
    PyObject *result, *pymul = PyInt_FromSsize_t(mul);
    if (unlikely(!pymul))
        return NULL;
    result = PyNumber_Multiply(seq, pymul);
    Py_DECREF(pymul);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PySequence_Multiply(PyObject *seq, Py_ssize_t mul) {
#if CYTHON_USE_TYPE_SLOTS
    PyTypeObject *type = Py_TYPE(seq);
    if (likely(type->tp_as_sequence && type->tp_as_sequence->sq_repeat)) {
        return type->tp_as_sequence->sq_repeat(seq, mul);
    } else
#endif
    {
        return __Pyx_PySequence_Multiply_Generic(seq, mul);
    }
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
    if (is_list || !PyMapping_Check(o))
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* RaiseUnboundLocalError */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__2);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_0_8
#define __PYX_HAVE_RT_ImportType_3_0_8
static PyTypeObject *__Pyx_ImportType_3_0_8(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_8 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_8 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_8 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    __Pyx_TypeName obj_type_name;
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_7cpython_5array_array)) return __pyx_pw_7cpython_5array_5array_1__getbuffer__(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
                 "'" __Pyx_FMT_TYPENAME "' does not have the buffer interface",
                 obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_7cpython_5array_array)) __pyx_pw_7cpython_5array_5array_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


/* MemviewSliceIsContig */
static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* CIntFromPy */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint16_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint16_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint16_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint16_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(uint16_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint16_t) >= 2 * PyLong_SHIFT)) {
                                return (uint16_t) (((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint16_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint16_t) >= 3 * PyLong_SHIFT)) {
                                return (uint16_t) (((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint16_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint16_t) >= 4 * PyLong_SHIFT)) {
                                return (uint16_t) (((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint16_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(uint16_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint16_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(uint16_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint16_t) (((uint16_t)-1)*(((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(uint16_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint16_t) ((((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint16_t) (((uint16_t)-1)*(((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint16_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint16_t) ((((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint16_t) (((uint16_t)-1)*(((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint16_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint16_t) ((((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(uint16_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint16_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            uint16_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (uint16_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (uint16_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (uint16_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (uint16_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (uint16_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(uint16_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((uint16_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(uint16_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((uint16_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((uint16_t) 1) << (sizeof(uint16_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (uint16_t) -1;
        }
    } else {
        uint16_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint16_t) -1;
        val = __Pyx_PyInt_As_uint16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint16_t");
    return (uint16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint16_t");
    return (uint16_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint8_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT)) {
                                return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT)) {
                                return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT)) {
                                return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(uint8_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(uint8_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(uint8_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            uint8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (uint8_t) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (uint8_t) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (uint8_t) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (uint8_t) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (uint8_t) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(uint8_t) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((uint8_t) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(uint8_t) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((uint8_t) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((uint8_t) 1) << (sizeof(uint8_t) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (uint8_t) -1;
        }
    } else {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* MemviewSliceCopyTemplate */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                                return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                                return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                                return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(int) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (int) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (int) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (int) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (int) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (int) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((int) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((int) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* MemviewSliceInit */
static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#if PY_VERSION_HEX >= 0x030A0000 || defined(HAVE_STDARG_PROTOTYPES)
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int_type *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int_type *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    __pyx_nonatomic_int_type old_acquisition_count;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        return;
    }
    old_acquisition_count = __pyx_add_acquisition_count(memview);
    if (unlikely(old_acquisition_count <= 0)) {
        if (likely(old_acquisition_count == 0)) {
            if (have_gil) {
                Py_INCREF((PyObject *) memview);
            } else {
                PyGILState_STATE _gilstate = PyGILState_Ensure();
                Py_INCREF((PyObject *) memview);
                PyGILState_Release(_gilstate);
            }
        } else {
            __pyx_fatalerror("Acquisition count is %d (line %d)",
                             old_acquisition_count+1, lineno);
        }
    }
}
static CYTHON_INLINE void __Pyx_XCLEAR_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    __pyx_nonatomic_int_type old_acquisition_count;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    old_acquisition_count = __pyx_sub_acquisition_count(memview);
    memslice->data = NULL;
    if (likely(old_acquisition_count > 1)) {
        memslice->memview = NULL;
    } else if (likely(old_acquisition_count == 1)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         old_acquisition_count-1, lineno);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                                return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                                return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                                return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(long) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(long) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (long) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (long) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (long) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (long) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (long) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((long) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((long) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(char) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(char, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(char) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) >= 2 * PyLong_SHIFT)) {
                                return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(char) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) >= 3 * PyLong_SHIFT)) {
                                return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(char) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) >= 4 * PyLong_SHIFT)) {
                                return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(char) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(char) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(char, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(char) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                                return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(char) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                                return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                                return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(char) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                                return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 4 * PyLong_SHIFT)) {
                                return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(char) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(char) - 1 > 4 * PyLong_SHIFT)) {
                                return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(char) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(char) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (char) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (char) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (char) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (char) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (char) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(char) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((char) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(char) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((char) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((char) 1) << (sizeof(char) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__22);
    }
    return name;
}
#endif

/* CheckBinaryVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
    cobj = PyCapsule_New(p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr_3_0_8
#define __PYX_HAVE_RT_ImportVoidPtr_3_0_8
static int __Pyx_ImportVoidPtr_3_0_8(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
